namespace mc {

  // Forward declarations
  class Grid;
  class Cell;
  class Trans;
  class TransBB;
  class TransBF;
  class Atom;
  class IState;
  class BState;
  class BStateSelector;
  class TransBBSelector;
  class TransPList;
  class MarkovPList;
  class CTReaction;

  // MC channel - New!
  enum MCChannel { cNone, cHeating, cExcitation, cIonization, cFreeFree, cBoundFree, cCollisional, cChargeTransfer };

  // Quantity dependence
  enum Dependence { dpNone=0, dpRad=1, dpRho=2, dpT=4, dpN=8, dpNe=16, dpAll=31 };

  // General update flag
  const double update=-1;

  // No temperature flag
  const double noT=-1;

  // No recombination coefficient flag
  const double noAlfa=-1;

  // Transition type
  enum TransType { radiativeBB, radiativeBF, chargeTransferBF, collisional, internal }; // Changed!

  // Charge transfer reaction direction - New!
  enum CTDirection { ctForward, ctBackward };

  // Transition probability
  class TransP {
  public :
    // Probability
    //double P; // Changed back! - Testing!
    // Is this precision ok?
    float P;
    // Transition
    Trans* trans;
    // Type
    TransType type;
    // Upward
    bool upward;
    // Index - New! - Removed!
    unsigned short index;
  };

   // Transition probability list iterator
  typedef vector<TransP>::iterator TransPIterator;

  // Transition probability list
  class TransPList : public vector<TransP> {
  public :
    // Constructor - Add this, or?
    //TransPList();
    // Copy constructors - Add this, or?
    //TransPList(const TransPList& transPList);
    // Assignment operator - Add this, or?
    //TransPList& operator=(const TransPList& transPList);
    // Compare transition probabilities
    static bool compare(TransP transP1, TransP transP2);
    // Sort list in descending transition probability
    void sort();
    // Get memory allocated - New!
    double mem() const;
  };

  // Markov transition probability
  class MarkovP {
  public :
    // Probability
    //double P; // Changed back! - Testing!
    // Is this precision ok?
    float P;
    // Next bound state
    BState* nextState;
  };

   // Markov transition probability list iterator
  typedef vector<MarkovP>::iterator MarkovPIterator;

  // Markov transition probability list
  class MarkovPList : public vector<MarkovP> {
  public :
    // Constructor - Is this necessary?
    MarkovPList();
    // Copy constructor - Is this necessary?
    MarkovPList(const MarkovPList& markovPList);
    // Assignment operator - Is this necessary?
    MarkovPList& operator=(const MarkovPList& markovPList);
    // Compare transition probabilities
    static bool compare(MarkovP markovP1, MarkovP markovP2);
    // Sort list in descending transition probability
    void sort();
    // Get memory allocated - New!
    double mem() const;
  };
  
  // Bound state
  class BState { 
  private :
    // Bound state data pointer - Moved here - Testing!!!
    BStateData* data;
    // Ionization state back reference pointer
    IState* pIState;
    // Number fraction
    double pX;
    // Number density - New - Testing!
    double pn;
    // Total energy outflow (needed for macro-atom machinery)
    double pD;
  public :
    // *****************************************
    // *** New functions for boost *************
    // *****************************************
    // Radiative transition boost factor - Made public - Write access functions!
    double pTransPRBoost;
    // Collisional transition boost factor - Made public - Write access functions!
    double pTransPCBoost;
    // Transition boost factor - Made public - Write access functions!
    double pTransPBoost;
    // Markov transition boost factor - Made public - Write access functions!
    double pMarkovPBoost;
    // Bound state index - Made public when testing
    int index;
    // Transition probability list
    TransPList transPList;
    // Markov transition probability list
    MarkovPList markovPList;
    // Total radiative de-activation probability
    double PRTot;
    // Total collisional de-activation probability
    double PCTot;
    // Constructor
    BState();
    // Copy constructor
    BState(const BState& bState);
    // Assignment operator
    BState operator=(const BState& bState);
    // Destructor - Testing!
    virtual ~BState();
    // LTE reference state
    BState* LTERefState;
    // LTE group - New!
    vector<BState*> LTEGroup;
    // List of downward bound-bound transitions
    vector<TransBB*> downTransBB;
    // List of upward bound-bound transitions
    vector<TransBB*> upTransBB;
    // List of downward bound-free transitions
    vector<TransBF*> downTransBF;
    // List of upward bound-free transitions
    vector<TransBF*> upTransBF;
    // Get index - Testing.
    int getIndex() const;
    // Get ionization state back reference pointer
    IState* iState() const;
    // Get number fraction
    double X() const;
    // Set number fraction
    void X(const double& _X);
    // Get number density
    //double n() const;
    double n(); // Changed! - Testing!
    // Set number density
    void n(const double& _n);
    // Get energy above ground level
    double E() const;
    // Get energy above ground level of the neutral atom
    double Etot() const;
    // Get statistical weight
    double g() const;
    // Get cumulative A value
    double A() const;
    // Is metastable
    bool isMetastable() const;
    // Return LTE group master
    BState* LTEGroupMaster();
    // Bound state is part of LTE group - New!
    bool partOfLTEGroup() const;
    // Draw bound state from LTE group - New!
    BState* drawFromLTEGroup(bool boundFree=false, bool coll=false, bool nonThermal=false, bool twoPhoton=false, bool chargeTransfer=false);
    // Total energy outflow in upward transitions (needed for macro-atom machinery)
    double ED(bool boundFree=false, bool coll=false, bool twoPhoton=false, bool chargeTransfer=false) const; // Changed!
    // Total energy outflow in downward transitions (needed for macro-atom machinery)
    double EU(bool boundFree=false, bool coll=false, bool nonThermal=false, bool chargeTransfer=false) const; // Changed!
    // Total radiative energy outflow (needed for macro-atom machinery)
    double ER(bool boundFree=false, bool twoPhoton=false, bool chargeTransfer=false) const; // Changed!
    // Total collisonal energy outflow (needed for macro-atom machinery)
    double EC(bool boundFree=false) const;
    // Total charge-transfer energy outflow (needed for macro-atom machinery) - New!
    double ECT() const;
    // Total energy outflow (needed for macro-atom machinery)
    double D(bool boundFree=false, bool coll=false, bool nonThermal=false, bool twoPhoton=false, bool chargeTransfer=false, bool sumLTEGroup=false); // Changed!
    // Total bound-free opacity
    double boundFreeKappa(const double& nu, bool cutOff=false) const;
    // Total bound-free opacity
    double boundFreeKappa(FrequencyBinIterator nu, EdgeType edgeType=noEdge, bool cutOff=false) const;
    // Total bound-free opacity number derivative
    double boundFreedKappadN(const double& nu);
    // Total bound-free emissivity
    double boundFreeEta(const double& nu, bool cutOff=false) const;
    // Total bound-free emissivity
    double boundFreeEta(FrequencyBinIterator nu, EdgeType edgeType=noEdge, bool cutOff=false) const;
    // Total bound-free emissivity number derivative
    double boundFreedEtadN(const double& nu);
    // Total bound-free cooling rate
    double boundFreeCooling(double T=noT) const;
    // Total (spontaneous) bound-free cooling rate
    double boundFreeCoolingSP(double T=noT) const;
    // Total bound-free heating rate
    double boundFreeHeating(double T=noT) const;
    // Total collisional cooling rate
    double collCooling(double T=noT) const;
    // Total collisional heating rate
    double collHeating(double T=noT) const;
    // Total charge-transfer cooling rate - New!
    double CTCooling(double T=noT) const;
    // Total charge-transfer heating rate - New!
    double CTHeating(double T=noT) const;
    // De-excite (implements the macro-atom machinery)
    Trans* deExcite(TransType& transType, BState* root, bool& deActivate, bool boundFree, bool coll, bool nonThermal, bool twoPhoton, bool chargeTransfer); // Changed!
    // De-excite (implements the macro-atom machinery) - Recursive version 
    Trans* deExciteRec(TransType& transType, BState* root, bool& deActivate, bool boundFree=false, bool coll=false, bool nonThermal=false, bool twoPhoton=false, bool chargeTransfer=false); // Changed!
    // De-excite (implements the macro-atom machinery) - New version
    Trans* deExciteNew(TransType& transType, BState* root, bool& deActivate, bool boundFree=false, bool coll=false, bool nonThermal=false, bool twoPhoton=false, bool chargeTransfer=false); // Changed!
    // De-excite (implements the macro-atom machinery) - Table version
    Trans* deExciteTable(TransType& transType, BState* root, bool& deActivate, bool boundFree, bool coll, bool nonThermal=false, bool twoPhoton=false, bool chargeTransfer=false); // Changed!
    // De-excite (implements the macro-atom machinery) - Markov table version
    Trans* deExciteMarkovTable(TransType& transType, BState* root, bool& deActivate, bool boundFree, bool coll, bool nonThermal, bool twoPhoton, bool chargeTransfer); // Changed!
    // Update transition probability list
    void updateTransPList();
    // Get ionization cross-section
    double boundFreeSigmaRad(const double& nu) const;
    // Reset derived quantities (to be updated)
    void reset(int dependence=dpAll);
    // Export statistics
    void exportStat(YAML::Emitter& data, const BState& bStateLTE, const double& tauLim=0.0);
    // Load bound state
    void loadState(const YAML::Node& data);
    // Save bound state data
    void save(YAML::Emitter& data) const;
    // Save bound state
    void saveState(YAML::Emitter& data);
    // Set back references
    void setBackReference(IState* iState);
    // *****************************************
    // *** New functions for boost *************
    // *****************************************
    // Boost transition probabilities - New!
    void boostTransP();
    // Boost radiative transition probabilities - New!
    void boostTransPR();
    // Boost collisional transition probabilities - New!
    void boostTransPC();
    // Boost Markov transition probabilities - New!
    void boostMarkovP();
    // Get radiative transition probability boost factor - New!
    double transPRBoost();
    // Get collisional transition probability boost factor - New!
    double transPCBoost();
    // Get transition probability boost factor - New!
    double transPBoost();
    // Get Markov transition probability boost factor - New!
    double markovPBoost();
    // *****************************************
    // *** New functions for alfa correction ***
    // *****************************************
    // Testing!
    bool alfaSimple() const;
    // Recombination correction flag
    bool alfaCorrect() const;
    // Recombination fraction (if any)
    double alfaFrac() const;
    // Total spontaneus recombination coefficient - New!
    double alfaSPTotal(double T=noT);
    // Original spontaneous recombination coefficient - New!
    double alfaSPOriginal(double T=noT);
    // Spontaneous recombination coefficient - New!
    double alfaSP(double T=noT, bool includeOSP=true);
    // Additional (non-emissive) recombination coefficient - New!
    double alfaAdd();
    // Update recombination correction
    void updateAlfaCorr(double alfaSPIState, double alfaDiIState);
    // *****************************************
    // *****************************************
    // *****************************************
    // Get memory allocated for probability lists - New!
    double memPList() const;
    // Get memory allocated - New!
    double mem() const;
    // Give Grid class access to private data 
    friend class Grid;
    // Give Cell class access to private data
    friend class Cell;
    // Give Atom class access to private data
    friend class Atom;
    // Give IState class access to private data
    friend class IState;
    // Give TransBF class access to private data - Testing!
    friend class TransBF;
    // Give CTReaction class access to private data - Testing!
    friend class CTReaction;
  };

  // Transition type - New!
  enum TransClass { transClassBB, transClassBF };

  // Transition
  class Trans {
  public :
    // Constructor
    Trans() : upperState(NULL), lowerState(NULL) { }; // Changed!
    // Destructor
    virtual ~Trans() {};
    // Upper bound state pointer
    BState* upperState;
    // Lower bound state pointer
    BState* lowerState;
    // Get transition type - New!
    virtual TransClass type() const {};
    // Get transition energy
    virtual double Xi() const {};
    // Get transition frequency
    virtual double nu() const {};
    // Get downward radiative rate
    virtual double RD() {};
    // Get upward radiative rate
    virtual double RU() {};
    // Get downward collisional rate
    virtual double CD(double T=noT) {};
    // Get upward collisional rate
    virtual double CU(double T=noT) {};
    // Get total downward rate (per particle)
    // Including stimulated emission
    virtual double gammaD() {};
    // Get total upward rate (per particle)
    // Excluding stimulated emission
    virtual double gammaU() {};
    // Get emissivity boost factor - New!
    virtual double etaBoost() {};
    // Get opacity boost factor - New!
    virtual double kappaBoost() {};
    // Is LTE group internal - New!
    virtual bool isLTEGroupInternal() const;
    // Give Grid class access to private data
    friend class Grid;
    // Give IState class access to private data
    friend class IState;
  };
  
  // Bound-bound transition
  class TransBB : public Trans {
  private :
    // Transition data pointer
    TransBBData* data;
    // Line downward radiative rate - New!
    double pR1D;
    // Two-photon downward radiative rate - New!
    double pR2D;
    // Upward radiative rate
    double pRU;
    // Downward collisional rate
    double pCD;
    // Upward collisional rate
    double pCU;
    // Upward non-thermal collisional rate
    double pCNU;
    // Non-thermal flag - New! - Removed!
    //bool pNonThermal;
    // Processes flags - New!
    unsigned int pProcesses;
  public :
    // Line emissivity boost factor - New! - Made public - Write access functions!
    double pEta1Boost;
    // Two-photon emissivity boost factor - New! - Made public - Write access functions!
    double pEta2Boost;
    // Boosted two-photon emission probability - New! - Made public - Write access functions!
    vector<double> eta2P;
    // Constructor
    TransBB() : Trans() { pR1D=update; pR2D=update; pRU=update; pCD=update; pCU=update; pCNU=0; pProcesses=pNone; }; // Changed!
    // Destructor - Testing!
    virtual ~TransBB();
    // Get transition type
    virtual TransClass type() const;
    // Get transition energy
    virtual double Xi() const;
    // Get transition frequency
    virtual double nu() const;
    // Get downward radiative rate - Changed!
    virtual double RD(bool twoPhoton=false);
    // Get line downward radiative rate - New!
    virtual double R1D();
    // Get two-photon downward radiative rate - New!
    virtual double R2D();
    // Get upward radiative rate
    virtual double RU(bool preventNegative=true);
    // Get net (upward) radiative rate
    virtual double RNet(bool twoPhoton=false);
    // Get net (upward) radiative rate number derivative
    double dRNetdN(BState* bState, bool twoPhoton=false, bool _dBetadN=true);
    // Get downward collisional rate
    virtual double CD(double T=noT);
    // Get upward collisional rate
    virtual double CU(double T=noT);
    // Get net (upward) collisional rate
    virtual double CNet(double T=noT);
    // Get updward non-thermal collisional rate
    double CNU() const;
    // Set updward non-thermal collisional rate
    void CNU(const double& _CNU);
    // Get non-thermal flag - New! - Removed!
    //bool nonThermal() const;
    // Set non-thermal flag - New! - Removed!
    //void nonThermal(const bool& _nonThermal);
    // Get processes flags - New!
    void getProcesses();
    // Process is active - New!
    bool active(const ProcessType& processType);
    // Activate process - New!
    void activate(const ProcessType& processType);
    // De-activate process - New!
    void deActivate(const ProcessType& processType);
    // Get total downward rate (per particle)
    // Including stimulated emission - Changed!
    virtual double gammaD(bool collisional=false, bool twoPhoton=false);
    // Get total upward rate (per particle)
    // Excluding stimulated emission
    virtual double gammaU(bool collisional=false, bool nonThermal=false);
    // Get Einstein spontaneous de-excitation (emission) coefficient
    double A() const;
    // Get two-photon Einstein spontaneous de-excitation (emission) coefficient - New!
    double A2() const;
    // Get Einstein stimulated de-excitation (emission) coefficient
    double BD() const;
    // Get Einstein excitation (absorption) coefficient
    double BU() const;
    // Get collisional cross-section
    double sigmaColl(double T=noT) const;
    // Get optical depth
    double tau(bool preventNegative=true) const;
    // Get optical depth number derivative
    double dTaudN(BState* bState, bool preventNegative=true) const;
    // Get escape fraction
    double beta() const;
    // Get escape fraction number derivative
    double dBetadN(BState* bState) const;
    // Draw emission type - New!
    int etaType();
    // Get two-photon emissivity - New!
    double eta2(const double& _nu);
    // Get two-photon emissivity cumulative distribution function - New!
    double eta2F(double _nu);
    // Get two-photon emissivity cumulative distribution function (inverse version) - New!
    double eta2FInv(double F);
    // Draw two-pohoton emission frequency - New!
    double eta2Nu();
    // Draw two-photon emission frequency (intervall version) - New!
    double eta2Nu(double nuMin, double nuMax);
    // Draw two-photon emission frequency (normal version) - New!
    double eta2NuNormal();
    // Draw two-photon emission frequency (boost version) - New!
    double eta2NuBoost();
    // Reset derived quantities (to be updated)
    void reset(int dependence=dpAll);
    // Export data
    void exportStat(YAML::Emitter& data, const double& tauLim=0.0);
    // Save state
    void saveState(YAML::Emitter& data);
    // Load state
    void loadState(const YAML::Node& data);
    // Boost emissivity - New!
    void boostEta();
    // Boost line emissivity - New!
    void boostEta1();
    // Boost two-photon emissivity - New!
    void boostEta2();
    // Boost opacity - New!
    void boostKappa();
    // Get emissivity boost factor - New!
    virtual double etaBoost();
    // Get line emissivity boost factor - New!
    virtual double eta1Boost();
    // Get two-photon emissivity boost factor - New!
    virtual double eta2Boost();
    // Get opacity boost factor - New!
    virtual double kappaBoost();
    // Get memory allocated - New!
    double mem() const;
    // Give Grid class access to private data
    friend class Grid;
    // Give IState class access to private data
    friend class IState;
  };
  
  // Line list iterator
  typedef vector<TransBB*>::iterator LineIterator;
  
  // Line list
  class LineList : public vector<TransBB*> {
  public :
    // Constructors
    LineList();
    LineList(int size);
    // Find line with largest transition energy smaller than Xi
    // (Assuming the list is sorted in reverse Xi)
    LineIterator find(const double& Xi) const;
    // Sort line list in Xi
    void sort();
  };

  // Edge list iterator
  typedef vector<TransBF*>::iterator EdgeIterator;
  
  // Edge list
  class EdgeList : public vector<TransBF*> {
  public :
    // Constructors
    EdgeList();
    EdgeList(int size);
    // Find line with largest transition energy smaller than Xi
    // (Assuming the list is sorted in reverse Xi)
    EdgeIterator find(const double& Xi) const;
    // Sort line list in Xi
    void sort();
  };

  const double noSplit=-1;

  // Bound-free transition
  class TransBF : public Trans {
  private :
    // Transition data pointer
    TransBFData* data;
    // Downward radiative rate
    double pRD;
    // Upward radiative rate
    double pRU;
    // Downward radiative rate number derivative
    double pdRDdN;
    // Upward radiative rate number derivaive
    double pdRUdN;
    // Downward collisional rate
    double pCD;
    // Upward collisional rate
    double pCU;
    // Downward collisional rate (per particle and electron number density) - Testing !
    //double pcD;
    // Upward collisional rate (per particle) - Testing !
    //double pcU;
    // Upward non-thermal collisional rate
    double pCNU;
    // Downward charge-transfer rate - New!
    double pCTD;
    // Upward charge-transfer rate - New!
    double pCTU;
    // Non-thermal flag - New! - Removed!
    //bool pNonThermal;
    // Processes flags - New!
    unsigned int pProcesses;
    // (spontaneous) cooling
    double pCoolingSP;
    // Photo-ionization coefficient
    double pGammaOriginal;
    double pGammaModified;
    double pGammaOriginalOSP;
    double pGammaModifiedOSP;
    // (Corrected) photo-ionization coefficient
    double pGammaSTOriginal;
    double pGammaSTModified;
    double pGammaSTOriginalOSP;
    double pGammaSTModifiedOSP;
    // (Spontaneous) recombination rate coefficient
    double pAlfaSPOriginal;
    double pAlfaSPModified;
    double pAlfaSPOriginalOSP;
    double pAlfaSPModifiedOSP;
    // (Stimulated) recombination rate coefficient
    double pAlfaSTOriginal;
    double pAlfaSTModified;
    double pAlfaSTOriginalOSP;
    double pAlfaSTModifiedOSP;
    // On-the-spot (new) flag
    bool pOnTheSpotNew;
    // Net (upward) radiative rate in on-the-spot region
    double pROSP;
    // Net radiative heating in on-the-spot region
    double pGOSP;
    // Net (upward) radiative rate
    double pRNet;
    // LTE number ratio
    double pPhi;
    // Cut-off frequency
    double pNuCutOff;
    // Radiative cross-section at the edge frequency - Testing!!!
    double sigmaRad0;
    // *************************************
    // **** variables for alfa correcion ***
    // *************************************
    double pAlfaCorr;
    double pAlfaAdd;
    double pAlfaSPNoCorr;
    // *************************************
    // *************************************
    // *************************************
  public :
    // Downward charge-transfer reactions - New!
    vector<CTReaction*> CTDReaction;
    // Upward charge-transfer reactions - New!
    vector<CTReaction*> CTUReaction;
    // Boosted emission probability - New! - Made public - Write access functions!
    vector<double> etaP;
    // Emissivity boost factor - New! - Made public - Write access functions!
    double pEtaBoost;
    // Opacity boost factor - New! - Made public - Write access functions!
    double pKappaBoost;
    // Emission probability table - New! - Made public - Write access functions!
    vector<double> etaPTable;
    // Constructor
    TransBF() : Trans() { pRD=update; pRU=update; pdRDdN=update; pdRUdN=update; pCD=update; pCU=update; pCNU=0; pCTD=update; pCTU=update; pProcesses=pNone; pCoolingSP=update; pGammaOriginal=update; pGammaModified=update; pGammaOriginalOSP=update; pGammaModifiedOSP=update; pGammaSTOriginal=update; pGammaSTModified=update; pGammaSTOriginalOSP=update; pGammaSTModifiedOSP=update; pAlfaSPOriginal=update; pAlfaSPModified=update; pAlfaSPOriginalOSP=update; pAlfaSPModifiedOSP=update; pAlfaSTOriginal=update; pAlfaSTModified=update; pAlfaSTOriginalOSP=update; pAlfaSTModifiedOSP=update; pROSP=update; pGOSP=update; pRNet=update; pPhi=update; pNuCutOff=1e99; pAlfaCorr=update; pAlfaAdd=update, pAlfaSPNoCorr=update; sigmaRad0=update; }; // Changed!
    // Testing !
    //TransBF() { pRD=update; pRU=update; pdRDdN=update; pdRUdN=update; pCD=update; pCU=update; pcD=update; pcU=update; pCoolingSP=update; pGammaOriginal=update; pGammaModified=update; pGammaOriginalOSP=update; pGammaModifiedOSP=update; pGammaSTOriginal=update; pGammaSTModified=update; pGammaSTOriginalOSP=update; pGammaSTModifiedOSP=update; pAlfaSPOriginal=update; pAlfaSPModified=update; pAlfaSPOriginalOSP=update; pAlfaSPModifiedOSP=update; pAlfaSTOriginal=update; pAlfaSTModified=update; pAlfaSTOriginalOSP=update; pAlfaSTModifiedOSP=update; pROSP=update; pGOSP=update; pRNet=update; pPhi=update; pNuCutOff=1e99; };
    // Copy constructor - New!
    TransBF(const TransBF& trans);
    // Assigment operator - New!
    TransBF& operator=(const TransBF& trans);
    // Destructor - Testing!
    virtual ~TransBF();
    // Get transition type
    virtual TransClass type() const;
    // Get transition energy
    virtual double Xi() const;
    // Get transition frequency
    virtual double nu() const;
    // Get detailed balance method - New!
    DBMethod dBMethod() const;
    // Get downward radiative rate
    virtual double RD();
    // Get upward radiative rate
    virtual double RU();
    // Get downward radiative rate number derivative
    double dRDdN(BState* bState);
    // Get upward radiative rate number derivative
    double dRUdN(BState* bState);
    // Get net (upward) radiative rate number derivative - New!
    double dRNetdN(BState* bState, bool includeOSP=true, bool dNedN=true, bool expdN=true);
    // Get downward collisional rate
    virtual double CD(double T=noT);
    // Get upward collisional rate
    virtual double CU(double T=noT);
    // Get downward collisional rate (per particle and electron density) - Testing !
    //double cD(double T=noT);
    // Get upward collisional rate (per particle) - Testing !
    //double cU(double T=noT);
    // Get net (upward) collisional rate
    virtual double CNet(double T=noT);
    // Get net (upward) collisional rate number derivative - New!
    double dCNetdN(BState* bState, double T=noT, bool dNedN=true, bool expdN=true);
    // Get updward non-thermal collisional rate
    double CNU() const;
    // Set updward non-thermal collisional rate
    void CNU(const double& _CNU);
    // Get downward charge-transfer rate - New!
    double CTD();
    // Get upward charge-transfer rate - New!
    double CTU();
    // Get net (upward) charge-transfer rate - New!
    double CTNet();
    // Get net (upward) charge-transfer rate number derivative - New!
    double dCTNetdN(BState* bState);
    // Get non-thermal flag - New! - Removed!
    //bool nonThermal() const;
    // Set non-thermal flag - New! - Removed!
    //void nonThermal(const bool& _nonThermal);
    // Get processes flags - New!
    void getProcesses();
    // Process is active - New!
    bool active(const ProcessType& processType);
    // Activate process - New!
    void activate(const ProcessType& processType);
    // De-activate process - New!
    void deActivate(const ProcessType& processType);
    // Get total downward rate (per particle)
    // Including stimulated emission
    virtual double gammaD(bool collisional=false, bool chargeTransfer=false, bool includeOSP=false);
    // Get total upward rate (per particle)
    // Excluding stimulated emission
    virtual double gammaU(bool collisional=false, bool nonThermal=false, bool chargeTransfer=false, bool includeOSP=false);
    // Get total downward rate (per particle) number derivative
    double dGammaDdN(BState* bState);
    // Get total upward rate (per particle) number derivative
    double dGammaUdN(BState* bState);
    // Get radiative cross-section
    //double sigmaRad(const double& _nu, bool correct=true) const; // Changed!
    double sigmaRad(const double& _nu, bool correct=true); // Changed!
    // Get radiative cross-section
    //double sigmaRad(FrequencyBinIterator _nu, EdgeType edgeType=noEdge, bool correct=true) const; // Changed!
    double sigmaRad(FrequencyBinIterator _nu, EdgeType edgeType=noEdge, bool correct=true); // Changed!
    // Get radiative cross-section (corrected for stimulated emission)
    double sigmaRadST(const double& _nu);
    // Get radiative cross-section (corrected for stimulated emission)
    double sigmaRadST(FrequencyBinIterator _nu, EdgeType edgeType=noEdge);
    // Get collisional cross-section
    //double sigmaColl(double T) const;
    double sigmaColl(double T);
    // Get photo-ionization rate coefficient (per frequency and mean intensity)
    //double gammaNu(const double& _nu, bool modified=false) const;
    double gammaNu(const double& _nu, bool modified=false);
    // Get photo-ionization rate coefficient (per frequency and mean intensity)
    //double gammaNu(FrequencyBinIterator _nu, EdgeType edgeType=noEdge, bool modified=false) const;
    double gammaNu(FrequencyBinIterator _nu, EdgeType edgeType=noEdge, bool modified=false);
    // Get photo-ionization coefficient
    double gamma(bool modified=false, bool includeOSP=true);
    // Get photo-ionization coefficient number derivative
    double dGammadN(BState* bState, bool modified=false);
    // Get (corrected) photo-ionization rate coeffiecient (per frequency and mean intensity)
    double gammaSTNu(const double& _nu, bool modified=false);
    // Get (corrected) photo-ionization rate coeffiecient (per frequency and mean intensity);
    double gammaSTNu(FrequencyBinIterator _nu, EdgeType edgeType=noEdge, bool modified=false);
    // Get (corrected) photo-ionization coefficient
    double gammaST(bool modified=false, bool includeOSP=true);
    // Get (corrected) photo-ionization coefficient ratio
    double gammaSTRatio();
    // Get (spontaneous) recombination rate coefficient (per frequency)
    double alfaSPNu(const double& _nu, double T=noT, bool modified=false, bool correct=true); // Changed!
    // Get (spontaneous) recombination rate coefficient (per frequency) - Changed!
    double alfaSPNu(FrequencyBinIterator _nu, EdgeType edgeType=noEdge, double T=noT, bool modified=false, bool correct=true); // Changed!
    // Get (spontaneous) recombination rate coefficient
    double alfaSP(double T=noT, bool modified=false, bool includeOSP=true, bool correct=true); // Changed!
    // Get (spontaneous) hydrogenic recombination rate coefficient
    double alfaSPH(double T, double nuMin, double nuMax, bool modified=false, bool correct=true); // New!
    // Get (stimulated) recombination rate coefficient (per frequency and mean intensity)
    double alfaSTNu(const double& _nu, double T=noT, bool modified=false);
    // Get (stimulated) recombination rate coefficient (per frequency and mean intensity)
    double alfaSTNu(FrequencyBinIterator _nu, EdgeType edgeType=noEdge, double T=noT, bool modified=false);
    // Get (stimulated) recombination rate coefficient
    double alfaST(double T=noT, bool modified=false, bool includeOSP=true);
    // Get (stimulated) recombination rate coefficient number derivative
    double dAlfaSTdN(BState* bState, bool modified=false);
    // Get (total) recombination rate coefficient (per frequency)
    double alfaNu(const double& _nu, double T=noT, bool modified=false);
    // Get (total) recombination rate coefficient
    double alfa(double T=noT, bool modified=false, bool includeOSP=true, bool correct=true);
    // Get opacity
    double kappa(const double& _nu, bool cutOff=false);
    // Get opacity
    double kappa(FrequencyBinIterator _nu, EdgeType edgeType=noEdge, bool cutOff=false);
    // Get opacity number derivative
    double dKappadN(const double& _nu, BState* bState);
    // Get emissivity
    double eta(const double& _nu, bool cutOff=false);
    // Get emissivity
    double eta(FrequencyBinIterator _nu, EdgeType edgeType=noEdge, bool cutOff=false);
    // Get emissivity number derivative
    double dEtadN(const double& _nu, BState* bState);
    // Draw emission frequency
    double etaNu();
    // Draw emission frequency (intervall version) - New!
    double etaNu(double nuMin, double nuMax);
    // Draw emission frequency (normal version) - New!
    double etaNuNormal();
    // Draw emission frequency (boost version) - New!
    double etaNuBoost();
    // Draw emission frequency (table version) - New!
    double etaNuTable(double nuMin, double nuMax);
    // Update emission probality table - New!
    void updateEtaPTable();
    // Get cooling
    double cooling(double T=noT, bool includeOSP=false);
    // Get (spontaneous) cooling (per frequency)
    double coolingSPNu(const double& _nu, double T=noT);
    // Get (spontaneous) cooling (per frequency)
    double coolingSPNu(FrequencyBinIterator _nu, EdgeType edgeType=noEdge, double T=noT);
    // Get (spontaneous) cooling
    double coolingSP(double T=noT);
    // Get (spontaneous) hydrogenic cooling
    double coolingSPH(double T, double nuMin, double nuMax);
    // Get heating
    double heating(double T=noT, bool includeOSP=false);
    // Reset derived quantities (to be updated)
    void reset(int dependence=dpAll);
    // Export data
    void exportStat(YAML::Emitter& data);
    // Save state
    void saveState(YAML::Emitter& data);
    // Load state
    void loadState(const YAML::Node& data);
    // Get on-the-spot (new) flag
    bool onTheSpotNew() const;
    // Set on-the-spot (new) flag
    void onTheSpotNew(const bool& _onTheSpotNew);
    // Get net (upward) radiative rate
    double RNet();
    // Get net (upward) radiative rate in the on-the-spot region
    double ROSP();
    // Get (OSP) net radiative rate number derivative
    double dROSPdN(BState* bState);
    // Get net radiative heating in the on-the-spot region
    double GOSP();
    // Get (normalized) LTE number ratio (See Lucy 2003)
    double phi(double T=noT);
    // Set cut-off frequency
    void nuCutOff(const double& _nuCutOff);
    // Get cut-off frequency
    double nuCutOff() const;
    // Boost emissivity - New!
    void boostEta();
    // Boost opacity - New!
    void boostKappa();
    // Get emissivity boost factor - New!
    virtual double etaBoost();
    // Get opacity boost factor - New!
    virtual double kappaBoost();
    // Draw bound-free absorption MC channel - New!
    MCChannel absorptionChannel();
    // Draw emission type
    int etaType();
    // Draw downward charge-transfer reaction
    //CTReaction* drawCTDReaction();
    CTReaction* drawCTReaction(CTDirection& cTDirection);
    // *****************************************
    // **** New functions for alfa correcion ***
    // *****************************************
    // Testing!
    bool alfaSimple() const;
    // Get (spontaneos) emission correction factor - New!
    double alfaSPCorr(double _nu) const;
    // Get cross-section correction factor - New!
    double sigmaCorr(double _nu) const;
    // Get (emissive) recombination correction factor - New!
    double alfaCorr() const;
    // Get (non-emissive) recombination correction term - New!
    double alfaAdd();
    // Get downward radiative rate (non-emissive) correction - New!
    double RDAdd();
    // Update recombination correction - New!
    void updateAlfaCorr(double alfaSPBStateDist, double alfaSPIStateDist, double alfaDiIStateDist);
    // *****************************************
    // *****************************************
    // *****************************************
    // Get memory allocated - New!
    double mem() const;
    // Give Grid class access to private data
    friend class Grid;
    // Give IState class access to private data
    friend class IState;
    // Give Atom class access to private data
    friend class Atom;
  };

  // No charge flag
  const double noC=-1e99;

  // Forward declarations
  class IStateList;

  // Ionization state
  class IState {
  private :
    // Ionization state data pointer
    IStateData* data;
    // Ionization state index - New!
    int pIndex;
    // Atom back reference pointer
    Atom* pAtom;
    // Number fraction
    double pX;
    // Partition function
    double pU;
    // Total free-free heating
    double pFreeFreeHeating;
    // *****************************************
    // *** New functions for alfa correction ***
    // *****************************************
    // Alfa correction update flag
    bool alfaCorrUpdate;
    // Alfa correction update lock (openmp)
    omp_lock_t alfaCorrUpdateLock;
    // Additional (non-emmissive) recombination correction
    double pAlfaAdd;
    // Testing
    //int pUpdateCounter;
    // *****************************************
    // *** New functions for bound-free rates **
    // *****************************************
    // Upward collisional rate
    double pCU;
    // Downward collisional rate
    double pCD;
    // Upward non-thermal rate
    double pCNU;
    // Upward charge-transfer rate
    double pCTU;
    // Downward charge-transfer rate
    double pCTD;
    // Charge-transfer rate number derivate
    vector<vector<double> > pDCTdN;
    // Charge-transfer rate number derivative update flag
    bool dCTdNUpdate;
    // Charge-transfer rate number derivative update lock (openmp)
    omp_lock_t dCTdNUpdateLock;
    // Spontaneous recombination coeffiecient
    double pAlfaSP;
    double pAlfaSPOSP;
    // Stimulated recombination coefficient
    double pAlfaST;
    double pAlfaSTOSP;
    // Photo-ionization coefficient
    double pGamma;
    double pGammaOSP;
    // OSP radiative rate
    double pROSP;
    // OSP radiative rate number derivative
    vector<vector<double> > pDROSPdN;
    // OSP radiative rate number derivative update flag
    bool dROSPdNUpdate;
    // OSP radiative rate number derivative update lock (openmp)
    omp_lock_t dROSPdNUpdateLock;
    // *****************************************
    // *****************************************
    // *****************************************
  public :
    // Constructor
    IState();
    // Copy constructor
    IState(const IState& iState);
    // Assigment operator
    IState& operator=(const IState& iState);
    // Destructor
    virtual ~IState();
    // List of bound states
    vector<BState> bState;
    // Get index - New!
    int index() const;
    // Get atom back reference pointer
    Atom* atom() const;
    // Get number fraction
    double X() const;
    // Set number fraction
    void X(const double& _X);
    // Get number density
    double n() const;
    // Set number density
    void n(const double& _X);
    // Get atom charge
    double C() const;
    // Get ionization potential from ground state
    double Xi0() const;
    // Get partition function
    double U() const;
    // Set partition function
    void U(const double& _U);
    // Total free-free opacity
    double freeFreeKappa(const double& nu) const;
    // Total free-free emissivity
    double freeFreeEta(const double& nu, double T=noT) const;
    // Total free-free cooling
    double freeFreeCooling(double T=noT) const;
    // Free-free heating (per frequency and mean intensity)
    double freeFreeHeatingNu(const double& nu, double T=noT) const;
    // Total free-free heating
    //double freeFreeHeating(double T=noT) const;
    double freeFreeHeating(double T=noT);
    // Total bound-free opacity
    double boundFreeKappa(const double& nu, bool cutOff=false) const;
    // Total bound-free opacity
    double boundFreeKappa(FrequencyBinIterator nu, EdgeType edgeType=noEdge, bool cutOff=false) const;
    // Total bound-free opacity number derivative - New!
    double boundFreedKappadN(const double& nu);
    // Total bound-free emissivity
    double boundFreeEta(const double& nu, bool cutOff=false) const;
    // Total bound-free emissivity
    double boundFreeEta(FrequencyBinIterator nu, EdgeType edgeType=noEdge, bool cutOff=false) const;
    // Total bound-free opacity number derivative - New!
    double boundFreedEtadN(const double& nu);
    // Total bound-free cooling
    double boundFreeCooling(double T=noT) const;
    // Total (spontaneous) bound-free cooling
    double boundFreeCoolingSP(double T=noT) const;
    // Total bound-free heating
    double boundFreeHeating(double T=noT) const;
    // Total collisional cooling
    double collCooling(double T=noT) const;
    // Total collisional heating
    double collHeating(double T=noT) const;
    // Total charge-transfer cooling - New!
    double CTCooling(double T=noT) const;
    // Total charge-transfer heating - New!
    double CTHeating(double T=noT) const;
    // Get ionization state name
    string name() const;
    // Reset derived quantities (to be updated)
    void reset(int dependence=dpAll);
    // Export data
    void exportStat(YAML::Emitter& data, const IState& iStateLTE, const double& tauLim=0.0);
    // Load ionization state data
    void load(pstring& iStatePData, IStateDataList& iStateDataList, IState* iStateDown, TransBBSelector* transSelector, BStateSelector* bStateSelector, double C=noC);
    // Construct ionization state
    void construct(double _C, IStateDataList& iStateDataList, TransBBSelector* transSelector, BStateSelector* bStateSelector);
    // Reload bound-bound transitions
    void reloadTransBB(TransBBSelector* transSelector);
    // Load ionization state data
    void loadYAML(const YAML::Node& iStateData, IStateDataList& iStateDataList, TransBBSelector* transSelector, BStateSelector* bStateSelector);
    // Load ionization state
    void loadState(const YAML::Node& data, bool loadBState=true);
    // Save ionization state data
    void save(YAML::Emitter& data) const;
    // Save ionization state
    void saveState(YAML::Emitter& data);
    // Set back references
    void setBackReference(Atom* atom);
    // *****************************************
    // *** New functions for alfa correction ***
    // *****************************************
    // Get recombination distribution method - New!
    AlfaDistMethod alfaDistMethod() const;
    // Force recombination coefficient match - New!
    bool alfaForceEqual() const;
    // Total spontaneus recombination coefficient - New!
    double alfaSPTotal(double T=noT);
    // Di-electronic recombination coefficient - New!
    double alfaDi(double T=noT);
    // Original spontaneous recombination coefficient - New!
    double alfaSPOriginal(double T=noT);
    // Additional (non-emissice) recombination coefficient - New!
    double alfaAdd();
    // Testing
    //int updateCounter();
    // Update recombination correction - New!
    void updateAlfaCorr();
    // Get distributed total recombination coeffiecient - New!
    vector<double> alfaSPDist();
    // Get distributed di-electronic recombination coeffiecient - New!
    vector<double> alfaDiDist();
    // *****************************************
    // *** New functions for bound-free rates **
    // *****************************************
    // Upward collisional rate - New!
    double CU(double T=noT);
    // Downward collisional rate - New!
    double CD(double T=noT);
    // Upward non-thermal rate - New!
    double CNU();
    // Upward charge-transfer rate - New!
    double CTU();
    // Downward charge-transfer rate - New!
    double CTD();
    // Collisional rate (out of the state) - New!
    double CC(double T=noT);
    // Non-thermal rate (out of the state) - New!
    double CN();
    // Charge-transfer rate (out of the state) - New!
    double CT();
    // Collisional rate (out of the state) number derivative - New!
    double dCCdN(IState* iState, double T=noT, bool dNedN=true);
    // Non-thermal rate (out of the state) number derivative - New!
    double dCNdN(IState* iState);
    // Charge-transfer rate (out of the state) number derivative - New!
    double dCTdN(IState* iState);
    // Update charge-transfer rate (out of the state) number derivative - New!
    void updateDCTdN();
    // Get (spontaneous) recombination rate coefficient - New! - Testing!
    double alfaSPNu(const double& _nu, double T=noT, bool modified=false);
    // Get (spontaneous) recombination rate coefficient - New! - Testing!
    double alfaSPNu(FrequencyBinIterator _nu, EdgeType edgeType=noEdge, double T=noT, bool modified=false);
    // Spontaneous recombination coefficient - New!
    double alfaSP(double T=noT, bool includeOSP=true);
    // Get (stimulated) recombination rate coefficient - New! - Testing!
    double alfaSTNu(const double& _nu, double T=noT, bool modified=false);
    // Get (stimulated) recombination rate coefficient - New! - Testing!
    double alfaSTNu(FrequencyBinIterator _nu, EdgeType edgeType=noEdge, double T=noT, bool modified=false);
    // Get stimulated recombination coefficient - New!
    double alfaST(double T=noT, bool includeOSP=true);
    // Get recombination coefficient - New!
    double alfa(double T=noT, bool includeOSP=true);
    // Calculate photo-ionization rate coefficient - New! - Testing!
    double gammaNu(const double& _nu, bool modified=false);
    // Calculate photo-ionization rate coefficient - New! - Testing!
    double gammaNu(FrequencyBinIterator _nu, EdgeType edgeType=noEdge, bool modified=false);
    // Get photo-ionization coefficient - New!
    double gamma(bool includeOSP=true);
    // Get OSP radiative rate (out of the state) - New!
    double ROSP();
    // Get OSP radiative rate (out of the state) number derivative - New!
    double dROSPdN(IState* iState, bool dNedN=true);
    // Update OSP radiative rate (out of the state) number derivative - New!
    void updateDROSPdN(bool dNedN=true);
    // Get radiative rate (out of the state) - New!
    double R(bool includeOSP=true);
    // Get radiative rate (out of the state) number derivative - New!
    double dRdN(IState* iState, bool includeOSP=true, bool dNedN=true);
    // *****************************************
    // *****************************************
    // *****************************************
    // Get memory allocated for bound-bound transitions - New!
    double memTransBB() const;
    // Get memory allocated - New!
    double mem() const;
    // Get memory allocated for probaility lists - New!
    double memPList() const;
    // Give Grid class access to private data
    friend class Grid;
    // Give Cell class access to private data
    friend class Cell;
    // Give Atom class access to private data
    friend class Atom;
  };

  // Ionization state list
  class IStateList : public vector<IState> {
  public :
    // Find ionization state
    IState* find(const int& C);
  };

  // Charge transfer reaction - New!
  class CTReaction {
  private :
    // Reaction data pointer
    CTReactionData* data;
    // Reaction index
    int pIndex;
    // Cell back reference pointer
    Cell* pCell;
    // Find bound-free transition matching the CT transition data
    TransBF* findTrans(const CTTransData& transData, vector<Atom>& atomList);
  public :
    // Downward bound-free transition
    TransBF* upTrans;
    // Upward bound-free transition
    TransBF* downTrans;
    // Constructor
    CTReaction();
    // Destructor
    ~CTReaction();
    // Get index
    int index() const;
    // Get cell back reference pointer
    Cell* cell() const;
    // Reaction coefficient
    double alfa(double T=noT);
    // Number density type
    NStateType nType();
    // Reaction energy gain (in forward direction)
    double Xi();
    // Reaction rate
    double rate(double T=noT);
    // Reaction rate (forward direction)
    double forwardRate(double T=noT);
    // Reaction rate (backward direction)
    double backwardRate(double T=noT);
    // Cooling rate
    double cooling(double T=noT);
    // Cooling rate (forward direction)
    double forwardCooling(double T=noT);
    // Cooling rate (backward direction)
    double backwardCooling(double T=noT);
    // Heating rate
    double heating(double T=noT);
    // Heating rate (forward direction)
    double forwardHeating(double T=noT);
    // Heating rate (backward direction)
    double backwardHeating(double T=noT);
    // Heating channel probability
    double PHeating(CTDirection cTDirection);
    // Draw absorption channel
    MCChannel absorptionChannel(CTDirection cTDirection);
    // Construct charge-transfer reaction
    bool construct(CTReactionData* reactionData, vector<Atom>& atomList);
    // Set back references
    void setBackReference(Cell* cell);
    // Get memory allocated - New!
    double mem() const;
    // Give Cell class access to private data
    friend class Cell;
    // Give Grid class access to private data - Testing!
    friend class Grid;
  };

  // Isotope
  class Isotope {
  private :
    // Isotope data pointer
    IsotopeData* data;
    // Atom back reference pointer
    Atom* pAtom;
    // Mass fraction
    double pX;
  public :
    // Get atom back reference pointer
    Atom* atom() const;
    // Get density
    double rho() const;
    // Get number density
    double n() const;
    // Get mass fraction
    double X() const;
    // Set mass fraction
    void X(const double& _X);
    // Get number of electrons
    int N() const;
    // Atomic mass
    double m() const;
    // Load Isotope data
    void load(pstring& isotopePData, IsotopeDataList& isotopeDataList);
    // Load Isotope data
    void loadYAML(const YAML::Node& isotopeData, IsotopeDataList& isotopeDataList);
    // Load Isotope state
    void loadState(const YAML::Node& data);
    // Save Isotope data
    void save(YAML::Emitter& data) const;
    // Save Isotope state
    void saveState(YAML::Emitter& data);
    // Set back references
    void setBackReference(Atom* atom);
    // Get memory allocated - New!
    double mem() const;
    // Give cell class access to private data
    friend class Cell;
    // Give Grid class access to private data
    friend class Grid;
  };

  // Atom
  class Atom {
    // Atom data pointer
    AtomData* data;
    // Atom index - New!
    int pIndex;
    // Cell back reference pointer
    Cell* pCell;
    // Use on-the-spot approximation
    bool pOnTheSpot;
  public :
    // List of isotopes
    vector<Isotope> isotope;
    // List of ionisation states
    IStateList iState;
    // Constructor
    Atom();
    // Copy constructor
    Atom(const Atom& atom);
    // Assigment operator
    Atom& operator=(const Atom& atom);
    // Destructor
    virtual ~Atom();
    // Get index - New!
    int index() const;
    // Get cell back reference pointer
    Cell* cell() const;
    // Get mass fraction
    double X() const;
    // Get density
    double rho() const;
    // Get number density
    double n() const;
    // Get atomic mass
    double m() const;
    // Get number of protons
    double Z() const;
    // Total bound-free cooling - New!
    double boundFreeCooling(double T=noT) const;
    // Total bound-free heating - New!
    double boundFreeHeating(double T=noT) const;
    // Total collisional cooling
    double collCooling(double T=noT) const;
    // Total collisional heating - New!
    double collHeating(double T=noT) const;
    // Total charge-transfer cooling - New!
    double CTCooling(double T=noT) const;
    // Total charge-transfe heating - New!
    double CTHeating(double T=noT) const;
    // Update recombination correction - New!
    void updateAlfaCorr();
    // Calculate Markov S matrix
    vector<vector<double> > markovS(const vector<vector<int> >& matrixIndex);
    // Singular (SInv) matrix check - New!
    bool singular(Mat_DP& SInv, const vector<vector<int> >& matrixIndex, int& rowIndex, int& iStateIndex, int& bStateIndex);
    // Setup bound state index mapping
    vector<vector<int> > bMatrixIndex(int& nStates);
    // Count number of unique bound state indeces
    void countBMatrixIndex(int& nStates);
    // Update transition probability list
    void updateTransPList();
    // Update Markov transition probability list
    void updateMarkovPList();
    // Update Markov transition probability list
    void updateMarkovPList(const vector<vector<double> >& S, const vector<vector<int> >& matrixIndex);
    // Get atom name
    string name() const;
    // Get on-the-spot flag
    bool onTheSpot() const;
    // Set on-the-spot flag
    void onTheSpot(const bool& _onTheSpot);
    // Reset derived quantities (to be updated)
    void reset(int dependence=dpAll);
    // Export data
    void exportStat(YAML::Emitter& data, const Atom& atomLTE, const double& tauLim=0.0);
    // Load Atom data
    void load(pstring& atomPData, AtomDataList& atomDataList, TransBBSelector* transSelector, BStateSelector* bStateSelector);
    // Reload bound-bound transitions
    void reloadTransBB(TransBBSelector* transSelector);
    // Load Atom data
    void loadYAML(const YAML::Node& atomData, AtomDataList& atomDataList, TransBBSelector* transSelector, BStateSelector* bStateSelector);
    // Load Atom state
    void loadState(const YAML::Node& data, bool loadIState=true, bool loadBState=true);
    // Save Atom data
    void save(YAML::Emitter& data) const;
    // Save Atom state
    void saveState(YAML::Emitter& data);
    // Set back references
    void setBackReference(Cell* cell);
    // Get memory allocated for bound-free transitions - New!
    double memTransBF() const;
    // Get memory allocated - New!
    double mem() const;
    // Get memory allocated for probability lists - New!
    double memPList() const;
    // Get memory allocated for bound-bound transitions - New!
    double memTransBB() const;
    // Give Grid class access to private data
    friend class Grid;
    // Give Cell class acess to private data
    friend class Cell;
    // Give Isotope class access to private data
    friend class Isotope;
  };
  
}
