#include <cstdlib>
#include <iostream>
#include <iomanip>
#include <fstream>
#include <sstream>
#include <string>
#include <math.h>
#include <algorithm>
#include <limits>
#include "yaml-cpp/yaml.h"
#include "mc-util.h"
#include "mc-math.h"
#include "mc-physics.h"
#include "atomdata.h"

using namespace std;
using namespace util;
using namespace math;
using namespace physics;
using namespace atomdata;

// Get data file
string atomdata::getDataFile(const string& dataDir, const string& dataFile) {
  string newDataFile=dataDir;
  newDataFile.append("/");
  newDataFile.append(dataFile);
  return newDataFile;
}

// Get data directory
string atomdata::getDataDir(const string& dataFile) {
  int pos=dataFile.rfind('/');
  if (pos<dataFile.size()) return dataFile.substr(0,pos);
  else return "";
}

// Get (YAML) type name
string atomdata::typeName(const YAML::Node& data) {
  switch(data.Type()) {
  case YAML::NodeType::Null: return "Null";
  case YAML::NodeType::Scalar: return "Scalar";
  case YAML::NodeType::Sequence: return "Sequence";
  case YAML::NodeType::Map: return "Map";
  case YAML::NodeType::Undefined: return "Undefined";
  default: return "Invalid";
  }
}

// Resample bound-free cross-section table
Interpolator1D atomdata::resampleBFTable(const vector<double>& nu, const double& nuEdge, Interpolator1D& sigma) {
  vector<double> nuNew;
  vector<FunctionValue> tableWork;
  vector<FunctionValue> tableNew;
  Interpolator1D sigmaWork,sigmaNew;
  double x1,x2,fInt;
  int i,j,jj,jjj;

  // Select new frequency that exceeds the minumum frequency of the original table
  for (j=0;j<nu.size();j++) { if (nu[j]>=sigma.table()[0].x) break; }

  // For each original table entry
  for (i=0;i<sigma.table().size();i++) {
    // If within re-sampling range
    if (sigma.table()[i].x>nu[0] and sigma.table()[i].x<nu[nu.size()-1]) {
      // If original table frequency exceeds current frequency
      if (sigma.table()[i].x>=nu[j]) { 
	// Add current frequency to new frequenct list
	nuNew.push_back(nu[j]);
	// Select new frequency that exceeds the original table frequency
	for (j=j+1;j<nu.size();j++) { if (nu[j]>=sigma.table()[i].x) break; }
      }
    }
    // Else - Add original table frequency to new frequency list
    else nuNew.push_back(sigma.table()[i].x);
  }

  // Add edge frequency and sort new frequency list
  nuNew.push_back(nuEdge); sort(nuNew.begin(),nuNew.end());

  // Make sure the work table covers the edge frequency
  if (nuEdge<sigma.table()[0].x) { tableWork.resize(1); tableWork[0].x=nuEdge; tableWork[0].f=sigma.table()[0].f; }

  // For each new frequency
  j=0; // New!
  for (i=0;i<nuNew.size()-1;i++) { // Changed!
    // For each original table entry
    fInt=0; // New!
    for (jj=0;jj+j+1<sigma.table().size();jj++) { // Changed!
      // Integrate original table
      fInt+=0.5*(sigma.table()[j+jj+1].f+sigma.table()[jj+j].f)*(sigma.table()[j+jj+1].x-sigma.table()[j+jj].x);
      // If original entry frequency exceeds new frequency - Break loop
      if (sigma.table()[jj+j+1].x>=nuNew[i+1]) break;
    }
    // If more than one original table entry traversed
    if (jj>0){
      // Add work table entry using mean frequency and cross-section
      tableWork.resize(tableWork.size()+1);
      tableWork[tableWork.size()-1].x=0.5*(sigma.table()[jj+j+1].x+sigma.table()[j].x);
      tableWork[tableWork.size()-1].f=fInt/(sigma.table()[jj+j+1].x-sigma.table()[j].x);
      // Select next table entry
      j+=(jj+1);
    }
    // Else
    else {
      // Add copy of original table entries to work table
      for (jjj=0;jjj<=jj;jjj++) tableWork.push_back(sigma.table()[j+jjj]);
      // Select next table entry
      j+=(jj+1);
    }
  }
  // Add any remaining entries in original table to work table
  for (j=j;j<sigma.table().size();j++) tableWork.push_back(sigma.table()[j]);

  // Once again - Make sure the work table covers the edge frequency - Hack!
  if (nuEdge<tableWork[0].x) { tableWork.insert(tableWork.begin(),tableWork[0]); tableWork[0].x=nuEdge; }

  // Initiate work interpolator
  sigmaWork.init(tableWork);

  // Initiate new table
  tableNew.resize(nuNew.size());

  // For each new frequency
  for (i=0;i<nuNew.size();i++) {
    // Set new table frequency
    tableNew[i].x=nuNew[i];
    // Interpolate new table cross-section using work interpolator
    tableNew[i].f=sigmaWork(nuNew[i]);
  }

  // Initiate new cross-section interpolator with new table
  sigmaNew.init(tableNew);

  return sigmaNew;
}

// Get bound-bound collisional strength (value)
//double BBOmegaValue::operator ()(double T) const {
double BBOmegaValue::operator ()(double T) {
  return omega;
}

// Load (YAML) data
void BBOmegaValue::load(const YAML::Node& data) {

  // Get collisional strength
  omega=data.as<double>();
}

// Log data
void BBOmegaValue::log() {
  int i;

  // Log 
  cout<<"BBOmegaValue::log: omega="<<omega<<"\n";
}

// Create copy
BBOmega* BBOmegaValue::copy() {
  return new BBOmegaValue(*this);
}

// Get memory allocated - New!
double BBOmegaValue::mem() const {
  double _mem=sizeof(*this);

  return _mem;
}

// Get bound bound collisional strength (table)
//double BBOmegaTable::operator ()(double T) const {
double BBOmegaTable::operator ()(double T) {
  return omega(T);
}

// Load (YAML) data
void BBOmegaTable::load(const YAML::Node& data) {
  vector<FunctionValue> table(data.size());
  int i;

  // Get table data
  for (i=0;i<data.size();i++) {
    if (data[i][0]) table[i].x=data[i][0].as<double>();
    else { cerr<<"BBOmegaTable::load, Error: Could not get data\n"; exit(-1); }
    if (data[i][1]) table[i].f=data[i][1].as<double>();
    else { cerr<<"BBOmegaTable::load, Error: Could not get data\n"; exit(-1); }
  }

  // Initiate interpolator
  omega.init(table);
}

// Log data
void BBOmegaTable::log() {
  int i;

  // Log 
  cout<<"BBOmegaTable::log: Table loaded\n";
}

// Create copy
BBOmega* BBOmegaTable::copy() {
  return new BBOmegaTable(*this);
}

// Get memory allocated - New!
double BBOmegaTable::mem() const {
  double _mem=sizeof(*this);

  // Add memory allocated for table
  _mem+=omega.mem();

  return _mem;
}

// Get bound-bound collisional strength (powerlaw)
//double BBOmegaPLaw::operator ()(double T) const {
double BBOmegaPLaw::operator ()(double T) {
  return omega*pow(T/T0,p);
}

// Load (YAML) data
void BBOmegaPLaw::load(const YAML::Node& data) {
  
  // Get collisional strength
  if (data[0]) omega=data[0].as<double>();
  else { cerr<<"BBOmegaPLaw::load, Error: Could not get data\n"; exit(-1); }  

  // Get temperature
  if (data[1]) T0=data[1].as<double>();
  else { cerr<<"BBOmegaPLaw::load, Error: Could not get data\n"; exit(-1); }  

  // Get exponent
  if (data[2]) p=data[2].as<double>();
  else { cerr<<"BBOmegaPLaw::load, Error: Could not get data\n"; exit(-1); }  
}

// Log data
void BBOmegaPLaw::log() {
  int i;

  // Log 
  cout<<"BBOmegaPLaw::log: sigma="<<sigma<<", T0="<<T0<<", p="<<p<<"\n";
}

// Create copy
BBOmega* BBOmegaPLaw::copy() {
  return new BBOmegaPLaw(*this);
}

// Get memory allocated - New!
double BBOmegaPLaw::mem() const {
  double _mem=sizeof(*this);

  return _mem;
}

// Create emissivity cumulative distribution function table
void BBEta2::createFTable(const vector<double>& nu, const double& nu0) {
  vector<FunctionValue> table(0),tableInv(0);
  FunctionValue tableEntry;
  double nuMin,nuMax,F=0;
  int i;

  // Add first table entry
  tableEntry.x=nu[0];
  tableEntry.f=F;
  table.push_back(tableEntry);
        
  // For each frequency bin
  for (i=0;i<nu.size()-1;i++) {
    if (nu[i]<nu0) {
      // Get frequency bin limits
      nuMin=nu[i];
      nuMax=MIN(nu[i+1],nu0);
      // Sum cumulative probability
      F+=0.5*(operator()(nuMin)+operator()(nuMax))*(nuMax-nuMin);
      // Add table entry
      tableEntry.x=nuMax;
      tableEntry.f=F;
      table.push_back(tableEntry);
    }
  }
    
  // Normalize probability
  for (i=0;i<table.size();i++) table[i].f=table[i].f/F;

  // Resize inverse table
  tableInv.resize(table.size());

  // Construct inverse table
  for (i=0;i<table.size();i++) { tableInv[i].x=table[i].f; tableInv[i].f=table[i].x; }

  // Inititate cumulative distribution function interpolators
  pF.init(table);
  pFInv.init(tableInv);
}

// Get emissivity cumulative distribution function
double BBEta2::F(double nu) {
  double _F;

  // Interpolate cumulative probability (using constant extrapolation if necessary)
  _F=pF(MIN(MAX(nu,pF.domain().min),pF.domain().max));

  return _F;
}

// Get emissivity cumulative distribution function (inverse version)
double BBEta2::FInv(double _F) {
  double nu;

  // Interpolate frequency (using constant extrapolation if necessary)
  nu=pFInv(MIN(MAX(_F,pFInv.domain().min),pFInv.domain().max));

  return nu;
}

// Get memory allocated - New!
double BBEta2::mem() const {
  double _mem=0;

  // Add memory allocated for tables
  _mem+=pF.mem();
  _mem+=pFInv.mem();

  return _mem;
}

// Get emissivity distribution function (table)
double BBEta2Table::operator ()(double nu) {
  return eta(nu);
}

// Load (YAML) data
void BBEta2Table::load(const YAML::Node& data) {
  vector<FunctionValue> table(data.size());
  int i;

  // Get table data
  for (i=0;i<data.size();i++) {
    if (data[i][0]) table[i].x=data[i][0].as<double>();
    else { cerr<<"BBEta2Table::load, Error: Could not get data\n"; exit(-1); }
    if (data[i][1]) table[i].f=data[i][1].as<double>();
    else { cerr<<"BBEta2Table::load, Error: Could not get data\n"; exit(-1); }
  }

  // Initiate interpolator
  eta.init(table);
}

// Adjust edge - New!
void BBEta2Table::adjustEdge(const double& nuEdge) {
  int i;

  // Get table
  vector<FunctionValue> table=eta.table();

  // If no table - Return
  if (table.size()==0) return;

  // Get frequency scaling
  double nuScale=nuEdge/table[table.size()-1].x;
  
  // Adjust frequency scale
  for (i=0;i<table.size();i++) table[i].x=nuScale*table[i].x;

  // Re-initiate interpolator with adjusted table
  eta.init(table);
}

// Log data
void BBEta2Table::log() {
  int i;

  // Log 
  cout<<"BBEta2Table::log: Table loaded\n";
}

// Create copy
BBEta2* BBEta2Table::copy() {
  return new BBEta2Table(*this);
}

// Get memory allocated - New!
double BBEta2Table::mem() const {
  double _mem=sizeof(*this);

  // Add (dynamic) memory allocated for baseclass
  _mem+=BBEta2::mem();

  // Add memory allocated for table
  _mem+=eta.mem();

  return _mem;
}

// Get bound-free radiative cross-section (value)
double BFSigmaRadValue::operator ()(double nu) {
  return sigma;
}

// Load (YAML) data
void BFSigmaRadValue::load(const YAML::Node& data) {

  // Get cross-section
  sigma=data.as<double>();
}

// Log data
void BFSigmaRadValue::log() {
  int i;

  // Log 
  cout<<"BFSigmaRadValue::log: sigma: "<<sigma<<"\n";
}

// Create copy
BFSigmaRad* BFSigmaRadValue::copy() {
  return new BFSigmaRadValue(*this);
}

// Get memory allocated - New!
double BFSigmaRadValue::mem() const {
  double _mem=sizeof(*this);

  return _mem;
}

// Get bound-free radiative cross-section (table)
//double BFSigmaRadTable::operator ()(double nu) const {
double BFSigmaRadTable::operator ()(double nu) {
  //cout<<"BFSigmaRadTable::operator (): nu: "<<nu<<", sigma.domain.min: "<<sigma.domain().min<<", sigma.domain.max: "<<sigma.domain().max<<", sigma("<<sigma.domain().max<<"): "<<sigma(sigma.domain().max)<<", pow(nu/sigma.domain.max,-3): "<<pow(nu/sigma.domain().max,-3)<<"\n"; 
  // New! - Testing!
  double x,x3,sigmaMax;
  // If within table domain - Return interpolated cross-section
  if (nu<=sigma.domain().max and nu>=sigma.domain().min) return sigma(nu);
  // Else - If below domain - Return zero
  else if (nu<sigma.domain().min) return 0.0;
  // Else - Return cross-section extrapolated as nu**-3
  //else return sigma(sigma.domain().max)*pow(nu/sigma.domain().max,-3);
  else { x=sigma.domain().max/nu; x3=x*x*x; sigmaMax=sigma.table()[sigma.table().size()-1].f; return sigmaMax*x3; } // Changed! - Testing!
}

// Load (YAML) data
void BFSigmaRadTable::load(const YAML::Node& data) {
  vector<FunctionValue> table(data.size());
  int i;

  // Get table data
  for (i=0;i<data.size();i++) {
    if (data[i][0]) table[i].x=data[i][0].as<double>();
    else { cerr<<"BFSigmaRadTable::load, Error: Could not get data\n"; exit(-1); }
    if (data[i][1]) table[i].f=data[i][1].as<double>();
    else { cerr<<"BFSigmaRadTable::load, Error: Could not get data\n"; exit(-1); }
  }

  // Initiate interpolator
  sigma.init(table);
}

// Resample cross-section table
double BFSigmaRadTable::resampleTable(const vector<double>& nu, const double& nuEdge) {
  // Resample cross-section table
  sigma=resampleBFTable(nu,nuEdge,sigma);
}

// Adjust edge - New!
void BFSigmaRadTable::adjustEdge(const double& nuEdge) {
  int i;

  // Get table
  vector<FunctionValue> table=sigma.table();

  // If no table - Return
  if (table.size()==0) return;

  // Get frequency scaling
  double nuScale=nuEdge/table[0].x;
  
  // Adjust frequency scale
  for (i=0;i<table.size();i++) table[i].x=nuScale*table[i].x;

  // Re-initiate interpolator with adjusted table
  sigma.init(table);
}

// Log data
void BFSigmaRadTable::log() {
  int i;

  // Log 
  cout<<"BFSigmaRadTable::log: Table loaded\n";
}

// Create copy
BFSigmaRad* BFSigmaRadTable::copy() {
  return new BFSigmaRadTable(*this);
}

// Get memory allocated - New!
double BFSigmaRadTable::mem() const {
  double _mem=sizeof(*this);
  
  // Add memory allocated for table
  _mem+=sigma.mem();

  return _mem;
}

// Get bound-free radiative cross-section (powerlaw)
//double BFSigmaRadPLaw::operator ()(double nu) const {
double BFSigmaRadPLaw::operator ()(double nu) {
  //if (nu>=nu0) return sigma*pow(nu/(nu0),p);
  //else return 0;
  // Hack - Fix this in a better way !!!
  return sigma*pow(nu/(nu0),p);
}

// Load (YAML) data
void BFSigmaRadPLaw::load(const YAML::Node& data) {
  
  // Get radiative cross-section
  if (data[0]) sigma=data[0].as<double>();
  else { cerr<<"BFSigmaRadPLaw::load, Error: Could not get data\n"; exit(-1); }  

  // Get frequency
  if (data[1]) nu0=data[1].as<double>();
  else { cerr<<"BFSigmaRadPLaw::load, Error: Could not get data\n"; exit(-1); }  

  // Get exponent
  if (data[2]) p=data[2].as<double>();
  else { cerr<<"BFSigmaRadPLaw::load, Error: Could not get data\n"; exit(-1); }  
}

// Log data
void BFSigmaRadPLaw::log() {
  int i;

  // Log 
  cout<<"BFSimgaRadPlaw::log: sigma="<<sigma<<", nu0="<<nu0<<", p="<<p<<"\n";
}

// Create copy
BFSigmaRad* BFSigmaRadPLaw::copy() {
  return new BFSigmaRadPLaw(*this);
}

// Get memory allocated - New!
double BFSigmaRadPLaw::mem() const {
  double _mem=sizeof(*this);

  return _mem;
}

// Get bound-free radiative cross-section (powerlaw/gaunt)
//double BFSigmaRadPLawGaunt::operator ()(double nu) const {
double BFSigmaRadPLawGaunt::operator ()(double nu) {
  //if (nu>=nu0) {
  //cout<<"BFSigmaRadPLawGaunt: nu: "<<nu<<", nu0: "<<nu0<<", sigma: "<<sigma<<", p: "<<p<<", gaunt: "<<gaunt(nu)<<"\n";
  //return gaunt(nu)*sigma*pow(nu/(nu0),p);
  //}
  //else return 0;
  // Hack - Fix this in a better way !!!
  return gaunt(nu)*sigma*pow(nu/(nu0),p);
}

// Load (YAML) data
void BFSigmaRadPLawGaunt::load(const YAML::Node& data) {
  vector<FunctionValue> table;
  int i;
  
  // Get radiative cross-section
  if (data[0]) sigma=data[0].as<double>();
  else { cerr<<"BFSigmaRadPLawGaunt::load, Error: Could not get data\n"; exit(-1); }  

  // Get frequency
  if (data[1]) nu0=data[1].as<double>();
  else { cerr<<"BFSigmaRadPLawGaunt::load, Error: Could not get data\n"; exit(-1); }  

  // Get exponent
  if (data[2]) p=data[2].as<double>();
  else { cerr<<"BFSigmaRadPLawGaunt::load, Error: Could not get data\n"; exit(-1); }

  // Get gaunt factor
  if (data[3]) {

    // Resize table
    table.resize(data[3].size());

    // Get table data
    for (i=0;i<data[3].size();i++) {
      if (data[3][i][0]) table[i].x=data[3][i][0].as<double>();
      else { cerr<<"BFSigmaRadPLawGaunt::load, Error: Could not get data\n"; exit(-1); }
      if (data[3][i][1]) table[i].f=data[3][i][1].as<double>();
      else { cerr<<"BFSigmaRadPLawGaunt::load, Error: Could not get data\n"; exit(-1); }    
    }

    // Initiate interpolator
    gaunt.init(table);
  }
  else { cerr<<"BFSigmaRadPLawGaunt::load, Error: Could not get data\n"; exit(-1); }
}

// Resample cross-section table
double BFSigmaRadPLawGaunt::resampleTable(const vector<double>& nu, const double& nuEdge) {
  // Resample gaunt-factor table
  gaunt=resampleBFTable(nu,nuEdge,gaunt);
}

// Log data
void BFSigmaRadPLawGaunt::log() {
  int i;

  // Log 
  cout<<"BFSimgaRadPlawGaunt::log: sigma="<<sigma<<", nu0="<<nu0<<", p="<<p<<", Gaunt factor loaded\n";
}

// Create copy
BFSigmaRad* BFSigmaRadPLawGaunt::copy() {
  return new BFSigmaRadPLawGaunt(*this);
}

// Get memory allocated - New!
double BFSigmaRadPLawGaunt::mem() const {
  double _mem=sizeof(*this);
  
  // Add memory allocated for table
  _mem+=gaunt.mem();

  return _mem;
}

// Get bound-free radiative cross-section (Hummer fit)
//double BFSigmaRadHummer::operator ()(double nu) const {
double BFSigmaRadHummer::operator ()(double nu) {
  /*double x=nu/nu0;
  if (x>=1) {
    if (x>=x1) return pow(10,a1+b1*log10(x));
    else return pow(10,a0+b0*log10(x)+c0*pow(log10(x),2)+d0*pow(log10(x),3));
    }*/
  double x=MAX(nu/nu0,1.0);
  // Should be like this, right? No !!!
  //double x=MIN(nu/nu0,1.0);
  double log10x=log10(x);
  //if (log10x>=x1) return pow(10,a1+log10x*b1);
  // But should be like this, right?
  if (x>=x1) return pow(10,a1+log10x*b1);
  else return pow(10,a0+log10x*(b0+log10x*(c0+log10x*d0)));
}

// Load (YAML) data
void BFSigmaRadHummer::load(const YAML::Node& data) {
  
  // Get edge frequency
  if (data[0]) nu0=data[0].as<double>();
  else { cerr<<"BFSigmaRadHummer::load, Error: Could not get data\n"; exit(-1); }  

  // Get fit parameters
  if (data[1] and data[2] and data[3] and data[4] and data[5] and data[6] and data[7]) { 
    a0=data[1].as<double>(); 
    b0=data[2].as<double>(); 
    c0=data[3].as<double>();
    d0=data[4].as<double>();
    x1=data[5].as<double>();
    a1=data[6].as<double>();
    b1=data[7].as<double>();
  }
  else { cerr<<"BFSigmaRadHummer::load, Error: Could not get data\n"; exit(-1); }  
}

// Log data
void BFSigmaRadHummer::log() {
  int i;

  // Log 
  cout<<"BFSimgaRadHummer::log: nu0="<<nu0<<", a0="<<a0<<", b0="<<b0<<", c0="<<c0<<", d0="<<d0<<", x1="<<x1<<", a1="<<a1<<", b1="<<b1<<"\n";
}

// Create copy
BFSigmaRad* BFSigmaRadHummer::copy() {
  return new BFSigmaRadHummer(*this);
}

// Get memory allocated - New!
double BFSigmaRadHummer::mem() const {
  double _mem=sizeof(*this);
  
  return _mem;
}

// Get bound-free radiative cross-section (Seaton fit)
//double BFSigmaRadSeaton::operator ()(double nu) const {
double BFSigmaRadSeaton::operator ()(double nu) {
  // Removed
  //double x=MIN(nu/nu0,1.0);
  // Should be like this, right?
  double x=MAX(nu/nu0,1.0);
  //double x=MAX((nu+nuOffset)/nu0,1.0);
  return sigma0*(alpha+(1.0-alpha)/x)*pow(x,-beta);
}

// Load (YAML) data
void BFSigmaRadSeaton::load(const YAML::Node& data) {
  
  // Get edge frequency
  if (data[0]) nu0=data[0].as<double>();
  else { cerr<<"BFSigmaRadSeaton::load, Error: Could not get data\n"; exit(-1); }  

  // Get fit parameters
  if (data[1] and data[2] and data[3]) { 
    sigma0=data[1].as<double>(); 
    alpha=data[2].as<double>(); 
    beta=data[3].as<double>();
  }
  else { cerr<<"BFSigmaRadSeaton::load, Error: Could not get data\n"; exit(-1); }  
  // If provided - Get frequency offset - New!
  //if (data[4]) nuOffset=data[1].as<double>();
  // Else - Set frequency offset to zero
  //else nuOffset=0.0;
}

// Log data
void BFSigmaRadSeaton::log() {
  int i;

  // Log 
  cout<<"BFSimgaRadSeaton::log: nu0="<<nu0<<", sigma0="<<sigma0<<", alpha="<<alpha<<", beta="<<beta<<"\n";
}

// Create copy
BFSigmaRad* BFSigmaRadSeaton::copy() {
  return new BFSigmaRadSeaton(*this);
}

// Get memory allocated - New!
double BFSigmaRadSeaton::mem() const {
  double _mem=sizeof(*this);
  
  return _mem;
}

// Get bound-free radiative cross-section (Verner fit)
// Note: Same as BFSigmaRadVerner95 except units
double BFSigmaRadVerner::operator ()(double nu) {
  double E,Q,y,sigma=0;
  int i;

  // For each n/l entry
  for (i=0;i<sigma0.size();i++) {
    E=h*nu;
    // If the energy is above the threshold
    if (E>=Eth[i]) {
      // Calculate and sum cross-section
      Q=5.5+l[i]-0.5*P[i];
      y=E/E0[i];
      sigma+=sigma0[i]*(pow(y-1,2)+pow(yw[i],2))*pow(y,-Q)*pow(1+sqrt(y/ya[i]),-P[i]);
    }
  }
  return sigma;
}

// Load (YAML) data
void BFSigmaRadVerner::load(const YAML::Node& data) {
  int i;
  
  // Resize fit parameter vectors
  n.resize(data.size());
  l.resize(data.size());
  Eth.resize(data.size());
  E0.resize(data.size());
  sigma0.resize(data.size());
  ya.resize(data.size());
  P.resize(data.size());
  yw.resize(data.size());

  // For each n/l entry
  for (i=0;i<data.size();i++) {
    // Get fit parameters
    if (data[i][0] and data[i][1] and data[i][2] and data[i][3] and data[i][4] and data[i][5] and data[i][6] and data[i][7]) {
      n[i]=data[i][0].as<int>();
      l[i]=data[i][1].as<int>();
      Eth[i]=data[i][2].as<double>();
      E0[i]=data[i][3].as<double>();
      sigma0[i]=data[i][4].as<double>(); 
      ya[i]=data[i][5].as<double>();
      P[i]=data[i][6].as<double>();
      yw[i]=data[i][7].as<double>(); 
    }
    else { cerr<<"BFSigmaRadVerner::load, Error: Could not get data\n"; exit(-1); } 
  }

}

// Log data
void BFSigmaRadVerner::log() {
  int i;

  // Log
  for (i=0;i<sigma0.size();i++) cout<<"BFSimgaRadVerner::log: n="<<n[i]<<", l="<<l[i]<<", Eth="<<Eth[i]<<", E0="<<E0[i]<<", sigma0="<<sigma0[i]<<", ya="<<ya[i]<<", P="<<P[i]<<", yw="<<yw[i]<<"\n";
}

// Create copy
BFSigmaRad* BFSigmaRadVerner::copy() {
  return new BFSigmaRadVerner(*this);
}

// Get memory allocated - New!
double BFSigmaRadVerner::mem() const {
  double _mem=sizeof(*this);
  int i;

  // Add memory allocated for fit parameter vectors
  for (i=0;i<sigma0.size();i++) _mem+=8*sizeof(double);
  
  return _mem;
}

// Get bound-free radiative cross-section (Verner 1995 fit)
double BFSigmaRadVerner95::operator ()(double nu) {
  double E,Q,y,sigma=0;
  int i;

  // For each n/l entry
  for (i=0;i<sigma0.size();i++) {
    E=h*nu;
    // If the energy is above the threshold
    if (E>=Eth[i]) {
      // Calculate and sum cross-section
      Q=5.5+l[i]-0.5*P[i];
      y=E/E0[i];
      sigma+=sigma0[i]*(pow(y-1,2)+pow(yw[i],2))*pow(y,-Q)*pow(1+sqrt(y/ya[i]),-P[i]);
    }
  }
  return sigma;
}

// Load (YAML) data
void BFSigmaRadVerner95::load(const YAML::Node& data) {
  int i;
  
  // Resize fit parameter vectors
  n.resize(data.size());
  l.resize(data.size());
  Eth.resize(data.size());
  E0.resize(data.size());
  sigma0.resize(data.size());
  ya.resize(data.size());
  P.resize(data.size());
  yw.resize(data.size());

  // For each n/l entry
  for (i=0;i<data.size();i++) {
    // Get fit parameters
    if (data[i][0] and data[i][1] and data[i][2] and data[i][3] and data[i][4] and data[i][5] and data[i][6] and data[i][7]) {
      n[i]=data[i][0].as<int>();
      l[i]=data[i][1].as<int>();
      Eth[i]=data[i][2].as<double>()*eV;
      E0[i]=data[i][3].as<double>()*eV;
      sigma0[i]=data[i][4].as<double>()*Mb;
      ya[i]=data[i][5].as<double>();
      P[i]=data[i][6].as<double>();
      yw[i]=data[i][7].as<double>(); 
    }
    else { cerr<<"BFSigmaRadVerner95::load, Error: Could not get data\n"; exit(-1); } 
  }

} 

// Adjust edge - New!
void BFSigmaRadVerner95::adjustEdge(const double& nuEdge) {
  
  // Calculate edge energy
  double EEdge=h*nuEdge;

  // Adjust energy scale for outer shell
  E0[0]=E0[0]*(EEdge/Eth[0]);
  // Adjust energy threshold for outer shell
  Eth[0]=EEdge;
}

// Log data
void BFSigmaRadVerner95::log() {
  int i;

  // Log
  for (i=0;i<sigma0.size();i++) cout<<"BFSimgaRadVerner95::log: n="<<n[i]<<", l="<<l[i]<<", Eth="<<Eth[i]<<", E0="<<E0[i]<<", sigma0="<<sigma0[i]<<", ya="<<ya[i]<<", P="<<P[i]<<", yw="<<yw[i]<<"\n";
}

// Create copy
BFSigmaRad* BFSigmaRadVerner95::copy() {
  return new BFSigmaRadVerner95(*this);
}

// Get memory allocated - New!
double BFSigmaRadVerner95::mem() const {
  double _mem=sizeof(*this);
  int i;

  // Add memory allocated for fit parameter vectors
  for (i=0;i<sigma0.size();i++) _mem+=8*sizeof(double);
  
  return _mem;
}

// Get bound-free radiative cross-section (Verner 1996 fit)
double BFSigmaRadVerner96::operator ()(double nu) {
  double E,x,y,sigma=0;

  E=h*nu;

  // If the energy is above the threshold
  if (E>=Eth) {
    // Calculate cross-section
    x=E/E0-y0;
    y=sqrt(pow(x,2)+pow(y1,2));
    sigma=sigma0*(pow(x-1,2)+pow(yw,2))*pow(y,0.5*P-5.5)*pow(1+sqrt(y/ya),-P);
  }

  return sigma;
}

// Load (YAML) data
void BFSigmaRadVerner96::load(const YAML::Node& data) {

  // Get fit parameters
  if (data.size()==9) {
      Eth=data[0].as<double>()*eV;
      Emax=data[1].as<double>()*eV;
      E0=data[2].as<double>()*eV;
      sigma0=data[3].as<double>()*Mb;
      ya=data[4].as<double>();
      P=data[5].as<double>();
      yw=data[6].as<double>(); 
      y0=data[7].as<double>();
      y1=data[8].as<double>();
  }
  else { cerr<<"BFSigmaRadVerner96::load, Error: Could not get data\n"; exit(-1); } 
}

// Adjust edge - New!
void BFSigmaRadVerner96::adjustEdge(const double& nuEdge) {

  // Calculate edge energy
  double EEdge=h*nuEdge;

  // Adjust energy scale
  E0=E0*(EEdge/Eth);
  // Adjust energy threshold
  Eth=EEdge;
}

// Log data
void BFSigmaRadVerner96::log() {

  // Log
  cout<<"BFSimgaRadVerner96::log: Eth="<<Eth<<", Emax="<<Emax<<", E0="<<E0<<", sigma0="<<sigma0<<", ya="<<ya<<", P="<<P<<", yw="<<yw<<", y0="<<y0<<", y1="<<y1<<"\n";
}

// Create copy
BFSigmaRad* BFSigmaRadVerner96::copy() {
  return new BFSigmaRadVerner96(*this);
}

// Get memory allocated - New!
double BFSigmaRadVerner96::mem() const {
  double _mem=sizeof(*this);
  
  return _mem;
}

// Get bound-free radiative cross-section (Verner 1995/1996 fit)
double BFSigmaRadVerner9596::operator ()(double nu) {
  double E,sigma;

  E=h*nu;

  // Get cross-section for outer shell (using Verner 1996 fit)
  sigma=v96(nu);
  // If E>Emax - Add cross-section for inner shells (using Verner 1995 fit)
  if (E>v96.Emax) sigma+=v95(nu);

  return sigma;
}

// Load (YAML) data
void BFSigmaRadVerner9596::load(const YAML::Node& data) {
  int i;

  if (data["v95"]) v95.load(data["v95"]);
  else { cerr<<"BFSigmaRadVerner9596::load, Error: Could not get v95 data\n"; exit(-1); } 

  if (data["v96"]) v96.load(data["v96"]);
  else { cerr<<"BFSigmaRadVerner9596::load, Error: Could not get v96 data\n"; exit(-1); } 
}

// Adjust edge - New!
void BFSigmaRadVerner9596::adjustEdge(const double& nuEdge) {

  // Adjust edge for outer shell
  v96.adjustEdge(nuEdge);
}


// Log data
void BFSigmaRadVerner9596::log() {
  int i;

  // Log Verner 1995 data
  v95.log();

  // Log Verner 1996 data
  v96.log();
}

// Create copy
BFSigmaRad* BFSigmaRadVerner9596::copy() {
  return new BFSigmaRadVerner9596(*this);
}

// Get memory allocated - New!
double BFSigmaRadVerner9596::mem() const {
  double _mem=sizeof(*this);
 
  // Add memory allocated for Verner 95 data
  _mem+=v95.mem();

  // Add memory allocated for Verner 96 data
  _mem+=v96.mem();
 
  return _mem;
}

// Get bound-free collisional cross-section (table)
//double BFSigmaCollTable::operator ()(double T) const {
double BFSigmaCollTable::operator ()(double T) {
  return sigma(T);
}

// Load (YAML) data
void BFSigmaCollTable::load(const YAML::Node& data) {
  vector<FunctionValue> table(data.size());
  int i;

  // Get table data
  for (i=0;i<data.size();i++) {
    if (data[i][0]) table[i].x=data[i][0].as<double>();
    else { cerr<<"BFSigmaCollTable::load, Error: Could not get data\n"; exit(-1); }
    if (data[i][1]) table[i].f=data[i][1].as<double>();
    else { cerr<<"BFSigmaCollTable::load, Error: Could not get data\n"; exit(-1); }    
  }

  // Initiate interpolator
  sigma.init(table);
}

// Log data
void BFSigmaCollTable::log() {
  int i;

  // Log 
  cout<<"BFSimgaCollTable::log: Table loaded\n";
}

// Create copy
BFSigmaColl* BFSigmaCollTable::copy() {
  return new BFSigmaCollTable(*this);
}

// Get memory allocated - New!
double BFSigmaCollTable::mem() const {
  double _mem=sizeof(*this);
 
  // Add memory allocated for table
  _mem+=sigma.mem();
 
  return _mem;
}

// Get bound-free collisional cross-section (powerlaw)
//double BFSigmaCollPLaw::operator ()(double T) const {
double BFSigmaCollPLaw::operator ()(double T) {
  return sigma*pow(T/(T0),p);
}

// Load (YAML) data
void BFSigmaCollPLaw::load(const YAML::Node& data) {
  
  // Get collisional strength
  if (data[0]) sigma=data[0].as<double>();
  else { cerr<<"BFSigmaCollPLaw::load, Error: Could not get data\n"; exit(-1); }  

  // Get temperature
  if (data[1]) T0=data[1].as<double>();
  else { cerr<<"BFSigmaCollPLaw::load, Error: Could not get data\n"; exit(-1); }  

  // Get exponent
  if (data[2]) p=data[2].as<double>();
  else { cerr<<"BFSigmaCollPLaw::load, Error: Could not get data\n"; exit(-1); }  
}

// Log data
void BFSigmaCollPLaw::log() {
  int i;

  // Log 
  cout<<"BFSimgaCollPlaw::log: sigma="<<sigma<<", T0="<<T0<<", p="<<p<<"\n";
}

// Create copy
BFSigmaColl* BFSigmaCollPLaw::copy() {
  return new BFSigmaCollPLaw(*this);
}

// Get memory allocated - New!
double BFSigmaCollPLaw::mem() const {
  double _mem=sizeof(*this);
 
  return _mem;
}

// Get bound-free collisional strength (value)
//double BFOmegaValue::operator ()(double T) const {
double BFOmegaValue::operator ()(double T) {
  return omega;
}

// Load (YAML) data
void BFOmegaValue::load(const YAML::Node& data) {

  // Get collisional strength
  omega=data.as<double>();
}

// Log data
void BFOmegaValue::log() {
  int i;

  // Log 
  cout<<"BFOmegaValue::log: omega="<<omega<<"\n";
}

// Create copy
BFOmega* BFOmegaValue::copy() {
  return new BFOmegaValue(*this);
}

// Get memory allocated - New!
double BFOmegaValue::mem() const {
  double _mem=sizeof(*this);
 
  return _mem;
}

// Get bound-free collisional strength (table)
//double BFOmegaTable::operator ()(double T) const {
double BFOmegaTable::operator ()(double T) {
  return omega(T);
}

// Load (YAML) data
void BFOmegaTable::load(const YAML::Node& data) {
  vector<FunctionValue> table(data.size());
  int i;

  // Get table data
  for (i=0;i<data.size();i++) {
    if (data[i][0]) table[i].x=data[i][0].as<double>();
    else { cerr<<"BFOmegaTable::load, Error: Could not get data\n"; exit(-1); }
    if (data[i][1]) table[i].f=data[i][1].as<double>();
    else { cerr<<"BFOmegaTable::load, Error: Could not get data\n"; exit(-1); }    
  }

  // Initiate interpolator
  omega.init(table);
}

// Log data
void BFOmegaTable::log() {
  int i;

  // Log 
  cout<<"BFOmegaTable::log: Table loaded\n";
}

// Create copy
BFOmega* BFOmegaTable::copy() {
  return new BFOmegaTable(*this);
}

// Get memory allocated - New!
double BFOmegaTable::mem() const {
  double _mem=sizeof(*this);
 
  // Add memory allocated for table
  _mem+=omega.mem();
 
  return _mem;
}

// Get bound-free collisional strength (powerlaw)
//double BFOmegaPLaw::operator ()(double T) const {
double BFOmegaPLaw::operator ()(double T) {
  return omega*pow(T/(T0),p);
}

// Load (YAML) data
void BFOmegaPLaw::load(const YAML::Node& data) {
  
  // Get collisional strength
  if (data[0]) omega=data[0].as<double>();
  else { cerr<<"BFOmegaPLaw::load, Error: Could not get data\n"; exit(-1); }  

  // Get temperature
  if (data[1]) T0=data[1].as<double>();
  else { cerr<<"BFOmegaPLaw::load, Error: Could not get data\n"; exit(-1); }  

  // Get exponent
  if (data[2]) p=data[2].as<double>();
  else { cerr<<"BFOmegaPLaw::load, Error: Could not get data\n"; exit(-1); }  
}

// Log data
void BFOmegaPLaw::log() {
  int i;

  // Log 
  cout<<"BFOmegaPlaw::log: omega="<<omega<<", T0="<<T0<<", p="<<p<<"\n";
}

// Create copy
BFOmega* BFOmegaPLaw::copy() {
  return new BFOmegaPLaw(*this);
}

// Get memory allocated - New!
double BFOmegaPLaw::mem() const {
  double _mem=sizeof(*this);
  
  return _mem;
}

// Load interpolation type
IPolType atomdata::loadIPolType(const YAML::Node& data) {
  if (data.as<string>().compare("log")==0) return ipolLog;
  else return ipolLinear;
}

// Log interpolation type
string atomdata::logIPolType(const IPolType& ipolType) {
  switch(ipolType) {
  case ipolLog : return "ipolLog";
  case ipolLinear : return "ipolLinear";
  default : return "unknown";
  }
}

// Load extrapolation type
EPolType atomdata::loadEPolType(const YAML::Node& data) {
  if (data.as<string>().compare("const")==0) return epolConst;
  else return epolNone;
}

// Log extrapolation type
string atomdata::logEPolType(const EPolType& epolType) {
  switch(epolType) {
  case epolConst : return "ipolConst";
  case epolNone : return "epolNone";
  default : return "unknown";
  }
}

// Get bound-free recombination coeffieicent (value)
double BFAlfaValue::operator ()(double T) {
  return alfa;
}

// Load (YAML) data
void BFAlfaValue::load(const YAML::Node& data) {

  // Get collisional strength
  alfa=data["data"].as<double>();
}

// Log data
void BFAlfaValue::log() {
  int i;

  // Log 
  cout<<"BFAlfaValue::log: alfa="<<alfa<<"\n";
}

// Create copy
BFAlfa* BFAlfaValue::copy() {
  return new BFAlfaValue(*this);
}

// Get memory allocated - New!
double BFAlfaValue::mem() const {
  double _mem=sizeof(*this);
  
  return _mem;
}

// Get bound-free recombination coeffiecient (table) - New!
double BFAlfaTable::operator ()(double T) {
  double x;
  
  // Transform interpolation quantity according to interpolation type
  if (ipolType==ipolLog) x=log10(T);
  else x=T;

  // If constant extrapolation - Restrict interpolation quantity to domain
  if (epolType==epolConst) x=MIN(MAX(x,alfa.domain().min),alfa.domain().max);
  // Else - No extrapolation

  // Return interpolated (or extrapolated) recombination coefficient
  return alfa(x);
}

// Load (YAML) data- New!
void BFAlfaTable::load(const YAML::Node& data) {
  vector<FunctionValue> table(data["data"].size());
  int i;

  // If provided- Get interpolation type
  if (data["ipolType"]) ipolType=loadIPolType(data["ipolType"]);
  // Else - Use linear interpolation
  else ipolType=ipolLinear;

  // If provided - Get extrapolation type
  if (data["epolType"]) epolType=loadEPolType(data["epolType"]);
  // Else - Use no extrapolation
  else epolType=epolNone;

  // Get table data
  for (i=0;i<data["data"].size();i++) {
    if (data["data"][i][0]) table[i].x=data["data"][i][0].as<double>();
    else { cerr<<"BFAlfaTable::load, Error: Could not get data\n"; exit(-1); }
    if (data["data"][i][1]) table[i].f=data["data"][i][1].as<double>();
    else { cerr<<"BFAlfaTable::load, Error: Could not get data\n"; exit(-1); }
    // Transform interpolation quantity according to interpolation type
    if (ipolType==ipolLog) table[i].x=log10(table[i].x);
  }

  // Initiate interpolator
  alfa.init(table);
}

// Log data - New!
void BFAlfaTable::log() {
  int i;

  // Log 
  cout<<"BFAlfaTable::log: Table loaded, ipolType: "<<logIPolType(ipolType)<<", epolType: "<<logEPolType(epolType)<<"\n";
}

// Create copy - New!
BFAlfa* BFAlfaTable::copy() {
  return new BFAlfaTable(*this);
}

// Get memory allocated - New!
double BFAlfaTable::mem() const {
  double _mem=sizeof(*this);
 
  // Add memory allocated to table
  _mem+=alfa.mem();
 
  return _mem;
}

// Get bound-free recombination coeffiecient (powerlaw))
//double BFAlfaPLaw::operator ()(double T) const {
double BFAlfaPLaw::operator ()(double T) {
  return alfa*pow(T/(T0),p);
}

// Load (YAML) data
void BFAlfaPLaw::load(const YAML::Node& data) {
  
  // Get collisional strength
  if (data["data"][0]) alfa=data["data"][0].as<double>();
  else { cerr<<"BFAlfaPLaw::load, Error: Could not get data\n"; exit(-1); }  

  // Get temperature
  if (data["data"][1]) T0=data["data"][1].as<double>();
  else { cerr<<"BFAlfaPLaw::load, Error: Could not get data\n"; exit(-1); }  

  // Get exponent
  if (data["data"][2]) p=data["data"][2].as<double>();
  else { cerr<<"BFAlfaPLaw::load, Error: Could not get data\n"; exit(-1); }  
}

// Log data
void BFAlfaPLaw::log() {
  int i;

  // Log 
  cout<<"BFAlfaPlaw::log: alfa="<<alfa<<", T0="<<T0<<", p="<<p<<"\n";
}

// Create copy
BFAlfa* BFAlfaPLaw::copy() {
  return new BFAlfaPLaw(*this);
}

// Get memory allocated - New!
double BFAlfaPLaw::mem() const {
  double _mem=sizeof(*this);
 
  return _mem;
}

// Get bound-free recombination coefficient (Verner et al. 1996 fit)
double BFAlfaVerner::operator ()(double T) {
  double ar=a*pow(sqrt(T/T0)*pow(1.0+sqrt(T/T0),1-b)*pow(1.0+sqrt(T/T1),1+b),-1)*1e-6;
  return ar;
}

// Load (YAML) data
void BFAlfaVerner::load(const YAML::Node& data) {
  
  // Get fit parameters
  if (data["data"].size()==4) { 
    a=data["data"][0].as<double>();
    b=data["data"][1].as<double>(); 
    T0=data["data"][2].as<double>();
    T1=data["data"][3].as<double>();
  }
  else { cerr<<"BFAlfaVerner::load, Error: Could not get data\n"; exit(-1); }  
}

// Log data
void BFAlfaVerner::log() {
  int i;

  // Log 
  cout<<"BFAlfaVerner::log: a="<<a<<", b="<<b<<", T0="<<T0<<", T1="<<T1<<"\n";
}

// Create copy
BFAlfa* BFAlfaVerner::copy() {
  return new BFAlfaVerner(*this);
}

// Get memory allocated - New!
double BFAlfaVerner::mem() const {
  double _mem=sizeof(*this);
 
  return _mem;
}

// Get bound-free recombination coefficient (Shull et al. 1982 fit)
double BFAlfaShull::operator ()(double T) {
  double ar=0,ad=0;
  if (ARad!=0) ar=ARad*pow(T/1e4,-XRad)*1e-6;
  if (ADi!=0) ad=ADi*pow(T,-3.0/2.0)*exp(-T0/T)*(1.0+BDi*exp(-T1/T))*1e-6;
  return (ar+ad);
}

// Load (YAML) data
void BFAlfaShull::load(const YAML::Node& data) {
  
  // Get fit parameters
  if (data["data"].size()==6) { 
    ARad=data["data"][0].as<double>();
    XRad=data["data"][1].as<double>(); 
    ADi=data["data"][2].as<double>();
    BDi=data["data"][3].as<double>();
    T0=data["data"][4].as<double>();
    T1=data["data"][5].as<double>();
  }
  else { cerr<<"BFAlfaShull::load, Error: Could not get data\n"; exit(-1); }  
}

// Log data
void BFAlfaShull::log() {
  int i;

  // Log 
  cout<<"BFAlfaShull::log: ARad="<<ARad<<", XRad="<<XRad<<", ADi="<<ADi<<", BDi="<<BDi<<", T0="<<T0<<", T1="<<T1<<"\n";
}

// Create copy
BFAlfa* BFAlfaShull::copy() {
  return new BFAlfaShull(*this);
}

// Get memory allocated - New!
double BFAlfaShull::mem() const {
  double _mem=sizeof(*this);
 
  return _mem;
}

// Get bound-free recombination coefficient (Pequignot et al. 1986 fit)
double BFAlfaPequignot::operator ()(double T) {
  double alfa=alfa100*pow(T/1e2,-eta)*1e-18;
  return alfa;
}

// Load (YAML) data
void BFAlfaPequignot::load(const YAML::Node& data) {
  
  // Get fit parameters
  if (data["data"].size()==2) { 
    alfa100=data["data"][0].as<double>();
    eta=data["data"][1].as<double>(); 
  }
  else { cerr<<"BFAlfaPequignot::load, Error: Could not get data\n"; exit(-1); }  
}

// Log data
void BFAlfaPequignot::log() {
  int i;

  // Log 
  cout<<"BFAlfaPequignot::log: alfa100="<<alfa100<<", eta="<<eta<<"\n";
}

// Create copy
BFAlfa* BFAlfaPequignot::copy() {
  return new BFAlfaPequignot(*this);
}

// Get memory allocated - New!
double BFAlfaPequignot::mem() const {
  double _mem=sizeof(*this);
 
  return _mem;
}

// Get bound-free recombination coefficient (Nussbaumer et al. 1986 fit)
double BFAlfaNussbaumer::operator ()(double T) {
  double t=T/1e4;
  double tm=6.0;
  // What about tl and (tm=6.0), the lower (and upper) limits for the validity of the fit?
  // Solution: Bound t to be in fit domain (i.e. constant extrapolation) - New!
  t=MIN(MAX(t,tl),tm);
  double alfa=1e-12*(a/t+b+c*t+d*pow(t,2))*pow(t,(-3.0/2.0))*exp(-f/t)*1e-6;
  return alfa;
}

// Load (YAML) data
void BFAlfaNussbaumer::load(const YAML::Node& data) {
  
  // Get fit parameters
  if (data["data"].size()==6) { 
    a=data["data"][0].as<double>();
    b=data["data"][1].as<double>();
    c=data["data"][2].as<double>();
    d=data["data"][3].as<double>();
    f=data["data"][4].as<double>();
    tl=data["data"][5].as<double>();
  }
  else { cerr<<"BFAlfaNussbaumer::load, Error: Could not get data\n"; exit(-1); }  
}

// Log data
void BFAlfaNussbaumer::log() {
  int i;

  // Log 
  cout<<"BFAlfaNussbaumer::log: a="<<a<<", b="<<b<<", c="<<c<<", d="<<d<<", f="<<f<<", tl="<<tl<<"\n";
}

// Create copy
BFAlfa* BFAlfaNussbaumer::copy() {
  return new BFAlfaNussbaumer(*this);
}

// Get memory allocated - New!
double BFAlfaNussbaumer::mem() const {
  double _mem=sizeof(*this);
 
  return _mem;
}

// Get bound-free recombination coefficient (Badnell 2006 fit)
double BFAlfaBadnell::operator ()(double T) {
  double BB=B+C*exp(-T2/T);
  double alfa=A*pow(sqrt(T/T0)*pow(1+sqrt(T/T0),1-BB)*pow(1+sqrt(T/T1),1+BB),-1);
  return alfa;
}

// Load (YAML) data
void BFAlfaBadnell::load(const YAML::Node& data) {
  
  // Get fit parameters
  if (data["data"].size()==6) { 
    A=data["data"][0].as<double>()*1e-6;
    B=data["data"][1].as<double>();
    T0=data["data"][2].as<double>();
    T1=data["data"][3].as<double>();
    C=data["data"][4].as<double>();
    T2=data["data"][5].as<double>();
  }
  else { cerr<<"BFAlfaBadnell::load, Error: Could not get data\n"; exit(-1); }  
}

// Log data
void BFAlfaBadnell::log() {
  int i;

  // Log 
  cout<<"BFAlfaBadnell::log: A="<<A<<", B="<<B<<", T0="<<T0<<", T1="<<T1<<", C="<<C<<", T2="<<T2<<"\n";
}

// Create copy
BFAlfa* BFAlfaBadnell::copy() {
  return new BFAlfaBadnell(*this);
}

// Get memory allocated - New!
double BFAlfaBadnell::mem() const {
  double _mem=sizeof(*this);
 
  return _mem;
}

// Constructor
TransBBData::TransBBData() : A(0), A2(0), omega(NULL), eta2(NULL), processes(pNone), edgeAdjust(false), TransData() { } // Changed!

// Copy constructor
TransBBData::TransBBData(const TransBBData& transData) : A(transData.A), A2(transData.A2), omega(NULL), eta2(NULL), processes(transData.processes), edgeAdjust(transData.edgeAdjust), TransData(transData) { // Changed!
  int i,j;

  // Copy collisional strength object
  if (transData.omega!=NULL) omega=transData.omega->copy();
  else omega=NULL;

  // Copy two-photon emissivity object - New!
  if (transData.eta2!=NULL) eta2=transData.eta2->copy();
  else eta2=NULL;
}

// Assignment operator
TransBBData& TransBBData::operator=(const TransBBData& transData) {
  int i,j;

  if (&transData!=this) {
    TransData::operator=(transData);
    A=transData.A;
    A2=transData.A2; // New!
    
    // Copy collisional strength object
    // This should be added, right?
    // if (omega!=NULL) delete omega;
    if (transData.omega!=NULL) omega=transData.omega->copy();
    else omega=NULL;

    // Copy two-photon normalized emissivity object - New!
    // This should be added, right?
    // if (eta2!=NULL) delete eta2;
    if (transData.eta2!=NULL) eta2=transData.eta2->copy();
    else eta2=NULL;

    // Copy processes flags - New!
    processes=transData.processes;

    // Copy adjust edge flag - New!
    edgeAdjust=transData.edgeAdjust;
  }
}

// Destructor
TransBBData::~TransBBData() { 
  if (omega!=NULL) delete omega;
  if (eta2!=NULL) delete eta2; // New!
}

// Load (YAML) data
void TransBBData::load(const YAML::Node& data) {
  string type;

  // Get transition energy
  if (data["Xi"]) Xi=data["Xi"].as<double>();
  else { cerr<<"TransBBData::load, Error: could not get Xi\n"; exit(-1); }
  
  // Get Einstein spontaneous emission coefficient
  if (data["A"]) A=data["A"].as<double>();
  else { cerr<<"TransBBData::load, Error: could not get A\n"; exit(-1); }

  // If collisional strength available
  if (data["omega"]) {

    // Get collisional strength data type
    if (data["omega"]["type"]) type=data["omega"]["type"].as<string>();
    else { cerr<<"TransBBData::load, Error: Could not get omega::type\n"; exit(-1); }
    
    // Create collisional strength data object
    if (type.compare("value")==0) omega = new BBOmegaValue;
    else if (type.compare("table")==0) omega = new BBOmegaTable;
    else if (type.compare("plaw")==0) omega = new BBOmegaPLaw;
    else { cerr<<"TransBBData::load, Error: omega::type "<<type<<" not supported\n"; exit(-1); }
    
    // Load collisional strength data
    if (data["omega"]["data"]) omega->load(data["omega"]["data"]);
    else { cerr<<"TransBBData::load, Error: Could not get omega::data\n"; exit(-1); }
  }

  // Set radiative and collisional processes flags - New!
  processes=processes|pRadiative|pCollisional;
}

// Load two-photon transition data - New!
void TransBBData::load2(const YAML::Node& data) {
  string type;
  
  // Get Einstein spontaneous emission coefficient
  if (data["A"]) A2=data["A"].as<double>();
  else { cerr<<"TransBBData::load2, Error: could not get A\n"; exit(-1); }

  // If normalized emissivity available
  if (data["eta"]) {

    // Get normalized emissivity data type
    if (data["eta"]["type"]) type=data["eta"]["type"].as<string>();
    else { cerr<<"TransBBData::load2, Error: Could not get eta::type\n"; exit(-1); }
    
    // Create normalized emissivity data object
    if (type.compare("table")==0) eta2 = new BBEta2Table;
    else { cerr<<"TransBBData::load2, Error: eta::type "<<type<<" not supported\n"; exit(-1); }
    
    // Load normalized emissivity data
    if (data["eta"]["data"]) eta2->load(data["eta"]["data"]);
    else { cerr<<"TransBBData::load2, Error: Could not get eta::data\n"; exit(-1); }
  }
  else { cerr<<"TransBBData::load2, Error: could not get eta\n"; exit(-1); }

  // If specified - Load adjust edge flag
  if (data["edgeAdjust"]) edgeAdjust=(data["edgeAdjust"].as<string>().compare("yes")==0);
  else edgeAdjust=false;

  // Set two-photon processes flag - New!
  processes=processes|pTwoPhoton;
}

// Update tables (using the provided frequency binning) - New!
void TransBBData::updateTables(const vector<double>& nu) {
  int i,j;

  // If two-photon emissivity available - create cumulative distribution function table
  if (eta2!=NULL) eta2->createFTable(nu,Xi/h);
}

// Adjust edge - New!
void TransBBData::adjustEdge() {

  // If requested and radiative cross-section available - Adjust edge
  if (eta2!=NULL and edgeAdjust) eta2->adjustEdge(Xi/h);
}

// Log data
void TransBBData::log() {
  int i;

  // Log index, energy, total energy and statistical weight
  cout<<"TransBBData::log: uppeState::index="<<upperState->index<<", lowerState::index="<<lowerState->index<<", Xi="<<Xi<<", A="<<A<<"\n";

  // Log collisional stength
  if (omega!=NULL) omega->log();
  else cout<<"TransBBData:log: omega not loaded\n";

  // Log two-photon normalized emissivity - New!
  if (eta2!=NULL) eta2->log();
  else if (A2!=0) cout<<"TransBBData:log: eta2 not loaded\n";
}

// Get memory allocated - New!
double TransBBData::mem() const {
  double _mem=sizeof(*this);

  // Add memory allocated for collisional strength
  if (omega!=NULL) _mem+=omega->mem();

  // Add memory allocated for two-photon normalized emissivity
  if (eta2!=NULL) _mem+=eta2->mem();

  return _mem;
}

// Load detailed balance method
DBMethod atomdata::loadDBMethod(const YAML::Node& data) {
  if (data.as<string>().compare("forceSigma")==0) return dbForceSigma;
  if (data.as<string>().compare("forceAlfa")==0) return dbForceAlfa;
  else return dbNone;
}

// Log detailed balance method
string atomdata::logDBMethod(const DBMethod& dBMethod) {
  switch(dBMethod) {
  case dbForceSigma : return "forceSigma";
  case dbForceAlfa : return "forceAlfa";
  case dbNone : return "none";
  default : return "unknown";
  }
}

// Constructor
TransBFData::TransBFData() : sigmaRad(NULL), sigmaColl(NULL), omega(NULL), processes(pNone), dBMethod(dbNone), edgeAdjust(false), TransData() { } // Changed!

// Copy constructor
TransBFData::TransBFData(const TransBFData& transData) : sigmaRad(NULL), sigmaColl(NULL), omega(NULL), processes(transData.processes), dBMethod(transData.dBMethod), edgeAdjust(transData.edgeAdjust), TransData(transData) { // Changed!
  int i,j;

  // Copy radiative cross-section object
  if (transData.sigmaRad!=NULL) sigmaRad=transData.sigmaRad->copy();
  else sigmaRad=NULL;

  // Copy collisional cross-section object
  if (transData.sigmaColl!=NULL) sigmaColl=transData.sigmaColl->copy();
  else sigmaColl=NULL;

  // Copy collisional strength object
  if (transData.omega!=NULL) omega=transData.omega->copy();
  else omega=NULL;
}

// Assignment operator
TransBFData& TransBFData::operator=(const TransBFData& transData) {
  int i,j;

  if (&transData!=this) {
    TransData::operator=(transData);
    
    if (sigmaRad!=NULL) delete sigmaRad;
    if (sigmaColl!=NULL) delete sigmaColl;
    if (omega!=NULL) delete omega;

    // Copy radiative cross-section object
    if (transData.sigmaRad!=NULL) sigmaRad=transData.sigmaRad->copy();
    else sigmaRad=NULL;

    // Copy collisional cross-section object
    if (transData.sigmaColl!=NULL) sigmaColl=transData.sigmaColl->copy();
    else sigmaColl=NULL;

    // Copy collisional strength object
    if (transData.omega!=NULL) omega=transData.omega->copy();
    else omega=NULL;

    // Copy processes flags - New!
    processes=transData.processes;

    // Copy detailed balance method - New!
    dBMethod=transData.dBMethod;

    // Copy adjust edge flag - New!
    edgeAdjust=transData.edgeAdjust;
  }
}

// Destructor
TransBFData::~TransBFData() { 
  if (sigmaRad!=NULL) delete sigmaRad; 
  if (sigmaColl!=NULL) delete sigmaColl;
  if (omega!=NULL) delete omega;
}

// Load (YAML) data
void TransBFData::load(const YAML::Node& data) {
  string type;

  // Get transition energy
  //if (data["Xi"]) Xi=data["Xi"].as<double>();
  //else { cerr<<"TransBFData::load, Error: Could not get Xi\n"; exit(-1); }

  // If radiative cross-section available
  if (data["sigmaR"]) {

    // Get radiative cross-section data type
    if (data["sigmaR"]["type"]) type=data["sigmaR"]["type"].as<string>();
    else { cerr<<"TransBFData::load, Error: Could not get sigmaR::type\n"; exit(-1); }

    // Create radiative cross-section data object
    if (type.compare("value")==0) sigmaRad = new BFSigmaRadValue;
    else if (type.compare("table")==0) sigmaRad = new BFSigmaRadTable;
    else if (type.compare("plaw")==0) sigmaRad = new BFSigmaRadPLaw;
    else if (type.compare("plawGaunt")==0) sigmaRad = new BFSigmaRadPLawGaunt;
    else if (type.compare("hummer")==0) sigmaRad = new BFSigmaRadHummer;
    else if (type.compare("seaton")==0) sigmaRad = new BFSigmaRadSeaton;
    else if (type.compare("verner")==0) sigmaRad = new BFSigmaRadVerner;
    else if (type.compare("verner95")==0) sigmaRad = new BFSigmaRadVerner95;
    else if (type.compare("verner96")==0) sigmaRad = new BFSigmaRadVerner96;
    else if (type.compare("verner9596")==0) sigmaRad = new BFSigmaRadVerner9596;
    else { cerr<<"TransBFData::load, Error: sigmaR::type "<<type<<" not supported\n"; exit(-1); }

    // Load radiative cross-section data
    if (data["sigmaR"]["data"]) sigmaRad->load(data["sigmaR"]["data"]);
    else { cerr<<"TransBFData::load, Error: Could not get sigmaR::data\n"; exit(-1); }
  }

  // If collisional crosss-section available
  if (data["sigmaC"]) {

    // Get collisional cross-section data type
    if (data["sigmaC"]["type"]) type=data["sigmaC"]["type"].as<string>();
    else { cerr<<"TransBFData::load, Error: Could not get sigmaC::type\n"; exit(-1); }

    // Create collisional cross-section data object
    if (type.compare("table")==0) sigmaColl = new BFSigmaCollTable;
    else if (type.compare("plaw")==0) sigmaColl = new BFSigmaCollPLaw;
    else { cerr<<"TransBFData::load, Error: sigmaC::type "<<type<<" not supported\n"; exit(-1); }
    
    // Load collisional cross-section data
    if (data["sigmaC"]["data"]) sigmaColl->load(data["sigmaC"]["data"]);
    else { cerr<<"TransBFData::load, Error: Could not get sigmaC::data\n"; exit(-1); }
  }

  // If collisional strength available
  if (data["omega"]) {

    // Get collisional strength data type
    if (data["omega"]["type"]) type=data["omega"]["type"].as<string>();
    else { cerr<<"TransBFData::load, Error: Could not get omega::type\n"; exit(-1); }

    // Create collisional strength data object
    if (type.compare("value")==0) omega = new BFOmegaValue;
    else if (type.compare("table")==0) omega = new BFOmegaTable;
    else if (type.compare("plaw")==0) omega = new BFOmegaPLaw;
    else { cerr<<"TransBFData::load, Error: omega::type "<<type<<" not supported\n"; exit(-1); }
    
    // Load collisional strength data
    if (data["omega"]["data"]) omega->load(data["omega"]["data"]);
    else { cerr<<"TransBFData::load, Error: Could not get omega::data\n"; exit(-1); }
  }

  // Set radiative and collisional processes flags
  processes=(pRadiative|pCollisional);

  // If specified - Load detailed balance method
  if (data["dBMethod"]) dBMethod=loadDBMethod(data["dBMethod"]);

  // If specified - Load adjust edge flag
  if (data["edgeAdjust"]) edgeAdjust=(data["edgeAdjust"].as<string>().compare("yes")==0);
  else edgeAdjust=false;
}

// Update tables (using the provided frequency binning) - New!
void TransBFData::updateTables(const vector<double>& nu) {

  // If radiative cross-section available - resample table (if present)
  if (sigmaRad!=NULL) sigmaRad->resampleTable(nu,Xi/h);
}

// Adjust edge - New!
void TransBFData::adjustEdge() {

  // If requested and radiative cross-section available - Adjust edge
  if (sigmaRad!=NULL and edgeAdjust) sigmaRad->adjustEdge(Xi/h);
}

// Log data
void TransBFData::log() {
  int i;

  // Log index, energy, total energy and statistical weight
  cout<<"TransBFData::log: uppeState::index="<<upperState->index<<", lowerState::index="<<lowerState->index<<", Xi="<<Xi<<"\n";

  // Log radiative cross-section
  if (sigmaRad!=NULL) sigmaRad->log();
  else cout<<"TransBFData:log: sigmaRad not loaded\n";

  // Log collisional cross-section
  if (sigmaColl!=NULL) sigmaColl->log();
  else cout<<"TransBFData:log: sigmaColl not loaded\n";

  // Log collisional strength
  if (omega!=NULL) omega->log();
  else cout<<"TransBFData:log: omega not loaded\n";

  // Log detailed balance method
  logDBMethod(dBMethod);
}

// Get memory allocated - New!
double TransBFData::mem() const {
  double _mem=sizeof(*this);

  // Add memory allocated for radiative cross-section
  if (sigmaRad!=NULL) _mem+=sigmaRad->mem();

  // Add memory allocated for collisional cross-section
  if (sigmaColl!=NULL) _mem+=sigmaColl->mem();

  // Add memory allocated for collisional strength
  if (omega!=NULL) _mem+=omega->mem();

  return _mem;
}

// Constructor
BStateData::BStateData() : index(0), E(0), Etot(0), g(0), A(0), alfa(NULL), LTERefBStateIndex(-1), LTERefIStateIndex(-1), LTERefState(NULL), alfaCorrect(false), alfaFrac(-1), alfaSimple(false), upTransBB(), downTransBB(), upTransBF(), downTransBF() {}; // Changed!

// Copy constructor - New!
BStateData::BStateData(const BStateData& bStateData) : index(bStateData.index), E(bStateData.E), Etot(bStateData.Etot), g(bStateData.g), A(bStateData.A), alfa(NULL), LTERefBStateIndex(bStateData.LTERefBStateIndex), LTERefIStateIndex(bStateData.LTERefIStateIndex), LTERefState(bStateData.LTERefState), alfaCorrect(bStateData.alfaCorrect), alfaFrac(bStateData.alfaFrac), alfaSimple(bStateData.alfaSimple), upTransBB(bStateData.upTransBB), downTransBB(bStateData.downTransBB), upTransBF(bStateData.upTransBF), downTransBF(bStateData.downTransBF) { // Changed!

  // Copy recombination coefficient object
  if (bStateData.alfa!=NULL) alfa=bStateData.alfa->copy();
  else alfa=NULL;
}

// Assignment operator - New!
BStateData& BStateData::operator=(const BStateData& bStateData) {

  if (&bStateData!=this) {
   
    index=bStateData.index;
    E=bStateData.E; 
    Etot=bStateData.Etot;
    g=bStateData.g; 
    A=bStateData.A;
    LTERefBStateIndex=bStateData.LTERefBStateIndex; // Changed!
    LTERefIStateIndex=bStateData.LTERefIStateIndex; // New!
    LTERefState=bStateData.LTERefState; 
    alfaCorrect=bStateData.alfaCorrect; 
    alfaFrac=bStateData.alfaFrac; 
    alfaSimple=bStateData.alfaSimple; // Testing!
    upTransBB=bStateData.upTransBB; 
    downTransBB=bStateData.downTransBB; 
    upTransBF=bStateData.upTransBF; 
    downTransBF=bStateData.downTransBF;

    if (alfa!=NULL) delete alfa;

    // Copy recombination coefficient object
    if (bStateData.alfa!=NULL) alfa=bStateData.alfa->copy();
    else alfa=NULL;
  }
}

// Destructor - New!
BStateData::~BStateData() { 
  if (alfa!=NULL) delete alfa; 
}

// Load (YAML) data
void BStateData::load(const YAML::Node& data) {
  string type;

  // Get index
  if (data["i"]) index=data["i"].as<int>()-1;
  else { cerr<<"BStateData::load, Error: Could not get i\n"; exit(-1); }
  
  // Get energy
  if (data["E"]) E=data["E"].as<double>();
  else { cerr<<"BStateData::load, Error: Could not get E\n"; exit(-1); }

  // Get statistical weight
  if (data["g"]) g=data["g"].as<double>();
  else { cerr<<"BStateData::load, Error: Could not get g\n"; exit(-1); }

  // Get LTE reference state index
  if (data["LTERef"]) { // Block changed!
    if (data.size()==2) {
      LTERefIStateIndex=data["LTERef"][0].as<int>()-1;
      LTERefBStateIndex=data["LTERef"][1].as<int>()-1;
    }
    else LTERefBStateIndex=data["LTERef"].as<int>()-1;
  }

  // If recombination coefficient data avaiable - New!
  if (data["alfa"]) {

    // Get recombination coefficient data type
    if (data["alfa"]["type"]) type=data["alfa"]["type"].as<string>();
    else { cerr<<"BstateData::load, Error: Could not get alfa::type\n"; exit(-1); }

    // Create recombination coefficient data object
    if (type.compare("value")==0) alfa = new BFAlfaValue;
    else if (type.compare("table")==0) alfa = new BFAlfaTable;
    else if (type.compare("plaw")==0) alfa = new BFAlfaPLaw;
    else { cerr<<"BStateData::load, Error: alfa::type "<<type<<" not supported\n"; exit(-1); }

    // Load recombination coefficient  data
    if (data["alfa"]["data"]) alfa->load(data["alfa"]);
    else { cerr<<"BStateData::load, Error: Could not get alfa::data\n"; exit(-1); }
  }
}

// Log data
void BStateData::log() {
  int i;

  // Log index, energy, total energy, statistical weight and LTE reference indeces
  cout<<"BStateData::log: index="<<index<<", E="<<E<<", Etot="<<Etot<<", g="<<g<<", LTERefBStateIndex: "<<LTERefBStateIndex<<", LTERefIStateIndex: "<<LTERefIStateIndex<<"\n";

  // Log data for each upward bound-bound transition
  cout<<"BStateData::log: Upward bound-bound transitions\n";
  for (i=0;i<upTransBB.size();i++) upTransBB[i]->log();

  // Log data for each downward bound-bound transition
  cout<<"BStateData::log: Downward bound-bound transitions\n";
  for (i=0;i<downTransBB.size();i++) downTransBB[i]->log();

  // Log data for each upward bound-free transition
  cout<<"BStateData::log: Upward bound-free transitions\n";
  for (i=0;i<upTransBF.size();i++) upTransBF[i]->log();

  // Log data for each downward bound-free transition
  cout<<"BStateData::log: Downward bound-free transitions\n";
  for (i=0;i<downTransBF.size();i++) downTransBF[i]->log();

  // Log ecombination coefficient - New!
  if (alfa!=NULL) alfa->log();
  else cout<<"BStateData:log: alfa not loaded\n";
}

// Get memory allocated - New!
double BStateData::mem() const {
  double _mem=sizeof(*this);
  int i;

  // Add memory allocated for each upward bound-bound transition pointer
  for (i=0;i<upTransBB.size();i++) _mem+=sizeof(TransBBData*);

  // Add memory allocated for each downward bound-bound transition pointer
  for (i=0;i<downTransBB.size();i++) _mem+=sizeof(TransBBData*);

  // Add memory allocated for each upward bound-free transition pointer
  for (i=0;i<upTransBF.size();i++) _mem+=sizeof(TransBFData*);

  // Add memory allocated for each downward bound-free transition pointer
  for (i=0;i<downTransBF.size();i++) _mem+=sizeof(TransBFData*);  

  // Add memory allocated for recombination coefficient
  if (alfa!=NULL) _mem+=alfa->mem();

  return _mem;
}

// Load recombination distribution method
AlfaDistMethod atomdata::loadAlfaDistMethod(const YAML::Node& data) {
  AlfaDistMethod alfaDistMethod;

  if (data.as<string>().compare("gReplace")==0) alfaDistMethod=gReplace;
  else alfaDistMethod=gAdd;

  return alfaDistMethod;
}

// Constructor
IStateData::IStateData() : alfaDistMethod(gAdd), alfaForceEqual(false), alfa(NULL), alfaDi(NULL), LTERefBStateIndex(-1), LTERefIStateIndex(-1) { } // Changed!

// Copy constructor
IStateData::IStateData(const IStateData& iStateData) : C(iStateData.C), index(iStateData.index), Xi0(iStateData.Xi0), alfaDistMethod(iStateData.alfaDistMethod), alfaForceEqual(iStateData.alfaForceEqual), LTERefBStateIndex(iStateData.LTERefBStateIndex), LTERefIStateIndex(iStateData.LTERefIStateIndex), bState(iStateData.bState), transBB(iStateData.transBB), alfa(NULL), alfaDi(NULL) { // Changed!
  int i,j;

  // Update the LTE reference state pointers - Removed!
  /*for (i=0;i<bState.size();i++) {
    if (bState[i].LTERefState!=NULL) bState[i].LTERefState=&(bState[iStateData.bState[i].LTERefState->index]);
    else bState[i].LTERefState=NULL;
    }*/

  // Update the bound-bound transition list bound state pointers
  for (i=0;i<transBB.size();i++) {
    transBB[i].upperState=&(bState[iStateData.transBB[i].upperState->index]);
    transBB[i].lowerState=&(bState[iStateData.transBB[i].lowerState->index]);
  }

  // Update the upward and downward bound-bound transition list transition pointers
  for (i=0;i<bState.size();i++) {
    for (j=0;j<bState[i].downTransBB.size();j++) bState[i].downTransBB[j]=&(transBB[iStateData.bState[i].downTransBB[j]->index]);
    for (j=0;j<bState[i].upTransBB.size();j++) bState[i].upTransBB[j]=&(transBB[iStateData.bState[i].upTransBB[j]->index]);
  }

  // Copy recombination coefficient object - New!
  if (iStateData.alfa!=NULL) alfa=iStateData.alfa->copy();
  else alfa=NULL;

  // Copy di-electronic recombination coefficient object - New!
  if (iStateData.alfaDi!=NULL) alfaDi=iStateData.alfaDi->copy();
  else alfaDi=NULL;
}

// Assignment operator
IStateData& IStateData::operator=(const IStateData& iStateData) {
  int i,j;

  if (&iStateData!=this) {
    C=iStateData.C;
    index=iStateData.index;
    Xi0=iStateData.Xi0;
    alfaDistMethod=iStateData.alfaDistMethod; // New!
    alfaForceEqual=iStateData.alfaForceEqual; // New!
    LTERefBStateIndex=iStateData.LTERefBStateIndex; // New!
    LTERefIStateIndex=iStateData.LTERefIStateIndex; // New!
    bState=iStateData.bState;
    transBB=iStateData.transBB;

    // Update the LTE reference state pointers - Removed!
    /*for (i=0;i<bState.size();i++) {
      if (bState[i].LTERefState!=NULL) bState[i].LTERefState=&(bState[iStateData.bState[i].LTERefState->index]);
      else bState[i].LTERefState=NULL;
      }*/

    // Update the bound-bound transition list bound state pointers
    for (i=0;i<transBB.size();i++) {
      transBB[i].upperState=&(bState[iStateData.transBB[i].upperState->index]);
      transBB[i].lowerState=&(bState[iStateData.transBB[i].lowerState->index]);
    }

    // Update the upward and downward bound-bound transition list transition pointers
    for (i=0;i<bState.size();i++) {
      for (j=0;j<bState[i].downTransBB.size();j++) bState[i].downTransBB[j]=&(transBB[iStateData.bState[i].downTransBB[j]->index]);
      for (j=0;j<bState[i].upTransBB.size();j++) bState[i].upTransBB[j]=&(transBB[iStateData.bState[i].upTransBB[j]->index]);
    }

    // If present - Delete recombination coefficient object - New!
    if (alfa!=NULL) delete alfa;

    // Copy recombination coefficient object - New!
    if (iStateData.alfa!=NULL) alfa=iStateData.alfa->copy();
    else alfa=NULL;

    // If present - Delete di-electronic recombination coefficient object - New!
    if (alfaDi!=NULL) delete alfaDi;

    // Copy di-electronic recombination coefficient object - New!
    if (iStateData.alfaDi!=NULL) alfaDi=iStateData.alfaDi->copy();
    else alfaDi=NULL;
  }
}

// Load (YAML) data
void IStateData::load(const YAML::Node& data, const string& dataDir) {
  YAML::Node bStateData, transBBData, transBB2Data, alfaData, alfaDiData; // Changed!
  string bStateDataFile,transBBDataFile,transBB2DataFile,alfaDataFile,alfaDiDataFile,type; // Changed!
  int bStateIndex,upperStateIndex,lowerStateIndex,alfaCorrMin,i,j;

  // Get C
  if (data["C"]) C=data["C"].as<double>();
  else { cerr<<"IStateData::load, Error: Could not get C\n"; exit(-1); }  

  // Get Xi0
  if (data["Xi0"]) Xi0=data["Xi0"].as<double>();
  else { cerr<<"IStateData::load, Error: Could not get Xi0\n"; exit(-1); }

  // Get bound state data file
  if (data["bState"]) bStateDataFile=getDataFile(dataDir,data["bState"].as<string>());
  else { cerr<<"IStateData::load, Error: Could not get bState\n"; exit(-1); }  // Changed!

  // Get LTE reference state index - New!
  if (data["LTERef"]) {
    if (data["LTERef"].size()==2) {
      LTERefIStateIndex=data["LTERef"][0].as<int>()-1;
      LTERefBStateIndex=data["LTERef"][1].as<int>()-1;
    }
    else { cerr<<"IStateData::load, Error: Could not get LTERef\n"; exit(-1); }
  }

  // Load bound state data from file
  bStateData=YAML::LoadFile(bStateDataFile);

  // Resize bound state list
  bState.resize(bStateData.size());

  // Load data for each bound state
  for (i=0;i<bStateData.size();i++) bState[i].load(bStateData[i]);

  // For each bound state - Removed!
  /*for (i=0;i<bState.size();i++) {
  // If LTE reference state index set - Set pointer
  if (bState[i].LTERefIndex!=-1) bState[i].LTERefState=&(bState[bState[i].LTERefIndex]);
  }*/

  cout<<"IStateData::load: Loaded "<<i<<" bound states for ionization state "<<index<<"\n";

  // Get bound-bound transition data file
  if (data["transBB"]) transBBDataFile=getDataFile(dataDir,data["transBB"].as<string>());
  else { cerr<<"IStateData::load, Error: Could not get transBB\n"; exit(-1); }  

  // Load bound-bound transition data from file
  transBBData=YAML::LoadFile(transBBDataFile);

  // Resize bound-bound transition list
  transBB.resize(transBBData.size());

  // For each bound-bound transition
  for (i=0;i<transBBData.size();i++) {

    // Set bound-bound transition index
    transBB[i].index=i;

    // Load transition data
    transBB[i].load(transBBData[i]);

    // Get upper state index
    if (transBBData[i]["j"]) bStateIndex=transBBData[i]["j"].as<int>();
    else { cerr<<"IStateData::load, Error: Could not get transBB::j\n"; exit(-1); }     
 
    // Set upper state pointer
    transBB[i].upperState=&(bState[bStateIndex-1]);
    // Add bound-bound transition to upper state downward bound-bound transition list
    bState[bStateIndex-1].downTransBB.push_back(&(transBB[i]));

    // Get lower state index
    if (transBBData[i]["i"]) bStateIndex=transBBData[i]["i"].as<int>();
    else { cerr<<"IStateData::load, Error: Could not get transBB::i\n"; exit(-1); }         

    // Set lower state pointer
    transBB[i].lowerState=&(bState[bStateIndex-1]);
    // Add bound-bound transition to lower state upward bound-bound transition list
    bState[bStateIndex-1].upTransBB.push_back(&(transBB[i]));

    // Add to upper state cumulative A value
    transBB[i].upperState->A+=transBB[i].A;
  }

  cout<<"IStateData::load: Loaded "<<i<<" bound-bound transitions for ionization state "<<index<<"\n";

  // If recombination coefficient data avaiable - New!
  if (data["alfa"]) {

    // Get recombination coefficient data file
    alfaDataFile=getDataFile(dataDir,data["alfa"].as<string>());

    // Load recombination coefficient data from file
    alfaData=YAML::LoadFile(alfaDataFile);

    // Get recombination coefficient data type
    if (alfaData["type"]) type=alfaData["type"].as<string>();
    else { cerr<<"IstateData::load, Error: Could not get alfa::type\n"; exit(-1); }

    // Create recombination coefficient data object
    if (type.compare("value")==0) alfa = new BFAlfaValue;
    else if (type.compare("table")==0) alfa = new BFAlfaTable;
    else if (type.compare("plaw")==0) alfa = new BFAlfaPLaw;
    else if (type.compare("verner")==0) alfa = new BFAlfaVerner;
    else if (type.compare("shull")==0) alfa = new BFAlfaShull;
    else if (type.compare("pequignot")==0) alfa = new BFAlfaPequignot;
    else if (type.compare("badnell")==0) alfa = new BFAlfaBadnell;
    else { cerr<<"IStateData::load, Error: alfa::type "<<type<<" not supported\n"; exit(-1); }

    // Load recombination coefficient  data
    if (alfaData["data"]) alfa->load(alfaData);
    else { cerr<<"IStateData::load, Error: Could not get alfa::data\n"; exit(-1); }

    cout<<"IStateData::load: Loaded recombination coefficient for ionization state "<<index<<"\n";
  }

  // If di-electronic recombination coefficient data avaiable - New!
  if (data["alfaDi"]) {

    // Get di-electionic recombination coefficient data file
    alfaDiDataFile=getDataFile(dataDir,data["alfaDi"].as<string>());

    // Load di-electronic recombination coefficient data from file
    alfaDiData=YAML::LoadFile(alfaDiDataFile);

    // Get di-electronic recombination coefficient data type
    if (alfaDiData["type"]) type=alfaDiData["type"].as<string>();
    else { cerr<<"IstateData::load, Error: Could not get alfaDi::type\n"; exit(-1); }

    // Create di-electronic recombination coefficient data object
    if (type.compare("value")==0) alfaDi = new BFAlfaValue;
    else if (type.compare("table")==0) alfaDi = new BFAlfaTable;
    else if (type.compare("plaw")==0) alfaDi = new BFAlfaPLaw;
    else if (type.compare("shull")==0) alfaDi = new BFAlfaShull;
    else if (type.compare("nussbaumer")==0) alfaDi = new BFAlfaNussbaumer;
    else { cerr<<"IStateData::load, Error: alfaDi::type "<<type<<" not supported\n"; exit(-1); }

    // Load di-electronic recombination coefficient data
    if (alfaDiData["data"]) alfaDi->load(alfaDiData);
    else { cerr<<"IStateData::load, Error: Could not get alfaDi::data\n"; exit(-1); }

    cout<<"IStateData::load: Loaded di-electronic recombination coefficient for ionization state "<<index<<"\n";
  }
  
  // If recombination distribution data available
  if (data["alfaDist"]) {
  
    // Load recombination distribution data
    loadAlfaDist(data["alfaDist"]);
  }

  // If two-photon transition data avaiable - New!
  if (data["transBB2"]) {

    // Get two-photon transition data file
    transBB2DataFile=getDataFile(dataDir,data["transBB2"].as<string>());

    // Load two-photon transition data from file
    transBB2Data=YAML::LoadFile(transBB2DataFile);

    // For each two-photon transition
    for (i=0;i<transBB2Data.size();i++) {

      // Get upper state index
      if (transBB2Data[i]["j"]) upperStateIndex=transBB2Data[i]["j"].as<int>()-1;
      else { cerr<<"IStateData::load, Error: Could not get transBB2::j\n"; exit(-1); }     

      // Get lower state index
      if (transBB2Data[i]["i"]) lowerStateIndex=transBB2Data[i]["i"].as<int>()-1;
      else { cerr<<"IStateData::load, Error: Could not get transBB2::i\n"; exit(-1); }         

      // If upper state index within range
      if (upperStateIndex<bState.size()) {
	// For each (downward) bound-bound transition from the upper state
	for (j=0;j<bState[upperStateIndex].downTransBB.size();j++) {
	  // If match
	  if (lowerStateIndex==bState[upperStateIndex].downTransBB[i]->lowerState->index) {
	    // Load two-photon transition data
	    bState[upperStateIndex].downTransBB[i]->load2(transBB2Data[i]);
	    // Break
	    break; 
	  }
	}
	if (j==bState[upperStateIndex].downTransBB.size()) { cerr<<"IStateData::load, Error: Could not find two-photon transition\n"; exit(-1); }
      }
      else { cerr<<"IStateData::load, Error: Could not find two-photon transition\n"; exit(-1); } 
    }

    cout<<"IStateData::load: Loaded "<<i<<" two-photon transitions for ionization state "<<index<<"\n";
  }

  // Adjust edge for each bound-bound (two-photon) transition
  for (i=0;i<transBB.size();i++) transBB[i].adjustEdge();

  cout<<"IStateData::load: Loaded ionization state "<<index<<"\n";
}

// Load (YAML) data
void IStateData::loadAlfaDist(const YAML::Node& data) {
  bool allWithNoAlfa=false;
  int i,j,iMin=0,iMax=bState.size()-1;
  // Testing!
  double alfaSimple=bState.size();
  
  if (data["method"]) alfaDistMethod=loadAlfaDistMethod(data["method"]);
  
  if (data["select"]) allWithNoAlfa=(data["select"].as<string>().compare("allWithNoAlfa")==0);

  if (data["forceEqual"]) alfaForceEqual=(data["forceEqual"].as<string>().compare("yes")==0);
  else alfaForceEqual=false;
   
  if (data["iMin"]) iMin=MAX(data["iMin"].as<int>()-1,0);
  if (data["iMax"]) iMax=MIN(data["iMax"].as<int>()-1,int(bState.size())-1);

  for (i=iMin;i<=iMax;i++) 
    if ((not allWithNoAlfa) or bState[i].alfa==NULL) bState[i].alfaCorrect=true;
  
  if (data["fractions"])
    for (i=0;i<data["fractions"].size();i++) {
      if (data["fractions"][i].size()==2) { 
	j=data["fractions"][i][0].as<int>()-1;
	if (j>=0 and j<bState.size()) bState[j].alfaFrac=data["fractions"][i][1].as<double>();
      }
    }

  // Testing!
  if (data["simple"]) alfaSimple=MAX(data["simple"].as<int>()-1,0);
  for (i=alfaSimple;i<bState.size();i++) bState[i].alfaSimple=true;
}

// Log data
void IStateData::log() {
  int i;

  // Log C and Xi0
  cout<<"IStateData::log: C="<<C<<", Xi0="<<Xi0<<"\n";

  // Log data for each bound state
  for (i=0;i<bState.size();i++) bState[i].log();

  // Log recombination coeffcient
  if (alfa!=NULL) alfa->log();
  else cout<<"IStateData:log: alfa not loaded\n";
}

// Get memory allocated - New!
double IStateData::mem() const {
  double _mem=sizeof(*this);
  int i;

  // Add memory allocated for each bound state
  for (i=0;i<bState.size();i++) _mem+=bState[i].mem();

  // Add memory allocated for each bound-bound transition
  for (i=0;i<transBB.size();i++) _mem+=transBB[i].mem();

  // Add memory allocated for recombination coefficient
  if (alfa!=NULL) _mem+=alfa->mem();

  // Add memory allocated for die-electronic recombination coefficient
  if (alfaDi!=NULL) _mem+=alfaDi->mem();

  return _mem;
}

// Find ionization state data
IStateData* IStateDataList::find(const int& C) {
  int i;
  for (i=0;i<(*this).size();i++) {
    if ((*this)[i].C==C) return &((*this)[i]);
  }
  return NULL;
}

// Load (YAML) data
void IsotopeData::load(const YAML::Node& data) {

  // Get N
  N=data["N"].as<double>();
  // Error handling ?
  
  // Get m
  m=data["m"].as<double>()*u;
  // Error handling ?
}

// Log data
void IsotopeData::log() {

  // Log N and m
  cout<<"IsotopeData::log: N="<<N<<", m="<<m<<"\n";
}

// Get memory allocated - New!
double IsotopeData::mem() const {
  double _mem=sizeof(*this);

  return _mem;
}

// Find isotope data
IsotopeData* IsotopeDataList::find(const int& N) {
  int i;
  for (i=0;i<(*this).size();i++) {
    if ((*this)[i].N==N) return &((*this)[i]);
  }
  return NULL;
}

// Constructor
AtomData::AtomData() { }

// Copy constructor
AtomData::AtomData(const AtomData& atomData) : index(atomData.index), Z(atomData.Z), isotope(atomData.isotope), iState(atomData.iState), transBF(atomData.transBF) {
  int i,j,k;

  //cout<<"AtomData::AtomData: index: "<<index<<", Z: "<<Z<<", iState.size: "<<iState.size()<<", iState.size: "<<atomData.iState.size()<<"\n";

  // Update the LTE reference state pointers - New!
  for (i=0;i<iState.size();i++)
    for (j=0;j<iState[i].bState.size();j++) {
      if (iState[i].bState[j].LTERefState!=NULL)
	iState[i].bState[j].LTERefState=&(iState[iState[i].bState[j].LTERefIStateIndex].bState[iState[i].bState[j].LTERefBStateIndex]);
      else iState[i].bState[j].LTERefState=NULL;
    }

  // Update the upward and downward bound-free transition list transition pointers
  for (i=0;i<iState.size();i++)
    for (j=0;j<iState[i].bState.size();j++) {
      for (k=0;k<iState[i].bState[j].downTransBF.size();k++) iState[i].bState[j].downTransBF[k]=&(transBF[atomData.iState[i].bState[j].downTransBF[k]->index]);
      for (k=0;k<iState[i].bState[j].upTransBF.size();k++) iState[i].bState[j].upTransBF[k]=&(transBF[atomData.iState[i].bState[j].upTransBF[k]->index]);
    }

  // Update the bound-free transition list bound state pointers
  for (i=0;i<iState.size();i++)
    for (j=0;j<iState[i].bState.size();j++) {
      for (k=0;k<iState[i].bState[j].downTransBF.size();k++) iState[i].bState[j].downTransBF[k]->upperState=&(iState[i].bState[j]);
      for (k=0;k<iState[i].bState[j].upTransBF.size();k++) iState[i].bState[j].upTransBF[k]->lowerState=&(iState[i].bState[j]);
    }
}

// Assignment operator
AtomData& AtomData::operator=(const AtomData& atomData) {
  int i,j,k;

  if (&atomData!=this) {
    index=atomData.index;
    Z=atomData.Z;
    isotope=atomData.isotope;
    iState=atomData.iState;
    transBF=atomData.transBF;

    //cout<<"AtomData::operator=: index: "<<index<<", Z: "<<Z<<", iState.size: "<<iState.size()<<", iState.size: "<<atomData.iState.size()<<"\n";

    // Update the LTE reference state pointers - New!
    for (i=0;i<iState.size();i++)
      for (j=0;j<iState[i].bState.size();j++) {
	if (iState[i].bState[j].LTERefState!=NULL)
	  iState[i].bState[j].LTERefState=&(iState[iState[i].bState[j].LTERefIStateIndex].bState[iState[i].bState[j].LTERefBStateIndex]);
	else iState[i].bState[j].LTERefState=NULL;
      }

    // Update the upward and downward bound-free transition list transition pointers
    for (i=0;i<iState.size();i++) 
      for (j=0;j<iState[i].bState.size();j++) {
	for (k=0;k<iState[i].bState[j].downTransBF.size();k++) iState[i].bState[j].downTransBF[k]=&(transBF[atomData.iState[i].bState[j].downTransBF[k]->index]);
	for (k=0;k<iState[i].bState[j].upTransBF.size();k++) iState[i].bState[j].upTransBF[k]=&(transBF[atomData.iState[i].bState[j].upTransBF[k]->index]);
      }

    // Update the bound-free transition list bound state pointers
    for (i=0;i<iState.size();i++) {
      //cout<<"AtomData::operator=: i: "<<i<<"\n";
      for (j=0;j<iState[i].bState.size();j++) {
	//cout<<"AtomData::operator=: j: "<<j<<"\n";
	for (k=0;k<iState[i].bState[j].downTransBF.size();k++) {
	  //cout<<"AtomData::operator=: down: i: "<<i<<", j: "<<j<<", k: "<<k<<"\n";
	  iState[i].bState[j].downTransBF[k]->upperState=&(iState[i].bState[j]);
	}
	for (k=0;k<iState[i].bState[j].upTransBF.size();k++) {
	  //cout<<"AtomData::operator=: up: i: "<<i<<", j: "<<j<<", k: "<<k<<"\n";
	  iState[i].bState[j].upTransBF[k]->lowerState=&(iState[i].bState[j]);
	}
      }
    }
  }
}

// Update energy above ground state of neutral atom
void AtomData::updateETot() {
  double ESmall,Xi0Sum=0;
  int i,j;
  
  // Hack - Introduce small shift of energy scale to avoid zero ground state energy
  ESmall=h*(c/(1e6*AA));

  // For each ionization state
  for (i=0;i<iState.size();i++) {
    // Sum ionization energy
    if (i>0) Xi0Sum+=iState[i-1].Xi0;
    // For each bound state
    for (j=0;j<iState[i].bState.size();j++) {
      // Calculate energy above ground state of neutral atom
      iState[i].bState[j].Etot=iState[i].bState[j].E+Xi0Sum+ESmall;
    }
  }
}

// Set LTE reference state pointers
void AtomData::setLTERefState() {
  int LTERefBStateIndex,LTERefIStateIndex,i,j;

  // For each ionization state
  for (i=0;i<iState.size();i++) {

    // For each bound state
    for (j=0;j<iState[i].bState.size();j++) {
      
      LTERefBStateIndex=-1;
      LTERefIStateIndex=-1;
      
      // If LTE reference indeces specified for ionization state - Get them
      if (iState[i].LTERefBStateIndex!=-1) {
	LTERefBStateIndex=iState[i].LTERefBStateIndex;
	LTERefIStateIndex=iState[i].LTERefIStateIndex;
      }
      // Else
      else {
	
	// If LTE reference indeces specified for bound state - Get them
	if (iState[i].bState[j].LTERefBStateIndex!=-1) {
	  LTERefBStateIndex=iState[i].bState[j].LTERefBStateIndex;
	  LTERefIStateIndex=iState[i].index;
	}
      }
      
      // If LTE reference indeces specified
      if (LTERefBStateIndex!=-1) {
	// Set LTE reference pointer
	if (LTERefIStateIndex!=-1) iState[i].bState[j].LTERefState=&(iState[LTERefIStateIndex].bState[LTERefBStateIndex]);
	else iState[i].bState[j].LTERefState=&(iState[i].bState[LTERefBStateIndex]);
	// Make sure (both) LTE reference indeces are set
	iState[i].bState[j].LTERefBStateIndex=LTERefBStateIndex;
	iState[i].bState[j].LTERefIStateIndex=LTERefIStateIndex;
      }
      // Otherwise - Set to NULL
      else iState[i].bState[j].LTERefState=NULL;

      // If self-reference - Set to NULL
      if (iState[i].bState[j].LTERefState==&(iState[i].bState[j])) iState[i].bState[j].LTERefState=NULL;

      //cout<<"AtomData::setLTERefState: Z: "<<Z<<", C: "<<iState[i].C<<", index: "<<iState[i].bState[j].index<<", LTERefState: "<<iState[i].bState[j].LTERefState<<", LTERefBStateIndex: "<<LTERefBStateIndex<<", LTERefIStateIndex: "<<LTERefIStateIndex<<"\n";
    }
  }
}

// Load (YAML) data
void AtomData::load(const YAML::Node& data, const string& dataDir) {
  YAML::Node isotopeData,iStateData,transBFData;
  string isotopeDataFile,iStateDataFile,transBFDataFile;
  double CMin=1e99; // New!
  int iStateIndex,bStateIndex,iStateIndexMin,i; // Changed!

  cout<<"data[Z]: "<<typeName(data["Z"])<<", size: "<<data["Z"].size()<<"\n";

  // Get Z from linestream
  if (data["Z"]) Z=data["Z"].as<double>();
  else { cerr<<"AtomData::load, Error: Could not get Z\n"; exit(-1); }

  // Get isotope data file
  if (data["isotope"]) isotopeDataFile=getDataFile(dataDir,data["isotope"].as<string>());
  else { cerr<<"AtomData::load, Error: Could not get isotope\n"; exit(-1); }  

  // Load isotope data from file
  isotopeData=YAML::LoadFile(isotopeDataFile);
 
  // Resize isotope list
  isotope.resize(isotopeData.size());

  // For each isotope
  for (i=0;i<isotopeData.size();i++) {

    // Set isotope index
    isotope[i].index=i;

    // Load isotope data
    isotope[i].load(isotopeData[i]);
  }

  cout<<"AtomData::load: Loaded "<<i<<" isotopes for atom Z="<<Z<<"\n";

  // Get ionization data file
  if (data["iState"]) iStateDataFile=getDataFile(dataDir,data["iState"].as<string>());
  else { cerr<<"AtomData::load, Error: Could not get iState\n"; exit(-1); }  

  // Load ionization data from file
  iStateData=YAML::LoadFile(iStateDataFile);

  // Resize ionization state list
  iState.resize(iStateData.size()+1);

  // For each ionization state
  for (i=0;i<iStateData.size();i++) {
    
    // Set ionization state index
    iState[i].index=i;

    // Load ionization state data
    iState[i].load(iStateData[i],getDataDir(iStateDataFile));

    // Determine minimum abs(C) - New!
    if (abs(iState[i].C)<CMin) CMin=abs(iState[i].C);
  }

  // Determine minimum iState index - New!
  iStateIndexMin=CMin;

  cout<<"AtomData::load: Loaded "<<i<<" ionization states for atom Z="<<Z<<"\n";

  // Add one further ionization state to list
  iState[i].C=iState[i-1].C-1;
  iState[i].Xi0=0;
  iState[i].index=i;
  iState[i].bState.resize(1);
  iState[i].bState[0].index=0;
  iState[i].bState[0].E=0;
  if (data["gIon"]) iState[i].bState[0].g=data["gIon"].as<double>();
  else iState[i].bState[0].g=1;
  iState[i].bState[0].A=0;
  iState[i].LTERefBStateIndex=iState[i-1].LTERefBStateIndex; // New!
  iState[i].LTERefIStateIndex=iState[i-1].LTERefIStateIndex; // New!

  // Set LTE reference state pointers - New!
  setLTERefState();

  // If bound-free transition data available
  if (data["transBF"]) {

    // Get bound-free transition data file
    transBFDataFile=getDataFile(dataDir,data["transBF"].as<string>());

    // Load bound-free transition data from file
    transBFData=YAML::LoadFile(transBFDataFile);

    // Resize bound-free transition list
    transBF.resize(transBFData.size());

    // For each bound-free transition
    for (i=0;i<transBFData.size();i++) {

      // Set bound-free transition index
      transBF[i].index=i;
    
      // Load transition data
      transBF[i].load(transBFData[i]);

      // Get upper ionization and bound state indices
      if (transBFData[i]["j"])
	if (transBFData[i]["j"].size()==2) {
	  iStateIndex=transBFData[i]["j"][0].as<int>()-iStateIndexMin; // Changed!
	  bStateIndex=transBFData[i]["j"][1].as<int>();
	}
	else { cerr<<"AtomData::load, Error: Could not get transBF::j\n"; exit(-1); }  
      else { cerr<<"AtomData::load, Error: Could not get transBF::j\n"; exit(-1); }  
 
      // Set upper state pointer
      transBF[i].upperState=&(iState[iStateIndex-1].bState[bStateIndex-1]);
      // Add bound-free transition to upper state downward bound-bound transition list
      iState[iStateIndex-1].bState[bStateIndex-1].downTransBF.push_back(&(transBF[i]));

      // Get lower ionization and bound state indices
      if (transBFData[i]["i"])
	if (transBFData[i]["i"].size()==2) {
	  iStateIndex=transBFData[i]["i"][0].as<int>()-iStateIndexMin; // Changed!
	  bStateIndex=transBFData[i]["i"][1].as<int>();
	}
	else { cerr<<"AtomData::load, Error: Could not get transBF::i\n"; exit(-1); }  
      else { cerr<<"AtomData::load, Error: Could not get transBF::i\n"; exit(-1); }  
                
      // Set lower state pointer
      transBF[i].lowerState=&(iState[iStateIndex-1].bState[bStateIndex-1]);
      // Add bound-free transition to lower state upward bound-bound transition list
      iState[iStateIndex-1].bState[bStateIndex-1].upTransBF.push_back(&(transBF[i]));

      // Update transition energy (only works for succesive ionization states)
      transBF[i].Xi=iState[iStateIndex-1].Xi0+(transBF[i].upperState->E-transBF[i].lowerState->E);
    }

    cout<<"AtomData::load: Loaded "<<i<<" bound-free transitions for atom Z="<<Z<<"\n";
  }

  // Update energy above ground state of neutral atom
  updateETot();

  // Adjust edge for each bound-free transition
  for (i=0;i<transBF.size();i++) transBF[i].adjustEdge();

  cout<<"AtomData::load: Loaded atom with Z="<<Z<<"\n";
}

// Log data
void AtomData::log() {
  int i;

  // Log Z
  cout<<"AtomData::log: Z="<<Z<<"\n";

  // Log data for each isotope
  for (i=0;i<isotope.size();i++) isotope[i].log();

  // Log data for each ionization date
  for (i=0;i<iState.size();i++) iState[i].log();
}

// Get bound-bound collisional strength (value)
double CTAlfaValue::operator ()(double T) {
  return alfa;
}

// Get memory allocated - New!
double AtomData::mem() const {
  double _mem=sizeof(*this);
  int i;

  // Add memory allocated for each ionization state
  for (i=0;i<iState.size();i++) _mem+=iState[i].mem();

  // Add memory allocated for each isotope
  for (i=0;i<isotope.size();i++) _mem+=isotope[i].mem();

  // Add memory allocated for each bound-free transition
  for (i=0;i<transBF.size();i++) _mem+=transBF[i].mem();

  return _mem;
}

// Load (YAML) data
void CTAlfaValue::load(const YAML::Node& data) {

  // Get reaction coefficient
  alfa=data["data"].as<double>();
}

// Log data
void CTAlfaValue::log() {

  // Log 
  cout<<"CTAlfaValue::log: alfa="<<alfa<<"\n";
}

// Create copy
CTAlfa* CTAlfaValue::copy() {
  return new CTAlfaValue(*this);
}

// Get memory allocated - New!
double CTAlfaValue::mem() const {
  double _mem=sizeof(*this);
 
  return _mem;
}

// Get charge-transfer reaction coeffiecient (table) - New!
double CTAlfaTable::operator ()(double T) {
  double x;
  
  // Transform interpolation quantity according to interpolation type
  if (ipolType==ipolLog) x=log10(T);
  else x=T;

  // If constant extrapolation - Restrict interpolation quantity to domain
  if (epolType==epolConst) x=MIN(MAX(x,alfa.domain().min),alfa.domain().max);
  // Else - No extrapolation

  // Return interpolated (or extrapolated) recombination coefficient
  return alfa(x);
}

// Load (YAML) data- New!
void CTAlfaTable::load(const YAML::Node& data) {
  vector<FunctionValue> table(data["data"].size());
  int i;

  // If provided- Get interpolation type
  if (data["ipolType"]) ipolType=loadIPolType(data["ipolType"]);
  // Else - Use linear interpolation
  else ipolType=ipolLinear;

  // If provided - Get extrapolation type
  if (data["epolType"]) epolType=loadEPolType(data["epolType"]);
  // Else - Use no extrapolation
  else epolType=epolNone;

  // Get table data
  for (i=0;i<data["data"].size();i++) {
    if (data["data"][i][0]) table[i].x=data["data"][i][0].as<double>();
    else { cerr<<"CTAlfaTable::load, Error: Could not get data\n"; exit(-1); }
    if (data["data"][i][1]) table[i].f=data["data"][i][1].as<double>();
    else { cerr<<"CTAlfaTable::load, Error: Could not get data\n"; exit(-1); }
    // Transform interpolation quantity according to interpolation type
    if (ipolType==ipolLog) table[i].x=log10(table[i].x);
  }

  // Initiate interpolator
  alfa.init(table);
}

// Log data - New!
void CTAlfaTable::log() {
  int i;

  // Log 
  cout<<"CTAlfaTable::log: Table loaded, ipolType: "<<logIPolType(ipolType)<<", epolType: "<<logEPolType(epolType)<<"\n";
}

// Create copy - New!
CTAlfa* CTAlfaTable::copy() {
  return new CTAlfaTable(*this);
}

// Get memory allocated - New!
double CTAlfaTable::mem() const {
  double _mem=sizeof(*this);
 
  // Add memory allocated for table
  _mem+=alfa.mem();

  return _mem;
}

// Get charge-transfer reaction coefficient (powerlaw)
double CTAlfaPLaw::operator ()(double T) {
  return alfa*pow(T/(T0),p);
}

// Load (YAML) data
void CTAlfaPLaw::load(const YAML::Node& data) {
  
  // Get charge-transfer reaction coefficient
  if (data["data"][0]) alfa=data["data"][0].as<double>();
  else { cerr<<"CTAlfaPLaw::load, Error: Could not get data\n"; exit(-1); }  

  // Get temperature
  if (data["data"][1]) T0=data["data"][1].as<double>();
  else { cerr<<"CTAlfaPLaw::load, Error: Could not get data\n"; exit(-1); }  

  // Get exponent
  if (data["data"][2]) p=data["data"][2].as<double>();
  else { cerr<<"CTAlfaPLaw::load, Error: Could not get data\n"; exit(-1); }  
}

// Log data
void CTAlfaPLaw::log() {
  int i;

  // Log 
  cout<<"CTAlfaPlaw::log: alfa="<<alfa<<", T0="<<T0<<", p="<<p<<"\n";
}

// Create copy
CTAlfa* CTAlfaPLaw::copy() {
  return new CTAlfaPLaw(*this);
}

// Get memory allocated - New!
double CTAlfaPLaw::mem() const {
  double _mem=sizeof(*this);
 
  return _mem;
}

// Log Arnaud et al. 1985 fit type
string atomdata::logArnaudFitType(ArnaudFitType fitType) {
  switch(fitType) {
  case fitTypeA : return "fitTypeA";
  case fitTypeB : return "fitTypeB";
  case fitTypeOH : return "fitTypeOH";
  default : return "unknown";
  }
}

// Get charge-transfer reaction coefficient (Arnaud fit)
double CTAlfaArnaud::operator ()(double T) {

  if (fitType==fitTypeA) return (a*1e-15)*pow(T/1e4,b)*(1.0+c*exp(d*(T/1e4)));
  else if (fitType==fitTypeB) return (a*1e-15)*pow(T/1e4,b)*exp(-c*(T/1e4))*exp(-(deltaE*eV)/(k*T));
  else return (a*1e-15)*exp(-deltaE/(k*T))*(1-0.93*exp(-c*(T/1e4)));
}

// Load (YAML) data
void CTAlfaArnaud::load(const YAML::Node& data) {
  
  // Get fit type
  if (data["fitType"]) {
    if (data["fitType"].as<string>().compare("A")==0) fitType=fitTypeA;
    else if (data["fitType"].as<string>().compare("B")==0) fitType=fitTypeB;
    else if (data["fitType"].as<string>().compare("OH")==0) fitType=fitTypeOH;
    else { cerr<<"CTAlfaArnaud::load, Error: fitType "<<data["fitType"].as<string>()<<" not supported\n"; exit(-1); }
  }
  else { cerr<<"CTAlfaArnaud::load, Error: Could not get fitType\n"; exit(-1); }  

  // Get fit parameters
  if (data["data"].size()==4) { 
    a=data["data"][0].as<double>(); 
    b=data["data"][1].as<double>(); 
    c=data["data"][2].as<double>();
    if (fitType==fitTypeA) d=data["data"][3].as<double>();
    else deltaE=data["data"][3].as<double>();
  }
  else { cerr<<"CTAlfaArnaud::load, Error: Could not get data\n"; exit(-1); }  
}

// Log data
void CTAlfaArnaud::log() {
  int i;

  // Log 
  if (fitType==fitTypeA) cout<<"CTAlfaArnaud::log: fitType="<<logArnaudFitType(fitType)<<", a="<<a<<", b="<<b<<", c="<<c<<", d="<<d<<"\n";
  else cout<<"CTAlfaArnaud::log: fitType="<<logArnaudFitType(fitType)<<", a="<<a<<", b="<<b<<", c="<<c<<", deltaE="<<deltaE<<"\n";
}

// Create copy
CTAlfa* CTAlfaArnaud::copy() {
  return new CTAlfaArnaud(*this);
}

// Get memory allocated - New!
double CTAlfaArnaud::mem() const {
  double _mem=sizeof(*this);
 
  return _mem;
}

// Constructor - New!
CTTransData::CTTransData() : atom(NULL), upperIState(NULL), upperBState(NULL), lowerIState(NULL), lowerBState(NULL) {};

// Destructor - New!
CTTransData::~CTTransData() {};

// Add to bound-free transition list (if necessary)
// Note: Does not update the (bound state) bound-free transition pointer lists.
void CTTransData::add() {
  TransBFData* transBF=NULL;
  TransBFData transBFNew;
  int i;

  // Check if bound-free transition present
  /*for (i=0;i<upperBState->downTransBF.size();i++)
    if (upperBState->downTransBF[i]->lowerState==lowerBState) { transBF=upperBState->downTransBF[i]; break; }*/
  // This should work better, right - No update of bound-free transition pointer lists is required!
  for (i=0;i<atom->transBF.size();i++)
    if (atom->transBF[i].upperState==upperBState and atom->transBF[i].lowerState==lowerBState) { transBF=&(atom->transBF[i]); break; }

  //cout<<"CTTransData::add: "<<atom->Z<<":"<<fabs(upperIState->C)<<":"<<upperBState->index<<"->"<<atom->Z<<":"<<fabs(lowerIState->C)<<":"<<lowerBState->index<<", transBF: "<<transBF<<"\n";
    
  // If present - Set charge-transfer processes flag
  if (transBF!=NULL) transBF->processes=transBF->processes|pChargeTransfer;
  // Else
  else {
    // Initiate new bound-free transition
    transBFNew.index=atom->transBF.size();
    transBFNew.upperState=upperBState;
    transBFNew.lowerState=lowerBState;
    transBFNew.Xi=lowerIState->Xi0+(upperBState->E-lowerBState->E);
    transBFNew.sigmaRad=new BFSigmaRadValue(); ((BFSigmaRadValue*) transBFNew.sigmaRad)->sigma=0;
    transBFNew.omega=new BFOmegaValue(); ((BFOmegaValue*) transBFNew.omega)->omega=0;
    transBFNew.processes=pChargeTransfer;
    // Add new bound-free transition to the atom
    atom->transBF.push_back(transBFNew);
    cout<<"CTTransData::add: Added "<<atom->Z<<":"<<fabs(upperIState->C)<<":"<<upperBState->index<<"->"<<atom->Z<<":"<<fabs(lowerIState->C)<<":"<<lowerBState->index<<" bound-free transition\n";
  }
}

// Constructor - New!
CTReactionData::CTReactionData() : index(0), upTrans(), downTrans(), alfa(NULL) {};

// Copy constructor - New!
CTReactionData::CTReactionData(const CTReactionData& reactionData) : index(reactionData.index), upTrans(reactionData.upTrans), downTrans(reactionData.downTrans), alfa(NULL) {

  // Copy reaction coefficient object
  if (reactionData.alfa!=NULL) alfa=reactionData.alfa->copy();
  else alfa=NULL;
}

// Assignment operator - New!
CTReactionData& CTReactionData::operator=(const CTReactionData& reactionData) {

  if (&reactionData!=this) {

    index=reactionData.index;
    upTrans=reactionData.upTrans;
    downTrans=reactionData.downTrans;
    
    // If present - Delete reaction coefficient object
    if (alfa!=NULL) delete alfa;

    // Copy reaction coefficient object
    if (reactionData.alfa!=NULL) alfa=reactionData.alfa->copy();
    else alfa=NULL;
  }
}

// Destructor - New!
CTReactionData::~CTReactionData() {
  if (alfa!=NULL) delete alfa;
}

// Load (YAML) data - New!
void CTReactionData::load(const YAML::Node& data) {
  string type;

  // If reaction data available
  if (data["alfa"]) {

    // Get reaction data type
    if (data["alfa"]["type"]) type=data["alfa"]["type"].as<string>();
    else { cerr<<"CTReactionData::load, Error: Could not get alfa::type\n"; exit(-1); }

    // Create reaction data object
    if (type.compare("value")==0) alfa = new CTAlfaValue;
    else if (type.compare("table")==0) alfa = new CTAlfaTable;
    else if (type.compare("plaw")==0) alfa = new CTAlfaPLaw;
    else if (type.compare("arnaud")==0) alfa = new CTAlfaArnaud;
    else { cerr<<"CTReactionData::load, Error: alfa::type "<<type<<" not supported\n"; exit(-1); }

    // Load reaction data
    if (data["alfa"]["data"]) alfa->load(data["alfa"]);
    else { cerr<<"CTReactionData::load, Error: Could not get alfa::data\n"; exit(-1); }
  }

  // If number density type available
  if (data["nType"]) {
    // Get number density type
    if (data["nType"].as<string>().compare("iState")==0) nType=nIState;
    else nType=nBState;
  }
  // Else - Set number density type to bound state
  else nType=nBState;
}

// Adjust transitions automatically - New!
void CTReactionData::adjustTrans() {
  double upTransXi0,downTransXi0,upTransXi,downTransXi,xiDiff,upTransXiDiff=1e99,downTransXiDiff=1e99;
  TransBFData* downTransBF;
  int i,lowerBStateIndex=0,upperBStateIndex=0;

  // Get upward and downward ground-state ionization energy
  upTransXi0=upTrans.lowerIState->Xi0;
  downTransXi0=downTrans.lowerIState->Xi0;

  // If lower state of downward transition not set
  if (downTrans.lowerBState==NULL) {
    // For each lower state of the downward transition
    for (i=0;i<downTrans.lowerIState->bState.size();i++) {
      // Calculate ionization energy
      downTransXi=downTrans.lowerIState->Xi0+(downTrans.upperBState->E-downTrans.lowerIState->bState[i].E);
      // Calculate ionization enery difference
      xiDiff=(downTransXi-upTransXi0);
      // If less than current minimum difference - Update minumum difference and set lower bound state
      if (fabs(xiDiff)<fabs(downTransXiDiff)) { downTransXiDiff=xiDiff; lowerBStateIndex=i; }
    }
  }

  // If upper state of upward transition not set
  if (upTrans.upperBState==NULL) {
    // For each upper state of the upward transition
    for (i=0;i<upTrans.upperIState->bState.size();i++) {
      // Calculate ionization energy
      upTransXi=upTrans.lowerIState->Xi0+(upTrans.upperIState->bState[i].E-upTrans.lowerBState->E);
      // Calculate ionization energy difference
      xiDiff=(downTransXi0-upTransXi);
      // If less than current minimum difference - Update minumum difference and set lower bound state
      if (fabs(xiDiff)<fabs(upTransXiDiff)) { upTransXiDiff=xiDiff; upperBStateIndex=i; }
    }
  }

  // If adjustment of downward transition gives miniumum energy difference
  if (fabs(downTransXiDiff)<fabs(upTransXiDiff)) {
    downTrans.lowerBState=&(downTrans.lowerIState->bState[lowerBStateIndex]);
    upTrans.upperBState=&(upTrans.upperIState->bState[0]); // Assumes that bState[0] is ground-state!
    xiDiff=downTransXiDiff;
  }
  // Else (i.e adjustment of upward transition gives miniumum energy difference)
  else {
    downTrans.lowerBState=&(downTrans.lowerIState->bState[0]); // Assumes that bState[0] is ground-state!
    upTrans.upperBState=&(upTrans.upperIState->bState[upperBStateIndex]);
    xiDiff=upTransXiDiff;
  }

  cout<<"CTReaction::adjustTrans: Adjusted "<<downTrans.atom->Z<<":"<<fabs(downTrans.upperIState->C)<<":"<<downTrans.upperBState->index<<"->"<<downTrans.atom->Z<<":"<<fabs(downTrans.lowerIState->C)<<":"<<downTrans.lowerBState->index<<" + "<<upTrans.atom->Z<<":"<<fabs(upTrans.lowerIState->C)<<":"<<upTrans.lowerBState->index<<"->"<<upTrans.atom->Z<<":"<<fabs(upTrans.upperIState->C)<<":"<<upTrans.upperBState->index<<" CT reaction, dE: "<<xiDiff/eV<<" eV\n";

  // Add downward and upward bound-free transitions to list (if necessary)
  // Note: Does not update the (bound state) bound-free transition pointer lists.
  downTrans.add();
  upTrans.add();
}

// Log data - New!
void CTReactionData::log() {

  // Log reaction coefficient
  if (alfa!=NULL) alfa->log();
  else cout<<"CTReactionData:log: alfa not loaded\n";
}

// Get memory allocated - New!
double CTReactionData::mem() const {
  double _mem=sizeof(*this);
  int i;

  // Add memory allocated for reaction coefficient
  if (alfa!=NULL) _mem+=alfa->mem();

  return _mem;
}

// Constructor - New!
AtomDataList::AtomDataList() { }

// Copy constructor - New!
AtomDataList::AtomDataList(const AtomDataList& atomDataList) : atom(atomDataList.atom), ctReaction(atomDataList.ctReaction) {
  CTTransData* trans;
  int i;

  // For each charge-transfer reaction
  for (i=0;i<ctReaction.size();i++) {

    // Update upward transition pointers
    trans = (CTTransData*) &(atomDataList.ctReaction[i].upTrans);
    ctReaction[i].upTrans.atom=&(atom[trans->atom->index]);
    ctReaction[i].upTrans.upperIState=&(atom[trans->atom->index].iState[trans->upperIState->index]);
    ctReaction[i].upTrans.upperBState=&(atom[trans->atom->index].iState[trans->upperIState->index].bState[trans->upperBState->index]);
    ctReaction[i].upTrans.lowerIState=&(atom[trans->atom->index].iState[trans->lowerIState->index]);
    ctReaction[i].upTrans.lowerBState=&(atom[trans->atom->index].iState[trans->lowerIState->index].bState[trans->lowerBState->index]);

    // Update downward transition pointers
    trans = (CTTransData*) &(atomDataList.ctReaction[i].downTrans);
    ctReaction[i].upTrans.atom=&(atom[trans->atom->index]);
    ctReaction[i].upTrans.upperIState=&(atom[trans->atom->index].iState[trans->upperIState->index]);
    ctReaction[i].upTrans.upperBState=&(atom[trans->atom->index].iState[trans->upperIState->index].bState[trans->upperBState->index]);
    ctReaction[i].upTrans.lowerIState=&(atom[trans->atom->index].iState[trans->lowerIState->index]);
    ctReaction[i].upTrans.lowerBState=&(atom[trans->atom->index].iState[trans->lowerIState->index].bState[trans->lowerBState->index]);
  }
}

// Assignment operator - New!
AtomDataList& AtomDataList::operator=(const AtomDataList& atomDataList) {
  CTTransData* trans;
  int i;

  if (&atomDataList!=this) {

    atom=atomDataList.atom;
    ctReaction=atomDataList.ctReaction;

    // For each charge-transfer reaction
    for (i=0;i<ctReaction.size();i++) {
      
      // Update upward transition pointers
      trans = (CTTransData*) &(atomDataList.ctReaction[i].upTrans);
      ctReaction[i].upTrans.atom=&(atom[trans->atom->index]);
      ctReaction[i].upTrans.upperIState=&(atom[trans->atom->index].iState[trans->upperIState->index]);
      ctReaction[i].upTrans.upperBState=&(atom[trans->atom->index].iState[trans->upperIState->index].bState[trans->upperBState->index]);
      ctReaction[i].upTrans.lowerIState=&(atom[trans->atom->index].iState[trans->lowerIState->index]);
      ctReaction[i].upTrans.lowerBState=&(atom[trans->atom->index].iState[trans->lowerIState->index].bState[trans->lowerBState->index]);
      
      // Update downward transition pointers
      trans = (CTTransData*) &(atomDataList.ctReaction[i].downTrans);
      ctReaction[i].upTrans.atom=&(atom[trans->atom->index]);
      ctReaction[i].upTrans.upperIState=&(atom[trans->atom->index].iState[trans->upperIState->index]);
      ctReaction[i].upTrans.upperBState=&(atom[trans->atom->index].iState[trans->upperIState->index].bState[trans->upperBState->index]);
      ctReaction[i].upTrans.lowerIState=&(atom[trans->atom->index].iState[trans->lowerIState->index]);
      ctReaction[i].upTrans.lowerBState=&(atom[trans->atom->index].iState[trans->lowerIState->index].bState[trans->lowerBState->index]);
    }
  }
}

// Load atom data
void AtomDataList::load(const string& dataFile, const AtomDataFormat& dataFormat, const bool& loadIStateData) {
  switch (dataFormat) {
  case aJ :
    loadAJ(dataFile,loadIStateData);
    break;
  case yAML :
    loadYAML(dataFile,loadIStateData);
    break;
  default :
    loadAJ(dataFile,loadIStateData);
  }

  //log();
}

// Log atom data
void AtomDataList::log() {
  int i;

  // Log data for each atom
  for (i=0;i<atom.size();i++) 
    atom[i].log();

  // Log data for each charge-transfer reaction - New!
  for (i=0;i<ctReaction.size();i++) 
    ctReaction[i].log();
}

// Load YAML atom data
void AtomDataList::loadYAML(const string& dataFile, const bool& loadIStateData) {
  YAML::Node data,ctReactionData;
  string ctReactionDataFile; // New!
  int nAtom,i;

  // Load atom data from file
  data=YAML::LoadFile(dataFile);
  // Error handling ?

  //cout<<"data: "<<typeName(data)<<", size: "<<data.size()<<"\n";

  //cout<<"data[atom]: "<<typeName(data["atom"])<<", size: "<<data["atom"].size()<<"\n";

  // Get number of atoms
  if (data["atom"]) nAtom=data["atom"].size();
  else { cerr<<"AtomDataList::loadYAML, Error: Could not get atom\n"; exit(-1); }  

  // Resize atom list
  atom.resize(nAtom);

  // For each atom
  for (i=0;i<nAtom;i++) {

    // Set atom index
    atom[i].index=i;

    //cout<<"data[atom]["<<i<<"]: "<<typeName(data["atom"][i])<<", size: "<<data["atom"][i].size()<<"\n";

    // Load atom data
    atom[i].load(data["atom"][i],getDataDir(dataFile));
  }

  // If charge-transfer reaction data avaiable - New!
  if (data["ctReaction"]) {

    // Get charge-transfer reaction data file
    ctReactionDataFile=getDataFile(getDataDir(dataFile),data["ctReaction"].as<string>());
    
    // Load charge-transfer reaction data from file
    ctReactionData=YAML::LoadFile(ctReactionDataFile);
    
    // Resize charge-transfer reaction list
    ctReaction.resize(ctReactionData.size());
    
    // For each charge-transfer reaction
    for (i=0;i<ctReactionData.size();i++) {

      //cout<<"AtomData::load, Load ctReaction "<<i<<"\n";

      // Set charge-transfer reaction index
      ctReaction[i].index=i;

      // Load upward transition data
      if (ctReactionData[i]["upTrans"]) ctReaction[i].upTrans=loadCTTrans(ctReactionData[i]["upTrans"]);
      else { cerr<<"AtomDataList::load, Error: Could not get ctReaction::upTrans\n"; exit(-1); }   
  
      // Load downward transition data
      if (ctReactionData[i]["downTrans"]) ctReaction[i].downTrans=loadCTTrans(ctReactionData[i]["downTrans"]);
      else { cerr<<"AtomDataList::load, Error: Could not get ctReaction::downTrans\n"; exit(-1); }   

      // If lower/upper state of downward/upward transition not set - Adjust transitions automatically
      if (ctReaction[i].downTrans.lowerBState==NULL or ctReaction[i].upTrans.upperBState==NULL) ctReaction[i].adjustTrans();
           
      // Load charge-transfer data
      ctReaction[i].load(ctReactionData[i]);

      cout<<"AtomDataList::load, CTReaction "<<i<<" loaded, downTrans: "<<ctReaction[i].downTrans.upperIState->index<<":"<<ctReaction[i].downTrans.upperBState->index<<"->"<<ctReaction[i].downTrans.lowerIState->index<<":"<<ctReaction[i].downTrans.lowerBState->index<<", upTrans: "<<ctReaction[i].upTrans.lowerIState->index<<":"<<ctReaction[i].upTrans.lowerBState->index<<"->"<<ctReaction[i].upTrans.upperIState->index<<":"<<ctReaction[i].upTrans.lowerBState->index<<"\n";
    }
    
    cout<<"AtomDataList::load: Loaded "<<i<<" charge-transfer reactions\n"; // Changed!

    // Update bound-free transition (pointer) lists - New!
    updateBFTrans();
  }
}

// Load charge-transfer transition data - New!
CTTransData AtomDataList::loadCTTrans(const YAML::Node& data) {
  CTTransData CTTrans;
  double Z;
  int i,upperIStateIndex,upperBStateIndex,lowerIStateIndex,lowerBStateIndex;

  // Load Z
  if (data["Z"]) Z=data["Z"].as<double>();
  else { cerr<<"AtomDataList::loadCTTrans, Error: Could not get Z\n"; exit(-1); }

  // Load upper ionization and bound state index
  if (data["j"]) 
    if (data["j"].size()==2){
      upperIStateIndex=data["j"][0].as<int>()-1; 
      if (data["j"][1].as<string>().compare("auto")==0) upperBStateIndex=-1;
      else upperBStateIndex=data["j"][1].as<int>()-1;
    }
    //if (data["j"].size()==2) { upperIStateIndex=data["j"][0].as<int>()-1; upperBStateIndex=data["j"][1].as<int>()-1; }
    else { cerr<<"AtomDataList::loadCTTrans, Error: Could not get j\n"; exit(-1); }
  else { cerr<<"AtomDataList::loadCTTrans, Error: Could not get j\n"; exit(-1); }

  // Load lower ionization and bound state index
  if (data["i"]) 
    if (data["i"].size()==2) { 
      lowerIStateIndex=data["i"][0].as<int>()-1;
      if (data["i"][1].as<string>().compare("auto")==0) lowerBStateIndex=-1;
      else lowerBStateIndex=data["i"][1].as<int>()-1; 
    }
    //if (data["i"].size()==2) { lowerIStateIndex=data["i"][0].as<int>()-1; lowerBStateIndex=data["i"][1].as<int>()-1; }
    else { cerr<<"AtomDataList::loadCTTrans, Error: Could not get i\n"; exit(-1); }
  else { cerr<<"AtomDataList::loadCTTrans, Error: Could not get i\n"; exit(-1); }

  // Set atom data pointer
  CTTrans.atom=find(Z);
  if (CTTrans.atom==NULL) { cerr<<"AtomDataList::loadCTTrans, Error: Could not get find atom Z="<<Z<<"\n"; exit(-1); }  

  // Set upper ionization and bound state pointer
  if (upperIStateIndex<CTTrans.atom->iState.size()) CTTrans.upperIState=&(CTTrans.atom->iState[upperIStateIndex]);
  else { cerr<<"AtomDataList::loadCTTrans, Error: Could not find ionization state "<<upperIStateIndex<<" for atom Z="<<Z<<"\n"; exit(-1); }
  if (upperBStateIndex<(int)CTTrans.upperIState->bState.size()) { // Note: Need to cast size to int here
    if (upperBStateIndex!=-1) CTTrans.upperBState=&(CTTrans.upperIState->bState[upperBStateIndex]);
    else CTTrans.upperBState=NULL;
  }
  //if (upperBStateIndex<CTTrans.upperIState->bState.size()) CTTrans.upperBState=&(CTTrans.upperIState->bState[upperBStateIndex]);
  else { cerr<<"AtomDataList::loadCTTrans, Error: Could not find bound state "<<upperIStateIndex<<":"<<upperBStateIndex<<" for atom Z="<<Z<<", upperIState.bState.size: "<<CTTrans.upperIState->bState.size()<<"\n"; exit(-1); }  

  // Set lower ionization and bound state pointer
  if (lowerIStateIndex<CTTrans.atom->iState.size()) CTTrans.lowerIState=&(CTTrans.atom->iState[lowerIStateIndex]);
  else { cerr<<"AtomDataList::loadCTTrans, Error: Could not find ionization state "<<lowerIStateIndex<<" for atom Z="<<Z<<"\n"; exit(-1); }
  if (lowerBStateIndex<(int)CTTrans.lowerIState->bState.size()) { // Note: Need to cast size to inte here
    if (lowerBStateIndex!=-1) CTTrans.lowerBState=&(CTTrans.lowerIState->bState[lowerBStateIndex]);
    else CTTrans.lowerBState=NULL;
  }
  //if (lowerBStateIndex<CTTrans.lowerIState->bState.size()) CTTrans.lowerBState=&(CTTrans.lowerIState->bState[lowerBStateIndex]);
  else { cerr<<"AtomDataList::loadCTTrans, Error: Could not find bound state "<<lowerIStateIndex<<":"<<lowerBStateIndex<<" for atom Z="<<Z<<", lowerIState.bState.size: "<<CTTrans.lowerIState->bState.size()<<"\n"; exit(-1); }  

  // If upper and lower bound state set - Add to bound-free transition list
  // Note: Does not update the (bound state) bound-free transition pointer lists.
  if (CTTrans.upperBState!=NULL and CTTrans.lowerBState!=NULL) CTTrans.add();

  // Return 
  return CTTrans;
}

// Update bound-free transition (pointer) lists
void AtomDataList::updateBFTrans() {
  int i,j,k;

  // For each bound state
  for (i=0;i<atom.size();i++) {
    for (j=0;j<atom[i].iState.size();j++) {
      for (k=0;k<atom[i].iState[j].bState.size();k++) {
	// Reset upward and downward bound-free transition (pointer) lists
	atom[i].iState[j].bState[k].upTransBF.resize(0);
	atom[i].iState[j].bState[k].downTransBF.resize(0);
      }
    }
  }

  // For each bound-free transition
  for (i=0;i<atom.size();i++) {
    for (j=0;j<atom[i].transBF.size();j++) {
      // Add to upward and downward bound-free transition (pointer) lists
      atom[i].transBF[j].lowerState->upTransBF.push_back(&(atom[i].transBF[j]));
      atom[i].transBF[j].upperState->downTransBF.push_back(&(atom[i].transBF[j]));
    }
  }
}

// Load A Jerkstrand atom data
// Deprecated!
void AtomDataList::loadAJ(const string& dataFile, const bool& loadIStateData) {
  fstream file;
  stringstream linestream;
  string line,iStateFile;
  int pos1,pos2,i,j,k,nAtoms,nIsotopes,nIStates,nBStates,nTrans;
  double Xi0Sum,ESmall;

  // Open atom data file
  file.open(dataFile.c_str());
  // Error handling
  if (file.fail()) {cerr<<"AtomDataList::loadAJ: Error1\n"; return;}
  // While not end of file
  while (!file.eof()) {
    // Get next line
    getline(file,line);
    linestream.str(line);
    linestream.clear();
    // Error handling
    if (file.fail() && !file.eof()) {cerr<<"AtomDataList::loadAJ: Error2\n"; return;}
    // No error - End of file
    if (file.fail() && file.eof()) break;
    // Ignore lines starting with # and empty lines
    if (line[0]!='#' && line.size()!=0) {
      // Read number of atoms from linestream
      linestream>>skipws>>nAtoms;
      // Error handling
      if (linestream.fail()) {cerr<<"AtomDataList::loadAJ: Error3\n"; return;}
      // Resize atom list
      atom.resize(nAtoms);
      // While atom data to read
      i=0;
      while (i<nAtoms) {
	// Get next line stream
	getline(file,line);
	linestream.str(line);
	linestream.clear();
	// Error handling
	if (file.fail()) {cerr<<"AtomDataList::loadAJ: Error4\n"; return;}
	// Ignore lines starting with # and empty lines
	if (line[0]!='#' && line.size()!=0) {
	  // Resize atom list
	  atom.resize(i+1);
	  // Set atom index
	  atom[i].index=i;
	  // Read Z from linestream
	  linestream>>skipws>>atom[i].Z;
	  // Error handling
	  if (linestream.fail()) {cerr<<"AtomDataList::loadAJ: Error5\n"; return;}
	  // Read number of isotopes from linestream
	  linestream>>skipws>>nIsotopes;
	  // Error handling
	  if (linestream.fail()) {cerr<<"AtomDataList::loadAJ: Error6\n"; return;}
	  // Resize isotope list
	  atom[i].isotope.resize(nIsotopes);
	  // Read number of ionization states from linestream
	  linestream>>skipws>>nIStates;
	  // Error handling
	  if (linestream.fail()) {cerr<<"Error"; return;}
	  // Resize ionization state list
	  atom[i].iState.resize(nIStates+1);
	  j=0;
	  // While isotope data to read
	  while (j<nIsotopes) {
	    // Get next line stream
	    getline(file,line);
	    linestream.str(line);
	    linestream.clear();
	    // Error handling
	    if (file.fail()) {cerr<<"AtomDataList::loadAJ: Error7\n"; return;}
	    // Ignore lines starting eith # and empty lines
	    if (line[0]!='#' && line.size()!=0) {
	      // Set isotope index
	      atom[i].isotope[j].index=j;
	      // Read N from linestream
	      linestream>>skipws>>atom[i].isotope[j].N;
	      // Error handling
	      if (linestream.fail()) {cerr<<"AtomDataList::loadAJ: Error8\n"; return;}
	      // Read m from linestream
	      linestream>>skipws>>atom[i].isotope[j].m;
	      // Convert to SI units
	      atom[i].isotope[j].m=atom[i].isotope[j].m*u;
	      // Error handling
	      if (linestream.fail()) {cerr<<"AtomDataList::loadAJ: Error9\n"; return;}
	      j++;
	    }
	  }
	  j=0;
	  // While ionization state data to read
	  while (j<nIStates) {
	    // Get next line stream
	    getline(file,line);
	    linestream.str(line);
	    linestream.clear();
	    // Error handling
	    if (file.fail()) {cerr<<"AtomDataList::loadAJ: Error10\n"; return;}
	    // Ignore lines starting with # and empty lines
	    if (line[0]!='#' && line.size()!=0) {
	      // Set ionization state index
	      atom[i].iState[j].index=j;
	      // Read C from linestream
	      linestream>>skipws>>atom[i].iState[j].C;
	      // Error handling
	      if (linestream.fail()) {cerr<<"AtomDataList::loadAJ: Error11\n"; return;}
	      // Read Xi0 from linestream
	      linestream>>skipws>>atom[i].iState[j].Xi0;
	      // Convert to SI units
	      atom[i].iState[j].Xi0=atom[i].iState[j].Xi0*eV;
	      // Error handling
	      if (linestream.fail()) {cerr<<"AtomDataList::loadAJ: Error12\n"; return;}
	      if (loadIStateData) {
		// Read ionization state file from linestream
		linestream>>skipws>>iStateFile;
		// Error handling
		if (linestream.fail()) {cerr<<"AtomDataList::loadAJ: Error13\n"; return;}
		// Read number of bound states from linestream
		linestream>>skipws>>nBStates;
		// Error handling
		if (linestream.fail()) {cerr<<"AtomDataList::loadAJ: Error14\n"; return;}
		//cout<<"Number of bound states: "<<nBStates<<"\n";
		// Resize bound state list
		atom[i].iState[j].bState.resize(nBStates);
		// Read number of bound-bound transitions from linestream
		linestream>>skipws>>nTrans;
		// Error handling
		if (linestream.fail()) {cerr<<"AtomDataList::loadAJ: Error15\n"; return;}
		// Resize bound-bound transition list
		atom[i].iState[j].transBB.resize(nTrans);
		// Load ionization state data from file
		loadIonStateAJ(atom[i].iState[j],iStateFile);
	      }
	      j++;
	    }
	    // Add one further ionization state to list (check this!!!)
	    atom[i].iState[j].C=atom[i].iState[j-1].C-1;
	    // Set ionization state index
	    atom[i].iState[j].index=j;
	  }
	  i++;
	}
      }
    }
  }

  // Temporary hack to add dummy ground state for highest ionization state
  for (i=0;i<atom.size();i++) {
    atom[i].iState[atom[i].iState.size()-1].bState.resize(1);
    atom[i].iState[atom[i].iState.size()-1].bState[0].index=0;
    atom[i].iState[atom[i].iState.size()-1].bState[0].E=0;
    atom[i].iState[atom[i].iState.size()-1].bState[0].g=1;
    atom[i].iState[atom[i].iState.size()-1].bState[0].A=0;
  }
  // Temporary hack to setup bound-free transition lists
  double nTransBF,nTransBFTot;
  for (i=0;i<atom.size();i++) {
    nTransBFTot=0;
    for (j=0;j<atom[i].iState.size()-1;j++) {
      if (atom[i].iState[j].bState.size()>=1000) nTransBF=1000;
      else nTransBF=atom[i].iState[j].bState.size();
      /*atom[i].iState[j].transBF.resize(nTransBF);
      for (k=0;k<nTransBF;k++) {
	atom[i].iState[j].transBF[k].index=k;
	atom[i].iState[j].transBF[k].upperState=&(atom[i].iState[j+1].bState[0]);
	atom[i].iState[j+1].bState[0].downTransBF.push_back(&(atom[i].iState[j].transBF[k]));
	atom[i].iState[j].transBF[k].lowerState=&(atom[i].iState[j].bState[k]);
	atom[i].iState[j].bState[k].upTransBF.push_back(&(atom[i].iState[j].transBF[k]));
	atom[i].iState[j].transBF[k].Xi=atom[i].iState[j].Xi0-atom[i].iState[j].bState[k].E;
	}*/
      atom[i].transBF.resize(nTransBFTot+nTransBF);
      for (k=0;k<nTransBF;k++) {
	atom[i].transBF[nTransBFTot+k].index=nTransBFTot+k;
	atom[i].transBF[nTransBFTot+k].upperState=&(atom[i].iState[j+1].bState[0]);
	atom[i].iState[j+1].bState[0].downTransBF.push_back(&(atom[i].transBF[nTransBFTot+k]));
	atom[i].transBF[nTransBFTot+k].lowerState=&(atom[i].iState[j].bState[k]);
	atom[i].iState[j].bState[k].upTransBF.push_back(&(atom[i].transBF[nTransBFTot+k]));
	atom[i].transBF[nTransBFTot+k].Xi=atom[i].iState[j].Xi0-atom[i].iState[j].bState[k].E;
      }
      nTransBFTot+=nTransBF;
    }
  }
  // Temporary hack to add bound state energy above ground state of the neutral atom
  for (i=0;i<atom.size();i++) {
    Xi0Sum=0;
    // Hack - introduce small shift of energy scale to avoid zero ground state energy
    ESmall=h*(c/(1e6*AA));
    for (j=0;j<atom[i].iState.size();j++) {
      if (j>0) Xi0Sum+=atom[i].iState[j-1].Xi0;
      for (k=0;k<atom[i].iState[j].bState.size();k++) {
	atom[i].iState[j].bState[k].Etot=atom[i].iState[j].bState[k].E+Xi0Sum+ESmall;
      }
    }
  }

  return;
}

// Load A Jerkstrand ionization state data
// Deprecated!
void AtomDataList::loadIonStateAJ(IStateData& iState, string dataFile) {
  ifstream file;
  stringstream linestream;
  string line;
  int index,i,j,k,l;

  // Open ionization state data file
  file.open(dataFile.c_str());
  // Error handling
  if (file.fail()) {cerr<<"AtomDataList::loadIonStateAJ: Could not open ionization state data file\n"; exit(1);}
  i=0;
  // Ignore two first lines
  getline(file,line);
  getline(file,line);
  // While bound state data to read
  while (i<iState.bState.size()) {
    // Get next line
    getline(file,line);
    linestream.str(line);
    linestream.clear();
    // Error handling
    if (file.fail()) {cerr<<"AtomDataList::loadIonStateAJ: Error2\n"; exit(1);}
    // cout<<"Read bound state "<<i<<" data\n";
    // Read index from linestream
    linestream>>skipws>>iState.bState[i].index;
    iState.bState[i].index-=1;
    // Error handling
    if (linestream.fail()) {cerr<<"AtomDataList::loadIonStateAJ: Error3\n"; exit(1);}
    // Read E from linestream
    linestream>>skipws>>iState.bState[i].E;
    // Convert to SI units
    iState.bState[i].E=h*(c*iState.bState[i].E/1e-2);
    // Error handling
    if (linestream.fail()) {cerr<<"AtomDataList::loadIonStateAJ: Error4\n"; exit(1);}
    // Read g from linestream
    linestream>>skipws>>iState.bState[i].g;
    // Error handling
    if (linestream.fail()) {cerr<<"AtomDataList::loadIonStateAJ: Error5\n"; exit(1);}
    // Initialize cumulative downward A value
    iState.bState[i].A=0;
    i++;
  }
  i=0;
  // Ignore two more lines
  getline(file,line);
  getline(file,line);
  // While not end of file
  while (i<iState.transBB.size()) {
    // Get next line stream
    getline(file,line);
    linestream.str(line);
    linestream.clear();
    // Error handling
    if (file.fail()) {cerr<<"AtomDataList::loadIonStateAJ: Error6\n"; exit(1);}
    // Set bound-bound transition index
    iState.transBB[i].index=i;
    // Read upper state index from linestream
    linestream>>skipws>>index;
    // Add bound-bound transition to ionization state bound-bound transition list
    iState.transBB[i].upperState=&(iState.bState[index-1]);
    // Add bound-bound transition to upper state downward bound-bound transition list
    iState.bState[index-1].downTransBB.push_back(&(iState.transBB[i]));
    // Error handling
    if (linestream.fail()) {cerr<<"AtomDataList::loadIonState: Error7\n"; exit(1);}
    // Read lower state index from index from linestream
    linestream>>skipws>>index;
    // Add bound-bound transition to ionization state bound-bound transition list
    iState.transBB[i].lowerState=&(iState.bState[index-1]);
    // Add bound-bound transition to lower state upward bound-bound transition list
    iState.bState[index-1].upTransBB.push_back(&(iState.transBB[i]));
    // Error handling
    if (linestream.fail()) {cerr<<"AtomDataList::loadIonsState: Error8\n"; exit(1);}
    // Read Xi from linestream
    linestream>>skipws>>iState.transBB[i].Xi;
    // Convert to SI units
    iState.transBB[i].Xi = h*(c/(iState.transBB[i].Xi*AA));
    // Error handling
    if (linestream.fail()) {cerr<<"AtomDataList::loadIonState: Error9\n"; exit(1);}
    // Read A from linestream
    linestream>>skipws>>iState.transBB[i].A;
    // Add to upper state cumulative A value
    iState.transBB[i].upperState->A+=iState.transBB[i].A;
    // Error handling
    if (linestream.fail()) {cerr<<"AtomDataList::loadIonState: Error10\n"; exit(1);}
    i++;
  }

  return;
}

// Find atom data
AtomData* AtomDataList::find(const int& Z) {
  int i;
  for (i=0;i<atom.size();i++) {
    if (atom[i].Z==Z) return &(atom[i]);
  }
  return NULL;
}

// Resample bound-free tables
// Deprecated - Use updateTables!
void AtomDataList::resampleBFTables(const vector<double>& nu) {
  int i,j;

  // For each atom
  for (i=0;i<atom.size();i++) {
    // For each bound-free transition
    for (j=0;j<atom[i].transBF.size();j++) {
      // If radiative cross-section available - resample table (if present)
      //cout<<"AtomDataList::resampleBFTables: atom: "<<i<<", upperState: "<<atom[i].transBF[j].upperState->index<<", lowerState: "<<atom[i].transBF[j].lowerState->index<<"\n";
      if (atom[i].transBF[j].sigmaRad!=NULL) atom[i].transBF[j].sigmaRad->resampleTable(nu,atom[i].transBF[j].Xi/h);
    }
  }
}

// Update tables (using the provided frequency binning) - New!
void AtomDataList::updateTables(const vector<double>& nu) {
  int i,j,k;

  // For each atom
  for (i=0;i<atom.size();i++) {

    // Update tables for each bound-free transition
    for (j=0;j<atom[i].transBF.size();j++) atom[i].transBF[j].updateTables(nu);
    
    // For each ionization state
    for (j=0;j<atom[i].iState.size();j++) {
      // Update tables for each bound-bound transition
      for (k=0;k<atom[i].iState[j].transBB.size();k++) atom[i].iState[j].transBB[k].updateTables(nu);
    }
  }
}

// Get memory allocated - New!
double AtomDataList::mem() const {
  double _mem=sizeof(*this);
  int i;

  // Add memory allocated for each atom
  for (i=0;i<atom.size();i++) _mem+=atom[i].mem();

  // Add memory allocated for each charge-transfer reaction
  for (i=0;i<ctReaction.size();i++) _mem+=ctReaction[i].mem();

  return _mem;
}
