#ifndef _ATOMDATA_
#define _ATOMDATA_

#include <string>
#include <vector>
#include "yaml-cpp/yaml.h"
#include "mc-util.h"
#include "mc-math.h"
#include "mc-physics.h"

using namespace std;
using namespace util;
using namespace math;
using namespace physics;

namespace atomdata {

  // Process type - Moved (from grid.h)!
  enum ProcessType { pNone=0, pEScatter=1, pCScatter=2, pBoundBound=4, pBoundFree=8, pFreeFree=16, pRadiative=32, pCollisional=64, pNonThermal=128 , pTwoPhoton=256, pChargeTransfer=512, pGScatter=1024, pAny=2047 }; // Changed!

  // Log level
  enum logLevel { logInfoBasic, logInfoAll, logDebug };

  // Get data file
  string getDataFile(const string& dataDir, const string& dataFile);

  // Get data directory
  string getDataDir(const string& dataFile);

  // Get (YAML) type name
  string typeName(const YAML::Node& data);

  // Resample bound-free cross-section table
  Interpolator1D resampleBFTable(const vector<double>& nu, const double& nuEdge, Interpolator1D& sigma);

  // Forward declarations
  class TransData;
  class TransBBData;
  class TransBFData;
  class BFAlfa;

  // Bound state data
  class BStateData {
  public :
    // Bound state index
    int index;
    // Energy above ground state
    double E;
    // Energy above ground state of the neutral atom
    double Etot;
    // Statistical weight
    double g;
    // Cumulative A value
    double A;
    // Recombination coefficient - New!
    BFAlfa* alfa;
    // LTE reference state index and pointer
    //int LTERefIndex;
    int LTERefBStateIndex,LTERefIStateIndex; // Changed!
    BStateData* LTERefState;
    // Recombination correction flag - New!
    bool alfaCorrect;
    // Recombination fraction - New!
    double alfaFrac;
    // Testing!
    bool alfaSimple;
    // List of (pointers to) downward bound-bound transitions
    vector<TransBBData*> upTransBB;
    // List of (pointers to) upward bound-bound transitions
    vector<TransBBData*> downTransBB;
    // List of (pointers to) downward bound-free transitions
    vector<TransBFData*> upTransBF;
    // List of (pointers to) upward bound-free transitions
    vector<TransBFData*> downTransBF;
    // Constructor
    BStateData();
    // Copy constructor
    BStateData(const BStateData& bStateData);
    // Assignment operator
    BStateData& operator=(const BStateData& bStateData);
    // Destructor
    virtual ~BStateData();
    // Load (YAML) data
    void load(const YAML::Node& data);
    // Log data
    void log();
    // Get memory allocated - New!
    double mem() const;
  };

  // Bound-bound collisional strength data
  class BBOmega {
  public :
    // Get collitional strength
    //virtual double operator ()(double T) const {};
    virtual double operator ()(double T) {};
    // Load (YAML) data
    virtual void load(const YAML::Node& data) {};
    // Log data
    virtual void log() {};
    // Create copy
    virtual BBOmega* copy() {};
    // Get memory allocated - New!
    virtual double mem() const {};
  };

  // Bound-bound collisional strength data (value)
  class BBOmegaValue : public BBOmega {
  public :
    // Collisional strength
    double omega;
    // Get collisional strength
    //virtual double operator ()(double T) const;
    virtual double operator ()(double T);
    // Load (YAML) data
    virtual void load(const YAML::Node& data);
    // Log data
    virtual void log();
    // Create copy
    virtual BBOmega* copy();
    // Get memory allocated - New!
    virtual double mem() const;
  };

  // Bound-bound collisional strength data (table)
  class BBOmegaTable : public BBOmega {
  public :
    // Collisional strength
    Interpolator1D omega;
    // Get collisional strength
    //virtual double operator ()(double T) const;
    virtual double operator ()(double T);
    // Load (YAML) data
    virtual void load(const YAML::Node& data);
    // Log data
    virtual void log();
    // Create copy
    virtual BBOmega* copy();
    // Get memory allocated - New!
    virtual double mem() const;
  };

  // Bound-bound collisional strength data (powerlaw)
  class BBOmegaPLaw : public BBOmega {
  public :
    // Collisional strength
    double omega;
    // Temperature
    double T0;
    // Power law coefficient
    double p;
    // Get cross-section
    //virtual double operator ()(double nu) const;
    virtual double operator ()(double nu);
    // Load (YAML) data
    virtual void load(const YAML::Node& data);
    // Log data
    virtual void log();
    // Create copy
    virtual BBOmega* copy();
    // Get memory allocated - New!
    virtual double mem() const;
  };
  
  // Two-photon emissivity data - New!
  class BBEta2 {
  private :
    // Emissivity cumulative distribution function table
    Interpolator1D pF;
    // Emissivity cumulative distribution function table (inverse version)
    Interpolator1D pFInv;
  public :
    // Get emissivity distribution function
    virtual double operator ()(double nu) {};
    // Get emissivity cumulative distribution function
    virtual double F(double nu);
    // Get emissivity cumulative distribution function (inverse version)
    virtual double FInv(double F);
    // Create emissivity cumulative distribution function table
    virtual void createFTable(const vector<double>& nu, const double& nu0);
    // Adjust edge (if necessary)
    virtual void adjustEdge(const double& nuEdge) {};
    // Load (YAML) data
    virtual void load(const YAML::Node& data) {};
    // Log data
    virtual void log() {};
    // Create copy
    virtual BBEta2* copy() {};
    // Get memory allocated - New!
    virtual double mem() const;
  };

  // Two-photon emissivity data (table) - New!
  class BBEta2Table : public BBEta2 {
  public :
    // Emissivity distribution function
    Interpolator1D eta;
    // Get emissivity distribution function
    virtual double operator ()(double nu);
    // Load (YAML) data
    virtual void load(const YAML::Node& data);
    // Adjust edge (if necessary)
    virtual void adjustEdge(const double& nuEdge);
    // Log data
    virtual void log();
    // Create copy
    virtual BBEta2* copy();
    // Get memory allocated - New!
    virtual double mem() const;
  };

  // Bound-free radiative cross-section data
  class BFSigmaRad {
  public :
    // Get radiative cross-section
    //virtual double operator ()(double nu) const {};
    virtual double operator ()(double nu) {};
    // Load (YAML) data
    virtual void load(const YAML::Node& data) {};
    // Resample table (if present)
    virtual double resampleTable(const vector<double>& nu, const double& nuEgde) {};
    // Adjust edge (if necessary) - New!
    virtual void adjustEdge(const double& nuEdge) {};
    // Log data
    virtual void log() {};
    // Create copy
    virtual BFSigmaRad* copy() {};
    // Get memory allocated - New!
    virtual double mem() const {};
  };

  // Bound-free radiative cross-cection data (value)
  class BFSigmaRadValue : public BFSigmaRad {
  public :
    // Cross-section
    double sigma;
    // Get cross-section
    virtual double operator ()(double nu);
    // Load (YAML) data
    virtual void load(const YAML::Node& data);
    // Log data
    virtual void log();
    // Create copy
    virtual BFSigmaRad* copy();
    // Get memory allocated - New!
    virtual double mem() const;
  };

  // Bound-free radiative cross-cection data (table)
  class BFSigmaRadTable : public BFSigmaRad {
  public :
    // Cross-section
    Interpolator1D sigma;
    // Get cross-section
    //virtual double operator ()(double nu) const;
    virtual double operator ()(double nu);
    // Load (YAML) data
    virtual void load(const YAML::Node& data);
    // Resample cross-section table
    virtual double resampleTable(const vector<double>& nu, const double& nuEdge);
    // Adjust edge - New!
    virtual void adjustEdge(const double& nuEdge);
    // Log data
    virtual void log();
    // Create copy
    virtual BFSigmaRad* copy();
    // Get memory allocated - New!
    virtual double mem() const;
  };

  // Bound-free radiative cross-cection data (powerlaw)
  class BFSigmaRadPLaw : public BFSigmaRad {
  public :
    // Cross-section
    double sigma;
    // Frequency
    double nu0;
    // Power law coefficient
    double p;
    // Get cross-section
    //virtual double operator ()(double nu) const;
    virtual double operator ()(double nu);
    // Load (YAML) data
    virtual void load(const YAML::Node& data);
    // Log data
    virtual void log();
    // Create copy
    virtual BFSigmaRad* copy();
    // Get memory allocated - New!
    virtual double mem() const;
  };

  // Bound-free radiative cross-cection data (powerlaw/gaunt)
  class BFSigmaRadPLawGaunt : public BFSigmaRad {
  public :
    // Cross-section
    double sigma;
    // Frequency
    double nu0;
    // Power law coefficient
    double p;
    // Gaunt factor
    Interpolator1D gaunt;
    // Get cross-section
    //virtual double operator ()(double nu) const;
    virtual double operator ()(double nu);
    // Load (YAML) data
    virtual void load(const YAML::Node& data);
    // Resample gaunt-factor table
    virtual double resampleTable(const vector<double>& nu, const double& nuEdge);
    // Log data
    virtual void log();
    // Create copy
    virtual BFSigmaRad* copy();
    // Get memory allocated - New!
    virtual double mem() const;
  };

  // Bound-free radiative cross-cection data (Hummer fir)
  class BFSigmaRadHummer : public BFSigmaRad {
  public :
    // Edge frequency
    double nu0;
    // Fit parameters
    double a0;
    double b0;
    double c0;
    double d0;
    double x1;
    double a1;
    double b1;
    // Get cross-section
    //virtual double operator ()(double nu) const;
    virtual double operator ()(double nu);
    // Load (YAML) data
    virtual void load(const YAML::Node& data);
    // Log data
    virtual void log();
    // Create copy
    virtual BFSigmaRad* copy();
    // Get memory allocated - New!
    virtual double mem() const;
  };

  // Bound-free radiative cross-cection data (Seaton fit)
  class BFSigmaRadSeaton : public BFSigmaRad {
  public :
    // Edge frequency
    double nu0;
    // Fit parameters
    double sigma0;
    double alpha;
    double beta;
    // Frequency offset - New!
    //double nuOffset;
    // Get cross-section
    //virtual double operator ()(double nu) const;
    virtual double operator ()(double nu);
    // Load (YAML) data
    virtual void load(const YAML::Node& data);
    // Log data
    virtual void log();
    // Create copy
    virtual BFSigmaRad* copy();
    // Get memory allocated - New!
    virtual double mem() const;
  };

  // Bound-free radiative cross-cection data (Verner fit)
  // Note: Same as BFSigmaRadVerner95 except units
  class BFSigmaRadVerner : public BFSigmaRad {
  public :
    // Fit parameters
    vector<double> n;
    vector<double> l;
    vector<double> Eth;
    vector<double> E0;
    vector<double> sigma0;
    vector<double> ya;
    vector<double> P;
    vector<double> yw;
    // Get cross-section
    virtual double operator ()(double nu);
    // Load (YAML) data
    virtual void load(const YAML::Node& data);
    // Log data
    virtual void log();
    // Create copy
    virtual BFSigmaRad* copy();
    // Get memory allocated - New!
    virtual double mem() const;
  };

  // Bound-free radiative cross-cection data (Verner 1995 fit)
  class BFSigmaRadVerner95 : public BFSigmaRad {
  public :
    // Fit parameters
    vector<double> n;
    vector<double> l;
    vector<double> Eth;
    vector<double> E0;
    vector<double> sigma0;
    vector<double> ya;
    vector<double> P;
    vector<double> yw;
    // Get cross-section
    virtual double operator ()(double nu);
    // Load (YAML) data
    virtual void load(const YAML::Node& data);
    // Adjust edge - New!
    virtual void adjustEdge(const double& nuEdge);
    // Log data
    virtual void log();
    // Create copy
    virtual BFSigmaRad* copy();
    // Get memory allocated - New!
    virtual double mem() const;
  };

  // Bound-free radiative cross-cection data (Verner 1996 fit)
  class BFSigmaRadVerner96 : public BFSigmaRad {
  public :
    // Fit parameters
    double Eth;
    double Emax;
    double E0;
    double sigma0;
    double ya;
    double P;
    double yw;
    double y0;
    double y1;
    // Get cross-section
    virtual double operator ()(double nu);
    // Load (YAML) data
    virtual void load(const YAML::Node& data);
    // Adjust edge - New!
    virtual void adjustEdge(const double& nuEdge);
    // Log data
    virtual void log();
    // Create copy
    virtual BFSigmaRad* copy();
    // Get memory allocated - New!
    virtual double mem() const;
  };

  // Bound-free radiative cross-cection data (Verner 1995/1996 fit)
  class BFSigmaRadVerner9596 : public BFSigmaRad {
  public :
    BFSigmaRadVerner95 v95;
    BFSigmaRadVerner96 v96;
    // Get cross-section
    virtual double operator ()(double nu);
    // Load (YAML) data
    virtual void load(const YAML::Node& data);
    // Adjust edge - New!
    virtual void adjustEdge(const double& nuEdge);
    // Log data
    virtual void log();
    // Create copy
    virtual BFSigmaRad* copy();
    // Get memory allocated - New!
    virtual double mem() const;
  };

  // Bound-free collisional cross-section data
  class BFSigmaColl {
  public :
    // Get cross-section
    //virtual double operator ()(double nu) const {};
    virtual double operator ()(double nu) {};
    // Load (YAML) data
    virtual void load(const YAML::Node& data) {};
    // Log data
    virtual void log() {};
    // Create copy
    virtual BFSigmaColl* copy() {};
    // Get memory allocated - New!
    virtual double mem() const {};
  };
  
  // Bound-free collisional cross-cection data (table)
  class BFSigmaCollTable : public BFSigmaColl {
  public :
    // Cross-section
    Interpolator1D sigma;
    // Get cross-section
    //virtual double operator ()(double T) const;
    virtual double operator ()(double T);
    // Load (YAML) data
    virtual void load(const YAML::Node& data);
    // Log data
    virtual void log();
    // Create copy
    virtual BFSigmaColl* copy();
    // Get memory allocated - New!
    virtual double mem() const;
  };

  // Bound-free collisional cross-cection data (powerlaw)
  class BFSigmaCollPLaw : public BFSigmaColl {
  public :
    // Cross-section
    double sigma;
    // Frequency
    double T0;
    // Power law coefficient
    double p;
    // Get cross-section
    //virtual double operator ()(double T) const;
    virtual double operator ()(double T);
    // Load (YAML) data
    virtual void load(const YAML::Node& data);
    // Log data
    virtual void log();
    // Create copy
    virtual BFSigmaColl* copy();
    // Get memory allocated - New!
    virtual double mem() const;
  };

  // Bound-free collisional strength data
  class BFOmega {
  public :
    // Get cross-section
    //virtual double operator ()(double nu) const {};
    virtual double operator ()(double nu) {};
    // Load (YAML) data
    virtual void load(const YAML::Node& data) {};
    // Log data
    virtual void log() {};
    // Create copy
    virtual BFOmega* copy() {};
    // Get memory allocated - New!
    virtual double mem() const {};
  };

  // Bound-free collisional strength data (value)
  class BFOmegaValue : public BFOmega {
  public :
    // Collisional strength
    double omega;
    // Get collisional strength
    //virtual double operator ()(double T) const;
    virtual double operator ()(double T);
    // Load (YAML) data
    virtual void load(const YAML::Node& data);
    // Log data
    virtual void log();
    // Create copy
    virtual BFOmega* copy();
    // Get memory allocated - New!
    virtual double mem() const;
  };

  // Bound-free collisional strength (table)
  class BFOmegaTable : public BFOmega {
  public :
    // Collisional strength
    Interpolator1D omega;
    // Get cross-section
    //virtual double operator ()(double T) const;
    virtual double operator ()(double T);
    // Load (YAML) data
    virtual void load(const YAML::Node& data);
    // Log data
    virtual void log();
    // Create copy
    virtual BFOmega* copy();
    // Get memory allocated - New!
    virtual double mem() const;
  };

  // Bound-free collisional strength data (powerlaw)
  class BFOmegaPLaw : public BFOmega {
  public :
    // Collisional strength
    double omega;
    // Temperature
    double T0;
    // Power law coefficient
    double p;
    // Get cross-section
    //virtual double operator ()(double T) const;
    virtual double operator ()(double T);
    // Load (YAML) data
    virtual void load(const YAML::Node& data);
    // Log data
    virtual void log();
    // Create copy
    virtual BFOmega* copy();
    // Get memory allocated - New!
    virtual double mem() const;
  };

  // Interpolation type
  enum IPolType { ipolLinear, ipolLog };
  // Load interpolation type
  IPolType loadIPolType(const YAML::Node& data);
  // Log interpolation type
  string logIPolType(const IPolType& ipolType);

  // Extrapolation type
  enum EPolType { epolNone, epolConst };
  // Load extrapolation type
  EPolType loadEPolType(const YAML::Node& data);
  // Log extrapolation type
  string logEPolType(const EPolType& epolType);

  // Bound-free recombination coefficient data - New!
  class BFAlfa {
  public :
    // Get radiative cross-section
    virtual double operator ()(double T) {};
    // Load (YAML) data
    virtual void load(const YAML::Node& data) {};
     // Log data
    virtual void log() {};
    // Create copy
    virtual BFAlfa* copy() {};
    // Get memory allocated - New!
    virtual double mem() const {};
  };

  // Bound-free recombination coeffiecient data (value)
  class BFAlfaValue : public BFAlfa {
  public :
    // Recombination coefficient
    double alfa;
    // Get collisional strength
    virtual double operator ()(double T);
    // Load (YAML) data
    virtual void load(const YAML::Node& data);
    // Log data
    virtual void log();
    // Create copy
    virtual BFAlfa* copy();
    // Get memory allocated - New!
    virtual double mem() const;
  };

  // Bound-free recombination coefficient data (table) - New!
  class BFAlfaTable : public BFAlfa {
  public :
    // Cross-section
    Interpolator1D alfa;
    IPolType ipolType;
    EPolType epolType; 
    // Get recombination coefficient
    virtual double operator ()(double T);
    // Load (YAML) data
    virtual void load(const YAML::Node& data);
    // Log data
    virtual void log();
    // Create copy
    virtual BFAlfa* copy();
    // Get memory allocated - New!
    virtual double mem() const;
  };

  // Bound-free recombination coefficient data (powerlaw)
  class BFAlfaPLaw : public BFAlfa {
  public :
    // Recombination coeffiecient
    double alfa;
    // Temperature
    double T0;
    // Power law coefficient
    double p;
    // Get cross-section
    virtual double operator ()(double T);
    // Load (YAML) data
    virtual void load(const YAML::Node& data);
    // Log data
    virtual void log();
    // Create copy
    virtual BFAlfa* copy();
    // Get memory allocated - New!
    virtual double mem() const;
  };

  // Bound-free recombination coeffiecent data (Verner et al. 1996 fit)
  class BFAlfaVerner : public BFAlfa {
  public :
    // Fit parameters
    double a;
    double b;
    double T0;
    double T1;
    // Get recombination coefficient
    virtual double operator ()(double T);
    // Load (YAML) data
    virtual void load(const YAML::Node& data);
    // Log data
    virtual void log();
    // Create copy
    virtual BFAlfa* copy();
    // Get memory allocated - New!
    virtual double mem() const;
  };

  // Bound-free recombination coeffiecent data (Shull et al. 1982 fit)
  class BFAlfaShull : public BFAlfa {
  public :
    // Fit parameters
    double ARad;
    double XRad;
    double ADi;
    double BDi;
    double T0;
    double T1;
    // Get recombination coefficient
    virtual double operator ()(double T);
    // Load (YAML) data
    virtual void load(const YAML::Node& data);
    // Log data
    virtual void log();
    // Create copy
    virtual BFAlfa* copy();
    // Get memory allocated - New!
    virtual double mem() const;
  };

  // Bound-free recombination coeffiecent data (Pequignot et al. 1986 fit)
  class BFAlfaPequignot : public BFAlfa {
  public :
    // Fit parameters
    double alfa100;
    double eta;
    // Get recombination coefficient
    virtual double operator ()(double T);
    // Load (YAML) data
    virtual void load(const YAML::Node& data);
    // Log data
    virtual void log();
    // Create copy
    virtual BFAlfa* copy();
    // Get memory allocated - New!
    virtual double mem() const;
  };

  // Bound-free recombination coeffiecent data (Nussbaumer et al. 1986 fit)
  class BFAlfaNussbaumer : public BFAlfa {
  public :
    // Fit parameters
    double a;
    double b;
    double c;
    double d;
    double f;
    double tl;
    // Get recombination coefficient
    virtual double operator ()(double T);
    // Load (YAML) data
    virtual void load(const YAML::Node& data);
    // Log data
    virtual void log();
    // Create copy
    virtual BFAlfa* copy();
    // Get memory allocated - New!
    virtual double mem() const;
  };

  // Bound-free recombination coeffiecent data (Badnell 2006 fit)
  class BFAlfaBadnell : public BFAlfa {
  public :
    // Fit parameters
    double A;
    double B;
    double T0;
    double T1;
    double C;
    double T2;
    // Get recombination coefficient
    virtual double operator ()(double T);
    // Load (YAML) data
    virtual void load(const YAML::Node& data);
    // Log data
    virtual void log();
    // Create copy
    virtual BFAlfa* copy();
    // Get memory allocated - New!
    virtual double mem() const;
  };
  
  // Transiton data
  class TransData {
  public:
    // Transaction index
    int index;
    // Upper bound state
    BStateData* upperState;
    // Lower bound state
    BStateData* lowerState;
    // Transition energy
    double Xi;
    // Constructor
    TransData() {};
    // Desctructor
    virtual ~TransData() {};
  };

  // Bound-bound transition data
  class TransBBData : public TransData {
  public:
    // Einstein spontaneous de-excitation coefficient
    double A;
    // Two-photon Einstein spontaneous de-excitation coefficient - New!
    double A2;
    // Collisional strength
    BBOmega* omega;
    // Two-photon normalized emissivity - New!
    BBEta2* eta2;
    // Processes flags
    unsigned int processes;
    // Adjust edge flag
    bool edgeAdjust;
    // Constructor
    TransBBData();
    // Copy constructor
    TransBBData(const TransBBData& transData);
    // Assignment operator
    TransBBData& operator=(const TransBBData& transData);
    // Destructor
    virtual ~TransBBData();
    // Load (YAML) data
    void load(const YAML::Node& data);
    // Load two-photon transition data - New!
    void load2(const YAML::Node& data);
    // Update tables (using the provided frequency binning) - New!
    void updateTables(const vector<double>& nu);
    // Adjust edge - New!
    void adjustEdge();
    // Log data
    virtual void log();
    // Get memory allocated - New!
    double mem() const;
    // Add copy constructor and assignment operator !!!
  };

  // Detailed balance method
  enum DBMethod { dbNone, dbForceSigma, dbForceAlfa };
  // Load detailed balance method
  DBMethod loadDBMethod(const YAML::Node& data);
  // Log detailed balance method
  string logDBMethod(const DBMethod& dBMethod);

  // Bound-free transiton data
  class TransBFData : public TransData { 
  public :
    // Radiative cross-section
    BFSigmaRad* sigmaRad;
    // Colisional cross-section
    BFSigmaColl* sigmaColl;
    // Colisional strength
    BFOmega* omega;
    // Processes flags
    unsigned int processes;
    // Detailed balance method
    DBMethod dBMethod;
    // Adjust edge flag
    bool edgeAdjust;
    // Constructor
    TransBFData();
    // Copy constructor
    TransBFData(const TransBFData& transData);
    // Assignment operator
    TransBFData& operator=(const TransBFData& transData);
    // Destructor
    virtual ~TransBFData();
    // Load (YAML) data
    void load(const YAML::Node& data);
    // Update tables (using the provided frequency binning) - New!
    void updateTables(const vector<double>& nu);
    // Adjust edge - New!
    void adjustEdge();
    // Log data
    virtual void log();
    // Get memory allocated - New!
    double mem() const;
    // Add copy constructor and assignment operator !!!
};

  // Recombination distribution method
  enum AlfaDistMethod { gAdd, gReplace };
  // Load recombination distribution method
  AlfaDistMethod loadAlfaDistMethod(const YAML::Node& data);

  // Ionization state data
  class IStateData {
  public:
    // Constructor
    IStateData();
    // Copy constructor
    IStateData(const IStateData& iStateData);
    // Assignment operator
    IStateData& operator=(const IStateData& iStateData);
    // Ionization state index
    int index;
    // Atom charge
    double C;
    // Ionization potential from ground state
    double Xi0;
    // List of bound states
    vector<BStateData> bState;
    // List of bound-bound transitions
    vector<TransBBData> transBB;
    // List of bound-free transitions
    //vector<TransBFData> transBF;
    // Recombination coefficient - New!
    BFAlfa* alfa;
    // Di-electronic recombination coefficient - New!
    BFAlfa* alfaDi;
    // Recombination distribution method - New!
    AlfaDistMethod alfaDistMethod;
    // Force recombination coefficient match
    bool alfaForceEqual;
    // LTE reference state index
    int LTERefBStateIndex,LTERefIStateIndex; // New!
    // Load (YAML) data
    void load(const YAML::Node& data, const string& dataDir);
    // Load recombination distribution data
    void loadAlfaDist(const YAML::Node& data);
    // Log data
    virtual void log();
    // Get memory allocated - New!
    double mem() const;
  };
  
  // Ionization state data list
  class IStateDataList : public vector<IStateData> {
  public :
    // Find ionization state data
    IStateData* find(const int& C);
  };
  
  // Isotope data
  class IsotopeData {
  public :
    // Isotope index
    int index;
    // Number of neutrons
    int N;
    // Atomic mass
    double m;
    // Load (YAML) data
    void load(const YAML::Node& data);
    // Log data
    virtual void log();
    // Get memory allocated - New!
    double mem() const;
  };
  
  // Isotope data list
  class IsotopeDataList : public vector<IsotopeData> {
  public :
    // Find isotope data
    IsotopeData* find(const int& N);
  };
  
  // Atom data
  class AtomData {
  public :
    // Constructor
    AtomData();
    // Copy constructor
    AtomData(const AtomData& atomData);
    // Assignment operator
    AtomData& operator=(const AtomData& atomData);
    // Atom index
    int index;
    // Number of protons
    int Z;
    // List of isotopes
    IsotopeDataList isotope;
    // List of ionisation states
    IStateDataList iState;
    // List of bound-free transitions
    vector<TransBFData> transBF;
    // Update energy above ground state of neutral atom
    void updateETot();
    // Set LTE reference state pointers
    void setLTERefState();
    // Load (YAML) data
    void load(const YAML::Node& data, const string& dataDir);
    // Log data
    virtual void log();
    // Get memory allocated - New!
    double mem() const;
  };

  // Number density type
  enum NStateType { nBState, nIState };

  // Charge-transfer reaction coefficient data - New!
  class CTAlfa {
  public :
    // Get reaction coeffiecient
    virtual double operator ()(double T) {};
    // Load (YAML) data
    virtual void load(const YAML::Node& data) {};
     // Log data
    virtual void log() {};
    // Create copy
    virtual CTAlfa* copy() {};
    // Get memory allocated - New!
    virtual double mem() const {};
  };

  // Charge-transfer reaction coefficient data (value)
  class CTAlfaValue : public CTAlfa {
  public :
    // Reaction coefficient
    double alfa;
    // Get reaction coefficient
    virtual double operator ()(double T);
    // Load (YAML) data
    virtual void load(const YAML::Node& data);
    // Log data
    virtual void log();
    // Create copy
    virtual CTAlfa* copy();
    // Get memory allocated - New!
    virtual double mem() const;
  };

  // Charge-transfer reaction coefficient data (table) - New!
  class CTAlfaTable : public CTAlfa {
  public :
    // Reaction coefficient
    Interpolator1D alfa;
    IPolType ipolType;
    EPolType epolType; 
    // Get reaction coefficient
    virtual double operator ()(double T);
    // Load (YAML) data
    virtual void load(const YAML::Node& data);
    // Log data
    virtual void log();
    // Create copy
    virtual CTAlfa* copy();
    // Get memory allocated - New!
    virtual double mem() const;
  };

  // Charge-transfer reaction coefficient data (powerlaw) - New!
  class CTAlfaPLaw : public CTAlfa {
  public :
    // Reaction coefficient
    double alfa;
    // Temperature
    double T0;
    // Power law coefficient
    double p;
    // Get cross-section
    virtual double operator ()(double nu);
    // Load (YAML) data
    virtual void load(const YAML::Node& data);
    // Log data
    virtual void log();
    // Create copy
    virtual CTAlfa* copy();
    // Get memory allocated - New!
    virtual double mem() const;
  };

  // Arnaud et al. 1985 fit type
  enum ArnaudFitType { fitTypeA, fitTypeB, fitTypeOH };

  // Log Arnaud et al. 1985 fit type
  string logArnaudFitType(ArnaudFitType fitType);

  // Charge-transfer reaction coefficient data (Arnud et al. 1985 fit) - New!
  class CTAlfaArnaud : public CTAlfa {
  public :
    // Fitting type
    ArnaudFitType fitType;
    // Fitting parameters
    double a;
    double b;
    double c;
    double d;
    double deltaE;
    // Get reaction coefficient
    virtual double operator ()(double T);
    // Load (YAML) data
    virtual void load(const YAML::Node& data);
    // Log data
    virtual void log();
    // Create copy
    virtual CTAlfa* copy();
    // Get memory allocated - New!
    virtual double mem() const;
  };

  // Charge-transfer transition - New!
  class CTTransData {
  public :
    AtomData* atom;
    IStateData* upperIState;
    BStateData* upperBState;
    IStateData* lowerIState;
    BStateData* lowerBState;
    // Constructor
    CTTransData();
    // Destructor
    ~CTTransData();
    // Add to bound-free transition list (if necessary)
    void add();
  };

  // Charge-transfer reaction data - New!
  class CTReactionData {
  public :
    // Reaction index
    int index;
    // Upward transition
    CTTransData upTrans;
    // Downward transition
    CTTransData downTrans;
    // Reaction coefficient
    CTAlfa* alfa;
    // Number density type
    NStateType nType;
    // Constructor
    CTReactionData();
    // Copy constructor
    CTReactionData(const CTReactionData& reactionData);
    // Assignment operator
    CTReactionData& operator=(const CTReactionData& reactionData);
    // Destructor
    virtual ~CTReactionData();
    // Load (YAML) data
    void load(const YAML::Node& data);
    // Adjust transitions automatically
    void adjustTrans();
    // Log data
    virtual void log();   
    // Get memory allocated - New!
    double mem() const;
  };
  
  // Atom data file format
  enum AtomDataFormat { aJ, yAML };
  
  // Atom data list
  class AtomDataList {
  private :
    // Load charge-transfer transition data - New!
    CTTransData loadCTTrans(const YAML::Node& data);
    // Update bound-free transition (pointer) lists - New!
    void updateBFTrans();
  public :
    // Constructor - New!
    AtomDataList();
    // Copy constructor - New!
    AtomDataList(const AtomDataList& atomDataList);
    // Assignment operator - New!
    AtomDataList& operator=(const AtomDataList& atomDataList);
    // List of Atoms
    vector<AtomData> atom;
    // List of charge-transfer reactions
    vector<CTReactionData> ctReaction;
    // Load atom data
    void load(const string& dataFile, const AtomDataFormat& format, const bool& loadIStateData);
    // Log atom data
    virtual void log();
    // Load YAML atom data
    void loadYAML(const string& dataFile, const bool& loadIStateData);
    // Load A Jerskstrand atom data
    // Deprecated!
    void loadAJ(const string& dataFile, const bool& loadIStateData);
    // Load A Jerkstrand ionization state data
    // Deprecated!
    void loadIonStateAJ(IStateData& iState, string dataFile);
    // Find atom
    AtomData* find(const int& Z);
    // Resample bound-free tables
    // Deprecated - Use updateTables!
    void resampleBFTables(const vector<double>& nu);
    // Update tables (using the provided frequency binning) - New!
    void updateTables(const vector<double>& nu);
    // Get memory allocated - New!
    double mem() const;
  };
}

#endif
