#ifndef _COMMON_
#define _COMMON_

#include <vector>
#include <boost/mpi.hpp>
#include <boost/serialization/string.hpp>
#include "mc-util.h"
#include "mc-math.h"
#include "mc-physics.h"

using namespace std;
using namespace util;
using namespace math;
using namespace physics;

namespace common {

  // No bin
  const int noBin=-1;
  
  // No frequency
  const double noNu=-1;

  // No (positive) number
  const int noN=-1;
 
  // Frequency bin
  class FrequencyBin {
  private :
    // MPI serialization
    friend class boost::serialization::access;
    template<class Archive> 
      void serialize (Archive & ar, const unsigned int version) 
      {
	ar & min;
	ar & max;
	ar & edge;
	// Testing
	ar & minorEdge;
      }
  public :
    // Constructors
    //FrequencyBin(const double& _min, const double& _max, bool _edge=false);
    // Testing
    FrequencyBin(const double& _min, const double& _max, bool _edge=false, bool _minorEdge=false);
    FrequencyBin();
    // Minimum frequency
    double min;
    // Maximum frequency
    double max;
    // Edge flag
    bool edge;
    // Testing
    // Minor (egde) flag
    bool minorEdge;
  };

  // Frequency bin list iterator
  typedef vector<FrequencyBin>::iterator FrequencyBinIterator;

  // Frequency bin list const iterator
  typedef vector<FrequencyBin>::const_iterator FrequencyBinConstIterator;

  // Frequency bin list
  class FrequencyBinList : public vector<FrequencyBin> {
  public :
    // Find frequency bin
    FrequencyBinIterator find(const double& _nu, int iGuess=noBin) const;
  };

  // Angle bin
  class AngleBin {
  private :
    // MPI serialization
    friend class boost::serialization::access;
    template<class Archive> 
      void serialize (Archive & ar, const unsigned int version) 
      {
	ar & min;
	ar & max;
      }
  public :
    // Constructors
    AngleBin(const double& _min, const double& _max);
    AngleBin();
    // Minimum frequency
    double min;
    // Maximum frequency
    double max;
  };

  // Angle bin list iterator
  typedef vector<AngleBin>::iterator AngleBinIterator;

  // Angle bin list
  class AngleBinList : public vector<AngleBin> {
  public :
    // Find angle bin
    AngleBinIterator find(const double& _mu, int iGuess=noBin) const;
  };  

  enum EdgeType { noEdge, blueEdge, redEdge };

  //const edgeWidth=2e-6;

  inline double nuEdge(FrequencyBinIterator nu, EdgeType edgeType) {
    return (edgeType==noEdge)?nu->min:((edgeType==redEdge)?(nu->min*(1.0-1e-6)):nu->min*(1.0+1e-6));
  }

  inline double nuEdge(double nu, EdgeType edgeType) {
    return (edgeType==noEdge)?nu:((edgeType==redEdge)?(nu*(1.0-1e-6)):nu*(1.0+1e-6));
  }

  /*enum NuType { ntVal, ntBin, nuBoth }
  
  class Nu {
    FrequencyBinIterator bin;
    double val;
    NuType type;
    EdgeType edgeType;
    Nu(const double& _value) : val(_val), type(ntVal) { }
    Nu(const FrequencyBinIterator& _bin) : bin(_bin), type(ntBin) {}
    Nu(const FrequencyBinIterator& _bin, const double& _val) : bin(_bin), val(_val), type(ntBoth) { }
    }*/

  // Frequency tabulator class
  template <class F> class NuTabulator {
  private :
    FrequencyBinIterator nuBegin;
    FrequencyBinIterator nuEnd;
    vector<double> fTable;
    vector<double> fRedTable;
    vector<double> fBlueTable;
    F f;
    // Testing
    int nBin;
    double edgeWidth;
  public :
    NuTabulator() {};
    ~NuTabulator() {};
    NuTabulator(FrequencyBinIterator _nuBegin, FrequencyBinIterator _nuEnd, const F& _f, double _edgeWidth=2e-6);
    void fixFTables(); // New!
    void init(FrequencyBinIterator _nuBegin, FrequencyBinIterator _nuEnd, const F& _f, double _edgeWidth=2e-6);
    void initOMP(FrequencyBinIterator _nuBegin, FrequencyBinIterator _nuEnd, const F& _f, double _edgeWidth=2e-6);
    vector<double>::const_iterator find(const double& nu) const;
    double operator ()(FrequencyBinIterator nuBin, EdgeType edgeType=noEdge) const;
    double operator ()(double nu, EdgeType edgeType=noEdge) const;
    double mem() const; // New!
  };

  template <class F> NuTabulator<F>::NuTabulator(FrequencyBinIterator _nuBegin, FrequencyBinIterator _nuEnd, const F& _f, double _edgeWidth) : nuBegin(_nuBegin), nuEnd(_nuEnd), f(_f), edgeWidth(_edgeWidth) {
    int i;
    
    // If any frequency bins to tabulate
    if (nuEnd>nuBegin) {
      // Initialize function table
      // Remove +1, right?
      fTable.resize(nuEnd-nuBegin+1);
      fRedTable.resize(nuEnd-nuBegin+1);
      fBlueTable.resize(nuEnd-nuBegin+1);
      // Calculate function values at the beginning of all frequency bins
      for (i=0;i<(nuEnd-nuBegin);i++) {
	//if (not (nuBegin+i)->minorEdge) { // New!
	fTable[i]=f(nuEdge(nuBegin+i,noEdge));
	if ((nuBegin+i)->edge) fRedTable[i]=f(nuEdge(nuBegin+i,redEdge));
	else fRedTable[i]=fTable[i];
	if ((nuBegin+i)->edge) fBlueTable[i]=f(nuEdge(nuBegin+i,blueEdge));
	else fBlueTable[i]=fTable[i];
	//}
      }
      // Fix function tables - New!
      //fixFTables();
      // Calculate funcion values at the end of the last frequency bin
      fTable.push_back(f(nuEdge((nuEnd-1)->max,noEdge)));
      fRedTable.push_back(f(nuEdge((nuEnd-1)->max,redEdge)));
      fBlueTable.push_back(f(nuEdge((nuEnd-1)->max,blueEdge)));
    }
    // Else
    else {
      // Reset function table
      fTable.resize(0);
      fRedTable.resize(0);
      fBlueTable.resize(0);
    }
  }

  // Fix function tables - New! - Testing!
  template <class F> void NuTabulator<F>::fixFTables() {
    Interpolator1D fIpol;
    vector<FunctionValue> fIpolTable(0);
    int i;

    for (i=0;i<(nuEnd-nuBegin);i++) {
      if (not (nuBegin+i)->minorEdge) {
	if ((nuBegin+i)->edge) { 
	  fIpolTable.push_back(FunctionValue(nuEdge(nuBegin+i,redEdge),fRedTable[i]));
	  fIpolTable.push_back(FunctionValue(nuEdge(nuBegin+i,blueEdge),fBlueTable[i]));
	}
	else fIpolTable.push_back(FunctionValue(nuEdge(nuBegin+i,noEdge),fTable[i]));
	}
    }
    fIpolTable.push_back(FunctionValue(nuEdge(nuEnd-1,noEdge),f(nuEdge(nuEnd-1,noEdge))));
    fIpol.init(fIpolTable);

    for (i=0;i<(nuEnd-nuBegin);i++) {
      if ((nuBegin+i)->minorEdge) {
	fTable[i]=fIpol(nuEdge(nuBegin+i,noEdge));
	if (nuEdge(nuBegin+i,noEdge)==0) cout<<"NuTabulator::fixFTables: nuEnd-nuBegin: "<<nuEnd-nuBegin<<", i: "<<i<<", zero frequency\n";
	fRedTable[i]=fTable[i]; fBlueTable[i]=fTable[i];
      }
      else if (not (nuBegin+i)->edge) { fRedTable[i]=fTable[i]; fBlueTable[i]=fTable[i]; }
    }
  }
  
  template <class F> void NuTabulator<F>::init(FrequencyBinIterator _nuBegin, FrequencyBinIterator _nuEnd, const F& _f, double _edgeWidth) {
    int i;
    
    // Initialize
    nuBegin=_nuBegin;
    nuEnd=_nuEnd;
    f=_f;
    edgeWidth=_edgeWidth;
    
    // If any frequency bins to tabulate
    if (nuEnd>nuBegin){
      // Initialize function table
      // Remove +1, right?
      fTable.resize(nuEnd-nuBegin+1);
      fRedTable.resize(nuEnd-nuBegin+1);
      fBlueTable.resize(nuEnd-nuBegin+1);
      // Calculate function values at the beginning of all frequency bins
      for (i=0;i<(nuEnd-nuBegin);i++) {
	//if (not (nuBegin+i)->minorEdge) { // New!
	fTable[i]=f(nuEdge(nuBegin+i,noEdge));
	if ((nuBegin+i)->edge) fRedTable[i]=f(nuEdge(nuBegin+i,redEdge));
	else fRedTable[i]=fTable[i];
	if ((nuBegin+i)->edge) fBlueTable[i]=f(nuEdge(nuBegin+i,blueEdge));
	else fBlueTable[i]=fTable[i];
	//}
      }
      // Fix function tables - New!
      //fixFTables();
      // Calculate function values at the end of the last frequency bin
      fTable.push_back(f(nuEdge((nuEnd-1)->max,noEdge)));
      fRedTable.push_back(f(nuEdge((nuEnd-1)->max,redEdge)));
      fBlueTable.push_back(f(nuEdge((nuEnd-1)->max,blueEdge)));
    }
    // Else
    else {
      // Reset function table
      fTable.resize(0);
      fRedTable.resize(0);
      fBlueTable.resize(0);
    }
  }
  
  template <class F> void NuTabulator<F>::initOMP(FrequencyBinIterator _nuBegin, FrequencyBinIterator _nuEnd, const F& _f, double _edgeWidth) {
    int i;
    
    // Initialize
    nuBegin=_nuBegin;
    nuEnd=_nuEnd;
    f=_f;
    edgeWidth=_edgeWidth;
    
    // If any frequency bins to tabulate
    if (nuEnd>nuBegin){
      // Calculate function values at the beginning of all frequency bin
      fTable=f(nuBegin,nuEnd,noEdge);
      fRedTable=f(nuBegin,nuEnd,redEdge);
      fBlueTable=f(nuBegin,nuEnd,blueEdge);
      // Fix function tables - New!
      //fixFTables();
      // To be on the safe side - Replace zeros in red and blue tables - Removed
      for (i=0;i<(nuEnd-nuBegin);i++)
	if (not ((nuBegin+i)->edge)) { fRedTable[i]=fTable[i]; fBlueTable[i]=fTable[i]; }
      // Calculate function values at the end of the last frequency bin
      fTable.push_back(f(nuEdge((nuEnd-1)->max,noEdge)));
      fRedTable.push_back(f(nuEdge((nuEnd-1)->max,redEdge)));
      fBlueTable.push_back(f(nuEdge((nuEnd-1)->max,blueEdge)));
    }
    // Else
    else {
      // Reset function table
      fTable.resize(0);
      fRedTable.resize(0);
      fBlueTable.resize(0);
    }
  }
  
  template <class F> vector<double>::const_iterator NuTabulator<F>::find(const double& nu) const {
    long iBegin,iMiddle,iEnd;
    
    if (nuEnd>nuBegin) {
      iBegin=0; iEnd=(nuEnd-nuBegin)-1; iMiddle=(iBegin+iEnd)/2;
      if (nu>=(nuBegin+iBegin)->min && nu<(nuBegin+iEnd)->min) {
	while (iEnd-iBegin>1) {
	  if (nu<(nuBegin+iMiddle)->min) {
	    iEnd=iMiddle;
	    iMiddle=(iBegin+iEnd)/2;
	  }
	  else {
	    iBegin=iMiddle;
	    iMiddle=(iBegin+iEnd)/2;
	  }
	}
      if (nu==(nuBegin+iBegin)->min) return fTable.begin()+iBegin;
      }
      else if (nu==(nuBegin+iEnd)->min) return fTable.begin()+iEnd;
      else if (nu==(nuBegin+iEnd)->max) return fTable.begin()+iEnd+1;
    }
    
    return fTable.end();
  }

  template <class F> double NuTabulator<F>::operator ()(FrequencyBinIterator nuBin, EdgeType edgeType) const {
    int i;
    
    if (nuBin>=nuBegin and nuBin<=nuEnd) {
      //if (nuBin->minorEdge) cout<<"NuTabulator: Warning: Minor edge\n";
      //if (nuBin->minorEdge) return f(nuEdge(nuBin,edgeType)); // New! - Testing!
      if (edgeType==noEdge) return fTable[(nuBin-nuBegin)];
      else if (edgeType==redEdge) return fRedTable[(nuBin-nuBegin)];
    else return fBlueTable[(nuBin-nuBegin)];
    }
    else { cout<<"NuTabulator: Error: Outside tabulated frequency range, nuEnd-nuBegin: "<<nuEnd-nuBegin<<", nuBegin-nuBin: "<<nuBegin-nuBin<<", nuBin-nuEnd: "<<nuBin-nuEnd<<"\n"; return 0; }
  }
  
  template <class F> double NuTabulator<F>::operator ()(double nu, EdgeType edgeType) const {
    vector<double>::const_iterator fIterator=find(nu);
    int i=fIterator-fTable.begin();
    
    if (fIterator==fTable.end()) return f(nuEdge(nu,edgeType));
    else {
      //if ((nuBegin+i)->minorEdge) cout<<"NuTabulator: Warning: Minor egde\n";
      //if ((nuBegin+i)->minorEdge) return f(nuEdge(nu,edgeType)); // New! - Testing!
      if (edgeType==noEdge) return fTable[i];
      else if (edgeType==redEdge) return fRedTable[i];
      else return fBlueTable[i];
    }
  }

  template <class F> double NuTabulator<F>::mem() const { // New!
    double _mem=0;
    int i;
    for (i=0;i<fTable.size();i++) _mem+=sizeof(double);
    for (i=0;i<fRedTable.size();i++) _mem+=sizeof(double);
    for (i=0;i<fBlueTable.size();i++) _mem+=sizeof(double);
    return _mem;
  }
  
  // Frequency interpolator MPI class
  class NuInterpolatorMPI {
  private :
    // MPI serialization
    friend class boost::serialization::access;
    template<class Archive> 
      void serialize (Archive & ar, const unsigned int version) 
      {
	ar & iBegin;
	ar & iEnd;
	ar & fTable;
	ar & fRedTable;
	ar & fBlueTable;
	ar & nBin;
	ar & edgeWidth;
      }
  public :
    int iBegin;
    int iEnd;
    vector<double> fTable;
    vector<double> fRedTable;
    vector<double> fBlueTable;
    int nBin;
    double edgeWidth;
  };

  // Frequency interpolator class
  template <class F> class NuInterpolator {
  private :
    FrequencyBinIterator nuBegin;
    FrequencyBinIterator nuEnd;
    vector<double> fTable;
    vector<double> fRedTable;
    vector<double> fBlueTable;
    F f;
    // Testing
    int nBin;
    double edgeWidth;
  public :
    NuInterpolator() {};
    ~NuInterpolator() {};
    NuInterpolator(FrequencyBinIterator _nuBegin, FrequencyBinIterator _nuEnd, const F& _f, double _edgeWidth=2e-6);
    //void initMap(FrequencyBinIterator _nuBeginOrig, FrequencyBinIterator _nuEndOrig); // Removed!
    //FrequencyBinIterator mapNuBin(FrequencyBinIterator nuBinOrig, bool& minorEdge) const; // Removed!
    void fixFTables(); // New!
    void init(FrequencyBinIterator _nuBegin, FrequencyBinIterator _nuEnd, const F& _f, double _edgeWidth=2e-6);
    void initOMP(FrequencyBinIterator _nuBegin, FrequencyBinIterator _nuEnd, const F& _f, double _edgeWidth=2e-6);
    vector<double>::const_iterator find(const double& nu) const;
    double operator ()(FrequencyBinIterator nuBin, EdgeType edgeType=noEdge) const;
    double operator ()(double nu) const;
    double operator ()(FrequencyBinIterator nuBin, double nu) const;
    double mem() const; // New!
    void pack(NuInterpolatorMPI& nuInterpolatorMPI, FrequencyBinList& nu);
    void unpack(NuInterpolatorMPI& nuInterpolarMPI, FrequencyBinList& nu, const F& _f);
  };

  template <class F> NuInterpolator<F>::NuInterpolator(FrequencyBinIterator _nuBegin, FrequencyBinIterator _nuEnd, const F& _f, double _edgeWidth) : nuBegin(_nuBegin), nuEnd(_nuEnd), f(_f), edgeWidth(_edgeWidth) { init(nuBegin,nuEnd,f,edgeWidth); }

  // Init frequency bin list map - New! - Testing! - Remove!
  /*template <class F> void NuInterpolator<F>::initMap(FrequencyBinIterator _nuBeginOrig, FrequencyBinIterator _nuEndOrig) {
    int i,ii=0;

    // Initialize
    nuBeginOrig=_nuBeginOrig;
    nuEndOrig=_nuEndOrig;
    
    // If any frequency bins to tabulate
    if (nuEndOrig>nuBeginOrig){
      // Initiate index map
      iMap.resize(nuEndOrig-nuBeginOrig);
      // Make index map
      for (i=0;i<iMap.size();i++)
	if ((not ((nuBeginOrig+i)->edge and (nuBeginOrig+i)->minorEdge)) or i==iMap.size()-1) iMap[i]=ii++;
	else iMap[i]=-ii;
      // Make inverse index map
      for (i=0;i<iMap.size();i++) if (iMap[i]>=0) iMapInv.push_back(i);
      // Make (new) frequency bin list
      nuBinList.resize(iMapInv.size());
      for (i=0;i<iMapInv.size();i++) {
	nuBinList[i].min=(nuBeginOrig+iMapInv[i])->min;
	if (i<iMapInv.size()-1) nuBinList[i].max=(nuBeginOrig+iMapInv[i+1])->min;
	else nuBinList[i].max=(nuBeginOrig+iMapInv[i])->max;
      }
    }

    nuBegin=nuBinList.begin();
    nuEnd=nuBinList.end();
    }*/

  // Map frequency bin - New! - Testing! - Remove!
  /*template <class F> FrequencyBinIterator NuInterpolator<F>::mapNuBin(FrequencyBinIterator nuBinOrig, bool& minorEdge) const {
    int i=nuBinOrig-nuBinOrig;
    if (i<0 or i>=iMap.size()) return nuEnd;
    int ii=iMap[i];
    if (ii>=0) { minorEdge=false; return nuBegin+ii; }
    else { minorEdge=true; return nuBegin-ii; }
    }*/

  // Fix function tables - New! - Testing!
  template <class F> void NuInterpolator<F>::fixFTables() {
    Interpolator1D fIpol;
    vector<FunctionValue> fIpolTable(0);
    int i;

    for (i=0;i<(nuEnd-nuBegin);i++) {
      if (not (nuBegin+i)->minorEdge) {
	if ((nuBegin+i)->edge) { 
	  fIpolTable.push_back(FunctionValue(nuEdge(nuBegin+i,redEdge),fRedTable[i]));
	  fIpolTable.push_back(FunctionValue(nuEdge(nuBegin+i,blueEdge),fBlueTable[i]));
	}
	else fIpolTable.push_back(FunctionValue(nuEdge(nuBegin+i,noEdge),fTable[i]));
      }
    }
    fIpolTable.push_back(FunctionValue(nuEdge(nuEnd-1,noEdge),f(nuEdge(nuEnd-1,noEdge))));
    fIpol.init(fIpolTable);

    for (i=0;i<(nuEnd-nuBegin);i++) {
      if ((nuBegin+i)->minorEdge) {
	fTable[i]=fIpol(nuEdge(nuBegin+i,noEdge));
	if (nuEdge(nuBegin+i,noEdge)==0) cout<<"NuInterpolator::fixFTables: nuEnd-nuBegin: "<<nuEnd-nuBegin<<", i: "<<i<<", zero frequency\n";
	fRedTable[i]=fTable[i]; fBlueTable[i]=fTable[i];
      }
      else if (not (nuBegin+i)->edge) { fRedTable[i]=fTable[i]; fBlueTable[i]=fTable[i]; }
    }
  }
      
  template <class F> void NuInterpolator<F>::init(FrequencyBinIterator _nuBegin, FrequencyBinIterator _nuEnd, const F& _f, double _edgeWidth) {
    int i;
    
    // Initialize
    nuBegin=_nuBegin;
    nuEnd=_nuEnd;
    f=_f;
    edgeWidth=_edgeWidth;
    
    // If any frequency bins to tabulate
    if (nuEnd>nuBegin){
      // Initialize function table
      // Remove +1, right?
      fTable.resize(nuEnd-nuBegin+1);
      fRedTable.resize(nuEnd-nuBegin+1);
      fBlueTable.resize(nuEnd-nuBegin+1);
      // Calculate function values at the beginning of all frequency bins
      for (i=0;i<(nuEnd-nuBegin);i++) {
	//if (not (nuBegin+i)->minorEdge) { // New!
	fTable[i]=f(nuEdge(nuBegin+i,noEdge));
	if ((nuBegin+i)->edge) fRedTable[i]=f(nuEdge(nuBegin+i,redEdge));
	else fRedTable[i]=fTable[i];
	if ((nuBegin+i)->edge) fBlueTable[i]=f(nuEdge(nuBegin+i,blueEdge));
	else fBlueTable[i]=fTable[i];
	//}
      }
      // Fix function tables - New!
      //fixFTables();
      // Calculate function values at the end of the last frequency bin
      fTable.push_back(f(nuEdge((nuEnd-1)->max,noEdge)));
      fRedTable.push_back(f(nuEdge((nuEnd-1)->max,redEdge)));
      fBlueTable.push_back(f(nuEdge((nuEnd-1)->max,blueEdge)));
    }
    // Else
    else {
      // Reset function table
      fTable.resize(0);
      fRedTable.resize(0);
      fBlueTable.resize(0);
    }
  }

  template <class F> void NuInterpolator<F>::initOMP(FrequencyBinIterator _nuBegin, FrequencyBinIterator _nuEnd, const F& _f, double _edgeWidth) {
    vector<double> fTableOrig,fRedTableOrig,fBlueTableOrig; // New!
    int i;
        
    // Initialize
    nuBegin=_nuBegin;
    nuEnd=_nuEnd;
    f=_f;
    edgeWidth=_edgeWidth;
    
    // If any frequency bins to tabulate
    if (nuEnd>nuBegin){
      // Calculate function values at the beginning of all frequency bin
      fTable=f(nuBegin,nuEnd,noEdge);
      fRedTable=f(nuBegin,nuEnd,redEdge);
      fBlueTable=f(nuBegin,nuEnd,blueEdge);
      // Fix function tables - New!
      //fixFTables();
      // To be on the safe side - Replace zeros in red and blue tables - Removed!
      for (i=0;i<(nuEnd-nuBegin);i++)
	if (not ((nuBegin+i)->edge)) { fRedTable[i]=fTable[i]; fBlueTable[i]=fTable[i]; }
      // Calculate function values at the end of the last frequency bin
      fTable.push_back(f(nuEdge((nuEnd-1)->max,noEdge)));
      fRedTable.push_back(f(nuEdge((nuEnd-1)->max,redEdge)));
      fBlueTable.push_back(f(nuEdge((nuEnd-1)->max,blueEdge)));
    }
    // Else
    else {
      // Reset function table
      fTable.resize(0);
      fRedTable.resize(0);
      fBlueTable.resize(0);
    }
  }
  
  template <class F> vector<double>::const_iterator NuInterpolator<F>::find(const double& nu) const {
    long iBegin,iMiddle,iEnd;
    
    if (nuEnd>nuBegin) {
      iBegin=0; iEnd=(nuEnd-nuBegin)-1; iMiddle=(iBegin+iEnd)/2;
      if (nu>=(nuBegin+iBegin)->min && nu<(nuBegin+iEnd)->min) {
	while (iEnd-iBegin>1) {
	  if (nu<(nuBegin+iMiddle)->min) {
	    iEnd=iMiddle;
	    iMiddle=(iBegin+iEnd)/2;
	  }
	  else {
	    iBegin=iMiddle;
	    iMiddle=(iBegin+iEnd)/2;
	  }
	}
	return fTable.begin()+iBegin;
      }
      else if (nu<=(nuBegin+iEnd)->max) return fTable.begin()+iEnd;
    }
    
    return fTable.end();
  }

  template <class F> double NuInterpolator<F>::operator ()(FrequencyBinIterator nuBin, EdgeType edgeType) const {
    int i;
    
    if (nuBin>=nuBegin and nuBin<=nuEnd) {
      //if (nuBin->minorEdge) cout<<"NuInterpolator: Warning: Minor edge\n";
      if (edgeType==noEdge) return fTable[(nuBin-nuBegin)];
      else if (edgeType==redEdge) return fRedTable[(nuBin-nuBegin)];
      else return fBlueTable[(nuBin-nuBegin)];
    }
    else { cout<<"NuInterpolator: Error: Outside tabulated frequency range, nuEnd-nuBegin: "<<nuEnd-nuBegin<<", nuBegin-nuBin: "<<nuBegin-nuBin<<", nuBin-nuEnd: "<<nuBin-nuEnd<<"\n"; return 0; }
  }
  
  template <class F> double NuInterpolator<F>::operator ()(double nu) const {
    vector<double>::const_iterator fIterator=find(nu);
    int i=fIterator-fTable.begin();
    double _f,dfdx;
    
    if (fIterator==fTable.end()) return f(nu);
    else {
      //if ((nuBegin+i)->minorEdge) cout<<"NuInterpolator: Warning: Minor egde\n";
      dfdx=(fRedTable[i+1]-fBlueTable[i])/(nuEdge((nuBegin+i)->max,redEdge)-nuEdge((nuBegin+i)->min,blueEdge));
      _f=fBlueTable[i]+dfdx*(nu-nuEdge((nuBegin+i)->min,blueEdge));
    }
    
    return _f;
  }
  
  template <class F> double NuInterpolator<F>::operator ()(FrequencyBinIterator nuBin, double nu) const {
    double _f,dfdx;
    
    if (nuBin>=nuBegin and nuBin<nuEnd) {
      //if (nuBin->minorEdge) cout<<"NuInterpolator: Warning: Minor edge\n";
      dfdx=(fRedTable[nuBin-nuBegin+1]-fBlueTable[nuBin-nuBegin])/(nuEdge(nuBin->max,redEdge)-nuEdge(nuBin->min,blueEdge));
      _f=fBlueTable[nuBin-nuBegin]+dfdx*(nu-nuEdge(nuBin->min,blueEdge));
    }
    else { cout<<"NuInterpolator: Error: Outside tabulated frequency range, nuEnd-nuBegin: "<<nuEnd-nuBegin<<", nuBegin-nuBin: "<<nuBegin-nuBin<<", nuBin-nuEnd: "<<nuBin-nuEnd<<"\n"; return 0; }
   
    return _f;
  }

  template <class F> double NuInterpolator<F>::mem() const { // New!
    double _mem=0;
    int i;
    for (i=0;i<fTable.size();i++) _mem+=sizeof(double);
    for (i=0;i<fRedTable.size();i++) _mem+=sizeof(double);
    for (i=0;i<fBlueTable.size();i++) _mem+=sizeof(double);
    return _mem;
  }

  // Pack frequency interpolator class (MPI)
  template <class F> void NuInterpolator<F>::pack(NuInterpolatorMPI& nuInterpolatorMPI, FrequencyBinList& nu) {
    nuInterpolatorMPI.iBegin=nuBegin-nu.begin();
    nuInterpolatorMPI.iEnd=nuEnd-nu.begin();
    nuInterpolatorMPI.fTable=fTable;
    nuInterpolatorMPI.fRedTable=fRedTable;
    nuInterpolatorMPI.fBlueTable=fBlueTable;
    nuInterpolatorMPI.nBin=nBin;
    nuInterpolatorMPI.edgeWidth=edgeWidth;
    //cout<<"NuInterpolator::pack: cell: "<<cellIndex<<", iBegin; "<<nuInterpolatorMPI.iBegin<<", iEnd: "<<nuInterpolatorMPI.iEnd<<", nu.size: "<<nu.size()<<", fTable.size: "<<fTable.size()<<", fRedTable.size: "<<fRedTable.size()<<", fBlueTable.size: "<<fBlueTable.size()<<", nBin: "<<nBin<<", edgeWidth: "<<edgeWidth<<"\n";
  }

  // Unpack frequency interpolator class (MPI)
  template <class F> void NuInterpolator<F>::unpack(NuInterpolatorMPI& nuInterpolatorMPI, FrequencyBinList& nu, const F& _f) {
    nuBegin=nu.begin()+nuInterpolatorMPI.iBegin;
    nuEnd=nu.begin()+nuInterpolatorMPI.iEnd;
    fTable=nuInterpolatorMPI.fTable;
    fRedTable=nuInterpolatorMPI.fRedTable;
    fBlueTable=nuInterpolatorMPI.fBlueTable;
    nBin=nuInterpolatorMPI.nBin;
    edgeWidth=nuInterpolatorMPI.edgeWidth;
    f=_f;
    //cout<<"NuInterpolator::unpack: cell: "<<cellIndex<<", iBegin; "<<nuInterpolatorMPI.iBegin<<", iEnd: "<<nuInterpolatorMPI.iEnd<<", nu.size: "<<nu.size()<<", nuBegin-nu.begin: "<<nuBegin-nu.begin()<<"; nuEnd-nu.begin: "<<nuEnd-nu.begin()<<", fTable.size: "<<fTable.size()<<", fRedTable.size: "<<fRedTable.size()<<", fBlueTable.size: "<<fBlueTable.size()<<", nBin: "<<nBin<<", edgeWidth: "<<edgeWidth<<"\n";
  }
  
  // Bound-free integrator
  template <class F> class BFIntegrator {
  public:
    // Trapezoidal integration
    //static double trapezoidal(const F& f, const double& nuMin, const double& nuMax, const FrequencyBinList& nuBin, int& nIter, double relTol=1e-3, double absTol=0.0, string test="", int nIterMin=1, int nIterMax=100, int nBinMax=16, bool adjustTail=true);
    static double trapezoidal(const F& f, const double& nuMin, const double& nuMax, FrequencyBinList& nuBin, int& nIter, double relTol=1e-3, double absTol=0.0, bool useIterator=false, string test="", int nIterMin=1, int nIterMax=100, int nBinMax=16, bool adjustTail=true);
    // Trapezoidal integration (piecewise)
    //static double trapezoidalPW(const F& f, const double& nuMin, const double& nuMax, const FrequencyBinList& nuBin, int& nIter, double relTol=1e-3, double absTol=0, string test="", int nIterMin=1, int nIterMax=100, int nBinMax=16, bool adjustTail=true);
    static double trapezoidalPW(const F& f, const double& nuMin, const double& nuMax, FrequencyBinList& nuBin, int& nIter, double relTol=1e-3, double absTol=0, bool useIterator=false, string test="", int nIterMin=1, int nIterMax=100, int nBinMax=16, bool adjustTail=true);
    // Initiate
    //static void init(const F& f, const double& nuMin, const double& nuMax, const FrequencyBinList& nuBin, vector<double>& nuVec, vector<double>& fVec);
    static void init(const F& f, const double& nuMin, const double& nuMax, FrequencyBinList& nuBin, vector<double>& nuVec, vector<FrequencyBinIterator>& nuIter, vector<double>& fVec, string test="", bool useIterator=false);
    // Initiate (piecewise)
    //static void initPW(const F& f, const double& nuMin, const double& nuMax, const FrequencyBinList& nuBin, vector<double>& nuVec, vector<bool>& edgeVec, vector<double>& fVec, vector<double>& fRed, string test="");
    static void initPW(const F& f, const double& nuMin, const double& nuMax, FrequencyBinList& nuBin, vector<double>& nuVec, vector<FrequencyBinIterator>& nuIter, vector<bool>& edgeVec, vector<double>& fVec, vector<double>& fRed, string test="", bool useIterator=false);
    // Initial binned integration
    //static void trapezInit(const F& f, const int& nBin, vector<double>& nuVec, vector<double>& fVec, double& fInt, bool adjustTail=false, double relTolTail=0.0, string test="");
    static void trapezInit(const F& f, const int& nBin, FrequencyBinList& nuBin, vector<double>& nuVec, vector<FrequencyBinIterator>& nuIter, vector<double>& fVec, double& fInt, bool adjustTail=false, double relTolTail=0.0, string test="", bool useIterator=false);
    // Initial binned integration (piecewise)
    //static void trapezInitPW(const F& f, const int& nBin, vector<double>& nuVec, vector<bool>& edgeVec, vector<double>& fVec, vector<double>& fRed, double& fInt, bool adjustTail=false, double relTolTail=0.0, string test="");
    static void trapezInitPW(const F& f, const int& nBin, FrequencyBinList& nuBin, vector<double>& nuVec, vector<FrequencyBinIterator>& nuIter, vector<bool>& edgeVec, vector<double>& fVec, vector<double>& fRed, double& fInt, bool adjustTail=false, double relTolTail=0.0, string test="", bool useIterator=false);
    // Refined binned intergration
    //static void trapezRefineBin(const F& f, const vector<double>& nuVec, const int& nBin, vector<double>& fVec, double& fInt, string test="");
    static void trapezRefineBin(const F& f, FrequencyBinList& nuBin, const vector<double>& nuVec, const vector<FrequencyBinIterator>& nuIter, const int& nBin, vector<double>& fVec, double& fInt, string test="", bool useIterator=false);
    // Refined binned intergration (piecewise)
    //static void trapezRefineBinPW(const F& f, const vector<double>& nuVec, const vector<bool>& edgeVec, const vector<double>& fRed, const int& nBin, vector<double>& fVec, double& fInt, string test="");
    static void trapezRefineBinPW(const F& f, FrequencyBinList& nuBin, const vector<double>& nuVec, const vector<FrequencyBinIterator>& nuIter, const vector<bool>& edgeVec, const vector<double>& fRed, const int& nBin, vector<double>& fVec, double& fInt, string test="", bool useIterator=false);
    // Refined trapezoidal integration
    static void trapezRefine(const F& f, const vector<double>& nuVec, const int& n, double& fInt, string test="");
  };

  //template <class F> double BFIntegrator<F>::trapezoidal(const F& f, const double& nuMin, const double& nuMax, const FrequencyBinList& nuBin, int& nIter, double relTol, double absTol, string test, int nIterMin, int nIterMax, int nBinMax, bool adjustTail) {
  template <class F> double BFIntegrator<F>::trapezoidal(const F& f, const double& nuMin, const double& nuMax, FrequencyBinList& nuBin, int& nIter, double relTol, double absTol, bool useIterator, string test, int nIterMin, int nIterMax, int nBinMax, bool adjustTail) {
    double fInt=0.0,fIntOld=0.0,epsilon=numeric_limits<double>::epsilon(),relTolTail;
    vector<double> nuVec;
    vector<FrequencyBinIterator> nuIter;
    vector<double> fVec;
    int n,nBin;
    // Testing
    bool log=(test.compare("")!=0);

    if (log) cout<<"BFIntegrator::trapezoidal: "<<test<<", nu: ("<<nuMin<<","<<nuMax<<"), relTol: "<<relTol<<", absTol: "<<absTol<<", nuBin: ("<<nuBin[0].min<<","<<nuBin[nuBin.size()-1].max<<", nuBin.size: "<<nuBin.size()<<"\n";

    //if (log) cout<<"BFIntegrator::trapezoidal: "<<test<<", nuBin.size: "<<nuBin.size()<<"\n";

    // Set relative tail tolerance
    //relTolTail=MAX(relTol*1e-1,epsilon);
    // Testing!
    relTolTail=MAX(relTol/2,epsilon);

    // Initiate
    //init(f,nuMin,nuMax,nuBin,nuVec,fVec);
    init(f,nuMin,nuMax,nuBin,nuVec,nuIter,fVec,test,useIterator);

    if (log) cout<<"BFIntegrator::trapezoidal: "<<test<<", nuVec.size: "<<nuVec.size()<<"\n";
    
    // Initiatal integration using the frequency bins
    //trapezInit(f,nBinMax,nuVec,fVec,fInt,adjustTail,relTolTail,test);
    trapezInit(f,nBinMax,nuBin,nuVec,nuIter,fVec,fInt,adjustTail,relTolTail,test,useIterator);

    if (log) cout<<"BFIntegrator::trapezoidal: "<<test<<", nuVec.size: "<<nuVec.size()<<"\n";

    if (!isnumber(fInt)) { cout<<"BFIntegrator::trapeziodal, Warning: Invalid integral: "<<fInt<<", nu: ("<<nuMin<<","<<nuMax<<"), nuVec.size: "<<nuVec.size()<<", nuBin.size: "<<nuBin.size()<<"\n"; return fInt; }
    
    // Backup integral value
    fIntOld=fInt;

    // If binning needs to be refined
    if (nBinMax>1) {

      // Binning refinement loop
      for (nBin=nBinMax;nBin>1;nBin=nBin/2) {

	// Refine binning
	//trapezRefineBin(f,nuVec,nBin,fVec,fInt,test);
	trapezRefineBin(f,nuBin,nuVec,nuIter,nBin,fVec,fInt,test,useIterator);

	if (log) cout<<"BFIntegrator::trapezoidal: "<<test<<", nBin: "<<nBin<<", fInt: "<<fInt<<", fIntOld: "<<fIntOld<<"\n";

	// If tolerance achieved - Return
	if ((fabs(fInt-fIntOld) < relTol*fabs(fIntOld)) or (fabs(fInt-fIntOld) < absTol) or (fInt==0.0 and fIntOld==0.0)) { 
	  if (log) cout<<"BFIntegrator::trapezoidal: "<<test<<", nBin: "<<nBin<<", fInt: "<<fInt<<", fIntOld: "<<fIntOld<<", nuVec.size: "<<nuVec.size()<<"\n"; 
	  if (!isnumber(fInt)) cout<<"BFIntegrator::trapezoidal: Invalid integral: "<<fInt<<", nBin: "<<nBin<<", nu: ("<<nuMin<<","<<nuMax<<"), nuVec.size: "<<nuVec.size()<<"\n";
	  nIter=1; return fInt; 
	}
	
	// Backup integral value
	fIntOld=fInt;
      }
    }
 
    // Refinement loop
    for (n=2;n<=nIterMax;n++) {
      
      // Refine integration
      trapezRefine(f,nuVec,n,fInt,test);
 
      //if (log) cout<<"BFIntegrator::trapezoidal: "<<test<<", n: "<<n<<", fInt: "<<fInt<<", fIntOld: "<<fIntOld<<"\n";
      
      // If minimum number of iterations performed
      if (n>=nIterMin) {
	// If tolerance achieved - Return
	if ((fabs(fInt-fIntOld) < relTol*fabs(fIntOld)) or (fabs(fInt-fIntOld) < absTol) or (fInt==0.0 and fIntOld==0.0)) { 
	  if (log) cout<<"BFIntegrator::trapezoidal: "<<test<<", nIter: "<<n<<", fInt: "<<fInt<<", fIntOld: "<<fIntOld<<", nuVec.size: "<<nuVec.size()<<"\n"; 
	  if (!isnumber(fInt)) cout<<"BFIntegrator::trapezoidal: Invalid integral: "<<fInt<<", n: "<<n<<", nu: ("<<nuMin<<","<<nuMax<<"), nuVec.size: "<<nuVec.size()<<"\n";
	  nIter=n; return fInt; 
	}
      }
      
      // Backup integral value
      fIntOld=fInt;
    }

    nIter=n;
    
    // Supress this warning - Testing!
    //cout<<"BFIntegrator::trapezoidal: Maximum number of iterations exceeded\n";
    
    return fInt;
  }

  //template <class F> double BFIntegrator<F>::trapezoidalPW(const F& f, const double& nuMin, const double& nuMax, const FrequencyBinList& nuBin, int& nIter, double relTol, double absTol, string test, int nIterMin, int nIterMax, int nBinMax, bool adjustTail) {
  template <class F> double BFIntegrator<F>::trapezoidalPW(const F& f, const double& nuMin, const double& nuMax, FrequencyBinList& nuBin, int& nIter, double relTol, double absTol, bool useIterator, string test, int nIterMin, int nIterMax, int nBinMax, bool adjustTail) {
    double fInt=0.0,fIntOld=0.0,epsilon=numeric_limits<double>::epsilon(),relTolTail;
    vector<double> nuVec;
    vector<FrequencyBinIterator> nuIter;
    vector<bool> edgeVec;
    vector<double> fVec;
    vector<double> fRed;
    int n,nBin;
    // Testing
    bool log=(test.compare("")!=0);
 
    if (log) cout<<"BFIntegrator::trapezoidalPW: "<<test<<", nu: ("<<nuMin<<","<<nuMax<<"), relTol: "<<relTol<<", absTol: "<<absTol<<", nIterMin: "<<nIterMin<<", nIterMax: "<<nIterMax<<"\n";

    // Set relative tail tolerance
    relTolTail=MAX(relTol*1e-1,epsilon);
    //relTolTail=MAX(relTol/2,epsilon);

    // Initiate
    //initPW(f,nuMin,nuMax,nuBin,nuVec,edgeVec,fVec,fRed,test);
    initPW(f,nuMin,nuMax,nuBin,nuVec,nuIter,edgeVec,fVec,fRed,test,useIterator);
    //initPW(f,nuMin,nuMax,nuBin,nuVec,edgeVec,fVec,fRed);
    
    // Initiatal integration using the frequency bins
    //trapezInitPW(f,nBinMax,nuVec,edgeVec,fVec,fRed,fInt,adjustTail,relTolTail,test);
    trapezInitPW(f,nBinMax,nuBin,nuVec,nuIter,edgeVec,fVec,fRed,fInt,adjustTail,relTolTail,test,useIterator);
    //trapezInitPW(f,nBinMax,nuVec,edgeVec,fVec,fRed,fInt,adjustTail,relTolTail);    

    if (!isnumber(fInt)) { cout<<"BFIntegrator::trapeziodalPW, Warning: Invalid integral: "<<fInt<<", nu: ("<<nuMin<<","<<nuMax<<"), nuVec.size: "<<nuVec.size()<<", nuBin.size: "<<nuBin.size()<<", f("<<nuMin<<"): "<<f(nuMin)<<"\n"; return fInt; }
    
    // Backup integral value
    fIntOld=fInt;

    // If binning needs to be refined
    if (nBinMax>1) {

      // Binning refinement loop
      for (nBin=nBinMax;nBin>1;nBin=nBin/2) {

	// Refine binning
	//trapezRefineBinPW(f,nuVec,edgeVec,fRed,nBin,fVec,fInt,test);
	trapezRefineBinPW(f,nuBin,nuVec,nuIter,edgeVec,fRed,nBin,fVec,fInt,test,useIterator);
	//trapezRefineBinPW(f,nuVec,edgeVec,fRed,nBin,fVec,fInt);
	
	if (log) cout<<"BFIntegrator::trapezoidalPW: "<<test<<", nBin: "<<nBin<<", fInt: "<<fInt<<", fIntOld: "<<fIntOld<<"\n";

	// If tolerance achieved - Return
	if ((fabs(fInt-fIntOld) < relTol*fabs(fIntOld)) or (fabs(fInt-fIntOld) < absTol) or (fInt==0.0 and fIntOld==0.0)) { 
	  if (log) cout<<"BFIntegrator::trapezoidalPW: "<<test<<", nBin: "<<nBin<<", fInt: "<<fInt<<", fIntOld: "<<fIntOld<<", nuVec.size: "<<nuVec.size()<<"\n"; 
	  if (!isnumber(fInt)) cout<<"BFIntegrator::trapezoidalPW: Invalid integral: "<<fInt<<", fIntOld: "<<fIntOld<<", nBin: "<<nBin<<", nu: ("<<nuMin<<","<<nuMax<<"), nuVec.size: "<<nuVec.size()<<", nuBin.size: "<<nuBin.size()<<"\n";
	  nIter=1; return fInt; 
	}
	
	// Backup integral value
	fIntOld=fInt;
      }
    }
 
    // Refinement loop
    for (n=2;n<=nIterMax;n++) {
      
      // Refine integration
      trapezRefine(f,nuVec,n,fInt);
 
      //if (log) cout<<"BFIntegrator::trapezoidalPW: "<<test<<", n: "<<n<<", fInt: "<<fInt<<", fIntOld: "<<fIntOld<<"\n";
     
      // If minimum number of iterations performed
      if (n>=nIterMin) {
	// If tolerance achieved - Return
	if ((fabs(fInt-fIntOld) < relTol*fabs(fIntOld)) or (fabs(fInt-fIntOld) < absTol) or (fInt==0.0 and fIntOld==0.0)) { 
	  if (log) cout<<"BFIntegrator::trapezoidalPW: "<<test<<", nIter: "<<n<<", fInt: "<<fInt<<", fIntOld: "<<fIntOld<<", nuVec.size: "<<nuVec.size()<<"\n"; 
	  if (!isnumber(fInt)) cout<<"BFIntegrator::trapezoidalPW: Invalid integral: "<<fInt<<", fIntOld: "<<fIntOld<<", n: "<<n<<", nuRange: ("<<nuMin<<","<<nuMax<<"), nuVec.size: "<<nuVec.size()<<", nuBin.size: "<<nuBin.size()<<"\n";
	  nIter=n; return fInt; }
      }
      
      // Backup integral value
      fIntOld=fInt;
    }
    
    nIter=n; // Corrected!

    // Supress this warning - Testing!
    //cout<<"BFIntegrator::trapezoidalPW: Maximum number of iterations exceeded\n";    
    
    return fInt;
  }  

  //template <class F> void BFIntegrator<F>::init(const F& f, const double& nuMin, const double& nuMax, const FrequencyBinList& nuBin, vector<double>& nuVec, vector<double>& fVec) {
  template <class F> void BFIntegrator<F>::init(const F& f, const double& nuMin, const double& nuMax, FrequencyBinList& nuBin, vector<double>& nuVec, vector<FrequencyBinIterator>& nuIter, vector<double>& fVec, string test,  bool useIterator) {
    int i;
    // Testing
    bool log=(test.compare("")!=0);
    
    // Construct frequency vector
    nuVec.push_back(nuMin);
    nuIter.push_back(nuBin.end());
    for (i=0;i<nuBin.size();i++) {
      if (nuBin[i].min>nuMin and nuBin[i].min<nuMax) { 
	// Testing - Exclude minor edges
	//if (nuBin[i].min>nuMin and nuBin[i].min<nuMax and (not nuBin[i].minorEdge)) {
	nuVec.push_back(nuBin[i].min); 
	nuIter.push_back(nuBin.begin()+i); 
      }
    }
    nuVec.push_back(nuMax);
    nuIter.push_back(nuBin.end());

    // Resize function vector
    fVec.resize(nuVec.size());
  }

  //template <class F> void BFIntegrator<F>::initPW(const F& f, const double& nuMin, const double& nuMax, const FrequencyBinList& nuBin, vector<double>& nuVec, vector<bool>& edgeVec, vector<double>& fVec, vector<double>& fRed, string test) {
  template <class F> void BFIntegrator<F>::initPW(const F& f, const double& nuMin, const double& nuMax, FrequencyBinList& nuBin, vector<double>& nuVec, vector<FrequencyBinIterator>& nuIter, vector<bool>& edgeVec, vector<double>& fVec, vector<double>& fRed, string test, bool useIterator) {
    int i,nEdge=0;
    double edgeWidth=2e-6; 
    // Testing
    bool log=(test.compare("")!=0);
   
    // Add minumum frequency to frequency vector
    // Testing - Removed
    nuVec.push_back(nuMin);
    edgeVec.push_back(false);
    nuIter.push_back(nuBin.end());

    // For each frequency bin
    for (i=0;i<nuBin.size();i++) {

      // If within specific range
      //if (nuBin[i].min>nuMin and nuBin[i].min<nuMax) {
      // Testing - Exclude minor edges
      if (nuBin[i].min>nuMin and nuBin[i].min<nuMax and (not nuBin[i].minorEdge)) {
      // Testing
      //if (nuBin[i].min>=nuMin and nuBin[i].min<=nuMax and (not nuBin[i].minorEdge)) {
	// If first frequency differs from minimum frequency
	// Testing
	/*if (nuVec.size()==0 and nuBin[i].min!=nuMin) {
	  // Add minumum frequency to frequency vector
	  nuVec.push_back(nuMin);
	  edgeVec.push_back(false);
	  nuIter.push_back(nuBin.end());
	  }*/
	// Add minimum frequency to frequency vector
	nuVec.push_back(nuBin[i].min);
	// Add edge flag to edge vector
	edgeVec.push_back(nuBin[i].edge);
	// Add frequency bin iterator
	nuIter.push_back(nuBin.begin()+i);
	//if (log and nuBin[i].edge) cout<<"BFIntegrator::initPW: "<<test<<", Edge added, nu: ("<<nuBin[i].min<<","<<nuBin[i].max<<")\n";
      }
    }

    // If last frequenct differs from maximum frequency
    // Testing
    //if (nuVec[nuVec.size()-1]!=nuMax) {
    // Add maximum frequency to frequency vector
    nuVec.push_back(nuMax);
    edgeVec.push_back(true);
    nuIter.push_back(nuBin.end());
    //}

    // Resize function vector
    fVec.resize(nuVec.size());
    // Resize red edge function vector
    fRed.resize(nuVec.size());

    // For each frequency
    for (i=0;i<nuVec.size();i++) {
      
      // If edge frequency
      if (edgeVec[i]) {
	// Calculate function value
	//fVec[i]=f(nuVec[i]);
	//fVec[i]=f(nuVec[i]*(1+0.5*edgeWidth));
	if (useIterator and nuIter[i]!=nuBin.end()) fVec[i]=f(nuIter[i],blueEdge);
	//else fVec[i]=f(nuVec[i]*(1+0.5*edgeWidth));
	else fVec[i]=f(nuVec[i],blueEdge);
	// Calculate red egde function value
	//fRed[i]=f(nuVec[i]*(1-0.5*edgeWidth));
	if (useIterator and nuIter[i]!=nuBin.end()) fRed[i]=f(nuIter[i],redEdge);
	//else fRed[i]=f(nuVec[i]*(1-0.5*edgeWidth));
	else fRed[i]=f(nuVec[i],redEdge);
	// Increase edge counter
	nEdge++;
      }
    }

    if (log) cout<<"BFIntegrator::initPW: "<<test<<", nuBin.size: "<<nuBin.size()<<", nuVec.size: "<<nuVec.size()<<", nEdge: "<<nEdge<<", nuBin: ("<<nuBin[0].min<<","<<nuBin[nuBin.size()-1].max<<"), nuVec: ("<<nuVec[0]<<","<<nuVec[nuVec.size()-1]<<")\n";
  }
  
  //template <class F> void BFIntegrator<F>::trapezInit(const F& f, const int& nBin, vector<double>& nuVec, vector<double>& fVec, double& fInt, bool adjustTail, double relTolTail, string test) {
  template <class F> void BFIntegrator<F>::trapezInit(const F& f, const int& nBin, FrequencyBinList& nuBin, vector<double>& nuVec, vector<FrequencyBinIterator>& nuIter, vector<double>& fVec, double& fInt, bool adjustTail, double relTolTail, string test, bool useIterator) {
    int i,iTail;
    double dfdnu,fIntTail=0.0,epsilon=numeric_limits<double>::epsilon();
    // Testing
    bool log=(test.compare("")!=0);

    // For each frequency
    for (i=0;i<nuVec.size();i+=nBin) {
      
      // Calculate function value
      //fVec[i]=f(nuVec[i]);
      if (useIterator and nuIter[i]!=nuBin.end()) fVec[i]=f(nuIter[i]);
      else fVec[i]=f(nuVec[i]);
      //if (log and nuIter[i]!=nuBin.end()) cout<<"BFIntegrator::trapezInit: "<<test<<", f1["<<i<<"]: "<<f(nuIter[i])<<", f2["<<i<<"]: "<<f(nuVec[i])<<", nu1: "<<nuIter[i]->min<<", nu2: "<<nuVec[i]<<"\n";
    }
    // Calculate function value for last frequency
    fVec[nuVec.size()-1]=f(nuVec[nuVec.size()-1]);

    // For each frequency bin
    // Unclear if muliple initializers work - Fix!
    iTail=0; // New!
    for (i=0;i<nuVec.size()-1;i+=nBin) { // Changed!
      
      // Calculate bin integral
      dfdnu=0.5*(fVec[i]+fVec[min(i+nBin,nuVec.size()-1)])*(nuVec[min(i+nBin,nuVec.size()-1)]-nuVec[i]);

      //if (log) cout<<"BFIntegrator::trapezInit: "<<test<<", Integrate fVec["<<i<<":"<<min(i+nBin,nuVec.size()-1)<<"]: ("<<fVec[i]<<","<<fVec[min(i+nBin,nuVec.size()-1)]<<")\n";
		 
      // Add to total integral
      fInt+=dfdnu;

      // If adjust tail
      if (adjustTail) {

	// Add to tail integral
	fIntTail+=dfdnu;

	// If tail integral exceeds tail tolerance - Reset tail integral
	// What if fInt=fIntTail=0! Fix!!!
	if (fabs(fIntTail)>fabs(fInt)*relTolTail) { fIntTail=0.0; iTail=i+nBin; }
      }
    }

    //if (log) cout<<"BFIntegrator::trapezInit: "<<test<<", iTail: "<<iTail<<", fInt: "<<fInt<<", fIntTail: "<<fIntTail<<"\n";

    // If adjust tail
    if (adjustTail) {
      // Cut tail from vectors
      nuVec.resize(min(iTail+1,nuVec.size())); 
      fVec.resize(min(iTail+1,nuVec.size()));
      nuIter.resize(min(iTail+1,nuVec.size()));
      // Subtract tail integral
      fInt=fInt-fIntTail; 
    }

    //if (log) cout<<"BFIntegrator::trapezInit: "<<test<<", nuVec.size: "<<nuVec.size()<<", nuVec: ("<<nuVec[0]<<","<<nuVec[nuVec.size()-1]<<")\n";
  }

  //template <class F> void BFIntegrator<F>::trapezInitPW(const F& f, const int& nBin, vector<double>& nuVec, vector<bool>& edgeVec, vector<double>& fVec, vector<double>& fRed, double& fInt, bool adjustTail, double relTolTail, string test) {
  template <class F> void BFIntegrator<F>::trapezInitPW(const F& f, const int& nBin, FrequencyBinList& nuBin, vector<double>& nuVec, vector<FrequencyBinIterator>& nuIter, vector<bool>& edgeVec, vector<double>& fVec, vector<double>& fRed, double& fInt, bool adjustTail, double relTolTail, string test, bool useIterator) {
    int i,j,n,iTail;
    double dfdnu,fIntTail=0.0,epsilon=numeric_limits<double>::epsilon();
    // Hack !!!
    double edgeWidth=2e-6;
    // Testing
    bool log=(test.compare("")!=0);

    // For each binned frequency
    for (i=0;i<nuVec.size();i=j) {

      // Calculate function value
      //if (i>0) fVec[i]=f(nuVec[i]);
      if (useIterator and nuIter[i]!=nuBin.end()) fVec[i]=f(nuIter[i]);
      else {
	//if (log) cout<<"BFIntegrator::trapezInitPW: "<<test<<", i: "<<i<<", nuVec[i]: "<<nuVec[i]<<", useIterator: "<<useIterator<<", nuIter[i]!=nuBin.end()): "<<(nuIter[i]!=nuBin.end())<<"\n";
	  //fVec[i]=f(nuVec[i]*(1.0+0.5*edgeWidth));
	if (i>0) fVec[i]=f(nuVec[i]);
	// Hack !!!
	else {
	  //if (log) cout<<"BFIntegrator::trapezInitPW: "<<test<<", f("<<nuVec[i]<<"): "<<f(nuVec[i])<<", f("<<nuVec[i]*(1.0+0.5*edgeWidth)<<"): "<<f(nuVec[i]*(1.0+0.5*edgeWidth))<<", f("<<nuVec[i]*(1.0+1.0*edgeWidth)<<"): "<<f(nuVec[i]*(1.0+1.0*edgeWidth))<<", f("<<nuVec[i]*(1.0+10.0*edgeWidth)<<"): "<<f(nuVec[i]*(1.0+10.0*edgeWidth))<<"\n";
	  //fVec[i]=f(nuVec[i]*(1.0+0.5*edgeWidth));
	  fVec[i]=f(nuVec[i],blueEdge);
	}
      }

      //if (log) cout<<"BFIntegrator::trapezInitPW: "<<test<<", fVec["<<i<<"]: "<<fVec[i]<<"\n";
      
      // Step through each frequency in the bin
      // Unclear if muliple initializers work - Fix!
      n=0; for (j=i+1;n<nBin and j<nuVec.size();j++) if (not edgeVec[j]) n++; // Changed!
      // Undo last increment - Improve !
      if (j<nuVec.size()) j--;
    }

    // For each frequency bin
    // Unclear if muliple initializers work - Fix!
    iTail=0; // New!
    for (i=0;i<nuVec.size()-1;i=j) { // Changed!

      // For each frequency sub-bin
      // Unclear if muliple initializers work - Fix!
      n=0; dfdnu=0; // New!
      for (j=i+1;n<nBin and j<nuVec.size();j++) { // Changed!

	// If ending on edge frequency
	if (edgeVec[j]) {

	  if (log) cout<<"BFIntegrator::trapezInitPW: "<<test<<", Edge integrate fVec["<<i<<":"<<j<<"]: ("<<fVec[i]<<","<<fRed[j]<<"), nu: ("<<nuVec[i]<<","<<nuVec[j]<<")\n";

	  // Add to bin integral (using red edge function value)
	  dfdnu+=0.5*(fVec[i]+fRed[j])*(nuVec[j]-nuVec[i]);

	  //if (!isnumber(dfdnu)) cout<<"BFIntegrator::trapezInitPW, Warning: Invalid edge integral:  "<<dfdnu<<", nuVec["<<i<<","<<j<<"]: ["<<nuVec[i]<<","<<nuVec[j]<<"], fVec["<<i<<"]: "<<fVec[i]<<", fRef["<<j<<"]: "<<fRed[j]<<"\n";

	  // Select new frequency
	  i=j;
	}
	// Else - increase sub-bin counter
	else n++;
      }
      // Undo last increment - Improve !
      j--;

      if (i!=j and log) cout<<"BFIntegrator::trapezInitPW: "<<test<<", Integrate fVec["<<i<<":"<<j<<"]: ("<<fVec[i]<<","<<fVec[j]<<"), nu: ("<<nuVec[i]<<","<<nuVec[j]<<")\n";

      // If any sub-bins left - Add to bin integral
      if (i!=j) dfdnu+=0.5*(fVec[i]+fVec[j])*(nuVec[j]-nuVec[i]);

      //if (i!=j and !isnumber(dfdnu)) cout<<"BFIntegrator::trapezInitPW, Warning: Invalid integral:  "<<dfdnu<<", nuVec["<<i<<","<<j<<"]: ["<<nuVec[i]<<","<<nuVec[j]<<"], fVec["<<i<<","<<j<<"]: ["<<fVec[i]<<","<<fVec[j]<<"]\n"; 
		 
      // Add to total integral
      fInt+=dfdnu;

      // If adjust tail
      if (adjustTail) {

	// Add to tail integral
	fIntTail+=dfdnu;

	// If tail integral exceeds tail tolerance - Reset tail integral
	// What if fInt=fIntTail=0! Fix!!!
	if (fabs(fIntTail)>fabs(fInt)*relTolTail) { fIntTail=0.0; iTail=j; }
      }
    }

    //if (log) cout<<"BFIntegrator::trapezInitPW: "<<test<<", iTail: "<<iTail<<", fInt: "<<fInt<<", fIntTail: "<<fIntTail<<"\n";

    // If adjust tail
    if (adjustTail) { 
      // Cut tail from vectors
      nuVec.resize(min(iTail+1,nuVec.size()));
      edgeVec.resize(min(iTail+1,nuVec.size()));
      fVec.resize(min(iTail+1,nuVec.size()));
      fRed.resize(min(iTail+1,nuVec.size()));
      nuIter.resize(min(iTail+1,nuVec.size()));
      // Subtract tail integral
      fInt=fInt-fIntTail; 
    }

    if (log) cout<<"BFIntegrator::trapezInitPW: "<<test<<", nuVec.size: "<<nuVec.size()<<", nuVec: ("<<nuVec[0]<<","<<nuVec[nuVec.size()-1]<<")\n";
  }
  
  //template <class F> void BFIntegrator<F>::trapezRefineBin(const F& f, const vector<double>& nuVec, const int& nBin, vector<double>& fVec, double& fInt, string test) {
  template <class F> void BFIntegrator<F>::trapezRefineBin(const F& f, FrequencyBinList& nuBin, const vector<double>& nuVec, const vector<FrequencyBinIterator>& nuIter, const int& nBin, vector<double>& fVec, double& fInt, string test, bool useIterator) {
    int i;
    // Testing
    bool log=(test.compare("")!=0);

    // For each new frequency
    for (i=nBin/2;i<nuVec.size();i+=nBin) {

      // Calculate function valyue
      //fVec[i]=f(nuVec[i]);
      if (useIterator and nuIter[i]!=nuBin.end()) fVec[i]=f(nuIter[i]);
      else fVec[i]=f(nuVec[i]);
      //if (log and nuIter[i]!=nuBin.end()) cout<<"BFIntegrator::trapezRefineBin: "<<test<<", f1["<<i<<"]: "<<f(nuIter[i])<<", f2["<<i<<"]: "<<f(nuVec[i])<<", nu1: "<<nuIter[i]->min<<", nu2: "<<nuVec[i]<<"\n";
    }

    // For each frequency bin
    // Unclear if muliple initializers work - Fix!
    fInt=0.0; // New!
    for (i=0;i<nuVec.size()-1;i+=nBin/2) { // Changed!
            		 
      // Calculate integral
      fInt+=0.5*(fVec[i]+fVec[min(i+nBin/2,nuVec.size()-1)])*(nuVec[min(i+nBin/2,nuVec.size()-1)]-nuVec[i]);

      //if (log) cout<<"BFIntegrator::trapezRefineBin: "<<test<<", Integrate fVec["<<i<<":"<<min(i+nBin/2,nuVec.size()-1)<<"]: ("<<fVec[i]<<","<<fVec[min(i+nBin/2,nuVec.size()-1)]<<")\n";
    }
  }

  //template <class F> void BFIntegrator<F>::trapezRefineBinPW(const F& f, const vector<double>& nuVec, const vector<bool>& edgeVec, const vector<double>& fRed, const int& nBin, vector<double>& fVec, double& fInt, string test) {
  template <class F> void BFIntegrator<F>::trapezRefineBinPW(const F& f, FrequencyBinList& nuBin, const vector<double>& nuVec, const vector<FrequencyBinIterator>& nuIter, const vector<bool>& edgeVec, const vector<double>& fRed, const int& nBin, vector<double>& fVec, double& fInt, string test, bool useIterator) {
    int i,j,n;
    // Testing
    bool log=(test.compare("")!=0);

    // Step through each frequency until the first binned frequency
    // Unclear if muliple initializers work - Fix!
    n=0; for (j=1;n<nBin/2 and j<nuVec.size();j++) if (not edgeVec[j]) n++; // Changed!
    // Undo last increment - Improve !
    if (j<nuVec.size()) j--;

    // For each binned frequency bin
    //for (i=nBin/2;i<nuVec.size();i=j) {
    for (i=j;i<nuVec.size();i=j) {

      // Calculate function value
      //fVec[i]=f(nuVec[i]);
      if (useIterator and nuIter[i]!=nuBin.end()) fVec[i]=f(nuIter[i]);
      else fVec[i]=f(nuVec[i]);

      if (log) cout<<"BFIntegrator::trapezRefineBinPW: "<<test<<", fVec["<<i<<"]: "<<fVec[i]<<"\n";

      // Step through each frequency in the bin
      // Unclear if muliple initializers work - Fix!
      n=0; for (j=i+1;n<nBin and j<nuVec.size();j++) if (not edgeVec[j]) n++; // Changed!
      // Undo last increment - Improve !
      if (j<nuVec.size()) j--;
    }

    // For each frequency bin
    // Unclear if muliple initializers work - Fix!
    fInt=0.0; // New!
    for (i=0;i<nuVec.size()-1;i=j) { // Changed!

      // For each frequency sub-bin 
      // Unclear if muliple initializers work - Fix!
      n=0; // New!
      for (j=i+1;n<nBin/2 and j<nuVec.size();j++) { // Changed!
	
	// If ending on edge frequency
	if (edgeVec[j]) {

	  if (log) cout<<"BFIntegrator::trapezRefineBinPW: "<<test<<", Edge integrate fVec["<<i<<":"<<j<<"]: ("<<fVec[i]<<","<<fRed[j]<<"), nu: ("<<nuVec[i]<<","<<nuVec[j]<<")\n";
	  
	  // Add to integral (using red edge function value)
	  fInt+=0.5*(fVec[i]+fRed[j])*(nuVec[j]-nuVec[i]);
	  
	  // Select new frequency
	  i=j;
	}
	// Else - increase sub-bin counter
	else n++;
      }
      // Undo last increment - Improve !
      j--;

      if (i!=j and log) cout<<"BFIntegrator::trapezRefineBinPW: "<<test<<", Integrate fVec["<<i<<":"<<j<<"]: ("<<fVec[i]<<","<<fVec[j]<<"), nu: ("<<nuVec[i]<<","<<nuVec[j]<<")\n";
    
      // If any sub-bins left - Add to integral
      if (i!=j) fInt+=0.5*(fVec[i]+fVec[j])*(nuVec[j]-nuVec[i]);
    }
  }

  template <class F> void BFIntegrator<F>::trapezRefine(const F& f, const vector<double>& nuVec, const int& n, double& fInt, string test) {
    int nNew,i,j;
    double dnu,nuMid,fSum;
    
    // Calculate reusable part of integral
    fInt=0.5*fInt;
    
    // Calculate number of new points
    // Unclear if muliple initializers work - Fix!
    nNew=1; for (j=1;j<n-1;j++) nNew <<= 1; // Changed!
    
    // For each frequency bin
    for (i=0;i<nuVec.size()-1;i++) {
      
      // Calculate delta and initial frequency
      dnu=(nuVec[i+1]-nuVec[i])/float(nNew);
      nuMid=nuVec[i]+0.5*dnu;

      //cout<<"BFIntegrator::trapezRefine: "<<test<<", nuVec["<<i<<"]: "<<nuVec[i]<<", nuVec["<<i+1<<"]: "<<nuVec[i+1]<<", nNew: "<<nNew<<", dnu: "<<dnu<<"\n";
      
      // Calculate and add refined part of integral
      // Unclear if muliple initializers work - Fix!
      // But how about mulitiple incrementors?
      fSum=0.0; // New!
      for (j=0;j<nNew;j++,nuMid+=dnu) { // Changed!
	fSum+= f(nuMid);
	//cout<<"BFIntegrator::trapezRefine: "<<test<<", n: "<<nNew<<", j: "<<j<<", f("<<nuMid<<"): "<<f(nuMid)<<"\n";
      }
      fInt+=0.5*dnu*fSum;
    }
  }


  /*  template <class F> double JIntegrator<F>::trapezoidal(const F& f, const double& nuMin, const double& nuMax, FrequencyBinList& nuBin, vector<double>& Jnu, vector<double>& nuMeanNu, int& nIter, double relTol, double absTol, bool useIterator, string test, int nIterMin, int nIterMax, int nBinMax, bool adjustTail) {
    double fInt=0.0,fIntOld=0.0,epsilon=numeric_limits<double>::epsilon(),relTolTail;
    vector<double> nuVec;
    vector<FrequencyBinIterator> nuIter;
    vector<double> fVec;
    int n,nBin;

    // Set relative tail tolerance
    relTolTail=MAX(relTol*1e-1,epsilon);

    // Initiate
    init(f,nuMin,nuMax,nuBin,nuVec,nuIter,fVec,test,useIterator);
    
    // Initiatal integration using the frequency bins
    trapezInit(f,nBinMax,nuBin,nuVec,nuIter,fVec,fInt,adjustTail,relTolTail,test,useIterator);

    if (!isnumber(fInt)) { cout<<"BFIntegrator::trapeziodal, Warning: Invalid integral: "<<fInt<<", nu: ("<<nuMin<<","<<nuMax<<"), nuVec.size: "<<nuVec.size()<<", nuBin.size: "<<nuBin.size()<<"\n"; return fInt; }
    
    // Backup integral value
    fIntOld=fInt;

    // If binning needs to be refined
    if (nBinMax>1) {

      // Binning refinement loop
      for (nBin=nBinMax;nBin>1;nBin=nBin/2) {

	// Refine binning
	trapezRefineBin(f,nuBin,nuVec,nuIter,nBin,fVec,fInt,test,useIterator);

	// If tolerance achieved - Return
	if ((fabs(fInt-fIntOld) < relTol*fabs(fIntOld)) or (fabs(fInt-fIntOld) < absTol) or (fInt==0.0 and fIntOld==0.0)) { 
	  if (!isnumber(fInt)) cout<<"BFIntegrator::trapezoidal: Invalid integral: "<<fInt<<", nBin: "<<nBin<<", nu: ("<<nuMin<<","<<nuMax<<"), nuVec.size: "<<nuVec.size()<<"\n";
	  nIter=1; return fInt; 
	}
	
	// Backup integral value
	fIntOld=fInt;
      }
    }
    
    cout<<"BFIntegrator::trapezoidal: Maximum number of iterations exceeded\n";
    
    return fInt;
  }

  template <class F> void JIntegrator<F>::trapezInit(const F& f, const int& nBin, FrequencyBinList& nuBin, vector<double>& nuVec, vector<FrequencyBinIterator>& nuIter, vector<double>& fVec, double& fInt, bool adjustTail, double relTolTail, string test, bool useIterator) {
    int i,iTail;
    double dfdnu,fIntTail=0.0,epsilon=numeric_limits<double>::epsilon();

    // For each frequency
    for (i=0;i<nuVec.size();i+=nBin) {
      
      // Calculate function value
      if (useIterator and nuIter[i]!=nuBin.end()) fVec[i]=f(nuIter[i]);
      else fVec[i]=f(nuVec[i]);
    }
    // Calculate function value for last frequency
    fVec[nuVec.size()-1]=f(nuVec[nuVec.size()-1]);

    // For each frequency bin
    iTail=0; // New!
    for (i=0;i<nuVec.size()-1;i+=nBin) { // Changed!
      
      // Calculate bin integral
      dfdnu=0.5*(fVec[i]+fVec[min(i+nBin,nuVec.size()-1)])*(nuVec[min(i+nBin,nuVec.size()-1)]-nuVec[i]);
		 
      // Add to total integral
      fInt+=dfdnu;

      // If adjust tail
      if (adjustTail) {

	// Add to tail integral
	fIntTail+=dfdnu;

	// If tail integral exceeds tail tolerance - Reset tail integral
	// What if fInt=fIntTail=0! Fix!!!
	if (fabs(fIntTail)>fabs(fInt)*relTolTail) { fIntTail=0.0; iTail=i+nBin; }
      }
    }

    // If adjust tail
    if (adjustTail) {
      // Cut tail from vectors
      nuVec.resize(min(iTail+1,nuVec.size())); 
      fVec.resize(min(iTail+1,nuVec.size()));
      nuIter.resize(min(iTail+1,nuVec.size()));
      // Subtract tail integral
      fInt=fInt-fIntTail; 
    }
    }*/
}

#endif
