#include <omp.h>
#include "mc.h"
#include "msolver.h"
#include "diffsolver.h"

using namespace mc;
using namespace msolver;
using namespace diffsolver;

// Testing!
void logSysMemory(string label) {
  stringstream header;
  int status;

  header.str(""); 
  header<<"echo \""<<label<<"\" >> mc-ps-$(hostname).log";
  status=system(header.str().c_str());
  status=system("./mc-ps.sh >> mc-ps-$(hostname).log");
}

// Constructors
DiffConvConfig::DiffConvConfig() : nIter(1000), nIterMin(1), TTol(1e-3), neTol(1e-2), dampT(0.8), TMin(500), dTMax(1000), derivative(dPartial), dTDerivative(1e-6) {}
DiffConvConfig::DiffConvConfig(const int& _nIter, const int& _nIterMin, const double& _TTol, const double& _neTol, const double& _dampT, const double& _TMin, const double& _dTMax, const DerivativeType& _derivative, const double& _dTDerivative) : nIter(_nIter), nIterMin(_nIterMin), TTol(_TTol), neTol(_neTol), dampT(_dampT), TMin(_TMin), dTMax(_dTMax), derivative(_derivative), dTDerivative(_dTDerivative) {}

// Load data
void DiffConvConfig::load(const YAML::Node& data) {

  // Load number of iterations
  if (data["nIter"]) nIter=data["nIter"].as<int>();

  // Load minimum number of iterations
  if (data["nIterMin"]) nIterMin=data["nIterMin"].as<int>();

  // Load temperature tolerance
  if (data["TTol"]) TTol=data["TTol"].as<double>();

  // Load electron number density tolerence
  if (data["neTol"]) neTol=data["neTol"].as<double>();

  // Load temperature damping factor
  if (data["dampT"]) dampT=data["dampT"].as<double>();

  // Load minimum temperature
  if (data["TMin"]) TMin=data["TMin"].as<double>();

  // Load maximum temperature change
  if (data["dTMax"]) dTMax=data["dTMax"].as<double>();

  // Load derivative type
  if (data["derivative"]) {
    if (data["derivative"].as<string>().compare("full")==0) derivative=dFull;
    else derivative=dPartial;
  }

  // Load derivative step
  if (data["dTDerivative"]) dTDerivative=data["dTDerivative"].as<double>();
}

// Log data
void DiffConvConfig::log() {
  cout<<"DiffConvConfig::log: nIter: "<<nIter<<", nIterMin: "<<nIterMin<<", TTol: "<<TTol<<", neTol: "<<neTol<<", dampT: "<<dampT<<", TMin: "<<TMin<<", dTMax: "<<dTMax<<", derivative: "<<derivative<<", dTDerivative: "<<dTDerivative<<"\n";
}

// Constructors
DiffTStepConfig::DiffTStepConfig() : set(false), refine(true), nIter(1000), nIterMin(1), drhoMax(0.05), dTMax(0.05), dtMax(1e99), dtFactorMin(2.0), dtFactorMax(1e99), lTLimit(true) {}
DiffTStepConfig::DiffTStepConfig(const bool& _set, const bool& _refine, const int& _nIter, const int& _nIterMin, const double& _drhoMax, const double& _dTMax, const double& _dtMax, const double& _dtFactorMin, const double& _dtFactorMax, const bool& _lTLimit) : set(_set), refine(_refine), nIter(_nIter), nIterMin(_nIterMin), drhoMax(_drhoMax), dTMax(_dTMax), dtMax(_dtMax), dtFactorMin(_dtFactorMin), dtFactorMax(_dtFactorMax), lTLimit(_lTLimit) {}

// Load data
void DiffTStepConfig::load(const YAML::Node& data) {

  // Load set time step flag
  if (data["set"]) set=(data["set"].as<string>().compare("yes")==0);

  // Load refine time step flag
  if (data["refine"]) refine=(data["refine"].as<string>().compare("yes")==0);

  // Load number of iterations
  if (data["nIter"]) nIter=data["nIter"].as<int>();

  // Load minimum number of iterations
  if (data["nIterMin"]) nIterMin=data["nIterMin"].as<int>();

  // Load maximum change in density
  if (data["drhoMax"]) drhoMax=data["drhoMax"].as<double>();

  // Load maximum change in temperature
  if (data["dTMax"]) dTMax=data["dTMax"].as<double>();

  // Load maximum timestep
  if (data["dtMax"]) dtMax=data["dtMax"].as<double>();

  // Load minimum decrease in timestep
  if (data["dtFactorMin"]) dtFactorMin=data["dtFactorMin"].as<double>();

  // Load maximum decrease in timestep
  if (data["dtFactorMax"]) dtFactorMax=data["dtFactorMax"].as<double>();

  // Load light-travel time limit flag
  if (data["lTLimit"]) lTLimit=(data["lTLimit"].as<string>().compare("yes")==0);
}

// Log data
void DiffTStepConfig::log() {
  cout<<"DiffTStepConfig::log: set: "<<set<<", refine: "<<refine<<", nIter: "<<nIter<<", nIterMin: "<<nIterMin<<", drhoMax: "<<drhoMax<<", dTMax: "<<dTMax<<", dtMax: "<<dtMax<<", dtFactorMin: "<<dtFactorMin<<", dtFactorMax: "<<dtFactorMax<<", lTLimit: "<<lTLimit<<"\n";
}

// Load boundary method
BoundaryMethod diffsolver::loadBoundaryMethod(const YAML::Node& data) {
  if (data.as<string>().compare("dLdmZero")==0) return dLdmZero;
  else if (data.as<string>().compare("LFixed")==0) return LFixed;
  else if (data.as<string>().compare("eddingtonMod")==0) return eddingtonMod;
  else return eddington;
}

// Constructors
DiffIBoundConfig::DiffIBoundConfig() : L(0.0), method(LFixed) {}
DiffIBoundConfig::DiffIBoundConfig(const double& _L, const BoundaryMethod& _method) : L(_L), method(_method) {}

// Load data
void DiffIBoundConfig::load(const YAML::Node& data) {

  // Load luminosity
  if (data["L"]) L=data["L"].as<double>();

  // Load method
  if (data["method"]) method=loadBoundaryMethod(data["method"]);
}

// Log data
void DiffIBoundConfig::log() {
  cout<<"DiffIBoundConfig::log: L: "<<L<<", method: "<<method<<"\n";
}

// Constructors
DiffOBoundConfig::DiffOBoundConfig() : LTFixed(false), L(0.0), method(eddington), phiZero(false), expCorr(false), useTJ(false) {}
DiffOBoundConfig::DiffOBoundConfig(const bool& _LTFixed, const double& _L, const BoundaryMethod& _method, const bool& _phiZero, const bool& _expCorr, const bool& _useTJ) : LTFixed(_LTFixed), L(_L), method(_method), phiZero(_phiZero), expCorr(_expCorr), useTJ(_useTJ) {}

// Load data
void DiffOBoundConfig::load(const YAML::Node& data) {

  // Load luminosity-temperature fixed flag
  if (data["LTFixed"]) LTFixed=(data["LTFixed"].as<string>().compare("yes")==0);

  // Load luminosity
  if (data["L"]) L=data["L"].as<double>();

  // Load method
  if (data["method"]) method=loadBoundaryMethod(data["method"]);

  // Load phi zero flag
  if (data["phiZero"]) phiZero=(data["phiZero"].as<string>().compare("yes")==0);

  // Load expansion correction flag
  if (data["expCorr"]) expCorr=(data["expCorr"].as<string>().compare("yes")==0);

  // Load use TJ flag
  if (data["useTJ"]) useTJ=(data["useTJ"].as<string>().compare("yes")==0);
}

// Log data
void DiffOBoundConfig::log() {
  cout<<"DiffOBoundConfig::log: LTFixed: "<<LTFixed<<", L: "<<L<<", method: "<<method<<", phiZero: "<<phiZero<<", expCorr: "<<expCorr<<", useTJ: "<<useTJ<<"\n";
}

// Constructors - New!
DiffKappaConfig::DiffKappaConfig() : processes(),  floor(false), MC(true) {}
DiffKappaConfig::DiffKappaConfig(const Processes& _processes, const bool& _floor, const bool& _MC) : processes(_processes), floor(_floor), MC(_MC) {}

// Load data - New!
void DiffKappaConfig::load(const YAML::Node& data) {

  // Load processes
  if (data["processes"]) processes.load(data["processes"]);

  // Load opacity floor flag
  if (data["floor"]) floor=(data["floor"].as<string>().compare("yes")==0);

  // Load opacity MC flag
  if (data["MC"]) MC=(data["MC"].as<string>().compare("yes")==0);

  log();
}

// Log data - New!
void DiffKappaConfig::log() {
  cout<<"DiffKappaConfig::log: floor: "<<floor<<", MC: "<<MC;
  processes.log();
  cout<<"\n";
}

// Constructors
DiffSolverConfig::DiffSolverConfig() : steadyState(true), includeRadio(false), nPacketGamma(1e5), phi(1.0), fluxLimiter(false), expansion(false), Lobs(0.0), convergence(), timeStep(), innerBoundary(), outerBoundary(), kappa(), testSwitch1(false), nuMin(-1), nuMax(-1), boundFreeCutOff(-1) {}; // Changed!
DiffSolverConfig::DiffSolverConfig(const bool& _steadyState, const bool& _includeRadio, const double& _nPacketGamma, const double& _phi, const bool& _fluxLimiter, const bool& _expansion, const double& _Lobs, const DiffConvConfig& _convergence, const DiffTStepConfig& _timeStep, const DiffIBoundConfig& _innerBoundary, const DiffOBoundConfig& _outerBoundary, const DiffKappaConfig& _kappa, const bool& _testSwitch1, const double& _nuMin, const double& _nuMax, const double& _boundFreeCutOff) : steadyState(_steadyState), includeRadio(_includeRadio), nPacketGamma(_nPacketGamma), phi(_phi), fluxLimiter(_fluxLimiter), Lobs(_Lobs), expansion(_expansion), convergence(_convergence), timeStep(_timeStep), innerBoundary(_innerBoundary), outerBoundary(_outerBoundary), kappa(_kappa), testSwitch1(_testSwitch1), nuMin(_nuMin), nuMax(_nuMax), boundFreeCutOff(_boundFreeCutOff) {}; // Changed!
DiffSolverConfig::DiffSolverConfig(const string& dataFile) { load(dataFile); }

// Load data
void DiffSolverConfig::load(const YAML::Node& data) {

  // Load steady state flag
  if (data["steadyState"]) steadyState=(data["steadyState"].as<string>().compare("yes")==0);

  // Load radioactive heating flag
  if (data["includeRadio"]) includeRadio=(data["includeRadio"].as<string>().compare("yes")==0);

  // Load number of gamma packets
  if (data["nPacketGamma"]) nPacketGamma=data["nPacketGamma"].as<int>();

  // Load phi
  if (data["phi"]) phi=data["phi"].as<double>();

  // Load flux limiter flag
  if (data["fluxLimiter"]) fluxLimiter=(data["fluxLimiter"].as<string>().compare("yes")==0);  

  // Load expansion flag
  if (data["expansion"]) steadyState=(data["expansion"].as<string>().compare("yes")==0);  

  // Load inner boundary config
  if (data["innerBoundary"]) innerBoundary.load(data["innerBoundary"]);

  // Load outer boundary config
  if (data["outerBoundary"]) outerBoundary.load(data["outerBoundary"]);

  // Load convergence config
  if (data["convergence"]) convergence.load(data["convergence"]);

  // Load timestep config
  if (data["timeStep"]) timeStep.load(data["timeStep"]);

  // Load opacity config - New!
  if (data["kappa"]) kappa.load(data["kappa"]);

  // Load test swith 1
  if (data["testSwitch1"]) testSwitch1=(data["testSwitch1"].as<string>().compare("yes")==0);

  // Load radiation field frequency limits
  if (data["nuMin"]) nuMin=data["nuMin"].as<double>();
  if (data["nuMax"]) nuMax=data["nuMax"].as<double>();

  // Load bound-free cut-off
  if (data["boundFreeCutOff"]) boundFreeCutOff=data["boundFreeCutOff"].as<double>();

  // Load opacity floor flag
  //if (data["kappaFloor"]) kappaFloor=(data["kappaFloor"].as<string>().compare("yes")==0);
  // Keep this for backward compatibility
  if (data["kappaFloor"]) kappa.floor=(data["kappaFloor"].as<string>().compare("yes")==0); // Changed!

  // Load opacity MC flag
  //if (data["kappaMC"]) kappaMC=(data["kappaMC"].as<string>().compare("yes")==0);
  // Keep this for backward compatibility
  if (data["kappaMC"]) kappa.MC=(data["kappaMC"].as<string>().compare("yes")==0); // Changed!

  log();
}

// Log data
void DiffSolverConfig::log() {
  //cout<<"DiffSolverConfig::log: steadyState: "<<steadyState<<", includeRadio: "<<includeRadio<<", nPacketGamma: "<<nPacketGamma<<", phi: "<<phi<<", fluxLimiter: "<<fluxLimiter<<", Lobs: "<<Lobs<<", nuMin: "<<nuMin<<", nuMax: "<<nuMax<<", boundFreeCutOff: "<<boundFreeCutOff<<", kappaFloor: "<<kappaFloor<<", kappaMC: "<<kappaMC<<"\n";
  cout<<"DiffSolverConfig::log: steadyState: "<<steadyState<<", includeRadio: "<<includeRadio<<", nPacketGamma: "<<nPacketGamma<<", phi: "<<phi<<", fluxLimiter: "<<fluxLimiter<<", Lobs: "<<Lobs<<", nuMin: "<<nuMin<<", nuMax: "<<nuMax<<", boundFreeCutOff: "<<boundFreeCutOff<<"\n"; // Changed!
  innerBoundary.log();
  outerBoundary.log();
  convergence.log();
  timeStep.log();
  kappa.log(); // New!
}

// Load data (from file)
void DiffSolverConfig::load(const string& dataFile) {
  YAML::Node data;

  // Load data from file
  data=YAML::LoadFile(dataFile);

  // Load data
  if (data["diffSolver"]) load(data["diffSolver"]);
}

// Constructors
DiffSolver::DiffSolver(DiffSolverConfig _config, LTESolverConfig _lteConfig, boost::mpi::communicator* _mpiCom) : config(_config), lteSolver(_lteConfig), pLog(true), mpiCom(_mpiCom) { lteSolver.config.temperature=LTETempConfig(); } // Changed! - Force LTE solver default temperature config - Hack!
DiffSolver::DiffSolver(const string& dataFile, boost::mpi::communicator* _mpiCom) : lteSolver(dataFile), pLog(true), mpiCom(_mpiCom) { config.load(dataFile); lteSolver.config.temperature=LTETempConfig(); } // Changed! - Force LTE solver default temperature config - Hack!

// Get temperature using Eddington approximation
double DiffSolver::TEdd(const double& _L) {
  double _TEdd,deltaTau,_F;
  int i=grid.cell.size()-1;

  deltaTau=kappa(i,front)*dm(i,front)/(4*pi*pow(r(i,front),2));
  _F=_L/(4*pi*pow(r(i,front),2));
  _TEdd=pow(_F*(deltaTau+(2.0/3.0))/((1.0/3.0)*a*c),(1.0/4.0));

  return _TEdd;
}

// Get luminosity using the Eddington approximation
//double DiffSolver::LEdd(const double& _T) const {
double DiffSolver::LEdd(const double& _T) {
  double _LEdd,deltaTau,_F;
  int i=grid.cell.size()-1;

  deltaTau=kappa(i,front)*dm(i,front)/(4*pi*pow(r(i,front),2));
  _F=(1.0/3.0)*a*c*pow(_T,4)/(deltaTau+(2.0/3.0));
  _LEdd=4*pi*pow(r(i,front),2)*_F;

  return _LEdd;
}

// Get luminosity using the modified Eddington approximation
//double DiffSolver::LEddMod(const double& _T) const {
double DiffSolver::LEddMod(const double& _T) {
  double _LEddMod,deltaTau,_F;
  int i=grid.cell.size()-1;

  deltaTau=kappa(i,front)*dm(i,front)/(4*pi*pow(r(i,front),2));
  _F=(1.0/3.0)*a*c*pow(_T,4)/(deltaTau+(1.0/3.0));  
  _LEddMod=4*pi*pow(r(i,front),2)*_F;
  
  return _LEddMod;
}

// Calculate temperature from front luminosity
// Does currently not work with flux-limiter
//double DiffSolver::TDiffFront(const int& i) const {
double DiffSolver::TDiffFront(const int& i) {
  double _T,_L;
  
  // Calculate temperature from front luminosity
  _L=4*pi*pow(r(i,front),2)*((CellSS*) grid.cell[i])->FFront();
  _T=pow(pow(T(i+1,center),4)+_L*(3*kappa(i,back))*dm(i,back)/(pow(4*pi*pow(r(i,back),2),2)*a*c),(1.0/4.0));

  return _T;
}

// Interpolate r following Falk & Arnett 1977
double DiffSolver::rIpol(const double& r1, const double& r2) const {
  return pow(0.5*(pow(r1,3)+pow(r2,3)),1.0/3.0);
}

// Interpolate density following Falk & Arnett (1977)
double DiffSolver::rhoIpol(const double& rho1, const double& rho2, const double& dm1, const double& dm2) const {
  return (rho1*dm1+rho2*dm2)/(dm1+dm2);
}

// Interpolate differential mass
double DiffSolver::dmIpol(const double& dm1, const double& dm2) const {
  return (dm1+dm2)/2;
}

// Interpolate opacity following Christy (1967)
double DiffSolver::kappaIpol(const double& T1, const double& T2, const double& kappa1, const double& kappa2) const {
  return (pow(T1,4)+pow(T2,4))/(pow(T1,4)/kappa1+pow(T2,4)/kappa2);
}

// Get radius
double DiffSolver::r(const int& i, const Location& location, const Epoch& epoch) const {
  double _r;

  // Select epoch
  const Grid* gridEpoch=(epoch==old?gridOld:&grid);
  
  // Switch on location
  switch(location) {
  case center : 
    // Interpolate
    _r=rIpol(r(i,back,epoch),r(i,front,epoch));
    break;
  case back :
    // Get inner border
    _r=((CellSS*) gridEpoch->cell[i])->rMin();
    break;
  case front :
    // Get outer border
    _r=((CellSS*) gridEpoch->cell[i])->rMax();
    break;
  }

  return _r;
}

// Set radius
void DiffSolver::r(const double& _r, const int& i, const Location& location, const Epoch& epoch) {
  int j; // New!

  // Select epoch
  const Grid* gridEpoch=(epoch==old?gridOld:&grid);

  // Switch on location
  switch(location) {
  case front :
    // Set outer radius
    ((CellSS*) gridEpoch->cell[i])->rMax(_r);
    // If cell container - Set outer radius of each subcell - New!
    if (gridEpoch->cell[i]->type()==cellContainer) for (j=0;j<gridEpoch->cell[i]->subCell.size();j++) ((CellSS*) gridEpoch->cell[i]->subCell[j])->rMax(_r);
    break;
  case back :
    // Set inner radius
    ((CellSS*) gridEpoch->cell[i])->rMin(_r);
    // If cell container - Set inner radius of each subcell - New!
    if (gridEpoch->cell[i]->type()==cellContainer) for (j=0;j<gridEpoch->cell[i]->subCell.size();j++) ((CellSS*) gridEpoch->cell[i]->subCell[j])->rMin(_r);
    break;
  case center : { cout<<"DiffSolver::r: Error: Center locatation not allowed\n"; exit(-1); } 
  }
}

// Get veclocity
double DiffSolver::v(const int& i, const Location& location, const Epoch& epoch) const {
  double _v;

  // Select epoch
  const Grid* gridEpoch=(epoch==old?gridOld:&grid);
  
  // Calculate v (assuming homology)
  _v=r(i,location,epoch)/gridEpoch->cell[i]->t();

  return _v;
}

// Get density
double DiffSolver::rho(const int& i, const Location& location, const Epoch& epoch) const {
  double _rho;

  // Select epoch
  const Grid* gridEpoch=(epoch==old?gridOld:&grid);

  // Switch on location
  switch(location) {
  case front :
    // If not outermost cell interpolate
    if (i<gridEpoch->cell.size()-1) _rho=rhoIpol(rho(i,center,epoch),rho(i+1,center,epoch),dm(i,center,epoch),dm(i+1,center,epoch));
    // Else - Use density at center
    else _rho=gridEpoch->cell[i]->rho();
    break;
  case back :
    // If not innermost cell interpolate
    if (i>0) _rho= _rho=rhoIpol(rho(i,center,epoch),rho(i-1,center,epoch),dm(i,center,epoch),dm(i-1,center,epoch));
    // Else - Use density at center
    else _rho=gridEpoch->cell[i]->rho();
    break;
  case center :
    // Get density at center
    _rho=gridEpoch->cell[i]->rho();
    break;
  }

  return _rho;
}

// Set density
// Location ignored (center assumed)
void DiffSolver::rho(const double& _rho, const int& i, const Location& location, const Epoch& epoch) {
  int j;

  // Select epoch
  const Grid* gridEpoch=(epoch==old?gridOld:&grid);

  // Calculate density scaling factor - New!
  double scale=_rho/gridEpoch->cell[i]->rho();

  // Set density
  gridEpoch->cell[i]->rho(_rho);
  // If cell container - Adjust density of each subcell using scaling factor - New!
  if (gridEpoch->cell[i]->type()==cellContainer) for (j=0;j<gridEpoch->cell[i]->subCell.size();j++) gridEpoch->cell[i]->subCell[j]->rho(gridEpoch->cell[i]->subCell[j]->rho()*scale);
}

// Get differential mass
double DiffSolver::dm(const int& i, const Location& location, const Epoch& epoch) const {
  double _dm=0;

  // Select epoch
  const Grid* gridEpoch=(epoch==old?gridOld:&grid);

  // Switch on location
  switch(location) {
  case front :
    // If not outermost cell interpolate
    if (i<gridEpoch->cell.size()-1) _dm=dmIpol(dm(i,center,epoch),dm(i+1,center,epoch));
    // Else - use mass at center
    else _dm=dm(i,center,epoch)/2;
    break;
  case back :
    // If not innermost cell interpolate
    if (i>0) _dm=dmIpol(dm(i,center,epoch),dm(i-1,center,epoch));
    // Else - use mass at center
    else _dm=dm(i,center,epoch)/2;
    break;
  case center :
    // Get differential mass at center
    //_dm=gridEpoch->cell[i]->mMax()-gridEpoch->cell[i]->mMin();
    _dm=gridEpoch->cell[i]->M();
    break;
  }

  return _dm;
}

// Get volume
// Location ignored (center assumed)
double DiffSolver::V(const int& i, const Location& location, const Epoch& epoch) const {
  double _V=0;

  // Select epoch
  const Grid* gridEpoch=(epoch==old?gridOld:&grid);

  // Get volume
  //_V=gridEpoch->cell[i]->V(gridEpoch->cell[i]->t());
  _V=gridEpoch->cell[i]->V();

  return _V;
}

// Get temperature
// Location ignored (center assumed)
double DiffSolver::T(const int& i, const Location& location, const Epoch& epoch) const {
  double _T,expFactor=1.0;

  // Select epoch
  const Grid* gridEpoch=(epoch==old?gridOld:&grid);

  // If inside outer boundary
  if (i<=cellIndexMax) {
    // Get temperature  
    _T=gridEpoch->cell[i]->T();
  }
  // Else (i.e. outer boundary) - New! - Testing!
  else {
    // Get temperature according to configuration
    if (config.outerBoundary.useTJ) _T=gridEpoch->cell[i]->TJ();
    else _T=gridEpoch->cell[i]->T();
    // If configured and current epoch - Correct temperature for expansion
    if (config.outerBoundary.expCorr and epoch==current) { expFactor=grid.t()/gridOld->t(); _T/=expFactor; }
  }

  //if (log()) cout<<"DiffSolver::T, i: "<<i<<", location: "<<location<<", epoch: "<<epoch<<", T: "<<_T<<", cellIndexMax: "<<cellIndexMax<<", expFactor: "<<expFactor<<"\n";

  return _T;
}

// Get pressure
// Location ignored (center assumed)
double DiffSolver::P(const int& i, const Location& location, const Epoch& epoch, double _T) const {
  double _P;

  // Select epoch
  const Grid* gridEpoch=(epoch==old?gridOld:&grid);

  // Get pressure
  _P=gridEpoch->cell[i]->P(_T);
  //cout<<"DiffSolver::P: P: "<<_P<<", T: "<<_T<<"\n";

  return _P;
}

// Get temperature derivative of gas pressure
// Location ignored (center assumed)
double DiffSolver::dPdT(const int& i, const Location& location, const Epoch& epoch) const {
  double _dPdT,_T;

  // Select epoch
  const Grid* gridEpoch=(epoch==old?gridOld:&grid);

  // Get temperature
  _T=T(i,location,epoch);

  // Calculate temperature derivative
  _dPdT=(P(i,location,epoch,_T*(1+dTDerivative/2))-P(i,location,epoch,_T*(1-dTDerivative/2)))/(_T*dTDerivative);

  return _dPdT;
}

// Get energy (per mass)
// Location ignored (center assumed)
double DiffSolver::E(const int& i, const Location& location, const Epoch& epoch, double _T) const {
  double _E;

  // Select epoch
  const Grid* gridEpoch=(epoch==old?gridOld:&grid);

  // Get energy (per mass)
  _E=gridEpoch->cell[i]->E(_T);

  return _E;
}

// Get temperature derivative of energy (per mass)
// Location ignored (center assumed)
double DiffSolver::dEdT(const int& i, const Location& location, const Epoch& epoch) const {
  double _dEdT,_T;

  // Select epoch
  const Grid* gridEpoch=(epoch==old?gridOld:&grid);

  // Get temperature
  _T=T(i,location,epoch);

  // Calculate temperature derivative
  _dEdT=(E(i,location,epoch,_T*(1+dTDerivative/2))-E(i,location,epoch,_T*(1-dTDerivative/2)))/(_T*dTDerivative);

  return _dEdT;
}

// Get (Rosseland mean) opactity
double DiffSolver::kappa(const int& i, const Location& location, const Epoch& epoch) {
  double _kappa;
  bool containerMC; // New! - Testing!

  // Select epoch
  const Grid* gridEpoch=(epoch==old?gridOld:&grid);
  
  // Switch on location
  switch(location) {
  case front :
    // If not outermost cell interpolate
    if (i<gridEpoch->cell.size()-1) _kappa=kappaIpol(T(i+1,center,epoch),T(i,center,epoch),kappa(i+1,center,epoch),kappa(i,center,epoch));
    // Else - Use opacity at center
    else _kappa=kappa(i,center,epoch);
    // Break
    break;
  case back :
    // If not innermost cell interpolate
    if (i>0) _kappa=kappaIpol(T(i-1,center,epoch),T(i,center,epoch),kappa(i-1,center,epoch),kappa(i,center,epoch));
    // Else - Use opacity at center
    else _kappa=kappa(i,center,epoch);
    // Break
    break;
  case center :
    // Removed!
    //if (config.testSwitch1) { _kappa=gridEpoch->cell[i]->electronScatterKappa(); break; }
    // New! - Testing!
    //if (config.testSwitch1) { _kappa=gridEpoch->cell[i]->rossContKappa(Processes(pEScatter),config.nuMin,config.nuMax,1e-3); break; }
    // Get cell entries in opacity and lock vectors
    vector<double>::iterator kappaEpoch=(epoch==old?pKappaOld.begin():pKappa.begin())+i;
    //vector<omp_lock_t>::iterator kappaLock=(epoch==old?pKappaLockOld.begin():pKappaLock.begin())+i;
    // Lock cell opacity
    /*omp_set_lock(&(*kappaLock));
    // If update
    if (*kappaEpoch==kappaUpdate) {
      // New! - Testing!
      containerMC=(i<=cellIndexMax);
      if (config.testSwitch1) *kappaEpoch=gridEpoch->cell[i]->rossContKappa(Processes(pEScatter),config.nuMin,config.nuMax,1e-3,containerMC);
      //if (config.testSwitch1) *kappaEpoch=gridEpoch->cell[i]->electronScatterKappa();
      else {
	// Get Rosseland mean opacity at center - Changed! - Testing!
	*kappaEpoch=gridEpoch->cell[i]->rossContKappa(Processes(pEScatter|pBoundFree|pFreeFree),config.nuMin,config.nuMax,1e-3,containerMC);
	// If configured - Apply opacity floor
	if (config.kappa.floor) *kappaEpoch=MAX(*kappaEpoch,gridEpoch->cell[i]->kappaFloor());
	// Testing!
	cout<<"DiffSolver::kappa: cell: "<<gridEpoch->cell[i]->index()<<", kappaFloor: "<<gridEpoch->cell[i]->kappaFloor()<<"\n";
      }
    }
    // Unlock cell opacity
    omp_unset_lock(&(*kappaLock));*/
    _kappa=*kappaEpoch;
    // Break
    break;
  }

  return _kappa;
}

// Get luminosity
double DiffSolver::L(const int& i, const Location& location, const Epoch& epoch, double _T) {
  double _L,dTdm;

  // Select epoch
  const Grid* gridEpoch=(epoch==old?gridOld:&grid);

  // If not specified - Get temperature  
  if (_T==noT) _T=T(i,center,epoch);

  // Switch on location
  switch(location) {
  case back :
    // If not innermost cell
    if (i>0) {
      // Calculate L
      _L=-pow(4*pi*pow(r(i,back,epoch),2),2)*(a*c*fluxLimit(i,back,epoch,_T)/(3*kappa(i,back,epoch)))*(pow(_T,4)-pow(T(i-1,center,epoch),4))/dm(i,back,epoch);
      // Hack !!!
      if (fluxLimit(i,back,epoch)<0.5) _L=fabs(_L);
      // Testing!
      //if (log()) cout<<"DiffSolver::L: i: "<<i<<", location: "<<location<<", epoch: "<<epoch<<", r: "<<r(i,back,epoch)<<", kappa: "<<kappa(i,back,epoch)<<", T: "<<_T<<", TPrev: "<<T(i-1,center,epoch)<<", dm: "<<dm(i,back,epoch)<<"\n";
      // Testing!
      //if (log()) cout<<"DiffSolver::L: cell: "<<i<<", r: "<<r(i,back,epoch)<<", kappa: "<<kappa(i,back,epoch)<<", T["<<i<<"]: "<<_T<<", T["<<i-1<<"]: "<<T(i-1,center,epoch)<<", dm: "<<dm(i,back,epoch)<<"\n";
    }
    // Else
    else {
      // Switch on inner boundary method
      switch(config.innerBoundary.method) {
      case LFixed : _L=config.innerBoundary.L; break;
      case dLdmZero : _L=L(i,front,epoch); break;
      default : _L=L(i,front,epoch);
      }
    }
    break;
  case front :
    // If not outermost cell
    if (i<gridEpoch->cell.size()-1) {
      // Calculate L
      _L=-pow(4*pi*pow(r(i,front,epoch),2),2)*(a*c*fluxLimit(i,front,epoch,_T)/(3*kappa(i,front,epoch)))*(pow(T(i+1,center,epoch),4)-pow(_T,4))/dm(i,front,epoch);
      // Hack !!!
      if (fluxLimit(i,front,epoch)<0.5) _L=fabs(_L);
      // Testing!
      //if (log()) cout<<"DiffSolver::L: i: "<<i<<", location: "<<location<<", epoch: "<<epoch<<", r: "<<r(i,front,epoch)<<", kappa: "<<kappa(i,front,epoch)<<", T: "<<_T<<", TNext: "<<T(i+1,center,epoch)<<", dm: "<<dm(i,front,epoch)<<"\n";
    }
    // Else
    else {
      // Switch on outer boundary method
      switch(config.outerBoundary.method) {
      case eddington : _L=LEdd(_T); break;
      case eddingtonMod : _L=LEddMod(_T); break;
      default : return _L=LEdd(_T);
      }
    }
    break;
  case center : { cout<<"DiffSolver::L: Error: Center locatation not allowed\n"; exit(-1); }
  }
  
  return _L;
}

// Get flux
double DiffSolver::F(const int& i, const Location& location, const Epoch& epoch, double _T) {
  double _F;

  // Changed!
  if (r(i,location,epoch)!=0) _F=L(i,location,epoch,_T)/(4*pi*pow(r(i,location,epoch),2));
  else _F=0;

  return _F;
}

// Set flux
void DiffSolver::F(const double& _F, const int& i, const Location& location, const Epoch& epoch) {

  // Select epoch
  const Grid* gridEpoch=(epoch==old?gridOld:&grid);

  switch(location) {
  case back : 
    ((CellSS*) gridEpoch->cell[i])->FBack(_F);
    // Testing!
    //if (log()) cout<<"DiffSolver::F: cell: "<<i<<", F: "<<_F<<"\n";
    // If cell containter - Set flux for each subcell - Fix!!!
    break;
  case front : 
    ((CellSS*) gridEpoch->cell[i])->FFront(_F);
    // Testing!
    //if (log()) cout<<"DiffSolver::F: cell: "<<i<<", F: "<<_F<<"\n";
    // If cell containter - Set flux for each subcell - Fix!!!
    break;
  case center : { cout<<"DiffSolver::F: Error: Center locatation not allowed\n"; exit(-1); }
  }
}

// Set observed flux
void DiffSolver::FObs(const double& _FObs, const int& i, const Location& location, const Epoch& epoch) {

  // Select epoch
  const Grid* gridEpoch=(epoch==old?gridOld:&grid);

  switch(location) {
  case back : 
    ((CellSS*) gridEpoch->cell[i])->FObsBack(_FObs);
    // If cell containter - Set flux for each subcell - Fix!!!
    break;
  case front : 
    ((CellSS*) gridEpoch->cell[i])->FObsFront(_FObs);
    // If cell containter - Set flux for each subcell - Fix!!!
    break;
  case center : { cout<<"DiffSolver::F: Error: Center locatation not allowed\n"; exit(-1); }
  }
}

// Get temperature derivative of luminosity
double DiffSolver::dLdT(const int& i, const Location& location, const Epoch& epoch, double _T) {
  double _dLdT;
  double dTDerivative=config.convergence.dTDerivative;

  // Select epoch
  const Grid* gridEpoch=(epoch==old?gridOld:&grid);

  // Get temperature
  _T=T(i,location,epoch);

  // If partial derivative
  if (config.convergence.derivative==dPartial) {

    // Calculate temperature derivative
    _dLdT=(L(i,location,epoch,_T*(1+dTDerivative/2))-L(i,location,epoch,_T*(1-dTDerivative/2)))/(_T*dTDerivative);  
  }
  // Else (i.e. full derivative)
  // Does not work with subcells!!!
  else {
    double L1,L2;

    // Evaluate L att higher temperature
    gridEpoch->cell[i]->T(_T*(1+dTDerivative/2));
    lteSolver.updateU(gridEpoch->cell[i]);
    lteSolver.updateIonState(gridEpoch->cell[i]);    
    lteSolver.updateBoundState(gridEpoch->cell[i]);
    gridEpoch->cell[i]->reset();
    L1=L(i,location,epoch);

    // Evaluate L at lower temperature
    gridEpoch->cell[i]->T(_T*(1-dTDerivative/2));
    lteSolver.updateU(gridEpoch->cell[i]);
    lteSolver.updateIonState(gridEpoch->cell[i]);    
    lteSolver.updateBoundState(gridEpoch->cell[i]);
    gridEpoch->cell[i]->reset();
    L2=L(i,location,epoch);

    // Reset temperature and matter state
    gridEpoch->cell[i]->T(_T);
    lteSolver.updateU(gridEpoch->cell[i]);
    lteSolver.updateIonState(gridEpoch->cell[i]);    
    lteSolver.updateBoundState(gridEpoch->cell[i]);
    gridEpoch->cell[i]->reset();

    // Calculate derivative of L
    _dLdT=(L1-L2)/(_T*dTDerivative);

    double dLdTP=(L(i,location,epoch,_T*(1+dTDerivative/2))-L(i,location,epoch,_T*(1-dTDerivative/2)))/(_T*dTDerivative);
  }
  
  return _dLdT;
}

// Get flux limit
double DiffSolver::fluxLimit(const int& i, const Location& location, const Epoch& epoch, double _T) {
  double R,_fluxLimit;

  // Select epoch
  const Grid* gridEpoch=(epoch==old?gridOld:&grid);

  // If not specified - Get temperature
  if (_T==noT) _T=T(i,center,epoch);

  // If use of the flux limiter requested
  if (config.fluxLimiter) {
    
    // Switch on location
    switch(location) {
    case back :
      // Calculate R
      R=((4*pi*pow(r(i,back,epoch),2))/(kappa(i,back,epoch)*dm(i,back,epoch)))*(fabs(pow(_T,4)-pow(T(i-1,center,epoch),4))/(0.5*(pow(_T,4)+pow(T(i-1,center,epoch),4))));
      break;
    case front :
      // If not outermost cell - Calculate R
      if (i<gridEpoch->cell.size()-1) R=((4*pi*pow(r(i,front,epoch),2))/(kappa(i,front,epoch)*dm(i,front,epoch)))*(fabs(pow(T(i+1,center,epoch),4)-pow(_T,4))/(0.5*(pow(T(i+1,center,epoch),4)+pow(_T,4))));
      // Else - Assume T=0 outside the surface (but only in the difference and not in the average)
      else R=((4*pi*pow(r(i,front,epoch),2))/(kappa(i,front,epoch)*dm(i,front,epoch)));
      break;
    case center : { cout<<"DiffSolver::fluxLimit: Error: Center locatation not allowed\n"; exit(-1); }
    }
    _fluxLimit=(6+3*R)/(6+3*R+pow(R,2));
  }
  // Otherwise just return 1
  else _fluxLimit=1.0;
  
  return _fluxLimit;
}

// Get adiabatic sound speed
// Location ignored (center assumed)
double DiffSolver::vs(const int& i, const Location& location, const Epoch& epoch) const {
  double _vs;

  // Select epoch
  const Grid* gridEpoch=(epoch==old?gridOld:&grid);

  // Calculate adiabatic sound speed
  // CP/CV=5/3 should be replaced by the effective CP/CV for a mixed gas consting of matter and radiation
  _vs=sqrt((5.0/3.0)*P(i,center,epoch)/rho(i,center,epoch));

  return _vs;
}

// Get radioactive heating
// Location ignored (center assumed)
double DiffSolver::eRadio(const int& i, const Location& location, const Epoch& epoch) const {
  double _eRadio;

  // Select epoch
  const Grid* gridEpoch=(epoch==old?gridOld:&grid);

  // Get radioactive heating
  _eRadio=gridEpoch->cell[i]->eRadio();

  return _eRadio;
}

// Solve for temperature
void DiffSolver::solveTemperature(vector<double>& dT) {
  Vec_DP aa(0.0,cellIndexMax+1),bb(0.0,cellIndexMax+1),cc(0.0,cellIndexMax+1),rhs(0.0,cellIndexMax+1),_dT(0.0,cellIndexMax+1);
  bool steadyState=config.steadyState;
  bool includeRadio=config.includeRadio;
  //double phi=(steadyState?1.0:config.phi);
  double phi;
  double t;
  int i;
  
  //if (log()) cout<<"DiffSolver::solveTemperature: cellIndexMax: "<<cellIndexMax<<"\n";
 
  // For each cell
  int iMax=cellIndexMax;
  #pragma omp parallel for firstprivate(phi,t)
  //for (i=0;i<=cellIndexMax;i++) {
  for (i=0;i<=iMax;i++) {

    // Set phi according to configuration
    if (config.steadyState) phi=1.0; 
    else {
      if (i==cellIndexMax and config.outerBoundary.phiZero) {
	phi=0;
      }
      else phi=config.phi;
    }

    // Testing
    t=grid.cell[i]->t();

    // Calculate the previous cell diagonal

    // Add (the derivative of) the dL/dm term
    if (i>0) aa[i]+=-phi*dLdT(i-1,front)*(dt/dm(i,center));

    // Calculate the cell diagonal
 
    // Add (the derivative of) the dE/dt term
    if (not steadyState) bb[i]+=dEdT(i,center);
    // Add (the derivative of) the P*d(1/rho)/dt term
    if (not steadyState) bb[i]+=0.5*dPdT(i,center)*(pow(rho(i,center),-1)-pow(rho(i,center,old),-1));
    // Testing
    else if (config.expansion) bb[i]+=3*dPdT(i,center)*pow(rho(i,center),-1)*(dt/t);
    // Add (the derivative of) the dL/dm term
    bb[i]+=phi*(dLdT(i,front)-dLdT(i,back))*(dt/dm(i,center));

    // Calculate the next cell diagonal

    // Add (the derivative of) the dL/dm term
    if (i<cellIndexMax) cc[i]+=phi*dLdT(i+1,back)*(dt/dm(i,center));

    // Calculate the right-hand side

    // Add the dEdt term
    if (not steadyState) rhs[i]=-(E(i,center)-E(i,center,old));
    // Add the P*d(1/rho)/dt term
    if (not steadyState) rhs[i]+=-0.5*(P(i,center)+P(i,center,old))*(pow(rho(i,center),-1)-pow(rho(i,center,old),-1));
    // Testing
    else if (config.expansion) rhs[i]+=-3*P(i,center)*pow(rho(i,center),-1)*(dt/t);
    // Add the dL/dm term
    rhs[i]+=-phi*(L(i,front)-L(i,back))*(dt/dm(i,center));
    if (not steadyState) rhs[i]+=-(1-phi)*(L(i,front,old)-L(i,back,old))*(dt/dm(i,center));
    // Add the radioactive heating term
    if (includeRadio) rhs[i]+=0.5*(eRadio(i,center)+eRadio(i,center,old))*dt;
  }

  // If luminosity and temperature fixed at the outer boundary
  if (config.outerBoundary.LTFixed) {
    
    // Shift equations one step (skipping the first)
    for (i=0;i<cellIndexMax;i++) { aa[i]=aa[i+1]; bb[i]=bb[i+1]; cc[i]=cc[i+1]; rhs[i]=rhs[i+1]; }

    // Add one new equation for luminosity and temperature at outer boundary
    aa[cellIndexMax]=dLdT(cellIndexMax,front)*(dt/dm(cellIndexMax,center)); 
    bb[cellIndexMax]=0; cc[cellIndexMax]=0;
    rhs[cellIndexMax]=-(L(cellIndexMax,front)-config.outerBoundary.L)*(dt/dm(cellIndexMax,center));

    // Solve the shifted tri-diagonal euqation system for dT
    tridag_upshift(aa,bb,cc,rhs,_dT);
  }
  // Else - Solve the tri-diagonal equation system for dT
  else tridag(aa,bb,cc,rhs,_dT);

  // Get temperature corrections
  for (i=0;i<=cellIndexMax;i++) {
    dT[i]=_dT[i];
  }
}

// Update matter state
// Old version.
/*void DiffSolver::updateMatterState(const vector<double>& dT) {
  double dampT=config.convergence.dampT;
  double dTMax=config.convergence.dTMax;
  double TMin=config.convergence.TMin;
  int i;

  // For each cell
  for (i=0;i<=cellIndexMax;i++) {

    // Update temperature
    grid.cell[i]->T(MAX(grid.cell[i]->T()+MAX(MIN(dT[i]*dampT,dTMax),-dTMax),TMin));

    // Update partition function
    lteSolver.updateU(grid.cell[i]);

    // Update ionization state
    lteSolver.updateIonState(grid.cell[i]);
    
    // Update bound state
    lteSolver.updateBoundState(grid.cell[i]);

    // Update edge list (needed by rossContKappa)
    grid.cell[i]->updateEdgeList();

    // Reset derived quantities (except bound-free cut-off)
    //grid.cell[i]->reset();
    grid.cell[i]->reset(dpAll,false);

    // Reset opacity
    pKappa[i]=kappaUpdate;
  }
  }*/

// Update (Rosseland mean) opacity for grid
// Does not work for subcells - Fix!
void DiffSolver::updateKappa(const Epoch& epoch) {
  if (mpiCom==NULL) updateKappaSingle(epoch);
  else updateKappaCluster(epoch);
}

// Update (Rosseland mean) opacity for grid (on a single computer)
// Does not work for subcells - Fix!
void DiffSolver::updateKappaSingle(const Epoch& epoch) {
  int i;

  // Select epoch
  const Grid* gridEpoch=(epoch==old?gridOld:&grid);

  int iMax=MIN(gridEpoch->cell.size()-1,cellIndexMax+1);

  // Update opacity for the cells
  #pragma omp parallel for
  for (i=0;i<=iMax;i++) updateKappa(i,epoch);
}

// Update (Rosseland mean) opacity for grid (on a cluster using MPI)
// Does not work for subcells - Fix!
void DiffSolver::updateKappaCluster(const Epoch& epoch) {
  int i,nResponsible=0;

  // Select epoch
  const Grid* gridEpoch=(epoch==old?gridOld:&grid);

  int iMax=MIN(gridEpoch->cell.size()-1,cellIndexMax+1);

  // Determine which processes are responsible for the cells
  for (i=0;i<=iMax;i++) gridEpoch->cell[i]->responsible=gridEpoch->cell[i]->index()%mpiCom->size();

  // Count number of cells the process is responsible for
  for (i=0;i<=iMax;i++) if (gridEpoch->cell[i]->responsible==mpiCom->rank()) nResponsible++;

  // Update opacity for the cells the process are responsible for
  #pragma omp parallel for
  for (i=0;i<=iMax;i++) {
    if (gridEpoch->cell[i]->responsible==mpiCom->rank()) {
      // Set number of threads to use in nested omp parallelization
      omp_set_num_threads((omp_get_num_threads()-1)/nResponsible+1);
      updateKappa(i,epoch);
    }
  }

  // Wait for all processes to reach this point
  mpiCom->barrier();

  // Synchronize opacity for each cell
  for (i=0;i<=iMax;i++) syncKappa(i,epoch);

  // Wait for all processes to reach this point
  mpiCom->barrier();
}

// Update (Rosseland mean) opacity for cell
// Does not work for subcells (see below) - Fix!
void DiffSolver::updateKappa(const int& i, const Epoch& epoch) {

  // Select epoch
  const Grid* gridEpoch=(epoch==old?gridOld:&grid);
  vector<double>::iterator kappaEpoch=(epoch==old?pKappaOld.begin():pKappa.begin())+i;
  
  //if (config.testSwitch1) *kappaEpoch=gridEpoch->cell[i]->rossContKappa(Processes(pEScatter),config.nuMin,config.nuMax,1e-3,config.kappa.MC);
  if (config.testSwitch1) 
    // Testing
    //if (config.kappaMC) *kappaEpoch=gridEpoch->cell[i]->rossKappaMC2(Processes(pEScatter),config.nuMin,config.nuMax,100000);
    if (config.kappa.MC) *kappaEpoch=gridEpoch->cell[i]->rossKappaMC2(Processes(pEScatter),config.nuMin,config.nuMax,100000);
    else *kappaEpoch=gridEpoch->cell[i]->rossContKappa(Processes(pEScatter),config.nuMin,config.nuMax,1e-3);
  else {
    // If bound-free processes included - New!
    if (config.kappa.processes(pBoundFree)) {
	// If needed (should be right) - Update of bound-free opacity cut-off and tables in advance - Testing!
	// This needs to be done for subcells as well!!!
	if (gridEpoch->cell[i]->boundFreeCutOffUpdate) { gridEpoch->cell[i]->updateBoundFreeCutOffOMP(false,config.boundFreeCutOff); gridEpoch->cell[i]->boundFreeCutOffUpdate=false; }
	if (gridEpoch->cell[i]->boundFreeTableUpdate) { gridEpoch->cell[i]->boundFreeAbsKappaTable(true); gridEpoch->cell[i]->boundFreeTableUpdate=false; }
      }
    // Get Rosseland mean opacity at center - Changed! - Testing!
    //*kappaEpoch=gridEpoch->cell[i]->rossContKappa(Processes(pEScatter|pBoundFree|pFreeFree),config.nuMin,config.nuMax,1e-3,config.kappa.MC);
    // Testing!
    //if (config.kappaMC) *kappaEpoch=gridEpoch->cell[i]->rossKappaMC2(Processes(pEScatter|pBoundFree|pFreeFree|pBoundBound),config.nuMin,config.nuMax,100000);
    //else *kappaEpoch=gridEpoch->cell[i]->rossContKappa(Processes(pEScatter|pBoundFree|pFreeFree),config.nuMin,config.nuMax,1e-3);
    if (config.kappa.MC) *kappaEpoch=gridEpoch->cell[i]->rossKappaMC2(config.kappa.processes,config.nuMin,config.nuMax,100000);
    else *kappaEpoch=gridEpoch->cell[i]->rossContKappa(config.kappa.processes,config.nuMin,config.nuMax,1e-3);
    // If configured - Apply opacity floor
    //if (config.kappaFloor) *kappaEpoch=MAX(*kappaEpoch,gridEpoch->cell[i]->kappaFloor());
    if (config.kappa.floor) *kappaEpoch=MAX(*kappaEpoch,gridEpoch->cell[i]->kappaFloor()); // Changed!
  }

  cout<<"DiffSolver: updateKappa: cell: "<<i<<", Updated, kappa: "<<*kappaEpoch<<"\n";
}

// Synchronize (Rosseland mean) opacity for cell
void DiffSolver::syncKappa(const int& i, const Epoch& epoch) {
  int j,k=0;

  // Select epoch
  const Grid* gridEpoch=(epoch==old?gridOld:&grid);
  vector<double>::iterator kappaEpoch=(epoch==old?pKappaOld.begin():pKappa.begin())+i;

  // Hmm, maybe use broadcast instead ... 

  // Initiate MPI requests
  int nRequest=(mpiCom->rank()==gridEpoch->cell[i]->responsible)?(mpiCom->size()-1):1;
  boost::mpi::request request[nRequest]; 

  // If responsible
  if (mpiCom->rank()==gridEpoch->cell[i]->responsible) {

    // For each process
    for (j=0;j<mpiCom->size();j++) {
      // If not self
      if (j!=mpiCom->rank()) {
	// Deliver opacity (non-blocking)
	request[k]=mpiCom->isend(j,gridEpoch->cell[i]->index(),*kappaEpoch);
	k++;
      }
    }
  }
  // Else (i.e. not responsible)
  else {

    // Collect opacity (non-blocking)
    request[0]=mpiCom->irecv(gridEpoch->cell[i]->responsible,gridEpoch->cell[i]->index(),*kappaEpoch);
  }

  // Wait for all MPI requests to complete
  boost::mpi::wait_all(request,request+nRequest);

  //if (log()) cout<<"DiffSolver: syncKappa: cell: "<<i<<", Synchronized, kappa: "<<*kappaEpoch<<"\n";
}

// Update matter state (grid version) - New!
void DiffSolver::updateMatterState(const vector<double>& dT) {
  double dampT=config.convergence.dampT;
  double dTMax=config.convergence.dTMax;
  double TMin=config.convergence.TMin;
  double _T;
  int i;

  // For each cell
  for (i=0;i<=cellIndexMax;i++) {
 
    // Calculate new temperature
    _T=MAX(grid.cell[i]->T()+MAX(MIN(dT[i]*dampT,dTMax),-dTMax),TMin);

    if (log()) cout<<"DiffSolver::updateMatterState: cellIndex: "<<i<<", T: "<<_T<<", grid.cell["<<i<<"].T: "<<grid.cell[i]->T()<<", T("<<i<<",center): "<<T(i,center)<<", dT: "<<dT[i]<<", dampT: "<<dampT<<", dTMax: "<<dTMax<<", TMin: "<<TMin<<"\n";

    // Update matter state
    updateMatterState(grid.cell[i],_T);

    if (log()) cout<<"DiffSolver::updateMatterState: cellIndex: "<<i<<", T: "<<_T<<", grid.cell["<<i<<"].T: "<<grid.cell[i]->T()<<", T("<<i<<",center): "<<T(i,center)<<"\n";

    // Reset opacity - Removed! - Testing!
    //pKappa[i]=kappaUpdate;
  }

  // Update opacity - New! - Testing!
  updateKappa(current);
}

// Update matter state - (cell version) - New
//void DiffSolver::updateMatterState(Cell* cell, const double& dT) {
void DiffSolver::updateMatterState(Cell* cell, const double& _T) {
  int i,j;

  // If container - Update matter state for each subcell and return
  //if (cell->type()==cellContainer) { for (j=0;j<cell->subCell.size();j++) updateMatterState(cell->subCell[j],dT); return; }
  if (cell->type()==cellContainer) { for (j=0;j<cell->subCell.size();j++) updateMatterState(cell->subCell[j],_T); return; }

  // Update temperature
  //cell->T(MAX(cell->T()+MAX(MIN(dT*dampT,dTMax),-dTMax),TMin));
  cell->T(_T);

  // Update diffusion temperature - Testing!
  cell->TDiff(_T);

  //if (log()) cout<<"DiffSolver::updateMatterState: cellIndex: "<<cell->index()<<", T: "<<_T<<", cell->T: "<<cell->T()<<"\n";

  // Update partition function
  lteSolver.updateU(cell);

  // Update ionization state
  lteSolver.updateIonState(cell);
    
  // Update bound state
  lteSolver.updateBoundState(cell);

  // Update edge list (needed by rossContKappa)
  cell->updateEdgeList();

  // Reset derived quantities (except bound-free cut-off)
  //cell->reset();
  cell->reset(dpAll,false);
}

// Update radiation state
void DiffSolver::updateRadiationState() {
  int i,j; // Changed!

  // For each cell
  for (i=0;i<=cellIndexMax;i++) {

    // If container - New!
    if (grid.cell[i]->type()==cellContainer)
      // Set blackbody radiation field at matter temperature for each subcell
      for (j=0;j<grid.cell[i]->subCell.size();j++) {
	grid.cell[i]->subCell[j]->TR(grid.cell[i]->subCell[j]->T());
	grid.cell[i]->subCell[j]->W(1.0);
      }
    // Set blackbody radiation field at matter temperature
    grid.cell[i]->TR(grid.cell[i]->T());
    grid.cell[i]->W(1.0);
    
    // Update flux at inner border
    F(F(i,back),i,back);
    FObs(F(i,back)+v(i,back)*(4.0/3.0)*a*pow(T(i,back),4),i,back);
    
    // Update flux at outer border
    if (i==cellIndexMax and config.outerBoundary.LTFixed) {
      F(config.outerBoundary.L/(4*pi*pow(r(i,front),2)),i,front);
      FObs(config.outerBoundary.L/(4*pi*pow(r(i,front),2))+v(i,front)*(4.0/3.0)*a*pow(T(i,front),4),i,front);
    }
    else {
      //if (log()) cout<<"DiffSolver::updateRadiationState(1): cell: "<<i<<", LFront: "<<((CellSS* ) grid.cell[i])->LFront()<<", L(front): "<<L(i,front)<<", kappa(front): "<<", FFront: "<<((CellSS* ) grid.cell[i])->FFront()<<", F(front): "<<F(i,front)<<", kappa(front): "<<kappa(i,front)<<"\n";
      F(F(i,front),i,front);
      FObs(F(i,front)+v(i,front)*(4.0/3.0)*a*pow(T(i,front),4),i,front);
      //if (log()) cout<<"DiffSolver::updateRadiationState(2): cell: "<<i<<", LFront: "<<((CellSS* ) grid.cell[i])->LFront()<<", L(front): "<<L(i,front)<<", kappa(front): "<<", FFront: "<<((CellSS* ) grid.cell[i])->FFront()<<", F(front): "<<F(i,front)<<", kappa(front): "<<kappa(i,front)<<"\n";
    }
  }
}

// Temperature convergence criteria
void DiffSolver::updateTConvergence(const vector<double>& dT) {
  vector<double> _rErrT(cellIndexMax+1);
  int i;

  // Calculate relative error in temperature in each cell
  for (i=0;i<=cellIndexMax;i++) {
    _rErrT[i]=fabs(dT[i]/T(i,center));
    if (log()) cout<<"DiffSolver::updateTConvergence: cellIndex: "<<i<<", T: "<<T(i,center)<<", dT: "<<dT[i]<<", rErrT: "<<_rErrT[i]<<"\n";
  }

  // Calculate (maximum) relative error in temperature
  rErrT=max(_rErrT);
}

// Temperature converged
bool DiffSolver::TConverged() {

  if (log()) cout<<"DiffSolver::converged: rErrT: "<<rErrT<<", TTol: "<<config.convergence.TTol<<"\n";

  // If error in temperature less than tolerance return true
  if (rErrT<config.convergence.TTol and n>=config.convergence.nIterMin) return true;
    
  return false;
}

// Evolve/update thermodynamical state of grid
void DiffSolver::evolveThermo() {
  int i,j,nIter=config.convergence.nIter; // Changed!
  vector<double> dT(cellIndexMax+1,0);
  double expFactor;

  //if (log()) cout<<"DiffSolver::evolveThermo, expansion: "<<config.expansion<<"\n";

  // If outermost cell has an outer neighbour and expansion correction on
  // Moved to DiffSolver:T function!!! - Testing!!!
  /*if (cellIndexMax<(grid.cell.size()-1) and config.outerBoundary.expCorr) {
    // Calculate expansion factor
    expFactor=grid.t()/gridOld->t();
    // Correct temperature in the outer neighbour of the outermost cell - Changed!
    if (grid.cell[cellIndexMax+1]->type()==cellContainer)
      for (j=0;j<grid.cell[cellIndexMax+1]->subCell.size();j++) grid.cell[cellIndexMax+1]->subCell[j]->T(gridOld->cell[cellIndexMax+1]->subCell[j]->T()/expFactor);
    else grid.cell[cellIndexMax+1]->T(gridOld->cell[cellIndexMax+1]->T()/expFactor);
    //if (log()) cout<<"DiffSolver::evolveThermo: Expansion corrected temperature in cell "<<cellIndexMax+1<<", T: "<<gridOld->cell[cellIndexMax+1]->T()<<"->"<<grid.cell[cellIndexMax+1]->T()<<"\n";
    }*/
  
  // Update matter state - Testing !!!
  updateMatterState(dT);

  // Iteration loop
  for (n=0;n<nIter;n++) {

    //if (log()) cout<<"DiffSolver::evolveThermo: n: "<<n<<"\n";
    
    // Solve for temperature
    solveTemperature(dT);

    //if (log()) cout<<"DiffSolver::evolveThermo: 1\n";

    // Update convergence criteria
    updateTConvergence(dT);

    //if (log()) cout<<"DiffSolver::evolveThermo: 2\n";

    // If converged break
    if (TConverged()) break;

    // Update matter state
    updateMatterState(dT);
  }
  if (n==nIter and log()) cout<<"DiffSolver::evolveThermo: State did not converge\n";
  else if (log()) cout<<"DiffSolver::evolveThermo: State converged in "<<n<<" iterations\n";
}

// Evolve dynamical state of grid (assuming homology)
void DiffSolver::evolveDyn() {
  double dtCenter;
  int i,j; // Changed!
 
  if (log()) cout<<"DiffSolver::evolveDyn\n";

  // For each cell
  for (i=0;i<grid.cell.size();i++) {
    
    // Calculate and set new (front) radius
    r(r(i,front,old)+v(i,front,old)*dt,i,front);
    // If not innermost cell - Set new (back) radius to (front) radius for previous cell
    if (i>0) r(r(i-1,front),i,back);
    // Else (i.e. innermost cell) - Calculate and set new (back radius) 
    else r(r(i,back,old)+v(i,back,old)*dt,i,back);
    
    // Calculate and set new density
    rho(rho(i,center,old)*pow(r(i,center,old)/r(i,center),3),i,center);

    // Removed - Does not work - Solve !

    // Calculate time centered time step
    // I don't think this works in general - Improve!
    //dtCenter=2*(gridOld->cell[i]->t()+dt-gridOld->cell[i]->tMax());

    // Update time borders
    //grid.cell[i]->tMin(gridOld->cell[i]->tMax());
    //grid.cell[i]->tMax(grid.cell[i]->tMin()+dtCenter);

    // Update time borders - Hack !
    grid.cell[i]->tMin(gridOld->cell[i]->t()+dt/2);
    grid.cell[i]->tMax(grid.cell[i]->tMin()+dt);
    // If container - Update time border for each subcell - New!
    if (grid.cell[i]->type()==cellContainer) 
      for (j=0;j<grid.cell[i]->subCell.size();j++) {
	grid.cell[i]->subCell[j]->tMin(grid.cell[i]->tMin());
	grid.cell[i]->subCell[j]->tMax(grid.cell[i]->tMax());
      }
  }

  // Removed - Does not work - Solve !

  // Calculate time centered time step
  // I don't think this works in general - Improve!
  //dtCenter=2*(gridOld->t()+dt-gridOld->tMax());

  // Update grid time borders
  //grid.tMin(gridOld->tMax());
  //grid.tMax(grid.tMin()+dtCenter);

  // Update grid time borders - Hack !
  grid.tMin(gridOld->t()+dt/2);
  grid.tMax(grid.tMin()+dt);

  if (log()) cout<<"DiffSolver::evolveDyn: dt: "<<dt/day<<", grid.tMin: "<<grid.tMin()/day<<", grid.tMax: "<<grid.tMax()/day<<"\n";
}

// Evolve/update radioactive state of grid
void DiffSolver::evolveRadio() {
  bool mCsteadyState=true;
  bool transformKappa=false;
  int i,j;

  if (log()) cout<<"DiffSolver::evolveRadio\n";
  
  // If not steady-state
  if (not config.steadyState)
    // For each cell evolve radioactive state
    // Is this correct - should we not use gridOld.t()!?!
    for (i=0;i<grid.cell.size();i++) grid.cell[i]->evolveRadio(grid.t(),dt);

  // Setup gamma-ray MC experiment
  //PacketConfig packetConfig = PacketConfig(mCsteadyState,false,false,false,false,false,false,false,0,false,false,false,false,urOnTheFly,false,false,-1,-1,false,1.0,false,1.0,transformKappa,false);
  // Testing!
  PacketConfig packetConfig = PacketConfig(mCsteadyState,false,false,false,false,false,false,false,false,0,false,false,false,false,false,false,false,urOnTheFly,true,false,-1,-1,false,1.0,false,1.0,false,transformKappa,false); // Changed!
  MCConfig mCConfig = MCConfig(&packetConfig);
  PacketSamplerGamma packetSampler = PacketSamplerGamma(config.nPacketGamma,&grid,&packetConfig,mpiCom);
  MCExperiment mCExperiment = MCExperiment(&packetSampler,NULL,&mCConfig,mpiCom);

  // Reset g-packet MC statistics
  for (i=0;i<grid.cell.size();i++) grid.cell[i]->resetGMCStatistics();

  // Run gamma-ray MC experiment
  mCExperiment.run();

  // Update radioactive energy deposition
  // Only works with on-the-fly update method !
  for (i=0;i<grid.cell.size();i++) grid.cell[i]->updateERadio();

  // Restore radioactive energy deposition in the outer region - Hack! - Changed!
  for (i=cellIndexMax+1;i<grid.cell.size();i++) {
    if (grid.cell[i]->type()==cellContainer)
      for (j=0;j<grid.cell[i]->subCell.size();j++) grid.cell[i]->subCell[j]->eRadio(gridOld->cell[i]->subCell[j]->eRadio());
    grid.cell[i]->eRadio(gridOld->cell[i]->eRadio());
  }

  // Update non-thermal rates (in the inner region) - Changed!
  for (i=0;i<=cellIndexMax;i++) {
    // If cell container - Update non-thermal rates for each subcell - New!
    if (grid.cell[i]->type()==cellContainer) {
      for (j=0;j<grid.cell[i]->subCell.size();j++) {
	grid.cell[i]->subCell[j]->nonThermalHeating(MAX(grid.cell[i]->subCell[j]->eRadio()*grid.cell[i]->subCell[j]->rho(),1e-99)); // Changed!
	grid.cell[i]->subCell[j]->nonThermalIonization(0.0);
	grid.cell[i]->subCell[j]->nonThermalExcitation(0.0);
	if (log()) cout<<"DiffSolver::evolveRadio: cell: "<<i<<", subCell: "<<j<<", nonThermalHeating: "<<grid.cell[i]->subCell[j]->nonThermalHeating()<<", eRadio: "<<grid.cell[i]->subCell[j]->eRadio()<<"\n";
      }
    }
    // Else
    else {
      grid.cell[i]->nonThermalHeating(MAX(grid.cell[i]->eRadio()*grid.cell[i]->rho(),1e-99));
      grid.cell[i]->nonThermalIonization(0.0);
      grid.cell[i]->nonThermalExcitation(0.0);
    }
    if (log()) cout<<"DiffSolver::evolveRadio: cell: "<<i<<", type: "<<grid.cell[i]->type()<<", nonThermalHeating: "<<grid.cell[i]->nonThermalHeating()<<", eRadio: "<<grid.cell[i]->eRadio()<<"\n";
  }
}

// Evolve state (the given timestep)
void DiffSolver::evolve() {

  // Evolve dynamical state of grid
  evolveDyn();
  // Evolve radioactive state of grid
  if (config.includeRadio) evolveRadio();
  // Evolve thermodynamical state of grid
  evolveThermo();
}

// Set timestep
void DiffSolver::setTimeStep() {
  double C=0.5,dtC;
  int i;
  
  // Calculate the timestep from the Courant condition
  dt=config.timeStep.dtMax;
  for (i=0;i<=cellIndexMax;i++) {
    dtC=C*(r(i,front)-r(i,back))/vs(i,center);
    if (dtC<dt) dt=dtC;
  }

  if (log()) cout<<"DiffSolver::setTimeStep: Time step dt="<<dt<<"\n";
}

// Refine timestep
void DiffSolver::refineTimeStep() {

  // If temperature converged
  if (TConverged()) {
    // Chose new timestep based on the changes in density and temperature
    dt=dt/MIN(MAX(MAX(drhoMax/config.timeStep.drhoMax,dTMax/config.timeStep.dTMax),config.timeStep.dtFactorMin),config.timeStep.dtFactorMax);
  }
  // Else
  else {
    // Half the time step
    dt=dt/2.0;
  }

  if (log()) cout<<"DiffSolver::refineTimeStep: New time step dt="<<dt<<"\n";
}

// Update convergence criteria
void DiffSolver::updateConvergence() {
  double drho,dT;
  int i;

  drhoMax=0;
  dTMax=0;

  // For each cell
  for (i=0;i<=cellIndexMax;i++) {

    // Get maximum change in density
    drho=fabs(rho(i,center)-rho(i,center,old))/rho(i,center);
    if (drho>drhoMax) drhoMax=drho;

    // Get maximum change in temperature
    dT=fabs(T(i,center)-T(i,center,old))/T(i,center);
    if (dT>dTMax) dTMax=dT;
  }
}

// Solution converged
bool DiffSolver::converged() {

  // If maxiumum changes in density and temperature below limits return true
  if (TConverged() and drhoMax<config.timeStep.drhoMax and dTMax<config.timeStep.dTMax and m>=config.timeStep.nIterMin) return true;

  if (log()) cout<<"DiffSolver::converged: drhoMax: "<<drhoMax<<", limit: "<<config.timeStep.drhoMax<<", dTMax: "<<dTMax<<", limit: "<<config.timeStep.dTMax<<"\n";

  return false;
}

// Evolve state
void DiffSolver::evolve(Grid* _grid, double& _dt, const int& _cellIndexMax, bool initiate, bool initiateState) {
  int nIter=config.timeStep.nIter;
  // Hack !!!
  int i,j; // Changed!

  //if (log()) cout<<"DiffSolver::evolve: 1\n";
  
  // If requested initiate
  if (initiate) init(_grid,initiateState);

  //if (log()) cout<<"DiffSolver::evolve: 2\n";

  // Store maximum cell index
  cellIndexMax=_cellIndexMax;

  // Update bound-free cut-off (based on opacities only) - Moved to updateKappa! - Testing!
  // Note - updateBoundFreeCutOff not set to false - Is this ok ?!
  //for (i=0;i<=cellIndexMax;i++) { _grid->cell[i]->updateBoundFreeCutOffOMP(false,config.boundFreeCutOff); _grid->cell[i]->boundFreeCutOffUpdate=false; }
  //pragma omp parallel for
  /*for (i=0;i<=_cellIndexMax;i++) { 
    // If cell container - Update bound-free cut-off for each subcell - New!
    if (_grid->cell[i]->type()==cellContainer) {
      for (j=0;j<_grid->cell[i]->subCell.size();j++) {
	_grid->cell[i]->subCell[j]->updateBoundFreeCutOff(false,config.boundFreeCutOff); 
	_grid->cell[i]->subCell[j]->boundFreeCutOffUpdate=false; 
      }
    }
    // Else
    else {
      _grid->cell[i]->updateBoundFreeCutOff(false,config.boundFreeCutOff); 
      _grid->cell[i]->boundFreeCutOffUpdate=false;
    }
    }*/

  // Store (pointer to) grid
  gridOld=_grid;

  // Update old grid opacity - New! - Testing!
  updateKappa(old);

  // Testing!
  logSysMemory("D1");

  // Copy grid
  //grid=(*_grid);
  copyGrid(); // Changed - Testing!!!

  // Testing!
  logSysMemory("D2");

  //if (log()) cout<<"DiffSolver::evolve: 3\n";

  // Set timestep
  if (config.timeStep.set) setTimeStep();
  else dt=_dt;

  // Hack !!!
  if (cellIndexMax<0) { evolveDyn(); (*_grid)=grid; return; }
   
  // Iteration loop
  for (m=0;m<nIter;m++) {

    if (log()) cout<<"DiffSolver::evolve: iteration: "<<m<<"\n";

    // Copy grid
    //if (m>0) grid=(*_grid);
    if (m>0) copyGrid();

    // Evolve state of grid (the given timestep)
    evolve();

    // Update convergence criteria
    updateConvergence();
    
    // If converged break
    if (converged()) break;

    // Refine timestep
    //refineTimeStep();
    // If not last iteration - Refine timestep - Changed!
    if (m<nIter-1) refineTimeStep();
  }
  if (m==nIter and log()) cout<<"DiffSolver::update: State did not converge\n";
  else if (log()) cout<<"DiffSolver::update: State converged in "<<m+1<<" iterations\n";

  // Update radiation state
  updateRadiationState();

  // Testing!
  if (log()) for (i=0;i<grid.cell.size();i++) cout<<"DiffSolver::update(1): cell: "<<i<<", LFront: "<<((CellSS* ) grid.cell[i])->LFront()<<", L(front): "<<L(i,front)<<", kappa(front): "<<kappa(i,front)<<"\n";

  //if (log()) cout<<"DiffSolver::evolve: 4\n";

  // If outermost cell has an outer neighbour and expansion correction on
  // Removed!!! - Testing!!!
  /*if (cellIndexMax<(grid.cell.size()-1) and config.outerBoundary.expCorr) {
    // Restore temperature in the outer neighbour of the outermost cell - Changed!
    if (grid.cell[cellIndexMax+1]->type()==cellContainer)
      for (j=0;j<grid.cell[cellIndexMax+1]->subCell.size();j++) grid.cell[cellIndexMax+1]->subCell[j]->T(gridOld->cell[cellIndexMax+1]->subCell[j]->T());
    else grid.cell[cellIndexMax+1]->T(gridOld->cell[cellIndexMax+1]->T());
    }*/

  //if (log()) cout<<"DiffSolver::evolve: 5\n";

  // Update timestep
  _dt=dt;

  //if (log()) cout<<"DiffSolver::evolve: 3\n";

  // Testing!
  logSysMemory("D3");

  // Testing!
  //if (log()) for (i=0;i<grid.cell.size();i++) cout<<"DiffSolver::update(2): cell: "<<i<<", LFront: "<<((CellSS* ) grid.cell[i])->LFront()<<", L(front): "<<L(i,front)<<", kappa(front): "<<kappa(i,front)<<"\n";

  // Update grid
  //(*_grid)=grid;
  updateGrid();

  // Testing!
  logSysMemory("D4");

  //if (log()) cout<<"DiffSolver::evolve: 4\n";

  // Testing!
  //if (log()) for (i=0;i<grid.cell.size();i++) cout<<"DiffSolver::update(3): cell: "<<i<<", LFront: "<<((CellSS* ) grid.cell[i])->LFront()<<", L(front): "<<L(i,front)<<", kappa(front): "<<kappa(i,front)<<"\n";
  
  // Testing!
  //if (log()) for (i=0;i<grid.cell.size();i++) cout<<"DiffSolver::update: cell: "<<i<<", electronScatterKappa: "<<grid.cell[i]->electronScatterKappa()<<"\n";
}

// Update state (i.e. assume steady state)
// Not up-to-date!!!
void DiffSolver::update(Grid* _grid, const int& _cellIndexMax, bool initiate, bool initiateState) {

  if (log()) cout<<"DiffSolver::update\n";
  
  // If requested initiate
  if (initiate) init(_grid,initiateState);

  // Store maximum cell index
  cellIndexMax=_cellIndexMax;

  // Store (pointer to) grid
  gridOld=_grid;

  // Copy grid
  grid=(*_grid);

  // Set dummy timestep
  dt=1.0;

  // Update radioactive state of grid
  //evolveRadio();

  // Update thermodynamical state of grid
  evolveThermo();

  // Update radiation state
  updateRadiationState();

  // Update grid
  (*_grid)=grid;

  //if (log()) cout<<"DiffSolver::evolve: lowerState: "<<_grid->cell[6]->atom[0].iState[0].bState[1].downTransBB[0]->lowerState<<", upperState: "<<_grid->cell[6]->atom[0].iState[0].bState[1].downTransBB[0]->upperState<<"\n";
}

// Initiate
void DiffSolver::init(Grid* grid, bool initiateState) {
  int i,j; // Changed!

  // If initiate state
  if (initiateState) {

    // Initiate matter state for each cell using LTE solver - Changed!
    for (i=0;i<grid->cell.size();i++) {
      // If container - Initiate matter state for each subcell using LTE solver - New!
      if (grid->cell[i]->type()==cellContainer) {
	for (j=0;j<grid->cell[i]->subCell.size();j++) {
	  lteSolver.update(grid->cell[i]->subCell[j]);
	  grid->cell[i]->subCell[j]->reset();
	}
      }
      // Else
      else {
	// Initiate matter state using LTE solver
	lteSolver.update(grid->cell[i]);
	grid->cell[i]->reset();
      }
    }

    // Initiate radioactive state
    if (config.includeRadio) initRadio(grid);
  }

  // Initiate opacity vectors
  pKappa.resize(grid->cell.size(),kappaUpdate);
  pKappaOld.resize(grid->cell.size(),kappaUpdate);

  // Initiate opacity locks (for openmp parallelization)
  pKappaLock.resize(grid->cell.size());
  pKappaLockOld.resize(grid->cell.size());
  for (i=0;i<pKappaLock.size();i++) omp_init_lock(&(pKappaLock[i]));
  for (i=0;i<pKappaLockOld.size();i++) omp_init_lock(&(pKappaLockOld[i]));
}

// Initiate radioactive state of grid
void DiffSolver::initRadio(Grid* grid) {
  bool mCsteadyState=true;
  bool transformKappa=false;
  int i;

  if (log()) cout<<"DiffSolver::initRadio\n";

  // Setup gamma-ray MC experiment
  //PacketConfig packetConfig = PacketConfig(mCsteadyState,false,false,false,false,false,false,false,0,false,false,false,false,urOnTheFly,false,false,-1,-1,false,1.0,false,1.0,transformKappa,false);
  // Testing!
  PacketConfig packetConfig = PacketConfig(mCsteadyState,false,false,false,false,false,false,false,false,0,false,false,false,false,false,false,false,urOnTheFly,true,false,-1,-1,false,1.0,false,1.0,false,transformKappa,false); // Changed!
  MCConfig mCConfig = MCConfig(&packetConfig);
  PacketSamplerGamma packetSampler = PacketSamplerGamma(config.nPacketGamma,grid,&packetConfig,mpiCom);
  MCExperiment mCExperiment = MCExperiment(&packetSampler,NULL,&mCConfig,mpiCom);

  // Reset g-packet MC statistics
  for (i=0;i<grid->cell.size();i++) grid->cell[i]->resetGMCStatistics();

  // Run gamma-ray MC experiment
  mCExperiment.run();

  // Update radioactive energy deposition
  // Only works with on-the-fly update method !
  for (i=0;i<grid->cell.size();i++) grid->cell[i]->updateERadio();
}

// Get logging flag
bool DiffSolver::log() const {

  // Return logging flag
  return pLog;
}

// Set logging flag
void DiffSolver::log(const bool& _log) {

  // Set logging flag
  pLog=_log;

  // Set LTE solver logging flag
  lteSolver.log(_log);
}

// Get memory allocated - New!
double DiffSolver::mem() const {
  double _mem=sizeof(*this);
  int i;

  // Add memory allocated for (new) grid
  _mem+=grid.mem();

  // Add memory allocated for opacity vectors
  for (i=0;i<pKappa.size();i++) _mem+=sizeof(double);
  for (i=0;i<pKappaOld.size();i++) _mem+=sizeof(double);

  return _mem;
}

// Get memory allcoated for probability lists
double DiffSolver::memPList() const { return grid.memPList(); }

// Get memory allocated for bound-bound transitions - New!
double DiffSolver::memTransBB() const { return grid.memTransBB(); }

// Get memory allocated for bound-free transition - New!
double DiffSolver::memTransBF() const { return grid.memTransBF(); }

// Copy grid - New!
// Workaround for memory leak in grid copy
void DiffSolver::copyGrid() {
  int i,j;

  // If initial copy - Make it complete
  if (grid.cell.size()==0) grid=(*gridOld);
  // Else
  else {
    // For each cell
    for (i=0;i<gridOld->cell.size();i++) {
      // Update cell
      updateCell(gridOld->cell[i],grid.cell[i]);
      // Update each subcell
      for(j=0;j<gridOld->cell[i]->subCell.size();j++) updateCell(gridOld->cell[i]->subCell[j],grid.cell[i]->subCell[j]);
    }
  }

  // Update time-borders
  grid.tMax(gridOld->tMax());
  grid.tMin(gridOld->tMin());
}

// Update grid - New!
// Workaround for memory leak in grid copy
void DiffSolver::updateGrid() {
  int i,j;

  // For each cell
  for (i=0;i<grid.cell.size();i++) {
    // Update cell
    updateCell(grid.cell[i],gridOld->cell[i]);
    // Update each subcell
    for(j=0;j<grid.cell[i]->subCell.size();j++) updateCell(grid.cell[i]->subCell[j],gridOld->cell[i]->subCell[j]);
  }

  // Update time-borders
  gridOld->tMax(grid.tMax());
  gridOld->tMin(grid.tMin());
}

// Update cell - New!
// Workaround for memory leak in grid copy
void DiffSolver::updateCell(Cell* cell1, Cell* cell2) {
  MStateMPI mState;
  
  // Copy matter state
  mState.pack(cell1);
  mState.unpack(cell2);
  // Reset
  cell2->reset();
  // Copy diffusion temperature - Testing!
  // Note: Include in MState instead, right?
  cell2->TDiff(cell1->TDiff());  
  // Copy border fluxes
  ((CellSS*) cell2)->FBack(((CellSS*) cell1)->FBack());
  ((CellSS*) cell2)->FFront(((CellSS*) cell1)->FFront());
  // Copy spatial borders
  ((CellSS*) cell2)->rMin(((CellSS*) cell1)->rMin());
  ((CellSS*) cell2)->rMax(((CellSS*) cell1)->rMax());
  // Copy time borders
  cell2->tMin(cell1->tMin());
  cell2->tMax(cell1->tMax());
  // Copy DBB field
  cell2->TR(cell1->TR());
  cell2->W(cell1->W());
  // Copy TJ
  cell2->TJ(cell1->TJ());
  // Copy radioactive energy deposition - New!
  cell2->eRadio(cell1->eRadio());
}
