#ifndef _DIFFSOLVER_
#define _DIFFSOLVER_

#include <omp.h>
#include "mc.h"
#include "msolver.h"

using namespace mc;
using namespace msolver;

namespace diffsolver {

  // Location
  enum Location { back, center, front };
  
  // Epoch
  enum Epoch { old, current };

  // Temperature differential
  const double dTDerivative=1e-6;

  // Update opacity flag
  const double kappaUpdate=-1;

  // Diffusion solver convergence config
  class DiffConvConfig {
  public :
    int nIter;
    int nIterMin;
    double TTol;
    double neTol;
    double dampT;
    double TMin;
    double dTMax;
    DerivativeType derivative;
    double dTDerivative;
    // Constructors
    DiffConvConfig();
    DiffConvConfig(const int& _nIter, const int& _nIterMin, const double& _TTol, const double& _neTol, const double& _dampT, const double& _TMin, const double& _dTMax, const DerivativeType& _derivative, const double& _dTDerivative);
    // Load data
    void load(const YAML::Node& data);
    // Log data
    void log();
  };

  // Diffusion solver timestep config
  class DiffTStepConfig {
  public :
    bool set;
    bool refine;
    int nIter;
    int nIterMin;
    double drhoMax;
    double dTMax;
    double dtMax;
    double dtFactorMax;
    double dtFactorMin;
    bool lTLimit;
    // Constructors
    DiffTStepConfig();
    DiffTStepConfig(const bool& _set, const bool& _refine, const int& _nIter, const int& _nIterMin, const double& _drhoMax, const double& _dTMax, const double& _dtMax, const double& _dtFactorMin, const double& _dtFactorMax, const bool& _lTLimit);
    // Load data
    void load(const YAML::Node& data);
    // Log data
    void log();
  };

  // Boundary method
  enum BoundaryMethod { eddington, eddingtonMod, LFixed, dLdmZero };

  // Load boundary method
  BoundaryMethod loadBoundaryMethod(const YAML::Node& data);

  // Diffusion solver inner boundary config
  class DiffIBoundConfig {
  public :
    double L;
    BoundaryMethod method;
    // Constructors
    DiffIBoundConfig();
    DiffIBoundConfig(const double& _L, const BoundaryMethod& _method);
    // Load data
    void load(const YAML::Node& data);
    // Log data
    void log();
  };

  // Diffusion solver outer boundary config
  class DiffOBoundConfig {
  public :
    bool LTFixed;
    double L;
    BoundaryMethod method;
    bool phiZero;
    bool expCorr;
    bool useTJ;
    // Constructors
    DiffOBoundConfig();
    DiffOBoundConfig(const bool& _LTFixed, const double& _L, const BoundaryMethod& _method, const bool& _phiZero, const bool& _expCorr, const bool& _useTJ);
    // Load data
    void load(const YAML::Node& data);
    // Log data
    void log();
  };



  // Diffusion solver opacity config
  class DiffKappaConfig {
  public :
    Processes processes;
    bool floor;
    bool MC;
    // Constructors
    DiffKappaConfig();
    DiffKappaConfig(const Processes& _processes, const bool& _kappaFloor, const bool& _kappaMC);
    DiffKappaConfig(const string& dataFile);
    // Load data
    void load(const YAML::Node& data);
    void load(const string& dataFile);
    // Log data
    void log();
  };

  // Diffusion solver config
  class DiffSolverConfig {
  public :
    bool steadyState;
    bool includeRadio;
    double nPacketGamma;
    double phi;
    bool fluxLimiter;
    bool expansion;
    double Lobs;
    DiffConvConfig convergence;
    DiffTStepConfig timeStep;
    DiffIBoundConfig innerBoundary;
    DiffOBoundConfig outerBoundary;
    DiffKappaConfig kappa; // New!
    bool testSwitch1;
    double nuMin;
    double nuMax;
    double boundFreeCutOff;
    //bool kappaFloor; // Moved to kappa
    //bool kappaMC; // Moved to kappa
    // Constructors
    DiffSolverConfig();
    DiffSolverConfig(const bool& _steadyState, const bool& _includeRadio, const double& _nPacketGamma, const double& _phi, const bool& _fluxLimiter, const bool& _expansion, const double& _Lobs, const DiffConvConfig& _convergence, const DiffTStepConfig& _timeStep, const DiffIBoundConfig& _innerBoundary, const DiffOBoundConfig& _outerBoundary, const DiffKappaConfig& _kappa, const bool& _testSwitch1, const double& _nuMin, const double& _nuMax, const double& _boundFreeCutOff); // Changed!
    DiffSolverConfig(const string& dataFile);
    // Load data
    void load(const YAML::Node& data);
    void load(const string& dataFile);
    // Log data
    void log();
  };

  // Diffusion solver
  class DiffSolver : public StateSolver {
  protected :
    LTESolver lteSolver;
    Grid grid;
    Grid* gridOld;
    double rErrT,dTMax,drhoMax,dt,cellIndexMax,cellIndexMaxMax;
    int m,n;
    vector<double> pKappa;
    vector<double> pKappaOld;
    vector<omp_lock_t> pKappaLock;
    vector<omp_lock_t> pKappaLockOld;
    bool pLog;
    boost::mpi::communicator* mpiCom;
    // Get temperature using Eddington approximation
    double TEdd(const double& _L);
    // Get luminosity using the Eddington approximation
    //double LEdd(const double& _T) const;
    double LEdd(const double& _T);
    // Get luminosity using the modified Eddington approximation
    //double LEddMod(const double& _T) const;
    double LEddMod(const double& _T);
    // Calculate temperature from front luminosity
    //double TDiffFront(const int& i) const;
    double TDiffFront(const int& i);
    // Interpolate r following Falk & Arnett 1977
    double rIpol(const double& r1, const double& r2) const;
    // Interpolate density following Falk & Arnett (1977)
    double rhoIpol(const double& rho1, const double& rho2, const double& dm1, const double& dm2) const;
    // Interpolate differential mass
    double dmIpol(const double& dm1, const double& dm2) const;
    // Interpolate opacity following Christy (1967)
    double kappaIpol(const double& T1, const double& T2, const double& kappa1, const double& kappa2) const;
    // Get radius
    double r(const int& i, const Location& location, const Epoch& epoch=current) const;
    // Set radius
    void r(const double& _r, const int& i, const Location& location, const Epoch& epoch=current);
    // Get veclocity
    double v(const int& i, const Location& location, const Epoch& epoch=current) const;
    // Get density
    double rho(const int& i, const Location& location, const Epoch& epoch=current) const;
    // Set density (center assumed)
    void rho(const double& _rho, const int& i, const Location& location, const Epoch& epoch=current);
    // Get differential mass
    double dm(const int& i, const Location& location, const Epoch& epoch=current) const;
    // Get volume
    double V(const int& i, const Location& location, const Epoch& epoch=current) const;
    // Get temperature
    double T(const int& i, const Location& location, const Epoch& epoch=current) const;
    // Get pressure
    double P(const int& i, const Location& location, const Epoch& epoch=current, double _T=noT) const;
    // Get temperature derivative of gas pressure (center assumed)
    double dPdT(const int& i, const Location& location, const Epoch& epoch=current) const;
    // Get energy (per mass)
    double E(const int& i, const Location& location, const Epoch& epoch=current, double _T=noT) const;
    // Get temperature derivative of energy (per mass)
    double dEdT(const int& i, const Location&, const Epoch& epoch=current) const;
    // Get (Rosseland mean) opactity
    double kappa(const int& i, const Location& location, const Epoch& epoch=current);
    // Get luminosity
    double L(const int& i, const Location& location, const Epoch& epoch=current, double _T=noT);
    // Get temperature derivative of luminosity
    double dLdT(const int& i, const Location& location, const Epoch& epoch=current, double _T=noT);
    // Get flux
    double F(const int& i, const Location& location, const Epoch& epoch=current, double _T=noT);
    // Set flux
    void F(const double& _F, const int& i, const Location& location, const Epoch& epoch=current);
    // Set observed flux
    void FObs(const double& _FObs, const int& i, const Location& location, const Epoch& epoch=current);
    // Get flux limit
    double fluxLimit(const int& i, const Location& location, const Epoch& epoch, double _T=noT);
    // Get adiabatic sound speed
    double vs(const int& i, const Location& location, const Epoch& epoch=current) const;
    // Get radioactive heating
    double eRadio(const int& i, const Location& location, const Epoch& epoch=current) const;
    // Solve for temperature
    void solveTemperature(vector<double>& dT);
    // Update (Rosseland mean) opacity for grid
    void updateKappa(const Epoch& epoch);
    // Update (Rosseland mean) opacity for grid (on a single computer)
    void updateKappaSingle(const Epoch& epoch);
    // Update (Rosseland mean) opacity for grid (on a cluster using MPI)
    void updateKappaCluster(const Epoch& epoch);
    // Update (Rosseland mean) opacity for cell
    void updateKappa(const int& i, const Epoch& epoch);
    // Synchronize (Rosseland mean) opacity for cell
    void syncKappa(const int& i, const Epoch& epoch);
    // Update matter state
    void updateMatterState(const vector<double>& dT);
    // Update matter state - (cell version) - New!
    //void updateMatterState(Cell* cell, const double& dT);
    void updateMatterState(Cell* cell, const double& _T);
    // Update radiation state
    void updateRadiationState(); 
    // Update temperature convergence criteria
    void updateTConvergence(const vector<double>& dT);
    // Temperature converged
    bool TConverged();
    // Evolve thermodynamical state
    void evolveThermo();
    // Evolve dynamical state of grid (assuming homology)
    void evolveDyn();
    // Evolve radioactive state
    void evolveRadio();
    // Evolve state
    void evolve();
    // Set timestep
    void setTimeStep();
    // Refine timestep
    void refineTimeStep();
    // Update convergence criteria
    void updateConvergence();
    // Solution converged
    bool converged();
    // Initiate
    void init(Grid* grid, bool initiateState=true);
    // Initiate radioactive state
    void initRadio(Grid* grid);
  public :
    DiffSolverConfig config;
    // Constructors
    DiffSolver(DiffSolverConfig _config=DiffSolverConfig(), LTESolverConfig _lteConfig=LTESolverConfig(), boost::mpi::communicator* _mpiCom=NULL);
    DiffSolver(const string& dataFile, boost::mpi::communicator* _mpiCom=NULL);
    // Evolve state
    virtual void evolve(Grid* _grid, double& _dt, const int& _cellIndexMax, bool initiate=false, bool initiateState=true);
    // Update state (i.e. assume steady state)
    virtual void update(Grid* _grid, const int& _cellIndexMax, bool initiate=false, bool initiateState=true);
    // Get logging flag
    bool log() const;
    // Set logging flag
    void log(const bool& _log);
    // Get memory allocated - New!
    double mem() const;
    // Get memory allcoated for probability lists
    double memPList() const;
    // Get memory allocated for bound-bound transitions - New!
    double memTransBB() const;
    // Get memory allocated for bound-free transition - New!
    double memTransBF() const;
    // Copy grid - New!
    void copyGrid();
    // Update grid - New!
    void updateGrid();
    // Update cell - New!
    void updateCell(Cell* cell1, Cell* cell2);
  };
}

#endif
