#include "mc.h"
#include "eos.h"

using namespace mc;
using namespace eos;

// Constructors
EOSBasicConfig::EOSBasicConfig() : EIon(false) {}
EOSBasicConfig::EOSBasicConfig(const bool& _EIon) : EIon(_EIon) {}
EOSBasicConfig::EOSBasicConfig(const string& dataFile) { load(dataFile); }

// Load data (from file)
void EOSBasicConfig::load(const string& dataFile) {
  YAML::Node data;

  // Load data from file
  data=YAML::LoadFile(dataFile);

  // If configuration provided
  if (data["EOSBasic"]) {

    // Load ionization energy flag
    if (data["EIon"]) EIon=(data["EIon"].as<string>().compare("yes")==0);
  }

  log();
}

// Log data
void EOSBasicConfig::log() {
  cout<<"EOSBasicConfig::log: EIon: "<<EIon<<"\n";
}

// Constructors
EOSBasic::EOSBasic(EOSBasicConfig _config) : config(_config) { }
EOSBasic::EOSBasic(const string& dataFile) { config.load(dataFile); }

// Get gas pressure
double EOSBasic::PGas(Cell* cell, double T) const {
  double _P;
  if (T==noT) T=cell->T();
  _P=cell->n()*k*T+cell->ne()*k*T;
  return _P;
}

// Get radiation pressure
double EOSBasic::PRad(Cell* cell, double T) const {
  double _P,fE=1.0/3.0;
  if (T==noT) T=cell->T();
  _P=fE*ERad(cell,T)*cell->rho();
  //cout<<"EOSBasic::PRad: P: "<<_P<<", T: "<<T<<"\n";
  return _P;
}

// Get pressure
double EOSBasic::P(Cell* cell, double T) const {
  //cout<<"EOSBasic::P\n";
  double _P;
  _P=PRad(cell,T)+PGas(cell,T);
  //cout<<"EOSBasic::P: P: "<<_P<<", T: "<<T<<"\n";
  return _P;
}

// Get ionization energy
double EOSBasic::EIon(Cell* cell) const {
  double Xi,_EIon=0;
  int i,j;
  for (i=0; i<cell->atom.size(); i++) {
    Xi=0;
    for (j=1; j<cell->atom[i].iState.size(); j++) {
      Xi+=cell->atom[i].iState[j-1].Xi0();
      _EIon+=cell->atom[i].iState[j].n()*Xi/cell->rho();
    }
  }
  return _EIon;
}

// Get gas energy (per mass)
double EOSBasic::EGas(Cell* cell, double T) const {
  double _E;
  _E=(3.0/2.0)*PGas(cell,T)/cell->rho();
  if (config.EIon) _E+=EIon(cell);
  return _E;
}

// Get radiation energy (per mass)
double EOSBasic::ERad(Cell* cell, double T) const {
  double _E;
  if (T==noT) T=cell->T();
  // Calculate energy (per mass) for blackbody radiation
  _E=a*pow(T,4)/cell->rho();
  return _E;
}

// Get energy (per mass)
double EOSBasic::E(Cell* cell, double T) const {
  double _E;
  _E=ERad(cell,T)+EGas(cell,T);
  return _E;
}
