#include "atom.h"

namespace mc {

  // Forward declarations
  class Grid;
  class Cell;
  class Trans;
  class TransBB;
  class TransBF;
  class Atom;
  class IState;
  class BState;
  class PacketEvent;
  //class FrequencyBin;
  class MCExperiment;
  class BStateSelector;
  class TransBBSelector;
  class GridFactory;

  typedef vector<Cell*> CellList;
  typedef vector<Cell*>::iterator CellIterator;

  // Ionization balance method
  enum IMethod { neIteration, newtonRaphson };
  
  // Initial guess method for ionization balance
  enum IGuess { halfIonized, nonCoupled };

  // Radiation field estimator
  enum RadEstimator { rfBB, rfDBB, rfLucy }; // Changed!

  // Line selector
  class LineSelector {
  protected :
    FrequencyBin nu;
    double tauLim;
  public :
    LineSelector();
    LineSelector(const FrequencyBin& _nu, const double& _tauLim);
    double getTauLim();
    virtual bool use(TransBB* trans, const double& lineTau);
  };

  // MC statistics update method
  enum UpdateMethod { urPacketHistory, urOnTheFly };

  // Infinite frequency
  const double nuInf=1e99;

  // No time
  const double noTime=-1;

  // Mean intensity integrand
  class JIntegrand {
  public :
    virtual double operator()(double nu) const {};
    virtual double operator()(FrequencyBinIterator nu, EdgeType edgeType=noEdge) const {};
  };

  // Frequency integrand
  class NuIntegrand {
  public :
    virtual double operator()(double nu) const {};
    virtual double operator()(FrequencyBinIterator nu, EdgeType edgeType=noEdge) const {};
  };

  // Frequency number derivative integrand
  class NuDNIntegrand {
  public :
    virtual double F(double nu) const {};
    virtual double F(FrequencyBinIterator nu, EdgeType edgeType=noEdge) const {}; // New!
    virtual double dFdN(double nu, BState* bState) const {};
    virtual double dFdN(FrequencyBinIterator nu, BState* bState, EdgeType edgeType=noEdge) const {}; // New!
  };

  // Frequency number derivative integrand (ionization state version) - New!
  class NuDNIIntegrand {
  public :
    virtual double F(double nu) const {};
    virtual double F(FrequencyBinIterator nu, EdgeType edgeType=noEdge) const {}; // New!
    virtual double dFdN(double nu, IState* iState) const {};
    virtual double dFdN(FrequencyBinIterator nu, IState* iState, EdgeType edgeType=noEdge) const {}; // New!
  };

  // Matter state solver
  class MStateSolver {
  public :
    // Update matter state
    virtual void update(Cell* cell, bool initiate=false, bool initiateState=true) {};
    // Update matter state using LTE
    virtual void updateLTE(Cell* cell) {};
    // Load data
    virtual void load(const string& dataFile) {};
    // Initiate - New!
    virtual void init(Cell* cell, bool initiateState=true) {};
    // Create copy
    virtual MStateSolver* copy() {};
  };

  // Equation of state 
  class EOS {
  public :
    virtual double P(Cell* cell, double T=noT) const { cout<<"EOS::P\n"; };
    virtual double E(Cell* cell, double T=noT) const { cout<<"EOS::E\n"; };
  };

  // (Bound-bound) Transition identifier
  class TransBBIdentifier {
  public :
    string iStateName;
    int upperStateIndex;
    int lowerStateIndex;
  };

  // (Bound-free) Transition identifier
  class TransBFIdentifier {
  public :
    string lowerStateName;
    string upperStateName;
    int upperStateIndex;
    int lowerStateIndex;
  };

  // Use on-the-spot approximation
  class OnTheSpot {
  private :
    vector<string> atomName;
  public :
    OnTheSpot();
    OnTheSpot(const string& specifier);
    bool operator() (Atom& atom) const;
  };

  // Use on-the-spot approximation (new)
  class OnTheSpotNew {
  private :
    vector<TransBFIdentifier> transID;
  public :
    OnTheSpotNew();
    OnTheSpotNew(const string& specifier);
    bool operator() (TransBF* trans) const;
  };

  // Use free-free on-the-spot approximation
  class FFOnTheSpot {
  public :
    bool apply;
    double tauLim;
    FFOnTheSpot();
    FFOnTheSpot(const bool& _apply, const double& _tauLim);
  };

  // Processes class
  class Processes : public vector<unsigned int> {
  public :
    // Constructors
    Processes();
    Processes(unsigned int type, unsigned int subType=pNone);
    // Get type from name
    ProcessType type(const string& name);
    // Get name from type
    string name(const ProcessType& type);
    // Load data
    void load(const YAML::Node& data);
    // Log data
    void log();
    // Contains process
    bool operator () (unsigned int type, unsigned int subtype=pAny) const;
  };

  // Non-thermal transition for MPI comminucation - New!
  class NTTransMPI {
  private :
    // MPI serialization
    friend class boost::serialization::access;
    template<class Archive> 
      void serialize (Archive & ar, const unsigned int version) 
      {
	ar & index;
	ar & CNU;
      }
  public:
    int index;
    double CNU;
  };

  // Non-thermal bound-bound transition list for MPI comminucation - New!
  class NTTransBBListMPI {
  private :
    // MPI serialization
    friend class boost::serialization::access;
    template<class Archive> 
      void serialize (Archive & ar, const unsigned int version) 
      {
	ar & ntTransList;
      }
  public :
    vector<NTTransMPI> ntTransList;
    void pack(vector<TransBB*>& transBB);
    void unpack(vector<TransBB*>& transBB) const;
  };

  // Non-thermal bound-free transition list for MPI comminucation - New!
  class NTTransBFListMPI {
  private :
    // MPI serialization
    friend class boost::serialization::access;
    template<class Archive> 
      void serialize (Archive & ar, const unsigned int version) 
      {
	ar & ntTransList;
      }
  public :
    vector<NTTransMPI> ntTransList;
    void pack(vector<TransBF*>& transBF);
    void unpack(vector<TransBF*>& transBF) const;
  };

  // Non-thermal (matter state) for MPI communication - New!
  class NTStateMPI {
  private :
    // MPI serialization
    friend class boost::serialization::access;
    template<class Archive> 
      void serialize (Archive & ar, const unsigned int version) 
      {
	ar & nonThermalHeating;
	ar & nonThermalIonization;
	ar & nonThermalExcitation;
	ar & ntTransBB;
	ar & ntTransBF;
      }
  public :
    double nonThermalHeating;
    double nonThermalIonization;
    double nonThermalExcitation;
    vector<vector<vector<NTTransBBListMPI> > > ntTransBB;
    vector<vector<vector<NTTransBFListMPI> > > ntTransBF;
    void pack(Cell* cell);
    void unpack(Cell* cell) const;
  };

  // Matter state for MPI communication
  class MStateMPI {
  private :
    // MPI serialization
    friend class boost::serialization::access;
    template<class Archive> 
      void serialize (Archive & ar, const unsigned int version) 
      {
	ar & T;
	ar & rho;
	ar & ne;
	ar & ntState; // New!
	ar & XIsotope;
	ar & XIon;
	ar & XBound;
      }
  public :
    double T;
    double rho;
    double ne;
    NTStateMPI ntState; // New!
    vector<vector<double> > XIsotope;
    vector<vector<double> > XIon;
    vector<vector<vector<double> > > XBound;
    void pack(Cell* cell);
    void unpack(Cell* cell) const;
  };

  // MC statistics for MPI communication
  class MCStatMPI {
  private :
    // MPI serialization
    friend class boost::serialization::access;
    template<class Archive> 
      void serialize (Archive & ar, const unsigned int version) 
      {
	ar & Jnu;
	ar & Hnu;
	ar & Knu;
	ar & Inu;
	ar & etaNu;
	ar & kappaJNu; // New!
	ar & kappaBFJNu; // New!
	ar & kappaFFJNu; // New!
	ar & kappaBBJNu; // New!
	ar & kappaESJNu; // New!
	ar & kappaHNu; // New!
	ar & kappaBFHNu; // New!
	ar & kappaFFHNu; // New!
	ar & kappaBBHNu; // New!
	ar & kappaESHNu; // New!
	/*ar & nAbsNu; // New!
	ar & nAbsBFNu; // New!
	ar & nAbsFFNu; // New!
	ar & nAbsBBNu; // New!
	ar & nAbsESNu; // New!
	ar & lNu; // New!*/
	ar & nuMeanNu;
	ar & JLine;
	ar & eRadio;
	ar & nJ;
	ar & nJLine;
	ar & neRadio; // New!
      }
  public :
    vector<double> Jnu;
    vector<double> Hnu;
    vector<double> Knu;
    vector<vector<double> > Inu;
    vector<double> etaNu;
    vector<double> kappaJNu; // New!
    vector<double> kappaBFJNu; // New!
    vector<double> kappaFFJNu; // New!
    vector<double> kappaBBJNu; // New!
    vector<double> kappaESJNu; // New!
    vector<double> kappaHNu; // New!
    vector<double> kappaBFHNu; // New!
    vector<double> kappaFFHNu; // New!
    vector<double> kappaBBHNu; // New!
    vector<double> kappaESHNu; // New!
    /*vector<uint64_t> nAbsNu; // New!
    vector<uint64_t> nAbsBFNu; // New!
    vector<uint64_t> nAbsFFNu; // New!
    vector<uint64_t> nAbsBBNu; // New!
    vector<uint64_t> nAbsESNu; // New!
    vector<double> lNu; // New!*/
    vector<double> nuMeanNu;
    vector<double> JLine;
    double eRadio;
    vector<uint64_t> nJ;
    vector<uint64_t> nJLine;
    uint64_t neRadio; // New!
    // Sum MC statistics
    void operator+=(const MCStatMPI& mCStat);
    // Pack MC statistics
    void pack(Cell* cell);
    // Unpack MC statistics
    void unpack(Cell* cell) const;
  };

  // G-packet MC statistics for MPI communication - New!
  class GMCStatMPI {
  private :
    // MPI serialization
    friend class boost::serialization::access;
    template<class Archive> 
      void serialize (Archive & ar, const unsigned int version) 
      {
	ar & eRadio;
	ar & neRadio;
      }
  public :
    double eRadio;
    uint64_t neRadio;
    // Sum g-packet MC statistics
    void operator+=(const GMCStatMPI& gMCStat);
    // Pack g-packet MC statistics
    void pack(Cell* cell);
    // Unpack g-packet MC statistics
    void unpack(Cell* cell) const;
  };

  // Bound-free opacity table for MPI communication - New!
  class BoundFreeAbsKappaMPI {
  private :
    // MPI serialization
    friend class boost::serialization::access;
    template<class Archive> 
      void serialize (Archive & ar, const unsigned int version) 
      {
	ar & boundFreeAbsKappaTable;
	ar & boundFreeTableUpdate;
      }
  public :
    NuInterpolatorMPI boundFreeAbsKappaTable;
    bool boundFreeTableUpdate;
    // Pack bound-free opacity table
    void pack(Cell* cell);
    // Unpack bound-free opacity table
    void unpack(Cell* cell);
  };

  // Bound-free frequency cut-off for MPI communication - New!
  class BoundFreeCutOffMPI {
  private :
    // MPI serialization
    friend class boost::serialization::access;
    template<class Archive> 
      void serialize (Archive & ar, const unsigned int version) 
      {
	ar & minorEdge;
	ar & minorEdgeOld;
	ar & nuCutOffTransBF;
	ar & boundFreeCutOffUpdate; // New!
      }
  public :
    vector<double> minorEdge;
    vector<double> minorEdgeOld;
    vector<vector<vector<vector<double> > > > nuCutOffTransBF;
    bool boundFreeCutOffUpdate; // New!
    void pack(Cell* cell);
    void unpack(Cell* cell);
  };

  // Bound-state ID for MPI comminucation - New!
  class BStateIDMPI {
  private :
    // MPI serialization
    friend class boost::serialization::access;
    template<class Archive> 
      void serialize (Archive & ar, const unsigned int version) 
      {
	//ar & index;
	ar & atomIndex;
	ar & iStateIndex;
	ar & bStateIndex;
      }
  public:
    //unsigned int index;
    unsigned char atomIndex;
    unsigned char iStateIndex;
    short bStateIndex;
    void pack(BState*& bState, Cell* cell);
    void unpack(BState*& bState, Cell* cell);
  };

  // Transition ID for MPI comminucation - New!
  class TransIDMPI {
  private :
    // MPI serialization
    friend class boost::serialization::access;
    template<class Archive> 
      void serialize (Archive & ar, const unsigned int version) 
      {
	ar & bState;
	//ar & flags;
	ar & index;
	ar & upward;
	ar & boundFree;
      }
  public:
    BStateIDMPI bState;
    //unsigned char flags;
    unsigned short index;
    bool upward;
    bool boundFree;
    void pack(Trans*& trans, unsigned short int& _index, bool& _upward, Cell* cell);
    void unpack(Trans*& trans, unsigned short int& _index, bool& _upward, Cell* cell);
  };

  // Markov transition probability for MPI comminucation - New!
  class MarkovPMPI {
  private :
    // MPI serialization
    friend class boost::serialization::access;
    template<class Archive> 
      void serialize (Archive & ar, const unsigned int version) 
      {
	ar & P;
	ar & nextState;
      }
  public:
    float P;
    BStateIDMPI nextState;
  };

  // Transition probability for MPI comminucation - New!
  class TransPMPI {
  private :
    // MPI serialization
    friend class boost::serialization::access;
    template<class Archive> 
      void serialize (Archive & ar, const unsigned int version) 
      {
	ar & P;
	ar & trans;
	ar & type;
      }
  public:
    float P;
    TransIDMPI trans;
    TransType type;
  };

  // Markov transition probability list for MPI comminucation - New!
  class MarkovPListMPI {
  private :
    // MPI serialization
    friend class boost::serialization::access;
    template<class Archive> 
      void serialize (Archive & ar, const unsigned int version) 
      {
	ar & markovPListMPI;
      }
  public :
    vector<MarkovPMPI> markovPListMPI;
    void pack(vector<MarkovP>& markovPList, Cell* cell);
    void unpack(vector<MarkovP>& markovPList, Cell* cell);
  };

  // Transition probability list for MPI comminucation - New!
  class TransPListMPI {
  private :
    // MPI serialization
    friend class boost::serialization::access;
    template<class Archive> 
      void serialize (Archive & ar, const unsigned int version) 
      {
	ar & transPListMPI;
      }
  public :
    vector<TransPMPI> transPListMPI;
    void pack(vector<TransP>& transPList, Cell* cell);
    void unpack(vector<TransP>& transPList, Cell* cell);
  };

  // Transition emission probability table for MPI comminucation - New!
  class EtaPTableTransMPI {
  private :
    // MPI serialization
    friend class boost::serialization::access;
    template<class Archive> 
      void serialize (Archive & ar, const unsigned int version) 
      {
	ar & index;
	ar & etaPTable;
      }
  public:
    int index;
    vector<double> etaPTable;
  };

  // Transition emission probability list for MPI comminucation - New!
  class EtaPTableTransListMPI {
  private :
    // MPI serialization
    friend class boost::serialization::access;
    template<class Archive> 
      void serialize (Archive & ar, const unsigned int version) 
      {
	ar & etaPTableTransList;
      }
  public :
    vector<EtaPTableTransMPI> etaPTableTransList;
    void pack(vector<TransBF*>& transBF);
    void unpack(vector<TransBF*>& transBF);
  };

  // Transition list for MPI comminucation - New!
  /*class TransListMPI {
  private :
    // MPI serialization
    friend class boost::serialization::access;
    template<class Archive> 
      void serialize (Archive & ar, const unsigned int version) 
      {
	ar & transList;
      }
  public :
    vector<TransIDMPI> transList;
    void pack(vector<Trans*>& trans);
    void unpack(vector<Trans*>& trans) const;
    };
  */

  // Bound-bound transition boost for MPI comminucation - New!
  class BoostTransBBMPI {
  private :
    // MPI serialization
    friend class boost::serialization::access;
    template<class Archive> 
      void serialize (Archive & ar, const unsigned int version) 
      {
	ar & eta1Boost;
	ar & eta2Boost;
	ar & eta2P;
      }
  public :
    double eta1Boost;
    double eta2Boost;
    vector<double> eta2P;
    void pack(TransBB* transBB);
    void unpack(TransBB* transBB);
  };

    // Bound-free transition boost for MPI comminucation - New!
  class BoostTransBFMPI {
  private :
    // MPI serialization
    friend class boost::serialization::access;
    template<class Archive> 
      void serialize (Archive & ar, const unsigned int version) 
      {
	ar & kappaBoost;
	ar & etaBoost;
	ar & etaP;
      }
  public :
    double kappaBoost;
    double etaBoost;
    vector<double> etaP;
    void pack(TransBF* transBF);
    void unpack(TransBF* transBF);
  };

  // Bound-state boost for MPI comminucation - New!
  class BoostBStateMPI {
  private :
    // MPI serialization
    friend class boost::serialization::access;
    template<class Archive> 
      void serialize (Archive & ar, const unsigned int version) 
      {
	ar & transPRBoost;
	ar & transPCBoost;
	ar & transPBoost;
	ar & markovPBoost;
      }
  public :
    double transPRBoost;
    double transPCBoost;
    double transPBoost;
    double markovPBoost;
    void pack(BState* bState);
    void unpack(BState* bState);
  };

  // Derived MC quantities for MPI communication - New!
  class BoostMPI {
  private :
    // MPI serialization
    friend class boost::serialization::access;
    template<class Archive> 
      void serialize (Archive & ar, const unsigned int version) 
      {
	ar & boost;
	ar & freeFreeEtaBoost;
	ar & freeFreeKappaBoost;
	ar & boundFreeKappaBoostTable;
	ar & comptonKappaBoost;
	ar & boundFreeCoolingBoost;
	ar & freeFreeCoolingBoost;
	ar & collCoolingBoost;
	ar & radCoolingBoost;
	ar & coolingBoost;
	ar & nonThermalHeatingBoost;
	ar & nonThermalExcIonBoost;
	ar & nonThermalBoost;
	ar & markovPThermBoost;
	ar & markovPCTBoost;
	ar & freeFreeEtaP;
	ar & boostBState;
	ar & boostTransBB;
	ar & boostTransBF;
      }
  public :
    vector<double> boost;
    double freeFreeEtaBoost;
    double freeFreeKappaBoost;
    Interpolator1D boundFreeKappaBoostTable;
    double comptonKappaBoost;
    double boundFreeCoolingBoost;
    double freeFreeCoolingBoost;
    double collCoolingBoost;
    double radCoolingBoost;
    double coolingBoost;
    double nonThermalHeatingBoost;
    double nonThermalExcIonBoost;
    double nonThermalBoost;
    double markovPThermBoost;
    vector<double> markovPCTBoost;
    vector<double> freeFreeEtaP;
    vector<vector<vector<BoostBStateMPI> > > boostBState;
    vector<vector<vector<vector<BoostTransBBMPI> > > > boostTransBB;
    vector<vector<vector<vector<BoostTransBFMPI> > > > boostTransBF;
    void pack(Cell* cell);
    void unpack(Cell* cell);
  };

  // Derived MC quantities for MPI communication - New!
  class MCDerivedMPI {
  private :
    // MPI serialization
    friend class boost::serialization::access;
    template<class Archive> 
      void serialize (Archive & ar, const unsigned int version) 
      {
	ar & boost;
	ar & markovPListTherm;
	ar & markovPListCT;
	ar & collExcitePList;
	//ar & CTExcitePList;
	ar & nonThermalExcitePList;
	ar & markovPList;
	ar & transPList;
	ar & PRTot;
	ar & PCTot;
	ar & etaPTableTransBF;
      }
  public :
    BoostMPI boost;
    MarkovPListMPI markovPListTherm;
    vector<MarkovPListMPI> markovPListCT;
    TransPListMPI collExcitePList;
    //TransPListMPI CTExcitePList;
    TransPListMPI nonThermalExcitePList;
    vector<vector<vector<MarkovPListMPI> > > markovPList;
    vector<vector<vector<TransPListMPI> > > transPList;
    vector<vector<vector<double> > > PRTot;
    vector<vector<vector<double> > > PCTot;
    vector<vector<vector<EtaPTableTransListMPI> > > etaPTableTransBF;
    void pack(Cell* cell);
    void unpack(Cell* cell);
  };

  // MC statistics update info
  class MCUpdateInfo {
  public :
    int nuBin;
    int muBin;
    int nuBinLine;
    MCUpdateInfo();
  };

  // Blackbody temperature type
  enum BBTType { bbTCur, bbTPrev, bbTJ, bbTJBB };

  // Cell config
  class CellConfig {
  public :
    // Steady state
    bool steadyState;
    // Ionization balance method
    IMethod iMethod;
    // Line selector
    LineSelector* lineSelector;
    UpdateMethod updateMethod;
    bool boundFreeTable;
    double boundFreeTableMin;
    double boundFreeTableMax;
    double boundFreeTableNBin;
    double boundFreeTableEdgeWidth;
    bool boundFreeSTable;
    int boundFreeSTableNBin;
    double boundFreeSTableEdgeWidth;
    RadEstimator radEstimator;
    RadEstimator radEstimatorBB;
    bool radEstimatorExp;
    bool transTable;
    bool markovTable;
    bool markovTableExt;
    bool markovTableExtSplit;
    double markovTableLim;
    double markovTableThermLim;
    double markovTableCTLim; // New!
    bool collExciteTable;
    double collExciteTableLim;
    bool nonThermalExciteTable;
    bool freeFreeMacro;
    bool boundFreeMacro;
    bool collMacro;
    bool nonThermalMacro;
    bool twoPhotonMacro; // New!
    bool chargeTransferMacro; // New!
    OnTheSpot onTheSpot;
    OnTheSpotNew onTheSpotNew;
    FFOnTheSpot ffOnTheSpot;
    bool continuumBB;
    bool boundFreeBB;
    bool freeFreeBB;
    double boundFreeBBTauLim;
    double continuumBBTauLim;
    BBTType continuumBBT;
    bool exportEta;
    bool transformRay;
    bool limitFreeFreeCRange;
    bool JLineFallback;
    double JLineFallbackFactor;
    int JLineFallbackCount;
    bool JLineFallbackDBB;
    bool trackI;
    bool trackHK;
    double boundFreeCutOff;
    double boundFreeCutOffRoss;
    bool boost;
    bool alfaCorrect; // New!
    bool alfaCorrectEmit; // New!
    bool alfaSimpleEmit; // New!
    bool alfaSPH; // New!
    bool forceDBalance; // New!
    bool forceDBalanceOSP; // New!
    bool chargeTransferEmit; // New!
    bool etaTable; // New!
    bool exportSBF; // New!
    bool boundFreeTableOSPOnly; // New! - Testing!
    double nuMax; // New!
    double nuMin; // New!
    int nNuBin; // New!
    int nMuBin; // New!
    bool LTEGroupMacro; // New!
    bool exportKappaMC2; // New!
    // Constructors
    CellConfig();
    CellConfig(const bool& _steadyState, const IMethod& _iMethod, LineSelector* lineSelector, const UpdateMethod& _updateMethod, const bool& _boundFreeTable, const double& _boundFreeTableMin, const double& _boundFreeTableMax, const int& _boundFreeTableNBin, const bool& _boundFreeSTable, const int& _boundFreeSTableNBin, const RadEstimator& _radEstimator, const RadEstimator& _radEstimatorBB, const bool& _radEstimatorExp, const bool& _transTable, const bool& _markovTable, const bool& _markovTableExt, const bool& _markovTableExtSplit, const double& _markovTableLim, const double& _markovTableThermLim, const double& _markovTableCTLim, const bool& _collExciteTable, const double& _collExciteTableLim, const bool& _nonThermalExciteTable, const bool& _freeFreeMacro, const bool& _boundFreeMacro, const bool& _collMacro, const bool& _nonThermalMacro, const bool& _twoPhotonMacro, const bool& _chargeTransferMacro, const OnTheSpot& _onTheSpot, const OnTheSpotNew& _onTheSpotNew, const FFOnTheSpot& _ffOnTheSpot, const bool& _continuumBB, const bool& _boundFreeBB, const double& _continuumBBTauLim, const double& _boundFreeBBTauLim, const BBTType& _continuumBBT, const bool& _exportEta, const bool& _transformRay, const bool& _limitFreeFreeCRange, const bool& _JLineFallback, const double& _JLineFallbackFactor, const int& _JLineFallbackCount, const bool& _JLineFallbackDBB, const bool& _trackI, const bool& _trackHK, const double& _boundFreeCutOff, const double& _boundFreeCutOffRoss, const bool& _boost, const bool& _alfaCorrect, const bool& _alfaCorrectEmit, const bool& _alfaSimpleEmit, const bool& _alfaSPH, const bool& _forceDBalance, const bool& _forceDBalanceOSP, const bool& _chargeTransferEmit, const bool& _etaTable, const bool& _exportSBF, const double& _nuMin, const double& _nuMax, const int& _nNuBin, const int& _nMuBin, const bool& _LTEGroupMacro, const bool& _exportKappaMC2); // Changed!
  };

  // Cell location
  enum CellLocation { cellBorder, cellInterior };

  // Border type
  enum BorderType { btCellBorder, btCellBorderInner, btCellBorderOuter, btSubCellBorder, btTimeBorder };

  // Check if cell border
  bool isCellBorder(const BorderType& borderType);

  // Cell type - Moved!
  enum CellType { cellNormal, cellContainer, cellSub };  

  // Load cell type - Moved!
  CellType loadCellType(const YAML::Node& data);

  // Save cell type - Moved!
  string saveCellType(const CellType& cellType);

  // Log cell type
  string logCellType(const CellType& cellType);

  // Cell base interface
  class CellBase {
  public :
    // Constructor
    CellBase() {};
    // Destructor
    virtual ~CellBase() {};
    // Get grid back reference pointer
    virtual Grid* grid() const {};
    // Get cell type - New!
    virtual int type() const {};
    // Get cell index
    virtual int index() const {};
    // Get sub cell index
    virtual int subIndex() const {};
    // Get minimum time
    virtual double tMin() const {};
    // Set minimum time
    virtual void tMin(const double& _tMin) {};
    // Get maximum time
    virtual double tMax() const {};
    // Set maximum time
    virtual void tMax(const double& _tMax) {}; 
    // Get expansion factor
    virtual double fExp() const {};
    // Get rho
    virtual double rho() const {};
    // Set rho
    virtual void rho(const double& _rho) {};
    // Get number density
    virtual double n() const {};
    // Get T
    virtual double T() const {};
    // Get energy (per mass)
    virtual double E(double _T=noT) {};
    // Get pressure
    virtual double P(double _T=noT) {};
    // Get electron number density
    virtual double ne() const {};
    // Get electron fraction
    virtual double xe() const {};
    // Get total mass
    virtual double M() {};
    // Get total mass of atom
    virtual double M(const double& Z) {};
    // Get total mass of isotope
    virtual double M(const double& Z, const double& N) {};
    // Get velocity
    virtual double v(const Vector3D& position, double _t=noTime) const {};
    // Get velocity gradient
    //virtual double dvdr() const {};
    virtual double dvdr(double _t=noTime) const {};
    // Get radiation temperature 
    virtual double TR() const {};
    // Get dilution factor
    virtual double W() const {};
    // Get radioacive energy deposition
    //virtual double LRadio() const {}; // Changed!
    // In cell - Changed!
    virtual bool inCell(const Vector3D& position, double t=noTime, bool virtualCell=false) {};
    // Calculate border distance (to be implemented by derived class) - Changed!
    virtual double borderDistance(const Vector3D& position, const Vector3D& direction, BorderType& borderType, double t=noTime, bool virtualCell=false, bool log=false) {};
    // Get volume
    virtual double V(double t=noTime) {};
    // Get random direction
    virtual Vector3D randomDirection(const CellLocation& cellLocation) {};
    // Get random position - Changed!
    virtual Vector3D randomPosition(const CellLocation& cellLocation, double _t=noTime, bool virtualCell=false) {};
    // Get typical size of cell
    virtual double size() {};
    // Select cell
    virtual CellBase* selectCell(const Vector3D& position, const Vector3D& direction, double t=noTime) {};
    // Select cell neighbour
    virtual CellBase* selectCellNeighbour(const Vector3D& entryPosition, const Vector3D& entryDirection, const double& entryTime) {};
    // Select next cell
    virtual CellBase* selectNextCell(const Vector3D& position, const Vector3D& direction, const double& t) {};
    // Select next resonant line
    virtual LineIterator selectNextLine(const double& nu, const Vector3D& position, const Vector3D& direction, const double& t) const {};
    // Check if end of line list
    virtual bool endOfLineList(const LineIterator& lineIterator) const {};
    // Calculate distance to next resonant line (knowing the line)
    virtual double lineDistance(const LineIterator& lineIteraor, const double& nu, const Vector3D& position, const Vector3D& direction, const double& t) const {};
    // Calculate the distance to next resonant line (knowing the line and the frequency)
    virtual double lineDistance(const LineIterator& lineIterator, const double& nuCMF, double _t=noTime) {};
    // Select next ionization edge
    virtual EdgeIterator selectNextEdge(const double& nu, const Vector3D& position, const Vector3D& direction, const double& t) const {};
    // Check if end of ionization edge list
    virtual bool endOfEdgeList(const EdgeIterator& edgeIterator) const {};
    // Calculate distance to next ionization edge (knowing the edge)
    virtual double edgeDistance(const EdgeIterator& edgeIteraor, const double& nu, const Vector3D& position, const Vector3D& direction, const double& t) const {};
    // Calculate the distance to next ionization edge (knowing the edge and the frequency)
    //virtual double edgeDistance(const EdgeIterator& edgeIterator, const double& nuCMF) {};
    virtual double edgeDistance(const EdgeIterator& edgeIterator, const double& nuCMF, double _t=noTime) {};
    // Select next boost bin - New!
    virtual FrequencyBinIterator selectNextBoostBin(const double& nu, const Vector3D& position, const Vector3D& direction, const double& t) const {};
    // Check if end of boost bin list - New!
    virtual bool endOfBoostBinList(const FrequencyBinIterator& boostBinIterator) const {};
    // Get end of boost bin list - New!
    virtual FrequencyBinIterator endOfBoostBinList() {};
    // Check if beginning of boost bin list - New!
    virtual bool beginOfBoostBinList(const FrequencyBinIterator& boostBinIterator) const {};
    // Calculate distance to next boost bin (knowing the boost bin) - New!
    virtual double boostBinDistance(const FrequencyBinIterator& boostBinIteraor, const double& nu, const Vector3D& position, const Vector3D& direction, const double& t) const {};
    // Calculate the distance to next boost bin (knowing the boost bin and the frequency) - New!
    virtual double boostBinDistance(const FrequencyBinIterator& boostBinIterator, const double& nuCMF, double _t=noTime) {};
    // Calculate the opacity for electron scattering
    virtual double electronScatterKappa() const {};
    // Calculate the opacity for compton scattering (not supported yet)
    virtual double comptonScatterKappa(const double& nu) const {};
    // Get the opacity for grey scattering - New!
    virtual double greyScatterKappa() const {};
    // Set the opacity for grey scattering - New!
    virtual void greyScatterKappa(const double _greyScatterKappa) {};
    // Get the opacity for dust absorption - New!
    virtual double dustAbsKappa() const {};
    // Set the opacity for dust absorption - New!
    virtual void dustAbsKappa(const double _dustAbsKappa) {};
    // Calculate the opacity for free-free absorption
    virtual double freeFreeAbsKappa(const double& nu) const {};
    // Calculate the emissivity for free-free emission
    virtual double freeFreeEta(const double& nu) const {};
    // Calculate free-free cooling rate
    virtual double freeFreeCooling(double _T=noT) {};
    // Draw free-free emission frequency
    virtual double freeFreeNu() const {};
    // Calculate the opacity for bound-free absorption
    virtual double boundFreeAbsKappa(const double& nu, EdgeType edgeType=noEdge) {};
    // Calculate the emissivity for bound-free emission
    virtual double boundFreeEta(const double& nu, bool cutOff=true) {};
    // Calculate bound-free cooling rate
    virtual double boundFreeCooling(double _T=noT) const {};
    // Calculate (spontaneous) bound-free cooling rate
    virtual double boundFreeCoolingSP(double _T=noT) {};
    // Draw bound-free emission frequency
    virtual double boundFreeNu() {};
    // Draw bound-free (downward) transition
    virtual Trans* boundFreeDeExcite() {};
    // Draw bound-free (upward) transition
    virtual Trans* boundFreeExcite(const double& nu) {};
    // Calculate the opacity for compton absorption
    virtual double comptonAbsKappa(const double& comptonAbsSigma) const {};
    // Calculate optical depth for next resonant line (knowing the line)
    // Pointer version
    virtual double lineTau(TransBB* trans) const {};
    // Calculate optical depth for next resonant line (knowing the line)
    // Iterator version
    virtual double lineTau(const LineIterator& lineIterator) const {};
    // Calculate escape fraction for next resonant line (knowing the line)
    // Pointer version
    virtual double lineBeta(TransBB* trans) const {};
    // Calculate escape fraction for next resonant line (knowing the line)
    // Iterator version
    virtual double lineBeta(const LineIterator& lineIterator) const {};
    // Calculate collisional cooling rate
    virtual double collCooling(double _T=noT) {};
    // Get non-thermal collisional heating rate
    virtual double nonThermalHeating() const {}; // Changed!
    // Set non-thermal collisional heating rate
    virtual void nonThermalHeating(const double& _nonThermalHeating) {};
    // Get non-thermal collisional excitation rate
    virtual double nonThermalExcitation() const {};
    // Set non-thermal collisional excitation rate
    virtual void nonThermalExcitation(const double& _nonThermalExcitation) {};
    // Get non-thermal collisional ionization rate
    virtual double nonThermalIonization() const {};
    // Set non-thermal collisional ionization rate
    virtual void nonThermalIonization(const double& _nonThermalIonization) {};
    // Calculate cooling rate
    virtual double cooling(Processes& processes, double _T=noT) {};
    // Calculate heating rate
    //virtual double heating(Processes& processes, double _T=noT) const {};
    virtual double heating(Processes& processes, double _T=noT) {}; // Changed!
    // Calculate the heating-cooling difference
    virtual double G(Processes& processes, double _T=noT) {};
    // Calculate derivative of the heating-cooling difference
    virtual double dGdT(Processes& processes) {};
    // Total continuum opacity (as a function of frequency)
    virtual double contKappaNu(const double& nu, const Processes& processes) {}; // Changed!
    // Rosseland mean (continuum) opacity
    virtual double rossContKappa(const Processes& processes, double nuMin=-1, double nuMax=-1, double tol=1e-2, bool containerMC=false) {}; // Changed! - Testing!
    // Energy mean (continuum) opacity
    virtual double energyContKappa(const Processes& processes) {}; // Changed!
    // Draw collisional excitation transition
    virtual Trans* collExcite() {};
    // Draw charge-transfer excitation transition
    virtual Trans* CTExcite() {};
    // Draw non-thermal excitation transition
    virtual Trans* nonThermalExcite() {};
    // Get estimator for line mean intensity
    virtual double Jb(const double& nu) {}; // Changed!
    // Integrate quantity over the mean intensity
    virtual double JIntegrate(const JIntegrand& integrand, double nuMin, double nuMax, int& nIter, const double& relTol, double absTol=0.0, bool includeOSP=true, bool useIterator=false) {}; // Changed!
    // Update the radiation state (could be overloaded by derived class)
    virtual void updateRadiationState(const FrequencyBinList& nu) {};
    // Reset MC statistics
    virtual void resetMCStatistics(const FrequencyBinList& _nu) {};
    // Update MC statistics (using packet history)
    virtual void updateMCStatistics() {};
    // Update MC statistics (using a single ray)
    virtual void updateMCStatistics(const PacketEvent& startPoint, const PacketEvent& endPoint, MCUpdateInfo& updateInfo) {};
    // Update g-packet MC statistics (using packet history)
    virtual void updateGMCStatistics() {};
    // Update g-packet MC statistics (using a single ray)
    virtual void updateGMCStatistics(const PacketEvent& startPoint, const PacketEvent& endPoint) {};
    // Update the matter state
    //virtual void updateMatterState(bool initiate=false, bool initiateState=true) {};
    virtual void updateMatterState(bool initiate=false, bool initiateState=true, bool updateMCDerived=true) {}; // Changed!
    // Export statistics
    virtual void exportStat(YAML::Emitter& data, const double& tauLim=0.0, string prefix="", bool close=true) {};
    // Export statistics (could be overloaded by derived class)
    virtual void exportStatistics(ofstream& file, const double& tauLimit=1.0) {};
    // Draw cooling MC channel - New!
    virtual MCChannel coolingChannel(bool deActivate) {};
    // Draw non-thermal MC channel - New!
    virtual MCChannel nonThermalChannel() {};
    // Get boost factor - New!
    virtual double boostFactor(const double& _nu) {};
    // Get free-free opacity boost factor - New!
    virtual double freeFreeKappaBoost() {};
    // Get bound-free opacity boost factor - New!
    virtual double boundFreeKappaBoost(const double& _nu, EdgeType edgeType=noEdge) {};
    // Get compton opacity boost factor - New!
    virtual double comptonKappaBoost() {};
    // Get memory allocated
    virtual double mem() const {};
  };

  // Virtual cell
  class CellVirtual : public CellBase {
  protected :
    // Cell container
    CellBase* pCell;
    // Sub cell
    CellBase* pSubCell;
    // Position offset
    Vector3D positionOffset;
    // Theta offset
    double thetaOffset;
    // Phi offset
    double phiOffset;
  public :
    // Constructor
    CellVirtual();
    // Constructor
    CellVirtual(CellBase* cell, CellBase* subCell, const Vector3D& position, const Vector3D& direction, const CellLocation& cellLocation, double t=noTime);
    // Constructor - New!
    CellVirtual(CellBase* _cell, CellBase* _subCell, const Vector3D& _positionOffset);
    // Copy constructor
    CellVirtual(const CellVirtual& cellVirtual);
    // Asigment operator
    CellVirtual& operator=(const CellVirtual& cellVirtual);
    // Destructor
    virtual ~CellVirtual();
    // Get grid back reference pointer
    virtual Grid* grid() const { return pSubCell->grid(); };
    // Get cell type
    virtual int type() const { return pSubCell->type(); };
    // Get cell index
    virtual int index() const;
    // Get sub cell index
    virtual int subIndex() const;
    // Get minimum time
    virtual double tMin() const { return pSubCell->tMin(); };
    // Set minimum time
    virtual void tMin(const double& _tMin) { pSubCell->tMin(_tMin); };
    // Get maximum time
    virtual double tMax() const { return pSubCell->tMax(); };
    // Set maximum time
    virtual void tMax(const double& _tMax) { pSubCell->tMax(_tMax); };
    // Get expansion factor
    virtual double fExp() const { return pSubCell->fExp(); };
    // Get rho
    virtual double rho() const { return pSubCell->rho(); };
    // Set rho
    virtual void rho(const double& _rho) { return pSubCell->rho(_rho); }
    // Get number density
    virtual double n() const { return pSubCell->n(); };
    // Get T
    virtual double T() const { return pSubCell->T(); };
    // Get energy (per mass)
    virtual double E(double _T=noT) { return pSubCell->E(_T); };
    // Get pressure
    virtual double P(double _T=noT) { return pSubCell->P(_T); };
    // Get electron number density
    virtual double ne() const { return pSubCell->ne(); };
    // Get electron fraction
    virtual double xe() const {return pSubCell->xe(); };
    // Get total mass
    virtual double M() { return pSubCell->M(); };
    // Get total mass of atom
    virtual double M(const double& Z) { return pSubCell->M(Z); };
    // Get total mass of isotope
    virtual double M(const double& Z, const double& N) { return pSubCell->M(Z,N); };
    // Get velocity
    virtual double v(const Vector3D& position, double _t=noTime) const { return pSubCell->v(position,_t); };
    // Get velocity gradient
    virtual double dvdr(double _t=noTime) const { return pSubCell->dvdr(_t); };
    // Get radiation temperature 
    virtual double TR() const { return pSubCell->TR(); };
    // Get dilution factor
    virtual double W() const { return pSubCell->W(); };
    // Get radioacive energy deposition
    //virtual double LRadio() const {return pSubCell->LRadio(); }; // Changed!
    // In cell - Changed!
    virtual bool inCell(const Vector3D& position, double t=noTime, bool virtualCell=false);
    // Calculate border distance - Changed!
    virtual double borderDistance(const Vector3D& position, const Vector3D& direction, BorderType& borderType, double t=noTime, bool virtualCell=false, bool log=false);
    // Get volume
    virtual double V(double t=noTime) { return pSubCell->V(t); };
    // Get random direction
    virtual Vector3D randomDirection(const CellLocation& cellLocation) { return pSubCell->randomDirection(cellLocation); };
    // Get random position - Changed!
    virtual Vector3D randomPosition(const CellLocation& cellLocation, double _t=noTime, bool virtualCell=false) { return pSubCell->randomPosition(cellLocation,_t,virtualCell); };
    // Get typical size of cell
    virtual double size() { return pSubCell->size(); }
    // Select cell
    virtual CellBase* selectCell(const Vector3D& position, const Vector3D& direction, double t=noTime);
    // Select cell neighbour
    virtual CellBase* selectCellNeighbour(const Vector3D& entryPosition, const Vector3D& entryDirection, const double& entryTime);
    // Select next cell
    virtual CellBase* selectNextCell(const Vector3D& position, const Vector3D& direction, const double& t);
    // Select next resonant line
    virtual LineIterator selectNextLine(const double& nu, const Vector3D& position, const Vector3D& direction, const double& t) const {return pSubCell->selectNextLine(nu,position,direction,t); };
    // Check if end of line list
    virtual bool endOfLineList(const LineIterator& lineIterator) const {return pSubCell->endOfLineList(lineIterator); };
    // Calculate distance to next resonant line (knowing the line)
    virtual double lineDistance(const LineIterator& lineIterator, const double& nu, const Vector3D& position, const Vector3D& direction, const double& t) const { return pSubCell->lineDistance(lineIterator,nu,position,direction,t); };
    // Calculate the distance to next resonant line (knowing the line and the frequency)
    virtual double lineDistance(const LineIterator& lineIterator, const double& nuCMF, double _t=noTime) { return pSubCell->lineDistance(lineIterator,nuCMF,_t); };
    // Select next ionization edge
    virtual EdgeIterator selectNextEdge(const double& nu, const Vector3D& position, const Vector3D& direction, const double& t) const {return pSubCell->selectNextEdge(nu,position,direction,t); };
    // Check if end of ionization edge list
    virtual bool endOfEdgeList(const EdgeIterator& edgeIterator) const {return pSubCell->endOfEdgeList(edgeIterator); };
    // Calculate distance to next ionization edge (knowing the edge)
    virtual double egdeDistance(const EdgeIterator& edgeIterator, const double& nu, const Vector3D& position, const Vector3D& direction, const double& t) const { return pSubCell->edgeDistance(edgeIterator,nu,position,direction,t); }; 
    // Calculate the distance to next ionization edge (knowing the edge and the frequency)
    virtual double edgeDistance(const EdgeIterator& edgeIterator, const double& nuCMF, double _t=noTime) { return pSubCell->edgeDistance(edgeIterator,nuCMF,_t); };
    // Select next boost bin - New! - Changed!
    virtual FrequencyBinIterator selectNextBoostBin(const double& nu, const Vector3D& position, const Vector3D& direction, const double& t) const {return pCell->selectNextBoostBin(nu,position,direction,t); };
    // Check if end of boost bin list - New! - Changed!
    virtual bool endOfBoostBinList(const FrequencyBinIterator& boostBinIterator) const {return pCell->endOfBoostBinList(boostBinIterator); };
    // Get end of boost bin list - New!
    virtual FrequencyBinIterator endOfBoostBinList() { return pCell->endOfBoostBinList(); }
    // Check if beginning of boost bin list - New!
    virtual bool beginOfBoostBinList(const FrequencyBinIterator& boostBinIterator) const { return pCell->beginOfBoostBinList(boostBinIterator); }
    // Calculate distance to next boost bin (knowing the boost bin) - New! - Changed!
    virtual double boostBinDistance(const FrequencyBinIterator& boostBinIterator, const double& nu, const Vector3D& position, const Vector3D& direction, const double& t) const { return pCell->boostBinDistance(boostBinIterator,nu,position,direction,t); };
    // Calculate the distance to next boost bin (knowing the boost bin and the frequency) - New! - Changed!
    virtual double boostBinDistance(const FrequencyBinIterator& boostBinIterator, const double& nuCMF, double _t=noTime) { return pCell->boostBinDistance(boostBinIterator,nuCMF,_t); };
    // Calculate the opacity for electron scattering
    virtual double electronScatterKappa() const { return pSubCell->electronScatterKappa(); };
    // Calculate the opacity for compton scattering (not supported yet)
    virtual double comptonScatterKappa(const double& nu) const { return pSubCell->comptonScatterKappa(nu); };
    // Get the opacity for dust absorption - New!
    virtual double dustAbsKappa() const { return pSubCell->dustAbsKappa(); };
    // Set the opacity for dust absorption - New!
    virtual void dustAbsKappa(const double _dustAbsKappa) { return pSubCell->dustAbsKappa(_dustAbsKappa); };
    // Get the opacity for grey scattering - New!
    virtual double greyScatterKappa() const { return pSubCell->greyScatterKappa(); };
    // Set the opacity for grey scattering - New!
    virtual void greyScatterKappa(const double _greyScatterKappa) { return pSubCell->greyScatterKappa(_greyScatterKappa); };
    // Calculate the opacity for free-free absorption
    virtual double freeFreeAbsKappa(const double& nu) const { return pSubCell->freeFreeAbsKappa(nu); };
    // Calculate the emissivity for free-free emission
    virtual double freeFreeEta(const double& nu) const { return pSubCell->freeFreeEta(nu); };
    // Calculate free-free cooling rate
    virtual double freeFreeCooling(double _T=noT) { return pSubCell->freeFreeCooling(_T); };
    // Draw free-free emission frequency
    virtual double freeFreeNu() const { return pSubCell->freeFreeNu(); };
    // Calculate the opacity for bound-free absorption
    virtual double boundFreeAbsKappa(const double& nu, EdgeType edgeType=noEdge) { return pSubCell->boundFreeAbsKappa(nu,edgeType); };
    // Calculate the emissivity for bound-free emission
    virtual double boundFreeEta(const double& nu, bool cutOff=true) { return pSubCell->boundFreeEta(nu,cutOff); };
    // Calculate bound-free cooling rate
    virtual double boundFreeCooling(double _T=noT) const { return pSubCell->boundFreeCooling(_T); };
    // Calculate (spontaneous) bound-free cooling rate
    virtual double boundFreeCoolingSP(double _T=noT) { return pSubCell->boundFreeCoolingSP(_T); };
    // Draw bound-free emission frequency
    virtual double boundFreeNu() { return pSubCell->boundFreeNu(); };
    // Draw bound-free (downward) transition
    virtual Trans* boundFreeDeExcite() { return pSubCell->boundFreeDeExcite(); };
    // Draw bound-free (upward) transition
    virtual Trans* boundFreeExcite(const double& nu) { return pSubCell->boundFreeExcite(nu); };
    // Calculate the opacity for compton absorption
    virtual double comptonAbsKappa(const double& comptonAbsSigma ) const { return pSubCell->comptonAbsKappa(comptonAbsSigma); }; 
    // Calculate optical depth for next resonant line (knowing the line)
    // Pointer version
    virtual double lineTau(TransBB* trans) const { return pSubCell->lineTau(trans); };
    // Calculate optical depth for next resonant line (knowing the line)
    // Iterator version
    virtual double lineTau(const LineIterator& lineIterator) const { return pSubCell->lineTau(lineIterator); };
    // Calculate escape fraction for next resonant line (knowing the line)
    // Pointer version
    virtual double lineBeta(TransBB* trans) const { return pSubCell->lineBeta(trans); };
    // Calculate escape fraction for next resonant line (knowing the line)
    // Iterator version
    virtual double lineBeta(const LineIterator& lineIterator) const { return pSubCell->lineBeta(lineIterator); };
    // Calculate collisional cooling rate
    virtual double collCooling(double _T=noT) { return pSubCell->collCooling(_T); };
    // Get non-thermal collisional heating rate
    virtual double nonThermalHeating() const {return pSubCell->nonThermalHeating(); }; // Changed!
    // Set non-thermal collisional heating rate
    virtual void nonThermalHeating(const double& _nonThermalHeating) { return pSubCell->nonThermalHeating(_nonThermalHeating); };
    // Get non-thermal collisional excitation rate
    virtual double nonThermalExcitation() const { return pSubCell->nonThermalExcitation(); };
    // Set non-thermal collisional excitation rate
    virtual void nonThermalExcitation(const double& _nonThermalExcitation) { return pSubCell->nonThermalExcitation(_nonThermalExcitation); };
    // Get non-thermal collisional ionization rate
    virtual double nonThermalIonization() const { return pSubCell->nonThermalIonization(); };
    // Set non-thermal collisional ionization rate
    virtual void nonThermalIonization(const double& _nonThermalIonization) { return pSubCell->nonThermalIonization(_nonThermalIonization); };
    // Calculate cooling rate
    virtual double cooling(Processes& processes, double _T=noT) { return pSubCell->cooling(processes,_T); };
    // Calculate heating rate
    //virtual double heating(Processes& processes, double _T=noT) const { return pSubCell->heating(processes,_T); };
    virtual double heating(Processes& processes, double _T=noT) { return pSubCell->heating(processes,_T); }; // Changed!
    // Calculate the heating-cooling difference
    virtual double G(Processes& processes, double _T=noT) { return pSubCell->G(processes,_T); };
    // Calculate derivative of the heating-cooling difference
    virtual double dGdT(Processes& processes) { return pSubCell->dGdT(processes); };
    // Total continuum opacity (as a function of frequency)
    virtual double contKappaNu(const double& nu, const Processes& processes) { return pSubCell->contKappaNu(nu,processes); }; // Changed!
    // Rosseland mean (continuum) opacity
    virtual double rossContKappa(const Processes& processes, double nuMin=-1, double nuMax=-1, double tol=1e-2, bool containerMC=false) { return pSubCell->rossContKappa(processes,nuMin,nuMax,tol,containerMC); }; // Changed! - Testing!
    // Energy mean (continuum) opacity
    virtual double energyContKappa(const Processes& processes) { return pSubCell->energyContKappa(processes); }; // Changed!
    // Draw collisional excitation transition
    virtual Trans* collExcite() { return pSubCell->collExcite(); };
    // Draw charge-transfer excitation transition
    virtual Trans* CTExcite() { return pSubCell->CTExcite(); }
    // Draw non-thermal excitation transition
    virtual Trans* nonThermalExcite() { return pSubCell->nonThermalExcite(); };
    // Get estimator for line mean intensity
    virtual double Jb(const double& nu) { return pSubCell->Jb(nu); }; // Changed!
    // Integrate quantity over the mean intensity
    virtual double JIntegrate(const JIntegrand& integrand, double nuMin, double nuMax, int& nIter, const double& relTol, double absTol=0.0, bool includeOSP=true, bool useIterator=false) { return pSubCell->JIntegrate(integrand,nuMin,nuMax,nIter,relTol,absTol,includeOSP,useIterator); }; // Changed!
    // Update the radiation state
    virtual void updateRadiationState(const FrequencyBinList& nu) { return pSubCell->updateRadiationState(nu); };
    // Reset MC statistics
    virtual void resetMCStatistics(const FrequencyBinList& _nu) { return pSubCell->resetMCStatistics(_nu); };
    // Update MC statistics (using packet history)
    virtual void updateMCStatistics() { return pSubCell->updateMCStatistics(); };
    // Update MC statistics (using a single ray)
    virtual void updateMCStatistics(const PacketEvent& startPoint, const PacketEvent& endPoint, MCUpdateInfo& updateInfo) { return pSubCell->updateMCStatistics(startPoint,endPoint,updateInfo); };
    // Update g-packet MC statistics (using packet history)
    virtual void updateGMCStatistics() { return pSubCell->updateGMCStatistics(); };
    // Update g-packet MC statistics (using a single ray)
    virtual void updateGMCStatistics(const PacketEvent& startPoint, const PacketEvent& endPoint) { return pSubCell->updateGMCStatistics(startPoint,endPoint); };
    // Update the matter state
    //virtual void updateMatterState(bool initiate=false, bool initiateState=true) { return pSubCell->updateMatterState(initiate,initiateState); };
    virtual void updateMatterState(bool initiate=false, bool initiateState=true, bool updateMCDerived=true) { return pSubCell->updateMatterState(initiate,initiateState,updateMCDerived); }; // Changed
    // Export statistics
    virtual void exportStat(YAML::Emitter& data, const double& tauLim=0.0, string prefix="", bool close=true) { return pSubCell->exportStat(data,tauLim,prefix,close); };
    // Export statistics
    virtual void exportStatistics(ofstream& file, const double& tauLimit=1.0) { return pSubCell->exportStatistics(file,tauLimit); };
    // Draw cooling MC channel - New!
    virtual MCChannel coolingChannel(bool deActivate) { return pSubCell->coolingChannel(deActivate); };
    // Draw non-thermal MC channel - New!
    virtual MCChannel nonThermalChannel() { return pSubCell->nonThermalChannel(); }; 
    // Get boost factor - New!
    virtual double boostFactor(const double& _nu) { return pSubCell->boostFactor(_nu); };
    // Get free-free opacity boost factor - New!
    virtual double freeFreeKappaBoost() { return pSubCell->freeFreeKappaBoost(); };
    // Get bound-free opacity boost factor - New!
    virtual double boundFreeKappaBoost(const double& _nu, EdgeType edgeType=noEdge) { return pSubCell->boundFreeKappaBoost(_nu,edgeType); };
    // Get compton opacity boost factor - New!
    virtual double comptonKappaBoost() { return pSubCell->comptonKappaBoost(); };
    // Get memory allocated - Not supported yet!
    virtual double mem() const { return 0; };
    // Give CellVirtualState class access to private data
    friend class CellVirtualState;
  };

  // Virtual cell state
  class CellVirtualState {
  private :
    // MPI serialization
    friend class boost::serialization::access;
    template<class Archive> 
      void serialize (Archive & ar, const unsigned int version) 
      {
	ar & index;
	ar & subIndex;
	ar & positionOffset;
      }
  public :
    int index;
    int subIndex;
    Vector3D positionOffset;
    // Constructors
    CellVirtualState() {};
    CellVirtualState(CellVirtual* cell);
  };

  // No cell index
  const int noCell=-1;

  // Bound-free source function
  class BoundFreeS {
  private :
    Cell* cell;
  public :
    BoundFreeS() {};
    BoundFreeS(Cell* _cell);
    double operator()(double nu) const;
    vector<double> operator()(FrequencyBinIterator nuBegin, FrequencyBinIterator nuEnd, EdgeType edgeType=noEdge) const;
  };

  // Bound-free opacity - New!
  class BoundFreeKappa {
  private :
    Cell* cell;
  public :
    BoundFreeKappa() {};
    BoundFreeKappa(Cell* _cell);
    double operator()(double nu) const;
    vector<double> operator()(FrequencyBinIterator nuBegin, FrequencyBinIterator nuEnd, EdgeType edgeType=noEdge) const;
  };

  // Blackbody factor (2hnu^3/c^2)
  class BBFactor {
  public :
    BBFactor() {};
    double operator()(double nu) const;
    vector<double> operator()(FrequencyBinIterator nuBegin, FrequencyBinIterator nuEnd, EdgeType edgeType=noEdge) const {};
  };

  // Boltzman factor (e^(-kT/hnu))
  class BMFactor {
  private :
    double T;
  public :
    BMFactor() {};
    BMFactor(double _T);
    double operator()(double nu) const;
    vector<double> operator()(FrequencyBinIterator nuBegin, FrequencyBinIterator nuEnd, EdgeType edgeType=noEdge) const {};
  };
  
  // Cell
  class Cell : public CellBase {
  protected :
    // Cell type
    CellType pType;
    // Cell index
    int pIndex;
    // Grid back reference
    Grid* pGrid;
    // Minimum time
    double ptMin;
    // Maximum time
    double ptMax;
    // Expansion factor
    double pfExp;
    // Density
    double prho;
    // Temperature
    double pT;
    // Temperature from last matter state update
    double pTOld;
    // Temperature from last diffusion solver update - Testing!
    double pTDiff;
    // Electron number density
    double pne;
    // Mean frequency
    double nuMean;
    // Radiation temperature
    double pTR;
    // Dilution factor
    double pW;
    // Radiation energy density temperature
    double pTJ;
    // Radiation energy density temperature (based on blackbody regions);
    double pTJBB;
    // Filling Factor
    double pFillFactor;
    // Number of (virtuall) cells
    unsigned int pNCells;
    // Free-free cooling rate
    double pFreeFreeCooling;
    // Free-free on-the-spot (minimum) frequency
    double pFreeFreeOnTheSpotNu;
    // Spontaneous bound-free cooling rate
    double pBoundFreeCoolingSP;
    // Collisional cooling rate
    double pCollCooling;
    // Charge-transfer cooling rate - New1
    double pCTCooling;
    // MPI communicator
    boost::mpi::communicator* mpiCom;
    // Bound-free opacity table
    //Interpolator1D pBoundFreeAbsKappaTable;
    NuInterpolator<BoundFreeKappa> pBoundFreeAbsKappaTable;
    // Bound-free opacity table lock (openmp)
    omp_lock_t boundFreeTableLock;
    // Bound-free source-function table
    //Tabulator1D<BoundFreeS> pBoundFreeSTable;
    NuTabulator<BoundFreeS> pBoundFreeSTable;
    // Bound-free source-function table lock (openmp)
    omp_lock_t boundFreeSTableLock;
    // Blackbody factor table
    NuTabulator<BBFactor> pBBFactorTable;
    // Blackbody factor lock (openmp
    omp_lock_t bBFactorLock; 
    // Boltzmann factor table
    NuTabulator<BMFactor> pBMFactorTable;
    // Boltzmann factor lock (openmp)
    omp_lock_t bMFactorLock; 
    // Radiation field interpolator;
    Interpolator1D JInterpol;
    // Collisional excitation probability list
    TransPList collExcitePList;
    // Charge-transfer excitation probability list - New!
    TransPList CTExcitePList;
    // Non-thermal excitation probability list
    TransPList nonThermalExcitePList;
    // Thermal pool Markov transition probability list
    MarkovPList markovPListTherm;
    // Charge-transfer Markov transition probability list
    vector<MarkovPList> markovPListCT;
    // Non-thermal collisional heating rate
    double pNonThermalHeating;
    // Non-thermal collisional ionization rate
    double pNonThermalIonization;
    // Non-thermal collisional excitation rate
    double pNonThermalExcitation;
    // On-the-spot (new) range
    FrequencyBin pOnTheSpotNu;
    // Free-free emissivity boost factor - New!
    double pFreeFreeEtaBoost;
    // Free-free opacity boost factor - New!
    double pFreeFreeKappaBoost;
    // Bound-free opacity boost factor table - New!
    Interpolator1D pBoundFreeKappaBoostTable;
    // Compton opacity boost factor - New!
    double pComptonKappaBoost;
    // Bound-free cooling boost factor - New!
    double pBoundFreeCoolingBoost;
    // Free-free cooling boost factor - New!
    double pFreeFreeCoolingBoost;
    // Collisional cooling boost factor - New!
    double pCollCoolingBoost;
    // Radiative cooling boost factor - New!
    double pRadCoolingBoost;
    // Cooling boost factor - New!
    double pCoolingBoost;
    // Non-thermal heating boost factor - New!
    double pNonThermalHeatingBoost;
    // Non-thermal excitation boost factor - New!
    double pNonThermalExcBoost;
    // Non-thermal ionization boost factor - New!
    double pNonThermalIonBoost;
    // Non-thermal excitation/ionization boost factor - New!
    double pNonThermalExcIonBoost;
    // Non-thermal deposition boost factor - New!
    double pNonThermalBoost;
    // Thermal pool Markov transition probability boost factor - New!
    double pMarkovPThermBoost;
    // Charge-transfer Markov transition probability boost factor - New!
    vector<double> markovPCTBoost;
    // Boosted free-free emission probability - New!
    vector<double> freeFreeEtaP;
    // Grey scattering opacity - New!
    double pGreyScatterKappa;
    // Dust absorption opacity - New!
    double pDustAbsKappa;
    // Opacity floor - New!
    double pKappaFloor;
    // Testing!!!
    vector<TransBF*> majorEdge;
    // Update line list
    void updateLineList();
    // Calculate free-free heating rate
    //double freeFreeHeating(double _T=noT) const;
    double freeFreeHeating(double _T=noT);
    // Calculate bound-free heating rate
    double boundFreeHeating(double _T=noT) const;
    // Calculate collisional heating rate
    double collHeating(double _T=noT) const;
    // Update blackbody frequency bins - New!
    void updateNuBB();
    // Update mean intensity
    void updateJ();
    // Update H and K
    void updateHK();
    // Update opacity (per frequency bin)
    void updateKappaNu();
    // Update frequency mean
    void updateNuMean();
    // Update diluted blackbody radiation field
    void updateDBBField();
    // Update blackbody field
    double updateBBField();
    // Testing
    //vector<double> boundFreeAbsKappaOMP(FrequencyBinIterator nuBegin, FrequencyBinIterator nuEnd, EdgeType edgeType=noEdge, int nBin=1, bool cutOff=true);
    // Calculate the opacity for bound-free absorption (openmp version)
    //vector<double> boundFreeAbsKappaOMP(const vector<double>& _nu, bool cutOff=true);
    // Update collisional excitation probability list
    void updateCollExcitePList();
    // Draw collisional excitation transition - Real time version
    Trans* collExciteCalc();
    // Draw collisional excitation transition - Table version
    Trans* collExciteTable();
    // Update charge-transfer excitation probability list - New!
    void updateCTExcitePList();
    // Draw charge-transfer excitation transition - Table version - New!
    Trans* CTExciteTable();
    // Update non-thermal excitation probability list
    void updateNonThermalExcitePList();
    // Draw non-thermal excitation transition - Real time version
    Trans* nonThermalExciteCalc();
    // Draw non-thermal excitation transition - Table version
    Trans* nonThermalExciteTable();
    // Update emission probability tables - New!
    void updateEtaPTable();
    // Calculate bound-free opacity boost factor - New!
    double boundFreeKappaBoostCalc(const double& _nu);
    // Create bound-free opacity boost factor table - New!
    void boundFreeKappaBoostTable();

  public :
    // List of atoms
    vector<Atom> atom;
    // List of charge-transfer reactions - New!
    vector<CTReaction> ctReaction;
    // List of lines
    LineList line;
    // List of ionization edges
    EdgeList edge;
    // Sub cells
    CellList subCell;
    // Packet history
    vector<PacketEvent*> packetHistory;
    // Cell config - Made public when testing
    CellConfig* config;
    // MPI process responsible for the cell
    int responsible;
    // Mean intensity - Made public - Replace with public function?
    double J;
    // H - Made public - Replace with public function?
    double H;
    // K - Made public - Replace with public function?
    double K;
    // Frequency bins for MC statistics - Made public when testing
    FrequencyBinList nu;
    // Angle bins for MC statistics - Made public when testing
    AngleBinList mu;
    // Mean intensity (per frequency bin) in the comoving frame - Made public when testing
    vector<double> Jnu;
    // H (per frequency bin) in the comoving frame - Made public when testing
    vector<double> Hnu;
    // K (per frequency bin) in the comoving frame - Made public when testing
    vector<double> Knu;
    // Intensity (per frequency and angle bin) in the comoving frame - Made public when testing
    vector<vector<double> > Inu;
    // Mean emissivity (per frequency bin) in the comoving frame - Made public when testing
    vector<double> etaNu;
    // Mean opacity-intensity products (per frequency bin) in the comoving frame - Made public when testing - New!
    vector<double> kappaJNu;
    vector<double> kappaBFJNu; // New!
    vector<double> kappaFFJNu; // New!
    vector<double> kappaBBJNu; // New!
    vector<double> kappaESJNu; // New! 
    // Mean opacities (per frequency bin) in the comoving frame - Made public when testing - New!
    vector<double> kappaNu;
    vector<double> kappaBFNu; // New!
    vector<double> kappaFFNu; // New!
    vector<double> kappaBBNu; // New!
    vector<double> kappaESNu; // New!
    // Mean opacity-H products (per frequency bin) in the comoving frame - Made public when testing - New!
    vector<double> kappaHNu;
    vector<double> kappaBFHNu; // New!
    vector<double> kappaFFHNu; // New!
    vector<double> kappaBBHNu; // New!
    vector<double> kappaESHNu; // New! 
    /*// Total pathlength - Made public when testing - New!
    vector<double> lNu;
    // Absorption event count - Made public when testing - New!
    vector<uint64_t> nAbsNu; // New!
    vector<uint64_t> nAbsBFNu; // New!
    vector<uint64_t> nAbsFFNu; // New!
    vector<uint64_t> nAbsBBNu; // New!
    vector<uint64_t> nAbsESNu; // New!*/
    // Mean frequency (per frequency bin) in the comoving frame  - Made public when testing
    vector<double> nuMeanNu;
    // Line frequency bins for MC statistics - Made public when testing
    FrequencyBinList nuLine;
    // Mean intensity at the line frequency in the comoving frame - Made public when testing
    vector<double> JLine;
    // Blackbody frequency bins for MC statistics - Made public when testing
    FrequencyBinList nuBB;
    // Mean intensity event count - Made public when testing
    vector<uint64_t> nJ;
    // Line intensity event count - Made public when testing
    vector<uint64_t> nJLine;
    // Radioactive energy deposition event count - Made public when testing - New!
    uint64_t neRadio;
    // Matter state solver - Made public when testing
    MStateSolver* mStateSolver;
    // Equation of state - Made public when testing
    EOS* eos;
    // Bound-free opacity table update flag - Made public when testing
    bool boundFreeTableUpdate;
    // Bound-free source-function table update flag - Made public when testing
    bool boundFreeSTableUpdate;
    // Bound-free cut-off update flag - Made public when testing
    bool boundFreeCutOffUpdate;
    // Blackbody factor update flag - Made public when testing
    bool bBFactorUpdate;
    // Boltzmann factor update flag - Made public when testing
    bool bMFactorUpdate;
    // Hack
    bool radUpdated;
    FrequencyBinList nuOld;
    AngleBinList muOld;
    vector<double> JnuOld;
    vector<double> HnuOld;
    vector<double> KnuOld;
    vector<vector<double> > InuOld;
    vector<double> etaNuOld;
    vector<double> kappaJNuOld;
    vector<double> kappaBFJNuOld; // New!
    vector<double> kappaFFJNuOld; // New!
    vector<double> kappaBBJNuOld; // New!
    vector<double> kappaESJNuOld; // New!
    vector<double> kappaHNuOld;
    vector<double> kappaBFHNuOld; // New!
    vector<double> kappaFFHNuOld; // New!
    vector<double> kappaBBHNuOld; // New!
    vector<double> kappaESHNuOld; // New!
    /*vector<double> lNuOld; // New!
    vector<uint64_t> nAbsNuOld; // New!
    vector<uint64_t> nAbsBFNuOld; // New!
    vector<uint64_t> nAbsFFNuOld; // New!
    vector<uint64_t> nAbsBBNuOld; // New!
    vector<uint64_t> nAbsESNuOld; // New!*/
    vector<double> nuMeanNuOld;
    vector<double> JLineOld;
    FrequencyBinList nuLineOld;
    FrequencyBinList nuBBOld;
    vector<uint64_t> nJOld;
    vector<uint64_t> nJLineOld;
    uint64_t neRadioOld; // New!
    // Radioactive energy deposition (per mass) - Made public - Hack !!!
    double peRadio, peRadioOld;
    // MC boost bins
    FrequencyBinList nuBoost;
    // MC boost factors
    vector<double> boost;
    // Cell neighbours - Made public when testing
    vector<CellIterator> cellNeighbour;
    // Deafult constructor - Testing!!!
    //Cell();
    // Constructors
    Cell(MStateSolver* mStateSolver=NULL, EOS* eos=NULL, CellConfig* _config=NULL, boost::mpi::communicator* _mpiCom=NULL);
    // Copy constructors
    Cell(const Cell& cell);
    // Assignment operator
    Cell& operator=(const Cell& cell);
    // Destructor
    virtual ~Cell();
    // Set atom data references
    // Used during cell copy or assigment - The source cell needs to be provided as an argument
    void setAtomDataReferences(AtomDataList& atomDataList, Cell* cell);
    // Get grid back reference pointer
    virtual Grid* grid() const;
    // Get cell type
    virtual int type() const;
    // Get cell index
    virtual int index() const;
    // Set cell index - New!
    void index(const int& _index);
    // Get sub cell index
    virtual int subIndex() const;
    // Get minimum time
    virtual double tMin() const;
    // Set minimum time
    virtual void tMin(const double& _tMin);
    // Get maximum time
    virtual double tMax() const;
    // Set maximum time
    virtual void tMax(const double& _tMax);
    // Get (centered) time
    virtual double t() const;
    // Get expansion factor
    virtual double fExp() const;
    // Get rho
    virtual double rho() const;
    // Set rho
    virtual void rho(const double& _rho);
    // Get number density
    virtual double n() const;
    // Get T
    virtual double T() const;
    // Set T
    virtual void T(const double& _T);
    // Get TDiff - Testing!
    virtual double TDiff() const;
    // Set TDiff - Testing!
    virtual void TDiff(const double& _TDiff);
    // Get energy (per mass)
    virtual double E(double _T=noT);
    // Get pressure
    virtual double P(double _T=noT);
    // Get electron number density
    virtual double ne() const;
    // Set electron number density
    virtual void ne(const double& _ne);
    // Get electron number density number derivative
    virtual double dNedN(IState* iState);
    // Get electron number density number derivative
    virtual double dNedN(BState* bState);
    // Get electron fraction
    virtual double xe() const;
    // Get filling factor
    virtual double fillFactor() const;
    // Get number of (virtual) cells
    virtual unsigned int nCells() const;
    // Get total mass
    virtual double M();
    // Get total mass of atom
    virtual double M(const double& Z);
    // Get total mass of isotope
    virtual double M(const double& Z, const double& N);
    // Get velocity
    virtual double v(const Vector3D& position, double _t=noTime) const;
    // Get velocity gradient
    virtual double dvdr(double _t=noTime) const;
    // Get radiation temperature 
    virtual double TR() const;
    // Set radiation temperature 
    virtual void TR(const double& _TR);
    // Get dilution factor
    virtual double W() const;
    // Set diluton factor
    virtual void W(const double& _W);
    // Get (radition) energy density temperature
    double TJ() const;
    // Set (radiation) energy density temperature 
    virtual void TJ(const double& _TJ);
    // Get (radiation) energy density temperature (based on the blackbody regions)
    double TJBB() const;
    // Get radioacive energy deposition (per mass)
    virtual double eRadio() const;
    // Set radioacive energy deposition (per mass)
    virtual void eRadio(const double& _eRadio);
    // In cell (to be implemented by derived class) - Changed!
    virtual bool inCell(const Vector3D& position, double t=noTime, bool virtualCell=false);
    // Calculate border distance (to be implemented by derived class) - Changed!
    virtual double borderDistance(const Vector3D& position, const Vector3D& direction, BorderType& borderType, double t=noTime, bool virtualCell=false, bool log=false);
    // Get volume (to be implement by derived class)
    virtual double V(double t=noTime);
    // Get random direction
    virtual Vector3D randomDirection(const CellLocation& cellLocation);
    // Get random position - Changed!
    virtual Vector3D randomPosition(const CellLocation& cellLocation, double _t=noTime, bool virtualCell=false);
    // Get typical size of cell
    // Why is this necessary - Implemented on CellBase
    virtual double size() {};
    // Select cell
    virtual CellBase* selectCell(const Vector3D& position, const Vector3D& direction, double t=noTime);
    // Select cell neighbour
    virtual CellBase* selectCellNeighbour(const Vector3D& entryPosition, const Vector3D& entryDirection, const double& entryTime);
    // Select next cell
    virtual CellBase* selectNextCell(const Vector3D& position, const Vector3D& direction, const double& t);
    // Select next resonant line
    virtual LineIterator selectNextLine(const double& nu, const Vector3D& position, const Vector3D& direction, const double& t) const;
    // Check if end of line list
    virtual bool endOfLineList(const LineIterator& lineIterator) const;
    // Calculate distance to next resonant line (knowing the line)
    virtual double lineDistance(const LineIterator& lineIteraor, const double& nu, const Vector3D& position, const Vector3D& direction, const double& t) const;
    // Calculate the distance to next resonant line (knowing the line and the frequency)
    virtual double lineDistance(const LineIterator& lineIterator, const double& nuCMF, double _t=noTime);
    // Update ionization edge list - Made public - Ok?
    void updateEdgeList();
    // Select next ionization edge
    virtual EdgeIterator selectNextEdge(const double& nu, const Vector3D& position, const Vector3D& direction, const double& t) const;
    // Check if end of ionization edge list
    virtual bool endOfEdgeList(const EdgeIterator& edgeIterator) const;
    // Calculate distance to next ionization edge (knowing the edge)
    virtual double edgeDistance(const EdgeIterator& edgeIteraor, const double& nu, const Vector3D& position, const Vector3D& direction, const double& t) const;
    // Calculate the distance to next ionization edge (knowing the edge and the frequency)
    virtual double edgeDistance(const EdgeIterator& edgeIterator, const double& nuCMF, double _t=noTime);
    // Select next boost bin - New!
    virtual FrequencyBinIterator selectNextBoostBin(const double& nu, const Vector3D& position, const Vector3D& direction, const double& t) const;
    // Check if end of boost bin list - New!
    virtual bool endOfBoostBinList(const FrequencyBinIterator& boostBinIterator) const;
    // Get end of boost bin list - New!
    virtual FrequencyBinIterator endOfBoostBinList();
    // Check if beginning of boost bin list - New!
    virtual bool beginOfBoostBinList(const FrequencyBinIterator& boostBinIterator) const;
    // Calculate distance to next boost bin (knowing the boost bin) - New!
    virtual double boostBinDistance(const FrequencyBinIterator& boostBinIteraor, const double& nu, const Vector3D& position, const Vector3D& direction, const double& t) const;
    // Calculate the distance to next boost bin (knowing the boost bin and the frequency) - New!
    virtual double boostBinDistance(const FrequencyBinIterator& boostBinIterator, const double& nuCMF, double _t=noTime);
    // Calculate the opacity for electron scattering
    virtual double electronScatterKappa() const;
    // Calculate the opacity for compton scattering (not supported yet)
    virtual double comptonScatterKappa(const double& nu) const;
    // Get the opacity for dust absorption - New!
    virtual double dustAbsKappa() const;
    // Set the opacity for dust absorption - New!
    virtual void dustAbsKappa(const double _dustAbsKappa);
    // Get the opacity for grey scattering - New!
    virtual double greyScatterKappa() const;
    // Set the opacity for grey scattering - New!
    virtual void greyScatterKappa(const double _greyScatterKappa);
    // Get opacity floor - New!
    virtual double kappaFloor() const;
    // Set opacity floor - New!
    virtual void kappaFloor(const double _kappaFloor);
    // Calculate the opacity for free-free absorption
    virtual double freeFreeAbsKappa(const double& nu) const;
    // Calculate the emissivity for free-free emission
    virtual double freeFreeEta(const double& nu) const;
    // Calculate free-free cooling rate
    virtual double freeFreeCooling(double _T=noT);
    // Draw free-free emission frequency
    virtual double freeFreeNu() const;
    // Draw free-free emission frequency (intervall version) - New!
    double freeFreeNu(double nuMin, double nuMax) const;
    // Draw free-free emission frequency (normal version) - New!
    double freeFreeNuNormal() const;
    // Draw free-free emission frequency (boost version) - New!
    double freeFreeNuBoost() const;
    // Get free-free on-the-spot (minumum) frequency
    virtual double freeFreeOnTheSpotNu() const;
    // Calculate the opacity for bound-free absorption
    virtual double boundFreeAbsKappa(const double& nu, EdgeType edgeType=noEdge);
    // Calculate the opacity for bound-free absorption - New!
    virtual double boundFreeAbsKappa(FrequencyBinIterator _nu, EdgeType edgeType=noEdge);
    // Calculate bound-free opacity - Moved here when testing
    double boundFreeAbsKappaCalc(const double& nu, bool cutOff=true);
    // Calculate bound-free opacity (openmp version)
    vector<double> boundFreeAbsKappaOMP(FrequencyBinIterator nuBegin, FrequencyBinIterator nuEnd, EdgeType edgeType=noEdge, bool cutOff=true);
    // Create bound-free opacity table - Made public when testing
    void boundFreeAbsKappaTable(bool useOMP=false);
    // Synchronize bound-free opacity table (MPI) - New!
    void syncBoundFreeAbsKappaTable();
    // Calculate the emissivity for bound-free emission
    virtual double boundFreeEta(const double& nu, bool cutOff=true);
    // Calculate the emissivity for bound-free emission (openmp version)
    vector<double> boundFreeEtaOMP(FrequencyBinIterator nuBegin, FrequencyBinIterator nuEnd, EdgeType edgeType=noEdge, bool cutOff=true);
    // Calculate source-function for bound-free absorption
    virtual double boundFreeS(const double& _nu, EdgeType edgeType=noEdge);
    // Get the source-function for bound-free absorption - edgeType=noEdge ?!?
    virtual double boundFreeS(FrequencyBinIterator _nu, EdgeType edgeType);
    // Calculate bound-free source-function - Moved here when testing
    double boundFreeSCalc(const double& _nu);
    // Calculate the source-function for bound-free absorption (omp version)
    vector<double> boundFreeSOMP(FrequencyBinIterator nuBegin, FrequencyBinIterator nuEnd, EdgeType edgeType=noEdge);
    // Create bound-free sorce-function table - Moved here when testing
    void boundFreeSTable(bool useOMP=false);
    // Calculate the source-function for bound-free absorption
    virtual double boundFreedSdN(const double& _nu, BState* bState, EdgeType edgeType=noEdge);
    // Calculate the source-function for bound-free absorption - New!
    virtual double boundFreedSdN(FrequencyBinIterator _nu, BState* bState, EdgeType edgeType=noEdge) ;
    // Calculate the source-function for bound-free absorption (ionization state version)
    virtual double boundFreedSdN(const double& _nu, IState* iState, EdgeType edgeType=noEdge);
    // Calculate the source-function for bound-free absorption (ionization state version)
    virtual double boundFreedSdN(FrequencyBinIterator _nu, IState* iState, EdgeType edgeType=noEdge);
    // Calculate bound-free cooling rate
    virtual double boundFreeCooling(double _T=noT) const;
    // Calculate (spontaneous) bound-free cooling rate
    virtual double boundFreeCoolingSP(double _T=noT);
    // Calculate net bound-free heating rate (as a single integral)
    virtual double boundFreeG(double _T=noT);
    // Draw bound-free emission frequency
    virtual double boundFreeNu();
    // Draw bound-free (downward) transition
    virtual Trans* boundFreeDeExcite();
    // Draw bound-free (upward) transition
    virtual Trans* boundFreeExcite(const double& nu);
    // Update bound-free cut-off frequencies
    virtual void updateBoundFreeCutOff(bool checkEta=true, double boundFreeCutOff=-1);
    // Update bound-free cut-off frequencies (openmp version)
    virtual void updateBoundFreeCutOffOMP(bool checkEta=true, double boundFreeCutOff=-1);
    // Synchronize bound-free cut-off (MPI) - New!
    void syncBoundFreeCutOff();
    // Get blackbody factor (2hnu^3/c^2)
    double bBFactor(FrequencyBinIterator _nu, EdgeType edgeType=noEdge);
    // Create blackbody factor (2hnu^3/c^2) table - Private ?
    void bBFactorTable(); 
    // Get Boltzmann factor (e^(-hnu/kT))
    double bMFactor(FrequencyBinIterator _nu, EdgeType edgeType=noEdge);
    // Create Boltzman factor (e^(-hnu/kT)) table - Private ?
    void bMFactorTable();
    // Calculate the opacity for compton absorption
    virtual double comptonAbsKappa(const double& comptonAbsSigma) const;
    // Calculate optical depth for next resonant line (knowing the line)
    // Pointer version
    virtual double lineTau(TransBB* trans) const;
    // Calculate optical depth for next resonant line (knowing the line)
    // Iterator version
    virtual double lineTau(const LineIterator& lineIterator) const;
    // Calculate escape fraction for next resonant line (knowing the line)
    // Pointer version
    virtual double lineBeta(TransBB* trans) const;
    // Calculate escape fraction for next resonant line (knowing the line)
    // Iterator version
    virtual double lineBeta(const LineIterator& lineIterator) const;
    // Calculate collisional cooling rate
    virtual double collCooling(double _T=noT);
    // Calculate charge-transfer cooling rate - New!
    virtual double CTCooling(double _T=noT);
    // Calculate collisional heating rate - New!
    virtual double CTHeating(double _T=noT);
    // Get non-thermal collisional heating rate
    virtual double nonThermalHeating() const;
    // Set non-thermal collisional heating rate
    virtual void nonThermalHeating(const double& _nonThermalHeating);
    // Get non-thermal collisional excitation rate
    virtual double nonThermalExcitation() const;
    // Set non-thermal collisional excitation rate
    virtual void nonThermalExcitation(const double& _nonThermalExcitation);
    // Get non-thermal collisional ionization rate
    virtual double nonThermalIonization() const;
    // Set non-thermal collisional ionization rate
    virtual void nonThermalIonization(const double& _nonThermalIonization);
    // Calculate cooling rate
    virtual double cooling(Processes& processes, double _T=noT);
    // Calculate heating rate
    //virtual double heating(Processes& processes, double _T=noT) const;
    virtual double heating(Processes& processes, double _T=noT);
    // Update recombination correction - New!
    virtual void updateAlfaCorr();
    // Calculate the heating-cooling difference
    virtual double G(Processes& processes, double _T=noT);
    // Calculate derivative of the heating-cooling difference
    virtual double dGdT(Processes& processes);
    // Total continuum opacity (as a function of frequency)
    virtual double contKappaNu(const double& nu, const Processes& processes);
    // Total continuum opacity (as a function of frequency)
    virtual double contKappaNu(FrequencyBinIterator nu, const Processes& processes, EdgeType edgeType=noEdge);
    // Total MC opacity (as a function of frequency) - New!
    virtual double kappaMCNu(const double& nu, const Processes& processes);
    // Total MC opacity (as a function of frequency) - New!
    virtual double kappaMCNu(FrequencyBinIterator nu, const Processes& processes, EdgeType edgeType=noEdge);
    // Rosseland mean (continuum) opacity
    virtual double rossContKappa(const Processes& processes, double nuMin=-1, double nuMax=-1, double tol=1e-2, bool containerMC=false); // Changed! - Testing!
    // Rosseland mean (continuum) opacity (Monte-carlo version)
    virtual double rossKappaMC2(const Processes& processes, double nuMin, double nuMax, int nPacket=10000);
    // Rosseland mean (continuum) opacity (Monte-carlo version for container cells)
    virtual double rossKappaMC2Container(const Processes& processes, double nuMin, double nuMax, int nPacket=10000);
    // Rosseland mean (continuum) opacity (Monte-carlo version for normal cells)
    virtual double rossKappaMC2Normal(const Processes& processes, double nuMin, double nuMax, int nPacket=10000);
    // Energy mean (continuum) opacity
    virtual double energyContKappa(const Processes& processes);
    // Energy mean (total) MC-based opacity
    virtual double energyKappaMC(const Processes& processes);
    // Flux mean (total) MC-based opacity - New!
    virtual double fluxKappaMC(const Processes& processes);
    // Rosseland mean (total) MC-based opacity - New!
    virtual double rossKappaMC(const Processes& processes);
    // Draw collisional excitation transition
    virtual Trans* collExcite();
    // Draw non-thermal excitation transition - New!
    virtual Trans* nonThermalExcite();
    // Draw charge-transfer excitation transition
    virtual Trans* CTExcite();
    // Get estimator for line mean intensity
    virtual double Jb(const double& nu);
    // Integrate quantity over the mean intensity
    //virtual double JIntegrate(const JIntegrand& integrand, double nuMin, double nuMax, int& nIter, const double& relTol, double absTol=0.0, bool includeOSP=true);
    virtual double JIntegrate(const JIntegrand& integrand, double nuMin, double nuMax, int& nIter, const double& relTol, double absTol=0.0, bool includeOSP=true, bool useIterator=false, bool onlyOSP=false, string test=""); // Testing!
    // Integrate quantity over the mean intensity number derivative
    double dJdNIntegrate(const NuIntegrand& integrand, BState* bState, double nuMin, double nuMax, int& nIter, const double& relTol, double absTol=0.0);
    virtual double NetIntegrate(const NuIntegrand& integrand1, const NuIntegrand& integrand2, double nuMin, double nuMax, int& nIter, const double& relTol, double absTol=0.0, bool includeOSP=true, bool useIterator=false, bool onlyOSP=false); // Testing!
    // Integrate net rate over the on-the-spot region
    virtual double OSPIntegrate(const NuIntegrand& integrand1, const NuIntegrand& integrand2, double nuMin, double nuMax, const double& relTol, double absTol=0.0, bool useIterator=false);
    // Integrate (OSP) net rate number derivative
    virtual double dROSPdNIntegrate(const NuDNIntegrand& integrand1, const NuDNIntegrand& integrand2, BState* bState, double nuMin, double nuMax, const double& relTol, double absTol=0.0, bool useIterator=false);
    // Integrate (OSP) net rate number derivative (ionization state version) - New!
    virtual double dROSPdNIntegrate(const NuDNIIntegrand& integrand1, const NuDNIIntegrand& integrand2, IState* iState, double nuMin, double nuMax, const double& relTol, double absTol=0.0, bool useIterator=false);
    // Calculate blackbody intensity in the blackbody regions
    double BIntBB(const double& T, double tol) const;
    // Calculate mean intensity in the blackbody regions
    double JIntBB() const;
    // Update the radiation state (could be overloaded by derived class)
    virtual void updateRadiationState(const FrequencyBinList& nu);
    // Update radiation state for container (could be overloaded by derived class) - New!
    virtual void updateRadiationStateContainer(const FrequencyBinList& _nu);
    // Update radioactive energy deposition
    virtual void updateERadio();
    // Update radioactive energy deposition for container - New!
    virtual void updateERadioContainer();
    // Scale radiation field (with expansion factor)
    virtual void scaleRadiationField(const double& expFactor, bool excpCooling=false);
    // Scale radiation field for container (with expansion factor) - New!
    virtual void scaleRadiationFieldContainer(const double& expFactor, bool expCooling=false);
    // Reset MC statistics
    virtual void resetMCStatisticsOld(const FrequencyBinList& _nu); // Old version - Remove!
    virtual void resetMCStatistics(const FrequencyBinList& _nu);
    // Reset MC statistics for container
    virtual void resetMCStatisticsContainerOld(const FrequencyBinList& _nu); // Old version - Remove!
    virtual void resetMCStatisticsContainer(const FrequencyBinList& _nu);
    // Reset g-packet MC statistics
    virtual void resetGMCStatistics();
    // Reset g-packet MC statistics for container
    virtual void resetGMCStatisticsContainer();
    // Update MC statistics (using packet history)
    virtual void updateMCStatistics();
    // Update MC statistics (using a single ray)
    virtual void updateMCStatistics(const PacketEvent& startPoint, const PacketEvent& endPoint, MCUpdateInfo& updateInfo);
    // Update MC opacity - New!
    //virtual void updateMCKappa(const PacketEvent& endPoint, const FrequencyBinIterator& nuBin, const double& kappaJNuPacket);
    virtual void updateMCKappa(const PacketEvent& endPoint, const FrequencyBinIterator& nuBin, const double& JnuPacket, const double& l, const double& muCMF); // Changed!
    // Update g-packet MC statistics (using packet history)
    virtual void updateGMCStatistics();
    // Update g-packet MC statistics (using a single ray)
    virtual void updateGMCStatistics(const PacketEvent& startPoint, const PacketEvent& endPoint);
    // Synchronize MC statistics (MPI)
    virtual void syncMCStatistics();
    // Synchronize g-packet MC statistics (MPI)
    virtual void syncGMCStatistics();
    // Update the matter state
    //virtual void updateMatterState(bool initiate=false, bool initiateState=true);
    virtual void updateMatterState(bool initiate=false, bool initiateState=true, bool updateMCDerived=true); // Changed!
    // Update derived matter quantities
    //virtual void updateMatterDerived(bool initiate=false);
    virtual void updateMatterDerived(bool updateMCDerived=false); // Changed!
    // Synchronize matter state (MPI)
    virtual void syncMatterState();
    // Synchronize derived matter quantities (MPI) - New!
    virtual void syncMatterDerived(bool syncMCDerived=false);
    // Export mean intensity per wavelength
    void exportJlambda(const string& dataFile);
    // Export H and K per wavelength
    void exportHKlambda(const string& dataFile);
    // Export mean intensity per wavelength
    void exportJnu(const string& dataFile);
    // Load mean intensity per frequency
    void loadJnu(const string& dataFile);
    // Export intensity per frequency
    void exportInu(const string& dataFile);
    // Load intensity per frequency
    void loadInu(const string& dataFile);
    // Export line intensity
    void exportJLine(const string& dataFile);
    // Load line intensity
    void loadJLine(const string& dataFile);
    // Export mean emissivity per wavelength
    void exportEtaLambda(const string& dataFile);
    // Export opacity (as a function of wavelength)
    void exportKappaLambda(const string& dataFile);
    // Export MC-based opacity (as a function of wavelength)
    void exportKappaMCLambda(const string& dataFile);
    // Export bound-free source function (as a function of wavelength) - New!
    void exportSBFLambda(const string& dataFile);
    // Save boost factor - New!
    void saveBoost(const string& dataFile);
    // Load boost factor - New!
    void loadBoost(const string& dataFile);
    // De-excite via thermal pool (implements the macro-atom machinery)
    BState* deExciteMarkovTableTherm();
    // De-excite via charge-transfer (implements the macro-atom machinery) - New!
    BState* deExciteMarkovTableCT(CTReaction* _cTReaction, CTDirection cTDirection);
    // Calculate (row-normalized) Markov S matrix
    vector<vector<double> > markovS(const vector<vector<vector<int> > >& matrixIndex);
    // Calculate cell Markov S matrix - New!
    vector<vector<double> > markovSCell(const vector<vector<vector<double> > >& SAtom, const vector<vector<vector<int> > >& indexCell, const vector<vector<vector<int> > >& indexAtom);
    // Calculate thermal pool atom Markov S vector - New!
    vector<vector<double> > markovSAtomTherm(const vector<vector<vector<double> > >& SAtom, const vector<vector<vector<int> > >& indexAtom);
    // Calculate thermal pool Markov S matrix (original version)
    vector<double> markovSTherm(const vector<vector<vector<double> > >& SAtom, const vector<vector<vector<int> > >& indexAtom, const vector<vector<vector<int> > >& indexCell);
    // Calculate thermal pool Markov S vector (new version) - New!
    vector<double> markovSTherm(const vector<vector<double> >& SCell, const vector<vector<vector<double> > >& SAtom, const vector<vector<vector<int> > >& indexCell, const vector<vector<vector<int> > >& indexAtom);
    // Calculate charge-transfer Markov S vector - New!
    vector<vector<double> > markovSCT(const vector<vector<double> >& SCell, const vector<vector<vector<double> > >& SAtom, const vector<vector<vector<int> > >& indexCell, const vector<vector<vector<int> > >& indexAtom);
    // Singular (SInv) matrix check - New!
    bool singular(Mat_DP& SInv, const vector<vector<vector<int> > >& matrixIndex, int& rowIndex, int& atomIndex, int& iStateIndex, int& bStateIndex);
    // Singular (SThermInv) matrix check - New!
    bool singularTherm(Mat_DP& SInv, int& rowIndex);
    // Setup bound state index mapping
    vector<vector<vector<int> > > bMatrixIndex(int& nStates);
    // Count number of unique bound state indeces
    void countBMatrixIndex(int& nStates);
    // Update transition probability list
    void updateTransPList();
    // Update Markov transition probability list
    void updateMarkovPList();
    // Update split Markov transition probability list
    void updateMarkovPListSplit();
    // Update thermal pool Markov transition probability list - Changed!
    void updateMarkovPListTherm(const vector<vector<double> >& SCell, const vector<vector<vector<double> > >& SAtom, const vector<vector<vector<int> > >& indexCell, const vector<vector<vector<int> > >& indexAtom);
    // Update charge-transfer Markov transition probability list - New!
    void updateMarkovPListCT(const vector<vector<double> >& SCell, const vector<vector<vector<double> > >& SAtom, const vector<vector<vector<int> > >& indexCell, const vector<vector<vector<int> > >& indexAtom);
    // Reset derived quantities (to be updated)
    //void reset(int dependence=dpAll);
    void reset(int dependence=dpAll, bool boundFreeCutOff=true);
    // Export data
    virtual void exportStat(YAML::Emitter& data, const double& tauLim=0.0, string prefix="", bool close=true);
    // Export data for container - New!
    virtual void exportStatContainer(YAML::Emitter& data, const double& tauLim, string prefix="", bool close=true);
    // Export statistics (could be overloaded by derived class)
    virtual void exportStatistics(ofstream& file, const double& tauLimit=1.0);
    // Load Cell data
    virtual void load(pstring& cellPData, AtomDataList& atomDataList, GridFactory* gridFactory, TransBBSelector* transSelector, BStateSelector* bStateSelector);
    // Reload bound-bound transitions
    virtual void reloadTransBB(TransBBSelector* transSelector);
    // Load Cell data
    virtual void loadYAML(const YAML::Node& data, AtomDataList& atomDataList, GridFactory* gridFactory, TransBBSelector* transSelector, BStateSelector* bStateSelector);
    // Load Cell state
    virtual void loadState(const YAML::Node& data, bool loadIState=true, bool loadBState=true, bool loadRState=true); // Changed!
    // Load Cell state for container - New!
    virtual void loadStateContainer(const YAML::Node& data, bool loadIState=true, bool loadBState=true, bool loadRState=true); // Changed!
    // Load radiation state
    virtual void loadRadiationState(const YAML::Node& data);
    // Load matter state
    virtual void loadMatterState(const YAML::Node& data, bool loadIState=true, bool loadBState=true); // Changed!
    // Load grey scattering opacity
    virtual void loadGreyScatterKappa(const YAML::Node& data); // New!
    // Save Cell data (could be overloaded by derived class)
    virtual void save(YAML::Emitter& data, bool close=true) const;
    // Save Cell state (could be overloaded by derived class)
    virtual void saveState(YAML::Emitter& data, string prefix="", bool close=true);
    // Save Cell state for container and subcells (could be overloaded by derived class) - New!
    virtual void saveStateContainer(YAML::Emitter& data, string prefix="", bool close=true);
    // Set back references
    void setBackReference(Grid* grid);
    // Evolve radioactive state of cell
    virtual void evolveRadio(const double& t, const double& dt);
    // Give Grid class access to private data
    // Is master?
    bool master() const;
    // Set on-the-spot range
    void onTheSpotNu(const FrequencyBin& _onTheSpotNu);
    // Get on-the-spot range
    FrequencyBin onTheSpotNu() const;
    // Init frequency binning
    //void initNu(const double& nuMin, const double& nuMax, const int& nBin);
    void initNu(double nuMin=noNu, double nuMax=noNu, int nBin=noN); // Changed!
    // Init frequency binning for container
    //void initNuContainer(const double& nuMin, const double& nuMax, const int& nBin);
    void initNuContainer(double nuMin=noNu, double nuMax=noNu, int nBin=noN); // Changed!
    // Init angular binning - New!
    void initMu(int nBin=noN);
    // Init angular binning for container - New!
    void initMuContainer(int nBin=noN);
    // Boost MC interactions - New!
    void boostMC();
    // Boost emission - New!
    void boostEta();
    // Boost free-free emission - New!
    void boostFreeFreeEta();
    // Boost radiative cooling - New!
    void boostRadiativeCooling();
    // Boost bound-free cooling - New!
    void boostBoundFreeCooling();
    // Boost free-free cooling - New!
    void boostFreeFreeCooling();
    // Boost collisional cooling - New!
    void boostCollCooling();
    // Boost non-thermal rates - New!
    void boostNonThermal();
    // Boost transition probabilities - New!
    void boostTransP();
    // Boost thermal pool Markov transition probabilities
    void boostMarkovPTherm();
    // Boost charge-transfer Markov transition probabilities
    void boostMarkovPCT();
    // Boost opacity - New!
    void boostKappa();
    // Boost bound-bound opacity - New!
    void boostBoundBoundKappa();
    // Boost bound-free opacity - New!
    void boostBoundFreeKappa();
    // Boost free-free opacity - New!
    void boostFreeFreeKappa();
    // Boost Compton opacity - New!
    void boostComptonKappa();
    // Get boost factor - New!
    virtual double boostFactor(const double& _nu);
    // Get free-free emissivity boost factor - New!
    double freeFreeEtaBoost();
    // Get free-free opacity boost factor - New!
    virtual double freeFreeKappaBoost();
    // Get bound-free opacity boost factor - New!
    virtual double boundFreeKappaBoost(const double& _nu, EdgeType edgeType=noEdge);
    // Get compton opacity boost factor - New!
    virtual double comptonKappaBoost();
    // Get bound-free cooling boost factor - New!
    double boundFreeCoolingBoost();
    // Get free-free cooling boost factor - New!
    double freeFreeCoolingBoost();
    // Get collisional cooling boost factor - New!
    double collCoolingBoost();
    // Get radiative cooling boost factor - New!
    double radCoolingBoost();
    // Get cooling boost factor - New!
    double coolingBoost();
    // Get non-thermal heating boost factor - New!
    double nonThermalHeatingBoost();
    // Get non-thermal excitation boost factor - New!
    double nonThermalExcBoost();
    // Get non-thermal ionization boost factor - New!
    double nonThermalIonBoost();
    // Get non-thermal excitation/ionization boost factor - New!
    double nonThermalExcIonBoost();
    // Get non-thermal deposition boost factor - New!
    double nonThermalBoost();
    // Get thermal pool Markov transition probability boost factor - New!
    double markovPThermBoost();
    // Draw cooling MC channel - New!
    virtual MCChannel coolingChannel(bool deActivate);
    // Draw non-thermal MC channel - New!
    virtual MCChannel nonThermalChannel();
    // Get memory allocated - New!
    virtual double mem() const;
    // Get memory allocated for probability lists - New!
    virtual double memPList() const;
    // Get memory allocated for bound-bound transitions - New!
    virtual double memTransBB() const;
    // Get memory allocated for bound-free transition - New!
    double memTransBF() const;
    // Give Grid class acess to private data
    friend class Grid;
    // Give MC statistics (MPI) class access to private data
    friend class MCStatMPI;
    // Give bound-free opacity (MPI) class access to private data
    friend class BoundFreeAbsKappaMPI;
    // Give derived matter state quantities (MPI) class access to private data
    friend class MCDerivedMPI;
    // Give boost quantities (MPI) class access to private data
    friend class BoostMPI;
  };
  
  // Forward declarations
  class GridSS;
  //class FrequencyBin;

  // MC statistics for MPI communication
  class MCStatMPISS {
  private :
    // MPI serialization
    friend class boost::serialization::access;
    template<class Archive> 
      void serialize (Archive & ar, const unsigned int version) 
      {
	ar & FBack;
	ar & FFront;
	ar & FObsBack;
	ar & FObsFront;
      }
  public :
    double FBack;
    double FFront;
    double FObsBack;
    double FObsFront;
  };

  // Spherical symmetric cell
  class CellSS : public Cell {
  protected :
    // Inner border
    double prMin;
    // Outer border
    double prMax;
    // Flux at inner border
    double pFBack;
    // Flux at outer border
    double pFFront;
    // Observed flux at inner border
    double pFObsBack;
    // Observed flux at outer border
    double pFObsFront;
    // Calculate border distance (steady-state version) - Changed!
    virtual double borderDistanceSteady(const Vector3D& position, const Vector3D& direction, BorderType& borderType, bool virtualCell=false, bool log=false);
    // Calculate border distance (time-dependent version) - Changed!
    virtual double borderDistanceTime(const Vector3D& position, const Vector3D& direction, BorderType& borderType, double _t, bool virtualCell=false, bool log=false);
  public :
    // Constructors
    CellSS(MStateSolver* mStateSolver=NULL, EOS* eos=NULL, CellConfig* _config=NULL, boost::mpi::communicator* _mpiCom=NULL);
    // Copy constructor
    CellSS(const CellSS& cell);
    // Assigment operator
    CellSS& operator=(const CellSS& cell);
    // Destructor
    virtual ~CellSS();
    // Get inner border
    double rMin() const;
    // Get inner border (virtual-cell version)
    double rMinV(bool virtualCell=false);
    // Get inner border (time-dependent version)
    double rMinT(double _t=noTime, bool virtualCell=false);
    // Set inner border
    void rMin(const double& _rMin);
    // Get outer border
    double rMax() const;
    // Get outer border (virtual-cell version)
    double rMaxV(bool virtualCell=false);
    // Get outer border (time-dependent version)
    double rMaxT(double _t=noTime, bool virtualCell=false);
    // Set outer boundary
    void rMax(const double& _rMax);
    // Get flux at inner border
    double FBack() const;
    // Set flux at inner border
    void FBack(const double& _FBack);
    // Get flux at outer border
    double FFront() const;
    // Set flux at outer border
    void FFront(const double& _FFront);
    // Get observed flux at inner border
    double FObsBack() const;
    // Set observed flux at inner border
    void FObsBack(const double& _FObsBack);
    // Get observed flux at outer border
    double FObsFront() const;
    // Set observed flux at outer border
    void FObsFront(const double& _FObsFront);
    // Get luminosity at inner border
    double LBack() const;
    // Get luminosity at outer border
    double LFront() const;
    // In cell (overloaded) - Changed!
    virtual bool inCell(const Vector3D& position, double _t=noTime, bool virtualCell=false);
    // Calculate border distance (overloaded) - Changed!
    virtual double borderDistance(const Vector3D& position, const Vector3D& direction, BorderType& borderType, double _t=noTime, bool virtualCell=false, bool log=false);
    // Get volume (overloaded)
    virtual double V(double _t=noTime);
    // Get random direction (overloaded)
    virtual Vector3D randomDirection(const CellLocation& cellLocation);
    // Get random position (overloaded) - Changed!
    virtual Vector3D randomPosition(const CellLocation& cellLocation, double _t=noTime, bool virtualCell=false);
    // Get typical size of cell
    virtual double size();
    // Update radiation state for container (overloaded) - New!
    virtual void updateRadiationStateContainer(const FrequencyBinList& _nu);
    // Scale radiation field (with expansion factor)
    virtual void scaleRadiationField(const double& expFactor, bool expCooling=false);
    // Reset MC statistics (overloaded)
    virtual void resetMCStatistics(const FrequencyBinList& _nu);
    // Update MC statistics (overloaded)
    virtual void updateMCStatistics(const PacketEvent& startPoint, const PacketEvent& endPoint, MCUpdateInfo& updateInfo);
    // Synchronize MC statistics (MPI)
    virtual void syncMCStatistics();
    // Export statistics
    virtual void exportStat(YAML::Emitter& data, const double& tauLim=0.0, string prefix="", bool close=true);
    // Export statistics
    virtual void exportStatistics(ofstream& file, const double& tauLimit=1.0);
    // Load cell
    virtual void load(pstring& cellPData, AtomDataList& atomDataList, GridFactory* gridFactory, TransBBSelector* transSelector, BStateSelector* bStateSelector);
    // Load Cell data (overloaded)
    virtual void loadYAML(const YAML::Node& data, AtomDataList& atomDataList, GridFactory* gridFactory, TransBBSelector* transSelector, BStateSelector* bStateSelector);
    // Load Cell state (overloaded)
    virtual void loadState(const YAML::Node& data, bool loadIState=true, bool loadBState=true, bool loadRState=true); // Changed!
    // Load Cell state for container (overloaded) - New!
    virtual void loadStateContainer(const YAML::Node& data, bool loadIState=true, bool loadBState=true, bool loadRState=true); // Changed!
    // Load radiation state (overloaded)
    virtual void loadRadiationState(const YAML::Node& data);
    // Save Cell data (overloaded)
    virtual void save(YAML::Emitter& data, bool close=true) const;
    // Save Cell state (overloaded)
    virtual void saveState(YAML::Emitter& data, string prefix="", bool close=true);
    // Save Cell state for container (overloaded)
    virtual void saveStateContainer(YAML::Emitter& data, string prefix="", bool close=true);
    // Give Grid access to private data
    friend class GridSS;
  };

  // Bound state selector
  class BStateSelector {
  public :
    BStateSelector();
    virtual bool use(const BStateData& bState);
  };

  // Transition selector
  class TransBBSelector {
  protected :
    FrequencyBin nu;
  public :
    TransBBSelector();
    TransBBSelector(const FrequencyBin& _nu);
    virtual bool use(TransBBData* trans);
  };

  // State solver
  class StateSolver {
  public :
    // Update state
    virtual void update(Grid* _grid, const int& _cellIndexMax, bool initiate=false, bool initiateState=true) {};
    // Evolve state
    virtual void evolve(Grid* _grid, double& _dt, const int& _cellIndexMax, bool initiate=false, bool inititateState=true) {};
    // Load data
    virtual void load(const string& dataFile) {};
  };
  
  // Grid config
  class GridConfig {
  public :
    // Cell config
    CellConfig* cell;
    BStateSelector* bStateSelector;
    TransBBSelector* transSelector;
    UpdateMethod updateMethod;
    bool matterDerivedMPI; //New!
    bool loadIStateData;
    // Constructors
    GridConfig();
    GridConfig(CellConfig* _cell, BStateSelector* _bStateSelector, TransBBSelector* _transSelector, const UpdateMethod& _updateMethod, const bool& matterDerivedMPI, const bool& _loadIStateData=true);
  };

  // Grid object factory
  class GridFactory {
  public :
    virtual Cell* createCell(MStateSolver* mStateSolver, EOS* eos, CellConfig* config=NULL, boost::mpi::communicator* _mpiCom=NULL);
    virtual Cell* copyCell(Cell* cell);
    virtual GridFactory* copy();
  };

  // Grid
  class Grid {
  protected :
    GridConfig* config;
    GridFactory* factory;
    MStateSolver* mStateSolver;
    EOS* eos;
    boost::mpi::communicator* mpiCom;
    double ptMin;
    double ptMax;
  public :
    // Made public - Improve this !!!
    AtomDataList atomData;
    // Constructor
    Grid(MStateSolver* _mStateSolver=NULL, EOS* _eos=NULL, GridConfig* _config=NULL, boost::mpi::communicator* _mpiCom=NULL);
    // Constructor
    Grid(const string& dataFile, MStateSolver* _mStateSolver, EOS* _eos, GridConfig* _config=NULL, boost::mpi::communicator* _mpiCom=NULL);
    // Copy constructor
    Grid(const Grid& grid);
    // Assignment operator
    Grid& operator=(const Grid& grid);
    // Destructor
    virtual ~Grid();
    // List of cells
    CellList cell;
    // Get minimum time
    virtual double tMin() const;
    // Set minimum time
    virtual void tMin(const double& _tMin);
    // Get maximum time
    virtual double tMax() const;
    // Set maximum time
    virtual void tMax(const double& _tMax);
    // Get (centered) time
    virtual double t() const;
    // Is master?
    bool master() const;
    // Select cell
    virtual CellBase* selectCell(const Vector3D& position, const Vector3D& direction, const double& t);
    // Select subcell
    virtual CellBase* selectCell(CellVirtualState* cellVirtualState);
    // Load Grid data
    virtual void load(const string& dataFile);
    // Load Grid data
    virtual void loadYAML(const string& dataFile);
    // Load Grid state
    virtual void loadState(const string& dataFile, bool loadIState=true, bool loadBState=true, bool loadRState=true); // Changed!
    // Load grey scattering opacity
    virtual void loadGreyScatterKappa(const string& dataFile); // New!
    // Save Grid data
    // Save Grid data
    virtual void save(const string& dataFile) const;
    // Save Grid state
    virtual void saveState(const string& dataFile);
    // Update radiation state (all cells)
    void updateRadiationState(MCExperiment& mCExperiment);
    // Update radiation state (range of cells)
    void updateRadiationState(int iMin, int iMax, MCExperiment& mCExperiment);
    // Scale radiation fiel with expansion factor (all cells)
    void scaleRadiationField(const double& expFactor, bool expCooling=false);
    // Scale radiation field with expansion factor (range of cells)
    void scaleRadiationField(const double& expFactor, int iMin, int iMax, bool expCooling=false);
    // Update matter state (all cells)
    //void updateMatterState(bool initiate=false, bool initiateState=true);
    void updateMatterState(bool initiate=false, bool initiateState=true, bool updateMCDerived=true); // Changed!
    // Update matter state (range of cell)
    //void updateMatterState(int iMin, int iMax, bool initiate=false, bool initiateState=true);
    void updateMatterState(int iMin, int iMax, bool initiate=false, bool initiateState=true, bool updateMCDerived=true); // Changed!
    // Update matter state (on a single computer)
    //void updateMatterStateSingle(int iMin, int iMax, bool initiate=false, bool initiateState=true);
    void updateMatterStateSingle(int iMin, int iMax, bool initiate=false, bool initiateState=true, bool updateMCDerived=true); // Changed!
    // Update matter state (on a cluster using MPI)
    //void updateMatterStateCluster(int iMin, int iMax, bool initiate=false, bool initiateState=true);
    void updateMatterStateCluster(int iMin, int iMax, bool initiate=false, bool initiateState=true, bool updateMCDerived=true); // Changed!
    // Update derived matter state quantities (all cells) - New!
    void updateMatterDerived(bool updateMCDerived);
    // Update derived matter state quantities (range of cells) - New!
    void updateMatterDerived(int iMin, int iMax, bool updateMCDerived=true);
    // Update derived matter state quantities (on a single computer) - New!
    void updateMatterDerivedSingle(int iMin, int iMax, bool updateMCDerived=true);
    // Update derived matter state quantities (on a cluster using MPI) - New!
    void updateMatterDerivedCluster(int iMin, int iMax, bool updateMCDerived=true);
    // Reset MC statistics (all cells)
    void resetMCStatistics(const FrequencyBinList& nu);
    // Reset MC statistics (range of cells)
    void resetMCStatistics(int iMin, int iMax, const FrequencyBinList& nu);
    // Reset derived quantities (to be updated)
    void reset(int dependence=dpAll);
    // Export data
    virtual void exportStat(const string& statFile, const double& tauLim=0.0);
    // Export statistics (could be overloaded by derived class)
    virtual void exportStatistics(const string& dataFile, const double& tauLimit=1.0);
    // Init frequency binning
    //void initNu(const double& nuMin, const double& nuMax, const int& nBin);
    void initNu(double nuMin=noNu, double nuMax=noNu, int nBin=noN); // Changed!
    // Init angular binning - New!
    void initMu(int nBin=noN);
    // Get memory allocated
    virtual double mem() const;
    // Get memory allcoated for probability lists
    virtual double memPList() const;
    // Get memory allocated for bound-bound transitions - New!
    virtual double memTransBB() const;
    // Get memory allocated for bound-free transition - New!
    virtual double memTransBF() const;
  };

  // Cell spacing for processGridTemplate
  enum CellSpacing { csLinear, csLog };
  
  // Spherical symmetric grid object factor
  class GridSSFactory : public GridFactory {
  public :
    virtual Cell* createCell(MStateSolver* mStateSolver, EOS* eos, CellConfig* config=NULL, boost::mpi::communicator* _mpiCom=NULL);
    virtual Cell* copyCell(Cell* cell);
    virtual GridFactory* copy();
  };
  
  // Sperical symmetric grid
  class GridSS : public Grid {
  public :
    // Constructor
    GridSS(MStateSolver* _mStateSolver=NULL, EOS* _eos=NULL, GridConfig* _config=NULL, boost::mpi::communicator* _mpiCom=NULL);
    GridSS(const string& dataFile, MStateSolver* _mStateSolver, EOS* _eos, GridConfig* _config=NULL, boost::mpi::communicator* _mpiCom=NULL);
    // Copy constructor
    GridSS(const GridSS& grid);
    // Assigment operator
    GridSS& operator=(const GridSS& grid);
    // Destructor
    virtual ~GridSS();
    // Load grid data (overloaded)
    virtual void load(const string& dataFile);
    // Load grid data (and initialize radiation field)
    void load(const string& dataFile, const double& Rph, const double& Lph);
    void processGridTemplate(const string& gridTemplateFile, const string& gridFile);
  };
}

