#include <vector>
#include <math.h>
#include <string.h>
#include <unistd.h>
#include <omp.h>
#include "mc-util.h"
#include "mc-math.h"
#include "mc-physics.h"
#include "mc.h"
#include "msolver.h"
#include "eos.h"
#include "diffsolver.h"

using namespace util;
using namespace math;
using namespace physics;
using namespace mc;
//using namespace grid;
using namespace msolver;
using namespace eos;
using namespace diffsolver;

class BoostRegionConfig {
public:
  FrequencyBin nu;
  int nBin;
  double nJGood;
  double boostMin;
  double boostMax;
  // Constructors
  BoostRegionConfig();
  // Load data
  void load(const YAML::Node& data);
};

// Constructors
BoostRegionConfig::BoostRegionConfig() : nu(0,nuInf), nBin(50), nJGood(1e3), boostMin(0.25), boostMax(1e5) {};

// Load data
void BoostRegionConfig::load(const YAML::Node& data) {
										    
  // Load frequency range
  if (data["lambda"]) {
    if (data["lambda"].size()==2) {
      nu.min=(c/data["lambda"][1].as<double>())/AA;
      nu.max=(c/data["lambda"][0].as<double>())/AA;
    }
  }
  else { cout<<"BoostRegionConfig::load: Error: Could not get frequency range\n"; exit(-1); }
  
  // Load number of bins
  if (data["nBin"]) nBin=data["nBin"].as<int>();

  // Load good number of J events
  if (data["nJGood"]) nJGood=data["nJGood"].as<int>();

  // Load minimum boost factor
  if (data["boostMin"]) boostMin=data["boostMin"].as<double>();

  // Load number of bins
  if (data["boostMax"]) boostMax=data["boostMax"].as<double>();
}

// Boost config
class BoostConfig {
public :
  vector<BoostRegionConfig> region;
  // Constructors
  BoostConfig();
  BoostConfig(const string& dataFile);
  // Load data
  void load(const YAML::Node& data);
  void load(const string& dataFile);
};

// Constructors
BoostConfig::BoostConfig() : region(0) {};
BoostConfig::BoostConfig(const string& dataFile) { load(dataFile); };

// Load data
void BoostConfig::load(const YAML::Node& data) {
  int i;

  // Load boost regions
  region.resize(data.size());
  for (i=0;i<data.size();i++) region[i].load(data[i]);
}

// Load data (from file)
void BoostConfig::load(const string& dataFile) {
  YAML::Node data;

  // Load data from file
  data=YAML::LoadFile(dataFile);

  // Load data
  if (data["boost"]) load(data["boost"]);
}

// Testing!
void tableUpdate(Grid* grid, int diffBorder) {
  int i;

  for (i=diffBorder+1;i<grid->cell.size();i++) {
    if (grid->cell[i]->bBFactorUpdate) { grid->cell[i]->bBFactorTable(); grid->cell[i]->bBFactorUpdate=false; }
    if (grid->cell[i]->bMFactorUpdate) { grid->cell[i]->bMFactorTable(); grid->cell[i]->bMFactorUpdate=false; }
    if (grid->cell[i]->boundFreeSTableUpdate) { grid->cell[i]->boundFreeSTable(true); grid->cell[i]->boundFreeSTableUpdate=false; }
  }
}

// Update bound-free opacity table (if neccesary) - New!
// Subcells and containers: Works
/*void boundFreeTableUpdate(Grid* grid, int diffBorder) {
  // Should be configurable!
  int i,j;

  // For each cell the process is responsible for
  #pragma omp parallel for private(j)
  for (i=diffBorder+1;i<grid->cell.size();i++) { 
    if (grid->cell[i]->boundFreeCutOffUpdate) { 
      if (grid->cell[i]->type()==cellContainer)
	for (j=0;j<grid->cell[i]->subCell.size();j++) { grid->cell[i]->subCell[j]->updateBoundFreeCutOff(false); grid->cell[i]->subCell[j]->boundFreeCutOffUpdate=false; }
      else grid->cell[i]->updateBoundFreeCutOff(false); grid->cell[i]->boundFreeCutOffUpdate=false; 
    }
    if (grid->cell[i]->boundFreeTableUpdate) {
      if (grid->cell[i]->type()==cellContainer)
	for (j=0;j<grid->cell[i]->subCell.size();j++) { grid->cell[i]->subCell[j]->boundFreeAbsKappaTable(false); grid->cell[i]->subCell[j]->boundFreeTableUpdate=false; }
      else { grid->cell[i]->boundFreeAbsKappaTable(false); grid->cell[i]->boundFreeTableUpdate=false; }
    }
  }
  }*/

// Does not work for subcells (see below) - Fix!
/*void boundFreeTableUpdate(Grid* grid, int diffBorder, boost::mpi::communicator* mpiCom) {
  int i,j,nResponsible=0;
  bool master=false;

  if (mpiCom==NULL) master=true;
  else if (mpiCom->rank()==0) master=true;

  // Determine which processes are responsible for the cells
  for (i=diffBorder+1;i<grid->cell.size();i++) grid->cell[i]->responsible=grid->cell[i]->index()%mpiCom->size();

  // Count number of cells the process is responsible for
  for (i=diffBorder+1;i<grid->cell.size();i++) if (grid->cell[i]->responsible==mpiCom->rank()) nResponsible++;

  // For each cell the process is responsible for
  #pragma omp parallel for private(j)
  for (i=diffBorder+1;i<grid->cell.size();i++) { 
    if (grid->cell[i]->responsible==mpiCom->rank()) {
      // Set number of threads to use in nested omp parallelization
      omp_set_num_threads((omp_get_num_threads()-1)/nResponsible+1);
      // If needed - Update bound-free opacity cut-off and tables in advance - Testing!
      if (grid->cell[i]->boundFreeCutOffUpdate) { 
	if (grid->cell[i]->type()==cellContainer)
	  for (j=0;j<grid->cell[i]->subCell.size();j++) { grid->cell[i]->subCell[j]->updateBoundFreeCutOffOMP(false); grid->cell[i]->subCell[j]->boundFreeCutOffUpdate=false; }
	else grid->cell[i]->updateBoundFreeCutOffOMP(false); grid->cell[i]->boundFreeCutOffUpdate=false; 
      }
      if (grid->cell[i]->boundFreeTableUpdate) {
	if (grid->cell[i]->type()==cellContainer)
	  for (j=0;j<grid->cell[i]->subCell.size();j++) { grid->cell[i]->subCell[j]->boundFreeAbsKappaTable(true); grid->cell[i]->subCell[j]->boundFreeTableUpdate=false; }
	else { grid->cell[i]->boundFreeAbsKappaTable(true); grid->cell[i]->boundFreeTableUpdate=false; }
      }
    }
  }

  // Wait for all processes to reach this point
  mpiCom->barrier();

  // Syncronize bound-free cut-off and opacity table for each cell
  for (i=diffBorder+1;i<grid->cell.size();i++) {
    if (master) cout<<"mc-diff-evo::boundFreeTableUpdate: Synchronize cell "<<i<<" bound-free cut-off\n";
    if (grid->cell[i]->type()==cellContainer) for (j=0;j<grid->cell[i]->subCell.size();j++) grid->cell[i]->subCell[j]->syncBoundFreeCutOff();
    else grid->cell[i]->syncBoundFreeCutOff();
  }

  // Wait for all processes to reach this point
  mpiCom->barrier();

  // Syncronize bound-free cut-off and opacity table for each cell
  for (i=diffBorder+1;i<grid->cell.size();i++) {
    if (master) cout<<"mc-diff-evo::boundFreeTableUpdate: Synchronize cell "<<i<<" bound-free opacity table\n";
    if (grid->cell[i]->type()==cellContainer) for (j=0;j<grid->cell[i]->subCell.size();j++) grid->cell[i]->subCell[j]->syncBoundFreeAbsKappaTable();
    else grid->cell[i]->syncBoundFreeAbsKappaTable();
  }

  // Wait for all processes to reach this point
  mpiCom->barrier();
  }*/

// Update bound-free tables - New version!
void boundFreeTableUpdate(Grid* grid, int iMin, int iMax, boost::mpi::communicator* mpiCom, double boundFreeCutOff=-1) {
  int i,j,nResponsible=0;
  vector<Cell*> cellWork(0);
  bool master=false;

  if (mpiCom==NULL) master=true;
  else if (mpiCom->rank()==0) master=true;

  // Create work list of cells and subcells - New!
  for (i=iMin;i<=iMax;i++) {
    if (grid->cell[i]->type()==cellContainer) {
      for (j=0;j<grid->cell[i]->subCell.size();j++) {
	cellWork.push_back(grid->cell[i]->subCell[j]);
      }
    }
    else cellWork.push_back(grid->cell[i]);
  }
 
  // Determine which processes are responsible for the cells - Changed!
  for (i=0;i<cellWork.size();i++) cellWork[i]->responsible=cellWork[i]->index()%mpiCom->size();

  // Count number of cells the process is responsible for - New!
  for (i=0;i<cellWork.size();i++) if (cellWork[i]->responsible==mpiCom->rank()) nResponsible++;

  // Update matter state for the cells the process are responsible for - Changed!
  #pragma omp parallel for
  for (i=0;i<cellWork.size();i++) {
    if (cellWork[i]->responsible==mpiCom->rank()) {
      // Set number of threads to use in nested omp parallelization - New!
      omp_set_num_threads((omp_get_num_threads()-1)/nResponsible+1);
      // If needed - Update bound-free opacity cut-off and tables in advance - Testing!
      if (cellWork[i]->boundFreeCutOffUpdate) { cellWork[i]->updateBoundFreeCutOffOMP(false); cellWork[i]->boundFreeCutOffUpdate=false; }
      if (cellWork[i]->boundFreeTableUpdate) { cellWork[i]->boundFreeAbsKappaTable(true); cellWork[i]->boundFreeTableUpdate=false; }
    }
  }

  // Wait for all processes to reach this point
  mpiCom->barrier();

  // Syncronize bound-free cut-off and opacity table for each cell
  for (i=0;i<cellWork.size();i++) {
    if (master) cout<<"mc-diff-evo::boundFreeTableUpdate: Synchronize cell "<<cellWork[i]->index()<<" bound-free cut-off\n";
    cellWork[i]->syncBoundFreeCutOff();
  }

  // Wait for all processes to reach this point
  mpiCom->barrier();

  // Syncronize bound-free cut-off and opacity table for each cell
  for (i=0;i<cellWork.size();i++) {
    if (master) cout<<"mc-diff-evo::boundFreeTableUpdate: Synchronize cell "<<cellWork[i]->index()<<" bound-free opacity table\n";
    cellWork[i]->syncBoundFreeAbsKappaTable();
  }

  // Wait for all processes to reach this point
  mpiCom->barrier();
}

enum EPacketMethod { epLBased, epEBased };

class TransBBAllSelector : public TransBBSelector {
  virtual bool use(TransBBData* trans) { return true; }
};

class TransBBNoneSelector : public TransBBSelector {
  virtual bool use(TransBBData* trans) { return false; }
};

// Create statistics file name
string statFile(const string& statDir, const string& prefix, const int& index) {
  stringstream statFile;
  statFile.str("");
  statFile<<statDir<<"/"<<prefix<<"-"<<index<<".dat";
  return statFile.str();
}

// Calculate (Rosseland mean) optical depth in cells (on a single computer) - New!
// Does not work for subcells - Fix!
vector<double> getTauCellSingle(const Grid& grid, Processes processes=Processes(pEScatter|pBoundFree|pFreeFree)) {
  CellSS* cell;
  vector<double> tauCell(grid.cell.size());
  int i;
 
  // Calculate Rosseland mean optical depth within each cell
  #pragma omp parallel for private(cell)
  for (i=0;i<grid.cell.size();i++) { 
    cell = (CellSS*) grid.cell[i];
    //tauCell[i]=cell->rossContKappa(Processes(pEScatter|pBoundFree|pFreeFree))*cell->rho()*(cell->rMax()-cell->rMin());
    tauCell[i]=cell->rossContKappa(processes)*cell->rho()*(cell->rMax()-cell->rMin());
  }

  return tauCell;
}

// Calculate (Rosseland mean) optical depth in cells (on a cluster using MPI) - New!
// Subcellsand containers: Works
// Does not work for subcells (see below) - Fix!
/*vector<double> getTauCellCluster(const Grid& grid, boost::mpi::communicator* mpiCom, Processes processes=Processes(pEScatter|pBoundFree|pFreeFree)) {
  CellSS* cell;
  vector<double> tauCell(grid.cell.size());
  // Should be configurable!
  double boundFreeCutOff=1e-3;
  int i,j,nResponsible=0;

  // Determine which processes are responsible for the cells
  for (i=0;i<grid.cell.size();i++) grid.cell[i]->responsible=grid.cell[i]->index()%mpiCom->size();

  // Count number of cells the process is responsible for
  for (i=0;i<grid.cell.size();i++) if (grid.cell[i]->responsible==mpiCom->rank()) nResponsible++;

  // For each cell the process is responsible for
  #pragma omp parallel for private(cell,j)
  for (i=0;i<grid.cell.size();i++) { 
    if (grid.cell[i]->responsible==mpiCom->rank()) {
      // Set number of threads to use in nested omp parallelization
      omp_set_num_threads((omp_get_num_threads()-1)/nResponsible+1);
      // If needed - Update bound-free opacity cut-off and tables in advance - Testing!
      //if (grid.cell[i]->boundFreeCutOffUpdate) { 
      if (grid.cell[i]->boundFreeCutOffUpdate and processes(pBoundFree)) { 
	if (grid.cell[i]->type()==cellContainer)
	  for (j=0;j<grid.cell[i]->subCell.size();j++) { grid.cell[i]->subCell[j]->updateBoundFreeCutOffOMP(false,boundFreeCutOff); grid.cell[i]->subCell[j]->boundFreeCutOffUpdate=false; }
	else grid.cell[i]->updateBoundFreeCutOffOMP(false,boundFreeCutOff); grid.cell[i]->boundFreeCutOffUpdate=false; 
      }
      //if (grid.cell[i]->boundFreeTableUpdate) {
      if (grid.cell[i]->boundFreeTableUpdate and processes(pBoundFree)) {
	if (grid.cell[i]->type()==cellContainer)
	  for (j=0;j<grid.cell[i]->subCell.size();j++) 	{ grid.cell[i]->subCell[j]->boundFreeAbsKappaTable(true); grid.cell[i]->subCell[j]->boundFreeTableUpdate=false; }
	else { grid.cell[i]->boundFreeAbsKappaTable(true); grid.cell[i]->boundFreeTableUpdate=false; }
      }
      // Calculate (Rosseland mean) optical depth
      cell = (CellSS*) grid.cell[i];
      //tauCell[i]=cell->rossContKappa(Processes(pEScatter|pBoundFree|pFreeFree))*cell->rho()*(cell->rMax()-cell->rMin());
      tauCell[i]=cell->rossContKappa(processes)*cell->rho()*(cell->rMax()-cell->rMin());
    }
  }

  // Wait for all processes to reach this point
  mpiCom->barrier();

  // For each cell
  for (i=0;i<grid.cell.size();i++) {
    // Broadcast (Rosseland mean) optical depth
    boost::mpi::broadcast(*mpiCom,tauCell[i],grid.cell[i]->responsible);
  }

  // Wait for all processes to reach this point
  mpiCom->barrier();

  return tauCell;
  }*/

// Calculate (Rosseland mean) optical depth in cells (on a cluster using MPI) - New version!
// Subcells and containers: Works
//vector<double> getTauCellCluster(const Grid& grid, boost::mpi::communicator* mpiCom, Processes processes=Processes(pEScatter|pBoundFree|pFreeFree)) {
vector<double> getTauCellCluster(Grid& grid, boost::mpi::communicator* mpiCom, Processes processes=Processes(pEScatter|pBoundFree|pFreeFree)) {
  CellSS* cell;
  vector<double> tauCell(grid.cell.size());
  // Should be configurable!
  double boundFreeCutOff=1e-3;
  int i,j,nResponsible=0;

  // If bound-free processes included - Update bound-free tables
  if (processes(pBoundFree)) boundFreeTableUpdate(&grid,0,grid.cell.size()-1,mpiCom,boundFreeCutOff);

  // Determine which processes are responsible for the cells
  for (i=0;i<grid.cell.size();i++) grid.cell[i]->responsible=grid.cell[i]->index()%mpiCom->size();

  // Count number of cells the process is responsible for
  for (i=0;i<grid.cell.size();i++) if (grid.cell[i]->responsible==mpiCom->rank()) nResponsible++;

  // For each cell the process is responsible for
  #pragma omp parallel for private(cell,j)
  for (i=0;i<grid.cell.size();i++) { 
    if (grid.cell[i]->responsible==mpiCom->rank()) {
      // Set number of threads to use in nested omp parallelization
      omp_set_num_threads((omp_get_num_threads()-1)/nResponsible+1);
      // Calculate (Rosseland mean) optical depth
      cell = (CellSS*) grid.cell[i];
      //tauCell[i]=cell->rossContKappa(Processes(pEScatter|pBoundFree|pFreeFree))*cell->rho()*(cell->rMax()-cell->rMin());
      tauCell[i]=cell->rossContKappa(processes)*cell->rho()*(cell->rMax()-cell->rMin());
    }
  }

  // Wait for all processes to reach this point
  mpiCom->barrier();

  // For each cell
  for (i=0;i<grid.cell.size();i++) {
    // Broadcast (Rosseland mean) optical depth
    boost::mpi::broadcast(*mpiCom,tauCell[i],grid.cell[i]->responsible);
  }

  // Wait for all processes to reach this point
  mpiCom->barrier();

  return tauCell;
}

// Calculate (Rosseland mean) optical depth in cells - New!
// Does not work for subcells - Fix!
//vector<double> getTauCell(const Grid& grid, Processes processes=Processes(pEScatter|pBoundFree|pFreeFree), boost::mpi::communicator* mpiCom=NULL) {
vector<double> getTauCell(Grid& grid, Processes processes=Processes(pEScatter|pBoundFree|pFreeFree), boost::mpi::communicator* mpiCom=NULL) {
  if (mpiCom==NULL) return getTauCellSingle(grid,processes);
  else return getTauCellCluster(grid,mpiCom,processes);
}

// Find optical depth border
// Subcellsand containers: Works
//void findBorder(const Grid& grid, const double& tauLim, int& border, double& R, Processes processes=Processes(pEScatter|pBoundFree|pFreeFree), boost::mpi::communicator* mpiCom=NULL) {
void findBorder(Grid& grid, const double& tauLim, int& border, double& R, Processes processes=Processes(pEScatter|pBoundFree|pFreeFree), boost::mpi::communicator* mpiCom=NULL) {
  CellSS* cell;
  double tau;
  vector<double> tauCell(grid.cell.size());
  int i;
  bool master=false;

  if (mpiCom==NULL) master=true;
  else if (mpiCom->rank()==0) master=true;

  if (master) cout<<"mc-diff-evo::findBorder: Calculate Rosseland mean optical depth within each cell\n";

  // Get (Rosseland optical depth) in the cells
  tauCell=getTauCell(grid,processes,mpiCom);

  // Find border
  tau=0;
  for (i=grid.cell.size()-1;i>=0;i--) {
    cell = (CellSS*) grid.cell[i];
    border=i; R=cell->rMax();
    tau+=tauCell[i];
    if (master) cout<<"mc-diff-evo::findBorder: cell: "<<i<<", Accumulated Rosseland mean optical depth: "<<tau<<"\n";
    if (tau>tauLim) break;
  }
  // If no border
  if (tau<=tauLim) { border=-1; R=((CellSS*) grid.cell[0])->rMin(); }

  if (master) cout<<"mc-diff-evo::findBorder: border: "<<border<<"\n";
  // Testing!
  if (mpiCom!=NULL) cout<<"mc-diff-evo::findBorder: Process: "<<mpiCom->rank()<<"; border: "<<border<<"\n";
}

// Load bound-bound transitions
void loadTransBB(const Grid& grid, const int& cellIndexMin, const int& cellIndexMax) {
  TransBBAllSelector transAllSelector;
  int i;
  
  // Load all bound-bound transitions for each cell in the specified region
  for (i=cellIndexMin;i<=cellIndexMax;i++) grid.cell[i]->reloadTransBB(&transAllSelector);
}

// Update cell on-the-spot approximation
// Subcells and containers: Works
void updateOnTheSpotCell(Cell* cell, bool set, const double& tauEdgeLim,  const OnTheSpot& onTheSpot, const OnTheSpotNew& onTheSpotNew, bool master=true) { // Changed!
  FrequencyBin onTheSpotNu;
  TransBF* transBF;
  double tauEdge; // New!
  int i,j,k,l;

  // If container - Update on-the-spot approximation for each subcell and return
  if (cell->type()==cellContainer) { for (i=0;i<cell->subCell.size();i++) updateOnTheSpotCell(cell->subCell[i],set,tauEdgeLim,onTheSpot,onTheSpotNew,master); return; }

  // Reset on-the-spot (new) range
  onTheSpotNu.min=nuInf; onTheSpotNu.max=nuInf;

  // For each atom
  for (i=0;i<cell->atom.size();i++) {

    // Update on-the-spot flag
    if (set and onTheSpot(cell->atom[i])) cell->atom[i].onTheSpot(true);
    else cell->atom[i].onTheSpot(false);

    // For each ionization state
    for (j=0;j<cell->atom[i].iState.size();j++) {
      // For each bound state
      for (k=0;k<cell->atom[i].iState[j].bState.size();k++) {
	// For each (upward) bound-free transition
	for (l=0;l<cell->atom[i].iState[j].bState[k].upTransBF.size();l++) {
	  // Set transition pointer
	  transBF=cell->atom[i].iState[j].bState[k].upTransBF[l];
	  // If outside border and on-the-spot (new) requested
	  if (set and onTheSpotNew(transBF)) {
	    // Get transition (edge) optical depth in cell
	    tauEdge=transBF->kappa(nuEdge(transBF->nu(),blueEdge))*cell->rho()*(((CellSS*) cell)->rMax()-((CellSS*) cell)->rMin());
	    // If above limit
	    if (tauEdge>tauEdgeLim) {
	      // Set on-the-spot (new) flag to true
	      transBF->onTheSpotNew(true);
	      // If necessary - Adjust on-the-spot (new) range
	      if (transBF->nu()<onTheSpotNu.min) onTheSpotNu.min=transBF->nu();
	      if (master) cout<<"mc-diff-evo::updateOnTheSpotRegion: cell: "<<cell->index()<<", set OSP flag for "<<transBF->lowerState->iState()->name()<<":"<<transBF->lowerState->getIndex()<<", nuEdge: "<<transBF->nu()<<", tauEdge: "<<tauEdge<<"\n";	      
	    }
	  }
	  // Else - Set on-the-spot (new) flag to false
	  else transBF->onTheSpotNew(false);
	}
      }
    }
    // Update on-the-spot (new) range
    if (master) cout<<"mc-diff-evo::updateOnTheSpotRegion: cell: "<<cell->index()<<", onTheSpotNu: ("<<onTheSpotNu.min<<","<<onTheSpotNu.max<<")\n";
    cell->onTheSpotNu(onTheSpotNu);
  }
}

// Update on-the-spot approximation region
// Subcells and containers: Works
void updateOnTheSpotRegion(Grid& grid, const double& tauLim, const double& tauEdgeLim, const OnTheSpot& onTheSpot, const OnTheSpotNew& onTheSpotNew, Processes processes=Processes(pEScatter|pBoundFree|pFreeFree), boost::mpi::communicator* mpiCom=NULL) { // Changed!
  int border;
  double R;
  int i;
  bool master=false;

  if (mpiCom==NULL) master=true;
  else if (mpiCom->rank()==0) master=true;

  // Find on-the-spot border
  findBorder(grid,tauLim,border,R,processes,mpiCom);
  // Testing!!!
  //border=58;
  //border=41;

  // Update on-the-spot appoximation for each cell
  for (i=0;i<grid.cell.size();i++) updateOnTheSpotCell(grid.cell[i],(i>border),tauEdgeLim,onTheSpot,onTheSpotNew,master); // Changed!

  // Update derived matter quantities (in on-the-spot region)- Excluding derived MC quantities  
  grid.updateMatterDerived(border+1,grid.cell.size()-1,false);
}

// Initiate cell radiation state
// Subcells and containers: Works
void initRadiationStateCell(Cell* cell) {
  int i;

  // If container - Initiate radiation state for each subcell
  if (cell->type()==cellContainer) for (i=0;i<cell->subCell.size();i++) initRadiationStateCell(cell->subCell[i]);

  // Initialize to BB radiation field at matter temperature and reset
  cell->TR(cell->T());
  cell->W(1.0);
  cell->reset();
}

// Initiate grid radiation state
// Subcells and containers: Works
void initRadiationState(const Grid& grid) {
  int i;

  // Initiate radiation state for each cell
  for (i=0;i<grid.cell.size();i++) initRadiationStateCell(grid.cell[i]);
}

// Initiate cell time borders
// Subcells and containers: Works
void initTimeBordersCell(Cell* cell, const double& dt) {
  double t;
  int i;

  // If container - Set time borders for each subcell
  if (cell->type()==cellContainer) for (i=0;i<cell->subCell.size();i++) initTimeBordersCell(cell->subCell[i],dt);

  // Get time
  t=cell->t();

  // Set cell time borders
  cell->tMin(t-dt/2);
  cell->tMax(t+dt/2);
}

// Initiate grid time borders
// Subcells and containers: Works
void initTimeBorders(Grid& grid, const double& dt) {
  double t;
  int i;

  // Get time
  t=grid.t();

  // Set grid time borders
  grid.tMin(t-dt/2);
  grid.tMax(t+dt/2);

  // Initiate time borders for each cell
  for (i=0;i<grid.cell.size();i++) initTimeBordersCell(grid.cell[i],dt);
}

// Update cell time borders
// Subcells and containers: Works
void updateTimeBordersCell(Cell* cell, const double& dt, const double& tMin, const double& tMax) {
  double dtCenter,t=(tMin+tMax)/2;
  int i;

  // If container - Update time borders for each subcell
  if (cell->type()==cellContainer) for (i=0;i<cell->subCell.size();i++) updateTimeBordersCell(cell->subCell[i],dt,tMin,tMax);

  // Update cell time borders
  dtCenter=2*(t+dt-tMax);
  cell->tMin(tMax);
  cell->tMax(cell->tMin()+dtCenter);
}

// Update cell time borders
// Subcells and containers: Works
void updateTimeBorders(Grid& grid, const double& dt, const double& tMin, const double& tMax) {
  double dtCenter,t=(tMin+tMax)/2;
  int i;

  // Update grid time borders
  dtCenter=2*(t+dt-tMax);
  grid.tMin(tMax);
  grid.tMax(grid.tMin()+dtCenter);

  // Update time borders for each cell
  for (i=0;i<grid.cell.size();i++) updateTimeBordersCell(grid.cell[i],dt,tMin,tMax);
}

// Create frequency bin list
// Subcells and containers: Not applicable
FrequencyBinList createNuList(const FrequencyBin& nu, const int& nBin) {
  int i;
  vector<double> nuWork(nBin+1);
  FrequencyBinList nuList;
  
  double dlognu=(log(nu.max)-log(nu.min))/float(nBin);
  for (i=0;i<nuWork.size();i++) nuWork[i]=nu.min*exp(i*dlognu);
  nuList.resize(nBin);
  for (i=0;i<nuList.size();i++) nuList[i].min=nuWork[i];
  for (i=0;i<nuList.size();i++) nuList[i].max=nuWork[i+1];

  return nuList;
}

// Cell radioactive decay luminosity
// Subcells and containers: Works
double LRadioCell(Cell* cell, const double& t) {
  double L=0;
  double M56Ni,M56Co,M56Fe,M57Ni,M57Co,M57Fe,M44Ti,M44Ca;
  int i;

  // If container - Sum radioactive decay luminosity for each subcell
  if (cell->type()==cellContainer) { for (i=0;i<cell->subCell.size();i++) L+=LRadioCell(cell->subCell[i],t); return L; }

  // Calculate the 56Ni decay-chain luminosity
  M56Ni=cell->M(28,28);
  M56Co=cell->M(27,29);
  M56Fe=cell->M(26,30);
  decay56Ni(M56Ni,M56Co,M56Fe,cell->t(),t);
  L+=M56Ni*e56Ni+M56Co*e56Co;
  
  // Calculate the 57Ni decay-chain luminosity
  M57Ni=cell->M(28,29);
  M57Co=cell->M(27,30);
  M57Fe=cell->M(26,31);
  decay57Ni(M57Ni,M57Co,M57Fe,cell->t(),t);
  L+=M57Ni*e57Ni+M57Co*e57Co;
  
  // Calculate the 44Ti decay-chain luminosity
  M44Ti=cell->M(22,22);
  M44Ca=cell->M(20,24);
  decay44Ti(M44Ti,M44Ca,cell->t(),t);
  L+=M44Ti*e44Ti;

  return L;
}

// Grid radioactive decay luminosity
// Subcells and containers: Works
double LRadio(Grid* grid, const double& t) {
  double L=0;
  int i;

  // Sum radioactive decay luminosity for each cell
  for (i=0;i<grid->cell.size();i++) L+=LRadioCell(grid->cell[i],t);

  return L;
}

// Radioactive decay energy integrand
// Subcells and containers: Works
class ERadioIntegrand {
public:
  Grid* grid;
  ERadioIntegrand(Grid* _grid) : grid(_grid) {};
  double operator()(double t) const { return LRadio(grid,t); }
};

// Radioactive decay energy
// Subcells and containers: Works
double ERadio(Grid* grid, const double& t1, const double& t2) {
  double E;
  ERadioIntegrand integrand(grid);
  // Testing
  int nIter;

  E=Integrator<ERadioIntegrand>::romberg(integrand,t1,t2,nIter,1e-2);

  return E;
}

// Sample thermal pakets
// Subcells and containers: Check packetSamplerThermal!
void sampleThermalPackets(Grid* grid, const int& cellIndexMin, const int& cellIndexMax, const double& EPacket, const FrequencyBin& nu, PacketStateList& packet, PacketConfig* packetConfig, boost::mpi::communicator* mpiCom=NULL, double expFactorWeight=1.0, bool LCorr=false, bool eRadioCorr=false) {
  double expFactor,E=0;
  double ECell; // New!
  Packet sPacket;
  int i,nPacket;
  // Testing!
  vector<double> T(grid->cell.size());

  // Calculate thermal energy in region
  //for (i=cellIndexMin;i<=cellIndexMax;i++) E+=grid->cell[i]->E()*grid->cell[i]->M();
  // Take dilution factor into account - Hack!!!
  //for (i=cellIndexMin;i<=cellIndexMax;i++) E+=grid->cell[i]->W()*grid->cell[i]->E()*grid->cell[i]->M();
  // Changed again!
  for (i=cellIndexMin;i<=cellIndexMax;i++) {
    ECell=grid->cell[i]->E();
    if (grid->cell[i]->W()!=1) ECell*=grid->cell[i]->W()*pow(grid->cell[i]->TR()/grid->cell[i]->T(),4);
    E+=ECell*grid->cell[i]->M();
    // If requested - correct for luminosity at cell borders - New!
    if (LCorr) { E-=expFactorWeight*(((CellSS*) grid->cell[i])->LBack()-((CellSS*) grid->cell[i])->LFront())*(grid->tMax()-grid->tMin()); }
    // If requested - correct for radioactive deposition - New!
    if (eRadioCorr) { E-=expFactorWeight*(grid->cell[i]->eRadio()*grid->cell[i]->M())*(grid->tMax()-grid->tMin()); }	
  }
	
  // Calculate number of thermal packets (corrected for expansion cooling)
  expFactor=(grid->tMin()+expFactorWeight*(grid->tMax()-grid->tMin()))/grid->tMin();
  nPacket=E*expFactor/EPacket;
 
  // Create new thermal packet sampler - Changed + Bug Fix!
  PacketSamplerBBThermal packetSamplerThermal(nPacket,cellIndexMin,cellIndexMax,grid,nu,packetConfig,mpiCom,0,expFactorWeight,LCorr,eRadioCorr);
  
  // Sample thermal packets in new region and add to packet list
  #pragma omp parallel for firstprivate(sPacket)
  for (i=0;i<packetSamplerThermal.nPacket;i++) {
    // If MPI run
    if (mpiCom!=NULL) {
      // If the process is responsible for the packet
       if (i%mpiCom->size()==mpiCom->rank()) {
	// Sample packet
	sPacket=packetSamplerThermal.sample(i);
	// Add packet to list
        #pragma omp critical
	packet.push_back(PacketState(sPacket));
      }
    }
    // Else
   else {
     // Sample the packet
     sPacket=packetSamplerThermal.sample(i);
     // Add packet to list
     #pragma omp critical
     packet.push_back(PacketState(sPacket));
   }
  }
}

// Get cell (radiation) energy
vector<double> getE(Grid* grid, boost::mpi::communicator* mpiCom=NULL) {
  vector<double> E(grid->cell.size());
  int i;
  bool master=false;
  if (mpiCom==NULL) master=true;
  else if (mpiCom->rank()==0) master=true;

  for (i=0;i<grid->cell.size();i++) {
    //E[i]=grid->cell[i]->E()*grid->cell[i]->rho();
    E[i]=grid->cell[i]->E();
    if (grid->cell[i]->W()!=1) E[i]*=grid->cell[i]->W()*pow(grid->cell[i]->TR()/grid->cell[i]->T(),4);
    //if (master) cout<<"getE: cell: "<<i<<", E: "<<E[i]<<"\n";
  }
  return E;
}

// Get cell (radiation) energy time derivative
vector<double> getdEdt(Grid* grid, const vector<double>& EPrev, const double& tPrev, boost::mpi::communicator* mpiCom=NULL) {
  vector<double> dEdt(grid->cell.size());
  double E;
  int i;
  bool master=false;
  if (mpiCom==NULL) master=true;
  else if (mpiCom->rank()==0) master=true;

  for (i=0;i<grid->cell.size();i++) { 
    //E=grid->cell[i]->E()*grid->cell[i]->rho();
    E=grid->cell[i]->E();
    if (grid->cell[i]->W()!=1) E*=grid->cell[i]->W()*pow(grid->cell[i]->TR()/grid->cell[i]->T(),4);
    dEdt[i]=(E-EPrev[i])/(grid->t()-tPrev); 
    //if (master) cout<<"getdEdt: cell: "<<i<<", dEdt: "<<dEdt[i]<<", E: "<<E<<", EPrev: "<<EPrev[i]<<", t: "<<grid->t()<<", tPrev: "<<tPrev<<"\n";
  }
  return dEdt;
}

// Sample thermal pakets - New version! - Testing!
// Subcells and containers: Check packetSamplerThermal!
/*void sampleThermalPackets(Grid* grid, const int& cellIndexMin, const int& cellIndexMax, const double& EPacket, const FrequencyBin& nu, PacketStateList& packet, PacketConfig* packetConfig, boost::mpi::communicator* mpiCom=NULL, vector<double> dEdt=vector<double>(0)) {
  double expFactor,E=0;
  double ECell; // New!
  Packet sPacket;
  int i,nPacket;
  // Testing!
  vector<double> T(grid->cell.size());

  // Calculate thermal energy in region
  for (i=cellIndexMin;i<=cellIndexMax;i++) {
    ECell=grid->cell[i]->E()*grid->cell[i]->rho();
    if (grid->cell[i]->W()!=1) ECell*=grid->cell[i]->W()*pow(grid->cell[i]->TR()/grid->cell[i]->T(),4);
    // If provided - Scale to tMin using energy time derivative - New! - Testing!
    if (dEdt.size()!=0) ECell-=MIN(dEdt[i]*grid->cell[i]->rho()*(grid->cell[i]->t()-grid->cell[i]->tMin()),ECell);
    E+=ECell*grid->cell[i]->V();
  }
	
  // Calculate number of thermal packets
  nPacket=E/EPacket;
 
  // Create new thermal packet sampler - Changed + Bug Fix!
  PacketSamplerBBThermal packetSamplerThermal(nPacket,cellIndexMin,cellIndexMax,grid,nu,packetConfig,mpiCom,0,dEdt);
  
  // Sample thermal packets in new region and add to packet list
  #pragma omp parallel for firstprivate(sPacket)
  for (i=0;i<packetSamplerThermal.nPacket;i++) {
    // If MPI run
    if (mpiCom!=NULL) {
      // If the process is responsible for the packet
       if (i%mpiCom->size()==mpiCom->rank()) {
	// Sample packet
	sPacket=packetSamplerThermal.sample(i);
	// Add packet to list
        #pragma omp critical
	packet.push_back(PacketState(sPacket));
      }
    }
    // Else
   else {
     // Sample the packet
     sPacket=packetSamplerThermal.sample(i);
     // Add packet to list
     #pragma omp critical
     packet.push_back(PacketState(sPacket));
   }
  }
  }*/

// Sample radiation field
// Subcells and containers: Check packetSamplerI!
void sampleRadiationField(Grid* grid, const int& cellIndexMin, const int& cellIndexMax, const double& EPacket, PacketStateList& packet, PacketConfig* packetConfig, boost::mpi::communicator* mpiCom=NULL, double expFactorWeight=1.0) {
  double expFactor,E=0;
  int i,nPacket;
  Packet sPacket;
  bool master=false;

  if (mpiCom==NULL) master=true;
  else if (mpiCom->rank()==0) master=true;

  // Calculate radiative energy in region
  for (i=cellIndexMin;i<=cellIndexMax;i++) {
    // Testing!
    if (master) cout<<"mc-diff-evo::sampleRadiationField: cell: "<<i<<", J: "<<grid->cell[i]->J<<"\n";
    E+=(4*pi/c)*grid->cell[i]->J*grid->cell[i]->V();
  }
	
  // Calculate number of packets (corrected for expansion cooling)
  expFactor=(grid->tMin()+expFactorWeight*(grid->tMax()-grid->tMin()))/grid->tMin();
  nPacket=E*expFactor/EPacket;
  
  if (master) cout<<"mc-diff-evo::sampleRadiationField: EPacket: "<<EPacket<<", nPacket: "<<nPacket<<", expFactor: "<<expFactor<<", E: "<<E<<", tMin: "<<grid->tMin()<<", tMax: "<<grid->tMax()<<", expFactorWeight: "<<expFactorWeight<<"\n";

  // Create new intensity packet sampler
  PacketSamplerI packetSamplerI(nPacket,cellIndexMin,cellIndexMax,grid,packetConfig,mpiCom,0,expFactorWeight);

  // Sample packets in region and add to packet list
  #pragma omp parallel for firstprivate(sPacket)
  for (i=0;i<packetSamplerI.nPacket;i++) {
    // If MPI run
    if (mpiCom!=NULL) {
      // If the process is responsible for the packet
      if (i%mpiCom->size()==mpiCom->rank()) {
	// Sample packet
	sPacket=packetSamplerI.sample(i);
	// Add packet to list
        #pragma omp critical
	packet.push_back(PacketState(sPacket));
      }
    }
    // Else
    else {
      // Sample the packet
      sPacket=packetSamplerI.sample(i);
      // Add packet to list
      #pragma omp critical
      packet.push_back(PacketState(sPacket));
    }
  }
  if (master) cout<<"sampleRadiationField: Sampled "<<packet.size()<<" packets\n";
}

// Sample initial packets
// Subcells and containers: Does not work yet!
void sampleInitialPackets(Grid* grid, const int& cellIndexMin, const int& cellIndexMax, const double& EPacket, const FrequencyBin& nu, PacketStateList& packet, PacketConfig* packetConfig, boost::mpi::communicator* mpiCom=NULL, double expFactorWeight=1.0) {
  int i,radFieldBorder=cellIndexMax+1;

  // Find the border outside which we know the state of the radiation field
  for (i=0;i<grid->cell.size();i++) if (grid->cell[i]->Inu.size()!=0) { radFieldBorder=i; break; }
  
  // Sample thermal packets inside the border and add to packet list  
  //if (cellIndexMin<radFieldBorder) sampleThermalPackets(grid,cellIndexMin,radFieldBorder-1,EPacket,nu,packet,packetConfig,mpiCom,expFactorWeight);
  // Hmm - maybe we need the old version here ...
  if (cellIndexMin<radFieldBorder) sampleThermalPackets(grid,cellIndexMin,radFieldBorder-1,EPacket,nu,packet,packetConfig,mpiCom);

  // Sample radiation field outside the border and add to packet list
  if (radFieldBorder<=cellIndexMax) sampleRadiationField(grid,MAX(radFieldBorder,cellIndexMin),cellIndexMax,EPacket,packet,packetConfig,mpiCom,expFactorWeight);

  // Hack for restart - Fix in a (much) better way !!!
  // Does not work with subcells and containers yet!!!
  if (cellIndexMin>radFieldBorder) for (i=radFieldBorder;i<cellIndexMin;i++) {
      grid->cell[i]->radUpdated=false;
      // Testing
      grid->cell[i]->T(grid->cell[i]->TR()*pow(grid->cell[i]->W(),1.0/4.0));
    }
}

// Evolve radiation state
// Subcells and containers: Check packetSamplerBBL and packetSamplerBBLGamma
void evolveRadiationState(Grid* grid, const int& diffBorder, const int& diffBorderOld, const double& EPacket, const FrequencyBin& nu, const FrequencyBinList& nuList, const bool& includeRadio, PacketStateList& packetIn, PacketStateList& packetOut, PacketStat* packetStat, MCConfig* mCConfig, boost::mpi::communicator* mpiCom=NULL, bool injectPackets=true, double nInjectPacketMax=std::numeric_limits<int>::max()) { //Changed!
  double Ldiff;
  //int nPacket,nPacketGamma;
  double nPacket,nPacketGamma; // Changed!
  PacketSampler* packetSampler=NULL;
  double factor; // New!
  int i;
  bool master=false;

  if (mpiCom==NULL) master=true;
  else if (mpiCom->rank()==0) master=true;

  // If new packets are to be injected
  if (injectPackets) {

    // If diffusion border the same
    //if (diffBorder==diffBorderOld) {

    // Get luminosity at diffusion border
    // Negative luminosties are taken care of below
    if (diffBorder>=0) Ldiff=MAX(((CellSS*) grid->cell[diffBorder])->LFront(),0);
    else Ldiff=0;
 
    // Calculate number of packets to inject
    nPacket=Ldiff*(grid->tMax()-grid->tMin())/EPacket;
    
    if (master) cout<<"mc-diff-evo::evolveRadiationState: nPacket: "<<nPacket<<", tMin: "<<grid->tMin()<<", tMax: "<<grid->tMax()<<", Ldiff: "<<Ldiff<<", EPacket: "<<EPacket<<"\n";
    
    // Set fraction of packets to reflect at diffusion border
    // Takes care of negative luminosities
    //if (((CellSS*) grid->cell[diffBorder])->FFront()<0) mCConfig->packet->PReflect=MAX(1.0-fabs(((CellSS*) grid->cell[diffBorder])->FFront()/BlackBody::F((grid->cell[diffBorder]->T()+grid->cell[diffBorder+1]->T())/2)),0.0);
    //else mCConfig->packet->PReflect=1.0;
    mCConfig->packet->PReflectInner=1.0;

    // If include radioactive heating
    if (includeRadio) {
      
      // Calculate number of gamma-packets to inject
      nPacketGamma=ERadio(grid,grid->tMin(),grid->tMax())/EPacket;
      if (master) cout<<"mc-diff-evo::evolveRadiationState: nPacketGamma: "<<nPacketGamma<<", tMin: "<<grid->tMin()<<", tMax: "<<grid->tMax()<<", ERadio: "<<ERadio(grid,grid->tMin(),grid->tMax())<<", EPacket: "<<EPacket<<"\n";
    }

    // If number of injected packets exceeds limit - New!
    if ((nPacket+nPacketGamma)>nInjectPacketMax) {
      // Calculate reduction factor
      factor=(nPacket+nPacketGamma)/nInjectPacketMax;
      // Reduce packet energy
      //EPacket=EPacket*factor;
      // Reduce number of packets
      nPacket=nPacket/factor;
      nPacketGamma=nPacketGamma/factor;
      if (master) cout<<"mc-diff-evo::evolveRadiationState: Maximum number of packets exceeded - Packet energy increased and number of packets decreased, EPacket: "<<EPacket*factor<<", nPacket: "<<nPacket<<", nPacketGamma: "<<nPacketGamma<<"\n";
    }

    // Create new packetSampler
    if (includeRadio) packetSampler = new PacketSamplerBBLGamma(Ldiff,int(nPacket),int(nPacketGamma),nu,diffBorder+1,grid,mCConfig->packet,mpiCom); // Changed!
    else packetSampler = new PacketSamplerBBL(Ldiff,int(nPacket),diffBorder+1,grid,nu,mCConfig->packet); // Changed!
  
    //}
  }

  if (master) cout<<"mc-diff-evo::evolveRadiationState: 1\n";
      
  // Reset packet statistics
  if (includeRadio) packetStat->reset(grid,0,grid->cell.size()-1,true,nuList);
  else packetStat->reset(grid,diffBorder+1,grid->cell.size()-1,true,nuList);
  if (master) cout<<"mc-diff-evo::evolveRadiationState: 2\n";
  // Reset MC statistics (in outer region)
  grid->resetMCStatistics(diffBorder+1,grid->cell.size()-1,nuList);
  if (master) cout<<"mc-diff-evo::evolveRadiationState: 3\n";
  // If include radioactive energy deposition
  if (includeRadio) {
    // Reset g-packet MC statistics (in inner region)
    for (i=0;i<=diffBorder;i++) grid->cell[i]->resetGMCStatistics();
  }
  if (master) cout<<"mc-diff-evo::evolveRadiationState: 4\n";
        
  // Create new MC experiment
  MCExperiment mCExperiment(packetSampler,packetStat,mCConfig,mpiCom);

  // Testing!!!
  //tableUpdate(grid,diffBorder);
  //if (mpiCom!=NULL) mpiCom->barrier();

  // Make sure bound-free opacity tables are updated - Testing!
  if (master) cout<<"mc-diff-evo::evolveRadiationState: Update bound-free opacity tables\n";
  //boundFreeTableUpdate(grid,diffBorder,mpiCom);
  boundFreeTableUpdate(grid,diffBorder+1,grid->cell.size()-1,mpiCom);
  if (mpiCom!=NULL) mpiCom->barrier();

  // Evolve MC experiment
  mCExperiment.evolve(packetIn,packetOut,grid);
      
  // Update radiation state (in outer region)
  grid->updateRadiationState(diffBorder+1,grid->cell.size()-1,mCExperiment);

  // If include radioactive heating
  if (includeRadio) {
    // Update radioactive heating (in inner region)
    // Only works with on-the-fly update method !
    for (i=0;i<=diffBorder;i++) grid->cell[i]->updateERadio();
  }

  // Cleanup
  if (packetSampler!=NULL) delete packetSampler;
}

// Export statistics
// Subcells and containers: Works
void exportStat(GridSS& grid, PacketStat& packetStat, string& gridStatDir, string& mCStatDir, int& nExport) {

  // Export grid statistics
  if (gridStatDir.compare("")!=0) grid.exportStat(statFile(gridStatDir,"grid-stat",nExport));
  // Export packet statistics
  if (mCStatDir.compare("")!=0) packetStat.exportStat(statFile(mCStatDir,"mc-stat",nExport));
  // Export spectrum
  if (mCStatDir.compare("")!=0) packetStat.exportSpectrum(statFile(mCStatDir,"spectrum",nExport));
  // Export spectrum per transition type
  if (mCStatDir.compare("")!=0) packetStat.exportSpectrumTrans(statFile(mCStatDir,"spectrum-trans",nExport));
  // Export observed spectral evolution
  if (mCStatDir.compare("")!=0) packetStat.specEvoObs->exportUpdated(mCStatDir,"spec-obs");

  // Increas export counter
  nExport++;
}

double getEPacket(GridSS& grid, const double& nPacketEff, const double& tStop, const bool& includeRadio) {
  double EPacket,E=0;
  int i,radFieldBorder=grid.cell.size();

  // Find the border outside which we know the state of the radiation field
  for (i=0;i<grid.cell.size();i++) if (grid.cell[i]->Inu.size()!=0) { radFieldBorder=i; break; }

  // Calculate total thermal energy inside the border
  //for (i=0;i<radFieldBorder;i++) E+=grid.cell[i]->E()*grid.cell[i]->M();
  // Take dilution factor into account - Hack!!!
  for (i=0;i<radFieldBorder;i++) E+=grid.cell[i]->W()*grid.cell[i]->E()*grid.cell[i]->M();

  // Calculate total radiative energy outside the border
  for (i=radFieldBorder;i<grid.cell.size();i++) E+=(4*pi/c)*grid.cell[i]->J*grid.cell[i]->V();

  // If include radioactive heating
  if (includeRadio) {
    
    // Calculate and add total radioactive energy
    E+=ERadio(&grid,grid.t(),tStop);
  }
    
  // Calculate energy per packet
  EPacket=E/nPacketEff;

  return EPacket;
}

//double getEPacketL(GridSS& grid, const int& nPacketEff) {
// Testing!
double getEPacketL(GridSS& grid, const double& nPacketLdt, const double& dt) {
  double EPacket,L,E;

  // Get luminosity
  L=((CellSS*) grid.cell[grid.cell.size()-1])->LFront();
  
  // Calculate energy emitted in 100 days
  //E=L*100*day;

  // Calculate energy emitted in one time-step
  // Testing!
  E=L*dt;
    
  // Calculate energy per packet
  EPacket=E/nPacketLdt;

  return EPacket;
}

// Scale cell matter state with expansion factor
// Applies to quantities not automatically scaled when the cell is expanded
// Subcells and containers: Works
void scaleMatterStateCell(Cell* cell, double expFactor) {
  int i,j,k,l,m;
  BState* bState;

  // If container - Scale matter state for each subcell and return
  if (cell->type()==cellContainer) { for (i=0;i<cell->subCell.size();i++) scaleMatterStateCell(cell->subCell[i],expFactor); return; }
    
  // Scale electron number density with expansion factor
  cell->ne(cell->ne()/pow(expFactor,3));

  // Scale non-thermal heating rate with expansion factor
  cell->nonThermalHeating(cell->nonThermalHeating()/pow(expFactor,3));
  // Scale non-thermal ionization rate with expansion factor
  cell->nonThermalIonization(cell->nonThermalIonization()/pow(expFactor,3));
  // Scale non-thermal excitation rate with expansion factor
  cell->nonThermalExcitation(cell->nonThermalExcitation()/pow(expFactor,3));
  
  // For each bound state
  for (j=0;j<cell->atom.size();j++)
    for (k=0;k<cell->atom[j].iState.size();k++)
      for (l=0;l<cell->atom[j].iState[k].bState.size();l++) {
	bState=&(cell->atom[j].iState[k].bState[l]);
	// Scale non-thermal ionization rate with expansion factor
	for (m=0;m<bState->upTransBF.size();m++)
	  //if (bState->upTransBF[m]->nonThermal()) bState->upTransBF[m]->CNU(bState->upTransBF[m]->CNU()/pow(expFactor,3));
	  if (bState->upTransBF[m]->active(pNonThermal)) bState->upTransBF[m]->CNU(bState->upTransBF[m]->CNU()/pow(expFactor,3));
	// Scale non-thermal excitation rate with expansion factor
	for (m=0;m<bState->upTransBB.size();m++) 
	  //if (bState->upTransBB[m]->nonThermal()) bState->upTransBB[m]->CNU(bState->upTransBB[m]->CNU()/pow(expFactor,3));
	  if (bState->upTransBB[m]->active(pNonThermal)) bState->upTransBB[m]->CNU(bState->upTransBB[m]->CNU()/pow(expFactor,3));
      }
}

// Scale grid matter state with expansion factor
// Applies to quantities not automatically scaled when the grid is expanded
// Subcells and containers: Works
void scaleMatterState(Grid& grid, double expFactor, int cellIndexMin, int cellIndexMax) {
  int i,j,k,l,m;
  BState* bState;

  // Scale matter state with expansion fact for each cell in the region
  for (i=cellIndexMin;i<=cellIndexMax;i++) scaleMatterStateCell(grid.cell[i],expFactor);
}

// Update cell matter state assuming LTE
// Subcells and containers: Works
void updateMatterStateLTECell(Cell* cell, string configFile) {
  LTESolver lteSolver;
  int i;

  // If container - Update matter state assuming LTE for each subcell and return
  if (cell->type()==cellContainer) { for (i=0;i<cell->subCell.size();i++) updateMatterStateLTECell(cell->subCell[i],configFile); return; }

  // Update partition function
  lteSolver.updateU(cell);
  // Update ionization state
  lteSolver.updateIonState(cell);
  // Update bound state
  lteSolver.updateBoundState(cell);
  
  // Update edge list (needed by rossContKappa)
  cell->updateEdgeList();
  
  // Reset derived quantities
  cell->reset();
}

// Update grid matter state assuming LTE
// Subcells and containers: Works
void updateMatterStateLTE(Grid& grid, string configFile) {
  int i;

  // Update matter state assuming LTE for each cell
  for (i=0;i<grid.cell.size();i++) updateMatterStateLTECell(grid.cell[i],configFile);
}

// Update non-thermal state
// Subcells and containers: Works
void updateNTState(Grid& grid, string configFile) {
  Koz92Solver ntSolver(configFile);
  vector<Cell*> cellWork(0);
  int i,j;

  // Create work list of cells and subcells
  for (i=0;i<grid.cell.size();i++)
    if (grid.cell[i]->type()==cellContainer) for (j=0;j<grid.cell[i]->subCell.size();j++) cellWork.push_back(grid.cell[i]->subCell[j]);
    else cellWork.push_back(grid.cell[i]);

  // For each cell
  #pragma omp parallel for firstprivate(ntSolver)
  for (i=0;i<cellWork.size();i++) {

    // Update non-thermal state
    ntSolver.update(cellWork[i],true);
  }
}

// Init cell non-thermal state
// Subcells and containers: Works
void initNTStateCell(Cell* cell, string configFile) {
  int i;

  // If container - Init non-thermal state for each subcell and return
  if (cell->type()==cellContainer) { for (i=0;i<cell->subCell.size();i++) initNTStateCell(cell->subCell[i],configFile); return; }
    
  // Initiate non-thermal state
  cell->nonThermalHeating(1e-99);
  cell->nonThermalIonization(0.0);
  cell->nonThermalExcitation(0.0);
}

// Init grid non-thermal state
// Subcells and containers: Works
void initNTState(Grid& grid, string configFile) {
  int i;

  // Init non-thermal state for each cell
  for (i=0;i<grid.cell.size();i++) initNTStateCell(grid.cell[i],configFile);
}

// Evolve dynamical state of grid (assuming homology)
// Subcells and containers: Works
void evolveDyn(Grid& grid, double dt) {
  double t,rCenter,rCenterOld;
  CellSS* cell;
  CellSS* cellPrev;
  int i,j;
 
  // Get time
  t=grid.t();

  // For each cell
  for (i=0;i<grid.cell.size();i++) {
    
    // Set cell pointers
    cell=((CellSS*) grid.cell[i]);
    if (i>0) cellPrev=((CellSS*) grid.cell[i-1]);

    // Calculate (center) radius
    rCenterOld=(cell->rMin()+cell->rMax())/2;

    // Calculate and set new (front) radius
    cell->rMax(cell->rMax()+(cell->rMax()/t)*dt);
 
    // If not innermost cell - Set new (back) radius to (front) radius for previous cell
    if (i>0) cell->rMin(cellPrev->rMax());
    // Else (i.e. innermost cell) - Calculate and set new (back radius) 
    else cell->rMin(cell->rMin()+(cell->rMin()/t)*dt);

    // Calculate new (center) radius
    rCenter=(cell->rMin()+cell->rMax())/2;

    // If container
    if (cell->type()==cellContainer)
      // Set new front and back radii for each subcell
      for (j=0;j<cell->subCell.size();j++) { ((CellSS*) cell->subCell[j])->rMin(cell->rMin()); ((CellSS*) cell->subCell[j])->rMax(cell->rMax()); }

    // If container
    if (cell->type()==cellContainer)
      // Calculate and set new density for each subcell
      for (j=0;j<cell->subCell.size();j++) cell->subCell[j]->rho(cell->subCell[j]->rho()*pow(rCenterOld/rCenter,3));
    // Else - Calculate and set new density
    else cell->rho(cell->rho()*pow(rCenterOld/rCenter,3));
  }

  // Update time borders
  updateTimeBorders(grid,dt,grid.tMin(),grid.tMax());
}

// Evolve radioactive state of grid
// Subcells and containers: Check packetSamplerGamma
void evolveRadio(Grid& grid, const double& dt, const double EPacket, PacketStateList& packetIn, MCConfig* mCConfig, boost::mpi::communicator* mpiCom=NULL) {
  PacketSamplerGamma* packetSamplerGamma;
  PacketStateList packetGammaOut(0);
  PacketStateList packetGammaIn(0);
  int nPacketGamma,i;
  
  // For each cell evolve radioactive state (from the last time-step)
  for (i=0;i<grid.cell.size();i++) grid.cell[i]->evolveRadio(grid.t()-dt,dt);

  // Sort out gamma packets from list
  for (i=0;i<packetIn.size();i++) if (packetIn[i].type==gPacket) packetGammaIn.push_back(packetIn[i]);

  // Calculate number of gamma-packets to inject
  nPacketGamma=ERadio(&grid,grid.tMin(),grid.tMax())/EPacket;

  // Turn off conversion of g-packets
  mCConfig->packet->convertGPacket=false;

  // Create new g-packet sampler
  packetSamplerGamma = new PacketSamplerGamma(nPacketGamma,&grid,mCConfig->packet,mpiCom);

  // Reset g-packet MC statistics
  for (i=0;i<grid.cell.size();i++) grid.cell[i]->resetGMCStatistics();
        
  // Create new MC experiment
  MCExperiment mCExperiment(packetSamplerGamma,NULL,mCConfig,mpiCom);
      
  // Evolve MC experiment
  mCExperiment.evolve(packetGammaIn,packetGammaOut,&grid);

  // Turn on conversion of g-packets
  // Assuming this was set to true - Is this ok?
  mCConfig->packet->convertGPacket=true;
      
  // Update radioactive heating
  // Only works with on-the-fly update method !
  for (i=0;i<grid.cell.size();i++) grid.cell[i]->updateERadio();

  // Cleanup
  if (packetSamplerGamma!=NULL) delete packetSamplerGamma;
}

// Subcells and containers: May not work yet!
/*void initBoostFactor(GridSS& grid, int nBoostBin, FrequencyBin& nu, boost::mpi::communicator* mpiCom=NULL) {
  double nuMin,nuLyman,nuBalmer,nuMax,dlognu;
  vector<double> nuWork;
  int i,j;
  bool master=false;

  if (mpiCom==NULL) master=true;
  else if (mpiCom->rank()==0) master=true;

  int nBoostBin1=50;
  int nBoostBin2=50;
  int nBoostBin3=1;
  double dlognu1,dlognu2,dlognu3;

  // Get frequency ranges
  nuMin=nu.min;
  nuBalmer=grid.cell[0]->atom[0].iState[0].bState[1].upTransBF[0]->nu();
  nuLyman=grid.cell[0]->atom[0].iState[0].bState[0].upTransBF[0]->nu();
  nuMax=nu.max;

  // Calculate frequencies for boost bins
  dlognu1=(log(nuBalmer)-log(nuMin))/float(nBoostBin1);
  for (i=0;i<nBoostBin1;i++) nuWork.push_back(nuMin*exp(i*dlognu1));
  dlognu2=(log(nuLyman)-log(nuBalmer))/float(nBoostBin2);
  for (i=0;i<nBoostBin2;i++) nuWork.push_back(nuBalmer*exp(i*dlognu2));
  dlognu3=(log(nuMax)-log(nuLyman))/float(nBoostBin3);
  for (i=0;i<nBoostBin3;i++) nuWork.push_back(nuLyman*exp(i*dlognu3));
  nuWork.push_back(nuMax);

  // Sort frequencies (to be on the safe side)
  sort(nuWork.begin(),nuWork.end());

  // For each cell
  for (i=0;i<grid.cell.size();i++) {

    // Resize boost vectors
    grid.cell[i]->nuBoost.resize(nuWork.size()-1);
    grid.cell[i]->boost.resize(nuWork.size()-1);

    // Initiate boost bins
    for (j=0;j<grid.cell[i]->nuBoost.size();j++) { 
      grid.cell[i]->nuBoost[j].min=nuWork[j]; 
      grid.cell[i]->nuBoost[j].max=nuWork[j+1];
      grid.cell[i]->boost[j]=1.0;
      if (master and i==0) cout<<"mc-diff-evo::initBoostFactor: nuBin: "<<j<<", lambda: ["<<(c/grid.cell[i]->nuBoost[j].min)/AA<<":"<<(c/grid.cell[i]->nuBoost[j].max)/AA<<"]\n";
    }
  }
  }*/

// Subcells and containers: May not work yet!
//void adjustBoostFactor(GridSS& grid, int diffBorder, double nJGood, boost::mpi::communicator* mpiCom=NULL) {
/*void adjustBoostFactor(GridSS& grid, int diffBorder, double nJGood, double boostMin, double boostMax, boost::mpi::communicator* mpiCom=NULL) {
  double boost,boostMean,nJ,nBin;
  FrequencyBinIterator nuBin,nuBinMin,nuBinMax;
  int i,j,k;
  bool master=false;
  double boostOuter=1;

  int nBoostBin1=50;
  int nBoostBin2=50;
  int nBoostBin3=1;

  if (mpiCom==NULL) master=true;
  else if (mpiCom->rank()==0) master=true;

  // Find photosphere and grey treatment border
  int photosphere,borderGrey;
  double R,tauPhot=1.0,tauGrey=100.0;
  findBorder(grid,tauPhot,photosphere,R,false);
  findBorder(grid,tauGrey,borderGrey,R,false);

  // For each cell
  for (i=diffBorder+1;i<grid.cell.size();i++) {

    // If grey treatment
    if (i<=borderGrey) {
      
      // Initialize number of bins and mean intensity events
      nBin=0; nJ=0;
      
      // Count number of bins and mean intensity events
      for (j=0;j<grid.cell[i]->nJ.size();j++) { 
	nBin++; 
	nJ+=grid.cell[i]->nJ[j]; 
      }

      // Calculate boost factor needed
      boost=(nJGood*nBin/MAX(nJ,1.0));
      boostMean=0;
      for (j=0;j<grid.cell[i]->boost.size();j++) boostMean+=grid.cell[i]->boost[j]/grid.cell[i]->boost.size();
      for (j=0;j<grid.cell[i]->boost.size();j++) grid.cell[i]->boost[j]=MIN(boostMean*boost,boostMax);

      if (master) cout<<"mc-diff-evo::adjustBoostFactor: cell: "<<i<<", Grey boost, boost: "<<boost<<", (new) boostFactor: "<<boostMean*boost<<", nJ: "<<nJ/nBin<<"\n";
    }
    // Else
    else {

      // For each boost bin
      for (j=0;j<grid.cell[i]->nuBoost.size();j++) {

	// Find overlapping frequency bins
	nuBinMin=grid.cell[i]->nu.find(grid.cell[i]->nuBoost[j].min);
	nuBinMax=grid.cell[i]->nu.find(grid.cell[i]->nuBoost[j].max);
	
	// Initialize number of bins and mean intensity events
	nBin=0; nJ=0;
	
	// Count number of bins and mean intensity events
	for (nuBin=nuBinMin;nuBin!=nuBinMax;nuBin++) { 
	  // Exclude edge bins
	  if (not nuBin->edge) nBin++;
	  nJ+=grid.cell[i]->nJ[nuBin-grid.cell[i]->nu.begin()]; 
	}
	// To be on the safe side
	nBin=MAX(nBin,1);
	
	// Calculate boost factor needed
	if (j<nBoostBin1+nBoostBin2) {
	  //if (grid.cell[i]->nuBoost[j].min>c/(10000*AA)) boost=(2*nJGood*nBin/MAX(nJ,1.0));
	  //else boost=(nJGood*nBin/MAX(nJ,1.0));	  
	  if (grid.cell[i]->nuBoost[j].min>c/(10000*AA)) boost=(nJGood*nBin/MAX(nJ,1.0));
	  else boost=((nJGood/4)*nBin/MAX(nJ,1.0));
	  grid.cell[i]->boost[j]=MAX(MIN(grid.cell[i]->boost[j]*boost,boostMax),boostMin);
	}
	else {
	  boost=(100.0*nBin/MAX(nJ,1.0));
	}
	
	if (master) cout<<"mc-diff-evo::adjustBoostFactor: cell: "<<i<<", lambdaBoost: ["<<(c/grid.cell[i]->nuBoost[j].max)/AA<<":"<<(c/grid.cell[i]->nuBoost[j].min)/AA<<"], boost: "<<boost<<", (new) boostFactor: "<<grid.cell[i]->boost[j]<<", nJ: "<<nJ/nBin<<", neRadio: "<<grid.cell[i]->neRadio<<"\n";
      }   
    }
  }
  }*/

// New version!
// Subcells and containers: May not work yet!
void initBoostFactor(GridSS& grid, BoostConfig& boostConfig, FrequencyBin& nu, boost::mpi::communicator* mpiCom=NULL) {
  double dlognu;
  BoostRegionConfig* region;
  vector<double> nuWork1,nuWork2;
  vector<Cell*> cellWork(0);
  int i,j;
  bool master=false;

  if (mpiCom==NULL) master=true;
  else if (mpiCom->rank()==0) master=true;

  // Create work list of cells and subcells
  for (i=0;i<grid.cell.size();i++) {
    if (grid.cell[i]->type()==cellContainer) for (j=0;j<grid.cell[i]->subCell.size();j++) cellWork.push_back(grid.cell[i]->subCell[j]);
    cellWork.push_back(grid.cell[i]);
  }

  // Calculate frequencies for boost bins
  for (i=0;i<boostConfig.region.size();i++) {
    region=&(boostConfig.region[i]);
    if (region->nu.max>nu.min and region->nu.min<nu.max) {
      dlognu=(log(MIN(region->nu.max,nu.max))-log(MAX(region->nu.min,nu.min)))/float(region->nBin);
      for (j=0;j<region->nBin;j++) nuWork1.push_back(MAX(region->nu.min,nu.min)*exp(j*dlognu));
      nuWork1.push_back(MIN(region->nu.max,nu.max));
    }
  }

  // Add frequency limits 
  nuWork1.push_back(nu.min);
  nuWork1.push_back(nu.max);

  // Sort frequencies
  sort(nuWork1.begin(),nuWork1.end());

  // Remove doublets
  nuWork2.push_back(nuWork1[0]); 
  for (i=1;i<nuWork1.size();i++) if (nuWork1[i]!=nuWork1[i-1]) nuWork2.push_back(nuWork1[i]);

  // For each cell in work list
  for (i=0;i<cellWork.size();i++) {

    // Resize boost vectors
    cellWork[i]->nuBoost.resize(nuWork2.size()-1);
    cellWork[i]->boost.resize(nuWork2.size()-1);

    // Initiate boost bins
    for (j=0;j<cellWork[i]->nuBoost.size();j++) { 
      cellWork[i]->nuBoost[j].min=nuWork2[j]; 
      cellWork[i]->nuBoost[j].max=nuWork2[j+1];
      cellWork[i]->boost[j]=1.0;
      if (master and i==0) cout<<"mc-diff-evo::initBoostFactor: nuBin: "<<j<<", lambda: ["<<(c/cellWork[i]->nuBoost[j].min)/AA<<":"<<(c/cellWork[i]->nuBoost[j].max)/AA<<"]\n";
    }
  }
}

// New version!
// Subcells and containers: May not work yet!
void adjustBoostFactor(GridSS& grid, int diffBorder, BoostConfig& boostConfig, bool subBoost=false, boost::mpi::communicator* mpiCom=NULL) {
  double boost,nJ,nBin,nuMin,nuMax; // Changed!
  BoostRegionConfig* region;
  FrequencyBinIterator nuBin,nuBinMin,nuBinMax;
  vector<Cell*> cellWork(0);
  int i,j,k;
  bool master=false;

  if (mpiCom==NULL) master=true;
  else if (mpiCom->rank()==0) master=true;

  // Create work list of cells and subcells
  for (i=diffBorder+1;i<grid.cell.size();i++)
    if (grid.cell[i]->type()==cellContainer and subBoost) for (j=0;j<grid.cell[i]->subCell.size();j++) cellWork.push_back(grid.cell[i]->subCell[j]); // Changed!
    else cellWork.push_back(grid.cell[i]);

  // For each cell in work list
  for (i=0;i<cellWork.size();i++) {

    // For each boost bin
    for (j=0;j<cellWork[i]->nuBoost.size();j++) {

      // Testing
      cout<<"mc-spec::adjustBoostFactor: cellWork["<<i<<"]->nuBoost["<<j<<"]: ["<<cellWork[i]->nuBoost[j].max<<":"<<cellWork[i]->nuBoost[j].min<<"]\n";
      
      // Search boost region
      region=NULL;
      for (k=0;k<boostConfig.region.size();k++)
	if (cellWork[i]->nuBoost[j].max>boostConfig.region[k].nu.min and cellWork[i]->nuBoost[j].min<boostConfig.region[k].nu.max) { region=&(boostConfig.region[k]); break; }
      
      // If region found
      if (region!=NULL) {
	
	// Find overlapping frequency bins
	nuMin=cellWork[i]->nu[0].min; // New!
	nuMax=cellWork[i]->nu[cellWork[i]->nu.size()-1].max; // New!
	nuBinMin=cellWork[i]->nu.find(MAX(cellWork[i]->nuBoost[j].min,nuMin)); // Changed!
	nuBinMax=cellWork[i]->nu.find(MIN(cellWork[i]->nuBoost[j].max,nuMax)); // Changed!

	// Testing!
	if (nuBinMin==cellWork[i]->nu.end()) cout<<"mc-spec::adjustBoostFactor: Warning: nuBinMin not found, cellWork["<<i<<"]->nuBoost["<<j<<"].min: "<<cellWork[i]->nuBoost[j].min<<"\n";
	if (nuBinMax==cellWork[i]->nu.end()) cout<<"mc-spec::adjustBoostFactor: Warning: nuBinMax not found, cellWork["<<i<<"]->nuBoost["<<j<<"].max: "<<cellWork[i]->nuBoost[j].max<<"\n";

	// Initialize number of bins and mean intensity events
	nBin=0; nJ=0;
	
	// Count number of bins and mean intensity events
	for (nuBin=nuBinMin;nuBin!=nuBinMax;nuBin++) { 
	  // Exclude edge bins
	  if (not nuBin->edge) nBin++;
	  nJ+=cellWork[i]->nJ[nuBin-cellWork[i]->nu.begin()]; 
	}
	// To be on the safe side
	nBin=MAX(nBin,1);
	
	// Calculate boost factor needed
	boost=(region->nJGood*nBin/MAX(nJ,1.0));
	cellWork[i]->boost[j]=MAX(MIN(cellWork[i]->boost[j]*boost,region->boostMax),region->boostMin);
	
	if (master) cout<<"mc-spec::adjustBoostFactor: cell: "<<i<<", lambdaBoost: ["<<(c/cellWork[i]->nuBoost[j].max)/AA<<":"<<(c/cellWork[i]->nuBoost[j].min)/AA<<"], boost: "<<boost<<", (new) boostFactor: "<<cellWork[i]->boost[j]<<", nJ: "<<nJ/nBin<<", neRadio: "<<cellWork[i]->neRadio<<"\n";
      }
      // Else - Set boost factor to one
      else cellWork[i]->boost[j]=1.0;
    }
    
    // If cell-container and sub-boost flag not set - New!
    if (cellWork[i]->type()==cellContainer and not subBoost) {
      // Copy boost factor for container to sub-cells 
      for (j=0;j<cellWork[i]->subCell.size();j++) cellWork[i]->subCell[j]->boost=cellWork[i]->boost;
    }
  }
}

// Subcells and containers: May not work yet!
void adjustBoostFactorERadio(GridSS& grid, int diffBorder, double neRadioGood, boost::mpi::communicator* mpiCom=NULL) {
  double boost,boostMean,neRadio,nJ;
  int i,j;
  double boostMax=1e5;
  double boostMin=0.25;
  double nJMax=1e7;
  bool master=false;

  if (mpiCom==NULL) master=true;
  else if (mpiCom->rank()==0) master=true;

  // For each cell
  for (i=diffBorder+1;i<grid.cell.size();i++) {
     
    // Get number of radioactive heating events
    neRadio=grid.cell[i]->neRadio;

    // Get number of mean intensity events
    nJ=0;
    for (j=0;j<grid.cell[i]->nJ.size();j++) nJ+=grid.cell[i]->nJ[j]; 
      
    // Calculate boost factor needed
    boost=MIN(neRadioGood/MAX(neRadio,1.0),nJMax/nJ);
    boostMean=0;
    for (j=0;j<grid.cell[i]->boost.size();j++) boostMean+=grid.cell[i]->boost[j]/grid.cell[i]->boost.size();
    for (j=0;j<grid.cell[i]->boost.size();j++) grid.cell[i]->boost[j]=MAX(MIN(boostMean*boost,boostMax),boostMin);
    
    //if (master) cout<<"mc-diff-evo::adjustBoostFactorGamma: cell: "<<i<<", boost: "<<boost<<", (new) boostFactor: "<<MAX(MIN(boostMean*boost,boostMax),boostMin)<<", nJ: "<<nJ<<", neRadio: "<<neRadio<<"\n";
  }
}

// Subcells and containers: May not work yet! - Not used - Remove!
/*void updateBoostFactor(Grid& grid, int cellIndexMin=0) {
  int i;

  //Update boost factors
  #pragma omp parallel for
  for (i=cellIndexMin;i<grid.cell.size();i++) grid.cell[i]->boostMC();
  }*/

// Log memory usage
void logMemory(GridSS& grid, DiffSolver& diffSolver, PacketStateList& packetIn, PacketStateList& packetOut, PacketStat& packetStat, boost::mpi::communicator* mpiCom=NULL) {
  double memPacketIn=0,memPacketOut=0,memPacketStat=0;
  int i;

  for (i=0;i<packetIn.size();i++) memPacketIn+=sizeof(PacketState);
  for (i=0;i<packetOut.size();i++) memPacketOut+=sizeof(PacketState);

  if (mpiCom!=NULL) cout<<"mc-diff-evo: Process: "<<mpiCom->rank()<<", Allocated memory: "<<(grid.mem()+diffSolver.mem()+memPacketIn+memPacketOut+packetStat.mem())/1e9<<" GB, grid: "<<grid.mem()/1e9<<" ("<<grid.memPList()/1e9<<","<<grid.memTransBB()/1e9<<","<<grid.memTransBF()/1e9<<") GB , diffSolver: "<<diffSolver.mem()/1e9<<" ("<<diffSolver.memPList()/1e9<<","<<diffSolver.memTransBB()/1e9<<","<<diffSolver.memTransBF()/1e9<<") GB, packet(In/Out): "<<memPacketIn/1e9<<"/"<<memPacketOut/1e9<<" GB, packetStat: "<<packetStat.mem()/1e9<<" GB\n";
  else cout<<"mc-diff-evo: Allocated memory: "<<(grid.mem()+diffSolver.mem()+memPacketIn+memPacketOut+packetStat.mem())/1e9<<" GB, grid: "<<grid.mem()/1e9<<" GB, diffSolver: "<<diffSolver.mem()/1e9<<" GB, packet(In/Out): "<<memPacketIn/1e9<<"/"<<memPacketOut/1e9<<" GB, packetStat: "<<packetStat.mem()/1e9<<" GB\n";
}

void logSysMemory(int n, int m, string label) {
  stringstream header;
  int status;

  header.str(""); 
  header<<"echo \"n: "<<n<<", m: "<<m<<", "<<label<<"\" >> mc-ps-$(hostname).log"; 
  status=system(header.str().c_str());
  status=system("date >> mc-ps-$(hostname).log");
  status=system("./mc-ps.sh >> mc-ps-$(hostname).log");
}

void logAlfa(Grid& grid, int diffBorder) {
  int i,j,k;
  for (i=diffBorder+1;i<grid.cell.size();i++) {
    for (j=0;j<grid.cell[i]->atom.size();j++) {
      for (k=0;k<grid.cell[i]->atom[j].iState.size();k++) {
	cout<<"mc-diff-evo::logAlfa: cell: "<<grid.cell[i]->index()<<", "<<grid.cell[i]->atom[j].iState[k].name()<<", alfaSPOriginal: "<<grid.cell[i]->atom[j].iState[k].alfaSPOriginal()<<", alfaSPTotal: "<<grid.cell[i]->atom[j].iState[k].alfaSPTotal()<<", alfaSP: "<<grid.cell[i]->atom[j].iState[k].alfaSP()<<", alfaAdd: "<<grid.cell[i]->atom[j].iState[k].alfaAdd()<<"\n";
      }
    }
  }
}

// Set absorption processes
Processes setAbsProcesses(bool electronScatter, bool boundFree, bool freeFree, bool boundBound, bool greyScatter) {
  unsigned int processesFlags=0;
  if (electronScatter) processesFlags|=pEScatter;
  if (boundFree) processesFlags|=pBoundFree;
  if (freeFree) processesFlags|=pFreeFree;
  if (boundBound) processesFlags|=pBoundBound;
  if (greyScatter) processesFlags|=pGScatter;
  return Processes(processesFlags);
}

// Evolve inner region using diffusion solver
// (and the outer region just dynamically)
void evolveDiff(Grid& grid, DiffSolver& diffSolver, double dt, int diffBorder, bool initiate=false, bool initiateState=true) {
  double tMin,tMax,dtDiff,dtLeft,TBorder,expFactor;

  // Backup time borders
  tMin=grid.tMin(); tMax=grid.tMax();

  // Handle diffusion solver timestep
  dtDiff=dt; dtLeft=dt;

  // If outer-boundary expansion correction - Backup outer border temperature
  if (diffSolver.config.outerBoundary.expCorr and (diffBorder+1)<grid.cell.size()) {
    if (diffSolver.config.outerBoundary.useTJ) TBorder=grid.cell[diffBorder+1]->TJ();
    else TBorder=grid.cell[diffBorder+1]->T();
  }

  // Do
  do { 
    // Evolve inner region using diffusion solver
    // (and the outer region just dynamically)
    diffSolver.evolve(&grid,dtDiff,diffBorder,initiate,initiateState);
    // If outer-boundary expansion correction - Update outer border temperature
    if (diffSolver.config.outerBoundary.expCorr and (diffBorder+1)<grid.cell.size()) { 
      expFactor=grid.t()/(grid.t()-dtDiff);
      if (diffSolver.config.outerBoundary.useTJ) grid.cell[diffBorder+1]->TJ(grid.cell[diffBorder+1]->TJ()/expFactor);
      else grid.cell[diffBorder+1]->T(grid.cell[diffBorder+1]->T()/expFactor);
    }
    dtLeft-=dtDiff; dtDiff=dtLeft; initiate=false; 
  }
  // While time left
  while (dtLeft>0);

  // If outer-boundary expansion correction - Restore outer border temperature
  if (diffSolver.config.outerBoundary.expCorr and (diffBorder+1)<grid.cell.size()) {
    if (diffSolver.config.outerBoundary.useTJ) grid.cell[diffBorder+1]->TJ(TBorder);
    else grid.cell[diffBorder+1]->T(TBorder);
  }
  
  // Update time borders
  updateTimeBorders(grid,dt,tMin,tMax);
}

// Clean up
void cleanUp(MCConfig* mCConfig, PacketConfig* packetConfig, EOS* eos, MStateSolver* mStateSolver, GridConfig* gridConfig, BStateSelector* bStateSelector, TransBBSelector* transSelector, CellConfig* cellConfig, LineSelector* lineSelector, SpecEvo* specEvoObs, boost::mpi::communicator* mpiCom, boost::mpi::environment* mpiEnv) {
  if (mCConfig!=NULL) delete mCConfig;
  if (packetConfig!=NULL) delete packetConfig;
  if (eos!=NULL) delete eos;
  if (mStateSolver!=NULL) delete mStateSolver;
  if (gridConfig!=NULL) delete gridConfig;
  if (bStateSelector!=NULL) delete bStateSelector;
  if (transSelector!=NULL) delete transSelector;
  if (cellConfig!=NULL) delete cellConfig;
  if (lineSelector!=NULL) delete lineSelector;
  if (specEvoObs!=NULL) delete specEvoObs;
  if (mpiCom!=NULL) delete mpiCom;
  if (mpiEnv!=NULL) delete mpiEnv;
}

// Main program
int main (int argc, char** argv) {

  PacketStateList packetIn(0); // Changed!
  PacketStateList packetOut(0); // Changed!
  bool neConverged,initiate,initiateMSolver,initiateDiffSolver,initiateState,updateMCDerived; // Changed!
  double E=0,EOuter=0,ENew=0,EPacket,Rdiff,ROnTheSpot,Ldiff;
  double t,tMin,tMax,dt,dtDiff,dtLeft,dlogt=0.01/2.5,expFactorWeight,tPrev,nIterLambdaUse; // Changed!
  int nPacket,diffBorder,diffBorderNew,diffBorderOld=-1,diffBorderMax,onTheSpotBorder,n,m,i,j; // Changed!
  double small=1e-10;
  FrequencyBinList nuList;
  vector<double> nuVector;
  vector<double> EPrev;
  vector<double> dEdt;
  PacketSamplerBBL* packetSamplerL=NULL;
  PacketSamplerBBThermal* packetSamplerThermal=NULL;
  MCExperiment* mCExperiment;
  bool injectPackets;
  boost::mpi::environment* mpiEnv=NULL;
  boost::mpi::communicator* mpiCom=NULL;
  bool master=true;
  Processes absProcesses;
  // Testing
  bool combSolverScanT;
  int combSolverNIter;
  double scanTFac;
  // Hack
  OnTheSpotNew onTheSpotNewInit;
  // Testing!
  double memPacketIn,memPacketOut;
  // Testing!
  double LRadioAbs;

  /* Check if enough number of arguments provided */
  if (argc<2) {
    printf("Missing argument\n");
    exit(-1);
  }
  
  /* Get mandatory arguments */
  string gridDataFile=string(argv[1]);

  /* Set default values for optional arguments */
  int nIterInit=10;
  int nIter=100;
  //int nIterLambda=1;
  double nIterLambda=1; // Changed!
  //int nIterLambdaInit=1;
  double nIterLambdaInit=1; // Changed!
  int nIterInitRad=0;
  int tStop=100*day;
  double nPacketEff=10000000; // Testing!
  double nPacketLdt=1000000; // Testing!
  double nPacketMax=std::numeric_limits<int>::max(); // New!
  //int nBin=10000;
  int nBin=5000; // Changed to be in sync with Cell::resetMCStatistics! Fix this!!! Fixed!
  int nMuBin=20; // New!
  double tauLim=1e-3;
  double neTol=1e-2;
  double lambdaMin=1;
  double lambdaMax=25000;
  double tauDiff=2e2;
  bool seed=false;
  bool eScatter=true;
  bool lineScatter=true;
  bool gScatter=false;
  bool freeFreeAbs=false;
  bool boundFreeAbs=false;
  bool flourescence=false;
  bool collProc=false;
  bool nonThermalProc=false;
  bool twoPhotonProc=false;
  bool chargeTransferProc=false;
  bool chargeTransferEmit=false;
  bool transTable=false;
  bool markovTable=false;
  bool markovTableExt=false;
  bool markovTableExtSplit=false;
  double markovTableLim=0.0;
  double markovTableThermLim=0.0;
  double markovTableCTLim=0.0;
  bool collExciteTable=false;
  double collExciteTableLim=0.0;  
  bool nonThermalExciteTable=false;
  bool etaTable=false;
  UpdateMethod updateMethod=urOnTheFly;
  bool boundFreeTable=false;
  double boundFreeTableMin=1;
  double boundFreeTableMax=1e5;
  int boundFreeTableNBin=1e4;
  bool boundFreeSTable=false;
  int boundFreeSTableNBin=1e3;
  RadEstimator radEstimator=rfDBB;
  RadEstimator radEstimatorBB=rfDBB;
  bool radEstimatorExp=false;
  BBSampling bbSampling=bbsEqualEnergy;
  BBFrame bbFrame=bbfMixed;
  bool bbLimbDark=true;
  OnTheSpot onTheSpot("");
  OnTheSpotNew onTheSpotNew("");
  double tauOnTheSpot=10;
  double tauEdgeOnTheSpot=0;
  FFOnTheSpot ffOnTheSpot(false,10.0);
  bool boost=false;
  bool subBoost=false;
  string boostConfigFile="";
  double nJGood=1000;
  double boostMin=0.25;
  double boostMax=1e5;
  bool alfaCorrect=false;
  bool alfaCorrectEmit=false;
  bool alfaSimpleEmit=false;
  bool alfaSPH=false;
  bool forceDBalance=false;
  bool forceDBalanceOSP=false;
  bool boundFreeBB=false;
  double boundFreeBBTauLim=1e3;
  double continuumBBTauLim=1e99;
  bool convertGPacket=true;
  bool transformKappa=false;
  bool bbEPacket=false;
  string configFile="";
  SolverMethod solverMethod=sNebular;
  bool includeRadio=false;
  bool exportEta=false;
  bool exportSBF=false;
  bool reflectOuter=false;
  bool reflectInnerGP=false;
  bool transformRay=false;
  bool limitFreeFreeCRange=false;
  string gridTemplateFile="";
  string gridStatDir="";
  string mCStatDir="";
  string gridStateDir="";
  string mCStateDir=""; // New!
  int saveStateStep=1;
  int saveIStep=1; // New!
  int saveMCStateStep=1; // New!
  string gridStateFile="";
  string mCStateFile=""; // New!
  bool exportIterInit=false;
  bool exportIterLambda=false;
  bool exportIterGuess=false;
  bool exportIterInitRad=false;
  bool JLineFallback=false;
  double JLineFallbackFactor=1.1;
  int JLineFallbackCount=3;
  bool JLineFallbackDBB=false;
  bool trackI=false;
  bool trackHK=false;
  //double scanTFacInit=2.0;
  double boundFreeCutOff=0.0;
  double boundFreeCutOffRoss=0.0;
  bool loadTransAll=true;
  bool mpi=false;
  double dtInit=-1;
  bool loadDBB=false;
  bool diffSolverInit=true;
  //EPacketMethod ePacketMethod=epLBased;
  EPacketMethod ePacketMethod=epEBased;
  bool dustAbs=false;
  int nCellOuterMin=1;
  double STPExpWeight=1.0;
  bool STPLCorr=false;
  bool STPERadioCorr=false;
  bool testSwitch2=false;
  bool expCoolingGuess=false;
  bool halfIterLambda=false;
  bool scaleRState=true;
  bool scaleMState=true;
  bool matterDerivedMPI=false;
  bool LTEGroupMacro=false;
  bool exportKappaMC2=true;
  bool loadDt=false;
  bool loadBoost=false;
  string gridKappaDir="";

  /* Get optional arguments */
  for (i=2;i<argc;i++) {
    if (strcmp(argv[i],"--niter_init")==0 && i+1<argc) 
      {i++; nIterInit=atoi(argv[i]);}
    else if (strcmp(argv[i],"--niter")==0 && i+1<argc) 
      {i++; nIter=atoi(argv[i]);}
    else if (strcmp(argv[i],"--niter_lambda")==0 && i+1<argc) 
      //{i++; nIterLambda=atoi(argv[i]);}
      {i++; nIterLambda=(float)strtod(argv[i],NULL);} // Changed!
    else if (strcmp(argv[i],"--niter_lambda_init")==0 && i+1<argc) 
      //{i++; nIterLambdaInit=atoi(argv[i]);}
      {i++; nIterLambdaInit=(float)strtod(argv[i],NULL);} // Changed!
    else if (strcmp(argv[i],"--niter_init_rad")==0 && i+1<argc) 
      {i++; nIterInitRad=atoi(argv[i]);}
    else if (strcmp(argv[i],"--t_stop")==0 && i+1<argc) 
      {i++; tStop=(float)strtod(argv[i],NULL)*day;}
    else if (strcmp(argv[i],"--npacket_eff")==0 && i+1<argc)
      // {i++; nPacketEff=atoi(argv[i]);}
      {i++; nPacketEff=(float)strtod(argv[i],NULL);}
    else if (strcmp(argv[i],"--npacket_Ldt")==0 && i+1<argc) // Testing!
      // {i++; nPacketLdt=atoi(argv[i]);}
      {i++; nPacketLdt=(float)strtod(argv[i],NULL);}
    else if (strcmp(argv[i],"--npacket_max")==0 && i+1<argc)
      {i++; nPacketMax=(float)strtod(argv[i],NULL);}
    else if (strcmp(argv[i],"--nbin")==0 && i+1<argc)
      {i++; nBin=atoi(argv[i]);}
    else if (strcmp(argv[i],"--taulim")==0 && i+1<argc)
      {i++; tauLim=(float)strtod(argv[i],NULL);}
    else if (strcmp(argv[i],"--netol")==0 && i+1<argc)
      {i++; neTol=(float)strtod(argv[i],NULL);}
    else if (strcmp(argv[i],"--lambda_min")==0 && i+1<argc)
      {i++; lambdaMin=(float)strtod(argv[i],NULL);}
    else if (strcmp(argv[i],"--lambda_max")==0 && i+1<argc)
      {i++; lambdaMax=(float)strtod(argv[i],NULL);}
    else if (strcmp(argv[i],"--taudiff")==0 && i+1<argc)
      {i++; tauDiff=(float)strtod(argv[i],NULL);}
    else if (strcmp(argv[i],"--seed")==0 && i+1<argc)
      {i++; if (strcmp(argv[i],"yes")) seed=true;} // Bug! Fix!
    else if (strcmp(argv[i],"--escatter")==0 && i+1<argc)
      {i++; if (strcmp(argv[i],"no")==0) eScatter=false;} // Bug fix!
    else if (strcmp(argv[i],"--linescatter")==0 && i+1<argc)
      {i++; if (strcmp(argv[i],"no")==0) lineScatter=false;}
    else if (strcmp(argv[i],"--gscatter")==0 && i+1<argc)
      {i++; if (strcmp(argv[i],"yes")==0) gScatter=true;}
    else if (strcmp(argv[i],"--freefree_abs")==0 && i+1<argc)
      {i++; if (strcmp(argv[i],"yes")==0) freeFreeAbs=true;}
    else if (strcmp(argv[i],"--boundfree_abs")==0 && i+1<argc)
      {i++; if (strcmp(argv[i],"yes")==0) boundFreeAbs=true;}
    else if (strcmp(argv[i],"--flourescence")==0 && i+1<argc)
      {i++; if (strcmp(argv[i],"yes")==0) flourescence=true;}
    else if (strcmp(argv[i],"--coll_proc")==0 && i+1<argc)
      {i++; if (strcmp(argv[i],"yes")==0) collProc=true;}
    else if (strcmp(argv[i],"--non_thermal_proc")==0 && i+1<argc)
      {i++; if (strcmp(argv[i],"yes")==0) nonThermalProc=true;}
    else if (strcmp(argv[i],"--two_photon_proc")==0 && i+1<argc)
      {i++; if (strcmp(argv[i],"yes")==0) twoPhotonProc=true;}
    else if (strcmp(argv[i],"--charge_transfer_proc")==0 && i+1<argc)
      {i++; if (strcmp(argv[i],"yes")==0) chargeTransferProc=true;}
    else if (strcmp(argv[i],"--charge_transfer_emit")==0 && i+1<argc)
      {i++; if (strcmp(argv[i],"yes")==0) chargeTransferEmit=true;}
    else if (strcmp(argv[i],"--trans_table")==0 && i+1<argc)
      {i++; if (strcmp(argv[i],"yes")==0) transTable=true;}
    else if (strcmp(argv[i],"--markov_table")==0 && i+1<argc)
      {i++; if (strcmp(argv[i],"yes")==0) markovTable=true;}
    else if (strcmp(argv[i],"--markov_table_ext")==0 && i+1<argc)
      {i++; if (strcmp(argv[i],"yes")==0) markovTableExt=true;}
    else if (strcmp(argv[i],"--markov_table_lim")==0 && i+1<argc) 
      {i++; markovTableLim=(float)strtod(argv[i],NULL);}
    else if (strcmp(argv[i],"--markov_table_ext_split")==0 && i+1<argc)
      {i++; if (strcmp(argv[i],"yes")==0) markovTableExtSplit=true;}
    else if (strcmp(argv[i],"--markov_table_therm_lim")==0 && i+1<argc) 
      {i++; markovTableThermLim=(float)strtod(argv[i],NULL);}
    else if (strcmp(argv[i],"--markov_table_ct_lim")==0 && i+1<argc) 
      {i++; markovTableCTLim=(float)strtod(argv[i],NULL);}
    else if (strcmp(argv[i],"--coll_exc_table")==0 && i+1<argc)
      {i++; if (strcmp(argv[i],"yes")==0) collExciteTable=true;}
    else if (strcmp(argv[i],"--coll_exc_table_lim")==0 && i+1<argc) 
      {i++; collExciteTableLim=(float)strtod(argv[i],NULL);}
    else if (strcmp(argv[i],"--non_thermal_exc_table")==0 && i+1<argc)
      {i++; if (strcmp(argv[i],"yes")==0) nonThermalExciteTable=true;}
    else if (strcmp(argv[i],"--eta_table")==0 && i+1<argc)
      {i++; if (strcmp(argv[i],"yes")==0) etaTable=true;}
    else if (strcmp(argv[i],"--boundfree_table")==0 && i+1<argc)
      {i++; if (strcmp(argv[i],"yes")==0) boundFreeTable=true;}
    else if (strcmp(argv[i],"--boundfree_table_min")==0 && i+1<argc)
      {i++; boundFreeTableMin==(float)strtod(argv[i],NULL);}
    else if (strcmp(argv[i],"--boundfree_table_max")==0 && i+1<argc)
      {i++; boundFreeTableMax=(float)strtod(argv[i],NULL);}
    else if (strcmp(argv[i],"--boundfree_table_nbin")==0 && i+1<argc)
      {i++; boundFreeTableNBin=atoi(argv[i]);}
    else if (strcmp(argv[i],"--boundfree_stable")==0 && i+1<argc)
      {i++; if (strcmp(argv[i],"yes")==0) boundFreeSTable=true;}
    else if (strcmp(argv[i],"--boundfree_stable_nbin")==0 && i+1<argc)
      {i++; boundFreeSTableNBin=atoi(argv[i]);}
    else if (strcmp(argv[i],"--rad_estimator")==0 && i+1<argc)
      { 
	i++;
	if (strcmp(argv[i],"lucy")==0) radEstimator=rfLucy;
	if (strcmp(argv[i],"bb")==0) radEstimator=rfBB;
      }
    else if (strcmp(argv[i],"--rad_estimator_bb")==0 && i+1<argc)
      {
	i++; 
	if (strcmp(argv[i],"lucy")==0) radEstimatorBB=rfLucy;
	if (strcmp(argv[i],"bb")==0) radEstimatorBB=rfBB;
      }
    else if (strcmp(argv[i],"--rad_estimator_exp")==0 && i+1<argc)
      {i++; if (strcmp(argv[i],"yes")==0) radEstimatorExp=true;}
    else if (strcmp(argv[i],"--bb_sampling")==0 && i+1<argc)
      {i++; if (strcmp(argv[i],"log_frequency")==0) bbSampling=bbsLogFrequency;}
    else if (strcmp(argv[i],"--bb_frame")==0 && i+1<argc)
      {
	i++; 
	if (strcmp(argv[i],"rest")==0) bbFrame=bbfRest;
	if (strcmp(argv[i],"cmf")==0) bbFrame=bbfCMF;
      }
    else if (strcmp(argv[i],"--bb_limb_dark")==0 && i+1<argc)
      {i++; if (strcmp(argv[i],"no")==0) bbLimbDark=false;}
    else if (strcmp(argv[i],"--onthespot")==0 && i+1<argc)
      {i++; onTheSpot=OnTheSpot(argv[i]);}
    else if (strcmp(argv[i],"--onthespot_new")==0 && i+1<argc)
      {i++; onTheSpotNew=OnTheSpotNew(argv[i]);}
    else if (strcmp(argv[i],"--tau_onthespot")==0 && i+1<argc)
      {i++; tauOnTheSpot=(float)strtod(argv[i],NULL);}
    else if (strcmp(argv[i],"--tau_edge_onthespot")==0 && i+1<argc)
      {i++; tauEdgeOnTheSpot=(float)strtod(argv[i],NULL);}
    else if (strcmp(argv[i],"--ff_onthespot")==0 && i+1<argc)
      {i++; if (strcmp(argv[i],"yes")==0) ffOnTheSpot.apply=true;}
    else if (strcmp(argv[i],"--ff_onthespot_taulim")==0 && i+1<argc)
      {i++; ffOnTheSpot.tauLim=(float)strtod(argv[i],NULL);}
    else if (strcmp(argv[i],"--boost")==0 && i+1<argc)
      {i++; if (strcmp(argv[i],"yes")==0) boost=true;}
    else if (strcmp(argv[i],"--sub_boost")==0 && i+1<argc)
      {i++; if (strcmp(argv[i],"yes")==0) subBoost=true;}
    else if (strcmp(argv[i],"--boost_config_file")==0 && i+1<argc)
      {i++; boostConfigFile=string(argv[i]);}
    else if (strcmp(argv[i],"--nJGood")==0 && i+1<argc)
      {i++; nJGood=(float)strtod(argv[i],NULL);}
    else if (strcmp(argv[i],"--boostMin")==0 && i+1<argc)
      {i++; boostMin=(float)strtod(argv[i],NULL);}
    else if (strcmp(argv[i],"--boostMax")==0 && i+1<argc)
      {i++; boostMax=(float)strtod(argv[i],NULL);}   
    else if (strcmp(argv[i],"--alfa_correct")==0 && i+1<argc)
      {i++; if (strcmp(argv[i],"yes")==0) alfaCorrect=true;}
    else if (strcmp(argv[i],"--alfa_correct_emit")==0 && i+1<argc)
      {i++; if (strcmp(argv[i],"yes")==0) alfaCorrectEmit=true;}
    else if (strcmp(argv[i],"--alfa_simple_emit")==0 && i+1<argc)
      {i++; if (strcmp(argv[i],"yes")==0) alfaSimpleEmit=true;}
    else if (strcmp(argv[i],"--alfa_SPH")==0 && i+1<argc)
      {i++; if (strcmp(argv[i],"yes")==0) alfaSPH=true;}
    else if (strcmp(argv[i],"--force_dbalance")==0 && i+1<argc)
      {i++; if (strcmp(argv[i],"yes")==0) forceDBalance=true;}
    else if (strcmp(argv[i],"--force_dbalance_osp")==0 && i+1<argc)
      {i++; if (strcmp(argv[i],"yes")==0) forceDBalanceOSP=true;}
    else if (strcmp(argv[i],"--convert_gpacket")==0 && i+1<argc)
      {i++; if (strcmp(argv[i],"no")==0) convertGPacket=false;}
    else if (strcmp(argv[i],"--transform_kappa")==0 && i+1<argc)
      {i++; if (strcmp(argv[i],"yes")==0) transformKappa=true;}
    else if (strcmp(argv[i],"--bb_epacket")==0 && i+1<argc)
      {i++; if (strcmp(argv[i],"yes")==0) bbEPacket=true;}
    else if (strcmp(argv[i],"--boundfree_bb")==0 && i+1<argc)
      {i++; if (strcmp(argv[i],"yes")==0) boundFreeBB=true;}
    else if (strcmp(argv[i],"--boundfree_bb_taulim")==0 && i+1<argc)
      {i++; boundFreeBBTauLim=(float)strtod(argv[i],NULL);}
    else if (strcmp(argv[i],"--continuum_bb_taulim")==0 && i+1<argc)
      {i++; continuumBBTauLim=(float)strtod(argv[i],NULL);}
    else if (strcmp(argv[i],"--config_file")==0 && i+1<argc)
      {i++; configFile=string(argv[i]);}
    else if (strcmp(argv[i],"--msolver_method")==0 && i+1<argc)
      {
	i++; 
	if (strcmp(argv[i],"nlte")==0) solverMethod=sNLTE;
	if (strcmp(argv[i],"comb")==0) solverMethod=sComb;
	if (strcmp(argv[i],"lte")==0) solverMethod=sLTE;	
      }
    else if (strcmp(argv[i],"--include_radio")==0 && i+1<argc)
      {i++; if (strcmp(argv[i],"yes")==0) includeRadio=true;}
    else if (strcmp(argv[i],"--export_eta")==0 && i+1<argc)
      {i++; if (strcmp(argv[i],"yes")==0) exportEta=true;}
    else if (strcmp(argv[i],"--export_sbf")==0 && i+1<argc)
      {i++; if (strcmp(argv[i],"yes")==0) exportSBF=true;}
    else if (strcmp(argv[i],"--reflect_outer")==0 && i+1<argc)
      {i++; if (strcmp(argv[i],"yes")==0) reflectOuter=true;}
    else if (strcmp(argv[i],"--reflect_inner_gp")==0 && i+1<argc)
      {i++; if (strcmp(argv[i],"yes")==0) reflectInnerGP=true;}
    else if (strcmp(argv[i],"--transform_ray")==0 && i+1<argc)
      {i++; if (strcmp(argv[i],"yes")==0) transformRay=true;}
    else if (strcmp(argv[i],"--limit_freefree_crange")==0 && i+1<argc)
      {i++; if (strcmp(argv[i],"yes")==0) limitFreeFreeCRange=true;}
    else if (strcmp(argv[i],"--grid_template_file")==0 && i+1<argc)
      {i++; gridTemplateFile=string(argv[i]);}
    else if (strcmp(argv[i],"--grid_stat_dir")==0 && i+1<argc)
      {i++; gridStatDir=string(argv[i]);}
    else if (strcmp(argv[i],"--mc_stat_dir")==0 && i+1<argc)
      {i++; mCStatDir=string(argv[i]);}
    else if (strcmp(argv[i],"--grid_state_dir")==0 && i+1<argc)
      {i++; gridStateDir=string(argv[i]);}
    else if (strcmp(argv[i],"--mc_state_dir")==0 && i+1<argc) // New!
      {i++; mCStateDir=string(argv[i]);}
    if (strcmp(argv[i],"--save_state_step")==0 && i+1<argc) 
      {i++; saveStateStep=atoi(argv[i]);}
    if (strcmp(argv[i],"--save_I_step")==0 && i+1<argc) // New!
      {i++; saveIStep=atoi(argv[i]);}
    if (strcmp(argv[i],"--save_mc_state_step")==0 && i+1<argc) // New!
      {i++; saveMCStateStep=atoi(argv[i]);}
    else if (strcmp(argv[i],"--grid_state_file")==0 && i+1<argc)
      {i++; gridStateFile=string(argv[i]);}
    else if (strcmp(argv[i],"--mc_state_file")==0 && i+1<argc)
      {i++; mCStateFile=string(argv[i]);}
    else if (strcmp(argv[i],"--export_iter_init")==0 && i+1<argc)
      {i++; if (strcmp(argv[i],"yes")==0) exportIterInit=true;}
    else if (strcmp(argv[i],"--export_iter_lambda")==0 && i+1<argc)
      {i++; if (strcmp(argv[i],"yes")==0) exportIterLambda=true;}
    else if (strcmp(argv[i],"--export_iter_guess")==0 && i+1<argc)
      {i++; if (strcmp(argv[i],"yes")==0) exportIterGuess=true;}
    else if (strcmp(argv[i],"--export_iter_init_rad")==0 && i+1<argc)
      {i++; if (strcmp(argv[i],"yes")==0) exportIterInitRad=true;}
    else if (strcmp(argv[i],"--jline_fallback")==0 && i+1<argc)
      {i++; if (strcmp(argv[i],"yes")==0) JLineFallback=true;}
    else if (strcmp(argv[i],"--jline_fallback_factor")==0 && i+1<argc)
      {i++; JLineFallbackFactor=(float)strtod(argv[i],NULL);}
    else if (strcmp(argv[i],"--jline_fallback_count")==0 && i+1<argc) 
      {i++; JLineFallbackCount=atoi(argv[i]);}
    else if (strcmp(argv[i],"--jline_fallback_dbb")==0 && i+1<argc)
      {i++; if (strcmp(argv[i],"yes")==0) JLineFallbackDBB=true;}
    else if (strcmp(argv[i],"--track_I")==0 && i+1<argc)
      {i++; if (strcmp(argv[i],"yes")==0) trackI=true;}
    else if (strcmp(argv[i],"--track_HK")==0 && i+1<argc)
      {i++; if (strcmp(argv[i],"yes")==0) trackHK=true;}
    //else if (strcmp(argv[i],"--scanTfac_init")==0 && i+1<argc)
    //  {i++; scanTFacInit=(float)strtod(argv[i],NULL);}
    else if (strcmp(argv[i],"--boundfree_cutoff")==0 && i+1<argc)
      {i++; boundFreeCutOff=(float)strtod(argv[i],NULL);}
    else if (strcmp(argv[i],"--boundfree_cutoff_ross")==0 && i+1<argc)
      {i++; boundFreeCutOffRoss=(float)strtod(argv[i],NULL);}
    else if (strcmp(argv[i],"--mpi")==0 && i+1<argc)
      {i++; if (strcmp(argv[i],"yes")==0) mpi=true;}
    else if (strcmp(argv[i],"--load_trans_all")==0 && i+1<argc)
      {i++; if (strcmp(argv[i],"no")==0) loadTransAll=false;}
    else if (strcmp(argv[i],"--dt_init")==0 && i+1<argc)
      {i++; dtInit=(float)strtod(argv[i],NULL)*day;}
    else if (strcmp(argv[i],"--load_dbb")==0 && i+1<argc)
      {i++; if (strcmp(argv[i],"yes")==0) loadDBB=true;}
    else if (strcmp(argv[i],"--dust_abs")==0 && i+1<argc)
      {i++; if (strcmp(argv[i],"yes")==0) dustAbs=true;}
    else if (strcmp(argv[i],"--diffsolver_init")==0 && i+1<argc)
      {i++; if (strcmp(argv[i],"no")==0) diffSolverInit=false;}
    else if (strcmp(argv[i],"--epacket_method")==0 && i+1<argc)
      {
	i++; 
	if (strcmp(argv[i],"e_based")==0) ePacketMethod=epEBased;
	if (strcmp(argv[i],"l_based")==0) ePacketMethod=epLBased;
      }
    else if (strcmp(argv[i],"--ncell_outer_min")==0 && i+1<argc)
      {i++; nCellOuterMin=atoi(argv[i]);}
    else if (strcmp(argv[i],"--stp_exp_weight")==0 && i+1<argc)
      {i++; STPExpWeight=(float)strtod(argv[i],NULL);}
    else if (strcmp(argv[i],"--stp_L_corr")==0 && i+1<argc)
      {i++; if (strcmp(argv[i],"yes")==0) STPLCorr=true;}
    else if (strcmp(argv[i],"--stp_eradio_corr")==0 && i+1<argc)
      {i++; if (strcmp(argv[i],"yes")==0) STPERadioCorr=true;}
    else if (strcmp(argv[i],"--test_switch_2")==0 && i+1<argc)
      {i++; if (strcmp(argv[i],"yes")==0) testSwitch2=true;}
    else if (strcmp(argv[i],"--exp_cooling_guess")==0 && i+1<argc)
      {i++; if (strcmp(argv[i],"yes")==0) expCoolingGuess=true;}
    else if (strcmp(argv[i],"--half_iter_lambda")==0 && i+1<argc)
      {i++; if (strcmp(argv[i],"yes")==0) halfIterLambda=true;}
    else if (strcmp(argv[i],"--scale_r_state")==0 && i+1<argc)
      {i++; if (strcmp(argv[i],"no")==0) scaleRState=false;}
    else if (strcmp(argv[i],"--scale_m_state")==0 && i+1<argc)
      {i++; if (strcmp(argv[i],"no")==0) scaleMState=false;}
    else if (strcmp(argv[i],"--matter_derived_mpi")==0 && i+1<argc)
      {i++; if (strcmp(argv[i],"yes")==0) matterDerivedMPI=true;}
    else if (strcmp(argv[i],"--lte_group_macro")==0 && i+1<argc)
      {i++; if (strcmp(argv[i],"yes")==0) LTEGroupMacro=true;}
    else if (strcmp(argv[i],"--export_kappa_mc2")==0 && i+1<argc)
      {i++; if (strcmp(argv[i],"no")==0) exportKappaMC2=false;}
    else if (strcmp(argv[i],"--load_dt")==0 && i+1<argc)
      {i++; if (strcmp(argv[i],"yes")==0) loadDt=true;}
    else if (strcmp(argv[i],"--load_boost")==0 && i+1<argc)
      {i++; if (strcmp(argv[i],"yes")==0) loadBoost=true;}
    else if (strcmp(argv[i],"--grid_kappa_dir")==0 && i+1<argc)
      {i++; gridKappaDir=string(argv[i]);}
  }

  // If MPI
  if (mpi) {
    mpiEnv=new boost::mpi::environment;
    mpiCom=new boost::mpi::communicator;
    // Determine if master
    if (mpiCom->rank()!=0) master=false;
  }

  // Allow nested openmp parallelization
  omp_set_nested(1);

  // Use this for gprof
  //omp_set_nested(0);
  //omp_set_dynamic(0);
  //omp_set_num_threads(1);

  // If seed requested
  if (seed) {
    // If MPI -Seed with time+pid
    if (mpi) randomize(0,1,time(NULL)+getpid());
    // Else - Seed with time (+ thead number)
    else randomize(0,1,time(NULL));
  }
  // Else
  else {
    // If MPI -Seed with pid
    if (mpi) randomize(0,1,getpid());
    // Else - Seed with zero (+ thread number)
    else randomize(0,1,0);
  }

  // Testing!
  logSysMemory(0,0,"I1");

  if (master) cout<<"mc-diff-evo: sizeof(BStateIDMPI): "<<sizeof(BStateIDMPI)<<", sizeof(TransIDMPI): "<<sizeof(TransIDMPI)<<", sizeof(TransPMPI): "<<sizeof(TransPMPI)<<", sizeof(MarkovPMPI): "<<sizeof(MarkovPMPI)<<"\n";
 
  // Configure grid
  FrequencyBin nu(c/(lambdaMax*AA),c/(lambdaMin*AA));
  LineSelector* lineSelector = new LineSelector(nu,tauLim);
  // Hack !!!
  onTheSpotNewInit=(gridStateFile.compare("")==0) ? onTheSpotNewInit=onTheSpotNew : onTheSpotNewInit=OnTheSpotNew("");
  CellConfig* cellConfig = new CellConfig(false,neIteration,lineSelector,updateMethod,boundFreeTable,boundFreeTableMin,boundFreeTableMax,boundFreeTableNBin,boundFreeSTable,boundFreeSTableNBin,radEstimator,radEstimatorBB,radEstimatorExp,transTable,markovTable,markovTableExt,markovTableExtSplit,markovTableLim,markovTableThermLim,markovTableCTLim,collExciteTable,collExciteTableLim,nonThermalExciteTable,freeFreeAbs,boundFreeAbs,collProc,nonThermalProc,twoPhotonProc,chargeTransferProc,onTheSpot,onTheSpotNewInit,ffOnTheSpot,false,boundFreeBB,continuumBBTauLim,boundFreeBBTauLim,bbTJBB,exportEta,transformRay,limitFreeFreeCRange,JLineFallback,JLineFallbackFactor,JLineFallbackCount,JLineFallbackDBB,trackI,trackHK,boundFreeCutOff,boundFreeCutOffRoss,boost,alfaCorrect,alfaCorrectEmit,alfaSimpleEmit,alfaSPH,forceDBalance,forceDBalanceOSP,chargeTransferEmit,etaTable,exportSBF,nu.min,nu.max,nBin,nMuBin,LTEGroupMacro,exportKappaMC2);
  BStateSelector* bStateSelector= new BStateSelector();
  TransBBSelector* transSelector;
  if (loadTransAll) transSelector = new TransBBAllSelector();
  else transSelector = new TransBBNoneSelector();
  GridConfig* gridConfig = new GridConfig(cellConfig,bStateSelector,transSelector,updateMethod,matterDerivedMPI);

  // Configure MC experiment
  PacketConfig* packetConfig = new PacketConfig(false,false,lineScatter,eScatter,false,gScatter,freeFreeAbs,boundFreeAbs,true,0,dustAbs,flourescence,collProc,nonThermalProc,twoPhotonProc,chargeTransferProc,transTable,updateMethod,boost,convertGPacket,-1,-1,true,1.0,reflectOuter,1.0,reflectInnerGP,transformKappa,bbEPacket); // Changed!
  MCConfig* mCConfig = new MCConfig(packetConfig);

  // Testing!
  if (testSwitch2) { packetConfig->lineScatter=false; packetConfig->freeFreeAbs=false; packetConfig->boundFreeAbs=false; packetConfig->dustAbs=false; }

  // Set absorption processes
  absProcesses=setAbsProcesses(packetConfig->electronScatter,packetConfig->boundFreeAbs,packetConfig->freeFreeAbs,packetConfig->lineScatter,packetConfig->greyScatter);

  // If boost - Load config from file
  BoostConfig boostConfig;
  if (boost) boostConfig.load(boostConfigFile);

  // Create matter state solver
  MStateSolver* mStateSolver;
  switch (solverMethod) {
  case sNLTE : { mStateSolver = new NLTESolver(configFile); break; }
  case sComb : { mStateSolver = new CombSolver(configFile); break; }
  case sLTE : { mStateSolver = new LTESolver(configFile); break; }
  deafult : { mStateSolver = new NebSolver(configFile); } }

  // Create EOS
  EOS* eos = new EOSBasic(configFile);

  // Create and configure diffusion solver
  DiffSolver diffSolver(configFile,mpiCom);
 
  // Turn on diffusion solver logging for master
  diffSolver.log(master);

  // Set diffusion solver radioactive heating flag
  diffSolver.config.includeRadio=includeRadio;
  // Set diffusion solver frequency limits
  diffSolver.config.nuMin=nu.min;
  diffSolver.config.nuMax=nu.max;

  // Create grid
  GridSS grid(mStateSolver,eos,gridConfig,mpiCom);
  
  // Load grid
  grid.load(gridDataFile); 

  // Calculate maximum diffusion border index - New!
  diffBorderMax=grid.cell.size()-nCellOuterMin-1;

  // Testing!
  logSysMemory(0,0,"I2");

  //if (mpiCom!=NULL) cout<<"mce-diff-evo: Process: "<<mpiCom->rank()<<", Reached here!\n";
  //else cout<<"mce-diff-evo: Reached here!\n";
  //exit(0);

  // Init grid frequency list
  // Not syncronized with Cell::resetMCStatistics!!! Fix!!!
  //grid.initNu(nu.min,nu.max,nBin);

  // Init grid frequency and angular binning
  grid.initNu();
  grid.initMu();

  // Create frequency bin list
  nuList=createNuList(nu,nBin);

  // Create frequency vector
  nuVector.resize(nBin+1); nuVector[0]=nuList[0].min;
  for (i=0;i<nuList.size();i++) nuVector[i+1]=nuList[i].max;
  
  // Update atomdata tables (using the provided binning) - New!
  grid.atomData.updateTables(nuVector);
  
  if (mpiCom!=NULL) mpiCom->barrier();

  // If requested - Load grid state
  if (gridStateFile.compare("")!=0) {
    
    grid.loadState(gridStateFile);
  }
  // Else
  else {
    
    // If DBB field not loaded (from the grid data file)
    if (not loadDBB) {
      // Initiate radiation state
      initRadiationState(grid);
    }
    
    // Initiate matter state
    //grid.updateMatterState(true);
  }

  // If requested - Load MC state
  if (mCStateFile.compare("")!=0 and master) { packetIn.load(mCStateFile); }

  // Testing!
  if (mpiCom==NULL) cout<<"mc-diff-evo: packetIn.size(1): "<<packetIn.size()<<"\n";
  else cout<<"mc-diff-evo: Process: "<<mpiCom->rank()<<", packetIn.size(1): "<<packetIn.size()<<"\n";

  // If requested - Distribute MC state
  if (mCStateFile.compare("")!=0) { packetIn.distribute(mpiCom); }

  // Testing!
  if (mpiCom==NULL) cout<<"mc-diff-evo: packetIn.size(2): "<<packetIn.size()<<"\n";
  else cout<<"mc-diff-evo: Process: "<<mpiCom->rank()<<", packetIn.size(2): "<<packetIn.size()<<"\n";

  // Determine timestep factor
  dlogt=log10(tStop/grid.t())/nIter;

  // Determine initial time step
  if (dtInit!=-1) dt=dtInit;
  else 
    if (gridStateFile.compare("")!=0 and loadDt) dt=(1.0-small)*(grid.tMax()-grid.tMin());
    else dt=grid.t()*(pow(10,dlogt)-1);

  // Set initial time borders
  initTimeBorders(grid,dt);

  // If grid-state not loaded
  if (gridStateFile.compare("")==0) {

    // If initiation using diffsolver requested
    if (diffSolverInit) {

      // Backup time borders - Hack !
      tMin=grid.tMin(); tMax=grid.tMax();
      
      // Evolve grid one step using diffusion solver
      if (mpiCom!=NULL) cout<<"mc-diff-evo: Process: "<<mpiCom->rank()<<", diffSolver.evolve 1, dt: "<<dt<<"\n";
      // Hack !!!
      DiffSolverConfig diffSolverConfig=diffSolver.config;
      diffSolver.config.convergence.nIter=10;
      diffSolver.config.timeStep.nIter=1;
      diffSolver.evolve(&grid,dt,grid.cell.size()-1,true);
      diffSolver.config=diffSolverConfig;
      
      if (master) cout<<"mc-diff-evo: 1\n";
      
      // Update time borders - Hack !
      updateTimeBorders(grid,dt,tMin,tMax);
    }
    // Else 
    else {
      // Update matter state assuming LTE
      updateMatterStateLTE(grid,configFile);
    }
  }

  if (master) cout<<"mc-diff-evo: 2\n";
    
  // Find diffusion border
  findBorder(grid,tauDiff,diffBorder,Rdiff,absProcesses,mpiCom);
  diffBorder=MIN(diffBorder,diffBorderMax); // New!
  // Testing!!!
  //diffBorder=58;
  //diffBorder=41;

  // Update on-the-spot region
  updateOnTheSpotRegion(grid,tauOnTheSpot,tauEdgeOnTheSpot,onTheSpot,onTheSpotNew,absProcesses,mpiCom); // Changed!

  // Load bound-bound transitions in outer region
  if (not loadTransAll) loadTransBB(grid,diffBorder+1,grid.cell.size()-1);

  // Set borders for MC experiment
  packetConfig->cellIndexMin=diffBorder+1;
  packetConfig->cellIndexMax=grid.cell.size()-1;

  // Get packet energy
  // Testing!
  //if (ePacketMethod==epEBased) 
  EPacket=getEPacket(grid,nPacketEff,tStop,includeRadio);
  //else EPacket=getEPacketL(grid,nPacketEff);
  // Testing!
  //else EPacket=getEPacketL(grid,nPacketEff,dt);

  if (master) cout<<"mc-diff-evo: EPacket: "<<EPacket<<", nPacketEff: "<<nPacketEff<<"\n";
   
  // Create packet statistics object
  PacketStat packetStat(&grid,diffBorder+1,grid.cell.size()-1,true,nuList);

  // Create spectral evolution object and add to packet statistics object
  SpecEvo* specEvoObs = new SpecEvo(&grid,nuList,rfDistantObserver,grid.t(),tStop,nIter);
  packetStat.specEvoObs = specEvoObs;

  if (mpiCom!=NULL) mpiCom->barrier();
  
  // Set initial expansion factor weight
  if (gridStateFile.compare("")!=0) expFactorWeight=0.5; 
  else expFactorWeight=1.0;

  // Number of boost bins - Testing!
  int nBoostBin=10;
  // Good number of mean intensity events (per bin)
  //double nJGood=1000;
  // Good number of radioactive heating events
  double neRadioGood=100;

  // If boost mode and not loaded - Initiate boost factor
  //if (boost) initBoostFactor(grid,nBoostBin,nu,mpiCom);
  if (boost and not (gridStateFile.compare("")!=0 and loadBoost)) initBoostFactor(grid,boostConfig,nu,mpiCom); // Changed!

  // Sample inital packets and add to packet list
  // Testing !!!
  if (nIterInitRad==0 and mCStateFile.compare("")==0) sampleInitialPackets(&grid,diffBorder+1,grid.cell.size()-1,EPacket,nu,packetIn,packetConfig,mpiCom,expFactorWeight);

  // Testing!
  if (mpiCom==NULL) cout<<"mc-diff-evo: packetIn.size(3): "<<packetIn.size()<<"\n";
  else cout<<"mc-diff-evo: Process: "<<mpiCom->rank()<<", packetIn.size(3): "<<packetIn.size()<<"\n";

  if (mpiCom!=NULL) mpiCom->barrier();

  // Create electron density vector
  vector<double> ne(grid.cell.size());

  if (master) cout<<"mc-diff-evo: Initial lambda-iteration, t: "<<grid.t()/day<<" days, tBorder: ["<<grid.tMin()/day<<","<<grid.tMax()/day<<"] days, diffusion border: "<<diffBorder<<"\n";

  // Initialize export counter
  int nExport=0;
  
  // Initial radiation field loop
  // Testing
  for (n=0;n<nIterInitRad;n++) {

    // Empty initial packet list - Hack!
    if (n==0) packetIn.resize(0);

    // If first iteration - Initiate non-thermal state
    if (n==0) initNTState(grid,configFile);
      
    // Update derived matter quantities (in outer region)
    grid.updateMatterDerived(diffBorder+1,grid.cell.size()-1);

    if (mpiCom!=NULL) mpiCom->barrier();

    // Evolve radiation state in outer region
    evolveRadiationState(&grid,diffBorder,diffBorderOld,EPacket,nu,nuList,includeRadio,packetIn,packetOut,&packetStat,mCConfig,mpiCom,true,nPacketMax);

    // If boost mode - Adjust boost factor based on number of radioactive heating events
    if (boost) adjustBoostFactorERadio(grid,diffBorder,neRadioGood,mpiCom);

    // Pick all active packets
    packetIn=packetOut;

    // Reset time for all active packets
    for (i=0;i<packetIn.size();i++) packetIn[i].t=grid.tMin();

    // Export statitics according to configuration
    if (master and exportIterInitRad) exportStat(grid,packetStat,gridStatDir,mCStatDir,nExport);
  }

  // Testing!
  logMemory(grid,diffSolver,packetIn,packetOut,packetStat,mpiCom);

  // Testing!
  logSysMemory(0,0,"I3");

  // Initial lambda-iteration loop
  for (n=0;n<nIterInit;n++) {

    initiate=(n==0);
    initiateState=(gridStateFile.compare("")==0);
    
    // Store electron density
    for (i=0;i<grid.cell.size();i++) ne[i]=grid.cell[i]->ne();

    // Testing !!!
    //grid.updateMatterState(59,59,initiate,initiateState);
    //cout<<"mc-diff-evo: Test Finished\n";
    //exit(-1);

    // Update matter state (in outer region)
    grid.updateMatterState(diffBorder+1,grid.cell.size()-1,initiate,initiateState);

    // Testing!!!
    //grid.updateMatterDerived(diffBorder+1,grid.cell.size()-1);

    // Export statitisc according to configuration
    if (master and exportIterInit) exportStat(grid,packetStat,gridStatDir,mCStatDir,nExport);

    // Check electron density convergence
    // Does not work with subcells and containers!
    neConverged=true;
    for (i=0;i<grid.cell.size();i++) {
      if (fabs((grid.cell[i]->ne()-ne[i])/grid.cell[i]->ne())>neTol) neConverged=false;
    }
    
    // Break if electron density converged
    if (n>0 and neConverged) break;
    
    if (mpiCom!=NULL) mpiCom->barrier();

    // Testing!
    logSysMemory(n,0,"IL1");
    
    // Track intensity only if state is going to be saved
    if (gridStateDir.compare("")!=0 and n==nIterInit-1) cellConfig->trackI=true; else cellConfig->trackI=false;

    // If not last lambda iteration - Lock spectral evolution for update
    if (m<(nIterInit-1)) packetStat.specEvoObs->lock(true);
    
    // Evolve radiation state in outer region
    evolveRadiationState(&grid,diffBorder,diffBorderOld,EPacket,nu,nuList,includeRadio,packetIn,packetOut,&packetStat,mCConfig,mpiCom,true,nPacketMax);

    // If not last lambda iteration - Unlock spectral evolution for update
    if (m<(nIterInit-1)) packetStat.specEvoObs->lock(false);

    // Testing!!!
    //exit(-1);

    if (mpiCom!=NULL) mpiCom->barrier();

    // Testing!
    logSysMemory(n,0,"IL2");

    // If boost mode - Adjust boost factor
    //if (boost) adjustBoostFactor(grid,diffBorder,nJGood,boostMin,boostMax,mpiCom);
    if (boost) adjustBoostFactor(grid,diffBorder,boostConfig,subBoost,mpiCom); // Changed!
  }

  // If no initial iterations - initiate matter-state solver
  //if (nIterInit==0) for (i=0;i<grid.cell.size();i++) grid.cell[i]->mStateSolver->init(cell[i]);
  
  // Update packet (state) list
  if (nIterInit>0 or nIterInitRad>0) packetIn=packetOut;

  // Testing!
  if (mpiCom==NULL) cout<<"mc-diff-evo: packetIn.size(3): "<<packetIn.size()<<"\n";
  else cout<<"mc-diff-evo: Process: "<<mpiCom->rank()<<", packetIn.size(3): "<<packetIn.size()<<"\n";

  // Evolution iteration loop
  for (n=0;n<nIter;n++) {

    //initiate=(n==0 and gridStateFile.compare("")!=0);
    initiateDiffSolver=(n==0 and (gridStateFile.compare("")!=0 or not diffSolverInit)); // Bug fix!
    initiateState=(n==0 and (gridStateFile.compare("")==0)); // Bug fix!
    initiateMSolver=(n==0 and nIterInit==0); // New!

    // Testing!
    LRadioAbs=0; for (i=0;i<grid.cell.size();i++) LRadioAbs+=grid.cell[i]->eRadio()*grid.cell[i]->M();
    
    if (master) cout<<"mc-diff-evo: Evolve grid, iteration: "<<n<<", t: "<<grid.t()/day<<" days, dt: "<<dt/day<<" days, tBorder: ["<<grid.tMin()/day<<","<<grid.tMax()/day<<"] days, diffusion border: "<<diffBorder<<", LRadio: "<<LRadio(&grid,grid.t())<<", LRadioAbs: "<<LRadioAbs<<"\n";
    // Testing!
    //if (mpiCom!=NULL) cout<<"mc-diff-evo: Process: "<<mpiCom->rank()<<", Evolve grid, iteration: "<<n<<", t: "<<grid.t()/day<<" days, dt: "<<dt/day<<" days, tBorder: ["<<grid.tMin()/day<<","<<grid.tMax()/day<<"] days, diffusion border: "<<diffBorder<<"\n";
 
    // Testing!
    logSysMemory(n,0,"M0");

    // Testing!
    logMemory(grid,diffSolver,packetIn,packetOut,packetStat,mpiCom);

    // Make sure bound-free opacity tables are updated - Testing!
    if (master) cout<<"mc-diff-evo::evolveRadiationState: Update bound-free opacity tables\n";
    //boundFreeTableUpdate(&grid,0,mpiCom);
    boundFreeTableUpdate(&grid,0,grid.cell.size()-1,mpiCom);
    if (mpiCom!=NULL) mpiCom->barrier();

    // Export statistics
    if (master) exportStat(grid,packetStat,gridStatDir,mCStatDir,nExport);

    // Save grid state
    if (gridStateDir.compare("")!=0 and (n%saveStateStep)==0 and master) grid.saveState(statFile(gridStateDir,"grid-state",n));

    // collect MC state
    if (mCStateDir.compare("")!=0 and (n%saveMCStateStep)==0) { cout<<"mc-diff-evo: collect MC state\n"; packetOut.collect(mpiCom); }

    // Save MC state
    if (mCStateDir.compare("")!=0 and (n%saveMCStateStep)==0 and master) { cout<<"mc-diff-evo: save MC state\n"; packetOut.save(statFile(mCStateDir,"mc-state",n)); }

    // Log recombination rates - Testing!
    if (master and alfaCorrect) logAlfa(grid,diffBorder);

    // Backup cell energy and time - New! - Testing!
    EPrev=getE(&grid,mpiCom); tPrev=grid.t();
   
    // Determine timestep
    dt=grid.t()*(pow(10,dlogt)-1); 

    // Testing!!!
    if (mpiCom!=NULL) mpiCom->barrier();
    //if (mpiCom!=NULL) { mpiCom->barrier(); cout<<"mc-diff-evo: Process: "<<mpiCom->rank()<<", point M1\n"; }

    // Testing!
    logSysMemory(n,0,"M1");

    // If required - Load grey scatter opacity - New!
    if (gridKappaDir.compare("")!=0 and gScatter) grid.loadGreyScatterKappa(statFile(gridKappaDir,"grid-kappa",n+1));

    // If diffusion border
    if (diffBorder>=0) {
    //if (false) { // Testing!!!

      // Evolve inner region using diffusion solver
      // (and the outer region just dynamically)
      evolveDiff(grid,diffSolver,dt,diffBorder,initiateDiffSolver,initiateState); // New!
    }
    // Else
    else {
      // Evolve dynamical state
      evolveDyn(grid,dt);

      // Evolve radioactive state
      evolveRadio(grid,dt,EPacket,packetIn,mCConfig,mpiCom);
    }

    // Testing!!!
    //cleanUp(mCConfig,packetConfig,eos,mStateSolver,gridConfig,bStateSelector,transSelector,cellConfig,lineSelector,specEvoObs,mpiCom,mpiEnv);
    //return 0;
    
    // Testing!!!
    if (mpiCom!=NULL) mpiCom->barrier();
    //if (mpiCom!=NULL) { mpiCom->barrier(); cout<<"mc-diff-evo: Process: "<<mpiCom->rank()<<", point M2\n"; }

    // Testing!
    logSysMemory(n,0,"M2");

    // Scale radiation field (in outer region) with expansion factor
    // Approximate solution. Lambda-iterations would be more correct.
    if (scaleRState) grid.scaleRadiationField(pow(10,dlogt),diffBorder+1,grid.cell.size()-1,expCoolingGuess); // Changed!
    
    // Turn on blackbody radiation field for optically thick wavelength regions
    cellConfig->continuumBB=true;
      
    // Scale matter state (in outer region) with expansion factor
    // Applies to quantities not automatically scaled when the grid is expanded
    if (scaleMState) scaleMatterState(grid,pow(10,dlogt),diffBorder+1,grid.cell.size()-1);

    // Update matter state (in outer region) - Excluding derived MC quantities
    grid.updateMatterState(diffBorder+1,grid.cell.size()-1,initiateMSolver,true,false); // Changed!
    
    // Testing!!!
    if (mpiCom!=NULL) mpiCom->barrier();
    //if (mpiCom!=NULL) { mpiCom->barrier(); cout<<"mc-diff-evo: Process: "<<mpiCom->rank()<<", point M3\n"; }

    // Get cell energy time derivative
    dEdt=getdEdt(&grid,EPrev,tPrev,mpiCom);

    // Testing!
    logSysMemory(n,0,"M3");

    // Export statistics according to configuration
    if (master and exportIterGuess) exportStat(grid,packetStat,gridStatDir,mCStatDir,nExport);

    // Update on-the-spot region
    updateOnTheSpotRegion(grid,tauOnTheSpot,tauEdgeOnTheSpot,onTheSpot,onTheSpotNew,absProcesses,mpiCom); // Changed!

    // Testing!!!
    if (mpiCom!=NULL) mpiCom->barrier();
    //if (mpiCom!=NULL) { mpiCom->barrier(); cout<<"mc-diff-evo: Process: "<<mpiCom->rank()<<", point M4\n"; }

    // Testing!
    logSysMemory(n,0,"M4");

    // Find new diffusion border - Changed!
    findBorder(grid,tauDiff,diffBorderNew,Rdiff,absProcesses,mpiCom);

    // Adjust packet energy
    // Testing!
    //if (ePacketMethod==epLBased) EPacket=getEPacketL(grid,nPacketEff);
    if (ePacketMethod==epLBased) EPacket=getEPacketL(grid,nPacketLdt,dt);

    // Testing!!!
    if (mpiCom!=NULL) mpiCom->barrier();    
    //if (mpiCom!=NULL) { mpiCom->barrier(); cout<<"mc-diff-evo: Process: "<<mpiCom->rank()<<", point M5\n"; }

    // Testing!
    logSysMemory(n,0,"M5");
    
    // If new diffusion border within previous
    if (diffBorderNew<diffBorder) {
	
      if (master) cout<<"mc-diff-evo: Move diffusion border from cell: "<<diffBorder<<" to cell "<<diffBorderNew<<"\n";
      // Testing 
      //if (mpiCom!=NULL) cout<<"mc-diff-evo: Process: "<<mpiCom->rank()<<", Move diffusion border from cell: "<<diffBorder<<" to cell "<<diffBorderNew<<"\n";

      // Load bound-bound transitions in the new region
      if (not loadTransAll) loadTransBB(grid,diffBorderNew+1,diffBorder);

      // Set boost factors to those of the previously innermost cell
      for (i=diffBorderNew+1;i<diffBorder+1;i++) grid.cell[i]->boost=grid.cell[diffBorder+1]->boost;

      // Sample thermal packets in new region and add to packet list
      sampleThermalPackets(&grid,diffBorderNew+1,diffBorder,EPacket,nu,packetIn,packetConfig,mpiCom);
      // Testing!!!
      //sampleThermalPackets(&grid,diffBorderNew+1,diffBorder,EPacket,nu,packetIn,packetConfig,mpiCom,STPExpWeight,STPLCorr,STPERadioCorr);
      //sampleThermalPackets(&grid,diffBorderNew+1,diffBorder,EPacket,nu,packetIn,packetConfig,mpiCom,dEdt);

      // Testing!
      if (mpiCom==NULL) cout<<"mc-diff-evo: packetIn.size(4): "<<packetIn.size()<<"\n";
      else cout<<"mc-diff-evo: Process: "<<mpiCom->rank()<<", packetIn.size(4): "<<packetIn.size()<<"\n";
      
      // Initiate matter state (in new region) - Excluding derived MC quantities
      grid.updateMatterState(diffBorderNew+1,diffBorder,true,true,false);

      // Testing!!!
      if (mpiCom!=NULL) mpiCom->barrier();
      //if (mpiCom!=NULL) { mpiCom->barrier(); cout<<"mc-diff-evo: Process: "<<mpiCom->rank()<<", point M6\n"; }

      // Testing!
      logSysMemory(n,0,"M6");

      // Store old diffusion border
      diffBorderOld=diffBorder;
	      
      // Set new diffusion border
      diffBorder=diffBorderNew;
      
      // Adjust inner border for MC experiment
      packetConfig->cellIndexMin=diffBorder+1;
    }

    // Get number of lamda iterations
    nIterLambdaUse=((n==0)?MAX(nIterLambdaInit,nIterLambda):nIterLambda);

    // Inner lambda-iteration loop
    //for (m=0;m<nIterLambdaUse;m++) {
    for (m=0;m<int(ceil(nIterLambdaUse));m++) { // Changed

      // Testing!!!
      if (mpiCom!=NULL) mpiCom->barrier();
      //if (mpiCom!=NULL) { mpiCom->barrier(); cout<<"mc-diff-evo: Process: "<<mpiCom->rank()<<", point L1\n"; }

      // Testing!
      logSysMemory(n,m,"L1");

      if (master) cout<<"mc-diff-evo: Lambda iteration "<<m<<"\n";

      // Turn off blackbody radiation field for optically thick regions
      cellConfig->continuumBB=false;

      // Track intensity only if it is going to be saved - Changed!
      if (gridStateDir.compare("")!=0 and ((n+1)%saveIStep)==0 and m==nIterLambda-1) cellConfig->trackI=true; else cellConfig->trackI=false;
     
      // Update derived matter quantities (in outer region)
      grid.updateMatterDerived(diffBorder+1,grid.cell.size()-1);

      // Testing!!!
      if (mpiCom!=NULL) mpiCom->barrier();
      //if (mpiCom!=NULL) { mpiCom->barrier(); cout<<"mc-diff-evo: Process: "<<mpiCom->rank()<<", point L2\n"; }

      // Testing!
      logSysMemory(n,m,"L2");

      // If not last lambda iteration - Lock spectral evolution for update
      if (m<(nIterLambdaUse-1)) packetStat.specEvoObs->lock(true);

      // Testing!
      if (mpiCom==NULL) cout<<"mc-diff-evo: packetIn.size(5): "<<packetIn.size()<<"\n";
      else cout<<"mc-diff-evo: Process: "<<mpiCom->rank()<<", packetIn.size(5): "<<packetIn.size()<<"\n";

      // Evolve radiation state in outer region
      evolveRadiationState(&grid,diffBorder,diffBorderOld,EPacket,nu,nuList,includeRadio,packetIn,packetOut,&packetStat,mCConfig,mpiCom,true,nPacketMax);

      // If not last lambda iteration - Unlock spectral evolution for update
      if (m<(nIterLambdaUse-1)) packetStat.specEvoObs->lock(false);

      // Testing!!!
      if (mpiCom!=NULL) mpiCom->barrier();
      //if (mpiCom!=NULL) { mpiCom->barrier(); cout<<"mc-diff-evo: Process: "<<mpiCom->rank()<<", point L3\n"; }

      // Testing!
      logSysMemory(n,m,"L3");

      // If boost mode - Adjust boost factor
      //if (boost) adjustBoostFactor(grid,diffBorder,nJGood,boostMin,boostMax,mpiCom);
      if (boost) adjustBoostFactor(grid,diffBorder,boostConfig,subBoost,mpiCom); // Changed!

      // If half lambda-iteration - Break - Testing!
      if (halfIterLambda or (nIterLambdaUse-m)==0.5) { if (master) cout<<"mc-diff-evo: Break lambda-iteration loop at "<<(m+0.5)<<" iterations\n"; break; } // Changed!

      // Update matter state (in outer region) - Excluding derived MC quantities
      grid.updateMatterState(diffBorder+1,grid.cell.size()-1,false,true,false);

      // Testing!!!
      if (mpiCom!=NULL) mpiCom->barrier();
      //if (mpiCom!=NULL) { mpiCom->barrier(); cout<<"mc-diff-evo: Process: "<<mpiCom->rank()<<", point L4\n"; }

      // Export statistics according to configuration
      if (master and exportIterLambda and m<(nIterLambdaUse-1)) exportStat(grid,packetStat,gridStatDir,mCStatDir,nExport);
    }

    // Update packet (state) list
    packetIn=packetOut;
  }

  // Clean up
  cleanUp(mCConfig,packetConfig,eos,mStateSolver,gridConfig,bStateSelector,transSelector,cellConfig,lineSelector,specEvoObs,mpiCom,mpiEnv);
  /*delete mCConfig;
  delete packetConfig;
  delete eos;
  delete mStateSolver;
  delete gridConfig;
  delete bStateSelector;
  delete transSelector;
  delete cellConfig;
  delete lineSelector;
  delete specEvoObs;
  if (mpiCom!=NULL) delete mpiCom;
  if (mpiEnv!=NULL) delete mpiEnv;*/
}
