#include <vector>
#include <math.h>
#include <string.h>
#include <unistd.h>
#include "mc-util.h"
#include "mc-math.h"
#include "mc-physics.h"
#include "mc.h"
#include "msolver.h"
#include "eos.h"
#include "diffsolver.h"

using namespace util;
using namespace math;
using namespace physics;
using namespace mc;
//using namespace grid;
using namespace msolver;
using namespace eos;
using namespace diffsolver;

string statFile(const string& statDir, const string& prefix, const int& index) {
  stringstream statFile;
  statFile.str("");
  statFile<<statDir<<"/"<<prefix<<"-"<<index<<".dat";
  return statFile.str();
}

// Initiate cell radiation state
// Subcells and containers: Works
void initRadiationStateCell(Cell* cell) {
  int i;

  // If container - Initiate radiation state for each subcell
  if (cell->type()==cellContainer) for (i=0;i<cell->subCell.size();i++) initRadiationStateCell(cell->subCell[i]);

  // Initialize to BB radiation field at matter temperature and reset
  cell->TR(cell->T());
  cell->W(1.0);
  cell->reset();
}

// Initiate grid radiation state
// Subcells and containers: Works
void initRadiationState(const Grid& grid) {
  int i;

  // Initiate radiation state for each cell
  for (i=0;i<grid.cell.size();i++) initRadiationStateCell(grid.cell[i]);
}

// Initiate cell time borders
// Subcells and containers: Works
void initTimeBordersCell(Cell* cell, const double& dt) {
  double t;
  int i;

  // If container - Set time borders for each subcell
  if (cell->type()==cellContainer) for (i=0;i<cell->subCell.size();i++) initTimeBordersCell(cell->subCell[i],dt);

  // Get time
  t=cell->t();

  // Set cell time borders
  cell->tMin(t-dt/2);
  cell->tMax(t+dt/2);
}

// Initiate grid time borders
// Subcells and containers: Works
void initTimeBorders(Grid& grid, const double& dt) {
  double t;
  int i;

  // Get time
  t=grid.t();

  // Set grid time borders
  grid.tMin(t-dt/2);
  grid.tMax(t+dt/2);

  // Initiate time borders for each cell
  for (i=0;i<grid.cell.size();i++) initTimeBordersCell(grid.cell[i],dt);
}

// Update cell time borders
// Subcells and containers: Works
void updateTimeBordersCell(Cell* cell, const double& dt, const double& tMin, const double& tMax) {
  double dtCenter,t=(tMin+tMax)/2;
  int i;

  // If container - Update time borders for each subcell
  if (cell->type()==cellContainer) for (i=0;i<cell->subCell.size();i++) updateTimeBordersCell(cell->subCell[i],dt,tMin,tMax);

  // Update cell time borders
  dtCenter=2*(t+dt-tMax);
  cell->tMin(tMax);
  cell->tMax(cell->tMin()+dtCenter);
}

// Update cell time borders
// Subcells and containers: Works
void updateTimeBorders(Grid& grid, const double& dt, const double& tMin, const double& tMax) {
  double dtCenter,t=(tMin+tMax)/2;
  int i;

  // Update grid time borders
  dtCenter=2*(t+dt-tMax);
  grid.tMin(tMax);
  grid.tMax(grid.tMin()+dtCenter);

  // Update time borders for each cell
  for (i=0;i<grid.cell.size();i++) updateTimeBordersCell(grid.cell[i],dt,tMin,tMax);
}

/*void initRadiationState(const Grid& grid) {
  int i;

  // Initialize to BB radiation field at matter temperature
  for (i=0;i<grid.cell.size();i++) {
    grid.cell[i]->TR(grid.cell[i]->T());
    grid.cell[i]->W(1.0);
    grid.cell[i]->reset();
  }
  }*/

/*void initTimeBorders(Grid& grid, const double& dt) {
  double t;
  int i;

  // Get time
  t=grid.t();

  // Set grid time borders
  grid.tMin(t-dt/2);
  grid.tMax(t+dt/2);

  // Set cell time borders
  for (i=0;i<grid.cell.size();i++) { 
    grid.cell[i]->tMin(t-dt/2); 
    grid.cell[i]->tMax(t+dt/2); 
  }
  }*/

FrequencyBinList createNuList(const FrequencyBin& nu, const int& nBin) {
  int i;
  vector<double> nuWork(nBin+1);
  FrequencyBinList nuList;
  
  double dlognu=(log(nu.max)-log(nu.min))/float(nBin);
  for (i=0;i<nuWork.size();i++) nuWork[i]=nu.min*exp(i*dlognu);
  nuList.resize(nBin);
  for (i=0;i<nuList.size();i++) nuList[i].min=nuWork[i];
  for (i=0;i<nuList.size();i++) nuList[i].max=nuWork[i+1];

  return nuList;
}

int main (int argc, char** argv) {

  bool neConverged,initiate;
  double t,dt,dlogt=0.01/2.5,tMin,tMax,dtDiff,dtLeft;
  FrequencyBinList nuList;
  vector<double> nuVector;
  int diffBorder,n,i;
  boost::mpi::environment* mpiEnv=NULL;
  boost::mpi::communicator* mpiCom=NULL;
  bool master=true;

  /* Check if enough number of arguments provided */
  if (argc<2) {
    printf("Missing argument\n");
    exit(-1);
  }
  
  /* Get mandatory arguments */
  string gridDataFile=string(argv[1]);

  /* Set default values for optional arguments */
  int nIter=100;
  int tStop=100*day;
  //int nBin=10000;
  int nBin=5000; // Changed!
  int nMuBin=20; // New!
  double tauLim=1e-3;
  double lambdaMin=1;
  double lambdaMax=25000;
  bool seed=false;
  bool freeFreeAbs=false;
  bool boundFreeAbs=false;
  bool collProc=false;
  bool nonThermalProc=false;
  bool twoPhotonProc=false;
  bool chargeTransferProc=false;
  bool chargeTransferEmit=false; // Not read!
  bool transTable=false; // Not read!
  bool markovTable=false; // Not read!
  bool markovTableExt=false; // Not read!
  bool markovTableExtSplit=false; // Not read!
  double markovTableLim=0.0; // Not read!
  double markovTableThermLim=0.0; // Not read!
  double markovTableCTLim=0.0; // Not read!
  bool collExciteTable=false; // Not read!
  double collExciteTableLim=0.0; // Not read!
  bool nonThermalExciteTable=false; // Not read!
  bool etaTable=false; // Not read!
  UpdateMethod updateMethod=urOnTheFly; // Not read!
  bool boundFreeTable=false;
  double boundFreeTableMin=1; // Not read!
  double boundFreeTableMax=1e5; // Not read!
  int boundFreeTableNBin=1e4; // Not read!
  bool boundFreeSTable=false; // Not read!
  int boundFreeSTableNBin=1e3; // Not read!
  RadEstimator radEstimator=rfDBB; // Not read!
  RadEstimator radEstimatorBB=rfDBB; // Not read!
  bool radEstimatorExp=false; // Not read!
  OnTheSpot onTheSpot(""); // Not read!
  OnTheSpotNew onTheSpotNew(""); // Not read!
  FFOnTheSpot ffOnTheSpot(false,10.0); // Not read!
  bool boost=false; // Not read!
  bool alfaCorrect=false;
  bool alfaCorrectEmit=false; // Not read!
  bool alfaSimpleEmit=false; // Not read!
  bool alfaSPH=false; // Not read
  bool forceDBalance=false;
  bool forceDBalanceOSP=false;
  bool boundFreeBB=false; // Not read!
  double boundFreeBBTauLim=1e3; // Not read!
  string configFile="";
  SolverMethod solverMethod=sNebular;
  bool includeRadio=false;
  bool exportEta=false; // Not read!
  bool exportSBF=false; // Not read!
  bool transformRay=false; // Not read!
  bool limitFreeFreeCRange=false;
  string gridStatDir="";
  string gridStateDir=""; // New!
  int saveStateStep=1; // New!
  bool JLineFallback=false; // Not read!
  double JLineFallbackFactor=1.1; // Not read!
  int JLineFallbackCount=3; // Not read!
  bool JLineFallbackDBB=false; // Not read!
  bool trackI=false; // Not read!
  bool trackHK=false; // Not read!
  double boundFreeCutOff=0.0;
  double boundFreeCutOffRoss=0.0;
  bool matterDerivedMPI=false;
  bool LTEGroupMacro=false; // Not read!
  bool exportKappaMC2=true;
  bool mpi=false;

  /* Get optional arguments */
  for (i=2;i<argc;i++) {
    if (strcmp(argv[i],"--niter")==0 && i+1<argc) 
      {i++; nIter=atoi(argv[i]);}
    else if (strcmp(argv[i],"--t_stop")==0 && i+1<argc) 
      {i++; tStop=(float)strtod(argv[i],NULL)*day;}
    else if (strcmp(argv[i],"--nbin")==0 && i+1<argc)
      {i++; nBin=atoi(argv[i]);}
    else if (strcmp(argv[i],"--taulim")==0 && i+1<argc)
      {i++; tauLim=(float)strtod(argv[i],NULL);}
    else if (strcmp(argv[i],"--lambda_min")==0 && i+1<argc)
      {i++; lambdaMin=(float)strtod(argv[i],NULL);}
    else if (strcmp(argv[i],"--lambda_max")==0 && i+1<argc)
      {i++; lambdaMax=(float)strtod(argv[i],NULL);}
    else if (strcmp(argv[i],"--seed")==0 && i+1<argc)
      {i++; if (strcmp(argv[i],"yes")) seed=true;}
    else if (strcmp(argv[i],"--freefree_abs")==0 && i+1<argc)
      {i++; if (strcmp(argv[i],"yes")==0) freeFreeAbs=true;}
    else if (strcmp(argv[i],"--boundfree_abs")==0 && i+1<argc)
      {i++; if (strcmp(argv[i],"yes")==0) boundFreeAbs=true;}
    else if (strcmp(argv[i],"--coll_proc")==0 && i+1<argc)
      {i++; if (strcmp(argv[i],"yes")==0) collProc=true;}
    else if (strcmp(argv[i],"--non_thermal_proc")==0 && i+1<argc)
      {i++; if (strcmp(argv[i],"yes")==0) nonThermalProc=true;}
    else if (strcmp(argv[i],"--two_photon_proc")==0 && i+1<argc)
      {i++; if (strcmp(argv[i],"yes")==0) twoPhotonProc=true;}
    else if (strcmp(argv[i],"--charge_transfer_proc")==0 && i+1<argc)
      {i++; if (strcmp(argv[i],"yes")==0) chargeTransferProc=true;}
    else if (strcmp(argv[i],"--boundfree_table")==0 && i+1<argc)
      {i++; if (strcmp(argv[i],"yes")==0) boundFreeTable=true;}
    else if (strcmp(argv[i],"--alfa_correct")==0 && i+1<argc)
      {i++; if (strcmp(argv[i],"yes")==0) alfaCorrect=true;}
    else if (strcmp(argv[i],"--force_dbalance")==0 && i+1<argc)
      {i++; if (strcmp(argv[i],"yes")==0) forceDBalance=true;}
    else if (strcmp(argv[i],"--force_dbalance_osp")==0 && i+1<argc)
      {i++; if (strcmp(argv[i],"yes")==0) forceDBalanceOSP=true;}
    else if (strcmp(argv[i],"--config_file")==0 && i+1<argc)
      {i++; configFile=string(argv[i]);}
    else if (strcmp(argv[i],"--msolver_method")==0 && i+1<argc)
      {
	i++; 
	if (strcmp(argv[i],"nlte")==0) solverMethod=sNLTE;
	if (strcmp(argv[i],"comb")==0) solverMethod=sComb;
	if (strcmp(argv[i],"lte")==0) solverMethod=sLTE;	
      }
    else if (strcmp(argv[i],"--include_radio")==0 && i+1<argc)
      {i++; if (strcmp(argv[i],"yes")==0) includeRadio=true;}
    else if (strcmp(argv[i],"--limit_freefree_crange")==0 && i+1<argc)
      {i++; if (strcmp(argv[i],"yes")==0) limitFreeFreeCRange=true;}
    else if (strcmp(argv[i],"--grid_stat_dir")==0 && i+1<argc)
      {i++; gridStatDir=string(argv[i]);}
    else if (strcmp(argv[i],"--grid_state_dir")==0 && i+1<argc)
      {i++; gridStateDir=string(argv[i]);}
    else if (strcmp(argv[i],"--save_state_step")==0 && i+1<argc) 
      {i++; saveStateStep=atoi(argv[i]);}
    else if (strcmp(argv[i],"--boundfree_cutoff")==0 && i+1<argc)
      {i++; boundFreeCutOff=(float)strtod(argv[i],NULL);}
    else if (strcmp(argv[i],"--boundfree_cutoff_ross")==0 && i+1<argc)
      {i++; boundFreeCutOffRoss=(float)strtod(argv[i],NULL);}
    else if (strcmp(argv[i],"--matter_derived_mpi")==0 && i+1<argc)
      {i++; if (strcmp(argv[i],"yes")==0) matterDerivedMPI=true;}
    else if (strcmp(argv[i],"--mpi")==0 && i+1<argc)
      {i++; if (strcmp(argv[i],"yes")==0) mpi=true;}
    else if (strcmp(argv[i],"--export_kappa_mc2")==0 && i+1<argc)
      {i++; if (strcmp(argv[i],"no")==0) exportKappaMC2=false;}
  }

  // If MPI
  if (mpi) {
    mpiEnv=new boost::mpi::environment;
    mpiCom=new boost::mpi::communicator;
    // Determine if master
    if (mpiCom->rank()!=0) master=false;
  }

  // Allow nested openmp parallelization - New!
  omp_set_nested(1);

  // Use this for gprof
  //omp_set_nested(0);
  //omp_set_dynamic(0);
  //omp_set_num_threads(1);

  // If seed requested
  if (seed) {
    // If MPI -Seed with time+pid
    if (mpi) randomize(0,1,time(NULL)+getpid());
    // Else - Seed with time (+ thead number)
    else randomize(0,1,time(NULL));
  }
  // Else
  else {
    // If MPI -Seed with pid
    if (mpi) randomize(0,1,getpid());
    // Else - Seed with zero (+ thread number)
    else randomize(0,1,0);
  }

  // If seed requested
  //if (seed) srand(time(NULL));
  // If seed requested - Seed with time (+ thead number)
  //if (seed) randomize(0,1,time(NULL)); // Changed!
  // Else - Seed with zero (+ thread number)
  //else randomize(0,1,0); // New!

  // Configure grid
  FrequencyBin nu(c/(lambdaMax*AA),c/(lambdaMin*AA));
  LineSelector* lineSelector = new LineSelector(nu,tauLim);
  CellConfig* cellConfig = new CellConfig(false,neIteration,lineSelector,updateMethod,boundFreeTable,boundFreeTableMin,boundFreeTableMax,boundFreeTableNBin,boundFreeSTable,boundFreeSTableNBin,radEstimator,radEstimatorBB,radEstimatorExp,transTable,markovTable,markovTableExt,markovTableExtSplit,markovTableLim,markovTableThermLim,markovTableCTLim,collExciteTable,collExciteTableLim,nonThermalExciteTable,freeFreeAbs,boundFreeAbs,collProc,nonThermalProc,twoPhotonProc,chargeTransferProc,onTheSpot,onTheSpotNew,ffOnTheSpot,false,boundFreeBB,1e3,boundFreeBBTauLim,bbTCur,exportEta,transformRay,limitFreeFreeCRange,JLineFallback,JLineFallbackFactor,JLineFallbackCount,JLineFallbackDBB,trackI,trackHK,boundFreeCutOff,boundFreeCutOffRoss,boost,alfaCorrect,alfaCorrectEmit,alfaSimpleEmit,alfaSPH,forceDBalance,forceDBalanceOSP,chargeTransferEmit,etaTable,exportSBF,nu.min,nu.max,nBin,nMuBin,LTEGroupMacro,exportKappaMC2);
  BStateSelector* bStateSelector= new BStateSelector();
  TransBBSelector* transSelector = new TransBBSelector(FrequencyBin(0,1e99));
  GridConfig* gridConfig = new GridConfig(cellConfig,bStateSelector,transSelector,updateMethod,matterDerivedMPI);

  // Create matter state solver
  MStateSolver* mStateSolver;
  switch (solverMethod) {
  case sNLTE : { mStateSolver = new NLTESolver(configFile); break; }
  case sComb : { mStateSolver = new CombSolver(configFile); break; }
  case sLTE : { mStateSolver = new LTESolver(configFile); break; }
  deafult : { mStateSolver = new NebSolver(configFile); } }

  // Create EOS
  EOS* eos = new EOSBasic(configFile);

  // Create and configure diffusion solver - Changed!
  DiffSolver diffSolver(configFile,mpiCom);

  // Turn on diffusion solver logging for master - New!
  diffSolver.log(master);

  // Set diffusion solver radioactive heating flag
  diffSolver.config.includeRadio=includeRadio;

  // Create grid
  GridSS grid(mStateSolver,eos,gridConfig);

  // Load grid
  grid.load(gridDataFile);

  // Init grid frequency list - New!
  grid.initNu(nu.min,nu.max,nBin);

  // Create frequency bin list - New!
  nuList=createNuList(nu,nBin);

  // Create frequency vector - New!
  nuVector.resize(nBin+1); nuVector[0]=nuList[0].min;
  for (i=0;i<nuList.size();i++) nuVector[i+1]=nuList[i].max;
  
  // Update atomdata tables (using the provided binning) - New!
  grid.atomData.updateTables(nuVector);
 
  // Initiate radiation state
  initRadiationState(grid);

  // Update matter state of entire grid
  grid.updateMatterState(true);

  // Set diffusion border
  diffBorder=grid.cell.size()-1;

  // Determine timestep factor
  dlogt=log10(tStop/grid.t())/nIter;

  // Determine initial time step
  dt=grid.t()*(pow(10,dlogt)-1);

  cout<<"mc-diff-only: dt: "<<dt<<", dlogt: "<<dlogt<<", tStop: "<<tStop<<", grid.t: "<<grid.t()<<"\n";

  // Set initial time borders
  initTimeBorders(grid,dt);
      
  // Evolution iteration loop
  for (n=0;n<nIter;n++) {

    initiate=(n==0);

    cout<<"mc-diff-only: Evolve grid, t: "<<grid.t()/day<<" days, dt: "<<dt/day<<" days, tBorder: ["<<grid.tMin()/day<<","<<grid.tMax()/day<<"]\n";

    // Export grid statistics - Changed!
    if (master and gridStatDir.compare("")!=0) grid.exportStat(statFile(gridStatDir,"grid-stat",n));

    // Save grid state - New!
    if (master and gridStateDir.compare("")!=0 and (n%saveStateStep)==0) grid.saveState(statFile(gridStateDir,"grid-state",n));
    
    // Determine timestep
    dt=grid.t()*(pow(10,dlogt)-1);

    // Backup time borders - Hack !
    tMin=grid.tMin(); tMax=grid.tMax();

    // Handle diffusion solver timestep - Hack !
    dtDiff=dt; dtLeft=dt;

    // Evolve inner region using diffusion solver
    // (and the outer region just dynamically)
    do { diffSolver.evolve(&grid,dtDiff,diffBorder,initiate); dtLeft-=dtDiff; dtDiff=dtLeft; initiate=false; }
    while (dtLeft>0);

    // Evolve inner region using diffusion solver
    // (and the entire grid dynamically)
    //diffSolver.evolve(&grid,dt,diffBorder,initiate);

    // Update time borders - Hack !
    updateTimeBorders(grid,dt,tMin,tMax);
  }

  // Clean up
  delete eos;
  delete mStateSolver;
  delete gridConfig;
  delete bStateSelector;
  delete transSelector;
  delete cellConfig;
  delete lineSelector;
}
