#include <vector>
#include <math.h>
#include <string.h>
#include <unistd.h>
#include "mc-util.h"
#include "mc-math.h"
#include "mc-physics.h"
#include "mc.h"
#include "msolver.h"
#include "eos.h"

using namespace util;
using namespace math;
using namespace physics;
using namespace mc;
//using namespace grid;
using namespace msolver;
using namespace eos;

string statFile(const string& statDir, const string& prefix, const int& index) {
  stringstream statFile;
  statFile.str("");
  statFile<<statDir<<"/"<<prefix<<"-"<<index<<".dat";
  return statFile.str();
}

void evolveDyn(Grid& grid, double& dt) {
  CellSS* cell;
  int i;
  double tOld,tNew;

  // For each cell
  for (i=0;i<grid.cell.size();i++) {

    // Set cell pointer
    cell=((CellSS*) grid.cell[i]);

    // Get old (centered) time
    tOld=(cell->tMin()+cell->tMax())/2;

    // Update time borders
    cell->tMin(cell->tMax());
    cell->tMax(cell->tMin()+dt);

    // Get new (centered) time
    tNew=(cell->tMin()+cell->tMax())/2;

    // Update inner and outer radial border
    cell->rMin(cell->rMin()*tNew/tOld);
    cell->rMax(cell->rMax()*tNew/tOld);

    // Update density
    cell->rho(cell->rho()*pow(tNew/tOld,-3));
  }
}

int main (int argc, char** argv) {

  PacketStateList packetIn(0);
  PacketStateList packetOut(0);
  bool neConverged,initiate;
  double R,tMin,tMax,neTol;
  double t,dt,dlogt=0.01/2.5;
  int n,i;
  boost::mpi::environment* mpiEnv=NULL;
  boost::mpi::communicator* mpiCom=NULL;
  bool master=true;

  /* Check if enough number of arguments provided */
  if (argc<2) {
    printf("Missing argument\n");
    exit(-1);
  }
  
  /* Get mandatory arguments */
  string gridDataFile=string(argv[1]);

  /* Set default values for optional arguments */
  int nIterInit=10;
  int nIter=100;
  int tStop=100*day;
  int nPacket=1000000;
  //int nBin=10000;
  int nBin=5000; // Changed!
  int nMuBin=20; // New!
  double tauLim=1e-3;
  double TTol=1e-2;
  double LTol=5e-2;
  double lambdaMin=1;
  double lambdaMax=25000;
  bool seed=false;
  bool eScatter=true;
  bool lineScatter=true;
  bool gScatter=false;
  bool freeFreeAbs=false;
  bool boundFreeAbs=false;
  bool flourescence=false;
  bool collProc=false;
  bool nonThermalProc=false;
  bool twoPhotonProc=false;
  bool chargeTransferProc=false;
  bool chargeTransferEmit=false;
  bool transTable=false;
  bool markovTable=false;
  bool markovTableExt=false;
  bool markovTableExtSplit=false;
  double markovTableLim=0.0;
  double markovTableThermLim=0.0;
  double markovTableCTLim=0.0;
  bool collExciteTable=false;
  double collExciteTableLim=0.0;
  bool nonThermalExciteTable=false;
  bool etaTable=false;
  UpdateMethod updateMethod=urOnTheFly;
  bool boundFreeTable=false;
  double boundFreeTableMin=1;
  double boundFreeTableMax=1e5;
  int boundFreeTableNBin=1e4;
  bool boundFreeSTable=false;
  int boundFreeSTableNBin=1e3;
  RadEstimator radEstimator=rfDBB;
  RadEstimator radEstimatorBB=rfDBB;
  bool radEstimatorExp=false;
  BBSampling bbSampling=bbsEqualEnergy;
  BBFrame bbFrame=bbfMixed;
  bool bbLimbDark=true;
  OnTheSpot onTheSpot("");
  OnTheSpotNew onTheSpotNew("");
  FFOnTheSpot ffOnTheSpot(false,10.0);
  bool boost=false;
  bool alfaCorrect=false;
  bool alfaCorrectEmit=false;
  bool alfaSimpleEmit=false;
  bool alfaSPH=false;
  bool forceDBalance=false;
  bool forceDBalanceOSP=false;
  bool boundFreeBB=false;
  double boundFreeBBTauLim=1e3;
  bool convertGPacket=true;
  bool transformKappa=false;
  bool bbEPacket=false;
  string configFile="";
  SolverMethod solverMethod=sNebular;
  bool exportEta=false;
  bool exportSBF=false;
  bool transformRay=false;
  bool limitFreeFreeCRange;
  string gridTemplateFile="";
  string gridStatDir="";
  string mCStatDir="";
  bool JLineFallback=false;
  double JLineFallbackFactor=1.1;
  int JLineFallbackCount=3;
  bool JLineFallbackDBB=false;
  bool trackI=false;
  bool trackHK=false;
  double boundFreeCutOff=0.0;
  double boundFreeCutOffRoss=0.0;
  bool matterDerivedMPI=false;
  bool LTEGroupMacro=false;
  bool exportKappaMC2=true;
  bool mpi=false;
  bool dustAbs=false;

  /* Get optional arguments */
  for (i=2;i<argc;i++) {
    if (strcmp(argv[i],"--niter_init")==0 && i+1<argc) 
      {i++; nIterInit=atoi(argv[i]);}
    else if (strcmp(argv[i],"--niter")==0 && i+1<argc) 
      {i++; tStop=atoi(argv[i]);}
    else if (strcmp(argv[i],"--t_stop")==0 && i+1<argc) 
      {i++; tStop=atoi(argv[i]);}
    else if (strcmp(argv[i],"--npacket")==0 && i+1<argc)
      {i++; nPacket=atoi(argv[i]);}
    else if (strcmp(argv[i],"--nbin")==0 && i+1<argc)
      {i++; nBin=atoi(argv[i]);}
    else if (strcmp(argv[i],"--taulim")==0 && i+1<argc)
      {i++; tauLim=(float)strtod(argv[i],NULL);}
    else if (strcmp(argv[i],"--ltol")==0 && i+1<argc)
      {i++; LTol=(float)strtod(argv[i],NULL);}
    else if (strcmp(argv[i],"--ttol")==0 && i+1<argc)
      {i++; TTol=(float)strtod(argv[i],NULL);}
    else if (strcmp(argv[i],"--lambda_min")==0 && i+1<argc)
      {i++; lambdaMin=(float)strtod(argv[i],NULL);}
    else if (strcmp(argv[i],"--lambda_max")==0 && i+1<argc)
      {i++; lambdaMax=(float)strtod(argv[i],NULL);}
    else if (strcmp(argv[i],"--seed")==0 && i+1<argc)
      {i++; if (strcmp(argv[i],"yes")) seed=true;}
    else if (strcmp(argv[i],"--escatter")==0 && i+1<argc)
      {i++; if (strcmp(argv[i],"yes")==0) eScatter=true;}
    else if (strcmp(argv[i],"--linescatter")==0 && i+1<argc)
      {i++; if (strcmp(argv[i],"no")==0) lineScatter=false;}
    else if (strcmp(argv[i],"--gscatter")==0 && i+1<argc)
      {i++; if (strcmp(argv[i],"yes")==0) gScatter=true;}
    else if (strcmp(argv[i],"--freefree_abs")==0 && i+1<argc)
      {i++; if (strcmp(argv[i],"yes")==0) freeFreeAbs=true;}
    else if (strcmp(argv[i],"--boundfree_abs")==0 && i+1<argc)
      {i++; if (strcmp(argv[i],"yes")==0) boundFreeAbs=true;}
    else if (strcmp(argv[i],"--flourescence")==0 && i+1<argc)
      {i++; if (strcmp(argv[i],"yes")==0) flourescence=true;}
    else if (strcmp(argv[i],"--coll_proc")==0 && i+1<argc)
      {i++; if (strcmp(argv[i],"yes")==0) collProc=true;}
    else if (strcmp(argv[i],"--non_thermal_proc")==0 && i+1<argc)
      {i++; if (strcmp(argv[i],"yes")==0) nonThermalProc=true;}
    else if (strcmp(argv[i],"--two_photon_proc")==0 && i+1<argc)
      {i++; if (strcmp(argv[i],"yes")==0) twoPhotonProc=true;}
    else if (strcmp(argv[i],"--charge_transfer_proc")==0 && i+1<argc)
      {i++; if (strcmp(argv[i],"yes")==0) chargeTransferProc=true;}
    else if (strcmp(argv[i],"--charge_transfer_emit")==0 && i+1<argc)
      {i++; if (strcmp(argv[i],"yes")==0) chargeTransferEmit=true;}
    else if (strcmp(argv[i],"--trans_table")==0 && i+1<argc)
      {i++; if (strcmp(argv[i],"yes")==0) transTable=true;}
    else if (strcmp(argv[i],"--markov_table")==0 && i+1<argc)
      {i++; if (strcmp(argv[i],"yes")==0) markovTable=true;}
    else if (strcmp(argv[i],"--markov_table_ext")==0 && i+1<argc)
      {i++; if (strcmp(argv[i],"yes")==0) markovTableExt=true;}
    else if (strcmp(argv[i],"--markov_table_lim")==0 && i+1<argc) 
      {i++; markovTableLim=(float)strtod(argv[i],NULL);}
    else if (strcmp(argv[i],"--markov_table_ext_split")==0 && i+1<argc)
      {i++; if (strcmp(argv[i],"yes")==0) markovTableExtSplit=true;}
    else if (strcmp(argv[i],"--markov_table_therm_lim")==0 && i+1<argc) 
      {i++; markovTableThermLim=(float)strtod(argv[i],NULL);}
    else if (strcmp(argv[i],"--markov_table_ct_lim")==0 && i+1<argc) 
      {i++; markovTableCTLim=(float)strtod(argv[i],NULL);}
    else if (strcmp(argv[i],"--coll_exc_table")==0 && i+1<argc)
      {i++; if (strcmp(argv[i],"yes")==0) collExciteTable=true;}
    else if (strcmp(argv[i],"--coll_exc_table_lim")==0 && i+1<argc) 
      {i++; collExciteTableLim=(float)strtod(argv[i],NULL);}
    else if (strcmp(argv[i],"--non_thermal_exc_table")==0 && i+1<argc)
      {i++; if (strcmp(argv[i],"yes")==0) nonThermalExciteTable=true;}
    else if (strcmp(argv[i],"--eta_table")==0 && i+1<argc)
      {i++; if (strcmp(argv[i],"yes")==0) etaTable=true;}
    else if (strcmp(argv[i],"--boundfree_table")==0 && i+1<argc)
      {i++; if (strcmp(argv[i],"yes")==0) boundFreeTable=true;}
    else if (strcmp(argv[i],"--boundfree_table_min")==0 && i+1<argc)
      {i++; boundFreeTableMin==(float)strtod(argv[i],NULL);}
    else if (strcmp(argv[i],"--boundfree_table_max")==0 && i+1<argc)
      {i++; boundFreeTableMax=(float)strtod(argv[i],NULL);}
    else if (strcmp(argv[i],"--boundfree_table_nbin")==0 && i+1<argc)
      {i++; boundFreeTableNBin=atoi(argv[i]);}
    else if (strcmp(argv[i],"--boundfree_stable")==0 && i+1<argc)
      {i++; if (strcmp(argv[i],"yes")==0) boundFreeSTable=true;}
    else if (strcmp(argv[i],"--boundfree_stable_nbin")==0 && i+1<argc)
      {i++; boundFreeSTableNBin=atoi(argv[i]);}
    else if (strcmp(argv[i],"--rad_estimator")==0 && i+1<argc)
      {i++; if (strcmp(argv[i],"lucy")==0) radEstimator=rfLucy;}
    else if (strcmp(argv[i],"--rad_estimator_bb")==0 && i+1<argc)
      {i++; if (strcmp(argv[i],"lucy")==0) radEstimatorBB=rfLucy;}
    else if (strcmp(argv[i],"--rad_estimator_exp")==0 && i+1<argc)
      {i++; if (strcmp(argv[i],"yes")==0) radEstimatorExp=true;}
    else if (strcmp(argv[i],"--bb_sampling")==0 && i+1<argc)
      {i++; if (strcmp(argv[i],"log_frequency")==0) bbSampling=bbsLogFrequency;}
    else if (strcmp(argv[i],"--bb_frame")==0 && i+1<argc)
      {
	i++; 
	if (strcmp(argv[i],"rest")==0) bbFrame=bbfRest;
	if (strcmp(argv[i],"cmf")==0) bbFrame=bbfCMF;
      }
    else if (strcmp(argv[i],"--bb_limb_dark")==0 && i+1<argc)
      {i++; if (strcmp(argv[i],"no")==0) bbLimbDark=false;}
    else if (strcmp(argv[i],"--onthespot")==0 && i+1<argc)
      {i++; onTheSpot=OnTheSpot(argv[i]);}
    else if (strcmp(argv[i],"--onthespot_new")==0 && i+1<argc)
      {i++; onTheSpotNew=OnTheSpotNew(argv[i]);}
    else if (strcmp(argv[i],"--ff_onthespot")==0 && i+1<argc)
      {i++; if (strcmp(argv[i],"yes")==0) ffOnTheSpot.apply=true;}
    else if (strcmp(argv[i],"--ff_onthespot_taulim")==0 && i+1<argc)
      {i++; ffOnTheSpot.tauLim=(float)strtod(argv[i],NULL);}
    else if (strcmp(argv[i],"--boost")==0 && i+1<argc)
      {i++; if (strcmp(argv[i],"yes")==0) boost=true;}
    else if (strcmp(argv[i],"--alfa_correct")==0 && i+1<argc)
      {i++; if (strcmp(argv[i],"yes")==0) alfaCorrect=true;}
    else if (strcmp(argv[i],"--alfa_correct_emit")==0 && i+1<argc)
      {i++; if (strcmp(argv[i],"yes")==0) alfaCorrectEmit=true;}
    else if (strcmp(argv[i],"--alfa_simple_emit")==0 && i+1<argc)
      {i++; if (strcmp(argv[i],"yes")==0) alfaSimpleEmit=true;}
    else if (strcmp(argv[i],"--alfa_SPH")==0 && i+1<argc)
      {i++; if (strcmp(argv[i],"yes")==0) alfaSPH=true;}
    else if (strcmp(argv[i],"--force_dbalance")==0 && i+1<argc)
      {i++; if (strcmp(argv[i],"yes")==0) forceDBalance=true;}
    else if (strcmp(argv[i],"--force_dbalance_osp")==0 && i+1<argc)
      {i++; if (strcmp(argv[i],"yes")==0) forceDBalanceOSP=true;}
    else if (strcmp(argv[i],"--convert_gpacket")==0 && i+1<argc)
      {i++; if (strcmp(argv[i],"no")==0) convertGPacket=false;}
    else if (strcmp(argv[i],"--transform_kappa")==0 && i+1<argc)
      {i++; if (strcmp(argv[i],"yes")==0) transformKappa=true;}
    else if (strcmp(argv[i],"--bb_epacket")==0 && i+1<argc)
      {i++; if (strcmp(argv[i],"yes")==0) bbEPacket=true;}
    else if (strcmp(argv[i],"--boundfree_bb")==0 && i+1<argc)
      {i++; if (strcmp(argv[i],"yes")==0) boundFreeBB=true;}
    else if (strcmp(argv[i],"--boundfree_bb_taulim")==0 && i+1<argc)
      {i++; boundFreeBBTauLim=(float)strtod(argv[i],NULL);}
    else if (strcmp(argv[i],"--config_file")==0 && i+1<argc)
      {i++; configFile=string(argv[i]);}
    else if (strcmp(argv[i],"--msolver_method")==0 && i+1<argc)
      {
	i++; 
	if (strcmp(argv[i],"nlte")==0) solverMethod=sNLTE;
	if (strcmp(argv[i],"comb")==0) solverMethod=sComb;
	if (strcmp(argv[i],"lte")==0) solverMethod=sLTE;	
      } 
    else if (strcmp(argv[i],"--export_eta")==0 && i+1<argc)
      {i++; if (strcmp(argv[i],"yes")==0) exportEta=true;}
    else if (strcmp(argv[i],"--export_sbf")==0 && i+1<argc)
      {i++; if (strcmp(argv[i],"yes")==0) exportSBF=true;}
    else if (strcmp(argv[i],"--transform_ray")==0 && i+1<argc)
      {i++; if (strcmp(argv[i],"yes")==0) transformRay=true;}
    else if (strcmp(argv[i],"--limit_freefree_crange")==0 && i+1<argc)
      {i++; if (strcmp(argv[i],"yes")==0) limitFreeFreeCRange=true;}
    else if (strcmp(argv[i],"--grid_template_file")==0 && i+1<argc)
      {i++; gridTemplateFile=string(argv[i]);}
    else if (strcmp(argv[i],"--grid_stat_dir")==0 && i+1<argc)
      {i++; gridStatDir=string(argv[i]);}
    else if (strcmp(argv[i],"--mc_stat_dir")==0 && i+1<argc)
      {i++; mCStatDir=string(argv[i]);}
    else if (strcmp(argv[i],"--mpi")==0 && i+1<argc)
      {i++; if (strcmp(argv[i],"yes")==0) mpi=true;}
    else if (strcmp(argv[i],"--jline_fallback")==0 && i+1<argc)
      {i++; if (strcmp(argv[i],"yes")==0) JLineFallback=true;}
    else if (strcmp(argv[i],"--jline_fallback_factor")==0 && i+1<argc)
      {i++; JLineFallbackFactor=(float)strtod(argv[i],NULL);}
    else if (strcmp(argv[i],"--jline_fallback_count")==0 && i+1<argc) 
      {i++; JLineFallbackCount=atoi(argv[i]);}
    else if (strcmp(argv[i],"--jline_fallback_dbb")==0 && i+1<argc)
      {i++; if (strcmp(argv[i],"yes")==0) JLineFallbackDBB=true;}
    else if (strcmp(argv[i],"--track_I")==0 && i+1<argc)
      {i++; if (strcmp(argv[i],"yes")==0) trackI=true;}
    else if (strcmp(argv[i],"--track_HK")==0 && i+1<argc)
      {i++; if (strcmp(argv[i],"yes")==0) trackHK=true;}
    else if (strcmp(argv[i],"--boundfree_cutoff")==0 && i+1<argc)
      {i++; boundFreeCutOff=(float)strtod(argv[i],NULL);}
    else if (strcmp(argv[i],"--boundfree_cutoff_ross")==0 && i+1<argc)
      {i++; boundFreeCutOffRoss=(float)strtod(argv[i],NULL);}
    else if (strcmp(argv[i],"--matter_derived_mpi")==0 && i+1<argc)
      {i++; if (strcmp(argv[i],"yes")==0) matterDerivedMPI=true;}
    else if (strcmp(argv[i],"--lte_group_macro")==0 && i+1<argc)
      {i++; if (strcmp(argv[i],"yes")==0) LTEGroupMacro=true;}
    else if (strcmp(argv[i],"--export_kappa_mc2")==0 && i+1<argc)
      {i++; if (strcmp(argv[i],"no")==0) exportKappaMC2=false;}
  }

  // If MPI
  if (mpi) {
    mpiEnv=new boost::mpi::environment;
    mpiCom=new boost::mpi::communicator;
    // Determine if master
    if (mpiCom->rank()!=0) master=false;
  }

  // If seed requested
  if (seed) {
    // If MPI -Seed with time+pid
    if (mpi) srand(time(NULL)+getpid());
    // Else - Seed with time
    else srand(time(NULL));
  }
  // Else - Seed with pid
  else if (mpi) srand(getpid());

  // Configure grid
  FrequencyBin nu(c/(lambdaMax*AA),c/(lambdaMin*AA));
  LineSelector* lineSelector = new LineSelector(nu,tauLim);
  CellConfig* cellConfig = new CellConfig(false,neIteration,lineSelector,updateMethod,boundFreeTable,boundFreeTableMin,boundFreeTableMax,boundFreeTableNBin,boundFreeSTable,boundFreeSTableNBin,radEstimator,radEstimatorBB,radEstimatorExp,transTable,markovTable,markovTableExt,markovTableExtSplit,markovTableLim,markovTableThermLim,markovTableCTLim,collExciteTable,collExciteTableLim,nonThermalExciteTable,freeFreeAbs,boundFreeAbs,collProc,nonThermalProc,twoPhotonProc,chargeTransferProc,onTheSpot,onTheSpotNew,ffOnTheSpot,false,boundFreeBB,1e3,boundFreeBBTauLim,bbTCur,exportEta,transformRay,limitFreeFreeCRange,JLineFallback,JLineFallbackFactor,JLineFallbackCount,JLineFallbackDBB,trackI,trackHK,boundFreeCutOff,boundFreeCutOffRoss,boost,alfaCorrect,alfaCorrectEmit,alfaSimpleEmit,alfaSPH,forceDBalance,forceDBalanceOSP,chargeTransferEmit,etaTable,exportSBF,nu.min,nu.max,nBin,nMuBin,LTEGroupMacro,exportKappaMC2);
  BStateSelector* bStateSelector= new BStateSelector();
  TransBBSelector* transSelector = new TransBBSelector(FrequencyBin(0,1e99));
  GridConfig* gridConfig = new GridConfig(cellConfig,bStateSelector,transSelector,updateMethod,matterDerivedMPI);

  // Create matter state solver
  MStateSolver* mStateSolver;
  switch (solverMethod) {
  case sNLTE : { mStateSolver = new NLTESolver(configFile); break; }
  case sComb : { mStateSolver = new CombSolver(configFile); break; }
  case sLTE : { mStateSolver = new LTESolver(); break; }
  deafult : { mStateSolver = new NebSolver(configFile); } }

  // Create EOS
  EOS* eos = new EOSBasic(configFile);

  // Create grid
  GridSS grid(mStateSolver,eos,gridConfig,mpiCom);

  // Load grid
  grid.load(gridDataFile);

  // Initialize to BB radiation field at matter temperature
  for (i=0;i<grid.cell.size();i++) {
    grid.cell[i]->TR(grid.cell[i]->T());
    grid.cell[i]->W(1.0);
    grid.cell[i]->reset();
  }

  // Configure MC experiment
  PacketConfig* packetConfig = new PacketConfig(false,false,lineScatter,eScatter,false,gScatter,freeFreeAbs,boundFreeAbs,false,0,dustAbs,flourescence,collProc,nonThermalProc,twoPhotonProc,chargeTransferProc,transTable,updateMethod,boost,convertGPacket,-1,-1,true,1.0,false,1.0,true,transformKappa,bbEPacket);
  MCConfig* mCConfig = new MCConfig(packetConfig);

  // Get time from innermost cell
  t=(grid.cell[0]->tMin()+grid.cell[0]->tMax())/2;

  // Determine timestep factor
  dlogt=log10(tStop/t)/nIter;

  // Determine initial time step
  dt=t*(pow(10,dlogt)-1);

  if (master) cout<<"mc-evo: Evolve grid, t: "<<t/day<<" days, dt: "<<dt/day<<" days\n";

  // Set timestep for each cell
  for (i=0;i<grid.cell.size();i++) { grid.cell[i]->tMin(t-dt/2); grid.cell[i]->tMax(t+dt/2); }

  // Create thermal packet sampler
  PacketSamplerThermal packetSamplerThermal(nPacket,0,grid.cell.size()-1,&grid,packetConfig);
 
  // Create packet statistics object
  PacketStat packetStat(&grid);

  // Create MC experiment
  MCExperiment* mCExperiment = new MCExperiment(&packetSamplerThermal,&packetStat,mCConfig,mpiCom);

  // Create electron density vector
  vector<double> ne(grid.cell.size());

  FrequencyBinList nuList;
  nuList.resize(2);
  nuList[0].min=nu.min;
  nuList[0].max=(nu.min+nu.max)/2;
  nuList[1].min=nuList[0].max;
  nuList[1].max=nu.max;
 
  // Initial (steady-state) iteration loop
  for (n=0;n<nIterInit;n++) {

    initiate=(n==0);

    // Store electron density
    for (i=0;i<grid.cell.size();i++) ne[i]=grid.cell[i]->ne();
    
    // Update matter state
    grid.updateMatterState(initiate);

    // Check electron density convergence
    for (i=0,neConverged=true;i<grid.cell.size();i++) {
      if (master) cout<<"mc-evo: cell "<<i<<", ne: "<<grid.cell[i]->ne()<<", neErr: "<<fabs((grid.cell[i]->ne()-ne[i])/grid.cell[i]->ne())<<"\n";
      if (fabs((grid.cell[i]->ne()-ne[i])/grid.cell[i]->ne())>neTol) neConverged=false;
    }
        
    // Break if electron density converged
    if (n>0 and neConverged) break;

    // Reset packet statistics
    packetStat.reset();

    // Reset grid MC statistics
    grid.resetMCStatistics(nuList);

    // Reset packet sampler
    packetSamplerThermal.reset();
    
    // Evolve MC experiment
    mCExperiment->evolve(packetIn,packetOut,&grid);
    
    // Update radiation state
    grid.updateRadiationState(*mCExperiment);

    // Export grid statistics
    if (gridStatDir.compare("")!=0 and master) grid.exportStat(statFile(gridStatDir,"grid-stat",n));

    // Export packet statistics
    if (mCStatDir.compare("")!=0 and master) packetStat.exportStat(statFile(mCStatDir,"mc-stat",n));
  }

  // Update packet state
  packetIn=packetOut;

  // Delete MC experiment
  delete mCExperiment;
    
  // Create MC experiment
  mCExperiment =  new MCExperiment(NULL,&packetStat,mCConfig,mpiCom);
  
  // Evolution iteration loop
  for (n=0;n<nIter;n++) {
    
    // Export grid statistics
    if (gridStatDir.compare("")!=0 and master) grid.exportStat(statFile(gridStatDir,"grid-stat",n));

    // Export packet statistics
    if (mCStatDir.compare("")!=0 and master) packetStat.exportStat(statFile(mCStatDir,"mc-stat",n));
   
    // Get time from innermost cell
    t=(grid.cell[0]->tMin()+grid.cell[0]->tMax())/2;

    // Determine timestep
    dt=t*(pow(10,dlogt)-1);

    if (master) cout<<"mc-evo: Evolve grid, t: "<<t/day<<" days, dt: "<<dt/day<<" days\n";
       
    // Evolve dynamical state
    evolveDyn(grid,dt);
    
    // Update matter state
    grid.updateMatterState();

    // Reset packet statistics
    packetStat.reset();
    
    // Reset grid MC statistics
    grid.resetMCStatistics(nuList);
    
    // Evolve MC experiment
    mCExperiment->evolve(packetIn,packetOut,&grid);
    
    // Update packet state
    packetIn=packetOut;
    
    // Update radiation state
    grid.updateRadiationState(*mCExperiment);
  }

  // Clean up
  delete mCExperiment;
  delete mCConfig;
  delete packetConfig;
  delete eos;
  delete mStateSolver;
  delete gridConfig;
  delete bStateSelector;
  delete transSelector;
  delete cellConfig;
  delete lineSelector;
}
