#include <vector>
#include <math.h>
#include <string.h>
#include "mc-util.h"
#include "mc-math.h"
#include "mc-physics.h"
#include "mc.h"
#include "msolver.h"
#include "eos.h"
#include "diffsolver.h"

using namespace util;
using namespace math;
using namespace physics;
using namespace mc;
using namespace msolver;
using namespace eos;
using namespace diffsolver;

MCExperiment* runMCExperiment(const int& border, const double& RBorder, const double& LBorder, const int& nPacket, const int& nPacketGamma, const int& nBin, const FrequencyBin& nu, const BBSampling& bbSampling, const BBFrame& bbFrame, const bool& bbLimbDark, const bool& createSpectrum, Grid* grid, MCConfig* mCConfig) {    
  PacketSampler* packetSampler;
  PacketStat* packetStat;
  MCExperiment* mCExperiment;
  int i;

  double rMax=((CellSS*)grid->cell[grid->cell.size()-1])->rMax();
  double tMin=grid->cell[0]->tMin();
  double tMax=grid->cell[0]->tMax();
 
  // Create new MC experiment
  mCConfig->packet->cellIndexMin=border;
  if (nPacketGamma==0) packetSampler = new PacketSamplerBB(RBorder,LBorder,nPacket,nBin,nu,bbSampling,bbFrame,bbLimbDark,grid,mCConfig->packet);
  else packetSampler = new PacketSamplerBBGamma(RBorder,LBorder,nPacket,nPacketGamma,nBin,nu,bbSampling,bbFrame,bbLimbDark,grid,mCConfig->packet);
  packetStat = new PacketStat(grid,border,grid->cell.size()-1,createSpectrum,packetSampler->nu);
  mCExperiment = new MCExperiment(packetSampler,packetStat,mCConfig);

  // Reset MC statistics
  for (i=border;i<grid->cell.size();i++) grid->cell[i]->resetMCStatistics(packetSampler->nu);

  // Run small Monte Carlo experiment
  mCExperiment->run();

  // Return packet statistics
  return mCExperiment;
}

int main (int argc, char** argv) {

  // Initialize variables
  double Lph,L,Tph,Rdiff,Ldiff,Tdiff,Rth,Tth,T,TPrev,r,rPrev,dTdR,LMC,LobsMC,vDiff;
  double rMax,tMin,tMax,tau,tauPhot=1.0,tauTherm=1.0,statTauLim=-1e99;
  int n,i,diffBorder,photosphere,thermBorder;
  bool Tconverged,Lconverged,initiate;
  stringstream specfile,statfile;
  CellSS* cell;

  /* Check if enough number of arguments provided */
  if (argc<4) {
    printf("mc-lr-diff: Missing argument\n");
    exit(-1);
  }
  
  /* Get mandatory arguments */
  string gridDataFile=string(argv[1]);
  double Lobs=strtod(argv[2],NULL);
  double RBase=strtod(argv[3],NULL);
  double Rph=strtod(argv[4],NULL);

  /* Set default values for optional arguments */
  int nTrial=20;
  int nTrialInit=3;
  int nPacket=4000000;
  int nPacketSmall=2000000;
  int nPacketInit=500000;
  //int nBin=10000;
  int nBin=5000; // Changed!
  int nMuBin=20; // New!
  int nBinSmall=100;
  int nPacketGamma=0;
  double tauLim=1e-3;
  double TTol=1e-2;
  double LTol=5e-2;
  double lambdaMin=1;
  double lambdaMax=25000;
  bool seed=false;
  bool eScatter=true;
  bool lineScatter=true;
  bool gScatter=false;
  bool freeFreeAbs=false;
  bool boundFreeAbs=false;
  bool flourescence=false;
  bool collProc=false;
  bool nonThermalProc=false;
  bool twoPhotonProc=false;
  bool chargeTransferProc=false;
  bool chargeTransferEmit=false;
  bool transTable=false;
  bool markovTable=false;
  bool markovTableExt=false;
  bool markovTableExtSplit=false;
  double markovTableLim=0.0;
  double markovTableThermLim=0.0;
  double markovTableCTLim=0.0;
  bool collExciteTable=false;
  double collExciteTableLim=0.0;
  bool nonThermalExciteTable=false;
  bool etaTable=false;
  UpdateMethod updateMethod=urOnTheFly;
  bool boundFreeTable=false;
  double boundFreeTableMin=1;
  double boundFreeTableMax=1e5;
  int boundFreeTableNBin=1e4;
  bool boundFreeSTable=false;
  int boundFreeSTableNBin=1e3;
  RadEstimator radEstimator=rfDBB;
  RadEstimator radEstimatorBB=rfDBB;
  bool radEstimatorExp=false;
  BBSampling bbSampling=bbsEqualEnergy;
  BBFrame bbFrame=bbfMixed;
  bool bbLimbDark=true;
  OnTheSpot onTheSpot("");
  OnTheSpotNew onTheSpotNew("");
  FFOnTheSpot ffOnTheSpot(false,10.0);
  bool boost=false;
  bool alfaCorrect=false;
  bool alfaCorrectEmit=false;
  bool alfaSimpleEmit=false;
  bool alfaSPH=false;
  bool forceDBalance=false;
  bool forceDBalanceOSP=false;
  bool boundFreeBB=false;
  double boundFreeBBTauLim=1e3;
  bool convertGPacket=true;
  bool transformKappa=false;
  bool bbEPacket=false;
  string configFile="";
  SolverMethod solverMethod=sNebular;
  double tauDiff=10.0;
  bool exportEta=false;
  bool exportSBF=false;
  bool transformRay=false;
  bool limitFreeFreeCRange=false;
  string gridTemplateFile="";
  string gridStatDir="";
  string mCStatDir="";
  bool JLineFallback=false;
  double JLineFallbackFactor=1.1;
  int JLineFallbackCount=3;
  bool JLineFallbackDBB=false;
  bool trackI=false;
  bool trackHK=false;
  double boundFreeCutOff=0.0;
  double boundFreeCutOffRoss=0.0;
  bool matterDerivedMPI=false;
  bool LTEGroupMacro=false;
  bool exportKappaMC2=true;
  Lph=Lobs;
  bool dustAbs=false;

  /* Get optional arguments */
  for (i=4;i<argc;i++) {
    if (strcmp(argv[i],"--ntrial")==0 && i+1<argc) 
      {i++; nTrial=atoi(argv[i]);}
    if (strcmp(argv[i],"--ntrial_init")==0 && i+1<argc) 
      {i++; nTrialInit=atoi(argv[i]);}
    else if (strcmp(argv[i],"--npacket")==0 && i+1<argc)
      {i++; nPacket=atoi(argv[i]);}
    else if (strcmp(argv[i],"--npacket_small")==0 && i+1<argc)
      {i++; nPacketSmall=atoi(argv[i]);}
    else if (strcmp(argv[i],"--npacket_init")==0 && i+1<argc)
      {i++; nPacketInit=atoi(argv[i]);}
    else if (strcmp(argv[i],"--nbin")==0 && i+1<argc)
      {i++; nBin=atoi(argv[i]);}
    else if (strcmp(argv[i],"--nbin_small")==0 && i+1<argc)
      {i++; nBinSmall=atoi(argv[i]);}
    else if (strcmp(argv[i],"--npacket_gamma")==0 && i+1<argc)
      {i++; nPacketGamma=atoi(argv[i]);}
    else if (strcmp(argv[i],"--taulim")==0 && i+1<argc)
      {i++; tauLim=(float)strtod(argv[i],NULL);}
    else if (strcmp(argv[i],"--ltol")==0 && i+1<argc)
      {i++; LTol=(float)strtod(argv[i],NULL);}
    else if (strcmp(argv[i],"--ttol")==0 && i+1<argc)
      {i++; TTol=(float)strtod(argv[i],NULL);}
    else if (strcmp(argv[i],"--lambda_min")==0 && i+1<argc)
      {i++; lambdaMin=(float)strtod(argv[i],NULL);}
    else if (strcmp(argv[i],"--lambda_max")==0 && i+1<argc)
      {i++; lambdaMax=(float)strtod(argv[i],NULL);}
    else if (strcmp(argv[i],"--seed")==0 && i+1<argc)
      {i++; if (strcmp(argv[i],"yes")) seed=true;}
    else if (strcmp(argv[i],"--escatter")==0 && i+1<argc)
      {i++; if (strcmp(argv[i],"yes")==0) eScatter=true;}
    else if (strcmp(argv[i],"--linescatter")==0 && i+1<argc)
      {i++; if (strcmp(argv[i],"no")==0) lineScatter=false;}
     else if (strcmp(argv[i],"--gscatter")==0 && i+1<argc)
      {i++; if (strcmp(argv[i],"yes")==0) gScatter=true;}
    else if (strcmp(argv[i],"--freefree_abs")==0 && i+1<argc)
      {i++; if (strcmp(argv[i],"yes")==0) freeFreeAbs=true;}
    else if (strcmp(argv[i],"--boundfree_abs")==0 && i+1<argc)
      {i++; if (strcmp(argv[i],"yes")==0) boundFreeAbs=true;}
    else if (strcmp(argv[i],"--flourescence")==0 && i+1<argc)
      {i++; if (strcmp(argv[i],"yes")==0) flourescence=true;}
    else if (strcmp(argv[i],"--coll_proc")==0 && i+1<argc)
      {i++; if (strcmp(argv[i],"yes")==0) collProc=true;}
    else if (strcmp(argv[i],"--non_thermal_proc")==0 && i+1<argc)
      {i++; if (strcmp(argv[i],"yes")==0) nonThermalProc=true;}
    else if (strcmp(argv[i],"--two_photon_proc")==0 && i+1<argc)
      {i++; if (strcmp(argv[i],"yes")==0) twoPhotonProc=true;}
    else if (strcmp(argv[i],"--charge_transfer_proc")==0 && i+1<argc)
      {i++; if (strcmp(argv[i],"yes")==0) chargeTransferProc=true;}
    else if (strcmp(argv[i],"--charge_transfer_emit")==0 && i+1<argc)
      {i++; if (strcmp(argv[i],"yes")==0) chargeTransferEmit=true;}
    else if (strcmp(argv[i],"--trans_table")==0 && i+1<argc)
      {i++; if (strcmp(argv[i],"yes")==0) transTable=true;}
    else if (strcmp(argv[i],"--markov_table")==0 && i+1<argc)
      {i++; if (strcmp(argv[i],"yes")==0) markovTable=true;}
    else if (strcmp(argv[i],"--markov_table_ext")==0 && i+1<argc)
      {i++; if (strcmp(argv[i],"yes")==0) markovTableExt=true;}
    else if (strcmp(argv[i],"--markov_table_lim")==0 && i+1<argc) 
      {i++; markovTableLim=(float)strtod(argv[i],NULL);}
    else if (strcmp(argv[i],"--markov_table_ext_split")==0 && i+1<argc)
      {i++; if (strcmp(argv[i],"yes")==0) markovTableExtSplit=true;}
    else if (strcmp(argv[i],"--markov_table_therm_lim")==0 && i+1<argc) 
      {i++; markovTableThermLim=(float)strtod(argv[i],NULL);}
    else if (strcmp(argv[i],"--markov_table_ct_lim")==0 && i+1<argc) 
      {i++; markovTableCTLim=(float)strtod(argv[i],NULL);}
    else if (strcmp(argv[i],"--coll_exc_table")==0 && i+1<argc)
      {i++; if (strcmp(argv[i],"yes")==0) collExciteTable=true;}
    else if (strcmp(argv[i],"--coll_exc_table_lim")==0 && i+1<argc) 
      {i++; collExciteTableLim=(float)strtod(argv[i],NULL);}
    else if (strcmp(argv[i],"--non_thermal_exc_table")==0 && i+1<argc)
      {i++; if (strcmp(argv[i],"yes")==0) nonThermalExciteTable=true;}
    else if (strcmp(argv[i],"--eta_table")==0 && i+1<argc)
      {i++; if (strcmp(argv[i],"yes")==0) etaTable=true;}
    else if (strcmp(argv[i],"--boundfree_table")==0 && i+1<argc)
      {i++; if (strcmp(argv[i],"yes")==0) boundFreeTable=true;}
    else if (strcmp(argv[i],"--boundfree_table_min")==0 && i+1<argc)
      {i++; boundFreeTableMin==(float)strtod(argv[i],NULL);}
    else if (strcmp(argv[i],"--boundfree_table_max")==0 && i+1<argc)
      {i++; boundFreeTableMax=(float)strtod(argv[i],NULL);}
    else if (strcmp(argv[i],"--boundfree_table_nbin")==0 && i+1<argc)
      {i++; boundFreeTableNBin=atoi(argv[i]);}
    else if (strcmp(argv[i],"--boundfree_stable")==0 && i+1<argc)
      {i++; if (strcmp(argv[i],"yes")==0) boundFreeSTable=true;}
    else if (strcmp(argv[i],"--boundfree_stable_nbin")==0 && i+1<argc)
      {i++; boundFreeSTableNBin=atoi(argv[i]);}
    else if (strcmp(argv[i],"--rad_estimator")==0 && i+1<argc)
      {i++; if (strcmp(argv[i],"lucy")==0) radEstimator=rfLucy;}
    else if (strcmp(argv[i],"--rad_estimator_bb")==0 && i+1<argc)
      {i++; if (strcmp(argv[i],"lucy")==0) radEstimatorBB=rfLucy;}
    else if (strcmp(argv[i],"--rad_estimator_exp")==0 && i+1<argc)
      {i++; if (strcmp(argv[i],"yes")==0) radEstimatorExp=true;}
    else if (strcmp(argv[i],"--bb_sampling")==0 && i+1<argc)
      {i++; if (strcmp(argv[i],"log_frequency")==0) bbSampling=bbsLogFrequency;}
    else if (strcmp(argv[i],"--bb_frame")==0 && i+1<argc)
      {
	i++; 
	if (strcmp(argv[i],"rest")==0) bbFrame=bbfRest;
	if (strcmp(argv[i],"cmf")==0) bbFrame=bbfCMF;
      }
    else if (strcmp(argv[i],"--bb_limb_dark")==0 && i+1<argc)
      {i++; if (strcmp(argv[i],"no")==0) bbLimbDark=false;}
    else if (strcmp(argv[i],"--onthespot")==0 && i+1<argc)
      {i++; onTheSpot=OnTheSpot(argv[i]);}
    else if (strcmp(argv[i],"--onthespot_new")==0 && i+1<argc)
      {i++; onTheSpotNew=OnTheSpotNew(argv[i]);}
    else if (strcmp(argv[i],"--ff_onthespot")==0 && i+1<argc)
      {i++; if (strcmp(argv[i],"yes")==0) ffOnTheSpot.apply=true;}
    else if (strcmp(argv[i],"--ff_onthespot_taulim")==0 && i+1<argc)
      {i++; ffOnTheSpot.tauLim=(float)strtod(argv[i],NULL);}
    else if (strcmp(argv[i],"--boost")==0 && i+1<argc)
      {i++; if (strcmp(argv[i],"yes")==0) boost=true;}
    else if (strcmp(argv[i],"--alfa_correct")==0 && i+1<argc)
      {i++; if (strcmp(argv[i],"yes")==0) alfaCorrect=true;}
    else if (strcmp(argv[i],"--alfa_correct_emit")==0 && i+1<argc)
      {i++; if (strcmp(argv[i],"yes")==0) alfaCorrectEmit=true;}
    else if (strcmp(argv[i],"--alfa_simple_emit")==0 && i+1<argc)
      {i++; if (strcmp(argv[i],"yes")==0) alfaSimpleEmit=true;}
    else if (strcmp(argv[i],"--alfa_SPH")==0 && i+1<argc)
      {i++; if (strcmp(argv[i],"yes")==0) alfaSPH=true;}
    else if (strcmp(argv[i],"--force_dbalance")==0 && i+1<argc)
      {i++; if (strcmp(argv[i],"yes")==0) forceDBalance=true;}
    else if (strcmp(argv[i],"--force_dbalance_osp")==0 && i+1<argc)
      {i++; if (strcmp(argv[i],"yes")==0) forceDBalanceOSP=true;}
    else if (strcmp(argv[i],"--convert_gpacket")==0 && i+1<argc)
      {i++; if (strcmp(argv[i],"no")==0) convertGPacket=false;}
    else if (strcmp(argv[i],"--transform_kappa")==0 && i+1<argc)
      {i++; if (strcmp(argv[i],"yes")==0) transformKappa=true;}
    else if (strcmp(argv[i],"--bb_epacket")==0 && i+1<argc)
      {i++; if (strcmp(argv[i],"yes")==0) bbEPacket=true;}
    else if (strcmp(argv[i],"--boundfree_bb")==0 && i+1<argc)
      {i++; if (strcmp(argv[i],"yes")==0) boundFreeBB=true;}
    else if (strcmp(argv[i],"--boundfree_bb_taulim")==0 && i+1<argc)
      {i++; boundFreeBBTauLim=(float)strtod(argv[i],NULL);}
    else if (strcmp(argv[i],"--config_file")==0 && i+1<argc)
      {i++; configFile=string(argv[i]);}
    else if (strcmp(argv[i],"--msolver_method")==0 && i+1<argc)
      {
	i++; 
	if (strcmp(argv[i],"nlte")==0) solverMethod=sNLTE;
	if (strcmp(argv[i],"comb")==0) solverMethod=sComb;
      } 
    else if (strcmp(argv[i],"--export_eta")==0 && i+1<argc)
      {i++; if (strcmp(argv[i],"yes")==0) exportEta=true;}
    else if (strcmp(argv[i],"--export_sbf")==0 && i+1<argc)
      {i++; if (strcmp(argv[i],"yes")==0) exportSBF=true;}
    else if (strcmp(argv[i],"--transform_ray")==0 && i+1<argc)
      {i++; if (strcmp(argv[i],"yes")==0) transformRay=true;}
    else if (strcmp(argv[i],"--limit_freefree_crange")==0 && i+1<argc)
      {i++; if (strcmp(argv[i],"yes")==0) limitFreeFreeCRange=true;}
    else if (strcmp(argv[i],"--grid_template_file")==0 && i+1<argc)
      {i++; gridTemplateFile=string(argv[i]);}
    else if (strcmp(argv[i],"--grid_stat_dir")==0 && i+1<argc)
      {i++; gridStatDir=string(argv[i]);}
    else if (strcmp(argv[i],"--mc_stat_dir")==0 && i+1<argc)
      {i++; mCStatDir=string(argv[i]);}
    else if (strcmp(argv[i],"--tau_diff")==0 && i+1<argc)
      {i++; tauDiff=(float)strtod(argv[i],NULL);}
    else if (strcmp(argv[i],"--jline_fallback")==0 && i+1<argc)
      {i++; if (strcmp(argv[i],"yes")==0) JLineFallback=true;}
    else if (strcmp(argv[i],"--jline_fallback_factor")==0 && i+1<argc)
      {i++; JLineFallbackFactor=(float)strtod(argv[i],NULL);}
    else if (strcmp(argv[i],"--jline_fallback_count")==0 && i+1<argc) 
      {i++; JLineFallbackCount=atoi(argv[i]);}
    else if (strcmp(argv[i],"--jline_fallback_dbb")==0 && i+1<argc)
      {i++; if (strcmp(argv[i],"yes")==0) JLineFallbackDBB=true;}
    else if (strcmp(argv[i],"--track_I")==0 && i+1<argc)
      {i++; if (strcmp(argv[i],"yes")==0) trackI=true;}
    else if (strcmp(argv[i],"--track_HK")==0 && i+1<argc)
      {i++; if (strcmp(argv[i],"yes")==0) trackHK=true;}
    else if (strcmp(argv[i],"--boundfree_cutoff")==0 && i+1<argc)
      {i++; boundFreeCutOff=(float)strtod(argv[i],NULL);}
    else if (strcmp(argv[i],"--boundfree_cutoff_ross")==0 && i+1<argc)
      {i++; boundFreeCutOffRoss=(float)strtod(argv[i],NULL);}
    else if (strcmp(argv[i],"--matter_derived_mpi")==0 && i+1<argc)
      {i++; if (strcmp(argv[i],"yes")==0) matterDerivedMPI=true;}
    else if (strcmp(argv[i],"--lte_group_macro")==0 && i+1<argc)
      {i++; if (strcmp(argv[i],"yes")==0) LTEGroupMacro=true;}
    else if (strcmp(argv[i],"--export_kappa_mc2")==0 && i+1<argc)
      {i++; if (strcmp(argv[i],"no")==0) exportKappaMC2=false;}
  }

  // Seed if requested
  if (seed) srand(time(NULL));
    
  // Create, configure and load grid
  FrequencyBin nu(c/(lambdaMax*AA),c/(lambdaMin*AA));
  LineSelector* lineSelector = new LineSelector(nu,tauLim);
  CellConfig* cellConfig = new CellConfig(true,neIteration,lineSelector,updateMethod,boundFreeTable,boundFreeTableMin,boundFreeTableMax,boundFreeTableNBin,boundFreeSTable,boundFreeSTableNBin,radEstimator,radEstimatorBB,radEstimatorExp,transTable,markovTable,markovTableExt,markovTableExtSplit,markovTableLim,markovTableThermLim,markovTableCTLim,collExciteTable,collExciteTableLim,nonThermalExciteTable,freeFreeAbs,boundFreeAbs,collProc,nonThermalProc,twoPhotonProc,chargeTransferProc,onTheSpot,onTheSpotNew,ffOnTheSpot,false,boundFreeBB,1e3,boundFreeBBTauLim,bbTCur,exportEta,transformRay,limitFreeFreeCRange,JLineFallback,JLineFallbackFactor,JLineFallbackCount,JLineFallbackDBB,trackI,trackHK,boundFreeCutOff,boundFreeCutOffRoss,boost,alfaCorrect,alfaCorrectEmit,alfaSimpleEmit,alfaSPH,forceDBalance,forceDBalanceOSP,chargeTransferEmit,etaTable,exportSBF,nu.min,nu.max,nBin,nMuBin,LTEGroupMacro,exportKappaMC2);
  BStateSelector* bStateSelector= new BStateSelector();
  TransBBSelector* transSelector = new TransBBSelector(FrequencyBin(0,1e99));
  GridConfig* gridConfig = new GridConfig(cellConfig,bStateSelector,transSelector,updateMethod,matterDerivedMPI);
  MStateSolver* mStateSolver;
  switch (solverMethod) {
  case sNLTE : { mStateSolver = new NLTESolver(configFile); break; }
  case sComb : { mStateSolver = new CombSolver(configFile); break; }
  deafult : { mStateSolver = new NebSolver(configFile); } }
  EOS* eos = new EOSBasic(configFile);
  GridSS grid(mStateSolver,eos,gridConfig);
  if (gridTemplateFile.compare("")) grid.processGridTemplate(gridTemplateFile,gridDataFile);
  grid.load(gridDataFile,RBase,Lph);

  // Configure diffusion solver
  DiffSolver diffSolver(configFile);
  diffSolver.config.innerBoundary.method=dLdmZero;
  diffSolver.config.outerBoundary.method=eddington;
  diffSolver.config.outerBoundary.LTFixed=true;
  diffSolver.config.outerBoundary.L=Lobs;
  diffSolver.config.expansion=true;

  // Configure MC experiment
  PacketConfig* packetConfig = new PacketConfig(true,false,lineScatter,eScatter,false,gScatter,freeFreeAbs,boundFreeAbs,false,0,dustAbs,flourescence,collProc,nonThermalProc,twoPhotonProc,chargeTransferProc,transTable,updateMethod,boost,convertGPacket,-1,-1,false,1.0,false,1.0,false,transformKappa,bbEPacket); // Bug fix!
  MCConfig* mCConfig = new MCConfig(packetConfig);
  PacketSampler* packetSampler = NULL;
  PacketStat* packetStat;
  MCExperiment* mCExperiment = NULL;

  vector<double> TOld(grid.cell.size(),0);
  rMax=((CellSS*)grid.cell[grid.cell.size()-1])->rMax();
  tMin=grid.cell[0]->tMin();
  tMax=grid.cell[0]->tMax();
  diffBorder=grid.cell.size()-2;

  // Find cell containg the photosphere
  for (i=0;i<grid.cell.size();i++)
    if (Rph>=((CellSS*) grid.cell[i])->rMin() and Rph<((CellSS*) grid.cell[i])->rMax()) { 
      photosphere=i-1;
      Rph=((CellSS*) grid.cell[i])->rMin();
      break; 
    }

  cout<<"mc-lr-diff: photosphere: "<<photosphere<<", Rph: "<<Rph<<"\n";

  // Calculate BB luminosity at photosphere
  Tph=pow(Lph/(4*pi*pow(Rph,2)*sigma),(1.0/4.0));

  // Initiatal guess for radiation field in outer region
  for (i=photosphere+1;i<grid.cell.size();i++) {
    cell = (CellSS*) grid.cell[i];
    cell->TR(Tph);
    cell->W(0.5*(1-pow(1-pow(Rph/((cell->rMin()+cell->rMax())/2),2),0.5)));
  }

  // Initiation loop
  for (n=0;n<nTrialInit;n++) {

    // Set initiate flag
    initiate=(n==0);

    // Update matter state in outer region
    for (i=photosphere+1;i<grid.cell.size();i++) grid.cell[i]->updateMatterState(initiate);

    // If grid statistics directory provided
    /*if (gridStatDir.compare("")!=0) {
      // Export grid statistics
      statfile.str(""); statfile.clear(); statfile<<gridStatDir<<"/grid-stat-init-"<<n<<".dat";
      grid.exportStat(statfile.str(),statTauLim);
      }*/

    cout<<"mc-lr-diff: Run MC experiment: photosphere: "<<photosphere<<", Rph: "<<Rph<<", Lph: "<<Lph<<"\n";

    // Run MC experiment in outer region
    if (mCExperiment!=NULL) { delete mCExperiment->packetSampler; delete mCExperiment->packetStat; delete mCExperiment; }
    mCExperiment=runMCExperiment(photosphere+1,Rph,Lph,nPacketInit,nPacketGamma,nBinSmall,nu,bbSampling,bbFrame,true,false,&grid,mCConfig);

    // If MC statistics directory provided
    if (mCStatDir.compare("")!=0) {
      // Export MC statistics
      statfile.str(""); statfile.clear(); statfile<<mCStatDir<<"/mc-stat-init-"<<n<<".dat";
      mCExperiment->packetStat->exportStat(statfile.str());
    }

    // Update radiation state in outer region
    for (i=photosphere+1;i<grid.cell.size();i++) grid.cell[i]->updateRadiationState(mCExperiment->packetSampler->nu);

    // Adjust luminosity
    L=mCExperiment->packetStat->L();
    cout<<"mc-lr-diff: Lobs="<<Lobs<<", L="<<L<<"\n";
    Lph=Lph*(Lobs/L);
  }

  // Set diffusion border to photosphere
  diffBorder=photosphere;
  Rdiff=Rph;

  // Iteration loop
  for (n=0;n<nTrial;n++) {
    
    Lconverged=false;
    Tconverged=false;

    // Set initiate flag
    initiate=(n==0);
    
    // Store old temperature
    for (i=0;i<grid.cell.size();i++) TOld[i]=grid.cell[i]->T();

    // Set luminosity at diffusion border
    diffSolver.config.outerBoundary.L=4*pi*pow(Rdiff,2)*((CellSS*) grid.cell[diffBorder+1])->FBack();

    // Update inner region using diffusion solver
    diffSolver.update(&grid,diffBorder,initiate);

    // If initial iteration
    if (initiate) {
      // Find diffusion approximation border
      for (i=grid.cell.size()-1,tau=0;i>=0;i--) { 
	cell = (CellSS*) grid.cell[i];
	diffBorder=i; Rdiff=cell->rMax();
	tau+=cell->rossContKappa(Processes(pEScatter|pBoundFree|pFreeFree))*cell->rho()*(cell->rMax()-cell->rMin());
	cout<<"mc-lr-diff: i: "<<i<<", tau: "<<tau<<"\n";
	if (tau>tauDiff) break;
      }
      // Find thermalization border
      for (i=grid.cell.size()-1,tau=0;i>=0;i--) { 
	cell = (CellSS*) grid.cell[i];
	thermBorder=i; Rth=cell->rMax();
	tau+=cell->rossContKappa(Processes(pBoundFree|pFreeFree))*cell->rho()*(cell->rMax()-cell->rMin());
	if (tau>tauTherm) break;
      }
      // Calculate BB luminosity at thermalization border
      //Tth=pow(Lph/(4*pi*pow(Rth,2)*sigma),(1.0/4.0));
      Tth=(grid.cell[thermBorder]->T()+grid.cell[thermBorder+1]->T())/2;
      // Initiatal guess for radiation field between the thermalization border and the photosphere
      for (i=thermBorder+1;i<grid.cell.size();i++) {
	cell = (CellSS*) grid.cell[i];
	cell->TR(Tth);
	cell->W(0.5*(1-pow(1-pow(Rth/((cell->rMin()+cell->rMax())/2),2),0.5)));
      }
      cout<<"mc-lr-diff: diffBorder: "<<diffBorder<<", Rdiff: "<<Rdiff<<", thermBorder: "<<thermBorder<<", Rth: "<<Rth<<", Tth: "<<Tth<<"\n";
    }

    // Update matter state in outer region
    for (i=diffBorder+1;i<grid.cell.size();i++) {
      if (i<photosphere+1) grid.cell[i]->updateMatterState(initiate);
      else grid.cell[i]->updateMatterState();
    }

    // Calculate optical depth at diffusion border
    for (i=grid.cell.size()-1,tau=0;i>diffBorder;i--) {
      cell = (CellSS*) grid.cell[i];
      tau+=cell->rossContKappa(Processes(pEScatter|pBoundFree|pFreeFree))*cell->rho()*(cell->rMax()-cell->rMin());
      cout<<"mc-lr-diff: i: "<<i<<", tau: "<<tau<<"\n";
    }

    cout<<"mc-lr-diff: tauDiff: "<<tau<<"\n";
    
    // If grid statistics directory provided
    if (gridStatDir.compare("")!=0) {
      // Export grid statistics
      statfile.str(""); statfile.clear(); statfile<<gridStatDir<<"/grid-stat-"<<n<<".dat";
      grid.exportStat(statfile.str(),statTauLim);
    }

    // Calculate BB luminosity at diffusion border
    if (initiate) {
      Tdiff=(grid.cell[diffBorder]->T()+grid.cell[diffBorder+1]->T())/2;
      Ldiff=4*pi*pow(Rdiff,2)*sigma*pow(Tdiff,4);
    }
    else {
      Ldiff=Ldiff*(Lobs/L);
      Tdiff=pow(Ldiff/(4*pi*pow(Rdiff,2)*sigma),(1.0/4.0));
    }

    cout<<"mc-lr-diff: Run MC experiment: diffBorder: "<<diffBorder<<", Rdiff: "<<Rdiff<<", Ldiff: "<<Ldiff<<", Tdiff: "<<Tdiff<<"\n";

    // Run MC experiment in outer region
    if (mCExperiment!=NULL) { delete mCExperiment->packetSampler; delete mCExperiment->packetStat; delete mCExperiment; }
    mCExperiment=runMCExperiment(diffBorder+1,Rdiff,Ldiff,nPacketSmall,nPacketGamma,nBinSmall,nu,bbSampling,bbFrame,bbLimbDark,false,&grid,mCConfig);

    // If MC statistics directory provided
    if (mCStatDir.compare("")!=0) {
      // Export MC statistics
      statfile.str(""); statfile.clear(); statfile<<mCStatDir<<"/mc-stat-"<<n<<".dat";
      mCExperiment->packetStat->exportStat(statfile.str());
    }

    // Update radiation state in outer region
    for (i=diffBorder+1;i<grid.cell.size();i++) grid.cell[i]->updateRadiationState(mCExperiment->packetSampler->nu);

    // Calculate optical depth at diffusion border
    for (i=grid.cell.size()-1,tau=0;i>diffBorder;i--) {
      cell = (CellSS*) grid.cell[i];      
      tau+=cell->energyContKappa(Processes(pEScatter|pBoundFree|pFreeFree))*cell->rho()*(cell->rMax()-cell->rMin());
      cout<<"mc-lr-diff: i: "<<i<<", tau: "<<tau<<"\n";
    }

    cout<<"mc-lr-diff: tauDiff (energy): "<<tau<<"\n";

    //exit(-1);
 
    // Check cell temperature convergence
    if (not initiate) {
      Tconverged=true;
      for (i=0;i<grid.cell.size();i++) {
	cout<<"mc-lr-diff: Temperature error in cell "<<i<<" : "<<fabs((grid.cell[i]->T()-TOld[i])/grid.cell[i]->T())<<"\n";
	if (fabs((grid.cell[i]->T()-TOld[i])/grid.cell[i]->T())>TTol) { 
	  Tconverged=false;
	}
      }
    }
    
    // Check luminosity convergence
    L=mCExperiment->packetStat->L();
    cout<<"mc-lr-diff: Lobs="<<Lobs<<", L="<<L<<"\n";
    if (fabs((Lobs-L)/Lobs)<LTol) Lconverged=true;

    // End loop if luminosity and temperature converged
    if (Lconverged and Tconverged) break;
  }
  if (Lconverged and Tconverged) cout<<"Luminosity converged in "<<i+1<<" iterations and a total of "<<n+1<<" iterations\n";

  // Run MC experiment in outer region
  if (mCExperiment!=NULL) { delete mCExperiment->packetSampler; delete mCExperiment->packetStat; delete mCExperiment; }
  mCExperiment=runMCExperiment(diffBorder+1,Rdiff,Ldiff,nPacket,nPacketGamma,nBin,nu,bbSampling,bbFrame,bbLimbDark,true,&grid,mCConfig);

  cout<<"mc-lr-diff: Export spectrum and grid statistics\n";
  mCExperiment->packetStat->exportSpectrum("spectrum.dat");
  mCExperiment->packetStat->exportStat("mc-stat.dat");
  grid.exportStat("grid-stat.dat");
}
