#include <cstdlib>
#include <iostream>
#include <fstream>
#include <math.h>
#include <limits>
#include <omp.h>
#include "nr.h"
#include "mc-math.h"

using namespace std;
using namespace NR;
using namespace math;

// Constuctor
math::Vector3D::Vector3D() : x(0), y(0), z(0) { }

// Constructor
math::Vector3D::Vector3D(const double& x1, const double& x2, const double& x3, const CoordSystem& coordSystem) {
  switch (coordSystem) {
  case spherical :
    x=x1*sin(x2)*cos(x3);
    y=x1*sin(x2)*sin(x3);
    z=x1*cos(x2);
    break;
  case cylindrical :
    x=x1*cos(x2);
    y=x2*sin(x2);
    z=x3;
    break;
  default:
    x=x1;
    y=x2;
    z=x3;
    break;
  }
}

// Add vector
Vector3D math::Vector3D::operator+(const Vector3D& vector) const {
  Vector3D result;
  result.x=x+vector.x;
  result.y=y+vector.y;
  result.z=z+vector.z;

  return result;
}

// Subtract vector
Vector3D math::Vector3D::operator-(const Vector3D& vector) const {
  Vector3D result;
  result.x=x-vector.x;
  result.y=y-vector.y;
  result.z=z-vector.z;

  return result;
}

// Add scalar
Vector3D math::Vector3D::operator+(const double& scalar) const {
  Vector3D result;
  result.x=x+scalar;
  result.y=y+scalar;
  result.z=z+scalar;
  
  return result;
}

// Subtract scalar
Vector3D math::Vector3D::operator-(const double& scalar) const {
  Vector3D result;
  result.x=x-scalar;
  result.y=y-scalar;
  result.z=z-scalar;
  
  return result;
}

// Change sign
Vector3D math::Vector3D::operator-() const {
  Vector3D result;
  result.x=-x;
  result.y=-y;
  result.z=-z;
  
  return result;
}

// Multiply with vector (scalar product)
double math::Vector3D::operator*(const Vector3D& vector) const {
  double result;
  result=x*vector.x+y*vector.y+z*vector.z;
  return result;
}

// Multiply with scalar
Vector3D math::Vector3D::operator*(const double& scalar) const {
  Vector3D result;
  result.x=x*scalar;
  result.y=y*scalar;
  result.z=z*scalar;

  return result;
}

// Divide with scalar 
Vector3D math::Vector3D::operator/(const double& scalar) const {
  Vector3D result;
  result.x=x/scalar;
  result.y=y/scalar;
  result.z=z/scalar;
  
  return result;
}

// Get coordinates in specified coordinate system
// No support for cylidrical yet and the sperical case should be checked !!!
void math::Vector3D::coords(double& x1, double& x2, double& x3, const CoordSystem& coordSystem) const {
  switch (coordSystem) {
  case spherical :
    x1=sqrt(pow(x,2)+pow(y,2)+pow(z,2));
    x2=acos(z/(x1));
    x3=atan2(y,x);
    break;
  default :
    x1=x;
    x2=y;
    x3=z;
    break;
  }
}

// Calculate norm
double math::Vector3D::norm() const {
  return sqrt((*this)*(*this));
}

// Rotate
void math::Vector3D::rotate(const double& theta, const double& phi) {
  double x1,y1,z1,x2,y2,z2;
  x1=x*cos(theta)+z*sin(theta);
  y1=y;
  z1=z*cos(theta)-x*sin(theta);
  x2=x1*cos(phi)-y1*sin(phi);
  y2=y1*cos(phi)+x1*sin(phi);
  z2=z1;
  x=x2;
  y=y2;
  z=z2;
  return;
}

// Randomize
/*double math::randomize(const double& x1, const double& x2) {
  int r;
  double z;
  // Note: rand is not threadsafe
  //pragma omp critical
  r=rand();
  z=x1+double(r)*(x2-x1)/RAND_MAX;
  //cout<<"math:randomize: z: "<<z<<"\n";
  return z;
  }*/


// Randomize
/*double math::randomize(const double& x1, const double& x2, int seed) {
  static int idum;
  double z;
  #pragma omp critical
  {
    if (seed>=0) idum=-seed;
    z=x1+double(ran1(idum))*(x2-x1);
  }
  return z;
  }*/

// Randomize
// Does this work with nested openmp?
// And what about the static members of ran1?!  - Testing fix - see below.
double math::randomize(const double& x1, const double& x2, int seed) {
  static vector<int> idum;
  static vector<int> iy; // Testing!
  static vector<Vec_INT> iv; // Testing!
  double z;
  const int NTAB=32; // Testing!!!
  int i;
  
  //cout<<"math::randomize: seed: "<<seed<<"\n";

  // If seed specified
  if (seed>=0) {
    // Get maximum number of threads
    int n=omp_get_max_threads();
    // Resize seed vector
    idum.resize(n);
    // Testing!!!
    iy.resize(n,0);
    iv.resize(n); for (i=0;i<n;i++) iv[i]=Vec_INT(NTAB);
    // Set seed for each (posssible) thread
    for (i=0;i<n;i++) idum[i]=-(seed+n);
    //cout<<"math::randomize: Initiated "<<n<<" seeds\n";
  }

  // Get thread number
  i=omp_get_thread_num();

  //cout<<"math::randomize: Use seed "<<i<<", out of "<<idum.size()<<"\n";

  // Get random number
  //z=x1+double(ran1(idum[i]))*(x2-x1);
  // Testing!!!
  z=x1+double(ran1_new(idum[i],iy[i],iv[i]))*(x2-x1);

  return z;
}

/*
// Randomize
double math::randomize(const double& x1, const double& x2, bool critical, int seed) {
  static int idum;
  if (seed>=0) idum=-seed;
  if (critical) return randomize_critical(x1,x2,idum);
  else return randomize_normal(x1,x2,idum);
}

// Randomize (normal version)
double math::randomize_normal(const double& x1, const double& x2, int& idum) {
  double z=x1+double(ran1(idum))*(x2-x1);
  return z;
}

// Randomize (threadsafe version)
double math::randomize_critical(const double& x1, const double& x2, int& idum) {
  #pragma omp critical 
  double z=x1+double(ran1(idum))*(x2-x1);
  return z;
  }*/

// Calculate roots to second degree equation
void math::root_2p(const double& a, const double& b, const double& c, double& rootMin, double& rootMax) {
  double mprec=1e-15;
  double fprec=mprec*MAX(fabs(pow(b,2)),fabs(4*a*c));
  double d=pow(b,2)-4*a*c;
  double b1;
  if (d>=0) d=fprec*floor(d/fprec);
  else d=fprec*ceil(d/fprec);
  rootMin=(-b-sqrt(d))/(2*a);
  rootMax=(-b+sqrt(d))/(2*a);
  fprec=mprec*MAX(fabs(b/(2*a)),fabs(sqrt(d)/(2*a)));
  if (rootMin>=0) rootMin=fprec*floor(rootMin/fprec);
  else rootMin=fprec*ceil(rootMin/fprec);
  if (rootMax>=0) rootMax=fprec*floor(rootMax/fprec);
  else rootMax=fprec*ceil(rootMin/fprec);
  return;
}

void math::root_2p(const float& a, const float& b, const float& c, float& rootMin, float& rootMax) {
  double da = (double) a;
  double db = (double) b;
  double dc = (double) c;
  double drootMin;
  double drootMax;
  root_2p(da,db,dc,drootMin,drootMax);
  rootMin = (float) drootMin;
  rootMax = (float) drootMax;
}

// Calculate roots to second degree equation (fast version)
void math::root_2p_fast(const double& a, const double& b, const double& c, double& rootMin, double& rootMax) {
  double b2=pow(b,2);
  double fac=4*a*c;
  double ta=2*a;
  double fprec=1e-15*MAX(fabs(b2),fabs(fac));
  double d=b2-fac;
  if (d>=0) d=fprec*floor(d/fprec);
  else d=fprec*ceil(d/fprec);
  double sd=sqrt(d);
  rootMin=(-b-sd)/ta;
  rootMax=(-b+sd)/ta;
  fprec=1e-15*MAX(fabs(b/ta),fabs(sd/ta));
  if (rootMin>=0) rootMin=fprec*floor(rootMin/fprec);
  else rootMin=fprec*ceil(rootMin/fprec);
  if (rootMax>=0) rootMax=fprec*floor(rootMax/fprec);
  else rootMax=fprec*ceil(rootMin/fprec);
  return;
}

// Calculate roots to second degree equation (fast version 2)
// Is the precision problem (see above) taken care of here?
void math::root_2p_fast_2(const double& a, const double& b, const double& c, double& rootMin, double& rootMax) {
  double sign=(b>0)?1:((b<0)?-1:0); 
  double q=-0.5*(b+sign*sqrt(b*b-4*a*c));
  // What about a=0?
  double r1=q/a;
  // What about q=0?
  double r2=c/q;
  if (r1<=r2) { rootMin=r1; rootMax=r2; }
  else { rootMin=r2; rootMax=r1; }
  return;
}

// Get maximum value in vector
double math::max(const vector<double>& x) {
  double xMax=(x.size()>0?x[0]:0);
  int i;
  for (i=1;i<x.size();i++) if (x[i]>xMax) xMax=x[i];
  return xMax;
}

// Constructors
FunctionDomain::FunctionDomain() : min(0), max(0) { }
FunctionDomain::FunctionDomain(const double& _min, const double& _max) : min(_min), max(_max) { }

FunctionValueConstIterator Interpolator1D::findInt(const double& x) const {
  unsigned int iBegin,iMiddle,iEnd;
  
  if (pTable.size()!=0) {
    iBegin=0; iEnd=pTable.size()-1; iMiddle=(iBegin+iEnd)/2;
    // Check if x is in the current or the next bin
    if (iEnd>iLast) if (x>=pTable[iLast].x && x<pTable[iLast+1].x) return pTable.begin()+(iLast+1);
    if (iEnd>iLast+1) if (x>=pTable[iLast+1].x && x<pTable[iLast+2].x) return pTable.begin()+(iLast+2);
    // DO a binary search for the bin
    if (x>=pTable[iBegin].x && x<pTable[iEnd].x) {
      while (iBegin!=iEnd) {
	if (x<pTable[iMiddle].x) {
	  iEnd=iMiddle;
	  iMiddle=(iBegin+iEnd)/2;
	}
	else {
	  iBegin=iMiddle+1;
	  iMiddle=(iBegin+iEnd)/2;
	}
      }
      return pTable.begin()+iBegin;
    }
    else if (x==pTable[iEnd].x) return pTable.begin()+iEnd;
  }
  
  return pTable.end();
}

FunctionValueConstIterator Interpolator1D::findLong(const double& x) const {
  unsigned long iBegin,iMiddle,iEnd;
  
  if (pTable.size()!=0) {
    iBegin=0; iEnd=pTable.size()-1; iMiddle=(iBegin+iEnd)/2;
    // Check if x is in the current or the next bin
    if (iEnd>iLast) if (x>=pTable[iLast].x && x<pTable[iLast+1].x) return pTable.begin()+(iLast+1);
    if (iEnd>iLast+1) if (x>=pTable[iLast+1].x && x<pTable[iLast+2].x) return pTable.begin()+(iLast+2);
    // Do a binary search for the bin
    if (x>=pTable[iBegin].x && x<pTable[iEnd].x) {
      while (iBegin!=iEnd) {
	if (x<pTable[iMiddle].x) {
	  iEnd=iMiddle;
	  iMiddle=(iBegin+iEnd)/2;
	}
	else {
	  iBegin=iMiddle+1;
	  iMiddle=(iBegin+iEnd)/2;
	}
      }
      return pTable.begin()+iBegin;
    }
    else if (x==pTable[iEnd].x) return pTable.begin()+iEnd;
  }
  
  return pTable.end();
}

double Interpolator1D::init(const vector<FunctionValue>& table) {
  pTable=table;
  if (table.size()>0) pDomain=FunctionDomain(table[0].x,table[table.size()-1].x);
  else pDomain=FunctionDomain(0,0);
  iLast=0;
}

double Interpolator1D::operator ()(double x) {
  double f;
  FunctionValueConstIterator item;
  if (pTable.size()<=numeric_limits<unsigned int>::max()) item=findInt(x);
  else item=findLong(x);
  if (item==pTable.end()) { cout<<"Interpolator1D: Error: "<<x<<" outside function domain ("<<domain().min<<","<<domain().max<<")\n"; iLast=0; return 0; }
  else iLast=(item-1)-pTable.begin();
  // If x equals one of the interval endpoints - Just return tabulated value
  if (x==(item-1)->x) return (item-1)->f;
  if (x==item->x) return item->f;
  double dfdx=(item->f-(item-1)->f)/(item->x-(item-1)->x);
  f=(item-1)->f+dfdx*(x-(item-1)->x);

  return f;
}

FunctionDomain Interpolator1D::domain() const {
  return pDomain;
}

vector<FunctionValue>& Interpolator1D::table() {
  return pTable;
}

double Interpolator1D::mem() const { // New!
  double _mem=0;
  int i;
  for (i=0;i<pTable.size();i++) _mem+=sizeof(FunctionValue);
  return _mem;
}

// Constructors
Domain::Domain() : min(0), max(0) { }
Domain::Domain(DP _min, DP _max) : min(_min), max(_max) { }

// Rewrite of mnewt in Numerical Recipes
// Given an initial guess x[0..n-1] for a root in n dimensions, take ntrial Newton-Raphson
// steps to improve the root. Stop if the root converges in either summed absolute variable increments
// tolx os summed absolute function values tolf
void math::mnewt(const int ntrial, Vec_IO_DP& x, Vec_I_DOM& domain, const DP damp, void (*usrfun) (Vec_I_DP&, Vec_O_DP&, Mat_O_DP&, void*), bool (*xconv) (Vec_I_DP&, Vec_I_DP&, void*), bool (*fconv) (Vec_I_DP&, void*), void* object)
{
  int k,i;
  DP d;
  int n=x.size();
  Vec_INT indx(n);
  Vec_DP p(n),fvec(n);
  Vec_DP xold(x);
  Mat_DP fjac(n,n);
  for (k=0;k<ntrial;k++) {
    // User function supplies function values at x in
    // fvec and Jacobian matrix in fjac
    (*usrfun) (x,fvec,fjac,object);
    //errf=0.0;
    // Right-hand side of linear equation
    for (i=0;i<n;i++) p[i] = -fvec[i];
    // Solve linear equations using LU decomposition
    ludcmp(fjac,indx,d);
    lubksb(fjac,indx,p);
    // Update solution				       
    for (i=0;i<n;i++) {
      xold[i] = x[i];
      x[i] += p[i]*damp;
    }
    // limit x to function domain
    for (i=0;i<n;i++) {
      if (x[i]<domain[i].min) { x[i]=domain[i].min; cout<<"math::mnewt: OUT OF DOMAIN\n"; }
      if (x[i]>domain[i].max) { x[i]=domain[i].max; cout<<"math::mnewt: OUT OF DOMAIN\n"; }
    }
    if (((*xconv) (x,xold,object)) && ((*fconv) (fvec,object))) {
      cout<<"math::mnewt: Converged in "<<k+1<<" iterations\n";
      return;
    }
  }
  cerr<<"math::mnewt: Error: Did not converge\n";

  return;
}

// Variant on tridag for tridiagonal matrix shifted on step upwards (i.e. skipping the first equation)
// and with one additional (trivial) equation for a[n-1]
void math::tridag_upshift(Vec_I_DP &a, Vec_I_DP& b, Vec_I_DP& c, Vec_I_DP& r, Vec_O_DP& u) {
  int j;
  DP bet;

  int n=a.size();
  // One vector of workspace, gam is needed
  Vec_DP gam(n);
  // Leftmost column
  Vec_DP v(0.0,n); v[0]=a[0];
  if (b[0] == 0.0) nrerror("Error 1 in tridag_upshift");
  // If this happens then you should rewrite your equations as a set of order N-1, with u1
  // trivially eliminated
  u[0]=r[0]/(bet=b[0]);
  // Decomposition and forward substitution
  for (j=1;j<(n-1);j++) {
    gam[j]=c[j-1]/bet;
    bet=b[j]-a[j]*gam[j];
    // Algorithm fails; see Numerical Recipes 2.4
    if (bet == 0.0) nrerror("Error 2 in tridag_upshift");
    u[j]=(r[j]-a[j]*u[j-1])/bet;
    // Keep track of v (leftmost coloum)
    v[j]=(v[j]-a[j]*v[j-1])/bet;
  }
  for (j=(n-3);j>=0;j--) {
    // Back substitution
    u[j] -= gam[j+1]*u[j+1];
    // Keep track of v (leftmost column)
    v[j] -= gam[j+1]*v[j+1];
  }
  // Permutate u[0..n-2] and v[0..n-2] one step
  for (j=(n-2);j>=0;j--) { u[j+1]=u[j]; v[j+1]=v[j]; }
  u[0]=u[n-1]; v[0]=v[n-1];
  // Solve for u[n-1]
  u[n-1]=r[n-1]/a[n-1];
  cout<<"u["<<n-1<<"]: "<<u[n-1]<<", r["<<n-1<<"]: "<<r[n-1]<<", a["<<n-1<<"]: "<<a[n-1]<<"\n";
  // Solve for u[0]
  u[0]=(u[0]-u[n-1])/v[0];
  // Eliminate v (leftmost column)
  for (j=1;j<(n-1);j++) u[j] -= v[j]*u[0];
  // Recalculate u[0]
  u[0]=(r[0]-u[1]*b[0]-u[2]*c[0])/a[0];
}

// Test if double is a number
bool math::isnumber(double x) {
  // This may work !?!
  return (isnormal(x) or fpclassify(x)==FP_ZERO);
}

// Exponential with lower (non-zero) bound
double expLB(double x, double expLowerBound) { 
  return MAX(exp(x),expLowerBound);
}
