#ifndef _MC_MATH_
#define _MC_MATH_

#include <iostream>
#include <fstream>
#include <math.h>
#include <vector>
#include <boost/mpi.hpp> // New!
#include <boost/serialization/string.hpp> // New!
#include "nr.h"

using namespace std;
using namespace NR;

namespace math {

  const double pi=3.1415926535897931;

  enum CoordSystem { cartesian, cylindrical, spherical };

  // 3D vector
  class Vector3D {
  private :
    // MPI serialization
    friend class boost::serialization::access;
    template<class Archive> 
      void serialize (Archive & ar, const unsigned int version) 
      {
	ar & x;
	ar & y;
	ar & z;
      }
  protected :
    // Cartesian coordinates
    double x;
    double y;
    double z;
  public :
    // Constructor
    Vector3D();
    // Constructor
    Vector3D(const double& x1, const double& x2, const double& x3, const CoordSystem& coordSystem=cartesian);
    // Add vector
    Vector3D operator+(const Vector3D& vector) const;
    // Subtract vector
    Vector3D operator-(const Vector3D& vector) const;
    // Add scalar
    Vector3D operator+(const double& scalar) const;
    // Subtract scalar
    Vector3D operator-(const double& scalar) const;
    // Change sign
    Vector3D operator-() const;
    // Multiply with vector (scalar product)
    double operator*(const Vector3D& vector) const;
    // Multiply with scalar
    Vector3D operator*(const double& scalar) const;
    // Divide with scalar
    Vector3D operator/(const double& scalar) const;
    // Get coordinates in specified coordinate system
    void coords (double& x1, double& x2, double& x3, const CoordSystem& coordSystem=cartesian) const;
    // Calculate norm
    double norm() const;
    // Rotate vector
    void rotate(const double& theta, const double& phi);
  };

  const int noSeed=-1;

  // Randomize
  //double randomize(const double& x1, const double& x2);
  double randomize(const double& x1, const double& x2, int seed=noSeed);
  //double randomize(const double& x1, const double& x2, bool critical=true, int seed=noSeed);
  //double randomize_normal(const double& x1, const double& x2, int& idum);
  //double randomize_critical(const double& x1, const double& x2, int& idum);
  
  // Calculate roots to second degree equation
  void root_2p(const double& a, const double& b, const double& c, double& rootMin, double& rootMax);
  void root_2p(const float& a, const float& b, const float& c, float& rootMin, float& rootMax);
  // Calculate roots to second degree equation (fast version)
  void root_2p_fast(const double& a, const double& b, const double& c, double& rootMin, double& rootMax);
  // Calculate roots to second degree equation (fast version 2)
  void root_2p_fast_2(const double& a, const double& b, const double& c, double& rootMin, double& rootMax);

  // Get maxium value in vector
  double max(const vector<double>& x);

  // Integrator 
  template <class F> class Integrator {
  public:
    // Definite integration using trapezoidal method
    static double trapezoidal(const F& f, const double& x1, const double& x2, double tol=1e-3, int nIterMin=5, int nIterMax=100, bool _adjustLimits=true);
    // Definite integration using Simpsons method
    static double simpson(const F& f, const double& x1, const double& x2, double tol=1e-3, int nIterMin=5, int nIterMax=100, bool _adjustLimits=true);
    // Definite integration using Rombergs method
    //static double romberg(const F& f, const double& x1, const double& x2, int& nIter, double tol=1e-3, int nIterMin=5, int nIterMax=100, bool _adjustLimits=true);
    static double romberg(const F& f, const double& x1, const double& x2, int& nIter, double relTol=1e-3, double absTol=0.0, int nIterMin=5, int nIterMax=100, bool _adjustLimits=true);
    static void trapezRefine(const F& f, const double& x1, const double &x2, const int& n, double& fInt);
    static void adjustLimits(const F& f, double& x1, double &x2, int nInt);
  };

  // Definite integration using trapezoidal method
  template <class F> double Integrator<F>::trapezoidal(const F& f, const double& x1, const double& x2, double tol, int nIterMin, int nIterMax, bool _adjustLimits) {
    double x1New=x1,x2New=x2;
    double fInt,fIntOld=0.0;
    int n;
    if (_adjustLimits) adjustLimits(f,x1New,x2New,100);
    for (n=1;n<=nIterMax;n++) {
      trapezRefine(f,x1New,x2New,n,fInt);
      if (n>=nIterMin)
        if ((fabs(fInt-fIntOld) < tol*fabs(fIntOld)) or (fInt==0.0 and fIntOld==0.0)) {
          //cout<<"Integrator::trapezoidal: n: "<<n<<", fInt: "<<fInt<<", fIntOld: "<<fIntOld<<"\n";
	  return fInt;
        }
      fIntOld=fInt;
    }

    cout<<"Integrator::trapezoidal: Maximum number of iterations exceeded\n";    

    return fInt;
    }

  // Definite integration using Simpsons method
  template <class F> double Integrator<F>::simpson(const F& f, const double& x1, const double& x2, double tol, int nIterMin, int nIterMax, bool _adjustLimits) {
    double x1New=x1,x2New=x2;
    double fInt,fIntT,fIntOld=0.0,fIntTOld=0.0;
    int n;
    if (_adjustLimits) adjustLimits(f,x1New,x2New,100);
    for (n=1;n<=nIterMax;n++) {
      trapezRefine(f,x1New,x2New,n,fIntT);
      fInt=(4.0*fIntT-fIntTOld)/3.0;
      if (n>=nIterMin)
	if ((fabs(fInt-fIntOld) < tol*fabs(fIntOld)) or (fInt==0.0 and fIntOld==0.0)) {
	  //cout<<"Integrator::simpson: n: "<<n<<", fInt: "<<fInt<<", fIntOld: "<<fIntOld<<", fIntT: "<<fIntT<<", fIntTOld: "<<fIntTOld<<"\n";
	  return fInt;
	}
      fIntOld=fInt;
      fIntTOld=fIntT;
    }

    cout<<"Integrator::simpson: Maximum number of iterations exceeded\n";    

    return fInt;
  }

  // Definite integration using Rombergs method
  //template <class F> double Integrator<F>::romberg(const F& f, const double& x1, const double& x2, int& nIter, double tol, int nIterMin, int nIterMax, bool _adjustLimits) {
  template <class F> double Integrator<F>::romberg(const F& f, const double& x1, const double& x2, int& nIter, double relTol, double absTol, int nIterMin, int nIterMax, bool _adjustLimits) {
    int K=5;
    double x1New=x1,x2New=x2;
    DP fInt,fIntOld=0,fIntErr,fIntT;
    Vec_DP fIntTVec(nIterMax),h(nIterMax+1),s_t(K),h_t(K);
    int i,n;

    if (_adjustLimits) adjustLimits(f,x1New,x2New,100);
    //if ((x1!=x1New) or (x2!=x2New)) cout<<"Integrator::romberg: Limits adjusted: ("<<x1<<","<<x2<<")->("<<x1New<<","<<x2New<<")\n";
    //cout<<"Integrator::romberg: Limits ("<<x1<<","<<x2<<")->("<<x1New<<","<<x2New<<")\n";
    h[0]=1.0;
    for (n=1;n<=nIterMax;n++) {
      trapezRefine(f,x1New,x2New,n,fIntT);
      fIntTVec[n-1]=fIntT;
      if (n >= K) {
	for (i=0;i<K;i++) {
	  h_t[i]=h[n-K+i];
	  s_t[i]=fIntTVec[n-K+i];
	}
	//cout<<"Integrator::romberg; n: "<<n<<"\n";
	polint(h_t,s_t,0.0,fInt,fIntErr);
	//cout<<"Integrator::romberg: n: "<<n<<", fInt: "<<fInt<<", fIntOld: "<<fIntOld<<"\n";
	//if (fabs(fIntErr) <= tol*fabs(fInt)) {
	if ((fabs(fInt-fIntOld) < relTol*fabs(fIntOld)) or (fabs(fInt-fIntOld) < absTol) or (fInt==0.0 and fIntOld==0.0)) {	
	  //cout<<"Integrator::romberg: n: "<<n<<", fInt: "<<fInt<<", fIntErr: "<<fIntErr<<", fIntT: "<<fIntTVec[n-1]<<", fIntTOld: "<<fIntTVec[n-2]<<"\n";
	  nIter=n;
	  return fInt;
	}
	fIntOld=fInt;
      }
      h[n]=0.25*h[n-1];
    }

    cout<<"Integrator::romberg: Maximum number of iterations exceeded\n";

    return fInt;
  }

  // Refine trapezoidal integration
  template <class F> void Integrator<F>::trapezRefine(const F& f, const double& x1, const double &x2, const int& n, double& fInt) {
    double x,dx,npf,fSum;
    int i,np;

    if (n==1) fInt=0.5*(x2-x1)*(f(x1)+f(x2));
    else {
      // Unclear if muliple initializers work - Fix!
      np=1; for (i=1;i<n-1;i++) np <<= 1; // Changed!
      npf=np;
      dx=(x2-x1)/npf;
      x=x1+0.5*dx;
      // Unclear if muliple initializers work - Fix!
      // How about multiple incrementors?
      fSum=0.0; for (i=0;i<np;i++,x+=dx) fSum+= f(x); // Changed
      fInt=0.5*(fInt+(x2-x1)*fSum/npf);
    }
  }

  // Adjust limits
  template <class F> void Integrator<F>::adjustLimits(const F& f, double& x1, double &x2, int nInt) {
    double x,dx,x1New=x1,x2New=x2;
    int i;
    dx=(x2-x1)/float(nInt);
    // Unclear if muliple initializers work - Fix!
    // How about multiple incrementors?
    x=x1; for (i=0;i<nInt;i++,x+=dx) { // Changed!
      //cout<<"Integrator::adjustLimits: x: "<<x<<", f(x): "<<f(x)<<"\n";
      if (f(x)==0.0) x1New=x;
      else break;
    }
    // Unclear if muliple initializers work - Fix!
    // How about multiple incrementors?
    x=x2; for (i=0;i<nInt;i++,x-=dx) { // Changed!
      //cout<<"Integrator::adjustLimits: x: "<<x<<", f(x): "<<f(x)<<"\n";
      if (f(x)==0.0) x2New=x;
      else break;
    }
    if (x1New<x2New) {
      x1=x1New;
      x2=x2New;
    }
    //cout<<"Integrator::adjustLimits: x1: "<<x1<<", x2: "<<x2<<"\n";
  }

  // Exponential integral integrand
  class EnIntegrand {
  public:
    double n;
    double x;
    EnIntegrand(double _n, double _x) : n(_n), x(_x) {};
    double operator ()(double t) const { return pow(t,-n)*exp(-x*t); }
  };

  // Root finder
  template <class Eq> class RootFinder {
  public:
    // Find root of 1-D equation using Newton-Raphsons method
    static double newtonRaphson(const Eq& eq, const double& x0, const double& xMin, const double& xMax, double xtol=1e-3, double ftol=1e99, int nIterMin=5, int nIterMax=100);
  };

  // Find root of 1-D equation using Newton-Raphsons method
  template <class Eq> double RootFinder<Eq>::newtonRaphson(const Eq& eq, const double& x0, const double& xMin, const double& xMax, double xtol, double ftol, int nIterMin, int nIterMax) {
    double f,dfdx,dx,x=x0;
    int n;

    for (n=1;n<=nIterMax;n++) {
      f=eq.f(x); dfdx=eq.dfdx(x);
      if (dfdx==0) { cout<<"RootFinder::newtonRaphson: Error: Zero derivative\n"; return x; }
      dx=-f/dfdx;
      if ((x==xMax and dx>0) or (x==xMin and dx<0)) { cout<<"RootFinder::newtonRaphson: Error: No solution within boundaries\n"; return x; }
      //cout<<"RootFinder::newtonRaphson: n: "<<n<<", f: "<<eq.f(x)<<", dfdx: "<<eq.dfdx(x)<<", x: "<<x<<", dx: "<<dx<<"\n";
      x=MIN(MAX(x+dx,xMin),xMax);
      if (fabs(dx/x)<xtol and fabs(dfdx*(x/f))<ftol and n>=nIterMin) return x;
    }

    cout<<"RootFiner::newtonRaphson: Warning: Maximum number of iterations exceeded\n";    

    return x;
  }

  // Function value
  class FunctionValue {
  private :
    // MPI serialization - New!
    friend class boost::serialization::access;
    template<class Archive> 
      void serialize (Archive & ar, const unsigned int version) 
      {
	ar & x;
	ar & f;
      }
    // *****************
  public :
    FunctionValue() : x(0), f(0) {}; // New!
    FunctionValue(double _x, double _f) : x(_x), f(_f) {}; // New!
    double x;
    double f;
  };

  // Function domain
  class FunctionDomain {
  private :
    // MPI serialization - New!
    friend class boost::serialization::access;
    template<class Archive> 
      void serialize (Archive & ar, const unsigned int version) 
      {
	ar & min;
	ar & max;
      }
    // *****************
  public :
    // Constructors
    FunctionDomain();
    FunctionDomain(const double& _min, const double& _max);
    double min;
    double max;
  };

  // Function value list iterator
  typedef vector<FunctionValue>::iterator FunctionValueIterator;
  typedef vector<FunctionValue>::const_iterator FunctionValueConstIterator;

  // 1D interpolator class
  class Interpolator1D {
  private :
    // MPI serialization - New!
    friend class boost::serialization::access;
    template<class Archive> 
      void serialize (Archive & ar, const unsigned int version) 
      {
	ar & pTable;
	ar & pDomain;
	ar & iLast;
      }
    // *****************
    vector<FunctionValue> pTable;
    FunctionDomain pDomain;
    unsigned long iLast;
  public :
    Interpolator1D() {};
    ~Interpolator1D() {};
    Interpolator1D(const vector<FunctionValue>& table) : pTable(table), pDomain(0,0), iLast(0) {} ;
    double operator ()(double x);
    double init(const vector<FunctionValue>& table);
    FunctionValueConstIterator findInt(const double& x) const;
    FunctionValueConstIterator findLong(const double& x) const;
    FunctionDomain domain() const;
    vector<FunctionValue>& table();
    double mem() const; // New!
  };

  // 1D function class
  class Function1D {
  public :
    virtual double operator()(double x) const { cout<<"Function1D::operator(): x: "<<x<<"\n"; } ;
  };

  // 1D tabulator class
  template <class F> class Tabulator1D {
  private :
    vector<FunctionValue> table;
    F f;
  public :
    Tabulator1D() {};
    ~Tabulator1D() {};
    Tabulator1D(const vector<double>& x, const F& _f);
    double operator ()(double x) const;
    void init(const vector<double>& x, const F& _f);
    FunctionValueConstIterator find(const double& x) const;
  };

  template <class F> Tabulator1D<F>::Tabulator1D(const vector<double>& x, const F& _f) : f(_f) {
    int i;
    table.resize(x.size());
    for (i=0;i<table.size();i++) { table[i].x=x[i]; table[i].f=f(x[i]); }
  }

  template <class F> FunctionValueConstIterator Tabulator1D<F>::find(const double& x) const {
    long iBegin,iMiddle,iEnd;
  
    if (table.size()!=0) {
      iBegin=0; iEnd=table.size()-1; iMiddle=(iBegin+iEnd)/2;
      if (x>=table[iBegin].x && x<table[iEnd].x) {
	while (iEnd-iBegin>1) {
	  if (x<table[iMiddle].x) {
	    iEnd=iMiddle;
	    iMiddle=(iBegin+iEnd)/2;
	  }
	  else {
	    iBegin=iMiddle;
	    iMiddle=(iBegin+iEnd)/2;
	  }
	}
	if (x==table[iBegin].x) return table.begin()+iBegin;
      }
      else if (x==table[iEnd].x) return table.begin()+iEnd;
    }
  
    return table.end();
  }

  template <class F> void Tabulator1D<F>::init(const vector<double>& x, const F& _f) {
    //cout<<"Tabulator1D::init, x.size: "<<x.size()<<"\n";
    int i;
    f=_f;
    table.resize(x.size());
    for (i=0;i<table.size();i++) { 
      table[i].x=x[i]; table[i].f=f(x[i]);
      //cout<<"Tabulator1D::init: table.x: "<<table[i].x<<", table.f: "<<table[i].f<<"\n";
    }
  }
  
  template <class F> double Tabulator1D<F>::operator ()(double x) const {
    FunctionValueConstIterator item=find(x);
    
    if (item==table.end()) { 
      //cout<<"Tabulator1D::operator(): x="<<x<<" not found in table, evaluating function, f="<<f(x)<<"\n"; 
      return f(x); }
    else { 
      //cout<<"Tabulator1D::operator(): x="<<x<<", table.x="<<item->x<<", table.f="<<item->f<<"\n"; 
      return item->f; }
  }

  // Function domain (for use in nmewt)
  class Domain {
  public:
    // Constructors
    Domain();
    Domain(DP _min, DP _max);
    // Min
    DP min;
    // Max
    DP max;
  };
  typedef const NRVec<Domain> Vec_I_DOM;
  typedef NRVec<Domain> Vec_DOM, Vec_O_DOM, Vec_IO_DOM;

  // Slight rewrite of mnewt in Numerical Recipes passing usrfun as function pointer together 
  // with an optional pointer to an object for non-static class member function
  //void mnewt(const int ntrial, Vec_IO_DP& x, const DP tolx, const DP tolf, const Vec_I_DOM& domain, const DP damp, void (*usrfun)(Vec_I_DP&, Vec_O_DP&, Mat_O_DP&, void*), void* object);
  void mnewt(const int ntrial, Vec_IO_DP& x, const Vec_I_DOM& domain, const DP damp, void (*usrfun) (Vec_I_DP&, Vec_O_DP&, Mat_O_DP&, void*), bool (*xconv) (Vec_I_DP&, Vec_I_DP&, void*), bool (*fconv) (Vec_I_DP&, void*), void* object);

  // Rewrite of tridag for tridiagonal matrix shifted on step upwards (i.e. skipping the first equation)
  // and with one additional (trivial) equation for a[n-1]
  void tridag_upshift(Vec_I_DP &a, Vec_I_DP& b, Vec_I_DP& c, Vec_I_DP& r, Vec_O_DP& u);
  
  // Test if double is a number
  bool isnumber(double x);

  inline int min(const int& a, const int& b)
  {return b < a ? (b) : (a);}

  // Exponential with lower (non-zero) bound
  inline double expLB(double x, double lowerBound=1e-200) 
  { return MAX(exp(x),lowerBound); }
}


#endif
