#include <vector>
#include <math.h>
#include "mc-math.h"
#include "mc-physics.h"
#include <boost/math/special_functions/expint.hpp>

using namespace std;
using namespace math;
using namespace physics;

// Number density of next ionization state (Saha equation)
double physics::nIonLTE (const double& n1, const double& ne, const double& T, const double& Eion, const double& g1, const double& g2) {
  double n2;
  //n2=(n1/ne)*2*pow(pow(h,2)/(2*pi*me*k*T),-3.0/2.0)*(g2/g1)*exp(-Eion/(k*T));
  n2=(n1/ne)*2*pow(pow(h,2)/(2*pi*me*k*T),-3.0/2.0)*(g2/g1)*expLB(-Eion/(k*T)); // Changed! - Testing!
  return n2;
}

// Number density of next excitation state (Boltzman equation)
double physics::nBoundLTE (const double& n1, const double& T, const double& E1, const double& E2, const double& g1, const double& g2) {
  double n2;
  //n2=n1*(g2/g1)*exp(-(E2-E1)/(k*T));
  n2=n1*(g2/g1)*expLB(-(E2-E1)/(k*T)); // Changed! - Testing!
  return n2;
}

// NLTE number densities of bound states ignoring ionization and recombination
/*vector<double> physics::nBoundNLTE (const double& nIon, const double& T, const double& ne, const vector<vector<double>>& J, const vector<vector<double>>& A, const vector<vector<double>>& upsilon, const vector<vector<double>>& E, const vector<double>& g, const vector<vector<double>> beta) {
  DP d;
  int n=E.size();
  Vec_INT indx(n);
  Mat_DP B(n,n),C(n,n),R(n,n);
  Mat_DP a(n,n);
  Vec_DP b(n,n);
  vector<double> nBound(n);

  // Calculate the Einstein B-coefficients
  for (i=0;i<B.nrows();i++) {
    for (j=0;j<B.ncols();j++) {
      if (i!=j) {
	nu=fabs(E[i]-E[j])/h;
	B[i][j]=A[i][j]*(c**2/(2*h*nu**3));
	B[j][i]=B[i][j]*(g[i]/g[j]);
      }
      else B[i][j]=0;
    }
  }

  // Calculate the collisional transition rates
  for (i=0;i<C.nrows();i++) {
    for (j=0;j<C.ncols();j++) {
      if (i!=j) {
	C[i][j]=8.629e-6*upsilon[i][j]*ne/(pow(T,0.5)*g[i]);
	C[j][i]=C[i][j]*(g[i]/g[j])*exp(-(E[i]-E[j])/(k*T));
      }
      else C[i][j]=0;
    }
  } 
  
  // Calculate total transition rates
  for (i=0;i<R.nrows();j++) {
    for (j=0;j<R.ncols();j++) {
      if (i!=j) R[i][j]=A[i][j]*beta[i][j]+B[i][j]*J[i][j]*beta[i][j]+C[i][j];
      else R[i][j]=0;
    }
  }
  
  // Set the rate equations
  for(i=0;i<a.nrows();i++) {
    a[i][i]=0;
    for (j=0;j<a.ncols();j++) if (i!=j) a[i,i]=a[i,i]-R[i,j];
    for (j=0;j<a.ncols();j++) if (i!=j) a[i,j]=R[j,i];
    b[i]=0;
  }
  // Replace the last equation with number conservation
  for (i=0;i<a.ncols();i++) a[a.nrows()-1,i]=1;
  b[b.size()]=1;
    
  // Solve the linear equation system using LU decomposition
  ludcmp(a,indx,d);
  lubksb(a,indx,b);

  for (i=0;i<nBound.size();i++) nBound[i]=nIon*b[i]
  
  return nBound*/

// Partition function
double physics::ULTE(const double& T, const vector<double>& E, const vector<double>& g) {
  int i;
  float U=0;
  for (i=0;i<E.size() && i<g.size();i++) {
    U=U+g[i]*exp(-E[i]/(k*T));
  }
  return U;
}

// Sobolev optical depth
double physics::tauSobolev(const double& Eul, const double& nl, const double& nu, const double& A, const double& gu, const double& gl, const double& drdv ) {
  double lambda,flu,tau;
  lambda=h*c/Eul;
  tau=(1/(8*pi))*A*pow(lambda,3)*(gu/gl)*drdv*nl*(1-(gl/gu)*(nu/nl));
  return tau;
}

// Sobolev optical depth lower state number derivative
double physics::dTauSobolevdNl(const double& Eul, const double& nl, const double& nu, const double& A, const double& gu, const double& gl, const double& drdv ) {
  double lambda,dTaudNl;
  lambda=h*c/Eul;
  dTaudNl=(1/(8*pi))*A*pow(lambda,3)*(gu/gl)*drdv;
  return dTaudNl;
}

// Sobolev optical depth upper state number derivative
double physics::dTauSobolevdNu(const double& Eul, const double& nl, const double& nu, const double& A, const double& gu, const double& gl, const double& drdv ) {
  double lambda,dTaudNu;
  lambda=h*c/Eul;
  dTaudNu=-(1/(8*pi))*A*pow(lambda,3)*drdv;
  return dTaudNu;
}

// Free-free emissivity
// Eq. 5.14b in Ribicki & Lightman * (1/4pi)
double physics::FreeFree::eta(const double& ne, const double& n, const double& Z, const double& T, const double& nu) {
  // Set gaunt factor to 1
  double g=1.0;
  return (1.0/(4*pi))*6.8e-38*pow(Z,2)*ne*pow(cm,3)*n*pow(cm,3)*pow(T,-0.5)*exp(-(h*nu)/(k*T))*g*erg*pow(cm,-3);
}

// Free-free opacity (assuming thermal emission -> k_nu=j_nu/B_nu)
// Eq. 5.18b in Ribicki & Lightman
double physics::FreeFree::kappa(const double& ne, const double& n, const double& Z, const double& T, const double& nu, const double rho) {
  // Set gaunt factor to 1
  double g=1.0;
  //return (0.018*pow(T,-1.5)*pow(Z,2)*ne*n*pow(nu,-2)*(1-exp(-(h*nu)/(k*T)))*g*pow(cm,-1))/rho;
  //return (3.7e8*pow(T,-0.5)*pow(Z,2)*ne*pow(cm,3)*n*pow(cm,3)*pow(nu,-3)*(1-exp(-(h*nu)/(k*T)))*g*pow(cm,-1))/rho;
  // Ensure that Kirchoffs law is satisfied exactly
  //return eta(ne,n,Z,T,nu)/(BlackBody::B(nu,T)*rho);
  // This should be the same but avoids division by zero
  return (1.0/(4*pi))*6.8e-38*pow(Z,2)*ne*pow(cm,3)*n*pow(cm,3)*pow(T,-0.5)*g*erg*pow(cm,-3)*(1-exp(-(h*nu)/(k*T)))*pow(c,2)/(2*h*pow(nu,3)*rho);
}

// Free-free cooling rate
// Eq. 5.15b in Ribicki & Lightman
double physics::FreeFree::cooling(const double& ne, const double& n, const double& Z, const double& T) {
  // Set gaunt factor to 1
  double g=1.0;
  return 1.4e-27*pow(T,0.5)*pow(Z,2)*ne*pow(cm,3)*n*pow(cm,3)*g*erg*pow(cm,-3);
}

// Free-free heating rate (assuming diluted BB radiation fiels)
double physics::FreeFree::heatingDBB(const double& ne, const double& n, const double& Z, const double& nu, const double& T, const double& TR, const double& W) {
  double H,f1,f2,fMin=1e-6,fMax=1e6,g=1.0;
  //f1=exp(h*nu/(k*T))-1;
  //f2=exp(h*nu/(k*TR))-1;
  //cout<<"physics::FreeFree::heatingDiluteBB, nu: "<<nu<<", lambda: "<<(c/nu)/AA<<", T: "<<T<<", TR: "<<TR<<", W: "<<W<<", ne: "<<ne<<", n: "<<n<<", Z: "<<Z<<", j: "<<FreeFree::j(ne,n,Z,T,nu)<<", f1: "<<exp(h*nu/(k*T))-1<<", f2: "<<exp(h*nu/(k*TR))-1<<", f2/f1: "<<f1/f2<<", f1/f2: "<<TR/T<<", f1/f2: "<<exp(-(h*nu/k)*(1/TR-1/T))<<"\n";
  //if ((f1<fMin) and (f2<fMin)) H=W*FreeFree::j(ne,n,Z,T,nu)*TR/T;
  //else 
    //if ((f1>fMax) and (f2>fMax)) H=W*FreeFree::j(ne,n,Z,T,nu)*exp(-(h*nu/k)*(1/TR-1/T));
    //if ((f1>fMax) and (f2>fMax)) H=(1.0/(4*pi))*6.8e-38*pow(Z,2)*ne*pow(cm,3)*n*pow(cm,3)*pow(T,-0.5)*exp(-(h*nu)/(k*TR))*g*erg*pow(cm,-3);
    //else H=W*FreeFree::j(ne,n,Z,T,nu)*(exp(h*nu/(k*T))-1)/(exp(h*nu/(k*TR))-1);
  H=W*FreeFree::kappa(ne,n,Z,T,nu,1.0)*4*pi*BlackBody::B(nu,TR);
  //cout<<"physics::FreeFree::heatingDiluteBB, H: "<<H<<"\n";
  return H;
}

// Total frequency integrated heating rate (assuming diluted BB radiation field)
double physics::FreeFree::totalHeatingDBB(const double& ne, const double& n, const double& Z, const double& T, const double& TR, const double& W, const double& tol) {
  double Hint,HintOld,nu,dnu,nu1,nu2,xMin=1e-3,xMax=1e2;
  double* dH=NULL;
  int i,nIter;
  //nu1=MIN(BlackBody::nuMax(T),BlackBody::nuMax(TR))/2;
  //nu1=c/(1e6*AA);
  //nu2=MAX(BlackBody::nuMax(T),BlackBody::nuMax(TR))*2;
  //nu2=c/(1e3*AA);
  //x1=(h*nu)/(k*T);
  //x2=(h*nu)/(k*T);
  nu1=MIN(xMin*(k*T)/h,xMin*(k*TR)/h);
  nu2=xMax*(k*TR)/h;
  //cout<<"FreeFree::totalHeatingDBB: nu1: "<<nu1<<", nu2: "<<nu2<<", x1(T): "<<(h*nu1)/(k*T)<<", x1(TR): "<<(h*nu1)/(k*TR)<<", x2: "<<(h*nu2)/(k*TR)<<"\n";
  nIter=10;
  Hint=0;
  do {
    dnu=(nu2-nu1)/float(nIter);
    HintOld=Hint;
    Hint=0;
    dH = new double[nIter];
    //pragma omp parallel for private(nu)
    for (i=0;i<nIter;i++) {
      nu=nu1+i*dnu;
      dH[i]=dnu*(heatingDBB(ne,n,Z,nu,T,TR,W)+heatingDBB(ne,n,Z,nu+dnu,T,TR,W))/2;
    }
    for (i=0;i<nIter;i++) Hint+=dH[i];
    nIter=2*nIter;
    //nu1=nu1/2;
    //nu2=nu2*2;
    //delete(dH);
    delete[] dH;
    //cout<<"physics::FreeFree::heatingDiluteBB: Hint: "<<Hint<<", dH: "<<dH<<", nIter: "<<nIter<<"\n";
  }
  while (fabs((Hint-HintOld)/Hint)>tol);
  //cout<<"physics::FreeFree::heatingDiluteBB: Hint: "<<Hint<<", nIter: "<<nIter<<"\n";

  return Hint;
}

// Frequency that gives the (fractional) frequency intergrated emissivity z over the intervall (nuMin,nu)
//double physics::FreeFree::etaNu(const double& z, const double& T) {
double physics::FreeFree::etaNu(const double& z, const double& T, double nuMin, double nuMax) {
  double nu;
  //nu=(-k*T/h)*log(1-z);
  if (nuMin==0 and nuMax==1e99) nu=(-k*T/h)*log(1-z);
  else nu=(-k*T/h)*log(exp(-(h*nuMin)/(k*T))-z*(exp(-(h*nuMin)/(k*T))-exp(-(h*nuMax)/(k*T))));
  return nu;
}

// Photo-ionization cross-section (Hydrogenic approximation)
double physics::BoundFree::sigmaRadH(const double& C, const double& Xi0, const double& E, const double& nu) {
  double sigma,nu0,n,g=1.0;
  n=pow(1-E/Xi0,-0.5);
  nu0=(Xi0-E)/h;
  if (nu>=nu0) sigma=7.9E-18*pow(nu/nu0,-3)*g*n*pow(fabs(C)+1,-2)*pow(cm,2);
  else sigma=0.0;
  return sigma;
}

// Recombination coefficient (Spnotaneous; Hydrogenic approximation)
double physics::BoundFree::alfaSPH(const double& C, const double& Xi0, const double& E, const double& g1, const double& g2, const double& T, double nuMin, double nuMax, bool modified) {
  double nu0=(Xi0-E)/h;
  double x0=(h*nu0)/(k*T);
  double sigma0,phi,factor,xMin,xMax,alfa,alfaMod;

  nuMin=MAX(nuMin,nu0);
  if (nuMin>=nuMax) return 0.0;

  sigma0=sigmaRadH(C,Xi0,E,nu0);

  phi=1/nIonLTE(1.0,1.0,T,(Xi0-E),g1,g2);
  
  factor=4*pi*phi*sigma0*(2*pow(nu0,3)/pow(c,2));

  xMin=(h*nuMin)/(k*T);

  //if (modified) alfa=factor*exp(-xMin)/x0;
  //else alfa=factor*(-boost::math::expint(-xMin));

  alfa=factor*(-boost::math::expint(-xMin));
  if (modified) alfaMod=factor*exp(-xMin)/x0;

  if (nuMax!=1e99) {

    xMax=(h*nuMax)/(k*T);

    //if (modified) alfa-=factor*exp(-xMax)/x0;
    //else alfa-=factor*(-boost::math::expint(-xMax));

    alfa-=factor*(-boost::math::expint(-xMax));
    if (modified) alfaMod-=factor*exp(-xMax)/x0;
  }
  
  if (modified) alfa=alfaMod-alfa;
  
  return alfa;
}

// Photo-ionization cross-section (Lucy test case)
double physics::BoundFree::sigmaRadL(const double& C, const double& Xi0, const double& E, const double& nu) {
  double sigma,nu0,n;
  vector<double> sigma0(6);
  sigma0[0]=6.23e-18*pow(cm,2);
  sigma0[1]=1.15e-17*pow(cm,2);
  sigma0[2]=2.15e-17*pow(cm,2);
  sigma0[3]=2.92e-17*pow(cm,2);
  sigma0[4]=3.70e-17*pow(cm,2);
  sigma0[5]=4.48e-17*pow(cm,2);
  n=round(pow(1-E/Xi0,-0.5));
  nu0=(Xi0-E)/h;
  if (nu>=nu0) {
    if (n<=6) sigma=sigma0[int(n-1)]*pow(nu/nu0,-3);
    else sigma=sigmaRadH(C,Xi0,E,nu);
  }
  else sigma=0.0;
  return sigma;
}

// Collision cross-section (calculated from collisional strength)
// Not a cross-section - rate per particle !!!
double physics::BoundFree::sigmaColl(const double& Xi, const double& g1, const double& omega, const double& T, const double& ne) {
  //return ne*pow(2*pi/(k*T),0.5)*(pow(hBar,2)/pow(me,1.5))*(1/g1)*exp(-Xi/(k*T))*omega;
  return ne*pow(2*pi/(k*T),0.5)*(pow(hBar,2)/pow(me,1.5))*(1/g1)*expLB(-Xi/(k*T))*omega; // Changed! - Testing!
}

// Collisional cross-section (Seaton approximation)
// Not a cross-section - rate per particle !!!
double physics::BoundFree::sigmaCollS(const double& C, const double& Xi, const double& sigmaRad0, const double& T, const double& ne) {
  //cout<<"BoundFree::SigmaCollS\n";
  double sigma,nu0,n,g;
  double u0=Xi/(k*T);
  if ((fabs(C)+1)==1.0) g=0.1;
  else
    if ((fabs(C)+1)==2.0) g=0.2;
    else g=0.3;
  //sigma=1.55e13*ne*pow(T,-0.5)*g*sigmaRad0*exp(-u0)/u0;
  sigma=1.55e13*ne*pow(T,-0.5)*g*sigmaRad0*expLB(-u0)/u0; // Changed! - Tesing!
  return sigma;
}

/*
// Get opacity (per volume) from cross-section
double physics::BoundFree::k(const double& sigma, const double& n) {
  double _kappa;
  _kappa=sigma*n;
  return _kappa;
}

// Get opacity (per volume) from cross-section (corrected for stimulated emission)
double physics::BoundFree::kST(const double& sigma, const double& n, const double& nIon, const double& ne, const double& g, const double& gIon, const double& Xi, const double& T, const double& nu) {
  double _kappa,nLTE;
  xLTE=nIonLTE (1.0,ne,T,Xi,g,gIon);
  _kappa=sigma*(n-(nIon/xLTE)*exp(-(h*nu)/(k*T)));
  return _kappa;
}

// Get emissivity from cross-section (spontaneous emission only)
double physics::BoundFree::eta(const double& sigma, const double& nIon, const double& ne, const double& g, const double& gIon, const double& Xi, const double& T, const double& nu) {
  double _kappa,xLTE;
  xLTE=nIonLTE(1.0,ne,T,Xi,g,gIon);
  _eta=sigma*(nIon/xLTE)*((2*h*pow(nu,3))/(pow(c,2)))*exp(-(h*nu)/(k*T));
  return _eta;
}

// Get emissivity from cross-section (including stimulated emission)
double physics::BoundFree::etaST(const double& sigma, const double& nIon, const double& ne, const double& g, const double& gIon, const double& Xi, const double& T, const double& J, const double& nu) {
  double _kappa,xLTE;
  xLTE=nIonLTE(1.0,ne,T,Xi,g,gIon);
  _eta=sigma*(nIon/xLTE)*((2*h*pow(nu,3))/(pow(c,2)))*exp(-(h*nu)/(k*T));
  _eta=_eta+sigma*(nIon/xLTE)*J*exp(-(h*nu)/(k*T));
  return _eta;
  }

// Frequency that gives the (fractional) frequency intergrated (spontaneous) emissivity z over the intervall (0,nu)
double physics::BoundFree::etaNu(const double& z, const double& T, const double Xi0, const double& E) {
  nu0=(Xi0-E)/h;
  double nu=nu0*(1-(k*T)/(h*nu0))*log(z);
  return nu;
  }
*/

/*double physics::BoundBound::gamma(const double& g, const double& u0) {
  vector<FunctionValue> table(14);
  table[0].x=1e-99; table[0].f=22.4;
  table[1].x=1e-10; table[1].f=22.4;
  table[2].x=1e-9; table[2].f=20.1;
  table[3].x=1e-8; table[3].f=17.8;
  table[4].x=1e-7; table[4].f=15.5;
  table[5].x=1e-6; table[5].f=13.2;
  table[6].x=1e-5; table[6].f=10.9;
  table[7].x=1e-4; table[7].f=8.63;
  table[8].x=1e-3; table[8].f=6.34;
  table[9].x=1e-2; table[9].f=4.08;
  table[10].x=1e-1; table[10].f=2.01;
  table[11].x=1.0; table[11].f=0.596;
  table[12].x=1e1; table[12].f=0.0916;  
  table[13].x=1e99; table[13].f=0.0;
  Interpolator1D f(table);
  return MAX(g,0.276*f(u0));
  }
*/

double physics::BoundBound::gammaNeu(const double& u0) {
  vector<FunctionValue> table(20);
  table[0].x=1e-99; table[0].f=22.4;
  table[1].x=1e-10; table[1].f=22.4;
  table[2].x=1e-9; table[2].f=20.1;
  table[3].x=1e-8; table[3].f=17.8;
  table[4].x=1e-7; table[4].f=15.5;
  table[5].x=1e-6; table[5].f=13.2;
  table[6].x=1e-5; table[6].f=10.9;
  table[7].x=1e-4; table[7].f=8.63;
  table[8].x=1e-3; table[8].f=6.34;
  table[9].x=1e-2; table[9].f=4.08;
  table[10].x=0.01; table[10].f=1.160;
  table[11].x=0.02; table[11].f=0.956;
  table[12].x=0.04; table[12].f=0.758;
  table[13].x=0.1; table[13].f=0.493;
  table[14].x=0.2; table[14].f=0.331;
  table[15].x=0.4; table[15].f=0.209;
  table[16].x=1.0; table[16].f=0.100;
  table[17].x=2.0; table[17].f=0.063;
  table[18].x=4.0; table[18].f=0.040;
  table[19].x=10.0; table[19].f=0.023;
  Interpolator1D f(table);
  if (u0>10.0) return 0.066*pow(u0,-0.5);
  else return 0.276*f(u0);
}

double physics::BoundBound::gamma(const double& u0, const double& C) {
  double g=0.2;
  if (C==0) return BoundBound::gammaNeu(u0);
  else return MAX(g,BoundBound::gammaNeu(u0));
}

// Collision cross-section (calculated from collisional strength)
// Not a cross-section - rate per particle !!!
double physics::BoundBound::sigmaColl(const double& Xi, const double& g1, const double& omega, const double& T, const double& ne) {
  return ne*pow(2*pi/(k*T),0.5)*(pow(hBar,2)/pow(me,1.5))*(1/g1)*exp(-Xi/(k*T))*omega;
}

// Collisional cross-section (Regemorter approximation)
// Not a cross-section - rate per particle !!!
double physics::BoundBound::sigmaCollR(const double& Xi, const double& g1, const double& g2, const double& A, const double& C, const double& T, const double& ne) {
  double sigma,Xi0H=13.595*eV;
  double C0=pi*pow(a0,2)*pow((8*k)/(me*pi),0.5);
  double u0=Xi/(k*T);
  double f=A*((epsilon0*me*pow(c,3))/(2*pi*pow(Xi/h,2)*pow(ee,2)))*(g2/g1);
  //double f=A*((me*pow(c,3))/(4*pi*2*pi*pow(Xi/h,2)*pow(ee,2)))*(g2/g1);
  sigma=C0*ne*pow(T,0.5)*(14.5*f*pow(Xi0H/Xi,2))*u0*exp(-u0)*BoundBound::gamma(u0,C);
  //cout<<"BoundBound::sigmaCollR: XI: "<<Xi<<", g1: "<<g1<<", g2: "<<g2<<", A: "<<A<<", C: "<<C<<", T: "<<T<<", ne: "<<ne<<", u0: "<<u0<<", sigma: "<<sigma<<"\n";
  return sigma;
}

// Intensity
double physics::BlackBody::B(const double& nu, const double& T) {
  return (2*h*pow(nu,3)/pow(c,2))/(exp(h*nu/(k*T))-1);
}

// Frequency integrated intensity
double physics::BlackBody::B(const double& nu1, const double& nu2, const double& T, const double& tol) {
  double Bint,BintOld,nu,dnu;
  double* dB=NULL;
  int i,n;
  n=10;
  Bint=0;
  do {
    dnu=(nu2-nu1)/float(n);
    BintOld=Bint;
    Bint=0;
    dB = new double[n];
    //pragma omp parallel for private(nu)
    for (i=0;i<n;i++) {
      nu=nu1+i*dnu;
      dB[i]=dnu*(B(nu,T)+B(nu+dnu,T))/2;
    }
    for (i=0;i<n;i++) Bint+=dB[i];
    n=2*n;
    //delete(dB);
    delete[] dB;
  }
  while (fabs((Bint-BintOld)/Bint)>tol);

  return Bint;
}

// Total frequency integrated flux intensity
double physics::BlackBody::B(const double& T) {
  return sigma*pow(T,4)/pi;
}

// Temperature derivative of intensity
double physics::BlackBody::dBdT(const double& nu, const double& T) {
  //return (2*h*pow(nu,3)/pow(c,2))*(h*nu/k)/pow(T*(exp(h*nu/(k*T))-1),2);
  // Corrected! Should be like this, right!
  double exponent=h*nu/(k*T);
  if (exponent<100) return (2*h*pow(nu,3)/pow(c,2))*exp(exponent)*(h*nu/k)/pow(T*(exp(exponent)-1),2);
  else return (2*h*pow(nu,3)/pow(c,2))*exp(-exponent)*(h*nu/k)/pow(T,2);
}

// Temperature derivative of frequency integrated intensity - New!
double physics::BlackBody::dBdT(const double& nu1, const double& nu2, const double& T, const double& tol) {
  double dBdTint,dBdTintOld,nu,dnu;
  double* ddBdT=NULL;
  int i,n;
  n=10;
  dBdTint=0;
  do {
    dnu=(nu2-nu1)/float(n);
    dBdTintOld=dBdTint;
    dBdTint=0;
    ddBdT = new double[n];
    //pragma omp parallel for private(nu)
    for (i=0;i<n;i++) {
      nu=nu1+i*dnu;
      ddBdT[i]=dnu*(dBdT(nu,T)+dBdT(nu+dnu,T))/2;
    }
    for (i=0;i<n;i++) dBdTint+=ddBdT[i];
    n=2*n;
    //delete(dB);
    delete[] ddBdT;
  }
  while (fabs((dBdTint-dBdTintOld)/dBdTint)>tol);

  return dBdTint;
}

// Temperature derivative of total frequency integrated flux intensity
double physics::BlackBody::dBdT(const double& T) {
  return 4*sigma*pow(T,3)/pi;
}

// Peak intensity frequency
double physics::BlackBody::nuMax(const double& T) {
  return 2.82*k*T/h;
}

double physics::BlackBody::nuB(const double& _B, const double& T, const double& nu1, const double& bTol ) {
  int i,iTrial=100;
  double damp=0.7,small,f,dfdnu,nu2,nu2Old;
  nu2=nuMax(T);
  small=nu2/1e6;
  for (i=0;i<iTrial;i++) {
    f=_B-B(nu1,nu2,T);
    if (fabs(f/_B)<=bTol) break;
    dfdnu=-B(nu2,T);
    nu2Old=nu2;
    nu2=nu2-damp*f/dfdnu;
    if (nu2<small) nu2=small;
  }
  if (i==iTrial) { cerr<<"physics:BlackBody::nuB: Error\n"; return -1; }
  
  return nu2;
}

// Flux (isotropic radiation from a sphere)
double physics::BlackBody::F(const double& nu, const double& T) {
  return pi*B(nu,T);
}

// Frequency integrated flux (isotropic radiation from a sphere)
double physics::BlackBody::F(const double& nu1, const double& nu2, const double& T, const double& tol) {
  return pi*B(nu1,nu2,T,tol);
}

// Total frequency integrated flux (isotropic radiation from a sphere)
double physics::BlackBody::F(const double& T) {
  return sigma*pow(T,4);
}

// Frequency intervall corresponding to a given intergrated flux
double physics::BlackBody::nuF(const double& F, const double& T, const double& nu1, const double& bTol) {
  return nuB(F/pi,T,nu1,bTol);
}

// Calculate late new mass fractions in the decay chain of 56Ni
double physics::decay56Ni(double& X56Ni, double& X56Co, double& X56Fe, const double& t1, const double& t2) {
  double X56NiNew,X56CoNew,X56FeNew;
  double mprec=1e-15;

  // Calculate the new mass fraction of 56Ni
  X56NiNew=X56Ni*exp(-(t2-t1)/t56Ni);
  // Caclulate the new mass fraction of 56Co
  X56CoNew=X56Co*exp(-(t2-t1)/t56Co)+X56Ni*((exp(-(t2-t1)/t56Co)-exp(-(t2-t1)/t56Ni))/(1-t56Ni/t56Co));
  // Caclulate the new mass fraction of 56Fe
  if (((X56Ni+X56Co)-(X56NiNew+X56CoNew))/(X56Ni+X56Co)>mprec) X56FeNew=X56Fe+((X56Ni+X56Co)-(X56NiNew+X56CoNew));
  else X56FeNew=X56Fe;

  // Update
  X56Ni=X56NiNew;
  X56Co=X56CoNew;
  X56Fe=X56FeNew;
}

// Calculate late new mass fractions in the decay chain of 57Ni
double physics::decay57Ni(double& X57Ni, double& X57Co, double& X57Fe, const double& t1, const double& t2) {
  double X57NiNew,X57CoNew,X57FeNew;
  double mprec=1e-15;

  // Calculate the new mass fraction of 57Ni
  X57NiNew=X57Ni*exp(-(t2-t1)/t57Ni);
  // Caclulate the new mass fraction of 57Co
  X57CoNew=X57Co*exp(-(t2-t1)/t57Co)+X57Ni*((exp(-(t2-t1)/t57Co)-exp(-(t2-t1)/t57Ni))/(1-t57Ni/t57Co));
  // Caclulate the new mass fraction of 57Fe
  if (((X57Ni+X57Co)-(X57NiNew+X57CoNew))/(X57Ni+X57Co)>mprec) X57FeNew=X57Fe+((X57Ni+X57Co)-(X57NiNew+X57CoNew));
  else X57FeNew=X57Fe;

  // Update
  X57Ni=X57NiNew;
  X57Co=X57CoNew;
  X57Fe=X57FeNew;
}

  // Calculate late new mass fractions in the decay chain of 44Ti
double physics::decay44Ti(double& X44Ti, double& X44Ca, const double& t1, const double& t2) {
  double X44TiNew,X44CaNew;
  double mprec=1e-15;

  // Calculate the new mass fraction of 44Ti
  X44TiNew=X44Ti*exp(-(t2-t1)/t44Ti);
  // Caclulate the new mass fraction of 44Ca
  if ((X44Ti-X44TiNew)/X44Ti>mprec) X44CaNew=X44Ca+(X44Ti-X44TiNew);
  else X44CaNew=X44Ca;

  // Update
  X44Ti=X44TiNew;
  X44Ca=X44CaNew;
}
