#include <vector>
#include <math.h>
#include <string.h>
#include <unistd.h>
#include <omp.h>
#include "mc-util.h"
#include "mc-math.h"
#include "mc-physics.h"
#include "mc.h"
#include "msolver.h"
#include "eos.h"
#include "diffsolver.h"

using namespace util;
using namespace math;
using namespace physics;
using namespace mc;
//using namespace grid;
using namespace msolver;
using namespace eos;
using namespace diffsolver;

enum EPacketMethod { epLBased, epEBased };

class TransBBAllSelector : public TransBBSelector {
  virtual bool use(TransBBData* trans) { return true; }
};

class TransBBNoneSelector : public TransBBSelector {
  virtual bool use(TransBBData* trans) { return false; }
};

// Create statistics file name
string statFile(const string& statDir, const string& prefix, const int& index) {
  stringstream statFile;
  statFile.str("");
  statFile<<statDir<<"/"<<prefix<<"-"<<index<<".dat";
  return statFile.str();
}

// Calculate (Rosseland mean) optical depth in cells (on a single computer) - New!
// Does not work for subcells - Fix!
vector<double> getTauCellSingle(const Grid& grid) {
  CellSS* cell;
  vector<double> tauCell(grid.cell.size());
  int i;
 
  // Calculate Rosseland mean optical depth within each cell
  #pragma omp parallel for private(cell)
  for (i=0;i<grid.cell.size();i++) { 
    cell = (CellSS*) grid.cell[i];
    tauCell[i]=cell->rossContKappa(Processes(pEScatter|pBoundFree|pFreeFree))*cell->rho()*(cell->rMax()-cell->rMin());
  }

  return tauCell;
}

// Calculate (Rosseland mean) optical depth in cells (on a cluster using MPI) - New!
// Subcellsand containers: Works
// Does not work for subcells (see below) - Fix!
vector<double> getTauCellCluster(const Grid& grid, boost::mpi::communicator* mpiCom) {
  CellSS* cell;
  vector<double> tauCell(grid.cell.size());
  // Should be configurable!
  double boundFreeCutOff=1e-3;
  int i,nResponsible=0;

  // Determine which processes are responsible for the cells
  for (i=0;i<grid.cell.size();i++) grid.cell[i]->responsible=grid.cell[i]->index()%mpiCom->size();

  // Count number of cells the process is responsible for
  for (i=0;i<grid.cell.size();i++) if (grid.cell[i]->responsible==mpiCom->rank()) nResponsible++;

  // For each cell the process is responsible for
  #pragma omp parallel for private(cell)
  for (i=0;i<grid.cell.size();i++) { 
    if (grid.cell[i]->responsible==mpiCom->rank()) {
      // Set number of threads to use in nested omp parallelization
      omp_set_num_threads((omp_get_num_threads()-1)/nResponsible+1);
      // If needed - Update bound-free opacity cut-off and tables in advance - Testing!
      // This needs to be done for thesubcells as well!!!
      if (grid.cell[i]->boundFreeCutOffUpdate) { grid.cell[i]->updateBoundFreeCutOffOMP(false,boundFreeCutOff); grid.cell[i]->boundFreeCutOffUpdate=false; }
      //if (grid.cell[i]->boundFreeCutOffUpdate) { grid.cell[i]->updateBoundFreeCutOff(false,boundFreeCutOff); grid.cell[i]->boundFreeCutOffUpdate=false; }
      if (grid.cell[i]->boundFreeTableUpdate) { grid.cell[i]->boundFreeAbsKappaTable(true); grid.cell[i]->boundFreeTableUpdate=false; } 
      // Calculate (Rosseland mean) optical depth
      cell = (CellSS*) grid.cell[i];
      tauCell[i]=cell->rossContKappa(Processes(pEScatter|pBoundFree|pFreeFree))*cell->rho()*(cell->rMax()-cell->rMin());
    }
  }

  // Wait for all processes to reach this point
  mpiCom->barrier();

  // For each cell
  for (i=0;i<grid.cell.size();i++) {
    // Broadcast (Rosseland mean) optical depth
    boost::mpi::broadcast(*mpiCom,tauCell[i],grid.cell[i]->responsible);
  }

  // Wait for all processes to reach this point
  mpiCom->barrier();

  return tauCell;
}

// Calculate (Rosseland mean) optical depth in cells - New!
// Does not work for subcells - Fix!
vector<double> getTauCell(const Grid& grid, boost::mpi::communicator* mpiCom=NULL) {
  if (mpiCom==NULL) return getTauCellSingle(grid);
  else return getTauCellCluster(grid,mpiCom);
}

// Find optical depth border
// Subcellsand containers: Works
void findBorder(const Grid& grid, const double& tauLim, int& border, double& R, boost::mpi::communicator* mpiCom=NULL) {
  CellSS* cell;
  double tau;
  vector<double> tauCell(grid.cell.size());
  int i;
  bool master=false;

  if (mpiCom==NULL) master=true;
  else if (mpiCom->rank()==0) master=true;

  if (master) cout<<"mc-spec-evo::findBorder: Calculate Rosseland mean optical depth within each cell\n";

  // Get (Rosseland optical depth) in the cells
  tauCell=getTauCell(grid,mpiCom);

  // Find border
  tau=0;
  for (i=grid.cell.size()-1;i>=0;i--) {
    cell = (CellSS*) grid.cell[i];
    border=i; R=cell->rMax();
    tau+=tauCell[i];
    if (tau>tauLim) break;
  }
  // If no border
  if (tau<=tauLim) { border=-1; R=((CellSS*) grid.cell[0])->rMin(); }

  if (master) cout<<"mc-spec-evo::findBorder: border: "<<border<<"\n";
  // Testing!
  if (mpiCom!=NULL) cout<<"mc-spec-evo::findBorder: Process: "<<mpiCom->rank()<<"; border: "<<border<<"\n";
}

// Find optical depth border
// Subcellsand containers: Works
/*void findBorder(const Grid& grid, const double& tauLim, int& border, double& R, bool log=false) {
  CellSS* cell;
  double tau;
  vector<double> tauCell(grid.cell.size());
  int i;

  if (log) cout<<"mc-diff-evo::findBorder: Calculate Rosseland mean optical depth within each cell\n";
  
  // Calculate Rosseland mean optical depth within each cell
  #pragma omp parallel for private(cell)
  for (i=0;i<grid.cell.size();i++) { 
    cell = (CellSS*) grid.cell[i];
    tauCell[i]=cell->rossContKappa(Processes(pEScatter|pBoundFree|pFreeFree))*cell->rho()*(cell->rMax()-cell->rMin());
  }

  // Find border
  tau=0;
  for (i=grid.cell.size()-1;i>=0;i--) {
    cell = (CellSS*) grid.cell[i];
    border=i; R=cell->rMax();
    tau+=tauCell[i];
    if (log) cout<<"mc-diff-evo::findBorder: cell: "<<i<<", tau: "<<tau<<", tauLim: "<<tauLim<<"\n";
    if (tau>tauLim) break;
  }
  // If no border
  if (tau<=tauLim) { border=-1; R=((CellSS*) grid.cell[0])->rMin(); }
  }*/

// Update cell on-the-spot approximation
// Subcells and containers: Works
void updateOnTheSpotCell(Cell* cell, bool set, const OnTheSpot& onTheSpot, const OnTheSpotNew& onTheSpotNew, bool log=false) {
  FrequencyBin onTheSpotNu;
  TransBF* transBF;
  int i,j,k,l;

  // If container - Update on-the-spot approximation for each subcell and return
  if (cell->type()==cellContainer) { for (i=0;i<cell->subCell.size();i++) updateOnTheSpotCell(cell->subCell[i],set,onTheSpot,onTheSpotNew,log); return; }

  // Reset on-the-spot (new) range
  onTheSpotNu.min=nuInf; onTheSpotNu.max=nuInf;

  // For each atom
  for (i=0;i<cell->atom.size();i++) {

    // Update on-the-spot flag
    if (set and onTheSpot(cell->atom[i])) cell->atom[i].onTheSpot(true);
    else cell->atom[i].onTheSpot(false);

    // For each ionization state
    for (j=0;j<cell->atom[i].iState.size();j++) {
      // For each bound state
      for (k=0;k<cell->atom[i].iState[j].bState.size();k++) {
	// For each (upward) bound-free transition
	for (l=0;l<cell->atom[i].iState[j].bState[k].upTransBF.size();l++) {
	  // Set transition pointer
	  transBF=cell->atom[i].iState[j].bState[k].upTransBF[l];
	  // If outside border and on-the-spot (new) requested
	  if (set and onTheSpotNew(transBF)) {
	    // Set on-the-spot (new) flag to true
	    transBF->onTheSpotNew(true);
	    // If necessary - Adjust on-the-spot (new) range
	    if (transBF->nu()<onTheSpotNu.min) onTheSpotNu.min=transBF->nu();
	  }
	  // Else - Set on-the-spot (new) flag to false
	  else transBF->onTheSpotNew(false);
	}
      }
    }
    // Update on-the-spot (new) range
    cout<<"mc-diff-evo::updateOnTheSpotRegion: cell: "<<cell->index()<<", onTheSpotNu: ("<<onTheSpotNu.min<<","<<onTheSpotNu.max<<")\n";
    cell->onTheSpotNu(onTheSpotNu);
  }
}

// Update on-the-spot approximation region
// Subcells and containers: Works
void updateOnTheSpotRegion(Grid& grid, const double& tauLim, const OnTheSpot& onTheSpot, const OnTheSpotNew& onTheSpotNew, bool log=false) {
  int border;
  double R;
  int i;

  // Find on-the-spot border
  findBorder(grid,tauLim,border,R,log);
  // Testing!!!
  //border=58;
  //border=41;

  // Update on-the-spot appoximation for each cell
  for (i=0;i<grid.cell.size();i++) updateOnTheSpotCell(grid.cell[i],(i>border),onTheSpot,onTheSpotNew,log);

  // Update derived matter quantities (in on-the-spot region)- Excluding derived MC quantities  
  grid.updateMatterDerived(border+1,grid.cell.size()-1,false);
}

// Create frequency bin list
// Subcells and containers: Not applicable
FrequencyBinList createNuList(const FrequencyBin& nu, const int& nBin) {
  int i;
  vector<double> nuWork(nBin+1);
  FrequencyBinList nuList;
  
  double dlognu=(log(nu.max)-log(nu.min))/float(nBin);
  for (i=0;i<nuWork.size();i++) nuWork[i]=nu.min*exp(i*dlognu);
  nuList.resize(nBin);
  for (i=0;i<nuList.size();i++) nuList[i].min=nuWork[i];
  for (i=0;i<nuList.size();i++) nuList[i].max=nuWork[i+1];

  return nuList;
}

// Cell radioactive decay luminosity
// Subcells and containers: Works
double LRadioCell(Cell* cell, const double& t) {
  double L=0;
  double M56Ni,M56Co,M56Fe,M57Ni,M57Co,M57Fe,M44Ti,M44Ca;
  int i;

  // If container - Sum radioactive decay luminosity for each subcell
  if (cell->type()==cellContainer) { for (i=0;i<cell->subCell.size();i++) L+=LRadioCell(cell->subCell[i],t); return L; }

  // Calculate the 56Ni decay-chain luminosity
  M56Ni=cell->M(28,28);
  M56Co=cell->M(27,29);
  M56Fe=cell->M(26,30);
  decay56Ni(M56Ni,M56Co,M56Fe,cell->t(),t);
  L+=M56Ni*e56Ni+M56Co*e56Co;
  
  // Calculate the 57Ni decay-chain luminosity
  M57Ni=cell->M(28,29);
  M57Co=cell->M(27,30);
  M57Fe=cell->M(26,31);
  decay57Ni(M57Ni,M57Co,M57Fe,cell->t(),t);
  L+=M57Ni*e57Ni+M57Co*e57Co;
  
  // Calculate the 44Ti decay-chain luminosity
  M44Ti=cell->M(22,22);
  M44Ca=cell->M(20,24);
  decay44Ti(M44Ti,M44Ca,cell->t(),t);
  L+=M44Ti*e44Ti;

  return L;
}

// Grid radioactive decay luminosity
// Subcells and containers: Works
double LRadio(Grid* grid, const double& t) {
  double L=0;
  int i;

  // Sum radioactive decay luminosity for each cell
  for (i=0;i<grid->cell.size();i++) L+=LRadioCell(grid->cell[i],t);

  return L;
}

// Radioactive decay energy integrand
// Subcells and containers: Works
class ERadioIntegrand {
public:
  Grid* grid;
  ERadioIntegrand(Grid* _grid) : grid(_grid) {};
  double operator()(double t) const { return LRadio(grid,t); }
};

// Radioactive decay energy
// Subcells and containers: Works
double ERadio(Grid* grid, const double& t1, const double& t2) {
  double E;
  ERadioIntegrand integrand(grid);
  // Testing
  int nIter;

  E=Integrator<ERadioIntegrand>::romberg(integrand,t1,t2,nIter,1e-2);

  return E;
}

// Sample thermal pakets
// Subcells and containers: Check packetSamplerThermal!
void sampleThermalPackets(Grid* grid, const int& cellIndexMin, const int& cellIndexMax, const double& EPacket, const FrequencyBin& nu, vector<PacketState>& packet, PacketConfig* packetConfig, boost::mpi::communicator* mpiCom=NULL, double expFactorWeight=1.0) {
  double expFactor,E=0;
  Packet sPacket;
  int i,nPacket;

  // Calculate thermal energy in region
  //for (i=cellIndexMin;i<=cellIndexMax;i++) E+=grid->cell[i]->E()*grid->cell[i]->M();
  // Take dilution factor into account - Hack!!!
  for (i=cellIndexMin;i<=cellIndexMax;i++) E+=grid->cell[i]->W()*grid->cell[i]->E()*grid->cell[i]->M();
	
  // Calculate number of thermal packets (corrected for expansion cooling)
  expFactor=(grid->tMin()+expFactorWeight*(grid->tMax()-grid->tMin()))/grid->tMin();
  nPacket=E*expFactor/EPacket;
 
  // Create new thermal packet sampler
  PacketSamplerBBThermal packetSamplerThermal(nPacket,cellIndexMin,cellIndexMax,grid,nu,packetConfig,mpiCom,expFactorWeight);
  
  // Sample thermal packets in new region and add to packet list
  #pragma omp parallel for firstprivate(sPacket)
  for (i=0;i<packetSamplerThermal.nPacket;i++) {
    // If MPI run
    if (mpiCom!=NULL) {
      // If the process is responsible for the packet
       if (i%mpiCom->size()==mpiCom->rank()) {
	// Sample packet
	sPacket=packetSamplerThermal.sample(i);
	// Add packet to list
        #pragma omp critical
	packet.push_back(PacketState(sPacket));
      }
    }
    // Else
   else {
     // Sample the packet
     sPacket=packetSamplerThermal.sample(i);
     // Add packet to list
     #pragma omp critical
     packet.push_back(PacketState(sPacket));
   }
  }
}

// Sample radiation field
// Subcells and containers: Check packetSamplerI!
void sampleRadiationField(Grid* grid, const int& cellIndexMin, const int& cellIndexMax, const double& EPacket, vector<PacketState>& packet, PacketConfig* packetConfig, boost::mpi::communicator* mpiCom=NULL, double expFactorWeight=1.0) {
  double expFactor,E=0;
  int i,nPacket;
  Packet sPacket;
  bool master=false;

  if (mpiCom==NULL) master=true;
  else if (mpiCom->rank()==0) master=true;

  // Calculate radiative energy in region
  for (i=cellIndexMin;i<=cellIndexMax;i++) E+=(4*pi/c)*grid->cell[i]->J*grid->cell[i]->V();
	
  // Calculate number of packets (corrected for expansion cooling)
  expFactor=(grid->tMin()+expFactorWeight*(grid->tMax()-grid->tMin()))/grid->tMin();
  nPacket=E*expFactor/EPacket;
  
  if (master) cout<<"mc-diff-evo::sampleRadiationField: EPacket: "<<EPacket<<", nPacket"<<nPacket<<", expFactor: "<<expFactor<<", E: "<<E<<", tMin: "<<grid->tMin()<<", tMax: "<<grid->tMax()<<", expFactorWeight: "<<expFactorWeight<<"\n";

  // Create new intensity packet sampler
  PacketSamplerI packetSamplerI(nPacket,cellIndexMin,cellIndexMax,grid,packetConfig,mpiCom,0,expFactorWeight);

  // Sample packets in region and add to packet list
  #pragma omp parallel for firstprivate(sPacket)
  for (i=0;i<packetSamplerI.nPacket;i++) {
    // If MPI run
    if (mpiCom!=NULL) {
      // If the process is responsible for the packet
      if (i%mpiCom->size()==mpiCom->rank()) {
	// Sample packet
	sPacket=packetSamplerI.sample(i);
	// Add packet to list
        #pragma omp critical
	packet.push_back(PacketState(sPacket));
      }
    }
    // Else
    else {
      // Sample the packet
      sPacket=packetSamplerI.sample(i);
      // Add packet to list
      #pragma omp critical
      packet.push_back(PacketState(sPacket));
    }
  }
  if (master) cout<<"sampleRadiationField: Sampled "<<packet.size()<<" packets\n";
}

// Sample initial packets
// Subcells and containers: Does not work yet!
void sampleInitialPackets(Grid* grid, const int& cellIndexMin, const int& cellIndexMax, const double& EPacket, const FrequencyBin& nu, vector<PacketState>& packet, PacketConfig* packetConfig, boost::mpi::communicator* mpiCom=NULL, double expFactorWeight=1.0) {
  int i,radFieldBorder=cellIndexMax+1;

  // Find the border outside which we know the state of the radiation field
  for (i=0;i<grid->cell.size();i++) if (grid->cell[i]->Inu.size()!=0) { radFieldBorder=i; break; }
  
  // Sample thermal packets inside the border and add to packet list  
  if (cellIndexMin<radFieldBorder) sampleThermalPackets(grid,cellIndexMin,radFieldBorder-1,EPacket,nu,packet,packetConfig,mpiCom,expFactorWeight);

  // Sample radiation field outside the border and add to packet list
  if (radFieldBorder<=cellIndexMax) sampleRadiationField(grid,MAX(radFieldBorder,cellIndexMin),cellIndexMax,EPacket,packet,packetConfig,mpiCom,expFactorWeight);

  // Hack for restart - Fix in a (much) better way !!!
  // Does not work with subcells and containers yet!!!
  if (cellIndexMin>radFieldBorder) for (i=radFieldBorder;i<cellIndexMin;i++) {
      grid->cell[i]->radUpdated=false;
      // Testing
      grid->cell[i]->T(grid->cell[i]->TR()*pow(grid->cell[i]->W(),1.0/4.0));
    }
}

// Evolve radiation state
// Subcells and containers: Check packetSamplerBBL and packetSamplerBBLGamma
void evolveRadiationState(Grid* grid, const int& diffBorder, const int& diffBorderOld, const double& EPacket, const FrequencyBin& nu, const FrequencyBinList& nuList, const bool& includeRadio, vector<PacketState>& packetIn, vector<PacketState>& packetOut, PacketStat* packetStat, MCConfig* mCConfig, boost::mpi::communicator* mpiCom=NULL, bool injectPackets=true) {
  double Ldiff;
  int nPacket,nPacketGamma;
  PacketSampler* packetSampler=NULL;
  int i;
  bool master=false;

  if (mpiCom==NULL) master=true;
  else if (mpiCom->rank()==0) master=true;

  // If new packets are to be injected
  if (injectPackets) {

    // If diffusion border the same
    //if (diffBorder==diffBorderOld) {

    // Get luminosity at diffusion border
    // Negative luminosties are taken care of below
    if (diffBorder>=0) Ldiff=MAX(((CellSS*) grid->cell[diffBorder])->LFront(),0);
    else Ldiff=0;
  
    // Calculate number of packets to inject
    nPacket=Ldiff*(grid->tMax()-grid->tMin())/EPacket;

    // Set fraction of packets to reflect at diffusion border
    // Takes care of negative luminosities
    //if (((CellSS*) grid->cell[diffBorder])->FFront()<0) mCConfig->packet->PReflect=MAX(1.0-fabs(((CellSS*) grid->cell[diffBorder])->FFront()/BlackBody::F((grid->cell[diffBorder]->T()+grid->cell[diffBorder+1]->T())/2)),0.0);
    //else mCConfig->packet->PReflect=1.0;
    mCConfig->packet->PReflectInner=1.0;

    // If include radioactive heating
    if (includeRadio) {
      
      // Calculate number of gamma-packets to inject
      nPacketGamma=ERadio(grid,grid->tMin(),grid->tMax())/EPacket;
      if (master) cout<<"mc-diff-evo::evolveRadiationState: nPacketGamma: "<<nPacketGamma<<", tMin: "<<grid->tMin()<<", tMax: "<<grid->tMax()<<", ERadio: "<<ERadio(grid,grid->tMin(),grid->tMax())<<", EPacket: "<<EPacket<<"\n";
    }

    // Create new packetSampler
    if (includeRadio) packetSampler = new PacketSamplerBBLGamma(Ldiff,nPacket,nPacketGamma,nu,diffBorder+1,grid,mCConfig->packet,mpiCom);
    else packetSampler = new PacketSamplerBBL(Ldiff,nPacket,diffBorder+1,grid,nu,mCConfig->packet);
  
    //}
  }
      
  // Reset packet statistics
  if (includeRadio) packetStat->reset(grid,0,grid->cell.size()-1,true,nuList);
  else packetStat->reset(grid,diffBorder+1,grid->cell.size()-1,true,nuList);
  // Reset MC statistics (in outer region)
  grid->resetMCStatistics(diffBorder+1,grid->cell.size()-1,nuList);
  // If include radioactive energy deposition
  if (includeRadio) {
    // Reset g-packet MC statistics (in inner region)
    for (i=0;i<=diffBorder;i++) grid->cell[i]->resetGMCStatistics();
  }
        
  // Create new MC experiment
  MCExperiment mCExperiment(packetSampler,packetStat,mCConfig,mpiCom);
      
  // Evolve MC experiment
  mCExperiment.evolve(packetIn,packetOut,grid);
      
  // Update radiation state (in outer region)
  grid->updateRadiationState(diffBorder+1,grid->cell.size()-1,mCExperiment);

  // If include radioactive heating
  if (includeRadio) {
    // Update radioactive heating (in inner region)
    // Only works with on-the-fly update method !
    for (i=0;i<=diffBorder;i++) grid->cell[i]->updateERadio();
  }

  // Cleanup
  if (packetSampler!=NULL) delete packetSampler;
}

// Export statistics
// Subcells and containers: Works
void exportStat(GridSS& grid, PacketStat& packetStat, string& mCStatDir, int& nExport) {

  // Export packet statistics
  if (mCStatDir.compare("")!=0) packetStat.exportStat(statFile(mCStatDir,"mc-stat",nExport));
  // Export spectrum
  if (mCStatDir.compare("")!=0) packetStat.exportSpectrum(statFile(mCStatDir,"spectrum",nExport));
  // Export spectrum per transition type
  if (mCStatDir.compare("")!=0) packetStat.exportSpectrumTrans(statFile(mCStatDir,"spectrum-trans",nExport));

  // Increas export counter
  nExport++;
}

double getEPacket(GridSS& grid, const double& nPacketEff, const double& tStop, const bool& includeRadio) {
  double EPacket,E=0;
  int i,radFieldBorder=grid.cell.size();

  // Find the border outside which we know the state of the radiation field
  for (i=0;i<grid.cell.size();i++) if (grid.cell[i]->Inu.size()!=0) { radFieldBorder=i; break; }

  // Calculate total thermal energy inside the border
  //for (i=0;i<radFieldBorder;i++) E+=grid.cell[i]->E()*grid.cell[i]->M();
  // Take dilution factor into account - Hack!!!
  for (i=0;i<radFieldBorder;i++) E+=grid.cell[i]->W()*grid.cell[i]->E()*grid.cell[i]->M();

  // Calculate total radiative energy outside the border
  for (i=radFieldBorder;i<grid.cell.size();i++) E+=(4*pi/c)*grid.cell[i]->J*grid.cell[i]->V();

  // If include radioactive heating
  if (includeRadio) {
    
    // Calculate and add total radioactive energy
    E+=ERadio(&grid,grid.t(),tStop);
  }
    
  // Calculate energy per packet
  EPacket=E/nPacketEff;

  return EPacket;
}

double getEPacketL(GridSS& grid, const double& nPacketLdt, const double& dt) {
  double EPacket,L,E;

  // Get luminosity
  L=((CellSS*) grid.cell[grid.cell.size()-1])->LFront();

  // Calculate energy emitted in one time-step
  E=L*dt;
    
  // Calculate energy per packet
  EPacket=E/nPacketLdt;

  return EPacket;
}

// Update non-thermal state
// Subcells and containers: Works
void updateNTState(Grid& grid, string configFile) {
  Koz92Solver ntSolver(configFile);
  vector<Cell*> cellWork(0);
  int i,j;

  // Create work list of cells and subcells
  for (i=0;i<grid.cell.size();i++)
    if (grid.cell[i]->type()==cellContainer) for (j=0;j<grid.cell[i]->subCell.size();j++) cellWork.push_back(grid.cell[i]->subCell[j]);
    else cellWork.push_back(grid.cell[i]);

  // For each cell
  #pragma omp parallel for firstprivate(ntSolver)
  for (i=0;i<cellWork.size();i++) {

    // Update non-thermal state
    ntSolver.update(cellWork[i],true);
  }
}

// Init cell non-thermal state
// Subcells and containers: Works
void initNTStateCell(Cell* cell, string configFile) {
  int i;

  // If container - Init non-thermal state for each subcell and return
  if (cell->type()==cellContainer) { for (i=0;i<cell->subCell.size();i++) initNTStateCell(cell->subCell[i],configFile); return; }
    
  // Initiate non-thermal state
  cell->nonThermalHeating(1e-99);
  cell->nonThermalIonization(0.0);
  cell->nonThermalExcitation(0.0);
}

// Init grid non-thermal state
// Subcells and containers: Works
void initNTState(Grid& grid, string configFile) {
  int i;

  // Init non-thermal state for each cell
  for (i=0;i<grid.cell.size();i++) initNTStateCell(grid.cell[i],configFile);
}

// Evolve radioactive state of grid
// Subcells and containers: Check packetSamplerGamma
void evolveRadio(Grid& grid, const double& dt, const double EPacket, vector<PacketState>& packetIn, MCConfig* mCConfig, boost::mpi::communicator* mpiCom=NULL) {
  PacketSamplerGamma* packetSamplerGamma;
  vector<PacketState> packetGammaOut(0);
  vector<PacketState> packetGammaIn(0);
  int nPacketGamma,i;
  
  // For each cell evolve radioactive state (from the last time-step)
  for (i=0;i<grid.cell.size();i++) grid.cell[i]->evolveRadio(grid.t()-dt,dt);

  // Sort out gamma packets from list
  for (i=0;i<packetIn.size();i++) if (packetIn[i].type==gPacket) packetGammaIn.push_back(packetIn[i]);

  // Calculate number of gamma-packets to inject
  nPacketGamma=ERadio(&grid,grid.tMin(),grid.tMax())/EPacket;

  // Turn off conversion of g-packets
  mCConfig->packet->convertGPacket=false;

  // Create new g-packet sampler
  packetSamplerGamma = new PacketSamplerGamma(nPacketGamma,&grid,mCConfig->packet,mpiCom);

  // Reset g-packet MC statistics
  for (i=0;i<grid.cell.size();i++) grid.cell[i]->resetGMCStatistics();
        
  // Create new MC experiment
  MCExperiment mCExperiment(packetSamplerGamma,NULL,mCConfig,mpiCom);
      
  // Evolve MC experiment
  mCExperiment.evolve(packetGammaIn,packetGammaOut,&grid);

  // Turn on conversion of g-packets
  // Assuming this was set to true - Is this ok?
  mCConfig->packet->convertGPacket=true;
      
  // Update radioactive heating
  // Only works with on-the-fly update method !
  for (i=0;i<grid.cell.size();i++) grid.cell[i]->updateERadio();

  // Cleanup
  if (packetSamplerGamma!=NULL) delete packetSamplerGamma;
}

// Subcells and containers: May not work yet!
void initBoostFactor(GridSS& grid, int nBoostBin, FrequencyBin& nu, boost::mpi::communicator* mpiCom=NULL) {
  double nuMin,nuLyman,nuBalmer,nuMax,dlognu;
  vector<double> nuWork;
  int i,j;
  bool master=false;

  if (mpiCom==NULL) master=true;
  else if (mpiCom->rank()==0) master=true;

  int nBoostBin1=50;
  int nBoostBin2=50;
  int nBoostBin3=1;
  double dlognu1,dlognu2,dlognu3;

  // Get frequency ranges
  nuMin=nu.min;
  nuBalmer=grid.cell[0]->atom[0].iState[0].bState[1].upTransBF[0]->nu();
  nuLyman=grid.cell[0]->atom[0].iState[0].bState[0].upTransBF[0]->nu();
  nuMax=nu.max;

  // Calculate frequencies for boost bins
  dlognu1=(log(nuBalmer)-log(nuMin))/float(nBoostBin1);
  for (i=0;i<nBoostBin1;i++) nuWork.push_back(nuMin*exp(i*dlognu1));
  dlognu2=(log(nuLyman)-log(nuBalmer))/float(nBoostBin2);
  for (i=0;i<nBoostBin2;i++) nuWork.push_back(nuBalmer*exp(i*dlognu2));
  dlognu3=(log(nuMax)-log(nuLyman))/float(nBoostBin3);
  for (i=0;i<nBoostBin3;i++) nuWork.push_back(nuLyman*exp(i*dlognu3));
  nuWork.push_back(nuMax);

  // Sort frequencies (to be on the safe side)
  sort(nuWork.begin(),nuWork.end());

  // For each cell
  for (i=0;i<grid.cell.size();i++) {

    // Resize boost vectors
    grid.cell[i]->nuBoost.resize(nuWork.size()-1);
    grid.cell[i]->boost.resize(nuWork.size()-1);

    // Initiate boost bins
    for (j=0;j<grid.cell[i]->nuBoost.size();j++) { 
      grid.cell[i]->nuBoost[j].min=nuWork[j]; 
      grid.cell[i]->nuBoost[j].max=nuWork[j+1];
      grid.cell[i]->boost[j]=1.0;
      if (master and i==0) cout<<"mc-diff-evo::initBoostFactor: nuBin: "<<j<<", lambda: ["<<(c/grid.cell[i]->nuBoost[j].min)/AA<<":"<<(c/grid.cell[i]->nuBoost[j].max)/AA<<"]\n";
    }
  }
}

// Subcells and containers: May not work yet!
//void adjustBoostFactor(GridSS& grid, int diffBorder, double nJGood, boost::mpi::communicator* mpiCom=NULL) {
void adjustBoostFactor(GridSS& grid, int diffBorder, double nJGood, double boostMin, double boostMax, boost::mpi::communicator* mpiCom=NULL) {
  double boost,boostMean,nJ,nBin;
  FrequencyBinIterator nuBin,nuBinMin,nuBinMax;
  int i,j,k;
  bool master=false;
  double boostOuter=1;

  int nBoostBin1=50;
  int nBoostBin2=50;
  int nBoostBin3=1;

  if (mpiCom==NULL) master=true;
  else if (mpiCom->rank()==0) master=true;

  // Find photosphere and grey treatment border
  int photosphere,borderGrey;
  double R,tauPhot=1.0,tauGrey=100.0;
  findBorder(grid,tauPhot,photosphere,R,false);
  findBorder(grid,tauGrey,borderGrey,R,false);

  // For each cell
  for (i=diffBorder+1;i<grid.cell.size();i++) {

    // If grey treatment
    if (i<=borderGrey) {
      
      // Initialize number of bins and mean intensity events
      nBin=0; nJ=0;
      
      // Count number of bins and mean intensity events
      for (j=0;j<grid.cell[i]->nJ.size();j++) { 
	nBin++; 
	nJ+=grid.cell[i]->nJ[j]; 
      }

      // Calculate boost factor needed
      boost=(nJGood*nBin/MAX(nJ,1.0));
      boostMean=0;
      for (j=0;j<grid.cell[i]->boost.size();j++) boostMean+=grid.cell[i]->boost[j]/grid.cell[i]->boost.size();
      for (j=0;j<grid.cell[i]->boost.size();j++) grid.cell[i]->boost[j]=MIN(boostMean*boost,boostMax);

      if (master) cout<<"mc-diff-evo::adjustBoostFactor: cell: "<<i<<", Grey boost, boost: "<<boost<<", (new) boostFactor: "<<boostMean*boost<<", nJ: "<<nJ/nBin<<"\n";
    }
    // Else
    else {

      // For each boost bin
      for (j=0;j<grid.cell[i]->nuBoost.size();j++) {

	// Find overlapping frequency bins
	nuBinMin=grid.cell[i]->nu.find(grid.cell[i]->nuBoost[j].min);
	nuBinMax=grid.cell[i]->nu.find(grid.cell[i]->nuBoost[j].max);
	
	// Initialize number of bins and mean intensity events
	nBin=0; nJ=0;
	
	// Count number of bins and mean intensity events
	for (nuBin=nuBinMin;nuBin!=nuBinMax;nuBin++) { 
	  // Exclude edge bins
	  if (not nuBin->edge) nBin++;
	  nJ+=grid.cell[i]->nJ[nuBin-grid.cell[i]->nu.begin()]; 
	}
	// To be on the safe side
	nBin=MAX(nBin,1);
	
	// Calculate boost factor needed
	if (j<nBoostBin1+nBoostBin2) {
	  //if (grid.cell[i]->nuBoost[j].min>c/(10000*AA)) boost=(2*nJGood*nBin/MAX(nJ,1.0));
	  //else boost=(nJGood*nBin/MAX(nJ,1.0));	  
	  if (grid.cell[i]->nuBoost[j].min>c/(10000*AA)) boost=(nJGood*nBin/MAX(nJ,1.0));
	  else boost=((nJGood/4)*nBin/MAX(nJ,1.0));
	  grid.cell[i]->boost[j]=MAX(MIN(grid.cell[i]->boost[j]*boost,boostMax),boostMin);
	}
	else {
	  boost=(100.0*nBin/MAX(nJ,1.0));
	}
	
	if (master) cout<<"mc-diff-evo::adjustBoostFactor: cell: "<<i<<", lambdaBoost: ["<<(c/grid.cell[i]->nuBoost[j].max)/AA<<":"<<(c/grid.cell[i]->nuBoost[j].min)/AA<<"], boost: "<<boost<<", (new) boostFactor: "<<grid.cell[i]->boost[j]<<", nJ: "<<nJ/nBin<<", neRadio: "<<grid.cell[i]->neRadio<<"\n";
      }   
    }
  }
}

// Subcells and containers: May not work yet!
void updateBoostFactor(Grid& grid, int cellIndexMin=0) {
  int i;

  //Update boost factors
  #pragma omp parallel for
  for (i=cellIndexMin;i<grid.cell.size();i++) grid.cell[i]->boostMC();
}

// Main program
int main (int argc, char** argv) {

  vector<PacketState> packetIn(0);
  vector<PacketState> packetOut(0);
  bool neConverged,initiate,initiateState,updateMCDerived;
  double E=0,EOuter=0,ENew=0,EPacket,Rdiff,ROnTheSpot,Ldiff;
  double t,tMin,tMax,dt,dtDiff,dtLeft,dlogt=0.01/2.5,expFactorWeight;
  int nPacket,diffBorder,diffBorderNew,diffBorderOld=-1,onTheSpotBorder,nIterLambdaUse,n,m,i,j;
  FrequencyBinList nuList;
  vector<double> nuVector;
  PacketSamplerBBL* packetSamplerL=NULL;
  PacketSamplerBBThermal* packetSamplerThermal=NULL;
  MCExperiment* mCExperiment;
  bool injectPackets;
  boost::mpi::environment* mpiEnv=NULL;
  boost::mpi::communicator* mpiCom=NULL;
  bool master=true;
  // Testing
  bool combSolverScanT;
  int combSolverNIter;
  double scanTFac;
  // Hack
  OnTheSpotNew onTheSpotNewInit;

  /* Check if enough number of arguments provided */
  if (argc<2) {
    printf("Missing argument\n");
    exit(-1);
  }
  
  /* Get mandatory arguments */
  string gridDataFile=string(argv[1]);
  int nIterMin=atoi(argv[2]);
  int nIterMax=atoi(argv[3]);

  /* Set default values for optional arguments */
  double nPacketEff=10000000; // Testing!
  double nPacketLdt=1000000; // Testing!
  int nBin=10000;
  double tauLim=1e-3;
  double lambdaMin=1;
  double lambdaMax=25000;
  double tauDiff=2e2;
  bool seed=false;
  bool eScatter=true;
  bool lineScatter=true;
  bool freeFreeAbs=false;
  bool boundFreeAbs=false;
  bool flourescence=false;
  bool collProc=false;
  bool nonThermalProc=false;
  bool twoPhotonProc=false;
  bool chargeTransferProc=false;
  bool chargeTransferEmit=false;
  bool transTable=false;
  bool markovTable=false;
  bool markovTableExt=false;
  bool markovTableExtSplit=false;
  double markovTableLim=0.0;
  double markovTableThermLim=0.0;
  double markovTableCTLim=0.0;
  bool collExciteTable=false;
  double collExciteTableLim=0.0;  
  bool nonThermalExciteTable=false;
  bool etaTable=false;
  UpdateMethod updateMethod=urOnTheFly;
  bool boundFreeTable=false;
  double boundFreeTableMin=1;
  double boundFreeTableMax=1e5;
  int boundFreeTableNBin=1e4;
  bool boundFreeSTable=false;
  int boundFreeSTableNBin=1e3;
  RadEstimator radEstimator=rfDBB;
  RadEstimator radEstimatorBB=rfDBB;
  bool radEstimatorExp=false;
  BBSampling bbSampling=bbsEqualEnergy;
  BBFrame bbFrame=bbfMixed;
  bool bbLimbDark=true;
  OnTheSpot onTheSpot("");
  OnTheSpotNew onTheSpotNew("");
  double tauOnTheSpot=10;
  FFOnTheSpot ffOnTheSpot(false,10.0);
  bool boost=false;
  double nJGood=1000;
  double boostMin=0.25;
  double boostMax=1e5;
  bool alfaCorrect=false;
  bool alfaCorrectEmit=false;
  bool forceDBalance=false;
  bool forceDBalanceOSP=false;
  bool boundFreeBB=false;
  double boundFreeBBTauLim=1e3;
  double continuumBBTauLim=1e99;
  bool convertGPacket=true;
  bool transformKappa=false;
  bool bbEPacket=false;
  string configFile="";
  SolverMethod solverMethod=sNebular;
  bool includeRadio=false;
  bool exportEta=false;
  bool reflectOuter=false;
  bool transformRay=false;
  bool limitFreeFreeCRange=false;
  string gridTemplateFile="";
  string gridStatDir="";
  string mCStatDir="";
  string gridStateDir="";
  int saveStateStep=1;
  string gridStateFile="";
  bool exportIterInit=false;
  bool exportIterLambda=false;
  bool exportIterGuess=false;
  bool exportIterInitRad=false;
  bool JLineFallback=false;
  double JLineFallbackFactor=1.1;
  int JLineFallbackCount=3;
  bool JLineFallbackDBB=false;
  bool trackI=false;
  bool trackHK=false;
  //double scanTFacInit=2.0;
  double boundFreeCutOff=0.0;
  double boundFreeCutOffRoss=0.0;
  bool loadTransAll=true;
  bool mpi=false;
  double dtInit=-1;
  bool loadDBB=false;
  bool diffSolverInit=true;
  //EPacketMethod ePacketMethod=epLBased;
  EPacketMethod ePacketMethod=epEBased;
  bool dustAbs=false;

  /* Get optional arguments */
  for (i=2;i<argc;i++) {
    if (strcmp(argv[i],"--npacket_eff")==0 && i+1<argc)
      {i++; nPacketEff=(float)strtod(argv[i],NULL);}
    else if (strcmp(argv[i],"--npacket_Ldt")==0 && i+1<argc) // Testing!
      {i++; nPacketLdt=(float)strtod(argv[i],NULL);}
    else if (strcmp(argv[i],"--nbin")==0 && i+1<argc)
      {i++; nBin=atoi(argv[i]);}
    else if (strcmp(argv[i],"--taulim")==0 && i+1<argc)
      {i++; tauLim=(float)strtod(argv[i],NULL);}
    else if (strcmp(argv[i],"--lambda_min")==0 && i+1<argc)
      {i++; lambdaMin=(float)strtod(argv[i],NULL);}
    else if (strcmp(argv[i],"--lambda_max")==0 && i+1<argc)
      {i++; lambdaMax=(float)strtod(argv[i],NULL);}
    else if (strcmp(argv[i],"--taudiff")==0 && i+1<argc)
      {i++; tauDiff=(float)strtod(argv[i],NULL);}
    else if (strcmp(argv[i],"--seed")==0 && i+1<argc)
      {i++; if (strcmp(argv[i],"yes")) seed=true;}
    else if (strcmp(argv[i],"--escatter")==0 && i+1<argc)
      {i++; if (strcmp(argv[i],"yes")==0) eScatter=true;}
    else if (strcmp(argv[i],"--linescatter")==0 && i+1<argc)
      {i++; if (strcmp(argv[i],"no")==0) lineScatter=false;}
    else if (strcmp(argv[i],"--freefree_abs")==0 && i+1<argc)
      {i++; if (strcmp(argv[i],"yes")==0) freeFreeAbs=true;}
    else if (strcmp(argv[i],"--boundfree_abs")==0 && i+1<argc)
      {i++; if (strcmp(argv[i],"yes")==0) boundFreeAbs=true;}
    else if (strcmp(argv[i],"--flourescence")==0 && i+1<argc)
      {i++; if (strcmp(argv[i],"yes")==0) flourescence=true;}
    else if (strcmp(argv[i],"--coll_proc")==0 && i+1<argc)
      {i++; if (strcmp(argv[i],"yes")==0) collProc=true;}
    else if (strcmp(argv[i],"--non_thermal_proc")==0 && i+1<argc)
      {i++; if (strcmp(argv[i],"yes")==0) nonThermalProc=true;}
    else if (strcmp(argv[i],"--two_photon_proc")==0 && i+1<argc)
      {i++; if (strcmp(argv[i],"yes")==0) twoPhotonProc=true;}
    else if (strcmp(argv[i],"--charge_transfer_proc")==0 && i+1<argc)
      {i++; if (strcmp(argv[i],"yes")==0) chargeTransferProc=true;}
    else if (strcmp(argv[i],"--charge_transfer_emit")==0 && i+1<argc)
      {i++; if (strcmp(argv[i],"yes")==0) chargeTransferEmit=true;}
    else if (strcmp(argv[i],"--trans_table")==0 && i+1<argc)
      {i++; if (strcmp(argv[i],"yes")==0) transTable=true;}
    else if (strcmp(argv[i],"--markov_table")==0 && i+1<argc)
      {i++; if (strcmp(argv[i],"yes")==0) markovTable=true;}
    else if (strcmp(argv[i],"--markov_table_ext")==0 && i+1<argc)
      {i++; if (strcmp(argv[i],"yes")==0) markovTableExt=true;}
    else if (strcmp(argv[i],"--markov_table_lim")==0 && i+1<argc) 
      {i++; markovTableLim=(float)strtod(argv[i],NULL);}
    else if (strcmp(argv[i],"--markov_table_ext_split")==0 && i+1<argc)
      {i++; if (strcmp(argv[i],"yes")==0) markovTableExtSplit=true;}
    else if (strcmp(argv[i],"--markov_table_therm_lim")==0 && i+1<argc) 
      {i++; markovTableThermLim=(float)strtod(argv[i],NULL);}
    else if (strcmp(argv[i],"--markov_table_ct_lim")==0 && i+1<argc) 
      {i++; markovTableCTLim=(float)strtod(argv[i],NULL);}
    else if (strcmp(argv[i],"--coll_exc_table")==0 && i+1<argc)
      {i++; if (strcmp(argv[i],"yes")==0) collExciteTable=true;}
    else if (strcmp(argv[i],"--coll_exc_table_lim")==0 && i+1<argc) 
      {i++; collExciteTableLim=(float)strtod(argv[i],NULL);}
    else if (strcmp(argv[i],"--non_thermal_exc_table")==0 && i+1<argc)
      {i++; if (strcmp(argv[i],"yes")==0) nonThermalExciteTable=true;}
    else if (strcmp(argv[i],"--eta_table")==0 && i+1<argc)
      {i++; if (strcmp(argv[i],"yes")==0) etaTable=true;}
    else if (strcmp(argv[i],"--boundfree_table")==0 && i+1<argc)
      {i++; if (strcmp(argv[i],"yes")==0) boundFreeTable=true;}
    else if (strcmp(argv[i],"--boundfree_table_min")==0 && i+1<argc)
      {i++; boundFreeTableMin==(float)strtod(argv[i],NULL);}
    else if (strcmp(argv[i],"--boundfree_table_max")==0 && i+1<argc)
      {i++; boundFreeTableMax=(float)strtod(argv[i],NULL);}
    else if (strcmp(argv[i],"--boundfree_table_nbin")==0 && i+1<argc)
      {i++; boundFreeTableNBin=atoi(argv[i]);}
    else if (strcmp(argv[i],"--boundfree_stable")==0 && i+1<argc)
      {i++; if (strcmp(argv[i],"yes")==0) boundFreeSTable=true;}
    else if (strcmp(argv[i],"--boundfree_stable_nbin")==0 && i+1<argc)
      {i++; boundFreeSTableNBin=atoi(argv[i]);}
    else if (strcmp(argv[i],"--rad_estimator")==0 && i+1<argc)
      {i++; if (strcmp(argv[i],"lucy")==0) radEstimator=rfLucy;}
    else if (strcmp(argv[i],"--rad_estimator_bb")==0 && i+1<argc)
      {i++; if (strcmp(argv[i],"lucy")==0) radEstimatorBB=rfLucy;}
    else if (strcmp(argv[i],"--rad_estimator_exp")==0 && i+1<argc)
      {i++; if (strcmp(argv[i],"yes")==0) radEstimatorExp=true;}
    else if (strcmp(argv[i],"--bb_sampling")==0 && i+1<argc)
      {i++; if (strcmp(argv[i],"log_frequency")==0) bbSampling=bbsLogFrequency;}
    else if (strcmp(argv[i],"--bb_frame")==0 && i+1<argc)
      {
	i++; 
	if (strcmp(argv[i],"rest")==0) bbFrame=bbfRest;
	if (strcmp(argv[i],"cmf")==0) bbFrame=bbfCMF;
      }
    else if (strcmp(argv[i],"--bb_limb_dark")==0 && i+1<argc)
      {i++; if (strcmp(argv[i],"no")==0) bbLimbDark=false;}
    else if (strcmp(argv[i],"--onthespot")==0 && i+1<argc)
      {i++; onTheSpot=OnTheSpot(argv[i]);}
    else if (strcmp(argv[i],"--onthespot_new")==0 && i+1<argc)
      {i++; onTheSpotNew=OnTheSpotNew(argv[i]);}
    else if (strcmp(argv[i],"--tau_onthespot")==0 && i+1<argc)
      {i++; tauOnTheSpot=(float)strtod(argv[i],NULL);}
    else if (strcmp(argv[i],"--ff_onthespot")==0 && i+1<argc)
      {i++; if (strcmp(argv[i],"yes")==0) ffOnTheSpot.apply=true;}
    else if (strcmp(argv[i],"--ff_onthespot_taulim")==0 && i+1<argc)
      {i++; ffOnTheSpot.tauLim=(float)strtod(argv[i],NULL);}
    else if (strcmp(argv[i],"--boost")==0 && i+1<argc)
      {i++; if (strcmp(argv[i],"yes")==0) boost=true;}
    else if (strcmp(argv[i],"--nJGood")==0 && i+1<argc)
      {i++; nJGood=(float)strtod(argv[i],NULL);}
    else if (strcmp(argv[i],"--boostMin")==0 && i+1<argc)
      {i++; boostMin=(float)strtod(argv[i],NULL);}
    else if (strcmp(argv[i],"--boostMax")==0 && i+1<argc)
      {i++; boostMax=(float)strtod(argv[i],NULL);}   
    else if (strcmp(argv[i],"--alfa_correct")==0 && i+1<argc)
      {i++; if (strcmp(argv[i],"yes")==0) alfaCorrect=true;}
    else if (strcmp(argv[i],"--alfa_correct_emit")==0 && i+1<argc)
      {i++; if (strcmp(argv[i],"yes")==0) alfaCorrectEmit=true;}
    else if (strcmp(argv[i],"--force_dbalance")==0 && i+1<argc)
      {i++; if (strcmp(argv[i],"yes")==0) forceDBalance=true;}
    else if (strcmp(argv[i],"--force_dbalance_osp")==0 && i+1<argc)
      {i++; if (strcmp(argv[i],"yes")==0) forceDBalanceOSP=true;}
    else if (strcmp(argv[i],"--convert_gpacket")==0 && i+1<argc)
      {i++; if (strcmp(argv[i],"no")==0) convertGPacket=false;}
    else if (strcmp(argv[i],"--transform_kappa")==0 && i+1<argc)
      {i++; if (strcmp(argv[i],"yes")==0) transformKappa=true;}
    else if (strcmp(argv[i],"--bb_epacket")==0 && i+1<argc)
      {i++; if (strcmp(argv[i],"yes")==0) bbEPacket=true;}
    else if (strcmp(argv[i],"--boundfree_bb")==0 && i+1<argc)
      {i++; if (strcmp(argv[i],"yes")==0) boundFreeBB=true;}
    else if (strcmp(argv[i],"--boundfree_bb_taulim")==0 && i+1<argc)
      {i++; boundFreeBBTauLim=(float)strtod(argv[i],NULL);}
    else if (strcmp(argv[i],"--continuum_bb_taulim")==0 && i+1<argc)
      {i++; continuumBBTauLim=(float)strtod(argv[i],NULL);}
    else if (strcmp(argv[i],"--config_file")==0 && i+1<argc)
      {i++; configFile=string(argv[i]);}
    else if (strcmp(argv[i],"--msolver_method")==0 && i+1<argc)
      {
	i++; 
	if (strcmp(argv[i],"nlte")==0) solverMethod=sNLTE;
	if (strcmp(argv[i],"comb")==0) solverMethod=sComb;
	if (strcmp(argv[i],"lte")==0) solverMethod=sLTE;	
      }
    else if (strcmp(argv[i],"--include_radio")==0 && i+1<argc)
      {i++; if (strcmp(argv[i],"yes")==0) includeRadio=true;}
    else if (strcmp(argv[i],"--export_eta")==0 && i+1<argc)
      {i++; if (strcmp(argv[i],"yes")==0) exportEta=true;}
    else if (strcmp(argv[i],"--reflect_outer")==0 && i+1<argc)
      {i++; if (strcmp(argv[i],"yes")==0) reflectOuter=true;}
    else if (strcmp(argv[i],"--transform_ray")==0 && i+1<argc)
      {i++; if (strcmp(argv[i],"yes")==0) transformRay=true;}
    else if (strcmp(argv[i],"--limit_freefree_crange")==0 && i+1<argc)
      {i++; if (strcmp(argv[i],"yes")==0) limitFreeFreeCRange=true;}
    else if (strcmp(argv[i],"--grid_template_file")==0 && i+1<argc)
      {i++; gridTemplateFile=string(argv[i]);}
    else if (strcmp(argv[i],"--grid_stat_dir")==0 && i+1<argc)
      {i++; gridStatDir=string(argv[i]);}
    else if (strcmp(argv[i],"--mc_stat_dir")==0 && i+1<argc)
      {i++; mCStatDir=string(argv[i]);}
    else if (strcmp(argv[i],"--grid_state_dir")==0 && i+1<argc)
      {i++; gridStateDir=string(argv[i]);}
    if (strcmp(argv[i],"--save_state_step")==0 && i+1<argc) 
      {i++; saveStateStep=atoi(argv[i]);}
    else if (strcmp(argv[i],"--grid_state_file")==0 && i+1<argc)
      {i++; gridStateFile=string(argv[i]);}
    else if (strcmp(argv[i],"--export_iter_init")==0 && i+1<argc)
      {i++; if (strcmp(argv[i],"yes")==0) exportIterInit=true;}
    else if (strcmp(argv[i],"--export_iter_lambda")==0 && i+1<argc)
      {i++; if (strcmp(argv[i],"yes")==0) exportIterLambda=true;}
    else if (strcmp(argv[i],"--export_iter_guess")==0 && i+1<argc)
      {i++; if (strcmp(argv[i],"yes")==0) exportIterGuess=true;}
    else if (strcmp(argv[i],"--export_iter_init_rad")==0 && i+1<argc)
      {i++; if (strcmp(argv[i],"yes")==0) exportIterInitRad=true;}
    else if (strcmp(argv[i],"--jline_fallback")==0 && i+1<argc)
      {i++; if (strcmp(argv[i],"yes")==0) JLineFallback=true;}
    else if (strcmp(argv[i],"--jline_fallback_factor")==0 && i+1<argc)
      {i++; JLineFallbackFactor=(float)strtod(argv[i],NULL);}
    else if (strcmp(argv[i],"--jline_fallback_count")==0 && i+1<argc) 
      {i++; JLineFallbackCount=atoi(argv[i]);}
    else if (strcmp(argv[i],"--jline_fallback_dbb")==0 && i+1<argc)
      {i++; if (strcmp(argv[i],"yes")==0) JLineFallbackDBB=true;}
    else if (strcmp(argv[i],"--track_I")==0 && i+1<argc)
      {i++; if (strcmp(argv[i],"yes")==0) trackI=true;}
    else if (strcmp(argv[i],"--track_HK")==0 && i+1<argc)
      {i++; if (strcmp(argv[i],"yes")==0) trackHK=true;}
    //else if (strcmp(argv[i],"--scanTfac_init")==0 && i+1<argc)
    //  {i++; scanTFacInit=(float)strtod(argv[i],NULL);}
    else if (strcmp(argv[i],"--boundfree_cutoff")==0 && i+1<argc)
      {i++; boundFreeCutOff=(float)strtod(argv[i],NULL);}
    else if (strcmp(argv[i],"--boundfree_cutoff_ross")==0 && i+1<argc)
      {i++; boundFreeCutOffRoss=(float)strtod(argv[i],NULL);}
    else if (strcmp(argv[i],"--mpi")==0 && i+1<argc)
      {i++; if (strcmp(argv[i],"yes")==0) mpi=true;}
    else if (strcmp(argv[i],"--load_trans_all")==0 && i+1<argc)
      {i++; if (strcmp(argv[i],"yes")==0) loadTransAll=true;}
    else if (strcmp(argv[i],"--dt_init")==0 && i+1<argc)
      {i++; dtInit=(float)strtod(argv[i],NULL)*day;}
    else if (strcmp(argv[i],"--load_dbb")==0 && i+1<argc)
      {i++; if (strcmp(argv[i],"yes")==0) loadDBB=true;}
    else if (strcmp(argv[i],"--dust_abs")==0 && i+1<argc)
      {i++; if (strcmp(argv[i],"yes")==0) dustAbs=true;}
    else if (strcmp(argv[i],"--diffsolver_init")==0 && i+1<argc)
      {i++; if (strcmp(argv[i],"no")==0) diffSolverInit=false;}
    else if (strcmp(argv[i],"--epacket_method")==0 && i+1<argc)
      {
	i++; 
	if (strcmp(argv[i],"e_based")==0) ePacketMethod=epEBased;
	if (strcmp(argv[i],"l_based")==0) ePacketMethod=epLBased;
      }
  }

  // If MPI
  if (mpi) {
    mpiEnv=new boost::mpi::environment;
    mpiCom=new boost::mpi::communicator;
    // Determine if master
    if (mpiCom->rank()!=0) master=false;
  }

  // Allow nested openmp parallelization
  omp_set_nested(1);

  // Use this for gprof
  //omp_set_nested(0);
  //omp_set_dynamic(0);
  //omp_set_num_threads(1);

  // If seed requested
  if (seed) {
    // If MPI -Seed with time+pid
    if (mpi) randomize(0,1,time(NULL)+getpid());
    // Else - Seed with time (+ thead number)
    else randomize(0,1,time(NULL));
  }
  // Else
  else {
    // If MPI -Seed with pid
    if (mpi) randomize(0,1,getpid());
    // Else - Seed with zero (+ thread number)
    else randomize(0,1,0);
  }
  
  // Configure grid
  FrequencyBin nu(c/(lambdaMax*AA),c/(lambdaMin*AA));
  LineSelector* lineSelector = new LineSelector(nu,tauLim);
  // Hack !!!
  onTheSpotNewInit=(gridStateFile.compare("")==0) ? onTheSpotNewInit=onTheSpotNew : onTheSpotNewInit=OnTheSpotNew("");
  CellConfig* cellConfig = new CellConfig(false,neIteration,lineSelector,updateMethod,boundFreeTable,boundFreeTableMin,boundFreeTableMax,boundFreeTableNBin,boundFreeSTable,boundFreeSTableNBin,radEstimator,radEstimatorBB,radEstimatorExp,transTable,markovTable,markovTableExt,markovTableExtSplit,markovTableLim,markovTableThermLim,markovTableCTLim,collExciteTable,collExciteTableLim,nonThermalExciteTable,freeFreeAbs,boundFreeAbs,collProc,nonThermalProc,twoPhotonProc,chargeTransferProc,onTheSpot,onTheSpotNewInit,ffOnTheSpot,false,boundFreeBB,continuumBBTauLim,boundFreeBBTauLim,bbTJBB,exportEta,transformRay,limitFreeFreeCRange,JLineFallback,JLineFallbackFactor,JLineFallbackCount,JLineFallbackDBB,trackI,trackHK,boundFreeCutOff,boundFreeCutOffRoss,boost,alfaCorrect,alfaCorrectEmit,forceDBalance,forceDBalanceOSP,chargeTransferEmit,etaTable);
  BStateSelector* bStateSelector= new BStateSelector();
  TransBBSelector* transSelector = new TransBBAllSelector();
  GridConfig* gridConfig = new GridConfig(cellConfig,bStateSelector,transSelector,updateMethod);

  // Configure MC experiment
  PacketConfig* packetConfig = new PacketConfig(false,false,lineScatter,eScatter,false,freeFreeAbs,boundFreeAbs,true,0,dustAbs,flourescence,collProc,nonThermalProc,twoPhotonProc,chargeTransferProc,transTable,updateMethod,boost,convertGPacket,-1,-1,true,1.0,reflectOuter,1.0,transformKappa,bbEPacket);
  MCConfig* mCConfig = new MCConfig(packetConfig);

  // Create EOS
  EOS* eos = new EOSBasic(configFile);

  // Create grid
  GridSS grid(NULL,eos,gridConfig,mpiCom);
  
  // Load grid
  grid.load(gridDataFile);

  // Init grid frequency list
  grid.initNu(nu.min,nu.max,nBin);

  // Create frequency bin list
  nuList=createNuList(nu,nBin);

  // Create frequency vector
  nuVector.resize(nBin+1); nuVector[0]=nuList[0].min;
  for (i=0;i<nuList.size();i++) nuVector[i+1]=nuList[i].max;
  
  // Update atomdata tables (using the provided binning)
  grid.atomData.updateTables(nuVector);
  
  if (mpiCom!=NULL) mpiCom->barrier();

  // Load grid state
  grid.loadState(statFile(gridStateDir),"grid-state",nIterMin);
  
  // Find diffusion border
  findBorder(grid,tauDiff,diffBorder,Rdiff,master);

  // Update on-the-spot region
  updateOnTheSpotRegion(grid,tauOnTheSpot,onTheSpot,onTheSpotNew,master);

  // Set borders for MC experiment
  packetConfig->cellIndexMin=diffBorder+1;
  packetConfig->cellIndexMax=grid.cell.size()-1;

  // Get packet energy
  EPacket=getEPacket(grid,nPacketEff,tStop,includeRadio);
   
  // Create packet statistics object
  PacketStat packetStat(&grid,diffBorder+1,grid.cell.size()-1,true,nuList);

  if (mpiCom!=NULL) mpiCom->barrier();
  
  // If boost mode - Initiate boost factor
  if (boost) initBoostFactor(grid,nBoostBin,nu,mpiCom);
  
  // Sample inital packets and add to packet list
  sampleInitialPackets(&grid,diffBorder+1,grid.cell.size()-1,EPacket,nu,packetIn,packetConfig,mpiCom);
  
  // Evolution iteration loop
  for (n=nIterMin+1;n<=nIterMax;n++) {

    // Load grid state
    grid.loadState(statFile(gridStateDir),"grid-state",n);

    if (master) cout<<"mc-spec: iteration: "<<n<<", t: "<<grid.t()/day<<" days, dt: "<<dt/day<<" days, tBorder: ["<<grid.tMin()/day<<","<<grid.tMax()/day<<"] days, diffusion border: "<<diffBorder<<"\n";

    // Find diffusion border
    findBorder(grid,tauDiff,diffBorder,Rdiff,master);

    // Update on-the-spot region
    updateOnTheSpotRegion(grid,tauOnTheSpot,onTheSpot,onTheSpotNew,master);

    // If new diffusion border within previous
    if (diffBorderNew<diffBorder) {
	
      if (master) cout<<"mc-diff-evo: Move diffusion border from cell: "<<diffBorder<<" to cell "<<diffBorderNew<<"\n";

      // Set boost factors to those of the previously innermost cell
      for (i=diffBorderNew+1;i<diffBorder+1;i++) grid.cell[i]->boost=grid.cell[diffBorder+1]->boost;

      // Sample thermal packets in new region and add to packet list
      sampleThermalPackets(&grid,diffBorderNew+1,diffBorder,EPacket,nu,packetIn,packetConfig,mpiCom);
      
      // Store old diffusion border
      diffBorderOld=diffBorder;
	      
      // Set new diffusion border
      diffBorder=diffBorderNew;
      
      // Adjust inner border for MC experiment
      packetConfig->cellIndexMin=diffBorder+1;
    }

    // Update derived matter quantities (in outer region)
    grid.updateMatterDerived(diffBorder+1,grid.cell.size()-1);

    // Evolve radiation state in outer region
    evolveRadiationState(&grid,diffBorder,diffBorderOld,EPacket,nu,nuList,includeRadio,packetIn,packetOut,&packetStat,mCConfig,mpiCom);
      
    if (mpiCom!=NULL) mpiCom->barrier();
    
    // If boost mode - Adjust boost factor
    if (boost) adjustBoostFactor(grid,diffBorder,nJGood,boostMin,boostMax,mpiCom);
    
    // Export MC statistics
    exportStat(packetStat,mCStatDir,n);

    // Update packet (state) list
    packetIn=packetOut;
  }

  // Clean up
  delete mCConfig;
  delete packetConfig;
  delete eos;
  delete gridConfig;
  delete bStateSelector;
  delete transSelector;
  delete cellConfig;
  delete lineSelector;
  if (mpiCom!=NULL) delete mpiCom;
  if (mpiEnv!=NULL) delete mpiEnv;
}
