#include <vector>
#include <math.h>
#include <string.h>
#include <unistd.h>
#include <omp.h>
#include "mc-util.h"
#include "mc-math.h"
#include "mc-physics.h"
#include "mc.h"
#include "msolver.h"
#include "eos.h"
#include "diffsolver.h"

using namespace util;
using namespace math;
using namespace physics;
using namespace mc;
//using namespace grid;
using namespace msolver;
using namespace eos;
using namespace diffsolver;

class BoostRegionConfig {
public:
  FrequencyBin nu;
  int nBin;
  double nJGood;
  double boostMin;
  double boostMax;
  // Constructors
  BoostRegionConfig();
  // Load data
  void load(const YAML::Node& data);
};

// Constructors
BoostRegionConfig::BoostRegionConfig() : nu(0,nuInf), nBin(50), nJGood(1e3), boostMin(0.25), boostMax(1e5) {};

// Load data
void BoostRegionConfig::load(const YAML::Node& data) {
										    
  // Load frequency range
  if (data["lambda"]) {
    if (data["lambda"].size()==2) {
      nu.min=(c/data["lambda"][1].as<double>())/AA;
      nu.max=(c/data["lambda"][0].as<double>())/AA;
    }
  }
  else { cout<<"BoostRegionConfig::load: Error: Could not get frequency range\n"; exit(-1); }
  
  // Load number of bins
  if (data["nBin"]) nBin=data["nBin"].as<int>();

  // Load good number of J events
  if (data["nJGood"]) nJGood=data["nJGood"].as<int>();

  // Load minimum boost factor
  if (data["boostMin"]) boostMin=data["boostMin"].as<double>();

  // Load number of bins
  if (data["boostMax"]) boostMax=data["boostMax"].as<double>();
}

// Boost config
class BoostConfig {
public :
  vector<BoostRegionConfig> region;
  // Constructors
  BoostConfig();
  BoostConfig(const string& dataFile);
  // Load data
  void load(const YAML::Node& data);
  void load(const string& dataFile);
};

// Constructors
BoostConfig::BoostConfig() : region(0) {};
BoostConfig::BoostConfig(const string& dataFile) { load(dataFile); };

// Load data
void BoostConfig::load(const YAML::Node& data) {
  int i;

  // Load boost regions
  region.resize(data.size());
  for (i=0;i<data.size();i++) region[i].load(data[i]);
}

// Load data (from file)
void BoostConfig::load(const string& dataFile) {
  YAML::Node data;

  // Load data from file
  data=YAML::LoadFile(dataFile);

  // Load data
  if (data["boost"]) load(data["boost"]);
}

enum EPacketMethod { epLBased, epEBased };

class TransBBAllSelector : public TransBBSelector {
  virtual bool use(TransBBData* trans) { return true; }
};

class TransBBNoneSelector : public TransBBSelector {
  virtual bool use(TransBBData* trans) { return false; }
};

// Create statistics file name
string statFile(const string& statDir, const string& prefix, const int& index) {
  stringstream statFile;
  statFile.str("");
  statFile<<statDir<<"/"<<prefix<<"-"<<index<<".dat";
  return statFile.str();
}

// Update populations for cell
void updatePopStateCell(Cell* cell) {
  LTESolver popLTESolver;
  popLTESolver.config.temperature.update=false;
  int i;

  // If container - Update populations for each subcell and return
  if (cell->type()==cellContainer) { for (i=0;i<cell->subCell.size();i++) updatePopStateCell(cell->subCell[i]); return; }

  // Update populations
  popLTESolver.update(cell);
}

// Update populations for grid
void updatePopState(const GridSS& grid) {
  int i;

  // Update populations for each cell
  for (i=0;i<grid.cell.size();i++) updatePopStateCell(grid.cell[i]);
}

// Initiate radiation state for cell
void initRadStateCell(Cell* cell) {
  int i;

  // If container - Initiate radiation state for each subcell
  if (cell->type()==cellContainer) { for (i=0;i<cell->subCell.size();i++) initRadStateCell(cell->subCell[i]); }

  // Initiate to blackbody field at matter temperature
  cell->TR(cell->T());
  cell->W(1.0);
}

// Initiate radiation state for grid
void initRadState(const GridSS& grid, int cellIndexMax ) {
  int i;

  // Initi radiation state for each cell
  for (i=0;i<=cellIndexMax;i++) initRadStateCell(grid.cell[i]);
}

// Update the luminosity at the diffusion border
void updateLDiff(const GridSS& grid, int diffBorder) {
  double T,TNext,dm,dmNext,dmFront,kappa,kappaNext,kappaFront,rFront,vFront,FFront,FObsFront;

  // If no diffusion border - Just return
  if (diffBorder<0 or (diffBorder+1)>=grid.cell.size()) return;

  // Get temperature, mass and opacity for the cell
  T=grid.cell[diffBorder]->T();
  dm=grid.cell[diffBorder]->M();
  kappa=grid.cell[diffBorder]->rossKappaMC2(Processes(pEScatter|pBoundFree|pFreeFree|pBoundBound),-1,-1,100000);

  // Get temperature, mass and opacacity for next cell
  TNext=grid.cell[diffBorder+1]->T();
  dmNext=grid.cell[diffBorder+1]->M();
  kappaNext=grid.cell[diffBorder+1]->rossKappaMC2(Processes(pEScatter|pBoundFree|pFreeFree|pBoundBound),-1,-1,100000);

  // Calculate cell front mass and opacity
  dmFront=(dm+dmNext)/2;
  kappaFront=(pow(T,4)+pow(TNext,4))/(pow(T,4)/kappa+pow(TNext,4)/kappaNext);

  // Get cell front radius
  rFront=((CellSS*) grid.cell[diffBorder])->rMax();
  // Calculate cell front veclocity
  vFront=rFront/grid.t();

  // Calculate cell front flux in the rest and observer frames
  FFront=-4*pi*pow(rFront,2)*(a*c/(3*kappaFront))*(pow(TNext,4)-pow(T,4))/dmFront;
  FObsFront=FFront+vFront*(4.0/3.0)*a*pow(T,4);

  // Update the cell front flux in the rest and observer frames
  ((CellSS*) grid.cell[diffBorder])->FFront(FFront);
  ((CellSS*) grid.cell[diffBorder])->FObsFront(FObsFront);
}

// Update bound-free tables - New version!
void boundFreeTableUpdate(Grid* grid, int iMin, int iMax, boost::mpi::communicator* mpiCom, double boundFreeCutOff=-1) {
  int i,j,nResponsible=0;
  vector<Cell*> cellWork(0);
  bool master=false;

  if (mpiCom==NULL) master=true;
  else if (mpiCom->rank()==0) master=true;

  // Create work list of cells and subcells - New!
  for (i=iMin;i<=iMax;i++) {
    if (grid->cell[i]->type()==cellContainer) {
      for (j=0;j<grid->cell[i]->subCell.size();j++) {
	cellWork.push_back(grid->cell[i]->subCell[j]);
      }
    }
    else cellWork.push_back(grid->cell[i]);
  }
 
  // Determine which processes are responsible for the cells - Changed!
  for (i=0;i<cellWork.size();i++) cellWork[i]->responsible=cellWork[i]->index()%mpiCom->size();

  // Count number of cells the process is responsible for - New!
  for (i=0;i<cellWork.size();i++) if (cellWork[i]->responsible==mpiCom->rank()) nResponsible++;

  // Update matter state for the cells the process are responsible for - Changed!
  #pragma omp parallel for
  for (i=0;i<cellWork.size();i++) {
    if (cellWork[i]->responsible==mpiCom->rank()) {
      // Set number of threads to use in nested omp parallelization - New!
      omp_set_num_threads((omp_get_num_threads()-1)/nResponsible+1);
      // If needed - Update bound-free opacity cut-off and tables in advance - Testing!
      if (cellWork[i]->boundFreeCutOffUpdate) { cellWork[i]->updateBoundFreeCutOffOMP(false); cellWork[i]->boundFreeCutOffUpdate=false; }
      if (cellWork[i]->boundFreeTableUpdate) { cellWork[i]->boundFreeAbsKappaTable(true); cellWork[i]->boundFreeTableUpdate=false; }
    }
  }

  // Wait for all processes to reach this point
  mpiCom->barrier();

  // Syncronize bound-free cut-off and opacity table for each cell
  for (i=0;i<cellWork.size();i++) {
    if (master) cout<<"mc-spec::boundFreeTableUpdate: Synchronize cell "<<cellWork[i]->index()<<" bound-free cut-off\n";
    cellWork[i]->syncBoundFreeCutOff();
  }

  // Wait for all processes to reach this point
  mpiCom->barrier();

  // Syncronize bound-free cut-off and opacity table for each cell
  for (i=0;i<cellWork.size();i++) {
    if (master) cout<<"mc-spec::boundFreeTableUpdate: Synchronize cell "<<cellWork[i]->index()<<" bound-free opacity table\n";
    cellWork[i]->syncBoundFreeAbsKappaTable();
  }

  // Wait for all processes to reach this point
  mpiCom->barrier();
}

// Calculate (Rosseland mean) optical depth in cells (on a single computer) - New!
// Does not work for subcells - Fix!
vector<double> getTauCellSingle(const Grid& grid, Processes processes=Processes(pEScatter|pBoundFree|pFreeFree)) {
  CellSS* cell;
  vector<double> tauCell(grid.cell.size());
  int i;
 
  // Calculate Rosseland mean optical depth within each cell
  #pragma omp parallel for private(cell)
  for (i=0;i<grid.cell.size();i++) { 
    cell = (CellSS*) grid.cell[i];
    //tauCell[i]=cell->rossContKappa(Processes(pEScatter|pBoundFree|pFreeFree))*cell->rho()*(cell->rMax()-cell->rMin());
    tauCell[i]=cell->rossContKappa(processes)*cell->rho()*(cell->rMax()-cell->rMin());
  }

  return tauCell;
}

// Calculate (Rosseland mean) optical depth in cells (on a cluster using MPI) - New!
// Subcellsand containers: Works
// Does not work for subcells (see below) - Fix!
/*vector<double> getTauCellCluster(const Grid& grid, boost::mpi::communicator* mpiCom) {
  CellSS* cell;
  vector<double> tauCell(grid.cell.size());
  // Should be configurable!
  double boundFreeCutOff=1e-3;
  int i,nResponsible=0;

  // Determine which processes are responsible for the cells
  for (i=0;i<grid.cell.size();i++) grid.cell[i]->responsible=grid.cell[i]->index()%mpiCom->size();

  // Count number of cells the process is responsible for
  for (i=0;i<grid.cell.size();i++) if (grid.cell[i]->responsible==mpiCom->rank()) nResponsible++;

  // For each cell the process is responsible for
  #pragma omp parallel for private(cell)
  for (i=0;i<grid.cell.size();i++) { 
    if (grid.cell[i]->responsible==mpiCom->rank()) {
      // Set number of threads to use in nested omp parallelization
      omp_set_num_threads((omp_get_num_threads()-1)/nResponsible+1);
      // If needed - Update bound-free opacity cut-off and tables in advance - Testing!
      // This needs to be done for thesubcells as well!!!
      if (grid.cell[i]->boundFreeCutOffUpdate) { grid.cell[i]->updateBoundFreeCutOffOMP(false,boundFreeCutOff); grid.cell[i]->boundFreeCutOffUpdate=false; }
      //if (grid.cell[i]->boundFreeCutOffUpdate) { grid.cell[i]->updateBoundFreeCutOff(false,boundFreeCutOff); grid.cell[i]->boundFreeCutOffUpdate=false; }
      if (grid.cell[i]->boundFreeTableUpdate) { grid.cell[i]->boundFreeAbsKappaTable(true); grid.cell[i]->boundFreeTableUpdate=false; } 
      // Calculate (Rosseland mean) optical depth
      cell = (CellSS*) grid.cell[i];
      tauCell[i]=cell->rossContKappa(Processes(pEScatter|pBoundFree|pFreeFree))*cell->rho()*(cell->rMax()-cell->rMin());
    }
  }

  // Wait for all processes to reach this point
  mpiCom->barrier();

  // For each cell
  for (i=0;i<grid.cell.size();i++) {
    // Broadcast (Rosseland mean) optical depth
    boost::mpi::broadcast(*mpiCom,tauCell[i],grid.cell[i]->responsible);
  }

  // Wait for all processes to reach this point
  mpiCom->barrier();

  return tauCell;
  }*/

// Calculate (Rosseland mean) optical depth in cells (on a cluster using MPI) - New version!
// Subcells and containers: Works
vector<double> getTauCellCluster(Grid& grid, boost::mpi::communicator* mpiCom, Processes processes=Processes(pEScatter|pBoundFree|pFreeFree)) {
  CellSS* cell;
  vector<double> tauCell(grid.cell.size());
  // Should be configurable!
  double boundFreeCutOff=1e-3;
  int i,j,nResponsible=0;

  // If bound-free processes included - Update bound-free tables
  if (processes(pBoundFree)) boundFreeTableUpdate(&grid,0,grid.cell.size()-1,mpiCom,boundFreeCutOff);

  // Determine which processes are responsible for the cells
  for (i=0;i<grid.cell.size();i++) grid.cell[i]->responsible=grid.cell[i]->index()%mpiCom->size();

  // Count number of cells the process is responsible for
  for (i=0;i<grid.cell.size();i++) if (grid.cell[i]->responsible==mpiCom->rank()) nResponsible++;

  // For each cell the process is responsible for
  #pragma omp parallel for private(cell,j)
  for (i=0;i<grid.cell.size();i++) { 
    if (grid.cell[i]->responsible==mpiCom->rank()) {
      // Set number of threads to use in nested omp parallelization
      omp_set_num_threads((omp_get_num_threads()-1)/nResponsible+1);
      // Calculate (Rosseland mean) optical depth
      cell = (CellSS*) grid.cell[i];
      //tauCell[i]=cell->rossContKappa(Processes(pEScatter|pBoundFree|pFreeFree))*cell->rho()*(cell->rMax()-cell->rMin());
      tauCell[i]=cell->rossContKappa(processes)*cell->rho()*(cell->rMax()-cell->rMin());
    }
  }

  // Wait for all processes to reach this point
  mpiCom->barrier();

  // For each cell
  for (i=0;i<grid.cell.size();i++) {
    // Broadcast (Rosseland mean) optical depth
    boost::mpi::broadcast(*mpiCom,tauCell[i],grid.cell[i]->responsible);
  }

  // Wait for all processes to reach this point
  mpiCom->barrier();

  return tauCell;
}

// Calculate (Rosseland mean) optical depth in cells - New!
// Does not work for subcells - Fix!
vector<double> getTauCell(Grid& grid, Processes processes=Processes(pEScatter|pBoundFree|pFreeFree), boost::mpi::communicator* mpiCom=NULL) {
  if (mpiCom==NULL) return getTauCellSingle(grid,processes);
  else return getTauCellCluster(grid,mpiCom,processes);
}

// Find optical depth border
// Subcellsand containers: Works
void findBorder(Grid& grid, const double& tauLim, int& border, double& R, Processes processes=Processes(pEScatter|pBoundFree|pFreeFree), boost::mpi::communicator* mpiCom=NULL) {
  CellSS* cell;
  double tau;
  vector<double> tauCell(grid.cell.size());
  int i;
  bool master=false;

  if (mpiCom==NULL) master=true;
  else if (mpiCom->rank()==0) master=true;

  if (master) cout<<"mc-spec::findBorder: Calculate Rosseland mean optical depth within each cell\n";

  // Get (Rosseland optical depth) in the cells
  tauCell=getTauCell(grid,processes,mpiCom);

  // Find border
  tau=0;
  for (i=grid.cell.size()-1;i>=0;i--) {
    cell = (CellSS*) grid.cell[i];
    border=i; R=cell->rMax();
    tau+=tauCell[i];
    if (master) cout<<"mc-spec::findBorder: cell: "<<i<<", Accumulated Rosseland mean optical depth: "<<tau<<"\n";
    if (tau>tauLim) break;
  }
  // If no border
  if (tau<=tauLim) { border=-1; R=((CellSS*) grid.cell[0])->rMin(); }

  if (master) cout<<"mc-spec::findBorder: border: "<<border<<"\n";
}

// Find optical depth border
// Subcells and containers: Works
/*void findBorder(const Grid& grid, const double& tauLim, int& border, double& R, bool log=false) {
  CellSS* cell;
  double tau;
  vector<double> tauCell(grid.cell.size());
  int i;

  if (log) cout<<"mc-diff-evo::findBorder: Calculate Rosseland mean optical depth within each cell\n";
  
  // Calculate Rosseland mean optical depth within each cell
  #pragma omp parallel for private(cell)
  for (i=0;i<grid.cell.size();i++) { 
    cell = (CellSS*) grid.cell[i];
    tauCell[i]=cell->rossContKappa(Processes(pEScatter|pBoundFree|pFreeFree))*cell->rho()*(cell->rMax()-cell->rMin());
  }

  // Find border
  tau=0;
  for (i=grid.cell.size()-1;i>=0;i--) {
    cell = (CellSS*) grid.cell[i];
    border=i; R=cell->rMax();
    tau+=tauCell[i];
    if (log) cout<<"mc-diff-evo::findBorder: cell: "<<i<<", tau: "<<tau<<", tauLim: "<<tauLim<<"\n";
    if (tau>tauLim) break;
  }
  // If no border
  if (tau<=tauLim) { border=-1; R=((CellSS*) grid.cell[0])->rMin(); }
  }*/

// Find diffusion border
// Subcells and containers: Works
void findDiffBorder(const Grid& grid, int& diffBorder, double& Rdiff, bool log=false) {
  int i;

  diffBorder=grid.cell.size()-1; Rdiff=((CellSS*) grid.cell[diffBorder])->rMax();

  for (i=0;i<grid.cell.size();i++) if (grid.cell[i]->Jnu.size()!=0) { diffBorder=i-1; Rdiff=((CellSS*) grid.cell[i])->rMin(); break; }

  return;
}
 
// Update cell on-the-spot approximation
// Subcells and containers: Works
void updateOnTheSpotCell(Cell* cell, bool set, const double& tauEdgeLim,  const OnTheSpot& onTheSpot, const OnTheSpotNew& onTheSpotNew, bool master=true) { // Changed!
  FrequencyBin onTheSpotNu;
  TransBF* transBF;
  double tauEdge; // New!
  int i,j,k,l;

  // If container - Update on-the-spot approximation for each subcell and return
  if (cell->type()==cellContainer) { for (i=0;i<cell->subCell.size();i++) updateOnTheSpotCell(cell->subCell[i],set,tauEdgeLim,onTheSpot,onTheSpotNew,master); return; } // Changed!

  // Reset on-the-spot (new) range
  onTheSpotNu.min=nuInf; onTheSpotNu.max=nuInf;

  // For each atom
  for (i=0;i<cell->atom.size();i++) {

    // Update on-the-spot flag
    if (set and onTheSpot(cell->atom[i])) cell->atom[i].onTheSpot(true);
    else cell->atom[i].onTheSpot(false);

    // For each ionization state
    for (j=0;j<cell->atom[i].iState.size();j++) {
      // For each bound state
      for (k=0;k<cell->atom[i].iState[j].bState.size();k++) {
	// For each (upward) bound-free transition
	for (l=0;l<cell->atom[i].iState[j].bState[k].upTransBF.size();l++) {
	  // Set transition pointer
	  transBF=cell->atom[i].iState[j].bState[k].upTransBF[l];
	  // If outside border and on-the-spot (new) requested
	  if (set and onTheSpotNew(transBF)) {
	    // Get transition (edge) optical depth in cell
	    tauEdge=transBF->kappa(nuEdge(transBF->nu(),blueEdge))*cell->rho()*(((CellSS*) cell)->rMax()-((CellSS*) cell)->rMin());
	    // If above limit
	    if (tauEdge>tauEdgeLim) {
	      // Set on-the-spot (new) flag to true
	      transBF->onTheSpotNew(true);
	      // If necessary - Adjust on-the-spot (new) range
	      if (transBF->nu()<onTheSpotNu.min) onTheSpotNu.min=transBF->nu();
	      if (master) cout<<"mc-spec::updateOnTheSpotRegion: cell: "<<cell->index()<<", set OSP flag for "<<transBF->lowerState->iState()->name()<<":"<<transBF->lowerState->getIndex()<<", nuEdge: "<<transBF->nu()<<", tauEdge: "<<tauEdge<<"\n";	      
	    }
	  }
	  // Else - Set on-the-spot (new) flag to false
	  else transBF->onTheSpotNew(false);
	}
      }
    }
    // Update on-the-spot (new) range
    if (master) cout<<"mc-spec::updateOnTheSpotRegion: cell: "<<cell->index()<<", onTheSpotNu: ("<<onTheSpotNu.min<<","<<onTheSpotNu.max<<")\n";
    cell->onTheSpotNu(onTheSpotNu);
  }
}

// Update on-the-spot approximation region
// Subcells and containers: Works
void updateOnTheSpotRegion(Grid& grid, const double& tauLim, const double& tauEdgeLim, const OnTheSpot& onTheSpot, const OnTheSpotNew& onTheSpotNew, Processes processes=Processes(pEScatter|pBoundFree|pFreeFree), boost::mpi::communicator* mpiCom=NULL) { // Changed!
  int border;
  double R;
  int i;
  bool master=false;

  if (mpiCom==NULL) master=true;
  else if (mpiCom->rank()==0) master=true;

  // Find on-the-spot border
  findBorder(grid,tauLim,border,R,processes,mpiCom);
  // Testing!!!
  //border=58;
  //border=41;

  // Update on-the-spot appoximation for each cell
  for (i=0;i<grid.cell.size();i++) updateOnTheSpotCell(grid.cell[i],(i>border),tauEdgeLim,onTheSpot,onTheSpotNew,master); // Changed!

  // Update derived matter quantities (in on-the-spot region)- Excluding derived MC quantities  
  grid.updateMatterDerived(border+1,grid.cell.size()-1,false);
}

// Create frequency bin list
// Subcells and containers: Not applicable
FrequencyBinList createNuList(const FrequencyBin& nu, const int& nBin) {
  int i;
  vector<double> nuWork(nBin+1);
  FrequencyBinList nuList;
  
  double dlognu=(log(nu.max)-log(nu.min))/float(nBin);
  for (i=0;i<nuWork.size();i++) nuWork[i]=nu.min*exp(i*dlognu);
  nuList.resize(nBin);
  for (i=0;i<nuList.size();i++) nuList[i].min=nuWork[i];
  for (i=0;i<nuList.size();i++) nuList[i].max=nuWork[i+1];

  return nuList;
}

// Handle cell time borders
// Subcells and containers: Works
void handleTimeBordersCell(Cell* cell, double& tMin, double& tMax, double& dt) {
  int i;

  // If container - Update time borders for each subcell
  if (cell->type()==cellContainer) for (i=0;i<cell->subCell.size();i++) handleTimeBordersCell(cell->subCell[i],tMin,tMax,dt);

  if (cell->tMin()!=tMax and tMax!=-1) cell->tMin(tMax);
}

// Handle cell time borders
// Subcells and containers: Works
void handleTimeBorders(Grid& grid, double& tMin, double& tMax, double &dt) {
  double dtCenter,t=(tMin+tMax)/2;
  int i;

  // if tMin differs from previous tMax
  if (grid.tMin()!=tMax and tMax!=-1) {
    // Warn
    cout<<"mc-spec::checkTimeBorders: Warning: tMin ("<<grid.tMin()<<") differs from previous tMax ("<<tMax<<") - Adjusting tMin!\n";
    // Adjust
    grid.tMin(tMax);
  }

  // Check time borders for each cell
  for (i=0;i<grid.cell.size();i++) handleTimeBordersCell(grid.cell[i],tMin,tMax,dt);

  // If not intial call - Calculate dt
  if (tMin!=-1 and tMax!=-1) dt=grid.t()-(tMax+tMin)/2;

  // Save new timeborders
  tMin=grid.tMin(); tMax=grid.tMax();
}

// Cell radioactive decay luminosity
// Subcells and containers: Works
double LRadioCell(Cell* cell, const double& t) {
  double L=0;
  double M56Ni,M56Co,M56Fe,M57Ni,M57Co,M57Fe,M44Ti,M44Ca;
  int i;

  // If container - Sum radioactive decay luminosity for each subcell
  if (cell->type()==cellContainer) { for (i=0;i<cell->subCell.size();i++) L+=LRadioCell(cell->subCell[i],t); return L; }

  // Calculate the 56Ni decay-chain luminosity
  M56Ni=cell->M(28,28);
  M56Co=cell->M(27,29);
  M56Fe=cell->M(26,30);
  decay56Ni(M56Ni,M56Co,M56Fe,cell->t(),t);
  L+=M56Ni*e56Ni+M56Co*e56Co;
  
  // Calculate the 57Ni decay-chain luminosity
  M57Ni=cell->M(28,29);
  M57Co=cell->M(27,30);
  M57Fe=cell->M(26,31);
  decay57Ni(M57Ni,M57Co,M57Fe,cell->t(),t);
  L+=M57Ni*e57Ni+M57Co*e57Co;
  
  // Calculate the 44Ti decay-chain luminosity
  M44Ti=cell->M(22,22);
  M44Ca=cell->M(20,24);
  decay44Ti(M44Ti,M44Ca,cell->t(),t);
  L+=M44Ti*e44Ti;

  return L;
}

// Grid radioactive decay luminosity
// Subcells and containers: Works
double LRadio(Grid* grid, const double& t) {
  double L=0;
  int i;

  // Sum radioactive decay luminosity for each cell
  for (i=0;i<grid->cell.size();i++) L+=LRadioCell(grid->cell[i],t);

  return L;
}

// Radioactive decay energy integrand
// Subcells and containers: Works
class ERadioIntegrand {
public:
  Grid* grid;
  ERadioIntegrand(Grid* _grid) : grid(_grid) {};
  double operator()(double t) const { return LRadio(grid,t); }
};

// Radioactive decay energy
// Subcells and containers: Works
double ERadio(Grid* grid, const double& t1, const double& t2) {
  double E;
  ERadioIntegrand integrand(grid);
  // Testing
  int nIter;

  E=Integrator<ERadioIntegrand>::romberg(integrand,t1,t2,nIter,1e-2);

  return E;
}

// Set diffusion temperature - Testing!
void setTDiff(Grid* grid, const int& cellIndexMin, const int& cellIndexMax, vector<double>& TCell, vector<vector<double> >& TSubCell, vector<double>& WCell, vector<vector<double> >& WSubCell) {
  int i,j;

  TCell.resize(grid->cell.size());
  TSubCell.resize(grid->cell.size());
  WCell.resize(grid->cell.size());
  WSubCell.resize(grid->cell.size());
  for (i=cellIndexMin;i<=cellIndexMax;i++) {
    if (grid->cell[i]->type()!=cellContainer) {
      if (grid->cell[i]->TDiff()!=noT) { 
	TCell[i]=grid->cell[i]->T(); grid->cell[i]->T(grid->cell[i]->TDiff());
	WCell[i]=grid->cell[i]->W(); grid->cell[i]->W(1.0); 
      }
    }
    else {
      TSubCell[i].resize(grid->cell[i]->subCell.size());
      WSubCell[i].resize(grid->cell[i]->subCell.size());
      for (j=0;j<grid->cell[i]->subCell.size();j++) {
	if (grid->cell[i]->subCell[j]->TDiff()!=noT) { 
	  TSubCell[i][j]=grid->cell[i]->subCell[j]->T(); grid->cell[i]->subCell[j]->T(grid->cell[i]->subCell[j]->TDiff());
	  WSubCell[i][j]=grid->cell[i]->subCell[j]->W(); grid->cell[i]->subCell[j]->W(1.0);
	}
      }
    }
  }
}

// Reset temperature - Testing!
void resetT(Grid* grid, const int& cellIndexMin, const int& cellIndexMax, vector<double>& TCell, vector<vector<double> >& TSubCell, vector<double>& WCell, vector<vector<double> >& WSubCell) {
  int i,j;

  for (i=cellIndexMin;i<=cellIndexMax;i++) {
    if (grid->cell[i]->type()!=cellContainer) {
      if (grid->cell[i]->TDiff()!=noT) {
	grid->cell[i]->T(TCell[i]);
	grid->cell[i]->W(WCell[i]);	
      }
    }
    else {
      for (j=0;j<grid->cell[i]->subCell.size();j++) {
	if (grid->cell[i]->subCell[j]->TDiff()!=noT) {
	  grid->cell[i]->subCell[j]->T(TSubCell[i][j]);
	  grid->cell[i]->subCell[j]->W(WSubCell[i][j]);
	}
      }
    }
  }
}

// Sample thermal pakets
// Subcells and containers: Check packetSamplerThermal!
void sampleThermalPackets(Grid* grid, const int& cellIndexMin, const int& cellIndexMax, const double& EPacket, const FrequencyBin& nu, PacketStateList& packet, PacketConfig* packetConfig, boost::mpi::communicator* mpiCom=NULL, double expFactorWeight=1.0, bool LCorr=false, bool eRadioCorr=false, bool useTDiff=false ) {
  double expFactor,E=0;
  double ECell; // New!
  Packet sPacket;
  int i,nPacket;
  // Testing!
  vector<double> TCell;
  vector<vector<double> > TSubCell;
  vector<double> WCell;
  vector<vector<double> > WSubCell;

  // If use diffusion temperature - Temporarily change to diffusion temperature - Testing!
  if (useTDiff) setTDiff(grid,cellIndexMin,cellIndexMax,TCell,TSubCell,WCell,WSubCell);

  // Calculate thermal energy in region
  //for (i=cellIndexMin;i<=cellIndexMax;i++) E+=grid->cell[i]->E()*grid->cell[i]->M();
  // Take dilution factor into account - Hack!!!
  //for (i=cellIndexMin;i<=cellIndexMax;i++) E+=grid->cell[i]->W()*grid->cell[i]->E()*grid->cell[i]->M();
  // Changed again!
  for (i=cellIndexMin;i<=cellIndexMax;i++) {
    ECell=grid->cell[i]->E();
    if (grid->cell[i]->W()!=1) ECell*=grid->cell[i]->W()*pow(grid->cell[i]->TR()/grid->cell[i]->T(),4);
    E+=ECell*grid->cell[i]->M();
    // If requested - correct for luminosity at cell borders - New!
    if (LCorr) { E-=expFactorWeight*(((CellSS*) grid->cell[i])->LBack()-((CellSS*) grid->cell[i])->LFront())*(grid->tMax()-grid->tMin()); }
    // If requested - correct for radioactive deposition - New!
    if (eRadioCorr) { E-=expFactorWeight*(grid->cell[i]->eRadio()*grid->cell[i]->M())*(grid->tMax()-grid->tMin()); }	
  }

  // Calculate number of thermal packets (corrected for expansion cooling)
  expFactor=(grid->tMin()+expFactorWeight*(grid->tMax()-grid->tMin()))/grid->tMin();
  nPacket=E*expFactor/EPacket;
 
  // Create new thermal packet sampler - Changed + Bug fix!
  PacketSamplerBBThermal packetSamplerThermal(nPacket,cellIndexMin,cellIndexMax,grid,nu,packetConfig,mpiCom,0,expFactorWeight,LCorr,eRadioCorr);
  // Hack - We need the old version of PacketSamplerBBThermal here! Or?
  //PacketSamplerBBThermal packetSamplerThermal(nPacket,cellIndexMin,cellIndexMax,grid,nu,packetConfig,mpiCom,0);
  
  // Sample thermal packets in new region and add to packet list
  #pragma omp parallel for firstprivate(sPacket)
  for (i=0;i<packetSamplerThermal.nPacket;i++) {
    // If MPI run
    if (mpiCom!=NULL) {
      // If the process is responsible for the packet
       if (i%mpiCom->size()==mpiCom->rank()) {
	// Sample packet
	sPacket=packetSamplerThermal.sample(i);
	// Add packet to list
        #pragma omp critical
	packet.push_back(PacketState(sPacket));
      }
    }
    // Else
   else {
     // Sample the packet
     sPacket=packetSamplerThermal.sample(i);
     // Add packet to list
     #pragma omp critical
     packet.push_back(PacketState(sPacket));
   }
  }

  // If use diffusion temperature - Restore temperature - Testing!
  if (useTDiff) resetT(grid,cellIndexMin,cellIndexMax,TCell,TSubCell,WCell,WSubCell);
}

// Sample radiation field
// Subcells and containers: Check packetSamplerI!
void sampleRadiationField(Grid* grid, const int& cellIndexMin, const int& cellIndexMax, const double& EPacket, PacketStateList& packet, PacketConfig* packetConfig, boost::mpi::communicator* mpiCom=NULL, double expFactorWeight=1.0, bool LCorr=false, bool eRadioCorr=false) {
  double expFactor,E=0;
  int i,nPacket;
  Packet sPacket;
  bool master=false;

  if (mpiCom==NULL) master=true;
  else if (mpiCom->rank()==0) master=true;

  // Calculate radiative energy in region
  //for (i=cellIndexMin;i<=cellIndexMax;i++) E+=(4*pi/c)*grid->cell[i]->J*grid->cell[i]->V();
  // Changed!
  for (i=cellIndexMin;i<=cellIndexMax;i++) {
    E+=(4*pi/c)*grid->cell[i]->J*grid->cell[i]->V();
    // If requested - correct for luminosity at cell borders - New!
    if (LCorr) { E-=expFactorWeight*(((CellSS*) grid->cell[i])->LBack()-((CellSS*) grid->cell[i])->LFront())*(grid->tMax()-grid->tMin()); }
    // If requested - correct for radioactive deposition - New!
    if (eRadioCorr) { E-=expFactorWeight*(grid->cell[i]->eRadio()*grid->cell[i]->M())*(grid->tMax()-grid->tMin()); }	
  }
	
  // Calculate number of packets (corrected for expansion cooling)
  expFactor=(grid->tMin()+expFactorWeight*(grid->tMax()-grid->tMin()))/grid->tMin();
  nPacket=E*expFactor/EPacket;
  
  if (master) cout<<"mc-diff-evo::sampleRadiationField: EPacket: "<<EPacket<<", nPacket"<<nPacket<<", expFactor: "<<expFactor<<", E: "<<E<<", tMin: "<<grid->tMin()<<", tMax: "<<grid->tMax()<<", expFactorWeight: "<<expFactorWeight<<"\n";

  // Create new intensity packet sampler
  PacketSamplerI packetSamplerI(nPacket,cellIndexMin,cellIndexMax,grid,packetConfig,mpiCom,0,expFactorWeight,LCorr,eRadioCorr);

  // Sample packets in region and add to packet list
  #pragma omp parallel for firstprivate(sPacket)
  for (i=0;i<packetSamplerI.nPacket;i++) {
    // If MPI run
    if (mpiCom!=NULL) {
      // If the process is responsible for the packet
      if (i%mpiCom->size()==mpiCom->rank()) {
	// Sample packet
	sPacket=packetSamplerI.sample(i);
	// Add packet to list
        #pragma omp critical
	packet.push_back(PacketState(sPacket));
      }
    }
    // Else
    else {
      // Sample the packet
      sPacket=packetSamplerI.sample(i);
      // Add packet to list
      #pragma omp critical
      packet.push_back(PacketState(sPacket));
    }
  }
  if (master) cout<<"sampleRadiationField: Sampled "<<packet.size()<<" packets\n";
}

// Sample initial packets
// Subcells and containers: Does not work yet!
void sampleInitialPackets(Grid* grid, const int& cellIndexMin, const int& cellIndexMax, const double& EPacket, const FrequencyBin& nu, PacketStateList& packet, PacketConfig* packetConfig, boost::mpi::communicator* mpiCom=NULL, double expFactorWeight=1.0, bool LCorr=false, bool eRadioCorr=false) {
  int i,radFieldBorder=cellIndexMax+1;

  // Find the border outside which we know the state of the radiation field
  for (i=0;i<grid->cell.size();i++) if (grid->cell[i]->Inu.size()!=0) { radFieldBorder=i; break; }
  
  // Sample thermal packets inside the border and add to packet list  
  if (cellIndexMin<radFieldBorder) sampleThermalPackets(grid,cellIndexMin,radFieldBorder-1,EPacket,nu,packet,packetConfig,mpiCom,expFactorWeight,LCorr,eRadioCorr);

  // Sample radiation field outside the border and add to packet list
  if (radFieldBorder<=cellIndexMax) sampleRadiationField(grid,MAX(radFieldBorder,cellIndexMin),cellIndexMax,EPacket,packet,packetConfig,mpiCom,expFactorWeight,LCorr,eRadioCorr);

  // Hack for restart - Fix in a (much) better way !!!
  // Does not work with subcells and containers yet!!!
  if (cellIndexMin>radFieldBorder) for (i=radFieldBorder;i<cellIndexMin;i++) {
      grid->cell[i]->radUpdated=false;
      // Testing
      grid->cell[i]->T(grid->cell[i]->TR()*pow(grid->cell[i]->W(),1.0/4.0));
    }
}

// Evolve radiation state
// Subcells and containers: Check packetSamplerBBL and packetSamplerBBLGamma
void evolveRadiationState(Grid* grid, const int& diffBorder, const int& diffBorderOld, const double& EPacket, const FrequencyBin& nu, const FrequencyBinList& nuList, const bool& includeRadio, PacketStateList& packetIn, PacketStateList& packetOut, PacketStat* packetStat, MCConfig* mCConfig, boost::mpi::communicator* mpiCom=NULL, bool injectPackets=true, double nInjectPacketMax=std::numeric_limits<int>::max()) { // Changed!
  double Ldiff;
  //int nPacket,nPacketGamma;
  double nPacket,nPacketGamma; // Changed!
  PacketSampler* packetSampler=NULL;
  double factor; // New!
  int i;
  bool master=false;

  if (mpiCom==NULL) master=true;
  else if (mpiCom->rank()==0) master=true;

  // If new packets are to be injected
  if (injectPackets) {

    // If diffusion border the same - Testing!!!
    //if (diffBorder==diffBorderOld) {

    // Get luminosity at diffusion border
    // Negative luminosties are taken care of below
    if (diffBorder>=0) Ldiff=MAX(((CellSS*) grid->cell[diffBorder])->LFront(),0);
    else Ldiff=0;
  
    // Calculate number of packets to inject
    nPacket=Ldiff*(grid->tMax()-grid->tMin())/EPacket;

    // Set fraction of packets to reflect at diffusion border
    // Takes care of negative luminosities
    //if (((CellSS*) grid->cell[diffBorder])->FFront()<0) mCConfig->packet->PReflect=MAX(1.0-fabs(((CellSS*) grid->cell[diffBorder])->FFront()/BlackBody::F((grid->cell[diffBorder]->T()+grid->cell[diffBorder+1]->T())/2)),0.0);
    //else mCConfig->packet->PReflect=1.0;
    mCConfig->packet->PReflectInner=1.0;
    
    // If include radioactive heating
    if (includeRadio) {
      
      // Calculate number of gamma-packets to inject
      nPacketGamma=ERadio(grid,grid->tMin(),grid->tMax())/EPacket;
      if (master) cout<<"mc-spec::evolveRadiationState: nPacketGamma: "<<nPacketGamma<<", tMin: "<<grid->tMin()<<", tMax: "<<grid->tMax()<<", ERadio: "<<ERadio(grid,grid->tMin(),grid->tMax())<<", EPacket: "<<EPacket<<"\n";
    }

    // If number of injected packets exceeds limit - New!
    if ((nPacket+nPacketGamma)>nInjectPacketMax) {
      // Calculate reduction factor
      factor=(nPacket+nPacketGamma)/nInjectPacketMax;
      // Reduce packet energy
      //EPacket=EPacket*factor;
      // Reduce number of packets
      nPacket=nPacket/factor;
      nPacketGamma=nPacketGamma/factor;
      if (master) cout<<"mc-spec::evolveRadiationState: Maximum number of packets exceeded - Packet energy increased and number of packets decreased, EPacket: "<<EPacket*factor<<", nPacket: "<<nPacket<<", nPacketGamma: "<<nPacketGamma<<"\n";
    }
    
    // Create new packetSampler
    if (includeRadio) packetSampler = new PacketSamplerBBLGamma(Ldiff,int(nPacket),int(nPacketGamma),nu,diffBorder+1,grid,mCConfig->packet,mpiCom); // Changed!
    else packetSampler = new PacketSamplerBBL(Ldiff,int(nPacket),diffBorder+1,grid,nu,mCConfig->packet); // Changed!
    
    //}
  }
      
  // Reset packet statistics
  if (includeRadio) packetStat->reset(grid,0,grid->cell.size()-1,true,nuList);
  else packetStat->reset(grid,diffBorder+1,grid->cell.size()-1,true,nuList);
  // Reset MC statistics (in outer region)
  grid->resetMCStatistics(diffBorder+1,grid->cell.size()-1,nuList);
  // If include radioactive energy deposition
  if (includeRadio) {
    // Reset g-packet MC statistics (in inner region)
    for (i=0;i<=diffBorder;i++) grid->cell[i]->resetGMCStatistics();
  }
        
  // Create new MC experiment
  MCExperiment mCExperiment(packetSampler,packetStat,mCConfig,mpiCom);
      
  // Make sure bound-free opacity tables are updated - Testing!
  if (master) cout<<"mc-spec::evolveRadiationState: Update bound-free opacity tables\n";
  boundFreeTableUpdate(grid,diffBorder+1,grid->cell.size()-1,mpiCom);
  if (mpiCom!=NULL) mpiCom->barrier();

  // Evolve MC experiment
  mCExperiment.evolve(packetIn,packetOut,grid);
      
  // Update radiation state (in outer region)
  grid->updateRadiationState(diffBorder+1,grid->cell.size()-1,mCExperiment);

  // If include radioactive heating
  if (includeRadio) {
    // Update radioactive heating (in inner region)
    // Only works with on-the-fly update method !
    for (i=0;i<=diffBorder;i++) grid->cell[i]->updateERadio();
  }

  // Cleanup
  if (packetSampler!=NULL) delete packetSampler;
}

// Export statistics
// Subcells and containers: Works
void exportStat(GridSS& grid, PacketStat& packetStat, string& gridStatDir, string& mCStatDir, int& n) {

  // Export grid statistics
  if (gridStatDir.compare("")!=0) grid.exportStat(statFile(gridStatDir,"grid-stat",n));
  // Export packet statistics
  if (mCStatDir.compare("")!=0) packetStat.exportStat(statFile(mCStatDir,"mc-stat",n));
  // Export spectrum
  if (mCStatDir.compare("")!=0) packetStat.exportSpectrum(statFile(mCStatDir,"spectrum",n));
  // Export spectrum per transition type
  if (mCStatDir.compare("")!=0) packetStat.exportSpectrumTrans(statFile(mCStatDir,"spectrum-trans",n));
  // Export observed spectral evolution
  if (mCStatDir.compare("")!=0) packetStat.specEvoObs->exportUpdated(mCStatDir,"spec-obs");
}

double getEPacket(GridSS& grid, const double& nPacketEff, const double& tStop, const bool& includeRadio) {
  double EPacket,E=0;
  int i,radFieldBorder=grid.cell.size();

  // Find the border outside which we know the state of the radiation field
  for (i=0;i<grid.cell.size();i++) if (grid.cell[i]->Inu.size()!=0) { radFieldBorder=i; break; }

  // Calculate total thermal energy inside the border
  //for (i=0;i<radFieldBorder;i++) E+=grid.cell[i]->E()*grid.cell[i]->M();
  // Take dilution factor into account - Hack!!!
  for (i=0;i<radFieldBorder;i++) E+=grid.cell[i]->W()*grid.cell[i]->E()*grid.cell[i]->M();

  // Calculate total radiative energy outside the border
  for (i=radFieldBorder;i<grid.cell.size();i++) E+=(4*pi/c)*grid.cell[i]->J*grid.cell[i]->V();

  // If include radioactive heating
  if (includeRadio) {
    
    // Calculate and add total radioactive energy
    E+=ERadio(&grid,grid.t(),tStop);
  }
    
  // Calculate energy per packet
  EPacket=E/nPacketEff;

  return EPacket;
}

double getEPacketL(GridSS& grid, const double& nPacketLdt, const double& dt) {
  double EPacket,L,E;

  // Get luminosity
  L=((CellSS*) grid.cell[grid.cell.size()-1])->LFront();

  // Calculate energy emitted in one time-step
  E=L*dt;
    
  // Calculate energy per packet
  EPacket=E/nPacketLdt;

  return EPacket;
}

// Update non-thermal state (on a single computer)
void updateNTStateSingle(Grid& grid, Koz92Solver& ntSolver, int iMin, int iMax) {
  vector<Cell*> cellWork(0);
  int i,j;

  // Prevent faulty cell range
  iMin=(iMin>=0)?iMin:0; iMax=(iMax<=(grid.cell.size()-1))?iMax:(grid.cell.size()-1);

  // Create work list of cells and subcells
  for (i=iMin;i<=iMax;i++) {
    if (grid.cell[i]->type()==cellContainer) for (j=0;j<grid.cell[i]->subCell.size();j++) cellWork.push_back(grid.cell[i]->subCell[j]);
    else cellWork.push_back(grid.cell[i]);
  }
 
  // Update non-thermal state for the cells the process are responsible for
  #pragma omp parallel for
  for (i=0;i<cellWork.size();i++) {
    cout<<"mc-spec::updateNTState: Update cell "<<cellWork[i]->index()<<" non-thermal state\n";
    ntSolver.update(cellWork[i],true);
  }
}

// Update non-thermal state (on a cluster using MPI)
void updateNTStateCluster(Grid& grid, Koz92Solver& ntSolver, int iMin, int iMax, boost::mpi::communicator* mpiCom=NULL) {
  vector<Cell*> cellWork(0);
  int i,j;
  bool master=false;

  if (mpiCom==NULL) master=true;
  else if (mpiCom->rank()==0) master=true;

  // Prevent faulty cell range
  iMin=(iMin>=0)?iMin:0; iMax=(iMax<=(grid.cell.size()-1))?iMax:(grid.cell.size()-1);

  // Create work list of cells and subcells
  for (i=iMin;i<=iMax;i++) {
    if (grid.cell[i]->type()==cellContainer) for (j=0;j<grid.cell[i]->subCell.size();j++) cellWork.push_back(grid.cell[i]->subCell[j]);
    else cellWork.push_back(grid.cell[i]);
  }
 
  // Determine which processes are responsible for the cells
  for (i=0;i<cellWork.size();i++) cellWork[i]->responsible=cellWork[i]->index()%mpiCom->size();

  // Update non-thermal state for the cells the process are responsible for
  #pragma omp parallel for
  for (i=0;i<cellWork.size();i++) {
    //if (master) cout<<"mc-spec::updateNTState: Update cell "<<cellWork[i]->index()<<" non-thermal state\n";
    if (cellWork[i]->responsible==mpiCom->rank()) {
      cout<<"mc-spec::updateNTState: Update cell "<<cellWork[i]->index()<<" non-thermal state\n";
      // This line sometimes gives rise to core dumps - Investigate!
      ntSolver.update(cellWork[i],true);
    }
  }

  // Wait for all processes to reach this point
  mpiCom->barrier();

  // Synchronize matter state for each cell
  for (i=0;i<cellWork.size();i++) {
    if (master) cout<<"mce-spec::updateNTState: Synchronize cell "<<cellWork[i]->index()<<" non-thermal state\n";
    cellWork[i]->syncMatterState();
  }

  // Wait for all processes to reach this point
  mpiCom->barrier();
}

// Update non-thermal state
void updateNTState(Grid& grid, Koz92Solver& ntSolver, int iMin, int iMax, boost::mpi::communicator* mpiCom=NULL) {
  if (mpiCom!=NULL) updateNTStateCluster(grid,ntSolver,iMin,iMax,mpiCom);
  else updateNTStateSingle(grid,ntSolver,iMin,iMax);
}

// Init cell non-thermal state
// Subcells and containers: Works
void initNTStateCell(Cell* cell, string configFile) {
  int i;

  // If container - Init non-thermal state for each subcell and return
  if (cell->type()==cellContainer) { for (i=0;i<cell->subCell.size();i++) initNTStateCell(cell->subCell[i],configFile); return; }
    
  // Initiate non-thermal state
  cell->nonThermalHeating(1e-99);
  cell->nonThermalIonization(0.0);
  cell->nonThermalExcitation(0.0);
}

// Init grid non-thermal state
// Subcells and containers: Works
void initNTState(Grid& grid, string configFile) {
  int i;

  // Init non-thermal state for each cell
  for (i=0;i<grid.cell.size();i++) initNTStateCell(grid.cell[i],configFile);
}

// Evolve radioactive state of grid
// Subcells and containers: Check packetSamplerGamma
void evolveRadio(Grid& grid, const double& dt, const double EPacket, PacketStateList& packetIn, MCConfig* mCConfig, boost::mpi::communicator* mpiCom=NULL) {
  PacketSamplerGamma* packetSamplerGamma;
  PacketStateList packetGammaOut(0);
  PacketStateList packetGammaIn(0);
  int nPacketGamma,i;
  
  // For each cell evolve radioactive state (from the last time-step)
  for (i=0;i<grid.cell.size();i++) grid.cell[i]->evolveRadio(grid.t()-dt,dt);

  // Sort out gamma packets from list
  for (i=0;i<packetIn.size();i++) if (packetIn[i].type==gPacket) packetGammaIn.push_back(packetIn[i]);

  // Calculate number of gamma-packets to inject
  nPacketGamma=ERadio(&grid,grid.tMin(),grid.tMax())/EPacket;

  // Turn off conversion of g-packets
  mCConfig->packet->convertGPacket=false;

  // Create new g-packet sampler
  packetSamplerGamma = new PacketSamplerGamma(nPacketGamma,&grid,mCConfig->packet,mpiCom);

  // Reset g-packet MC statistics
  for (i=0;i<grid.cell.size();i++) grid.cell[i]->resetGMCStatistics();
        
  // Create new MC experiment
  MCExperiment mCExperiment(packetSamplerGamma,NULL,mCConfig,mpiCom);
      
  // Evolve MC experiment
  mCExperiment.evolve(packetGammaIn,packetGammaOut,&grid);

  // Turn on conversion of g-packets
  // Assuming this was set to true - Is this ok?
  mCConfig->packet->convertGPacket=true;
      
  // Update radioactive heating
  // Only works with on-the-fly update method !
  for (i=0;i<grid.cell.size();i++) grid.cell[i]->updateERadio();

  // Cleanup
  if (packetSamplerGamma!=NULL) delete packetSamplerGamma;
}

// New version!
// Subcells and containers: May not work yet!
void initBoostFactor(GridSS& grid, BoostConfig& boostConfig, FrequencyBin& nu, boost::mpi::communicator* mpiCom=NULL) {
  double dlognu;
  BoostRegionConfig* region;
  vector<double> nuWork1,nuWork2;
  vector<Cell*> cellWork(0);
  int i,j;
  bool master=false;

  if (mpiCom==NULL) master=true;
  else if (mpiCom->rank()==0) master=true;

  // Create work list of cells and subcells
  for (i=0;i<grid.cell.size();i++) {
    if (grid.cell[i]->type()==cellContainer) for (j=0;j<grid.cell[i]->subCell.size();j++) cellWork.push_back(grid.cell[i]->subCell[j]);
    cellWork.push_back(grid.cell[i]);
  }

  // Calculate frequencies for boost bins
  for (i=0;i<boostConfig.region.size();i++) {
    region=&(boostConfig.region[i]);
    if (region->nu.max>nu.min and region->nu.min<nu.max) {
      dlognu=(log(MIN(region->nu.max,nu.max))-log(MAX(region->nu.min,nu.min)))/float(region->nBin);
      for (j=0;j<region->nBin;j++) nuWork1.push_back(MAX(region->nu.min,nu.min)*exp(j*dlognu));
      nuWork1.push_back(MIN(region->nu.max,nu.max));
    }
  }

  // Add frequency limits 
  nuWork1.push_back(nu.min);
  nuWork1.push_back(nu.max);

  // Sort frequencies
  sort(nuWork1.begin(),nuWork1.end());

  // Remove doublets
  nuWork2.push_back(nuWork1[0]); 
  for (i=1;i<nuWork1.size();i++) if (nuWork1[i]!=nuWork1[i-1]) nuWork2.push_back(nuWork1[i]);

  // For each cell
  for (i=0;i<cellWork.size();i++) {

    // Resize boost vectors
    cellWork[i]->nuBoost.resize(nuWork2.size()-1);
    cellWork[i]->boost.resize(nuWork2.size()-1);

    // Initiate boost bins
    for (j=0;j<cellWork[i]->nuBoost.size();j++) { 
      cellWork[i]->nuBoost[j].min=nuWork2[j]; 
      cellWork[i]->nuBoost[j].max=nuWork2[j+1];
      cellWork[i]->boost[j]=1.0;
      if (master and i==0) cout<<"mc-spec::initBoostFactor: nuBin: "<<j<<", lambda: ["<<(c/cellWork[i]->nuBoost[j].min)/AA<<":"<<(c/cellWork[i]->nuBoost[j].max)/AA<<"]\n";
    }
  }
}

// New version!
// Subcells and containers: May not work yet!
void adjustBoostFactor(GridSS& grid, int diffBorder, BoostConfig& boostConfig, bool subBoost=false, boost::mpi::communicator* mpiCom=NULL) {
  double boost,nJ,nBin;
  BoostRegionConfig* region;
  FrequencyBinIterator nuBin,nuBinMin,nuBinMax;
  vector<Cell*> cellWork(0);
  int i,j,k;
  bool master=false;

  if (mpiCom==NULL) master=true;
  else if (mpiCom->rank()==0) master=true;

  // Create work list of cells and subcells
  for (i=diffBorder+1;i<grid.cell.size();i++)
    if (grid.cell[i]->type()==cellContainer and subBoost) for (j=0;j<grid.cell[i]->subCell.size();j++) cellWork.push_back(grid.cell[i]->subCell[j]); // Changed!
    else cellWork.push_back(grid.cell[i]);

  // For each cell in work list
  for (i=0;i<cellWork.size();i++) {

    // For each boost bin
    for (j=0;j<cellWork[i]->nuBoost.size();j++) {
      
      // Search boost region
      region=NULL;
      for (k=0;k<boostConfig.region.size();k++)
	if (cellWork[i]->nuBoost[j].max>boostConfig.region[k].nu.min and cellWork[i]->nuBoost[j].min<boostConfig.region[k].nu.max) { region=&(boostConfig.region[k]); break; }
      
      // If region found
      if (region!=NULL) {
	
	// Find overlapping frequency bins
	nuBinMin=cellWork[i]->nu.find(cellWork[i]->nuBoost[j].min);
	nuBinMax=cellWork[i]->nu.find(cellWork[i]->nuBoost[j].max);
	
	// Initialize number of bins and mean intensity events
	nBin=0; nJ=0;
	
	// Count number of bins and mean intensity events
	for (nuBin=nuBinMin;nuBin!=nuBinMax;nuBin++) { 
	  // Exclude edge bins
	  if (not nuBin->edge) nBin++;
	  nJ+=cellWork[i]->nJ[nuBin-cellWork[i]->nu.begin()]; 
	}
	// To be on the safe side
	nBin=MAX(nBin,1);
	
	// Calculate boost factor needed
	boost=(region->nJGood*nBin/MAX(nJ,1.0));
	cellWork[i]->boost[j]=MAX(MIN(cellWork[i]->boost[j]*boost,region->boostMax),region->boostMin);
	
	if (master) cout<<"mc-spec::adjustBoostFactor: cell: "<<i<<", lambdaBoost: ["<<(c/cellWork[i]->nuBoost[j].max)/AA<<":"<<(c/cellWork[i]->nuBoost[j].min)/AA<<"], boost: "<<boost<<", (new) boostFactor: "<<cellWork[i]->boost[j]<<", nJ: "<<nJ/nBin<<", neRadio: "<<cellWork[i]->neRadio<<"\n";
      }
      // Else - Set boost factor to one
      else cellWork[i]->boost[j]=1.0;
    }

    // If cell-container and sub-boost flag not set - New!
    if (cellWork[i]->type()==cellContainer and not subBoost) {
      // Copy boost factor for container to sub-cells 
      for (j=0;j<cellWork[i]->subCell.size();j++) cellWork[i]->subCell[j]->boost=cellWork[i]->boost;
    }
  }
}

// Subcells and containers: May not work yet! - Not used - Remove!
/*void updateBoostFactor(Grid& grid, int cellIndexMin=0) {
  int i;

  //Update boost factors
  #pragma omp parallel for
  for (i=cellIndexMin;i<grid.cell.size();i++) grid.cell[i]->boostMC();
  }*/

// Log memory usage
void logMemory(GridSS& grid, PacketStateList& packetIn, PacketStateList& packetOut, PacketStat& packetStat, boost::mpi::communicator* mpiCom=NULL) {
  double memPacketIn=0,memPacketOut=0,memPacketStat=0;
  int i;

  for (i=0;i<packetIn.size();i++) memPacketIn+=sizeof(PacketState);
  for (i=0;i<packetOut.size();i++) memPacketOut+=sizeof(PacketState);

  if (mpiCom!=NULL) cout<<"mc-diff-evo: Process: "<<mpiCom->rank()<<", Allocated memory: "<<(grid.mem()+memPacketIn+memPacketOut+packetStat.mem())/1e9<<" GB, grid: "<<grid.mem()/1e9<<" ("<<grid.memPList()/1e9<<","<<grid.memTransBB()/1e9<<","<<grid.memTransBF()/1e9<<") GB, packet(In/Out): "<<memPacketIn/1e9<<"/"<<memPacketOut/1e9<<" GB, packetStat: "<<packetStat.mem()/1e9<<" GB\n";
  else cout<<"mc-diff-evo: Allocated memory: "<<(grid.mem()+memPacketIn+memPacketOut+packetStat.mem())/1e9<<" GB, grid: "<<grid.mem()/1e9<<" GB, packet(In/Out): "<<memPacketIn/1e9<<"/"<<memPacketOut/1e9<<" GB, packetStat: "<<packetStat.mem()/1e9<<" GB\n";
}

void logSysMemory(int n, string label) {
  stringstream header;
  int status;

  header.str(""); 
  header<<"echo \"n: "<<n<<", "<<label<<"\" >> mc-spec-ps-$(hostname).log"; 
  status=system(header.str().c_str());
  status=system("date >> mc-ps-$(hostname).log");
  status=system("./mc-spec-ps.sh >> mc-spec-ps-$(hostname).log");
}

// Set absorption processes
Processes setAbsProcesses(bool electronScatter, bool boundFree, bool freeFree, bool boundBound, bool greyScatter) {
  unsigned int processesFlags=0;
  if (electronScatter) processesFlags|=pEScatter;
  if (boundFree) processesFlags|=pBoundFree;
  if (freeFree) processesFlags|=pFreeFree;
  if (boundBound) processesFlags|=pBoundBound;
  if (greyScatter) processesFlags|=pGScatter;
  return Processes(processesFlags);
}

// Main program
int main (int argc, char** argv) {

  PacketStateList packetIn(0);
  PacketStateList packetOut(0);
  bool initiate,updateMCDerived;
  double E=0,EOuter=0,ENew=0,EPacket,Rdiff,ROnTheSpot,Ldiff;
  double t,tMin=-1,tMax=-1,dt=-1,dtDiff,dtLeft,dlogt=0.01/2.5,expFactorWeight; // Changed!
  int nPacket,diffBorder,diffBorderNew,diffBorderOld=-1,diffBorderMax,onTheSpotBorder,nIterLambdaUse,n,m,i,j;
  FrequencyBinList nuList;
  vector<double> nuVector;
  PacketSamplerBBL* packetSamplerL=NULL;
  PacketSamplerBBThermal* packetSamplerThermal=NULL;
  MCExperiment* mCExperiment;
  bool injectPackets;
  boost::mpi::environment* mpiEnv=NULL;
  boost::mpi::communicator* mpiCom=NULL;
  bool master=true;
  Processes absProcesses;

  /* Check if enough number of arguments provided */
  if (argc<2) {
    printf("Missing argument\n");
    exit(-1);
  }
  
  /* Get mandatory arguments */
  string gridDataFile=string(argv[1]);
  int nIterMin=atoi(argv[2]);
  int nIterMax=atoi(argv[3]);

  /* Set default values for optional arguments */
  int tStop=100*day;
  double nPacketEff=10000000;
  double nPacketLdt=1000000;
  double nPacketMax=std::numeric_limits<int>::max(); // New!
  //int nBin=10000;
  int nBin=5000; // Changed!
  int nMuBin=20; // New!
  double tauLim=1e-3;
  double lambdaMin=1;
  double lambdaMax=25000;
  double tauDiff=2e2;
  bool useTauDiff=false;
  bool seed=false;
  bool eScatter=true;
  bool lineScatter=true;
  bool gScatter=false;
  bool freeFreeAbs=false;
  bool boundFreeAbs=false;
  bool flourescence=false;
  bool collProc=false;
  bool nonThermalProc=false;
  bool twoPhotonProc=false;
  bool chargeTransferProc=false;
  bool chargeTransferEmit=false;
  bool transTable=false;
  bool markovTable=false;
  bool markovTableExt=false;
  bool markovTableExtSplit=false;
  double markovTableLim=0.0;
  double markovTableThermLim=0.0;
  double markovTableCTLim=0.0;
  bool collExciteTable=false;
  double collExciteTableLim=0.0;  
  bool nonThermalExciteTable=false;
  bool etaTable=false;
  UpdateMethod updateMethod=urOnTheFly; // Not read!
  bool boundFreeTable=false;
  double boundFreeTableMin=1; // Not read - Deprecated!
  double boundFreeTableMax=1; // Not read - Deprecated!
  int boundFreeTableNBin=1; // Not read - Deprecated!
  bool boundFreeSTable=false;
  int boundFreeSTableNBin=1; // Not read - Deprecated!
  RadEstimator radEstimator=rfDBB;
  RadEstimator radEstimatorBB=rfDBB;
  bool radEstimatorExp=false;
  OnTheSpot onTheSpot("");
  OnTheSpotNew onTheSpotNew("");
  double tauOnTheSpot=10;
  double tauEdgeOnTheSpot=0;
  FFOnTheSpot ffOnTheSpot(false,10.0); // No read!
  bool boost=false;
  bool subBoost=false;
  string boostConfigFile="";
  int nBoostBin=1; // Not read - Deprecated!
  double nJGood=1000;
  double boostMin=0.25;
  double boostMax=1e5;
  bool alfaCorrect=false;
  bool alfaCorrectEmit=false; // Not read - Deprecated! - Wrong - Still used!
  bool alfaSimpleEmit=false;
  bool alfaSPH=false;
  bool forceDBalance=false;
  bool forceDBalanceOSP=false;
  bool boundFreeBB=false;
  double boundFreeBBTauLim=1e3;
  double continuumBBTauLim=1e99;
  bool convertGPacket=true;
  bool transformKappa=false;
  bool bbEPacket=false;
  string configFile="";
  bool includeRadio=false;
  bool exportEta=false;
  bool exportSBF=false;
  bool reflectOuter=false;
  bool reflectInnerGP=false;
  bool transformRay=false;
  bool limitFreeFreeCRange=false; // Not read!
  string gridStatDir="";
  string mCStatDir="";
  string gridStateDir="";
  bool JLineFallback=false;
  double JLineFallbackFactor=1.1;
  int JLineFallbackCount=3;
  bool JLineFallbackDBB=false;
  bool trackI=false; // Not read!
  bool trackHK=false; // Not read!
  double boundFreeCutOff=0.0;
  double boundFreeCutOffRoss=0.0;
  bool mpi=false;
  EPacketMethod ePacketMethod=epEBased;
  bool dustAbs=false;
  bool useTDiff=true;
  int nCellOuterMin=1;
  bool loadPop=true;
  bool loadRad=true;
  bool matterDerivedMPI=false;
  bool LTEGroupMacro=false;
  bool exportKappaMC2=true;

  /* Get optional arguments */
  for (i=2;i<argc;i++) {
    if (strcmp(argv[i],"--t_stop")==0 && i+1<argc) 
      {i++; tStop=(float)strtod(argv[i],NULL)*day;}
    else if (strcmp(argv[i],"--npacket_eff")==0 && i+1<argc)
      {i++; nPacketEff=(float)strtod(argv[i],NULL);}
    else if (strcmp(argv[i],"--npacket_Ldt")==0 && i+1<argc) // Testing!
      {i++; nPacketLdt=(float)strtod(argv[i],NULL);}
    else if (strcmp(argv[i],"--npacket_max")==0 && i+1<argc)
      {i++; nPacketMax=(float)strtod(argv[i],NULL);}
    else if (strcmp(argv[i],"--nbin")==0 && i+1<argc)
      {i++; nBin=atoi(argv[i]);}
    else if (strcmp(argv[i],"--taulim")==0 && i+1<argc)
      {i++; tauLim=(float)strtod(argv[i],NULL);}
    else if (strcmp(argv[i],"--lambda_min")==0 && i+1<argc)
      {i++; lambdaMin=(float)strtod(argv[i],NULL);}
    else if (strcmp(argv[i],"--lambda_max")==0 && i+1<argc)
      {i++; lambdaMax=(float)strtod(argv[i],NULL);}
    else if (strcmp(argv[i],"--taudiff")==0 && i+1<argc)
      {i++; tauDiff=(float)strtod(argv[i],NULL);}
    else if (strcmp(argv[i],"--use_taudiff")==0 && i+1<argc)
      {i++; if (strcmp(argv[i],"yes")==0) useTauDiff=true;}
    else if (strcmp(argv[i],"--seed")==0 && i+1<argc)
      {i++; if (strcmp(argv[i],"yes")) seed=true;} // Bug! Fix!
    else if (strcmp(argv[i],"--escatter")==0 && i+1<argc)
      {i++; if (strcmp(argv[i],"yes")==0) eScatter=true;}
    else if (strcmp(argv[i],"--linescatter")==0 && i+1<argc)
      {i++; if (strcmp(argv[i],"no")==0) lineScatter=false;}
     else if (strcmp(argv[i],"--gscatter")==0 && i+1<argc)
      {i++; if (strcmp(argv[i],"yes")==0) gScatter=true;}
    else if (strcmp(argv[i],"--freefree_abs")==0 && i+1<argc)
      {i++; if (strcmp(argv[i],"yes")==0) freeFreeAbs=true;}
    else if (strcmp(argv[i],"--boundfree_abs")==0 && i+1<argc)
      {i++; if (strcmp(argv[i],"yes")==0) boundFreeAbs=true;}
    else if (strcmp(argv[i],"--flourescence")==0 && i+1<argc)
      {i++; if (strcmp(argv[i],"yes")==0) flourescence=true;}
    else if (strcmp(argv[i],"--coll_proc")==0 && i+1<argc)
      {i++; if (strcmp(argv[i],"yes")==0) collProc=true;}
    else if (strcmp(argv[i],"--non_thermal_proc")==0 && i+1<argc)
      {i++; if (strcmp(argv[i],"yes")==0) nonThermalProc=true;}
    else if (strcmp(argv[i],"--two_photon_proc")==0 && i+1<argc)
      {i++; if (strcmp(argv[i],"yes")==0) twoPhotonProc=true;}
    else if (strcmp(argv[i],"--charge_transfer_proc")==0 && i+1<argc)
      {i++; if (strcmp(argv[i],"yes")==0) chargeTransferProc=true;}
    else if (strcmp(argv[i],"--charge_transfer_emit")==0 && i+1<argc)
      {i++; if (strcmp(argv[i],"yes")==0) chargeTransferEmit=true;}
    else if (strcmp(argv[i],"--trans_table")==0 && i+1<argc)
      {i++; if (strcmp(argv[i],"yes")==0) transTable=true;}
    else if (strcmp(argv[i],"--markov_table")==0 && i+1<argc)
      {i++; if (strcmp(argv[i],"yes")==0) markovTable=true;}
    else if (strcmp(argv[i],"--markov_table_ext")==0 && i+1<argc)
      {i++; if (strcmp(argv[i],"yes")==0) markovTableExt=true;}
    else if (strcmp(argv[i],"--markov_table_lim")==0 && i+1<argc) 
      {i++; markovTableLim=(float)strtod(argv[i],NULL);}
    else if (strcmp(argv[i],"--markov_table_ext_split")==0 && i+1<argc)
      {i++; if (strcmp(argv[i],"yes")==0) markovTableExtSplit=true;}
    else if (strcmp(argv[i],"--markov_table_therm_lim")==0 && i+1<argc) 
      {i++; markovTableThermLim=(float)strtod(argv[i],NULL);}
    else if (strcmp(argv[i],"--markov_table_ct_lim")==0 && i+1<argc) 
      {i++; markovTableCTLim=(float)strtod(argv[i],NULL);}
    else if (strcmp(argv[i],"--coll_exc_table")==0 && i+1<argc)
      {i++; if (strcmp(argv[i],"yes")==0) collExciteTable=true;}
    else if (strcmp(argv[i],"--coll_exc_table_lim")==0 && i+1<argc) 
      {i++; collExciteTableLim=(float)strtod(argv[i],NULL);}
    else if (strcmp(argv[i],"--non_thermal_exc_table")==0 && i+1<argc)
      {i++; if (strcmp(argv[i],"yes")==0) nonThermalExciteTable=true;}
    else if (strcmp(argv[i],"--eta_table")==0 && i+1<argc)
      {i++; if (strcmp(argv[i],"yes")==0) etaTable=true;}
    else if (strcmp(argv[i],"--boundfree_table")==0 && i+1<argc)
      {i++; if (strcmp(argv[i],"yes")==0) boundFreeTable=true;}
    else if (strcmp(argv[i],"--boundfree_stable")==0 && i+1<argc)
      {i++; if (strcmp(argv[i],"yes")==0) boundFreeSTable=true;}
    else if (strcmp(argv[i],"--rad_estimator")==0 && i+1<argc)
      {
	i++; 
	if (strcmp(argv[i],"lucy")==0) radEstimator=rfLucy;
	if (strcmp(argv[i],"bb")==0) radEstimator=rfBB;
      }
    else if (strcmp(argv[i],"--rad_estimator_bb")==0 && i+1<argc)
      {
	i++; 
	if (strcmp(argv[i],"lucy")==0) radEstimatorBB=rfLucy;
	if (strcmp(argv[i],"bb")==0) radEstimator=rfBB;
      }
    else if (strcmp(argv[i],"--rad_estimator_exp")==0 && i+1<argc)
      { i++; if (strcmp(argv[i],"yes")==0) radEstimatorExp=true; }
    else if (strcmp(argv[i],"--onthespot")==0 && i+1<argc)
      {i++; onTheSpot=OnTheSpot(argv[i]);}
    else if (strcmp(argv[i],"--onthespot_new")==0 && i+1<argc)
      {i++; onTheSpotNew=OnTheSpotNew(argv[i]);}
    else if (strcmp(argv[i],"--tau_onthespot")==0 && i+1<argc)
      {i++; tauOnTheSpot=(float)strtod(argv[i],NULL);}
    else if (strcmp(argv[i],"--tau_edge_onthespot")==0 && i+1<argc)
      {i++; tauEdgeOnTheSpot=(float)strtod(argv[i],NULL);}
    else if (strcmp(argv[i],"--ff_onthespot_taulim")==0 && i+1<argc)
      {i++; ffOnTheSpot.tauLim=(float)strtod(argv[i],NULL);}
    else if (strcmp(argv[i],"--boost")==0 && i+1<argc)
      {i++; if (strcmp(argv[i],"yes")==0) boost=true;}
    else if (strcmp(argv[i],"--sub_boost")==0 && i+1<argc)
      {i++; if (strcmp(argv[i],"yes")==0) subBoost=true;}
    else if (strcmp(argv[i],"--boost_config_file")==0 && i+1<argc)
      {i++; boostConfigFile=string(argv[i]);}
    else if (strcmp(argv[i],"--alfa_correct")==0 && i+1<argc)
      {i++; if (strcmp(argv[i],"yes")==0) alfaCorrect=true;}
    else if (strcmp(argv[i],"--alfa_correct_emit")==0 && i+1<argc) // Re-added - Still used!
      {i++; if (strcmp(argv[i],"yes")==0) alfaCorrectEmit=true;}
    else if (strcmp(argv[i],"--alfa_simple_emit")==0 && i+1<argc)
      {i++; if (strcmp(argv[i],"yes")==0) alfaSimpleEmit=true;}
    else if (strcmp(argv[i],"--alfa_SPH")==0 && i+1<argc)
      {i++; if (strcmp(argv[i],"yes")==0) alfaSPH=true;}
    else if (strcmp(argv[i],"--force_dbalance")==0 && i+1<argc)
      {i++; if (strcmp(argv[i],"yes")==0) forceDBalance=true;}
    else if (strcmp(argv[i],"--force_dbalance_osp")==0 && i+1<argc)
      {i++; if (strcmp(argv[i],"yes")==0) forceDBalanceOSP=true;}
    else if (strcmp(argv[i],"--convert_gpacket")==0 && i+1<argc)
      {i++; if (strcmp(argv[i],"no")==0) convertGPacket=false;}
    else if (strcmp(argv[i],"--transform_kappa")==0 && i+1<argc)
      {i++; if (strcmp(argv[i],"yes")==0) transformKappa=true;}
    else if (strcmp(argv[i],"--bb_epacket")==0 && i+1<argc)
      {i++; if (strcmp(argv[i],"yes")==0) bbEPacket=true;}
    else if (strcmp(argv[i],"--boundfree_bb")==0 && i+1<argc)
      {i++; if (strcmp(argv[i],"yes")==0) boundFreeBB=true;}
    else if (strcmp(argv[i],"--boundfree_bb_taulim")==0 && i+1<argc)
      {i++; boundFreeBBTauLim=(float)strtod(argv[i],NULL);}
    else if (strcmp(argv[i],"--continuum_bb_taulim")==0 && i+1<argc)
      {i++; continuumBBTauLim=(float)strtod(argv[i],NULL);}
    else if (strcmp(argv[i],"--config_file")==0 && i+1<argc)
      {i++; configFile=string(argv[i]);}
    else if (strcmp(argv[i],"--include_radio")==0 && i+1<argc)
      {i++; if (strcmp(argv[i],"yes")==0) includeRadio=true;}
    else if (strcmp(argv[i],"--export_eta")==0 && i+1<argc)
      {i++; if (strcmp(argv[i],"yes")==0) exportEta=true;}
    else if (strcmp(argv[i],"--export_sbf")==0 && i+1<argc)
      {i++; if (strcmp(argv[i],"yes")==0) exportSBF=true;}
    else if (strcmp(argv[i],"--reflect_outer")==0 && i+1<argc)
      {i++; if (strcmp(argv[i],"yes")==0) reflectOuter=true;}
    else if (strcmp(argv[i],"--reflect_inner_gp")==0 && i+1<argc)
      {i++; if (strcmp(argv[i],"yes")==0) reflectInnerGP=true;}
    else if (strcmp(argv[i],"--transform_ray")==0 && i+1<argc)
      {i++; if (strcmp(argv[i],"yes")==0) transformRay=true;}
    else if (strcmp(argv[i],"--grid_stat_dir")==0 && i+1<argc)
      {i++; gridStatDir=string(argv[i]);}
    else if (strcmp(argv[i],"--mc_stat_dir")==0 && i+1<argc)
      {i++; mCStatDir=string(argv[i]);}
    else if (strcmp(argv[i],"--grid_state_dir")==0 && i+1<argc)
      {i++; gridStateDir=string(argv[i]);}
    else if (strcmp(argv[i],"--jline_fallback")==0 && i+1<argc)
      {i++; if (strcmp(argv[i],"yes")==0) JLineFallback=true;}
    else if (strcmp(argv[i],"--jline_fallback_factor")==0 && i+1<argc)
      {i++; JLineFallbackFactor=(float)strtod(argv[i],NULL);}
    else if (strcmp(argv[i],"--jline_fallback_count")==0 && i+1<argc) 
      {i++; JLineFallbackCount=atoi(argv[i]);}
    else if (strcmp(argv[i],"--jline_fallback_dbb")==0 && i+1<argc)
      {i++; if (strcmp(argv[i],"yes")==0) JLineFallbackDBB=true;}
    else if (strcmp(argv[i],"--boundfree_cutoff")==0 && i+1<argc)
      {i++; boundFreeCutOff=(float)strtod(argv[i],NULL);}
    else if (strcmp(argv[i],"--boundfree_cutoff_ross")==0 && i+1<argc)
      {i++; boundFreeCutOffRoss=(float)strtod(argv[i],NULL);}
    else if (strcmp(argv[i],"--mpi")==0 && i+1<argc)
      {i++; if (strcmp(argv[i],"yes")==0) mpi=true;}
    else if (strcmp(argv[i],"--dust_abs")==0 && i+1<argc)
      {i++; if (strcmp(argv[i],"yes")==0) dustAbs=true;}
    else if (strcmp(argv[i],"--epacket_method")==0 && i+1<argc)
      {
	i++; 
	if (strcmp(argv[i],"e_based")==0) ePacketMethod=epEBased;
	if (strcmp(argv[i],"l_based")==0) ePacketMethod=epLBased;
      }
    else if (strcmp(argv[i],"--use_Tdiff")==0 && i+1<argc)
      {i++; if (strcmp(argv[i],"no")==0) useTDiff=false;} // Buf fix!
    else if (strcmp(argv[i],"--ncell_outer_min")==0 && i+1<argc)
      {i++; nCellOuterMin=atoi(argv[i]);}
    else if (strcmp(argv[i],"--load_pop")==0 && i+1<argc)
      {i++; if (strcmp(argv[i],"no")==0) loadPop=false;}
    else if (strcmp(argv[i],"--load_rad")==0 && i+1<argc)
      {i++; if (strcmp(argv[i],"no")==0) loadRad=false;}
    else if (strcmp(argv[i],"--matter_derived_mpi")==0 && i+1<argc)
      {i++; if (strcmp(argv[i],"yes")==0) matterDerivedMPI=true;}
    else if (strcmp(argv[i],"--lte_group_macro")==0 && i+1<argc)
      {i++; if (strcmp(argv[i],"yes")==0) LTEGroupMacro=true;}
    else if (strcmp(argv[i],"--export_kappa_mc2")==0 && i+1<argc)
      {i++; if (strcmp(argv[i],"no")==0) exportKappaMC2=false;}
  }

  // If MPI
  if (mpi) {
    mpiEnv=new boost::mpi::environment;
    mpiCom=new boost::mpi::communicator;
    // Determine if master
    if (mpiCom->rank()!=0) master=false;
  }

  // Allow nested openmp parallelization
  omp_set_nested(1);

  // Use this for gprof
  //omp_set_nested(0);
  //omp_set_dynamic(0);
  //omp_set_num_threads(1);

  // If seed requested
  if (seed) {
    // If MPI -Seed with time+pid
    if (mpi) randomize(0,1,time(NULL)+getpid());
    // Else - Seed with time (+ thead number)
    else randomize(0,1,time(NULL));
  }
  // Else
  else {
    // If MPI -Seed with pid
    if (mpi) randomize(0,1,getpid());
    // Else - Seed with zero (+ thread number)
    else randomize(0,1,0);
  }
  
  // Configure grid
  FrequencyBin nu(c/(lambdaMax*AA),c/(lambdaMin*AA));
  LineSelector* lineSelector = new LineSelector(nu,tauLim);
  CellConfig* cellConfig = new CellConfig(false,neIteration,lineSelector,updateMethod,boundFreeTable,boundFreeTableMin,boundFreeTableMax,boundFreeTableNBin,boundFreeSTable,boundFreeSTableNBin,radEstimator,radEstimatorBB,radEstimatorExp,transTable,markovTable,markovTableExt,markovTableExtSplit,markovTableLim,markovTableThermLim,markovTableCTLim,collExciteTable,collExciteTableLim,nonThermalExciteTable,freeFreeAbs,boundFreeAbs,collProc,nonThermalProc,twoPhotonProc,chargeTransferProc,onTheSpot,onTheSpotNew,ffOnTheSpot,false,boundFreeBB,continuumBBTauLim,boundFreeBBTauLim,bbTJBB,exportEta,transformRay,limitFreeFreeCRange,JLineFallback,JLineFallbackFactor,JLineFallbackCount,JLineFallbackDBB,trackI,trackHK,boundFreeCutOff,boundFreeCutOffRoss,boost,alfaCorrect,alfaCorrectEmit,alfaSimpleEmit,alfaSPH,forceDBalance,forceDBalanceOSP,chargeTransferEmit,etaTable,exportSBF,nu.min,nu.max,nBin,nMuBin,LTEGroupMacro,exportKappaMC2);
  BStateSelector* bStateSelector= new BStateSelector();
  TransBBSelector* transSelector = new TransBBAllSelector();
  GridConfig* gridConfig = new GridConfig(cellConfig,bStateSelector,transSelector,updateMethod,matterDerivedMPI);

  // Configure population LTE solver
  LTESolver popLTESolver;
  popLTESolver.config.temperature.update=false;

  // Configure MC experiment
  PacketConfig* packetConfig = new PacketConfig(false,false,lineScatter,eScatter,false,gScatter,freeFreeAbs,boundFreeAbs,true,0,dustAbs,flourescence,collProc,nonThermalProc,twoPhotonProc,chargeTransferProc,transTable,updateMethod,boost,convertGPacket,-1,-1,true,1.0,reflectOuter,1.0,reflectInnerGP,transformKappa,bbEPacket);
  MCConfig* mCConfig = new MCConfig(packetConfig);

  // Set absorption processes
  absProcesses=setAbsProcesses(packetConfig->electronScatter,packetConfig->boundFreeAbs,packetConfig->freeFreeAbs,packetConfig->lineScatter,packetConfig->greyScatter);

  // If boost - Load config from file
  BoostConfig boostConfig;
  if (boost) boostConfig.load(boostConfigFile);

  // Create (dummy) matter-state solver
  MStateSolver *mStateSolver = new NLTESolver(configFile);

  // Create non-thermal solver
  Koz92Solver ntSolver(configFile);

  // Create EOS
  EOS* eos = new EOSBasic(configFile);

  // Create grid
  GridSS grid(mStateSolver,eos,gridConfig,mpiCom);
 
  // Load grid
  grid.load(gridDataFile);

  // Calculate maximum diffusion border index - New!
  diffBorderMax=grid.cell.size()-nCellOuterMin-1;

  // Init grid frequency list
  //grid.initNu(nu.min,nu.max,nBin);

  // Init grid frequency and angular binning - New!
  grid.initNu();
  grid.initMu();

  // Create frequency bin list
  nuList=createNuList(nu,nBin);

  // Create frequency vector
  nuVector.resize(nBin+1); nuVector[0]=nuList[0].min;
  for (i=0;i<nuList.size();i++) nuVector[i+1]=nuList[i].max;
  
  // Update atomdata tables (using the provided binning)
  grid.atomData.updateTables(nuVector);
  
  // Load grid state
  grid.loadState(statFile(gridStateDir,"grid-state",nIterMin),loadPop,loadPop,loadRad);

  // If populations not loaded - Update - New!
  if (not loadPop) updatePopState(grid);

  // If radiation state not loaded - Initiate - New!
  if (not loadRad) initRadState(grid,grid.cell.size()-1);
  
  // Handle time borders - New!
  handleTimeBorders(grid,tMin,tMax,dt);

  // Make sure flux through innermost cell border is zero - Hack!!!
  //((CellSS*) grid.cell[0])->FBack(0); ((CellSS*) grid.cell[0])->FObsBack(0);
  
  // Find diffusion border
  if (useTauDiff) findBorder(grid,tauDiff,diffBorder,Rdiff,absProcesses,mpiCom); // Changed!
  else findDiffBorder(grid,diffBorder,Rdiff,master);
  diffBorder=MIN(diffBorder,diffBorderMax); // New!

  // If radiation state not loaded - Update diffusion border luminosity - New!
  if (not loadRad) updateLDiff(grid,diffBorder);

  // Update on-the-spot region
  updateOnTheSpotRegion(grid,tauOnTheSpot,tauEdgeOnTheSpot,onTheSpot,onTheSpotNew,absProcesses,mpiCom); // Changed!

  // Set borders for MC experiment
  packetConfig->cellIndexMin=diffBorder+1;
  packetConfig->cellIndexMax=grid.cell.size()-1;

  // Get packet energy
  EPacket=getEPacket(grid,nPacketEff,tStop,includeRadio);

  if (master) cout<<"mc-spec: EPacket: "<<EPacket<<", nPacketEff: "<<nPacketEff<<"\n";
   
  // Create packet statistics object
  PacketStat packetStat(&grid,diffBorder+1,grid.cell.size()-1,true,nuList);

  // Create spectral evolution object and add to packet statistics object
  SpecEvo* specEvoObs = new SpecEvo(&grid,nuList,rfDistantObserver,grid.t(),tStop,nIterMax-nIterMin);
  packetStat.specEvoObs = specEvoObs;
  
  // If boost mode - Initiate boost factor
  if (boost) initBoostFactor(grid,boostConfig,nu,mpiCom);
  
  // Sample inital packets and add to packet list
  sampleInitialPackets(&grid,diffBorder+1,grid.cell.size()-1,EPacket,nu,packetIn,packetConfig,mpiCom,0.5);

  // Wait for all processes to reach this point - Testing!
  if (mpiCom!=NULL) mpiCom->barrier();
  
  // Evolution iteration loop
  for (n=nIterMin;n<=nIterMax;n++) {

    // Set initiate flag
    //initiate=(n==nIterMin);

    // Testing!
    if (mpiCom!=NULL) mpiCom->barrier();
    logSysMemory(n,"M0");

    // Get time-step - Removed!
    //dt=grid.tMax()-grid.tMin();

    // If not first iteration
    if (n>nIterMin) {

      // Load grid state
      grid.loadState(statFile(gridStateDir,"grid-state",n),loadPop,loadPop,loadRad);

      // If populations not loaded - Update - New!
      if (not loadPop) updatePopState(grid);

      // If radiation state not loaded - Initiate in diffusion region - New!
      if (not loadRad) initRadState(grid,diffBorder);

      // Handle time borders - New!
      handleTimeBorders(grid,tMin,tMax,dt);

      // Testing!
      if (mpiCom!=NULL) mpiCom->barrier();
      logSysMemory(n,"M1");

      // Make sure flux through innermost cell border is zero - Hack!!!
      //((CellSS*) grid.cell[0])->FBack(0); ((CellSS*) grid.cell[0])->FObsBack(0);
      
      // Find diffusion border
      if (useTauDiff) findBorder(grid,tauDiff,diffBorderNew,Rdiff,absProcesses,mpiCom); // Changed!
      else findDiffBorder(grid,diffBorderNew,Rdiff,master);

      // If radiation state not loaded - Update diffusion border luminosity - New!
      if (not loadRad) updateLDiff(grid,diffBorderNew);
    
      if (master) cout<<"mc-spec: Evolve radiation field, iteration: "<<n<<", t: "<<grid.t()/day<<" days, dt: "<<dt/day<<" days, tBorder: ["<<grid.tMin()/day<<","<<grid.tMax()/day<<"] days, diffusion border: "<<diffBorder<<"\n";
           
      // Update on-the-spot region
      updateOnTheSpotRegion(grid,tauOnTheSpot,tauEdgeOnTheSpot,onTheSpot,onTheSpotNew,absProcesses,mpiCom); // Changed!
      
      // Adjust packet energy
      if (ePacketMethod==epLBased) EPacket=getEPacketL(grid,nPacketLdt,dt);
      
      if (master) cout<<"mc-spec: EPacket: "<<EPacket<<", nPacketLdt: "<<nPacketLdt<<"\n";

      // If new diffusion border within previous
      if (diffBorderNew<diffBorder) {
	
	if (master) cout<<"mc-spec: Move diffusion border from cell: "<<diffBorder<<" to cell "<<diffBorderNew<<"\n";
	
	// Set boost factors to those of the previously innermost cell
	for (i=diffBorderNew+1;i<diffBorder+1;i++) grid.cell[i]->boost=grid.cell[diffBorder+1]->boost;
	
	// Sample thermal packets in new region and add to packet list
	//sampleThermalPackets(&grid,diffBorderNew+1,diffBorder,EPacket,nu,packetIn,packetConfig,mpiCom,0.5,true,true);
	sampleThermalPackets(&grid,diffBorderNew+1,diffBorder,EPacket,nu,packetIn,packetConfig,mpiCom,0.5,not useTDiff,not useTDiff,useTDiff);

	// Testing!
	if (mpiCom!=NULL) mpiCom->barrier();
	logSysMemory(n,"M2");
	
	// Store old diffusion border
	diffBorderOld=diffBorder;
	
	// Set new diffusion border
	diffBorder=diffBorderNew;
      
	// Adjust inner border for MC experiment
	packetConfig->cellIndexMin=diffBorder+1;
      }
    }

    // Update non-thermal state
    updateNTState(grid,ntSolver,diffBorder+1,grid.cell.size()-1,mpiCom);

    // Testing!
    if (mpiCom!=NULL) mpiCom->barrier();
    logSysMemory(n,"M3");

    // Update derived matter quantities (in outer region)
    grid.updateMatterDerived(diffBorder+1,grid.cell.size()-1);

    // Testing!
    if (mpiCom!=NULL) mpiCom->barrier();
    logSysMemory(n,"M4");

    // Evolve radiation state in outer region
    evolveRadiationState(&grid,diffBorder,diffBorderOld,EPacket,nu,nuList,includeRadio,packetIn,packetOut,&packetStat,mCConfig,mpiCom);

    // If boost mode - Adjust boost factor
    if (boost) adjustBoostFactor(grid,diffBorder,boostConfig,subBoost,mpiCom);

    // Testing!
    if (mpiCom!=NULL) mpiCom->barrier();
    logSysMemory(n,"M5");

    // Testing!
    logMemory(grid,packetIn,packetOut,packetStat,mpiCom);
    
    // Export MC statistics
    if (master) exportStat(grid,packetStat,gridStatDir,mCStatDir,n);

    // Update packet (state) list
    packetIn=packetOut;
  }

  // Clean up
  delete mCConfig;
  delete packetConfig;
  delete eos;
  delete gridConfig;
  delete bStateSelector;
  delete transSelector;
  delete cellConfig;
  delete lineSelector;
  if (mpiCom!=NULL) delete mpiCom;
  if (mpiEnv!=NULL) delete mpiEnv;
}
