#include <vector>
#include <math.h>
#include <string.h>
#include <unistd.h>
#include <omp.h>
#include "mc-util.h"
#include "mc-math.h"
#include "mc-physics.h"
#include "mc.h"
#include "msolver.h"
#include "eos.h"
#include "diffsolver.h"

using namespace util;
using namespace math;
using namespace physics;
using namespace mc;
//using namespace grid;
using namespace msolver;
using namespace eos;
using namespace diffsolver;

class TransBBAllSelector : public TransBBSelector {
  virtual bool use(TransBBData* trans) { return true; }
};

class TransBBNoneSelector : public TransBBSelector {
  virtual bool use(TransBBData* trans) { return false; }
};

// Create statistics file name
string statFile(const string& statDir, const string& prefix, const int& index) {
  stringstream statFile;
  statFile.str("");
  statFile<<statDir<<"/"<<prefix<<"-"<<index<<".dat";
  return statFile.str();
}

// Find optical depth border
void findBorder(const Grid& grid, const double& tauLim, int& border, double& R, bool log=false) {
  CellSS* cell;
  double tau;
  vector<double> tauCell(grid.cell.size());
  int i;

  if (log) cout<<"mc-diff-evo::findBorder: Calculate Rosseland mean optical depth within each cell\n";
  
  // Calculate Rosseland mean optical depth within each cell
  #pragma omp parallel for private(cell)
  for (i=0;i<grid.cell.size();i++) { 
    cell = (CellSS*) grid.cell[i];
    tauCell[i]=cell->rossContKappa(Processes(pEScatter|pBoundFree|pFreeFree))*cell->rho()*(cell->rMax()-cell->rMin());
  }

  // Find border
  // Unclear if muliple initializers work - Fix!
  tau=0; // New!
  for (i=grid.cell.size()-1;i>=0;i--) { // Changed
    cell = (CellSS*) grid.cell[i];
    border=i; R=cell->rMax();
    tau+=tauCell[i];
    if (log) cout<<"mc-diff-evo::findBorder: cell: "<<i<<", tau: "<<tau<<", tauLim: "<<tauLim<<"\n";
    if (tau>tauLim) break;
  }
  // If no border
  if (tau<=tauLim) { border=-1; R=((CellSS*) grid.cell[0])->rMin(); }
}

// Update cell on-the-spot approximation
void updateOnTheSpotCell(Cell* cell, bool set, const OnTheSpot& onTheSpot, const OnTheSpotNew& onTheSpotNew, bool log=false) {
  FrequencyBin onTheSpotNu;
  TransBF* transBF;
  int i,j,k,l;

  // If container - Update on-the-spot approximation for each subcell and return
  if (cell->type()==cellContainer) { for (i=0;i<cell->subCell.size();i++) updateOnTheSpotCell(cell->subCell[i],set,onTheSpot,onTheSpotNew,log); return; }

  // Set on-the-splot flag in the O-zones only - Testing!!!
  if (cell->index()==102 or cell->index()==103 or cell->index()==104) set=true;
  else set=false;

  // Reset on-the-spot (new) range
  onTheSpotNu.min=nuInf; onTheSpotNu.max=nuInf;

  // For each atom
  for (i=0;i<cell->atom.size();i++) {

    // Update on-the-spot flag
    if (set and onTheSpot(cell->atom[i])) cell->atom[i].onTheSpot(true);
    else cell->atom[i].onTheSpot(false);

    // For each ionization state
    for (j=0;j<cell->atom[i].iState.size();j++) {
      // For each bound state
      for (k=0;k<cell->atom[i].iState[j].bState.size();k++) {
	// For each (upward) bound-free transition
	for (l=0;l<cell->atom[i].iState[j].bState[k].upTransBF.size();l++) {
	  // Set transition pointer
	  transBF=cell->atom[i].iState[j].bState[k].upTransBF[l];
	  // If outside border and on-the-spot (new) requested
	  if (set and onTheSpotNew(transBF)) {
	    // Set on-the-spot (new) flag to true
	    transBF->onTheSpotNew(true);
	    // If necessary - Adjust on-the-spot (new) range
	    if (transBF->nu()<onTheSpotNu.min) onTheSpotNu.min=transBF->nu();
	  }
	  // Else - Set on-the-spot (new) flag to false
	  else transBF->onTheSpotNew(false);
	}
      }
    }
    // Update on-the-spot (new) range
    cout<<"mc-diff-evo::updateOnTheSpotRegion: cell: "<<cell->index()<<", onTheSpotNu: ("<<onTheSpotNu.min<<","<<onTheSpotNu.max<<")\n";
    cell->onTheSpotNu(onTheSpotNu);
  }
}

// Update on-the-spot approximation region
void updateOnTheSpotRegion(Grid& grid, const double& tauLim, const OnTheSpot& onTheSpot, const OnTheSpotNew& onTheSpotNew, bool log=false) {
  int border;
  double R;
  int i;

  // Find on-the-spot border
  findBorder(grid,tauLim,border,R,log);

  // Update on-the-spot appoximation for each cell
  for (i=0;i<grid.cell.size();i++) updateOnTheSpotCell(grid.cell[i],(i>border),onTheSpot,onTheSpotNew,log);

  // Update derived matter quantities (in on-the-spot region)- Excluding derived MC quantities  
  grid.updateMatterDerived(border+1,grid.cell.size()-1,false);
}

// Initiate cell radiation state
void initRadiationStateCell(Cell* cell) {
  int i;

  // If container - Initiate radiation state for each subcell
  if (cell->type()==cellContainer) for (i=0;i<cell->subCell.size();i++) initRadiationStateCell(cell->subCell[i]);

  // Initialize to BB radiation field at matter temperature and reset
  cell->TR(cell->T());
  cell->W(1.0);
  cell->reset();
}

// Initiate grid radiation state
void initRadiationState(const Grid& grid) {
  int i;

  // Initiate radiation state for each cell
  for (i=0;i<grid.cell.size();i++) initRadiationStateCell(grid.cell[i]);
}

// Initiate cell time borders
void initTimeBordersCell(Cell* cell) {
  double t,dt=1.0;
  int i;

  // If container - Set time borders for each subcell
  if (cell->type()==cellContainer) for (i=0;i<cell->subCell.size();i++) initTimeBordersCell(cell->subCell[i]);

  // Get time
  t=cell->t();

  // Set cell time borders
  cell->tMin(t-dt/2);
  cell->tMax(t+dt/2);
}

// Initiate grid time borders
void initTimeBorders(Grid& grid) {
  double t,dt=1.0;
  int i;

  // Get time
  t=grid.t();

  // Set grid time borders
  grid.tMin(t-dt/2);
  grid.tMax(t+dt/2);

  // Initiate time borders for each cell
  for (i=0;i<grid.cell.size();i++) initTimeBordersCell(grid.cell[i]);
}

// Create frequency bin list
FrequencyBinList createNuList(const FrequencyBin& nu, const int& nBin) {
  int i;
  vector<double> nuWork(nBin+1);
  FrequencyBinList nuList;
  
  double dlognu=(log(nu.max)-log(nu.min))/float(nBin);
  for (i=0;i<nuWork.size();i++) nuWork[i]=nu.min*exp(i*dlognu);
  nuList.resize(nBin);
  for (i=0;i<nuList.size();i++) nuList[i].min=nuWork[i];
  for (i=0;i<nuList.size();i++) nuList[i].max=nuWork[i+1];

  return nuList;
}

// Update radiation state
void updateRadiationState(Grid* grid, const int& nPacket, const FrequencyBin& nu, const FrequencyBinList& nuList, PacketStat* packetStat, MCConfig* mCConfig, boost::mpi::communicator* mpiCom=NULL) {
  PacketSampler* packetSampler=NULL;
  int i;
  bool master=false;

  if (mpiCom==NULL) master=true;
  else if (mpiCom->rank()==0) master=true;

  // Set fraction of packets to reflect at diffusion border
  mCConfig->packet->PReflectInner=1.0;
      
  // Create new packet sampler
  packetSampler = new PacketSamplerGamma(nPacket,grid,mCConfig->packet,mpiCom);
      
  // Reset packet statistics
  packetStat->reset(grid,0,grid->cell.size()-1,true,nuList);
  // Reset MC statistics
  grid->resetMCStatistics(0,grid->cell.size()-1,nuList);
        
  // Create new MC experiment
  MCExperiment mCExperiment(packetSampler,packetStat,mCConfig,mpiCom);
      
  // Evolve MC experiment
  mCExperiment.run();
      
  // Update radiation state
  grid->updateRadiationState(0,grid->cell.size()-1,mCExperiment);

  // Cleanup
  if (packetSampler!=NULL) delete packetSampler;
}

// Export statistics
void exportStat(GridSS& grid, PacketStat& packetStat, string& gridStatDir, string& mCStatDir, int& nExport) {

  // Export grid statistics
  if (gridStatDir.compare("")!=0) grid.exportStat(statFile(gridStatDir,"grid-stat",nExport));
  // Export packet statistics
  if (mCStatDir.compare("")!=0) packetStat.exportStat(statFile(mCStatDir,"mc-stat",nExport));
  // Export spectrum
  if (mCStatDir.compare("")!=0) packetStat.exportSpectrum(statFile(mCStatDir,"spectrum",nExport));
  // Export spectrum per transition type
  if (mCStatDir.compare("")!=0) packetStat.exportSpectrumTrans(statFile(mCStatDir,"spectrum-trans",nExport));
  // Export spectrum per cell
  if (mCStatDir.compare("")!=0) packetStat.exportSpectrumCell(statFile(mCStatDir,"spectrum-cell",nExport));

  // Increas export counter
  nExport++;
}

// Update cell matter state assuming LTE
void updateMatterStateLTECell(Cell* cell, string configFile, bool log=false) {
  LTESolver lteSolver;
  int i;

  if (log) cout<<"mc-steady-state::updateMatterStateLTECell: cell: "<<cell->index()<<", type: "<<cell->type()<<"\n";

  // If container - Update matter state assuming LTE for each subcell and return
  if (cell->type()==cellContainer) { for (i=0;i<cell->subCell.size();i++) updateMatterStateLTECell(cell->subCell[i],configFile,log); return; }

  // Update partition function
  lteSolver.updateU(cell);
  // Update ionization state
  lteSolver.updateIonState(cell);
  // Update bound state
  lteSolver.updateBoundState(cell);
  
  // Update edge list (needed by rossContKappa)
  cell->updateEdgeList();
  
  // Reset derived quantities
  cell->reset();
}

// Update grid matter state assuming LTE
void updateMatterStateLTE(Grid& grid, string configFile, bool log=false) {
  int i;

  // Update matter state assuming LTE for each cell
  for (i=0;i<grid.cell.size();i++) updateMatterStateLTECell(grid.cell[i],configFile,log);
}

// Init cell non-thermal state
void initNTStateCell(Cell* cell, string configFile) {
  int i;

  // If container - Init non-thermal state for each subcell and return
  if (cell->type()==cellContainer) { for (i=0;i<cell->subCell.size();i++) initNTStateCell(cell->subCell[i],configFile); return; }
    
  // Initiate non-thermal state
  cell->nonThermalHeating(1e-99);
  cell->nonThermalIonization(0.0);
  cell->nonThermalExcitation(0.0);
}

// Init grid non-thermal state
void initNTState(Grid& grid, string configFile) {
  int i;

  // Init non-thermal state for each cell
  for (i=0;i<grid.cell.size();i++) initNTStateCell(grid.cell[i],configFile);
}

void logAlfa(Grid& grid) {
  int i,j,k,l;
  double alfaSP;
  for (i=0;i<grid.cell.size();i++) {
    grid.cell[i]->reset();
    for (j=0;j<grid.cell[i]->atom.size();j++) {
      for (k=0;k<grid.cell[i]->atom[j].iState.size();k++) {
	cout<<"mc-steady-state::logAlfa: cell: "<<grid.cell[i]->index()<<", "<<grid.cell[i]->atom[j].iState[k].name()<<", alfaSPOriginal: "<<grid.cell[i]->atom[j].iState[k].alfaSPOriginal()<<", alfaSPTotal: "<<grid.cell[i]->atom[j].iState[k].alfaSPTotal()<<", alfaSP: "<<grid.cell[i]->atom[j].iState[k].alfaSP()<<", alfaAdd: "<<grid.cell[i]->atom[j].iState[k].alfaAdd()<<"\n";
	if (grid.cell[i]->atom[j].iState[k].name().compare("HeI")==0 or grid.cell[i]->atom[j].iState[k].name().compare("OI")==0 or grid.cell[i]->atom[j].iState[k].name().compare("CI")==0 or grid.cell[i]->atom[j].iState[k].name().compare("CII")==0)
	  for (l=0;l<grid.cell[i]->atom[j].iState[k].bState.size();l++)  cout<<"mc-steady-state::logAlfa: cell: "<<grid.cell[i]->index()<<", "<<grid.cell[i]->atom[j].iState[k].name()<<":"<<grid.cell[i]->atom[j].iState[k].bState[l].getIndex()<<", alfaSPOriginal: "<<grid.cell[i]->atom[j].iState[k].bState[l].alfaSPOriginal()<<", alfaSPTotal: "<<grid.cell[i]->atom[j].iState[k].bState[l].alfaSPTotal()<<", alfaSP: "<<grid.cell[i]->atom[j].iState[k].bState[l].alfaSP()<<", alfaAdd: "<<grid.cell[i]->atom[j].iState[k].bState[l].alfaAdd()<<"\n";
      }
    }
  }
}

void logCT(Grid& grid) {
  int i,j,ii;
  for (i=0;i<grid.cell.size();i++) {
    for (ii=0;ii<grid.cell[i]->ctReaction.size();ii++) {
      cout<<"mc-steady-state::logCT: cell: "<<grid.cell[i]->index()<<", "<<grid.cell[i]->ctReaction[ii].upTrans->lowerState->iState()->name()<<":"<<grid.cell[i]->ctReaction[ii].upTrans->lowerState->getIndex()<<"+"<<grid.cell[i]->ctReaction[ii].downTrans->upperState->iState()->name()<<":"<<grid.cell[i]->ctReaction[ii].downTrans->upperState->getIndex()<<"->"<<grid.cell[i]->ctReaction[ii].upTrans->upperState->iState()->name()<<":"<<grid.cell[i]->ctReaction[ii].upTrans->upperState->getIndex()<<"+"<<grid.cell[i]->ctReaction[ii].downTrans->lowerState->iState()->name()<<":"<<grid.cell[i]->ctReaction[ii].downTrans->lowerState->getIndex()<<", forwardRate: "<<grid.cell[i]->ctReaction[ii].forwardRate()<<", backwardRate: "<<grid.cell[i]->ctReaction[ii].backwardRate()<<"\n";
    }
    for (j=0;j<grid.cell[i]->subCell.size();j++) {
      for (ii=0;ii<grid.cell[i]->subCell[j]->ctReaction.size();ii++) {
	cout<<"mc-steady-state::logCT: cell: "<<grid.cell[i]->subCell[j]->index()<<", "<<grid.cell[i]->subCell[j]->ctReaction[ii].upTrans->lowerState->iState()->name()<<":"<<grid.cell[i]->subCell[j]->ctReaction[ii].upTrans->lowerState->getIndex()<<"+"<<grid.cell[i]->subCell[j]->ctReaction[ii].downTrans->upperState->iState()->name()<<":"<<grid.cell[i]->subCell[j]->ctReaction[ii].downTrans->upperState->getIndex()<<"->"<<grid.cell[i]->subCell[j]->ctReaction[ii].upTrans->upperState->iState()->name()<<":"<<grid.cell[i]->subCell[j]->ctReaction[ii].upTrans->upperState->getIndex()<<"+"<<grid.cell[i]->subCell[j]->ctReaction[ii].downTrans->lowerState->iState()->name()<<":"<<grid.cell[i]->subCell[j]->ctReaction[ii].downTrans->lowerState->getIndex()<<", forwardRate: "<<grid.cell[i]->subCell[j]->ctReaction[ii].forwardRate()<<", backwardRate: "<<grid.cell[i]->subCell[j]->ctReaction[ii].backwardRate()<<"\n";
      }
    }
  }
}

// Main program
int main (int argc, char** argv) {

  bool initiate;
  double EPacket;
  int n,i;
  FrequencyBinList nuList;
  vector<double> nuVector;
  boost::mpi::environment* mpiEnv=NULL;
  boost::mpi::communicator* mpiCom=NULL;
  bool master=true;
  // Hack
  OnTheSpotNew onTheSpotNewInit;

  /* Check if enough number of arguments provided */
  if (argc<2) {
    printf("Missing argument\n");
    exit(-1);
  }
  
  /* Get mandatory arguments */
  string gridDataFile=string(argv[1]);

  /* Set configuration parameters not read */
  bool steadyState=true;
  bool radEstimatorExp=false;
  bool transformKappa=false;
  bool transformRay=false;
  bool boost=false;
  bool trackI=false;
  bool trackHK=false;
  bool reflectOuter=false;
  double continuumBBTauLim=1e99;

  /* Set default values for optional arguments */
  int nIter=100;
  int nPacket=10000000;
  //int nBin=10000;
  int nBin=5000; // Changed!
  int nMuBin=20; // New!
  double tauLim=1e-3;
  double lambdaMin=1;
  double lambdaMax=25000;
  bool seed=false;
  bool eScatter=true;
  bool lineScatter=true;
  bool gScatter=false;
  bool freeFreeAbs=false;
  bool boundFreeAbs=false;
  bool flourescence=false;
  bool collProc=false;
  bool nonThermalProc=false;
  bool twoPhotonProc=false;
  bool chargeTransferProc=false;
  bool chargeTransferEmit=false;
  bool transTable=false;
  bool markovTable=false;
  bool markovTableExt=false;
  bool markovTableExtSplit=false;
  double markovTableLim=0.0;
  double markovTableThermLim=0.0;
  double markovTableCTLim=0.0;
  bool collExciteTable=false;
  double collExciteTableLim=0.0;  
  bool nonThermalExciteTable=false;
  bool etaTable=false;
  UpdateMethod updateMethod=urOnTheFly;
  bool boundFreeTable=false;
  double boundFreeTableMin=1;
  double boundFreeTableMax=1e5;
  int boundFreeTableNBin=1e4;
  bool boundFreeSTable=false;
  int boundFreeSTableNBin=1e3;
  RadEstimator radEstimator=rfDBB;
  RadEstimator radEstimatorBB=rfDBB;
  OnTheSpot onTheSpot("");
  OnTheSpotNew onTheSpotNew("");
  double tauOnTheSpot=10;
  FFOnTheSpot ffOnTheSpot(false,10.0);
  bool alfaCorrect=false;
  bool alfaCorrectEmit=false;
  bool alfaSimpleEmit=false;
  bool alfaSPH=false;
  bool forceDBalance=false;
  bool forceDBalanceOSP=false;
  bool boundFreeBB=false;
  double boundFreeBBTauLim=1e3;
  bool convertGPacket=true;
  bool bbEPacket=false;
  string configFile="";
  SolverMethod solverMethod=sNebular;
  bool exportEta=false;
  bool exportSBF=false;
  bool reflectInnerGP=false;
  bool limitFreeFreeCRange=false;
  string gridTemplateFile="";
  string gridStatDir="";
  string mCStatDir="";
  string gridStateDir="";
  int saveStateStep=1;
  string gridStateFile="";
  bool JLineFallback=false;
  double JLineFallbackFactor=1.1;
  int JLineFallbackCount=3;
  bool JLineFallbackDBB=false;
  double boundFreeCutOff=0.0;
  double boundFreeCutOffRoss=0.0;
  bool mpi=false;
  bool loadDBB=false;
  bool dustAbs=false;
  bool matterDerivedMPI=false;
  bool LTEGroupMacro=false;
  bool exportKappaMC2=true;

  /* Get optional arguments */
  for (i=2;i<argc;i++) {
    if (strcmp(argv[i],"--niter")==0 && i+1<argc) 
      {i++; nIter=atoi(argv[i]);}
    else if (strcmp(argv[i],"--npacket")==0 && i+1<argc)
      {i++; nPacket=atoi(argv[i]);}
    else if (strcmp(argv[i],"--nbin")==0 && i+1<argc)
      {i++; nBin=atoi(argv[i]);}
    else if (strcmp(argv[i],"--taulim")==0 && i+1<argc)
      {i++; tauLim=(float)strtod(argv[i],NULL);}
    else if (strcmp(argv[i],"--lambda_min")==0 && i+1<argc)
      {i++; lambdaMin=(float)strtod(argv[i],NULL);}
    else if (strcmp(argv[i],"--lambda_max")==0 && i+1<argc)
      {i++; lambdaMax=(float)strtod(argv[i],NULL);}
    else if (strcmp(argv[i],"--seed")==0 && i+1<argc)
      {i++; if (strcmp(argv[i],"yes")) seed=true;}
    else if (strcmp(argv[i],"--escatter")==0 && i+1<argc)
      {i++; if (strcmp(argv[i],"yes")==0) eScatter=true;}
    else if (strcmp(argv[i],"--linescatter")==0 && i+1<argc)
      {i++; if (strcmp(argv[i],"no")==0) lineScatter=false;}
     else if (strcmp(argv[i],"--gscatter")==0 && i+1<argc)
      {i++; if (strcmp(argv[i],"yes")==0) gScatter=true;}
    else if (strcmp(argv[i],"--freefree_abs")==0 && i+1<argc)
      {i++; if (strcmp(argv[i],"yes")==0) freeFreeAbs=true;}
    else if (strcmp(argv[i],"--boundfree_abs")==0 && i+1<argc)
      {i++; if (strcmp(argv[i],"yes")==0) boundFreeAbs=true;}
    else if (strcmp(argv[i],"--flourescence")==0 && i+1<argc)
      {i++; if (strcmp(argv[i],"yes")==0) flourescence=true;}
    else if (strcmp(argv[i],"--coll_proc")==0 && i+1<argc)
      {i++; if (strcmp(argv[i],"yes")==0) collProc=true;}
    else if (strcmp(argv[i],"--non_thermal_proc")==0 && i+1<argc)
      {i++; if (strcmp(argv[i],"yes")==0) nonThermalProc=true;}
    else if (strcmp(argv[i],"--two_photon_proc")==0 && i+1<argc)
      {i++; if (strcmp(argv[i],"yes")==0) twoPhotonProc=true;}
    else if (strcmp(argv[i],"--charge_transfer_proc")==0 && i+1<argc)
      {i++; if (strcmp(argv[i],"yes")==0) chargeTransferProc=true;}
    else if (strcmp(argv[i],"--charge_transfer_emit")==0 && i+1<argc)
      {i++; if (strcmp(argv[i],"yes")==0) chargeTransferEmit=true;}
    else if (strcmp(argv[i],"--trans_table")==0 && i+1<argc)
      {i++; if (strcmp(argv[i],"yes")==0) transTable=true;}
    else if (strcmp(argv[i],"--markov_table")==0 && i+1<argc)
      {i++; if (strcmp(argv[i],"yes")==0) markovTable=true;}
    else if (strcmp(argv[i],"--markov_table_ext")==0 && i+1<argc)
      {i++; if (strcmp(argv[i],"yes")==0) markovTableExt=true;}
    else if (strcmp(argv[i],"--markov_table_lim")==0 && i+1<argc) 
      {i++; markovTableLim=(float)strtod(argv[i],NULL);}
    else if (strcmp(argv[i],"--markov_table_ext_split")==0 && i+1<argc)
      {i++; if (strcmp(argv[i],"yes")==0) markovTableExtSplit=true;}
    else if (strcmp(argv[i],"--markov_table_therm_lim")==0 && i+1<argc) 
      {i++; markovTableThermLim=(float)strtod(argv[i],NULL);}
    else if (strcmp(argv[i],"--markov_table_ct_lim")==0 && i+1<argc) 
      {i++; markovTableCTLim=(float)strtod(argv[i],NULL);}
    else if (strcmp(argv[i],"--coll_exc_table")==0 && i+1<argc)
      {i++; if (strcmp(argv[i],"yes")==0) collExciteTable=true;}
    else if (strcmp(argv[i],"--coll_exc_table_lim")==0 && i+1<argc) 
      {i++; collExciteTableLim=(float)strtod(argv[i],NULL);}
    else if (strcmp(argv[i],"--non_thermal_exc_table")==0 && i+1<argc)
      {i++; if (strcmp(argv[i],"yes")==0) nonThermalExciteTable=true;}
    else if (strcmp(argv[i],"--eta_table")==0 && i+1<argc)
      {i++; if (strcmp(argv[i],"yes")==0) etaTable=true;}
    else if (strcmp(argv[i],"--boundfree_table")==0 && i+1<argc)
      {i++; if (strcmp(argv[i],"yes")==0) boundFreeTable=true;}
    else if (strcmp(argv[i],"--boundfree_table_min")==0 && i+1<argc)
      {i++; boundFreeTableMin==(float)strtod(argv[i],NULL);}
    else if (strcmp(argv[i],"--boundfree_table_max")==0 && i+1<argc)
      {i++; boundFreeTableMax=(float)strtod(argv[i],NULL);}
    else if (strcmp(argv[i],"--boundfree_table_nbin")==0 && i+1<argc)
      {i++; boundFreeTableNBin=atoi(argv[i]);}
    else if (strcmp(argv[i],"--boundfree_stable")==0 && i+1<argc)
      {i++; if (strcmp(argv[i],"yes")==0) boundFreeSTable=true;}
    else if (strcmp(argv[i],"--boundfree_stable_nbin")==0 && i+1<argc)
      {i++; boundFreeSTableNBin=atoi(argv[i]);}
    else if (strcmp(argv[i],"--rad_estimator")==0 && i+1<argc)
      {i++; if (strcmp(argv[i],"lucy")==0) radEstimator=rfLucy;}
    else if (strcmp(argv[i],"--rad_estimator_bb")==0 && i+1<argc)
      {i++; if (strcmp(argv[i],"lucy")==0) radEstimatorBB=rfLucy;}
    else if (strcmp(argv[i],"--onthespot")==0 && i+1<argc)
      {i++; onTheSpot=OnTheSpot(argv[i]);}
    else if (strcmp(argv[i],"--onthespot_new")==0 && i+1<argc)
      {i++; onTheSpotNew=OnTheSpotNew(argv[i]);}
    else if (strcmp(argv[i],"--tau_onthespot")==0 && i+1<argc)
      {i++; tauOnTheSpot=(float)strtod(argv[i],NULL);}
    else if (strcmp(argv[i],"--ff_onthespot")==0 && i+1<argc)
      {i++; if (strcmp(argv[i],"yes")==0) ffOnTheSpot.apply=true;}
    else if (strcmp(argv[i],"--ff_onthespot_taulim")==0 && i+1<argc)
      {i++; ffOnTheSpot.tauLim=(float)strtod(argv[i],NULL);}
    else if (strcmp(argv[i],"--alfa_correct")==0 && i+1<argc)
      {i++; if (strcmp(argv[i],"yes")==0) alfaCorrect=true;}
    else if (strcmp(argv[i],"--alfa_correct_emit")==0 && i+1<argc)
      {i++; if (strcmp(argv[i],"yes")==0) alfaCorrectEmit=true;}
    else if (strcmp(argv[i],"--alfa_simple_emit")==0 && i+1<argc)
      {i++; if (strcmp(argv[i],"yes")==0) alfaSimpleEmit=true;}
    else if (strcmp(argv[i],"--alfa_SPH")==0 && i+1<argc)
      {i++; if (strcmp(argv[i],"yes")==0) alfaSPH=true;}
    else if (strcmp(argv[i],"--force_dbalance")==0 && i+1<argc)
      {i++; if (strcmp(argv[i],"yes")==0) forceDBalance=true;}
    else if (strcmp(argv[i],"--force_dbalance_osp")==0 && i+1<argc)
      {i++; if (strcmp(argv[i],"yes")==0) forceDBalanceOSP=true;}
    else if (strcmp(argv[i],"--convert_gpacket")==0 && i+1<argc)
      {i++; if (strcmp(argv[i],"no")==0) convertGPacket=false;}
    else if (strcmp(argv[i],"--bb_epacket")==0 && i+1<argc)
      {i++; if (strcmp(argv[i],"yes")==0) bbEPacket=true;}
    else if (strcmp(argv[i],"--boundfree_bb")==0 && i+1<argc)
      {i++; if (strcmp(argv[i],"yes")==0) boundFreeBB=true;}
    else if (strcmp(argv[i],"--boundfree_bb_taulim")==0 && i+1<argc)
      {i++; boundFreeBBTauLim=(float)strtod(argv[i],NULL);}
    else if (strcmp(argv[i],"--continuum_bb_taulim")==0 && i+1<argc)
      {i++; continuumBBTauLim=(float)strtod(argv[i],NULL);}
    else if (strcmp(argv[i],"--config_file")==0 && i+1<argc)
      {i++; configFile=string(argv[i]);}
    else if (strcmp(argv[i],"--msolver_method")==0 && i+1<argc)
      {
	i++; 
	if (strcmp(argv[i],"nlte")==0) solverMethod=sNLTE;
	if (strcmp(argv[i],"comb")==0) solverMethod=sComb;
	if (strcmp(argv[i],"lte")==0) solverMethod=sLTE;	
      }
    else if (strcmp(argv[i],"--export_eta")==0 && i+1<argc)
      {i++; if (strcmp(argv[i],"yes")==0) exportEta=true;}
    else if (strcmp(argv[i],"--export_sbf")==0 && i+1<argc)
      {i++; if (strcmp(argv[i],"yes")==0) exportSBF=true;}
    else if (strcmp(argv[i],"--reflect_inner_gp")==0 && i+1<argc)
      {i++; if (strcmp(argv[i],"yes")==0) reflectInnerGP=true;}
    else if (strcmp(argv[i],"--limit_freefree_crange")==0 && i+1<argc)
      {i++; if (strcmp(argv[i],"yes")==0) limitFreeFreeCRange=true;}
    else if (strcmp(argv[i],"--grid_template_file")==0 && i+1<argc)
      {i++; gridTemplateFile=string(argv[i]);}
    else if (strcmp(argv[i],"--grid_stat_dir")==0 && i+1<argc)
      {i++; gridStatDir=string(argv[i]);}
    else if (strcmp(argv[i],"--mc_stat_dir")==0 && i+1<argc)
      {i++; mCStatDir=string(argv[i]);}
    else if (strcmp(argv[i],"--grid_state_dir")==0 && i+1<argc)
      {i++; gridStateDir=string(argv[i]);}
    else if (strcmp(argv[i],"--save_state_step")==0 && i+1<argc) 
      {i++; saveStateStep=atoi(argv[i]);}
    else if (strcmp(argv[i],"--grid_state_file")==0 && i+1<argc)
      {i++; gridStateFile=string(argv[i]);}
    else if (strcmp(argv[i],"--jline_fallback")==0 && i+1<argc)
      {i++; if (strcmp(argv[i],"yes")==0) JLineFallback=true;}
    else if (strcmp(argv[i],"--jline_fallback_factor")==0 && i+1<argc)
      {i++; JLineFallbackFactor=(float)strtod(argv[i],NULL);}
    else if (strcmp(argv[i],"--jline_fallback_count")==0 && i+1<argc) 
      {i++; JLineFallbackCount=atoi(argv[i]);}
    else if (strcmp(argv[i],"--jline_fallback_dbb")==0 && i+1<argc)
      {i++; if (strcmp(argv[i],"yes")==0) JLineFallbackDBB=true;}
    else if (strcmp(argv[i],"--boundfree_cutoff")==0 && i+1<argc)
      {i++; boundFreeCutOff=(float)strtod(argv[i],NULL);}
    else if (strcmp(argv[i],"--boundfree_cutoff_ross")==0 && i+1<argc)
      {i++; boundFreeCutOffRoss=(float)strtod(argv[i],NULL);}
    else if (strcmp(argv[i],"--mpi")==0 && i+1<argc)
      {i++; if (strcmp(argv[i],"yes")==0) mpi=true;}
    else if (strcmp(argv[i],"--load_dbb")==0 && i+1<argc)
      {i++; if (strcmp(argv[i],"yes")==0) loadDBB=true;}
    else if (strcmp(argv[i],"--dust_abs")==0 && i+1<argc)
      {i++; if (strcmp(argv[i],"yes")==0) dustAbs=true;}
    else if (strcmp(argv[i],"--matter_derived_mpi")==0 && i+1<argc)
      {i++; if (strcmp(argv[i],"yes")==0) matterDerivedMPI=true;}
    else if (strcmp(argv[i],"--lte_group_macro")==0 && i+1<argc)
      {i++; if (strcmp(argv[i],"yes")==0) LTEGroupMacro=true;}
    else if (strcmp(argv[i],"--export_kappa_mc2")==0 && i+1<argc)
      {i++; if (strcmp(argv[i],"no")==0) exportKappaMC2=false;}
  }

  // If MPI
  if (mpi) {
    mpiEnv=new boost::mpi::environment;
    mpiCom=new boost::mpi::communicator;
    // Determine if master
    if (mpiCom->rank()!=0) master=false;
  }

  // Allow nested openmp parallelization
  omp_set_nested(1);

  // Use this for gprof
  //omp_set_nested(0);
  //omp_set_dynamic(0);
  //omp_set_num_threads(1);

  // If seed requested
  if (seed) {
    // If MPI -Seed with time+pid
    if (mpi) randomize(0,1,time(NULL)+getpid());
    // Else - Seed with time (+ thead number)
    else randomize(0,1,time(NULL));
  }
  // Else
  else {
    // If MPI -Seed with pid
    if (mpi) randomize(0,1,getpid());
    // Else - Seed with zero (+ thread number)
    else randomize(0,1,0);
  }
  
  // Configure grid
  FrequencyBin nu(c/(lambdaMax*AA),c/(lambdaMin*AA));
  LineSelector* lineSelector = new LineSelector(nu,tauLim);
  // Hack !!!
  onTheSpotNewInit=(gridStateFile.compare("")==0) ? onTheSpotNewInit=onTheSpotNew : onTheSpotNewInit=OnTheSpotNew("");
  CellConfig* cellConfig = new CellConfig(steadyState,neIteration,lineSelector,updateMethod,boundFreeTable,boundFreeTableMin,boundFreeTableMax,boundFreeTableNBin,boundFreeSTable,boundFreeSTableNBin,radEstimator,radEstimatorBB,radEstimatorExp,transTable,markovTable,markovTableExt,markovTableExtSplit,markovTableLim,markovTableThermLim,markovTableCTLim,collExciteTable,collExciteTableLim,nonThermalExciteTable,freeFreeAbs,boundFreeAbs,collProc,nonThermalProc,twoPhotonProc,chargeTransferProc,onTheSpot,onTheSpotNewInit,ffOnTheSpot,false,boundFreeBB,continuumBBTauLim,boundFreeBBTauLim,bbTJBB,exportEta,transformRay,limitFreeFreeCRange,JLineFallback,JLineFallbackFactor,JLineFallbackCount,JLineFallbackDBB,trackI,trackHK,boundFreeCutOff,boundFreeCutOffRoss,boost,alfaCorrect,alfaCorrectEmit,alfaSimpleEmit,alfaSPH,forceDBalance,forceDBalanceOSP,chargeTransferEmit,etaTable,exportSBF,nu.min,nu.max,nBin,nMuBin,LTEGroupMacro,exportKappaMC2);
  BStateSelector* bStateSelector= new BStateSelector();
  TransBBSelector* transSelector= new TransBBAllSelector();
  GridConfig* gridConfig = new GridConfig(cellConfig,bStateSelector,transSelector,updateMethod,matterDerivedMPI);

  // Configure MC experiment
  PacketConfig* packetConfig = new PacketConfig(steadyState,false,lineScatter,eScatter,false,gScatter,freeFreeAbs,boundFreeAbs,true,0,dustAbs,flourescence,collProc,nonThermalProc,twoPhotonProc,chargeTransferProc,transTable,updateMethod,boost,convertGPacket,-1,-1,true,1.0,reflectOuter,1.0,reflectInnerGP,transformKappa,bbEPacket);
  MCConfig* mCConfig = new MCConfig(packetConfig);

  // Create matter state solver
  MStateSolver* mStateSolver;
  switch (solverMethod) {
  case sNLTE : { mStateSolver = new NLTESolver(configFile); break; }
  case sComb : { mStateSolver = new CombSolver(configFile); break; }
  case sLTE : { mStateSolver = new LTESolver(configFile); break; }
  deafult : { mStateSolver = new NebSolver(configFile); } }

  // Create EOS
  EOS* eos = new EOSBasic(configFile);

  // Create grid
  GridSS grid(mStateSolver,eos,gridConfig,mpiCom);
  
  // Load grid
  grid.load(gridDataFile);

  // Init grid frequency list
  //grid.initNu(nu.min,nu.max,nBin);

  // Init grid frequency and angular binning - New!
  grid.initNu();
  grid.initMu();

  // Create frequency bin list
  nuList=createNuList(nu,nBin);

  // Create frequency vector
  nuVector.resize(nBin+1); nuVector[0]=nuList[0].min;
  for (i=0;i<nuList.size();i++) nuVector[i+1]=nuList[i].max;

  // Resample bound-free tables - Removed!
  //grid.atomData.resampleBFTables(nuVector);

  // Update atomdata tables (using the provided binning) - New!
  grid.atomData.updateTables(nuVector);
  
  if (mpiCom!=NULL) mpiCom->barrier();

  // Set time borders
  initTimeBorders(grid);

  // If requested - Load grid state
  if (gridStateFile.compare("")!=0) grid.loadState(gridStateFile);
  // Else
  else {
    // If DBB field not loaded - Initiate radiation state
    if (not loadDBB) initRadiationState(grid);

    // Update matter state assuming LTE
    updateMatterStateLTE(grid,configFile,master);
  }

  // Initiate non-thermal state
  initNTState(grid,configFile);
  
  // Set borders for MC experiment
  packetConfig->cellIndexMin=0;
  packetConfig->cellIndexMax=grid.cell.size()-1;
  
  // Create packet statistics object
  PacketStat packetStat(&grid,0,grid.cell.size()-1,true,nuList);

  // Initialize export counter
  int nExport=0;

  if (master) logAlfa(grid);
  if (master) logCT(grid);

  // Main iteration loop
  for (n=0;n<nIter;n++) {

    initiate=(n==0);

    if (master) cout<<"mc-steady-state: Main loop, iteration: "<<n<<"\n";
 
    // Update on-the-spot region
    updateOnTheSpotRegion(grid,tauOnTheSpot,onTheSpot,onTheSpotNew,master);
    
    // Update derived matter quantities (in outer region)
    grid.updateMatterDerived(0,grid.cell.size()-1);
    
    if (mpiCom!=NULL) mpiCom->barrier();

    // Track intensity only if state is going to be saved
    if (gridStateDir.compare("")!=0 and (n%saveStateStep)==0) cellConfig->trackI=true; else cellConfig->trackI=false;
    
    // Update radiation state
    updateRadiationState(&grid,nPacket,nu,nuList,&packetStat,mCConfig,mpiCom);
    
    if (mpiCom!=NULL) mpiCom->barrier();
       
    // Update matter state - Excluding derived MC quantities
    grid.updateMatterState(0,grid.cell.size()-1,initiate,true,false);

    if (master) logAlfa(grid);
    if (master) logCT(grid);
    
    // Export statistics
    if (master) exportStat(grid,packetStat,gridStatDir,mCStatDir,nExport);

    // Save grid state
    if (gridStateDir.compare("")!=0 and (n%saveStateStep)==0 and master) grid.saveState(statFile(gridStateDir,"grid-state",n));
  }

  // Clean up
  delete mCConfig;
  delete packetConfig;
  delete eos;
  delete mStateSolver;
  delete gridConfig;
  delete bStateSelector;
  delete transSelector;
  delete cellConfig;
  delete lineSelector;
  if (mpiCom!=NULL) delete mpiCom;
  if (mpiEnv!=NULL) delete mpiEnv;
}
