#ifndef _MC_
#define _MC_

#include <string>
#include <vector>
#include <boost/mpi.hpp>
#include <boost/serialization/string.hpp>
#include <omp.h>
#include "yaml-cpp/yaml.h"
#include "mc-util.h"
#include "mc-math.h"
#include "mc-physics.h"
#include "common.h"
#include "atomdata.h"

using namespace std;
using namespace util;
using namespace math;
using namespace physics;
using namespace common;
using namespace atomdata;

#include "grid.h"

//using namespace grid;

namespace mc {
  
  class Packet;

  // Packet event statistics
  class PacketEventStat {
  private :
    // MPI serialization
    friend class boost::serialization::access;
    template<class Archive> 
      void serialize (Archive & ar, const unsigned int version) 
      {
	ar & nEScatter;
	ar & nLScatter;
	ar & nGScatter; // New!
	ar & nLAbs;
	ar & nLEmission;
	ar & nFFAbs;
	ar & nFFEmission;
	ar & nBFAbs;
	ar & nBFEmission;
	ar & n2PEmission; // New!
	ar & nCCrossing;
	ar & nCICrossing; // New!
	ar & nCOCrossing; // New!
	ar & nTCrossing; // New!
	ar & nReflect;
	// Testing
	ar & nEmission;
	ar & ECEmission;
	ar & nNone;
	ar & ECNone;
	ar & nBCrossing;
	ar & nComptonAbs;
	ar & nRadioEmission;
	ar & n56NiGamma;
	ar & n56CoGamma;
	ar & n56CoPos;
	ar & n57NiGamma;
	ar & n57CoGamma;
	ar & n57CoPos;
	ar & n44TiGamma;
	ar & n44TiPos;
	ar & nDummyEmission; // New!
      }
  public :
    uint64_t nEScatter;
    uint64_t nLScatter;
    uint64_t nGScatter;
    uint64_t nLAbs; 
    uint64_t nLEmission; 
    uint64_t nFFAbs; 
    uint64_t nFFEmission; 
    uint64_t nBFAbs; 
    uint64_t nBFEmission;
    uint64_t n2PEmission; // New!
    uint64_t nCCrossing;
    uint64_t nCICrossing; // New!
    uint64_t nCOCrossing; // New!
    uint64_t nTCrossing; // New!
    uint64_t nReflect;
    // Testing
    uint64_t nEmission;
    double ECEmission;
    uint64_t nNone;
    double ECNone;
    uint64_t nBCrossing;
    uint64_t nComptonAbs;
    uint64_t nRadioEmission;
    uint64_t n56NiGamma;
    uint64_t n56CoGamma;
    uint64_t n56CoPos;
    uint64_t n57NiGamma;
    uint64_t n57CoGamma;
    uint64_t n57CoPos;
    uint64_t n44TiGamma;
    uint64_t n44TiPos;
    uint64_t nDummyEmission; // New!

    // Constructor
    PacketEventStat();
    // Update packet event statistics
    void update(const PacketEvent& event);
    // Export packet event statistics
    void exportStat(YAML::Emitter& data);
    // Sum statistics
    void operator+=(const PacketEventStat& eventStat);
    // Reset packet event statistics
    void reset();
  };

  // Reference frame - New!
  enum RefFrame { rfRest, rfObserver, rfDistantObserver };

  // Spectrum - New!
  class Spectrum {
    private :
    // MPI serialization
    friend class boost::serialization::access;
    template<class Archive> 
      void serialize (Archive & ar, const unsigned int version) 
      {
	ar & pL;
	ar & Llambda;
	ar & LlambdaFreeFree;
	ar & LlambdaBoundFree;
	ar & LlambdaBoundBound;
	ar & LlambdaCell;
	ar & LlambdaSubCell;
	}
    Grid* grid;
    int cellIndexMin;
    int cellIndexMax;
    FrequencyBinList nu;
    RefFrame refFrame;
    double pL;
    vector<double> Llambda;
    vector<double> LlambdaFreeFree;
    vector<vector<vector<double> > > LlambdaBoundFree;
    vector<vector<vector<double> > > LlambdaBoundBound;
    vector<vector<double> > LlambdaCell;
    vector<vector<vector<double> > > LlambdaSubCell;
    // Update observed luminosity
    void updateL(const Packet& packet);
    // Update observed spectrum
    void updateSpecTotal(const Packet& packet);
    // Update observed spectrum per transition type
    void updateSpecTrans(const Packet& packet);
    // Update observed spectrum per cell
    void updateSpecCell(const Packet& packet);
  public :
    double tMin,tMax;
    // Constructors
    Spectrum() {};
    // Constructors
    Spectrum(Grid* _grid, const FrequencyBinList& _nu, const RefFrame& refFrame, const double& _tMin, const double& _tMax);
    Spectrum(Grid* _grid, const int& _cellIndexMin, const int& _cellIndexMax, const FrequencyBinList& _nu, const RefFrame& _refFrame, const double& _tMin, const double& _tMax);
    // Destructor
    virtual ~Spectrum() {};
    // Sum spectrum
    void operator+=(const Spectrum& spectrum);
    // Update spectrum
    virtual void update(const Packet& packet);
    // Reset spectrum
    virtual void reset();
    // Reset spectrum
    void reset(Grid* _grid, const int& _cellIndexMin, const int& _cellIndexMax, const FrequencyBinList& _nu, const RefFrame& _refFrame, const double& _tMin, const double& _tMax);
    // Synchronize (MPI)
    void sync(boost::mpi::communicator* mpiCom);
    // Get luminosity
    double L();
    // Export spectrum
    void exportSpecTotal(const string& dataFile);
    // Export spectrum per transition type
    void exportSpecTrans(const string& dataFile);
    // Export spectrum per cell
    void exportSpecCell(const string& dataFile);
    // Get memory allocated - New!
    double mem() const;
  };

  // Spectral evolution - New!
  class SpecEvo {
    vector<bool> exported;
    vector<bool> synchronized;
    vector<Spectrum> spectrum;
    Grid* grid;
    int cellIndexMin;
    int cellIndexMax;
    FrequencyBinList nu;
    RefFrame refFrame;
    bool pLock;
  public :
    // Constructors
    SpecEvo(Grid* _grid, const FrequencyBinList& _nu, const RefFrame& _refFrame, const double& tBegin, const double tEnd, const double nBin);
    SpecEvo(Grid* _grid, const int& _cellIndexMin, const int& _cellIndexMax, const FrequencyBinList& _nu, const RefFrame& _refFrame, const double& tBegin, const double tEnd, const double nBin);
    // Reset
    void reset();
    // Get lock-flag
    bool lock() const;
    // Set lock-flag
    void lock(const bool& _lock);
    // Synchronize (MPI)
    void sync(boost::mpi::communicator* mpiCom);
    // Update
    void update(const Packet& packet);
    // Export
    void exportUpdated(const string& specDir, const string& prefix);
    // Get memory allocated - New!
    double mem() const;
  };

  // Packet statistics
  class PacketStat {
  private :
    // MPI serialization
    friend class boost::serialization::access;
    template<class Archive> 
      void serialize (Archive & ar, const unsigned int version) 
      {
	ar & cellIndexMin;
	ar & cellIndexMax;
	ar & pL;
	ar & pLC;
	ar & pNPacket; // New!
	ar & pNPacketInject; // New!
	ar & pEBoost; // New!
	ar & pEBoostInject; // New!
	ar & Llambda;
	ar & LlambdaFreeFree; // New!
	ar & LlambdaBoundFree; // New!
	ar & LlambdaBoundBound; // New!
	ar & LlambdaCell; // New!
	ar & LlambdaSubCell; // New!
	ar & eventStat;
	ar & createSpectrum;
      }
    Grid* grid;
    int cellIndexMin;
    int cellIndexMax;
    double pL;
    double pLC;
    int pNPacket; // New!
    int pNPacketInject; // New!
    double pEBoost; // New!
    double pEBoostInject; // New!
    FrequencyBinList nu;
    vector<double> Llambda;
    vector<double> LlambdaFreeFree; // New!
    vector<vector<vector<double> > > LlambdaBoundFree; // New!
    vector<vector<vector<double> > > LlambdaBoundBound; // New!
    vector<vector<double> > LlambdaCell; // New!
    vector<vector<vector<double> > > LlambdaSubCell; // New!
    vector<PacketEventStat> eventStat;
    bool createSpectrum;
    // Update observed luminosity
    void updateL(const Packet& packet);
    // Update number of packets - New!
    void updateNPacket(Packet& packet);
    // Update injected number of packets - New!
    void updateNPacketInject(Packet& packet);
    // Update energy boost - New!
    void updateEBoost(Packet& packet);
    // Update injected energy boost - New!
    void updateEBoostInject(Packet& packet);
    // Update observed spectrum
    void updateSpectrum(const Packet& packet);
    // Update observed spectrum per transition type - New!
    void updateSpectrumTrans(const Packet& packet);
    // Update observed spectrum per cell - New!
    void updateSpectrumCell(const Packet& packet);
  public :
    SpecEvo* specEvoObs; // New!
    // Constructors
    PacketStat() {};
    // Constructors
    PacketStat(Grid* _grid, const bool _createSpectrum=false, const FrequencyBinList _nu=FrequencyBinList());
    PacketStat(Grid* _grid, const int& _cellIndexMin, const int& _cellIndexMax, const bool _createSpectrum=false, const FrequencyBinList _nu=FrequencyBinList());
    // Destructor
    virtual ~PacketStat() {};
    // Sum statistics
    void operator+=(const PacketStat& packetStat);
    // Update packet statistics
    //virtual void update(const Packet& packet);
    virtual void update(Packet& packet); // Changed!
    // Update injected packet statistics
    virtual void updateInject(Packet& packet); // New!
    // Reset packet statistics
    virtual void reset();
    // Reset packet statistics
    void reset(Grid* _grid, const int& _cellIndexMin, const int& _cellIndexMax, const bool _createSpectrum=false, const FrequencyBinList _nu=FrequencyBinList());
    // Update packet event statistics
    virtual void updateEvent(const PacketEvent& packetEvent);
    // Get observed luminosity
    double L();
    // Get CMF luminosity
    double LC();
    // Get number of active packets - New!
    int nPacket();
    // Get injected number of packets - New!
    int nPacketInject();
    // Get energy boost - New!
    double EBoost();
    // Get injected energy boost - New!
    double EBoostInject();
    // Export spectrum
    void exportSpectrum(const string& dataFile);
    // Export spectrum per transition type - New!
    void exportSpectrumTrans(const string& dataFile);
    // Export spectrum per cell - New!
    void exportSpectrumCell(const string& dataFile);
    // Export packet statistics
    void exportStat(const string& statFile);
    // Get memory allocated - New!
    double mem() const;
  };

  // Packet event type
  enum PacketEventType { evNone, evEmission, evLineScatter, evLineAbs, evLineEmission, evElectronScatter, evComptonScatter, evGreyScatter, evFreeFreeAbs, evFreeFreeEmission, evBoundFreeAbs, evBoundFreeEmission, evComptonAbs, evDustAbs, evTwoPhotonEmission, evCellCrossing, ev56NiGammaDecay, ev56CoGammaDecay, ev56CoPosDecay, ev57NiGammaDecay, ev57CoGammaDecay, ev57CoPosDecay, ev44TiGammaDecay, ev44TiPosDecay, evReflect, evBoundaryCrossing, evCellCrossingInner, evCellCrossingOuter, evSubCellCrossing, evTimeBorderCrossing, evDummyEmission }; // Changed!

  // Testing!!!
  void logSysCPU(int i, string label);

  // Check if continuum event
  bool isContinuumEvent(const PacketEventType& eventType);
  // Check if line event
  bool isLineEvent(const PacketEventType& eventType);
  // Check if (true) emission event
  bool isEmissionEvent(const PacketEventType& eventType);
  // Check if (true) absorption event
  bool isTrueAbsEvent(const PacketEventType& eventType);
  // Check if absorption event
  bool isAbsEvent(const PacketEventType& eventType);
  // Check if scatter event
  bool isScatterEvent(const PacketEventType& eventType);
  // Check if radioactive emission event
  bool isRadioEmissionEvent(const PacketEventType& eventType);
  // Log packet event type
  string logPacketEventType(const PacketEventType& eventType);

  // Packet event
  class PacketEvent {
  public :
    PacketEventType type;
    Vector3D position;
    // Are these ok as floats ??? - Yes, maybe, changed back to double.
    double t;
    double E;
    double nu;
    double ECMF;
    double nuCMF;
    int cellIndex;
    int subCellIndex;
    int packetIndex;
    // Constructors
    PacketEvent();
    PacketEvent(const PacketEventType& _type, const Vector3D& _position, const double& t, const double& _E, const double& _nu, const double& _ECMF, const double& _nuCMF, int _cellIndex, int _subCellIndex, int _packetIndex);
    // Destructor
    ~PacketEvent();
    // Validate data
    bool valid() const;
  };

  // Packet type
  enum PacketType { rPacket, iPacket, kPacket, ePacket, gPacket };

  // Packet emission event
  class PacketEmissionEvent {
  private :
    // MPI serialization
    friend class boost::serialization::access;
    template<class Archive> 
      void serialize (Archive & ar, const unsigned int version) 
      {
	ar & cellIndex;
	ar & subCellIndex;
	ar & Z;
	ar & C;
      }
  public :
    PacketEventType type;
    int cellIndex;
    int subCellIndex;
    double Z;
    double C;
    // Constructors
    PacketEmissionEvent();
    PacketEmissionEvent(const PacketEvent& packetEvent, Trans* trans);
    // Copy constructor
    PacketEmissionEvent(const PacketEmissionEvent& event);
    PacketEmissionEvent& operator=(const PacketEmissionEvent& event);
    // Destructor
    ~PacketEmissionEvent();
  };

  // Packet state
  class PacketState {
  private :
    // MPI serialization
    friend class boost::serialization::access;
    template<class Archive> 
      void serialize (Archive & ar, const unsigned int version) 
      {
	ar & index;
	ar & type;
	ar & position;
	ar & direction;
	ar & t;
	ar & E;
	ar & nu;
	ar & boostFactor;
	ar & lastEmissionEvent;
	ar & cellVirtualState;
      }
  public :
    int index;
    PacketType type;
    Vector3D position;
    Vector3D direction;
    double t;
    double E;
    double nu;
    double boostFactor; // New!
    PacketEmissionEvent lastEmissionEvent; // New!
    CellVirtualState* cellVirtualState; // New!
    // Hack - Does this work with threading ?!?
    double comptonAbsSigma;
    // Constructors
    PacketState() {};
    PacketState(const Packet& packet);
  };

  // Packet config
  class PacketConfig {
  public :
    bool steadyState;
    bool addVTh;
    bool lineScatter;
    bool electronScatter;
    bool comptonScatter;
    bool greyScatter; // New!
    bool freeFreeAbs;
    bool boundFreeAbs;
    bool comptonAbs;
    double comptonAbsSigma;
    bool dustAbs; // New!
    bool flourescence;
    bool collProc;
    bool nonThermalProc;
    bool twoPhotonProc; // New!
    bool chargeTransferProc; // New!
    // Move this one to grid
    bool transTable;
    UpdateMethod updateMethod;
    bool boost; // New!
    bool convertGPacket;
    int cellIndexMin;
    int cellIndexMax;
    int cellIndexSplit;
    bool reflectInner;
    double PReflectInner;
    bool reflectOuter;
    double PReflectOuter;
    bool reflectInnerGP; // New!
    bool transformKappa;
    bool bbEPacket;
    // Constructor
    PacketConfig();
    PacketConfig(const bool& _steadyState, const bool& _addVTh, const bool& _lineScatter, const bool& _electronScatter, const bool& _comptonScatter, const bool& _greyScatter, const bool& _freeFreeAbs, const bool& _boundFreeAbs, const bool& _comptonAbs, const double& comptonAbsSigma, const bool& _dustAbs, const bool& _flourescence, const bool& _collProc, const bool& _nonThermalProc, const bool& _twoPhotonProc, const bool& _chargeTransferProc, const bool& _transTable, const UpdateMethod& _updateMethod, const bool& _boost, const bool& _convertGPacket, const int& _cellIndexMin, const int& _cellIndexMax, const bool& _reflectInner, const double& _PReflectInner, const bool& _reflectOuter, const double& _PReflectOuter, const bool& _reflectInnerGP, const bool& _transformKappa, const bool& _bbEPacket); // Changed!
    PacketConfig(const PacketConfig& packetConfig);
    // Destructor
    virtual ~PacketConfig();
  };
  
  // Distance update flag
  const double dUpdate=-1;
 
  // Opacity update flag
  const double kUpdate=-1;
  
  // Energy packet
  class Packet {
  private :
    // Packet index
    int index;
    // Packet type
    PacketType type; 
    // Packet config
    PacketConfig* config;
    // Frequency (in reference frame)
    double nu;
    // Frequency (in the comoving frame)
    double nuCMF;
    // Energy (in reference frame)
    double E;
    // Energy (in the comoving frame)
    double ECMF;
    // Packet position
    Vector3D position;
    // Packet direction
    Vector3D direction;
    // Packet time
    double t;
    // Optical depth to next event
    double eventTau;
    // Accumulated optical depth
    double accumulatedTau;
    // Grid pointer
    Grid* grid;
    // Cell pointer
    CellBase* cell;
    // Line iterator
    LineIterator line;
    // Edge iterator
    EdgeIterator edge;
    // Boost bin iterator - New!
    FrequencyBinIterator boostBin;
    // Cell border distance
    double pBorderDistance;
    // Cell border type - New!
    BorderType pBorderType;
    // Distance to next continuum event
    double pContinuumEventDistance;
    // Distance to next resonant line
    double pLineDistance;
    // Distance to next ionization edge
    double pEdgeDistance;
    // Distance to next boost bin - New!
    double pBoostBinDistance;
    // Electron scatter opacity
    double pElectronScatterKappa;
    // Compton scatter opacity
    double pComptonScatterKappa;
    // Grey scatter opacity - New!
    double pGreyScatterKappa;
    // Free-free absorption opacity
    double pFreeFreeAbsKappa;
    // Bound-free absorption opacity
    double pBoundFreeAbsKappa;
    // Compton absorption opacity
    double pComptonAbsKappa;
    // Dust absorption opacity - New!
    double pDustAbsKappa;
    // Continuum event opacity
    double pContinuumEventKappa;
    // Packet terminated flag
    bool pTerminated;
    // Cosine of polar angle (about the radius vector)
    double mu;
    // Cell velocity
    double v;
    // Last event (only used whitg urOnTheFly update method)
    PacketEvent lastEvent;
    // Last event handled flag (only used whith urOnTheFly update method)
    bool lastEventHandled;
    // Last emission event
    PacketEmissionEvent lastEmissionEvent;
    // Boost factor - New!
    double boostFactor;
    // Pointer to packet statistics object
    PacketStat* packetStat;
    // MC statistics update info
    MCUpdateInfo mCUpdateInfo;
    // Testing
    double ECMFOrig;
    // Testing!!!
    //stringstream logStream;
    // Testing!!!
    bool testLog;
    // Check if packet is on grid
    bool onGrid();
    // Check if packet terminated
    bool terminated();
    // Adjust energy (ignored in normal mode) - New!
    void adjustEnergy();
    // Select next resonant line
    void selectNextLine();
    // End of line list
    bool endOfLineList();
    // Select next ionization edge
    void selectNextEdge();
    // End of ionization edge list
    bool endOfEdgeList();
    // Select next boost bin - New!
    void selectNextBoostBin();
    // End of boost bin list - New!
    bool endOfBoostBinList();
    // Calculate cell border distance
    double borderDistance();
    // Get cell border type - New!
    BorderType borderType();
    // Calculate distance to next continuum event
    double continuumEventDistance();
    // Calculate distance to next resonant line
    double lineDistance();
    // Calculate distance to next ionization edge
    double edgeDistance();
    // Calculate distance to next boost bin - New!
    double boostBinDistance();
    // Calculate opacity for electron scattering
    double electronScatterKappa();
    // Calculate opacity for compton scattering (not supported yet)
    double comptonScatterKappa();
    // Calculate opacity for grey scattering
    double greyScatterKappa();
    // Calculate opacity for free-fee absorption
    double freeFreeAbsKappa();
    // Calculate opacity for bound-free absorption
    double boundFreeAbsKappa();
    // Calculate opacity for electron scattering
    double comptonAbsKappa();
    // Calculate opacity for dust absorption - New!
    double dustAbsKappa();
    // Calculate continuum event opacity
    //double continuumEventKappa();
    // Testing
    double continuumEventKappa(bool restFrame=true);
    // Calculate optical depth of next line
    double lineTau();
    // Calculate the optical depth for continuum events
    double continuumEventTau(const double& distance);
    // Update position
    void updatePosition(const double& distance);
    // Get continuum event type
    PacketEventType getContinuumEventType();
    // Get event treatment (absorption/emission) - New!
    void getEventTreatment(const PacketEventType& eventType, bool& emission, bool& absorption);
    // Handle continuum event
    //void continuumEvent(const PacketEventType& eventType);
    void continuumEvent(const PacketEventType& eventType, bool emission=true, bool absorption=true);
    // Handle electron scattering
    void electronScatter();
    // Handle compton scattering (not supported yet)
    void comptonScatter();
    // Handle grey scattering
    void greyScatter();
    // Handle free-free absorption
    //void freeFreeAbs();
    void freeFreeAbs(bool emission=true, bool absorption=true);
    // Handle bound-free absorption
    //void boundFreeAbs();
    void boundFreeAbs(bool emission=true, bool absorption=true);
    // Handle Compton absorption
    //void comptonAbs();
    void comptonAbs(bool emission=true, bool absorption=true);
    // Handle line absorption
    //void lineScatter();
    //void lineAbs();
    void lineAbs(bool emission=true, bool absorption=true);
    // Handle dust absorption - New!
    void dustAbs();
    // Re-emit packet isotropically
    void reEmit(double m=-1);
    // Select next cell
    void selectNextCell();
    // Enter new cell
    void newCell();
    // Enter new boost bin - New!
    void newBoostBin();
    // Pass bound-free new edge - New!
    void boundFreeEdge();
    // Is inner border crossing
    bool innerBorderCrossing();
    // Is outer border crossing
    bool outerBorderCrossing();
    // Is g-packet inner border crossing - New!
    bool innerBorderCrossingGP();
    // Reflect packet
    void reflect();
    // Handle packet event
    //void handleEvent(const PacketEvent& event);
    void handleEvent(const PacketEvent& event, bool initial=false, Trans* trans=NULL); // Changed!
    // Process i-packet
    Trans* processIPacket(PacketEventType& eventType, Trans* trans);
    // Process k-packet
    Trans* processKPacket(PacketEventType& eventType, bool deActivate=false);
    // Process e-packet
    Trans* processEPacket(PacketEventType& eventType);
    // Process BB e-packet
    Trans* processBBEPacket(PacketEventType& eventType);
    // Propagate r-packet
    void propagateRPacket();
    // Propagate g-packet
    void propagateGPacket();
    // Reset r-packet at current position, direction and frequency
    void reset();
    // Terminate packet - New!
    void terminate();
    // Remove current child packet (keeping the descendents) - New!
    void removeCurrentChild();
    // Split and emit child packet - New!
    void emissionSplit(const PacketEventType& eventType);
    // Split packet at border crossing - New!
    void borderSplit();
    // Merge packet at border crossing - New!
    void borderMerge();
  public :
    // Packet history
    vector<PacketEvent> history;
    // Packet childs
    //vector<Packet> child;
    list<Packet> child; // Changed!
    // Constructors
    Packet(PacketConfig* _config=NULL);
    Packet(const double& _E, const FrequencyBin& _nu, const Vector3D& _position, const Vector3D& _direction, const double& _t, CellBase* _cell, Grid* grid, int _index, const PacketType& _type, const PacketEventType& eventType, PacketConfig* _config=NULL, double boostFactor=1.0); // Changed!
    Packet(const PacketState& packetState, const PacketEventType& eventType, Grid* _grid, PacketConfig* _config);
    // Copy constructor
    Packet(const Packet& packet, bool deep=true);
    // Assignment operator
    Packet& operator=(const Packet& packet);
    // Destructor
    ~Packet();
    // Propagate packet
    void propagate(PacketStat* packetStat=NULL);
    // Validate data
    bool valid() const;
    // Give MCExperiment access to private data
    friend class MCExperiment;
    // Give PacketStat access to private data
    friend class PacketStat;
    // Give Spectrum access to private data
    friend class Spectrum;
    // Give SpecEvo access to private data
    friend class SpecEvo;
    // Give PacketState access to private data
    friend class PacketState;
  };

  // Packet state list
  class PacketStateList : public vector<PacketState> {
  private :
    // MPI serialization
    friend class boost::serialization::access;
    template<class Archive> 
      void serialize (Archive & ar, const unsigned int version) 
      { ar & *((vector<PacketState> *) this); }
  public :
    // Constructors
    PacketStateList();
    PacketStateList(int size);
    // Load packet (state) list from file
    void load(const string& dataFile);
    // Save packet (state) list to file
    void save(const string& dataFile) const;
    // Collect packet (state) list (MPI)
    void collect(boost::mpi::communicator* mpiCom);
    // Distribute packet (state) list (MPI)
    void distribute(boost::mpi::communicator* mpiCom);
  };
  
  // Monte Carlo experiment config
  class MCConfig {
  public :
    PacketConfig* packet;
    // Constuctor
    MCConfig();
    MCConfig(PacketConfig* _packet);
  };

  // Forward declarations
  class PacketSampler;
  class PacketStat;

  // MPI messages
  enum { mPacketStat, mCellMCStat, mCellSSMCStat, mSpectrum, mPacketList };
  
  // Monte Carlo experiment
  class MCExperiment {
  protected :
    // MPI communicator
    boost::mpi::communicator* mpiCom;
    // Synchronize packet statistics (MPI)
    void syncPacketStat();
    // Run Monte Carlo experiment (on a single computer)
    void runSingle();
    // Run Monte Carlo experiment (on a cluster using MPI)
    void runCluster();
    // Evolve the Monte-Carlo experiment (on a single computer)
    void evolveSingle(const PacketStateList& packetIn, PacketStateList& packetOut, Grid* grid, int chunkSize=1000); // Changed!
    // Evolve the Monte-Carlo experiment (on a cluster using MPI)
    void evolveCluster(const PacketStateList& packetIn, PacketStateList& packetOut, Grid* grid, int chunkSize=1000); // Changed!
    // Determine if packet is (spatially) on grid
    bool onGrid(const Packet& packet, Grid* grid);
    // Store active (mother and child) packets in packet state list (recursive) - New!
    void storeActivePackets(Packet& packet, PacketStateList& packetState, int& nChild, Grid* grid);
  public :
    // Config
    MCConfig* config;
    // Packet list
    vector<Packet> packet;
    // Frequency bin
    FrequencyBinList nu;
    // Packet sampler
    PacketSampler* packetSampler;
    // Packet statistics
    PacketStat* packetStat;
    // Constructors
    MCExperiment(MCConfig* _config=NULL);
    MCExperiment(PacketSampler* _packetSampler, PacketStat* _packetStat, MCConfig* _config, boost::mpi::communicator* _mpiCom=NULL);
    // Destructor
    virtual ~MCExperiment();
    // Run the Monte Carlo experiment
    void run(void);
    // Evolve the Monte-Carlo experiment
    void evolve(const PacketStateList& packetIn, PacketStateList& packetOut, Grid* grid, int chunkSize=1000); // Changed!
  };

  const int noIndex=-1;

  // Packet sampler
  class PacketSampler {
  protected :
    // Number of bins
    int nBin;
    // Number of packets in bins
    int nPacketBin;
    // Packet index
    int packetIndex;
    // Packet index base
    int packetIndexBase;
  public :
    // Number of packets
    int nPacket;
    // Frequency bin list
    FrequencyBinList nu;
    // Constructor
    PacketSampler() {};
    // Destructor
    virtual ~PacketSampler() {};
    // Sample packet
    virtual Packet sample(int _packetIndex=noIndex) {};
    // Reset packet sampler
    virtual void reset();
  };

  enum BBSampling { bbsEqualEnergy, bbsLogFrequency };
  enum BBFrame { bbfRest, bbfCMF, bbfMixed };

  // Blackbody packet sampler
  class PacketSamplerBB : public PacketSampler {
  private :
    double Teff,Rph,Lph,Eph,t,dt,FBin;
    PacketConfig* packetConfig;
    Grid* grid;
    CellSS* cell;
    BBSampling bbSampling;
    BBFrame bbFrame;
    bool bbLimbDark;
  public :
    // Constructors
    PacketSamplerBB() {};
    PacketSamplerBB(const double& _Rph, const double& _Lph, const int& _nPacket, const int& _nBin, const FrequencyBin& nuLim, const BBSampling& _bbSampling, const BBFrame& _bbFrame, const bool& _bbLimbDark, Grid* _grid, PacketConfig* _packetConfig, int _packetIndexBase=0);
    // Destructor
    virtual ~PacketSamplerBB() {};
    // Sample packet from blackbody distribution
    virtual Packet sample(int _packetIndex=noIndex);
  };

  // Blackbody (luminosity) packet sampler
  /*class PacketSamplerBBL : public PacketSampler {
  private :
    double L,T,EPacket,t,etaBoost;
    vector<double> etaP;
    FrequencyBin nuLim;
    PacketConfig* packetConfig;
    Grid* grid;
    //CellBase* cell;
    Cell* cell; // Changed!
    boost::mpi::communicator* mpiCom;
    // Boost emission probabilities - New!
    void boostEta();
    // Select random frequency - New!
    double randomNu();
    // Select random frequency (intervall version) - New!
    double randomNu(double nuMin, double nuMax);
    // Select random frequency (normal version) - New!
    double randomNuNormal();
    // Select random frequency (boost version) - New!
    double randomNuBoost();
  public :
    // Constructors
    PacketSamplerBBL() {};
    PacketSamplerBBL(const double& _L, const int& _nPacket, const int& cellIndex, Grid* _grid, const FrequencyBin& nuLim, PacketConfig* _config, boost::mpi::communicator* _mpiCom=NULL, int _packetIndexBase=0);
    // Destructor
    virtual ~PacketSamplerBBL() {};
    // Sample packet from blackbody distribution
    virtual Packet sample(int _packetIndex=noIndex);
    };*/

  // Cell blackbody (luminosity) packet sampler - New!
  class CellPacketSamplerBBL : public PacketSampler {
  private :
    double L,T,EPacket,t,etaBoost;
    vector<double> etaP;
    FrequencyBin nuLim;
    PacketConfig* packetConfig;
    Grid* grid;
    Cell* cell;
    CellBase* container;
    boost::mpi::communicator* mpiCom;
    // Boost emission probabilities - New!
    void boostEta();
    // Select random frequency - New!
    double randomNu();
    // Select random frequency (intervall version) - New!
    double randomNu(double nuMin, double nuMax);
    // Select random frequency (normal version) - New!
    double randomNuNormal();
    // Select random frequency (boost version) - New!
    double randomNuBoost();
  public :
    // Constructors
    CellPacketSamplerBBL() {};
    CellPacketSamplerBBL(Grid* _grid, Cell* _cell, CellBase* _container, const FrequencyBin& nuLim, PacketConfig* _config, const double& _EPacket, const double& _L, const double& _T, boost::mpi::communicator* _mpiCom=NULL, int _packetIndexBase=0);
    // Destructor
    virtual ~CellPacketSamplerBBL() {};
    // Sample packet from blackbody distribution
    virtual Packet sample(int _packetIndex=noIndex);
    };

  // Blackbody (luminosity) packet sampler - Changed!
  class PacketSamplerBBL : public PacketSampler {
  private :
    vector<CellPacketSamplerBBL> cellPacketSampler;
    double L,T,EPacket,t;
    PacketConfig* packetConfig;
    Grid* grid;
    Cell* cell;
    boost::mpi::communicator* mpiCom;
  public :
    // Constructors
    PacketSamplerBBL() {};
    PacketSamplerBBL(const double& _L, const int& _nPacket, const int& cellIndex, Grid* _grid, const FrequencyBin& nuLim, PacketConfig* _config, boost::mpi::communicator* _mpiCom=NULL, int _packetIndexBase=0);
    // Destructor
    virtual ~PacketSamplerBBL() {};
    // Sample packet from blackbody distribution
    virtual Packet sample(int _packetIndex=noIndex);
    // Reset
    void reset();
    };

  // Cell gamma-ray/lepton packet sampler
  class CellPacketSamplerGamma : public PacketSampler {
  private :
    double EPacket,t;
    PacketConfig* packetConfig;
    Grid* grid;
    CellBase* cell;
    CellBase* container; // New!
    int nPacketGamma56Ni;
    int nPacketGamma56Co;
    int nPacketPos56Co;
    int nPacketGamma57Ni;
    int nPacketGamma57Co;
    int nPacketPos57Co;
    int nPacketGamma44Ti;
    int nPacketPos44Ti;
    boost::mpi::communicator* mpiCom;
  public :
    // Constructors
    CellPacketSamplerGamma() {};
    CellPacketSamplerGamma(Grid* _grid, CellBase* _cell, CellBase* _container, PacketConfig* _config, const double& _EPacket, boost::mpi::communicator* _mpiCom=NULL, int _packetIndexBase=0); // Changed!
    // Destructor
    virtual ~CellPacketSamplerGamma() {};
    // Sample packet from blackbody distribution
    virtual Packet sample(int _packetIndex=noIndex);
  };

  // Gamma-ray/lepton packet sampler
  class PacketSamplerGamma : public PacketSampler {
  private :
    vector<CellPacketSamplerGamma> cellPacketSampler;
    double EPacket;
    PacketConfig* packetConfig;
    Grid* grid;
    boost::mpi::communicator* mpiCom;
  public :
    // Constructors
    PacketSamplerGamma() {};
    PacketSamplerGamma(const int& _nPacket, Grid* _grid, PacketConfig* _config, boost::mpi::communicator* _mpiCom=NULL, int _packetIndexBase=0);
    // Destructor
    virtual ~PacketSamplerGamma() {};
    // Sample packet from blackbody distribution
    virtual Packet sample(int _packetIndex=noIndex);
  };

  // Cell thermal packet sampler
  class CellPacketSamplerThermal : public PacketSampler {
  private :
    double EPacket,t;
    PacketConfig* packetConfig;
    Grid* grid;
    CellBase* cell;
    CellBase* container; // New!
  public :
    // Constructors
    CellPacketSamplerThermal() {};
    CellPacketSamplerThermal(Grid* _grid, CellBase* _cell, CellBase* _container, PacketConfig* _config, const double& _EPacket, int _packetIndexBase=0); // Changed!
    // Destructor
    virtual ~CellPacketSamplerThermal() {};
    // Sample packet from blackbody distribution
    virtual Packet sample(int _packetIndex=noIndex);
  };

  // Thermal packet sampler
  class PacketSamplerThermal : public PacketSampler {
  private :
    vector<CellPacketSamplerThermal> cellPacketSampler;
    double EPacket;
    PacketConfig* packetConfig;
    Grid* grid;
  public :
    // Constructors
    PacketSamplerThermal() {};
    PacketSamplerThermal(const int& _nPacket, int cellIndexMin, int cellIndexMax, Grid* _grid, PacketConfig* _config, int _packetIndexBase=0);
    // Destructor
    virtual ~PacketSamplerThermal() {};
    // Sample packet from blackbody distribution
    virtual Packet sample(int _packetIndex=noIndex);
    // Reset
    void reset();
  };

  // Cell thermal packet sampler
  class CellPacketSamplerBBThermal : public PacketSampler {
  private :
    //double EPacket,t,T,E,dEdr,etaBoost; // Changed!
    double EPacket,t,T,E,dEdrBack,dEdrFront,etaBoost; // Changed! - Testing!
    vector<double> rMinSub,rMaxSub,PSub,etaP; // Changed!
    FrequencyBin nuLim;
    PacketConfig* packetConfig;
    Grid* grid;
    //CellBase* cell;
    Cell* cell; // Changed!
    CellBase* container; // New!
    boost::mpi::communicator* mpiCom;
    // Get random position
    Vector3D randomPosition() const;
    // Boost emission probabilities - New!
    void boostEta();
    // Select random frequency - New!
    double randomNu();
    // Select random frequency (intervall version) - New!
    double randomNu(double nuMin, double nuMax);
    // Select random frequency (normal version) - New!
    double randomNuNormal();
    // Select random frequency (boost version) - New!
    double randomNuBoost();
  public :
    // Constructors
    CellPacketSamplerBBThermal() {};
    //CellPacketSamplerBBThermal(Grid* _grid, Cell* _cell, CellBase* _container, const FrequencyBin& _nuLim, PacketConfig* _config, const double& _EPacket, const double& _E, const double& _dEdr, boost::mpi::communicator* _mpiCom=NULL, int _packetIndexBase=0, double expFactorWeight=1.0); // Changed!
    CellPacketSamplerBBThermal(Grid* _grid, Cell* _cell, CellBase* _container, const FrequencyBin& _nuLim, PacketConfig* _config, const double& _EPacket, const double& _E, const double& _dEdrBack, const double& _dEdrFront, boost::mpi::communicator* _mpiCom=NULL, int _packetIndexBase=0, double expFactorWeight=1.0); // Changed! - Testing!
    //CellPacketSamplerBBThermal(Grid* _grid, Cell* _cell, CellBase* _container, const FrequencyBin& _nuLim, PacketConfig* _config, const double& _EPacket, const double& _E, const double& _dEdrBack, const double& _dEdrFront, boost::mpi::communicator* _mpiCom=NULL, int _packetIndexBase=0); // Changed! - Testing!
    // Destructor
    virtual ~CellPacketSamplerBBThermal() {};
    // Sample packet from blackbody distribution
    virtual Packet sample(int _packetIndex=noIndex);
  };

  // Thermal packet sampler
  class PacketSamplerBBThermal : public PacketSampler {
  private :
    vector<CellPacketSamplerBBThermal> cellPacketSampler;
    double EPacket;
    PacketConfig* packetConfig;
    Grid* grid;
    boost::mpi::communicator* mpiCom;
    double expFactorWeight; // New! - Removed! - Testing!
    bool LCorr; // New! - Removed! - Testing!
    bool eRadioCorr; // New! - Removed! - Testing!
    //vector<double> dEdt; // New! - Testing!
    // Calculate energy (per volume) // New!
    double ECell(const int& i, int j=noCell) const;
    // Calculate energy (per volume) gradient
    double dEdr(const int& i, int j=noCell) const;
    // Calculate energy (per volume) gradient (back version) - New! - Testing!
    double dEdrBack(const int& i, int j=noCell) const;
    // Calculate energy (per volume) gradient (front version) - New! - Testing!
    double dEdrFront(const int& i, int j=noCell) const;
  public :
    // Constructors
    PacketSamplerBBThermal() {};
    PacketSamplerBBThermal(const int& _nPacket, int cellIndexMin, int cellIndexMax,  Grid* _grid, const FrequencyBin& nuLim, PacketConfig* _config, boost::mpi::communicator* _mpiCom=NULL, int _packetIndexBase=0, double _expFactorWeight=1.0, bool _LCorr=false, bool _eRadioCorr=false); // Changed!
    //PacketSamplerBBThermal(const int& _nPacket, int cellIndexMin, int cellIndexMax,  Grid* _grid, const FrequencyBin& nuLim, PacketConfig* _config, boost::mpi::communicator* _mpiCom=NULL, int _packetIndexBase=0, vector<double> _dEdt=vector<double>(0)); // Changed! - Testing!
    // Destructor
    virtual ~PacketSamplerBBThermal() {};
    // Sample packet from blackbody distribution
    virtual Packet sample(int _packetIndex=noIndex);
    // Reset
    void reset();
  };

  // Cell intensity packet sampler
  class CellPacketSamplerI : public PacketSampler {
  private :
    double EPacket,t,T,E,dEdr; // Changed!
    vector<double> rMinSub,rMaxSub,PSub;
    PacketConfig* packetConfig;
    Grid* grid;
    Cell* cell;
    Cell* container; // New!
    boost::mpi::communicator* mpiCom;
    // Get random position
    Vector3D randomPosition() const;
    // Get random frequency and polar angle
    void randomNuAndTheta(double& nu, double& theta, const double& v) const;
  public :
    // Constructors
    CellPacketSamplerI() {};
    CellPacketSamplerI(Grid* _grid, Cell* _cell, Cell* _container, PacketConfig* _config, const double& _EPacket, const double& _E, const double& _dEdr, boost::mpi::communicator* _mpiCom=NULL, int _packetIndexBase=0, double expFactorweight=1.0); // Changed!
    // Destructor
    virtual ~CellPacketSamplerI() {};
    // Sample packet from blackbody distribution
    virtual Packet sample(int _packetIndex=noIndex);
  };

  // Intensity packet sampler
  class PacketSamplerI : public PacketSampler {
  private :
    vector<CellPacketSamplerI> cellPacketSampler;
    double EPacket;
    PacketConfig* packetConfig;
    Grid* grid;
    boost::mpi::communicator* mpiCom;
    double expFactorWeight; // New!
    bool LCorr; // New!
    bool eRadioCorr; // New!
    // Calculate (radiative) energy
    double ECell(const int& i, int j=noCell) const; // New!
    // Calculate (radiative) energy gradient
    double dEdr(const int& i, int j=noCell) const;
  public :
    // Constructors
    PacketSamplerI() {};
    PacketSamplerI(const int& _nPacket, int cellIndexMin, int cellIndexMax,  Grid* _grid, PacketConfig* _config, boost::mpi::communicator* _mpiCom=NULL, int _packetIndexBase=0, double _expFactorWeight=1.0, bool _LCorr=false, bool _eRadioCorr=false); // Changed!
    // Destructor
    virtual ~PacketSamplerI() {};
    // Sample packet from blackbody distribution
    virtual Packet sample(int _packetIndex=noIndex);
    // Reset
    void reset();
  };

  // Combined blackbody and gamma-ray/lepton packet sampler
  class PacketSamplerBBGamma : public PacketSampler {
  private :
    PacketSamplerGamma packetSamplerGamma;
    PacketSamplerBB packetSamplerBB;
    int nPacketGamma;
    int nPacketBB;
  public :
    // Constructors
    PacketSamplerBBGamma() {};
    PacketSamplerBBGamma(const double& Rph, const double& Lph, const int& nPacketBB, const int& nPacketGamma, const int& nBin, const FrequencyBin& nuLim, const BBSampling& bbSampling, const BBFrame& bbFrame, const bool& bbLimbDark, Grid* grid, PacketConfig* packetConfig, int _packetIndexBase=0);
    // Destructor
    virtual ~PacketSamplerBBGamma() {};
    // Sample packet from blackbody distribution
    virtual Packet sample(int _packetIndex=noIndex);
  };

  // Combined blackbody (luminosity) and gamma-ray/lepton packet sampler
  class PacketSamplerBBLGamma : public PacketSampler {
  private :
    PacketSamplerGamma packetSamplerGamma;
    PacketSamplerBBL packetSamplerBB;
    int nPacketGamma;
    int nPacketBB;
    boost::mpi::communicator* mpiCom;
  public :
    // Constructors
    PacketSamplerBBLGamma() {};
    PacketSamplerBBLGamma(const double& L, const int& _nPacketBB, const int& _nPacketGamma, const FrequencyBin& nuLim, const int& cellIndex, Grid* grid, PacketConfig* packetConfig, boost::mpi::communicator* _mpiCom=NULL, int _packetIndexBase=0);
    // Destructor
    virtual ~PacketSamplerBBLGamma() {};
    // Sample packet from blackbody distribution
    virtual Packet sample(int _packetIndex=noIndex);
  };
}

#endif
