#ifndef _MSOLVER_
#define _MSOLVER_

#include "mc.h"
#include "ntsolver.h"

using namespace mc;
using namespace ntsolver;

namespace msolver {

  // No limit
  const double noLim=-1.0;

  // Path
  class Path : public vector<int> {
  public :
    // Constructors
    Path() {};
    Path(int i);
    // Concatenate operator
    Path operator << (int& i) const;
  };

  // Initialize N/dN for cell
  void initN(Cell* cell, vector<vector<vector<double> > >& N);

  // Initialize N/dN for atom
  void initN(const Atom& atom, vector<vector<double> >& N);

  // Initialize XIState for cell
  void initXIState(Cell* cell, vector<vector<double> >& XIState);

  // Initialize XBState for cell
  void initXBState(Cell* cell, vector<vector<vector<double> > >& XBState);

  // Atom specific config entry
  class AtomConfigEntry {
  public :
    vector<string> atomName;
    // Load data
    virtual void load(const YAML::Node& data);
    // Log data
    virtual void log();
  };

  // Atom specific config
  template <class AtomConfigEntryLike> class AtomConfig {
  private :
    vector<AtomConfigEntryLike> entry;
  public :
    // Get entry
    AtomConfigEntryLike getEntry(const string& _atomName);
    // Get entry
    AtomConfigEntryLike getEntry(Atom& atom);
    // Get entry
    AtomConfigEntryLike getEntry(IState& iState);
    // Load data
    virtual void load(const YAML::Node& data);
    // Log data
    virtual void log();
  };

  // Solver method
  enum SolverMethod { sNone, sNLTE, sNebular, sLTE, sComb };

  // Load solver method
  SolverMethod loadSolverMethod(const YAML::Node& data);

  // Covergence fauilure action
  enum ConvFailAction { cfIgnore, cfAbort };

  // Load convergence failure action
  ConvFailAction loadConvFailAction(const YAML::Node& data);

  // LTE temperature type
  enum LTETempType { lteBB, lteDBB, lteGas }; // Changed!

  // Convergence config
  class LTEConvConfig {
  public :
    double neTol;
    double TMin; // New!
    double TMax; // New!
    // Constructor
    LTEConvConfig();
    // Load data
    void load(const YAML::Node& data);
    // Log data
    void log();
  };

  // Temperature config
  class LTETempConfig {
  public :
    LTETempType type;
    LTETempType typeIon;
    LTETempType typeBound;
    bool update; // New! - Testing!
    // Constructor
    LTETempConfig();
    // Load data
    void load(const YAML::Node& data);
    // Log data
    void log();
  };


  // LTE solver config
  class LTESolverConfig {
  public :
    LTEConvConfig convergence;
    LTETempConfig temperature;
    bool testSwitch1;
    // Constructors
    LTESolverConfig();
    LTESolverConfig(const LTEConvConfig& _convergence, const LTETempConfig& _temperature, const bool& _testSwitch1);
    LTESolverConfig(const string& dataFile);
    // Load data
    void load(const YAML::Node& data);
    void load(const string& dataFile);
    // Log data
    void log();
  };

  // LTE solver
  class LTESolver : public MStateSolver {
  protected :
    //LTESolverConfig config;
    bool pLog;
  public :
    LTESolverConfig config; // Moved here! Hack!
    // Constructors
    LTESolver(LTESolverConfig _config=LTESolverConfig());
    LTESolver(const string& dataFile);
    // Create copy
    virtual MStateSolver* copy();
    // Solve for ionization state of atom
    void solveIonState(Atom& atom, const double& ne, const double& T, const double& TR, const double& W, vector<double>& X);
    // Update ionization state for atom
    void updateIonState(Atom& atom, const double& ne, const double& T, const double& TR, const double& W);
    // Update ionization state for cell (decoupled version)
    void updateIonStateDec(Cell* cell);
    // Update ionization state for cell
    virtual void updateIonState(Cell* cell);
    // Solve for bound state for ion
    void solveBoundState(IState& iState, const double& T, const double& TR, const double& W, vector<double>& X);
    // Update bound state for ion
    void updateBoundState(IState& iState, const double& T, const double& TR, const double& W);
    // Update bound state for atom
    void updateBoundState(Atom& atom, const double& T, const double& TR, const double& W);
    // Update bound state for cell
    virtual void updateBoundState(Cell* cell);
    // Update temperature for cell
    virtual void updateTemperature(Cell* cell);
    // Update partition function for ion
    virtual void updateU(IState& iState, const double& T);
    // Update partition function for atom
    void updateU(Atom& atom, const double& T);
    // Update partition function for cell
    void updateU(Cell* cell);
    // Update matter state assuming LTE
    virtual void updateLTE(Cell* cell);
    // Update matter state
    virtual void update(Cell* cell, bool initiate=false, bool initiateState=true);
    // Return the product of the ionization state number fraction and the electron density 
    virtual double xIon(const IState& iState1, const IState& iState2, const double& T, const double& TR, const double& W);
    // Return the bound state number fraction (relative the ground state)
    virtual double xBound(const BState& groundState, const BState& upperState, const double& T, const double& TR, const double& W);
    // Get logging flag
    bool log() const;
    // Set logging flag
    void log(const bool& _log);
  };

  enum IonizationMethod { ionLTE, ionNebular };
  enum ExcitationMethod { excLTE, excNebular };

  // Population config
  class NebPopConfig {
  public :
    IonizationMethod ionization;
    ExcitationMethod excitation;
    double EIonEdge;
    bool ionExcited;
    // Constructor
    NebPopConfig();
    // Load data
    void load(const YAML::Node& data);
    // Log data
    void log();
  };
  
  // Temperature config
  class NebTempConfig {
  public :
    SolverMethod method;
    // Constructor
    NebTempConfig();
    // Load data
    void load(const YAML::Node& data);
    // Log data
    void log();
  };

  // Convergence config
  class NebConvConfig {
  public :
    double neTol;
    double TMin; // New!
    double TMax; // New!
    // Constructor
    NebConvConfig();
    // Load data
    void load(const YAML::Node& data);
    // Log data
    void log();
  };

  // Nebular solver config
  class NebSolverConfig {
  public :
    NebPopConfig population;
    NebTempConfig temperature;
    NebConvConfig convergence;
    // Constructors
    NebSolverConfig();
    NebSolverConfig(const NebPopConfig& _population, const NebTempConfig& _temperature, const NebConvConfig& _convergence);
    NebSolverConfig(const string& dataFile);
    // Load data
    void load(const YAML::Node& data);
    void load(const string& dataFile);
    // Log data
    void log();
  };

  // Nebular solver
  class NebSolver : public LTESolver {
  protected :
    NebSolverConfig nebConfig;
    LTESolver lteSolver;
  public :
    // Constructors
    //NebSolver(NebSolverConfig _config=NebSolverConfig());
    NebSolver(LTESolverConfig _config=LTESolverConfig(), NebSolverConfig _nebConfig=NebSolverConfig()); // Changed!
    NebSolver(const string& dataFile);
    // Create copy
    virtual MStateSolver* copy();
    // Update temperature for cell
    virtual void updateTemperature(Cell* cell);
    // Return the bound state number fraction (relative the ground sate)
    virtual double xBound(const BState& groundState, const BState& upperState, const double& T, const double& TR, const double& W);
    // Return the product of the ionization state fraction and the electron density
    virtual double xIon(const IState& iState1, const IState& iState2, const double& T, const double& TR, const double& W);
    // Equation (10) in Mazzali & Lucy 93.
    double nIonGround (const double& n1, const double& ne, const double& Te, const double& TR, const double& EIon, const double& g1, const double& g2, const double& W, const double& zeta);
    // Second term in Equation (11) in Mazzali & Lucy 93.
    double nIonExcited (const double& n1, const double& ne, const double& Te, const double& TR, const double& EIon, const double& g1, const double& g2, const double& W, const double& zeta);
    // Equation (14) in Mazzali & Lucy 93.
    double nIonMaz93 (const double& n1, const double& ne, const double& Te, const double& TR, const double& EIon, const double& g1, const double& g2, const double& W, const double& zeta, const double& EIonEdge, const double& b1, const bool& ionExcited);
    // Delta factor as given by equations (15) and (20) in Mazzali & Lucy 93.
    double delta (const double& Te, const double& TR, const double& EIon, const double& W, const double& EIonEdge, const double& b1);
    // Update matter state assuming LTE
    virtual void updateLTE(Cell* cell);
  };

  // Populations config entry
  class NLTEPopConfigEntry : public AtomConfigEntry {
  public :
    Processes processes;
    bool coupled;
    // Constructor
    NLTEPopConfigEntry();
    // Load data
    virtual void load(const YAML::Node& data);
    // Log data
    virtual void log();
  };

  // Population config
  class NLTEPopConfig : public AtomConfig<NLTEPopConfigEntry> {
  public :
    Processes processes(Atom& atom);
    bool coupled(Atom& atom);
  };

  enum DerivativeType { dPartial, dFull, dFullNe };

  // Temperature config entry
  class NLTETempConfigEntry : public AtomConfigEntry {
  public :
    Processes processes;
    // Constructor
    NLTETempConfigEntry();
    // Load data
    virtual void load(const YAML::Node& data);
    // Log data
    virtual void log();
  };

  // Temperature config
  class NLTETempConfig : public AtomConfig<NLTETempConfigEntry> {
  public :
    Processes processes (Atom& atom);
  };

  // Number density error method
  enum ErrNMethod { enMax, enMean };

  enum ScanTMethod { scanTBisection, scanTMixed };

  // Load number denisty error method
  ErrNMethod loadErrNMethod(const YAML::Node& data);

  // Convergence config
  class NLTEConvConfig {
  public :
    int nIter;
    int nIterMin;
    ConvFailAction maxIterAction; // New!
    double nTol;
    double TTol;
    double neTol;
    double dampN;
    double dampT;
    double dampNMin;
    double dampTMin;
    double dampTMax;
    double nMin;
    double TMin;
    double TMax; // New!
    double dTMax;
    double tauLim;
    double XILim; // New!
    double XBLim; // New!
    double XALim; // New!
    ErrNMethod errNMethod;
    DerivativeType derivative;
    double dTDerivative;
    bool scanT;
    //double scanTFac;
    //int scanTDiv;
    //double scanTFacSmall;
    //int scanTDivSmall;
    vector<double> scanTFac;
    double scanTTol;
    ScanTMethod scanTMethod;
    double scanTBAcc;
    int scanTNIter;
    ConvFailAction scanTMaxIterAction; // New!
    double scanTDampN;
    double scanTNeTol;
    double scanTNTol;
    double scanTNTolInit;
    double scanTNTNeTol; // New!
    double scanTNTNeTolInit; // New!
    ConvFailAction scanTNoZeroAction; // New!
    double scanTTauLim; // New!
    double scanTXILim; // New!
    double scanTXBLim; // New!
    double scanTXALim; // New!
    bool iterT;
    double NTNeTol; // New!
    bool dNedN; // New!
    // Constructors
    NLTEConvConfig();
    //NLTEConvConfig(const int& _nIter, const int& _nIterMin, const double& _nTol, const double& _TTol, const double& _neTol, const double& _dampN, const double& _dampT, const double& _dampNMin, const double& _dampTMin, const double& _dampTMax, const double& _nMin, const double& _TMin, const double& _dTMax, const double& _tauLim, const ErrNMethod& _errNMethod, const DerivativeType& _derivative, const double& _dTDerivative, const bool& _scanT, const double& _scanTFac, const double& _scanTDiv, const double& _scanTFacSmall, const double& _scanTDivSmall, const double& _scanTNIter, const double& _scanTDampN, const double& _scanTNeTol, const double& _scanTNTol, const bool& _iterT);
    NLTEConvConfig(const int& _nIter, const int& _nIterMin, const ConvFailAction& _maxIterAction, const double& _nTol, const double& _TTol, const double& _neTol, const double& _dampN, const double& _dampT, const double& _dampNMin, const double& _dampTMin, const double& _dampTMax, const double& _nMin, const double& _TMin, const double& _TMax, const double& _dTMax, const double& _tauLim, const double& _XILim, const double& _XBLim, const double& _XALim, const ErrNMethod& _errNMethod, const DerivativeType& _derivative, const double& _dTDerivative, const bool& _scanT, const vector<double>& _scanTFac, const double& _scanTTol, const ScanTMethod& _scanTMethod, const double& _scanTBAcc, const double& _scanTNIter, const ConvFailAction& _scanTMaxIterAction, const double& _scanTDampN, const double& _scanTNeTol, const double& _scanTNTol, const double& _scanTNTolInit, const double& _scanTNTNeTol, const double& _scanTNTNeTolInit, const ConvFailAction& _scanTNoZeroAction, const double& _scanTTauLim, const double& _scanTXILim, const double& _scanTXBLim, const double& _scanTXALim, const bool& _iterT, const double& _NTNeTol, const bool& _dNedN); // Changed!
    // Load data
    void load(const YAML::Node& data);
    // Log data
    void log();
  };

  // NLTE solver config
  class NLTESolverConfig {
  public :
    NLTEPopConfig population;
    NLTETempConfig temperature;
    NLTEConvConfig convergence;
    bool steadyState;
    bool coupled;
    bool ionCoupled; // New!
    // Constructors
    NLTESolverConfig();
    NLTESolverConfig(const NLTEPopConfig& _population, const NLTETempConfig& _temperature, const NLTEConvConfig& _convergence, const bool& _steadyState, const bool& _coupled, const bool& _ionCoupled); // Changed!
    NLTESolverConfig(const string& dataFile);
    // Load data
    void load(const YAML::Node& data);
    void load(const string& dataFile);
    // Log data
    void log();
  };

  // No index
  const int noIndex=-1;

  // NLTE solver
  class NLTESolver : public MStateSolver {
  protected :
    LTESolver lteSolver;
    Koz92Solver ntSolver;
    int n;
    double rErrN,rErrT,rErrNe;
    double rErrNPrev,rErrTPrev,rErrNePrev;
    double dampN,dampT;
    vector<double> excludeAtom;
    vector<vector<double> > excludeIState;
    vector<vector<vector<double> > > excludeBState;
    bool pLog;
    // Count number of transitions for bound state
    int nTrans(const BState& bState, const Processes& processes);
    // Count number of transitions for ion
    int nTrans(const IState& iState, const Processes& processes);
    // Count number of transitions for atom
    int nTrans(const Atom& atom, const Processes& processes);
    // Calculate maximum optical depth for bound-bound transitions
    double bbTauMax(BState& bState);
    // Calculate maximum optical depth for bound-bound transitions (ionization state version) - New!
    double bbTauMax(IState& iState);
    // Force neutral state of atom
    double forceNeutralState(Atom& atom);
    // Force ground state of ion
    double forceGroundState(IState& iState);
    // Exclude atom, ionization state or bound state
    bool exclude(const Path& path);
  public :
    NLTESolverConfig config;
    // Constructors
    NLTESolver(NLTESolverConfig _config=NLTESolverConfig(), LTESolverConfig _lteConfig=LTESolverConfig(), Koz92SolverConfig _ntSolverConfig=Koz92SolverConfig()); // Changed!
    NLTESolver(const string& dataFile);
    // Copy constructor - New!
    NLTESolver(const NLTESolver& nlteSolver);
    // Create copy
    virtual MStateSolver* copy();
    // Make bound state NLTE matrix index map for ion
    vector<int> bMatrixIndex(IState& iState, int& nStates, const Path& path);
    // Make bound state NLTE matrix index map for atom
    vector<vector<int> > bMatrixIndex(Atom& atom, int& nStates, const Path& path);
    // Make bound state NLTE matrix index map for cell
    vector<vector<vector<int> > > bMatrixIndex(Cell* cell, int& nStates);
    // Make ionization state NLTE matrix index map for atom
    vector<int> iMatrixIndex(Atom& atom, int& nStates, const Path& path);
    // Make ionization state NLTE matrix index map for cell - New!
    vector<vector<int> > iMatrixIndex(Cell* cell, int& nStates);
    // Populate bound state NLTE matrix for ion
    void populateBMatrix(const IState& iState, Mat_DP& matrix, Vec_DP& rhs, const vector<int>& index, const Processes& processes, const Path& path, bool nConserve=true, bool boundFree=false);
    // Populate bound state NLTE matrix for ion (in terms of dN)
    //void populateBMatrixDN(IState& iState, Mat_DP& matrix, Vec_DP& rhs, const vector<int>& index, const Processes& processes, const Path& path, bool nConserve=true, bool boundFree=false);
    void populateBMatrixDN(IState& iState, Mat_DP& matrix, Vec_DP& rhs, const vector<int>& index, const Processes& processes, const Path& path, bool nConserve=true, bool boundFree=false); // Changed! - Testing!
    // Populate bound state NLTE matrix for atom
    void populateBMatrix(const Atom& atom, Mat_DP& matrix, Vec_DP& rhs, const vector<vector<int> >& index, const Processes& processes, const Path& path);
    // Populate bound state NLTE matrix for atom (in terms of dN)
    void populateBMatrixDN(Atom& atom, Mat_DP& matrix, Vec_DP& rhs, const vector<vector<int> >& index, const Processes& processes, const Path& path); // Changed! Removed const before Atom.
    // Populate bound state NLTE matrix for cell (in terms of dN)
    void populateBMatrixDN(Cell* cell, Mat_DP& matrix, Vec_DP& rhs, const vector<vector<vector<int> > >& index);
    // Populate ionization state NLTE matrix for atom
    void populateIMatrix(Atom& atom, Mat_DP& matrix, Vec_DP& rhs, const vector<int>& index, const Processes& processes, const Path& path);
    // Populate ionization state NLTE matrix for atom (in term of dN) - New!
    void populateIMatrixDN(Atom& atom, Mat_DP& matrix, Vec_DP& rhs, const vector<int>& index, const Processes& processes, const Path& path);
    // Populate ionization state NLTE matrix for cell - New!
    void populateIMatrix(Cell* cell, Mat_DP& matrix, Vec_DP& rhs, const vector<vector<int> >& index);
    // Populate ionization state NLTE matrix for cell (in term of dN) - New!
    void populateIMatrixDN(Cell* cell, Mat_DP& matrix, Vec_DP& rhs, const vector<vector<int> >& index);
    // Populate ionization state NLTE matrix for atom (test version 1)
    void populateIMatrixDN1(Atom& atom, Mat_DP& matrix, Vec_DP& rhs, const vector<int>& index, const Processes& processes, const Path& path);
    // Populate ionization state NLTE matrix for cell (test version 1)
    void populateIMatrixDN1(Cell* cell, Mat_DP& matrix, Vec_DP& rhs, const vector<vector<int> >& index);
    // Populate ionization state NLTE matrix for cell (test version 2)
    void populateIMatrixDN2(Cell* cell, Mat_DP& matrix, Vec_DP& rhs, const vector<vector<int> >& index);
    // Log transition rates for cell
    void logGamma(Cell* cell);
    // Get LTE reference factor
    double xLTERef(const BState& bState, const double& T);
    // Get LTE reference number density
    //double nLTERef(const BState& bState, const double& T);
    double nLTERef(BState& bState, const double& T); // Changed! - Testing!
    // (Internal) LTE group transition
    bool LTEGroupTrans(TransBB* transBB) const;
    // Expand LTE groups (in terms of number fraction/density)
    void expandLTEGroups(IState& iState, vector<double>& X, const Path& path);
    // Expand LTE groups (in terms of number density) - New!
    void expandLTEGroupsN(Atom& atom, vector<vector<double> >& N, const Path& path);
    // Expand LTE groups (in terms of number density corrections)
    void expandLTEGroupsDN(IState& iState, vector<double>& dN, const Path& path);
    // Expand LTE groups (in terms of number density corrections) - New!
    void expandLTEGroupsDN(Atom& atom, vector<vector<double> >& dN, const Path& path);
    // Solve for bound state of ion
    virtual void solveBoundState(IState& iState, vector<double>& X, const Processes& processes, const Path& path);
    // Solve for ionization state of atom - New!
    virtual void solveIonState(Atom& atom, vector<double>& dN, const Processes& processes, const Path& path);
    // Solve for ionization state of atom (in terms of number fractions) - Renamed (from solveIonState)!
    virtual void solveIonStateX(Atom& atom, vector<double>& X, const Processes& processes, const Path& path);
    // Solve for bound state of atom
    virtual void solveBoundState(Atom& atom, vector<vector<double> >& dN, const Processes& processes, const Path& path);
    // Solve for bound state of atom (in terms of number densities)
    virtual void solveBoundStateN(Atom& atom, vector<vector<double> >& N, const Processes& processes, const Path& path);
    // Solve for bound state of atom (in terms of number fractions)
    virtual void solveBoundStateX(Atom& atom, vector<double>& XIState, vector<vector<double> >& XBState, const Processes& processes, const Path& path);
    // Solve for bound state of atom (Decoupled version)
    void solveBoundStateDecoupled(Atom& atom, vector<vector<double> >& dN, const Processes& processes, const Path& path);
    // Solve for bound state of atom (in terms of number fractions, Decoupled version)
    void solveBoundStateDecoupledX(Atom& atom, vector<double>& XIState, vector<vector<double> >& XBState, const Processes& processes, const Path& path);
    // Solve for bound state of cell - virtual !?!
    void solveBoundState(Cell* cell, vector<vector<vector<double> > >& dN);
    // Solve for bound state of cell (in terms of number fractions) - virtual !?!
    void solveBoundStateX(Cell* cell, vector<vector<double> >& XIState, vector<vector<vector<double> > >& XBState);
    // Solve for bound state of cell (Coupled version)
    void solveBoundStateCoupled(Cell* cell, vector<vector<vector<double> > >& dN);
    // Solve for ionization state of cell - New!
    //void solveIonState(Cell* cell, vector<vector<double> >& X);
    // Solve for ionization state of cell (coupled version) - New!
    void solveIonStateCoupled(Cell* cell, vector<vector<vector<double> > >& dN);
    // Calculate heating-cooling balance
    //double G(const IState& iState, const vector<double>& dN, const Processes& processes, const Path& path, const double& T);
    double G(IState& iState, const vector<double>& dN, const Processes& processes, const Path& path, const double& T);
    // Calculate heating-cooling balance
    double G(Cell* cell, const vector<vector<vector<double> > >& dN, Processes processes=Processes(), const double& T=noT);
    // Log heating-cooling
    void logHC(Cell* cell, const vector<vector<vector<double> > >& dN, Processes processes=Processes(), const double& T=noT);
    // Calculate derivative of heating-cooling balance
    //double dGdT(const IState& iState, const Processes& processes, const Path& path);
    double dGdT(IState& iState, const Processes& processes, const Path& path);
    // Calculate derivative of heating-cooling balance
    double dGdT(Cell* cell, Processes processes=Processes());
    // Solve for temperature
    void solveTemperature(Cell* cell, const vector<vector<vector<double> > >& dN, double& dT);
    // Scan for temperature
    void scanTemperature(Cell* cell, double& dT, bool update=false);
    // Update matter state at given temperature - Changed!
    void updateMatterStateT(Cell* cell, const double& T, const int& nIter, double dampN=1.0, double neTol=0.0, double NTol=1e99, double NTNeTol=1e99, double tauLim=0.0, double XILim=0.0, double XBLim=0.0, double XALim=0.0);
    // Get change in electron number density
    double getdNe(Cell* cell, const vector<vector<vector<double> > >& dN);
    // Update bound state for atom
    void updateBoundState(Atom& atom, const vector<vector<double> >& dN, const Path& path, bool damp=true);
    // Update bound state for atom (in terms of number fractions)
    void updateBoundStateX(Atom& atom, const vector<double>& XIState, const vector<vector<double> >& XBState, const Path& path);
    //void updateBoundState(Atom& atom, const vector<vector<double> >& dN, const Path& path);
    // Update bound state for cell
    void updateBoundState(Cell* cell, const vector<vector<vector<double> > >& dN, bool damp=true);
    // Update bound state for cell (in terms of number fractions)
    void updateBoundStateX(Cell* cell, const vector<vector<double> >& XIState, const vector<vector<vector<double> > >& XBState);
    // Update bound state for cell (takes care of excluded atoms and states)
    void updateBoundStateExcluded(Cell* cell);
    // Update matter state for cell
    void updateMatterState(Cell* cell, const vector<vector<vector<double> > >& dN, const double& dT, bool damp=true);
    // Update matter state for cell (in terms of number fractions)
    void updateMatterStateX(Cell* cell, const vector<vector<double> >& XIState, const vector<vector<vector<double> > >& XBState, const double& dT);
    // Get maximum relative error in number density - Changed!
    double rErrNMax(Cell* cell, const vector<vector<vector<double> > >& dN, double tauLim=noLim, double XILIm=noLim, double XBLIm=noLim, double XALIm=noLim);
    // Get average relative error in number density - Changed!
    double rErrNMean(Cell* cell, const vector<vector<vector<double> > >& dN, double tauLim=noLim, double XILIm=noLim, double XBLIm=noLim, double XALIm=noLim);
    // Get average relative error in ion number density - New! - Changed!
    double rErrNIMean(Cell* cell, const vector<vector<vector<double> > >& dN, double tauLim=noLim, double XILIm=noLim, double XBLIm=noLim, double XALIm=noLim);
    // Update errors
    void updateErrors(Cell* cell, const vector<vector<vector<double> > >& dN, const double& dT);
    // Update damping factors
    void updateDamping();
    // Update convergence criteria
    void updateConvergence(Cell* cell, const vector<vector<vector<double> > >& dN, const double& dT);
    // Solution converged
    bool converged();
    // Log level inversions
    void logInversion(Cell* cell);
    // Update matter state (overloaded)
    virtual void update(Cell* cell, bool initiate=false, bool initiateState=true);
    // Initiate
    virtual void init(Cell* cell, bool initiateState=true);
    // Reset
    void reset(Cell* cell, vector<vector<vector<double> > >& dN);
    // Update matter state assuming LTE
    virtual void updateLTE(Cell* cell);
    // Load data
    virtual void load(const string& dataFile);
    // Get logging flag
    bool log() const;
    // Set logging flag
    void log(const bool& _log);
  };

  // Population config entry
  class CombPopConfigEntry : public AtomConfigEntry {
  public :
    SolverMethod method;
    SolverMethod excitationMethod;
    SolverMethod ionizationMethod;
    // Constructor
    CombPopConfigEntry();
    // Load data
    void load(const YAML::Node& data);
    // Log data
    virtual void log();
  };

  // Population config
  class CombPopConfig : public AtomConfig<CombPopConfigEntry> {
  public :
    SolverMethod method(Atom& atom);
    SolverMethod excitationMethod(Atom& atom);
    SolverMethod excitationMethod(IState& iState);
    SolverMethod ionizationMethod(Atom& atom);
  };

  // Temperature config
  class CombTempConfig {
  public :
    SolverMethod method;
    // Constructor
    CombTempConfig();
    // Load data
    void load(const YAML::Node& data);
    // Log data
    void log();
  };

  // Combined NLTE and nebular solver configuration
  class CombSolverConfig {
  public :
    CombPopConfig population;
    CombTempConfig temperature;
    // Constructors
    CombSolverConfig();
    CombSolverConfig(const CombPopConfig& _population, const CombTempConfig& _temperature);
    CombSolverConfig(const string& dataFile);
    // Load data
    void load(const YAML::Node& data);
    void load(const string& dataFile);
    // Log data
    void log();
  };

  // Combined NLTE and nebular solver
  class CombSolver : public NLTESolver {
  protected:
    NebSolver nebSolver;
  public:
    CombSolverConfig combConfig;
    // Constructors
    CombSolver(NLTESolverConfig _config=NLTESolverConfig(), CombSolverConfig _combConfig=CombSolverConfig(), NebSolverConfig _nebConfig=NebSolverConfig(), LTESolverConfig _lteSolverConfig=LTESolverConfig());
    CombSolver(const string& dataFile);
    // Copy constructor - New!
    CombSolver(const CombSolver& combSolver);
    // Create copy
    virtual MStateSolver* copy();
    // Solve for bound state of ion (overloaded)
    virtual void solveBoundState(IState& iState, vector<double>& X, const Processes& processes, const Path& path);
    // Solve for ionization state of atom (overloaded) - New!
    virtual void solveIonState(Atom& atom, vector<double>& dN, const Processes& processes, const Path& path);
    // Solve for ionization state of atom (in terms of number fractions, overloaded) - Renamed (from solveIonState)!
    virtual void solveIonStateX(Atom& atom, vector<double>& X, const Processes& processes, const Path& path);
    // Solve for bound state of atom (overloaded)
    virtual void solveBoundState(Atom& atom, vector<vector<double> >& dN, const Processes& processes, const Path& path);
    // Solve for bound state of atom (in terms of number fractions, overloaded)
    virtual void solveBoundStateX(Atom& atom, vector<double>& XIState, vector<vector<double> >& XBState, const Processes& processes, const Path& path);
    // Solve for bound state of cell (overloaded) - virtual !?!
    void solveBoundState(Cell* cell, vector<vector<vector<double> > >& dN);
    // Solve for bound state of cell (in terms of number fractions, overloaded) - virtual !?!
    void solveBoundStateX(Cell* cell, vector<vector<double> >& XIState, vector<vector<vector<double> > >& XBState);
    // Update matter state (overloaded)
    virtual void update(Cell* cell, bool initiate, bool initiateState=true);
    // Initiate
    virtual void init(Cell* cell, bool initiateState=true);
  };
}

#endif
