#include <math.h>
#include "yaml-cpp/yaml.h"
#include "mc-math.h"
#include "mc-physics.h"
#include "mc.h"
#include "ntsolver.h"
// Testing!!!
#include <omp.h>

using namespace std;
using namespace math;
using namespace physics;
using namespace mc;
using namespace ntsolver;

// SF Fortran wrapper
extern"C" {
  void sf_wrapper_(float* xelin, float* ab, float* xa, int* eflag, float* ph, float* peie, float* pem, float* pim);
}

// SF C++ wrapper
void ntsolver::sfWrapper(const double& xelin, const vector<double>& ab, const vector<vector<double> >& xa, const vector<vector<int> >& eflag, double& ph, vector<vector<double> >& peie, vector<vector<vector<double> > >& pem, vector<vector<double> >& pim) {
  float xelin_f;
  float ab_f[mIe];
  float xa_f[mIe*mIon];
  int eflag_f[mIe*mIon];
  float ph_f;
  float peie_f[mIe*mIon];
  float pem_f[mIe*mIon*mIl];
  float pim_f[mIe*mIon];
  int i,j,k;

  // Pack in-parameters
  xelin_f=xelin;
  for (i=0;i<mIe;i++) {
    //if (ab[i]!=0) cout<<"ntsolver::sfWrapper: ab["<<i<<"]: "<<ab[i]<<"\n";
    ab_f[i]=ab[i];
    for (j=0;j<mIon;j++) {
      xa_f[i+j*mIe]=xa[i][j];
      eflag_f[i+j*mIe]=eflag[i][j];
      //if (xa[i][j]!=0) cout<<"ntsolver::sfWrapper: xa["<<i<<"]["<<j<<"]: "<<xa[i][j]<<"\n";
    }
  }

  // Call the SF Fortran wrapper
  // SF does currently not work with omp - Fix!!!
  //cout<<"ntsolver::sfWrapper: 1, "<<omp_get_thread_num()<<"\n";
  #pragma omp critical
  sf_wrapper_(&xelin_f,ab_f,xa_f,eflag_f,&ph_f,peie_f,pem_f,pim_f);
  //cout<<"ntsolver::sfWrapper: 2, "<<omp_get_thread_num()<<"\n";

  // Unpack out-parameters
  ph=ph_f;
  for (i=0;i<mIe;i++) {
    for (j=0;j<mIon;j++) {
      peie[i][j]=peie_f[i+j*mIe];
      pim[i][j]=pim_f[i+j*mIe];
      //if (peie[i][j]!=0) cout<<"ntsolver::sfWrapper: peie["<<i<<"]["<<j<<"]: "<<peie[i][j]<<"\n";
      //if (pim[i][j]!=0) cout<<"ntsolver::sfWrapper: pim["<<i<<"]["<<j<<"]: "<<pim[i][j]<<"\n";
      for (k=0;k<mIl;k++) {
	pem[i][j][k]=pem_f[i+j*mIe+k*(mIe*mIon)];
	//if (pem[i][j][k]!=0) cout<<"ntsolver::sfWrapper: pem["<<i<<"]["<<j<<"]["<<k<<"]: "<<pem[i][j][k]<<"\n";
      }
    }
  }
}

// Get SF atom index
double ntsolver::sfIndexAtom(double Z) {
  switch(int(Z)) {
  case 1 : return 1;
  case 2 : return 2;
  case 6 : return 3;
  case 7 : return 4;
  case 8 : return 5;
  case 10 : return 6;
  case 11 : return 7;
  case 12 : return 8;
  case 14 : return 9;
  case 16 : return 10;
  case 18 : return 11;
  case 20 : return 12;
  case 26 : return 13;
  case 27 : return 14;
  case 28 : return 15;
  }
}

// Get SF ionization state index
double ntsolver::sfIndexIState(double C) {
  return abs(C)+1;
}

// Find bound-free ground-state transition
TransBF* ntsolver::findGroundStateBF(BState* bState) {
  TransBF* transBF=NULL;
  int i;
  for (i=0;i<bState->upTransBF.size();i++) {
    if (bState->upTransBF[i]->upperState->E()==0) { transBF=bState->upTransBF[i]; break; }
  }
  return transBF;
}

// Find bound-bound ground-state transition
TransBB* ntsolver::findGroundStateBB(BState* bState) {
  TransBB* transBB=NULL;
  int i;
  for (i=0;i<bState->downTransBB.size();i++) {
    if (bState->downTransBB[i]->lowerState->E()==0) { transBB=bState->downTransBB[i]; break; }
  }
  return transBB;
}

// Find bound-bound transition
TransBB* ntsolver::findTransBB(BState* lowerState, BState* upperState) {
  TransBB* transBB=NULL;
  int i;
  for (i=0;i<lowerState->upTransBB.size();i++) {
    if (lowerState->upTransBB[i]->upperState==upperState) { transBB=lowerState->upTransBB[i]; break; }
  }
  return transBB;
}
		
// Constructor
FitParam::FitParam(const double& _C, const double& a1, const double& a2, const double& a3, const double& a4) : C(_C) { a[0]=a1; a[1]=a2; a[2]=a3; a[3]=a4; }

// Load data
void FitParam::load(const YAML::Node& data) {
  int i;

  // Load fit parameters
  if (data.size()==5) {
    C=data[0].as<double>();
    for (i=0;i<4;i++) a[i]=data[i+1].as<double>();
  }
  else { cerr<<"Koz92Param::load, Error: Could not get fit parameters\n"; exit(-1); }  
  
}

// Constructor
IonEntry::IonEntry(const double& _Z, const double& _C, const FitParam& _param) : Z(_Z), C(_C), param(_param) {}

// Load data
void IonEntry::load(const YAML::Node& data) {

  // Load number of protons
  if (data["Z"]) Z=data["Z"].as<double>();
  else { cerr<<"IonEntry::load, Error: Could not get Z\n"; exit(-1); }  

  // Load ion charge
  if (data["C"]) C=data["C"].as<double>();
  else { cerr<<"IonEntry::load, Error: Could not get C\n"; exit(-1); }  

  // Load ionization fit parameters
  if (data["param"]) param.load(data["param"]);
  else { cerr<<"IonEntry::load, Error: Could not get ionization fit parameters\n"; exit(-1); }  
}

// Load data
void IonTable::load(const YAML::Node& data) {
  IonEntry ionEntry;
  int i;
 
 // Load each table entry
  for (i=0;i<data.size();i++) {
    ionEntry.load(data[i]);
    push_back(ionEntry);
  }
}
 
// Find entry in ionization table
FitParam* IonTable::find(const double& Z, const double& C) {
  int i;
  for (i=0;i<size();i++) {
    if ((*this)[i].Z==Z and (*this)[i].C==C) return &((*this)[i].param);
  }
  return NULL;
}

// Constructor
ExcEntry::ExcEntry(const double& _Z, const double& _C, const int& _upperStateIndex, const double& _fraction) : Z(_Z), C(_C), upperStateIndex(_upperStateIndex), fraction(_fraction) {}

// Load data
void ExcEntry::load(const YAML::Node& data) {

  // Load number of protons
  if (data["Z"]) Z=data["Z"].as<double>();
  else { cerr<<"IonEntry::load, Error: Could not get Z\n"; exit(-1); }  

  // Load ion charge
  if (data["C"]) C=data["C"].as<double>();
  else { cerr<<"IonEntry::load, Error: Could not get C\n"; exit(-1); }  

  // Load bound state index
  if (data["i"]) upperStateIndex=data["i"].as<int>()-1;
  else { cerr<<"IonEntry::load, Error: Could not get bound state index\n"; exit(-1); }  

  // Load excitation fraction
  if (data["frac"]) fraction=data["frac"].as<double>();
  else { cerr<<"IonEntry::load, Error: Could not get excitation fraction\n"; exit(-1); }  
}

// Load data
void ExcTable::load(const YAML::Node& data) {
  ExcEntry excEntry;
  int i;
 
 // Load each table entry
  for (i=0;i<data.size();i++) {
    excEntry.load(data[i]);
    push_back(excEntry);
  }
}

// Find entry in excitation table
double* ExcTable::find(const double& Z, const double& C, const int& upperStateIndex) {
  int i;
  for (i=0;i<size();i++) {
    //cout<<"ExcTable::find: Z: "<<(*this)[i].Z<<","<<Z<<", C: "<<(*this)[i].C<<","<<C<<", upperStateIndex: "<<(*this)[i].upperStateIndex<<","<<upperStateIndex<<"\n";
    if ((*this)[i].Z==Z and (*this)[i].C==C and (*this)[i].upperStateIndex==upperStateIndex) return &((*this)[i].fraction);
  }
  return NULL;
}

// Load
void Koz92Fitter::load(const string& dataFile) {
  YAML::Node data;

  // Load data from file
  data=YAML::LoadFile(dataFile);
  // Error handling ?

  //cout<<"Koz92Fitter::load: data.size: "<<data.size()<<"\n";

  // Load ionization table
  if (data["ionTable"]) ionTable.load(data["ionTable"]);
  else { cerr<<"Koz92Fitter::load, Error: Could not get ionization table\n"; exit(-1); }

  // Load excitation table
  if (data["excTable"]) excTable.load(data["excTable"]);
  else { cerr<<"Koz92Fitter::load, Error: Could not get excitation table\n"; exit(-1); }  

  // Load heating fit parameters
  if (data["heatParam"]) heatParam.load(data["heatParam"]);
  else { cerr<<"Koz92Fitter::load, Error: Could not get heating fit parameters\n"; exit(-1); }  

  // Load ionization fit parameters
  if (data["ionParam"]) ionParam.load(data["ionParam"]);
  else { cerr<<"Koz92Fitter::load, Error: Could not get ionization fit parameters\n"; exit(-1); }  

  // Load excitation fit parameters
  if (data["excParam"]) excParam.load(data["excParam"]);
  else { cerr<<"Koz92Fitter::load, Error: Could not get excitation fit parameters\n"; exit(-1); }  
}

// Get effective ionization potential (see Kozma 1992)
double Koz92Fitter::XiEff(const double& Z, const double& C, const double& xe) {
  FitParam* param=ionTable.find(Z,C);
  if (param!=NULL) return fit(xe,*param)*eV;
  else return 1e99;
}

// Apply fitting formula (see Kozma 1992)
double Koz92Fitter::fit(const double& xe, const FitParam& param) const {
  double value,alpha=0;
  int i;

  if (xe<pow(10,-4) or xe>pow(10,-0.5)) cout<<"Koz92Fitter::fit: xe="<<xe<<" outside interpolation range\n";

  for (i=0; i<4; i++) alpha+=param.a[i]*pow(log10(xe)+4,i+1);
  value=param.C*pow(10,alpha);

  return value;
}

// Get fractional energy going into heating
double Koz92Fitter::heatFrac(const double& xe) const {
  double frac;
  frac=fit(xe,heatParam);
  if (frac<0) cout<<"Koz92Fitter::heatFrac: Warning: Negative fraction: "<<frac<<"\n";
  return frac;
}

// Get fractional energy going into ionization
double Koz92Fitter::ionFrac(const double& xe) const {
  double frac;
  frac=fit(xe,ionParam);
  if (frac<0) cout<<"Koz92Fitter::ionFrac: Warning: Negative fraction: "<<frac<<"\n";
  return frac;
}

// Get fractional energy going into ionization (for a specified ion)
double Koz92Fitter::ionFrac(const double& Z, const double& C, const double& Xi, const double& X, const double& xe) {
  double frac;
  frac=(Xi/XiEff(Z,C,xe))*X;
  if (frac<0) cout<<"Koz92Fitter::ionFrac: Z: "<<Z<<", C: "<<C<<", Warning: Negative fraction: "<<frac<<"\n";
  return frac;
}

// Get fractional energy going into excitation
double Koz92Fitter::excFrac(const double& xe) const {
  double frac;
  frac=fit(xe,excParam);
  if (frac<0) cout<<"Koz92Fitter::excFrac: Warning: Negative fraction: "<<frac<<"\n";
  return frac;
}

// Get fractional energy going into excitation (to a specified state for a specified ion)
double Koz92Fitter::excFrac(const double& Z, const double& C, const double& upperStateIndex, const double& X, const double& xe) {
  double frac=0.0;
  double* excFracTrans=excTable.find(Z,C,upperStateIndex);
  if (excFracTrans!=NULL) frac=(*excFracTrans)*excFrac(xe)*X;
  if (frac<0) cout<<"Koz92Fitter::excFrac: Z: "<<Z<<", C: "<<C<<", upperState: "<<upperStateIndex<<", Warning: Negative fraction: "<<frac<<"\n";
  return frac;
}

void Koz92Fitter::update(Cell* cell, bool initiate) {
  double Z,C,X,excFracTrans,excFracTot=0,ionFracTrans,ionFracTot=0;
  TransBB* transBB;
  TransBF* transBF;
  int i,j,k,l;

  // Update non-thermal heating rate
  cell->nonThermalHeating(cell->eRadio()*cell->rho()*heatFrac(cell->xe()));
  cout<<"Koz92Fitter::update: nonThermalHeating: "<<cell->nonThermalHeating()<<"\n";

  // Update non-thermal ionization rate
  cell->nonThermalIonization(cell->eRadio()*cell->rho()*ionFrac(cell->xe()));
  cout<<"Koz92Fitter::update: nonThermalIonization: "<<cell->nonThermalIonization()<<"\n";

  // Update non-thermal excitation rate
  cell->nonThermalExcitation(cell->eRadio()*cell->rho()*excFrac(cell->xe()));
  cout<<"Koz92Fitter::update: nonThermalExcitation: "<<cell->nonThermalExcitation()<<"\n";

  // For each atom
  for (i=0;i<cell->atom.size();i++) {
    // Get number of protons
    Z=cell->atom[i].Z();
    // For each ionization state
    for (j=0;j<cell->atom[i].iState.size();j++) {
      // Get ion charge
      C=cell->atom[i].iState[j].C();
      // Get ion number fraction
      X=cell->atom[i].iState[j].n()/cell->n();
      // For each bound state
      for (k=0;k<cell->atom[i].iState[j].bState.size();k++) {
	//if (k==0) cout<<"Koz92Fitter::update: "<<cell->atom[i].iState[j].name()<<", nTransBB: "<<cell->atom[i].iState[j].bState[k].upTransBB.size()<<"\n";
	// For each upward bound-bound transition
	for (l=0;l<cell->atom[i].iState[j].bState[k].upTransBB.size();l++) {
	  transBB=cell->atom[i].iState[j].bState[k].upTransBB[l];
	  // If transition from the ground state
	  if (k==0) {
	    // Get fractional energy going into excitation
	    excFracTrans=excFrac(Z,C,transBB->upperState->getIndex(),X,cell->xe());
	    excFracTot+=excFracTrans;
	    // Update non-thermal excitation rate
	    transBB->CNU(cell->eRadio()*cell->rho()*excFracTrans/transBB->Xi());
	    //cout<<"Koz92Fitter::update: BB transition 0->"<<transBB->upperState->getIndex()<<", excFracTrans: "<<excFracTrans<<", excFracTot: "<<excFracTot<<", CNU: "<<transBB->CNU()<<"\n";
	  }
	  // Else - Set excitation rate to zero
	  else transBB->CNU(0);
	}
	// For each upward bound-free transition from the ground state
	for (l=0;l<cell->atom[i].iState[j].bState[k].upTransBF.size();l++) {
	  transBF=cell->atom[i].iState[j].bState[k].upTransBF[l];
	  // If transition from the ground state
	  if (k==0) {
	    // Get fractional energy going into ionization
	    ionFracTrans=ionFrac(Z,C,transBB->Xi(),X,cell->xe());
	    ionFracTot+=ionFracTrans;
	    // Update non-thermal ionization rate
	    transBF->CNU(cell->eRadio()*cell->rho()*ionFracTrans/transBF->Xi());
	    //cout<<"Koz92Fitter::update: BF transition 0->"<<transBF->upperState->getIndex()<<", ionFracTrans: "<<ionFracTrans<<", ionFracTot: "<<ionFracTot<<", CNU: "<<transBF->CNU()<<"\n";
	  }
	  // Else - Set ionization rate to zero
	  else transBF->CNU(0);
	}
      }
    }
  }

  // For each atom
  for (i=0;i<cell->atom.size();i++) {
    // For each ionization state
    for (j=0;j<cell->atom[i].iState.size();j++) {
      // For each bound state
      for (k=0;k<cell->atom[i].iState[j].bState.size();k++) {
	// For each upward bound-bound transition
	for (l=0;l<cell->atom[i].iState[j].bState[k].upTransBB.size();l++) {
	  transBB=cell->atom[i].iState[j].bState[k].upTransBB[l];
	  // If transition from the ground state - Normalize non-thermal excitation rate
	  if (k==0 and excFracTot!=0) transBB->CNU(transBB->CNU()*(excFrac(cell->xe())/excFracTot));
	  //if (k==0) cout<<"Koz92Fitter::update: BB transiton "<<transBB->lowerState->getIndex()<<"->"<<transBB->upperState->getIndex()<<", CNU: "<<transBB->CNU()<<"\n";
	}
	// For each upward bound-free transition from the ground state
	for (l=0;l<cell->atom[i].iState[j].bState[k].upTransBF.size();l++) {
	  transBF=cell->atom[i].iState[j].bState[k].upTransBF[l];
	  // If transition from the ground state - Normalize non-thermal ionization rate
	  if (k==0 and ionFracTot!=0) transBF->CNU(transBF->CNU()*(ionFrac(cell->xe())/ionFracTot));
	  //if (k==0) cout<<"Koz92Fitter::update: BF transition "<<transBF->lowerState->getIndex()<<"->"<<transBF->upperState->getIndex()<<", CNU: "<<transBF->CNU()<<"\n";
	}
      }
    }
  }
}

// Constructors - New!
SFTransBB::SFTransBB() : pLowerIndex(noIndex), pUpperIndex(noIndex), pSFIndex(noIndex) {};

// Load data - New!
void SFTransBB::load(const YAML::Node& data) {
  int i;

  // Get lower index (and convert to zero-based scale)
  if (data["lowerIndex"]) pLowerIndex=data["lowerIndex"].as<int>()-1;
  else {cout<<"SFTransBB::load: Error: Could not get lower index\n"; exit(-1);}

  // Get upper index (and convert to zero-based scale)
  if (data["upperIndex"]) pUpperIndex=data["upperIndex"].as<int>()-1;
  else {cout<<"SFTransBB::load: Error: Could not get upper index\n"; exit(-1);}

  // Get SF index (and convert to zero-based scale)
  if (data["sfIndex"]) pSFIndex=data["sfIndex"].as<int>()-1;
  else {cout<<"SFTransBB::load: Error: Could not get SF index\n"; exit(-1);}

  return;
}

// Get lower index - New!
double SFTransBB::lowerIndex() const {
  return pLowerIndex;
}

// Get upper index - New!
double SFTransBB::upperIndex() const {
  return pUpperIndex;
}

// Get SF index - New!
double SFTransBB::sfIndex() const {
  return pSFIndex;
}

// Constructors
SFBState::SFBState() {};

// Load data - Changed!
void SFBState::load(const YAML::Node& data) {
  int i;

  // If index data present
  if (data["index"]) {

    // If scalar value
    if (data["index"].Type()==YAML::NodeType::Scalar) {

      // Resize index vector
      index.resize(1);
      
      // Get index(and convert to zero-based scale
      index[0]=data["index"].as<int>()-1;
    }
    // Else
    else {
      
      // Resize index vector
      index.resize(data["index"].size());
      
      // Get each index (and convert to zero-based scale)
      for (i=0;i<data["index"].size();i++) index[i]=data["index"][i].as<int>()-1;
    }
  }
  else {cout<<"SFBState::load: Error: Could not get index\n"; exit(-1);}

  // If SF index data present
  if (data["sfIndex"]) {
    
    // If scalar value
    if (data["sfIndex"].Type()==YAML::NodeType::Scalar) {
      
      // Resize SF index vector
      sfIndex.resize(1);
      
      // Get SF index(and convert to zero-based scale
      sfIndex[0]=data["sfIndex"].as<int>()-1;
    }
    // Else
    else {
      
      // Resize SF index vector
      sfIndex.resize(data["sfIndex"].size());

      // Get each SF index (and convert to zero-based scale)
      for (i=0;i<data["sfIndex"].size();i++) sfIndex[i]=data["sfIndex"][i].as<int>()-1;
    }
  }
  else {cout<<"SFBState::load: Error: Could not get SF index\n"; exit(-1);}

  return;
}

// Log data
void SFBState::log() {};

// Get index
/*double SFBState::index() const {
  return pIndex;
  }*/

// Constructors
SFIState::SFIState() : pIndex(noIndex), pIonization(true) {};

// Load data
void SFIState::load(const YAML::Node& data) {
  int i;

  // Get C
  if (data["C"]) pC=data["C"].as<double>();
  else {cout<<"SFIState::load: Error: Could not get C\n"; exit(-1);}

  // Get ionization flag
  if (data["ionization"]) pIonization=(data["ionization"].as<string>().compare("yes")==0);
  else pIonization=true;

  // Get SF index (and convert to zero-based scale)
  pSFIndex=sfIndexIState(pC)-1;

  // If bound state data present
  if (data["bState"]) {

    // Resize bound state vector
    bState.resize(data["bState"].size());
    
    // Load each bound state
    for (i=0;i<data["bState"].size();i++) bState[i].load(data["bState"][i]);
  }

  // If bound-bound transition data present - New!
  if (data["transBB"]) {

    // Resize bound-bound transition vector
    transBB.resize(data["transBB"].size());
    
    // Load each bound-bound transition
    for (i=0;i<data["transBB"].size();i++) transBB[i].load(data["transBB"][i]);
  }

  return;
}

// Log data
void SFIState::log() {};

// Get C
double SFIState::C() const {
  return pC;
}

// Get index
int SFIState::index() const {
  return pIndex;
}

// Set index
void SFIState::index(const int& _index) {
  pIndex=_index;
}

// Get SF index
int SFIState::sfIndex() const {
  return pSFIndex;
}

// Get ionization flag
bool SFIState::ionization() const {
  return pIonization;
}

// Constructors
SFAtom::SFAtom() : pIndex(noIndex) {};

// Load data
void SFAtom::load(const YAML::Node& data) {
  int i;

  // Get Z
  if (data["Z"]) pZ=data["Z"].as<double>();
  else {cout<<"SFAtom::load: Error: Could not get Z\n"; exit(-1);}

  // Get SF index (and convert to zero-based scale)
  pSFIndex=sfIndexAtom(pZ)-1;

  // If ionization state data present
  if (data["iState"]) {

    // Resize ionization state vector
    iState.resize(data["iState"].size());
    
    // Load each ionization state
    for (i=0;i<data["iState"].size();i++) iState[i].load(data["iState"][i]);
  }

  return;
}

// Log data
void SFAtom::log() {};

// Get Z
double SFAtom::Z() const {
  return pZ;
}

// Get index
int SFAtom::index() const {
  return pIndex;
}

// Set index
void SFAtom::index(const int& _index) {
  pIndex=_index;
}

// Get SF index
int SFAtom::sfIndex() const {
  return pSFIndex;
}

// Constructors
SFMapping::SFMapping() {};

// Load data
void SFMapping::load(const YAML::Node& data) {
  int i;

  // If atom data present
  if (data["atom"])  {

    // Resize atom vector
    atom.resize(data["atom"].size());
    
    // Load each atom
    for (i=0;i<data["atom"].size();i++) atom[i].load(data["atom"][i]);
  }

  return;
}

// Log data
void SFMapping::log() {};

// Initiate
void SFMapping::init(Cell* cell) {
  int i,j,k,l,ii,jj;
  SFIState* iState;
  SFBState* bState;
  SFTransBB* sfTransBB;
  BState* lowerState;
  BState* upperState;
  TransBF* upTransBF;
  TransBB* downTransBB;
  TransBB* transBB;
  double g;

  // For each (SF) atom
  for (i=0;i<atom.size();i++) {

    // For each atom
    for (ii=0;ii<cell->atom.size();ii++) {
      // If Z match - Set index and break
      if (atom[i].Z()==cell->atom[ii].Z()) { atom[i].index(ii); break; }
    }

    // If Z match found
    if (atom[i].index()!=noIndex) {

      // For each (SF) ionization state
      for (j=0;j<atom[i].iState.size();j++) {
	
	// For each ionization state
	for (jj=0;jj<cell->atom[ii].iState.size();jj++) {
	  // If C match - Set index and break
	  if (atom[i].iState[j].C()==cell->atom[ii].iState[jj].C()) { atom[i].iState[j].index(jj); break; }
	}
      }
    }
  }

  // For each (SF) atom
  for (i=0;i<atom.size();i++) {

    // If index not set - Erase from list and continue
    if (atom[i].index()==noIndex) { atom.erase(atom.begin()+i); i--; continue; }

    // For each (SF) ionization state
    for (j=0;j<atom[i].iState.size();j++) {

      // If index not set - Erase from list and continue
      if (atom[i].iState[j].index()==noIndex) { atom[i].iState.erase(atom[i].iState.begin()+j); j--; continue; }
    }
  }

  // For each (SF) atom
  for (i=0;i<atom.size();i++) {

    cout<<"SFMapping::init: cell: "<<cell->index()<<", Added "<<cell->atom[atom[i].index()].name()<<"\n";

    // For each (SF) ionization state
    for (j=0;j<atom[i].iState.size();j++) {

      // Set (SF) ionization state pointer
      iState=&(atom[i].iState[j]);

      // If ionization flag set
      if (iState->ionization()) {

	// Set lower state pointer
	lowerState=&(cell->atom[atom[i].index()].iState[iState->index()].bState[0]);

	// Find ground state transition - New!
	// Has to be improved to include ground-state multiplets!
	upTransBF=findGroundStateBF(lowerState);
	// If not found - Log error and exit
	if (upTransBF==NULL) { cout<<"SFMapping::init: cell: "<<cell->index()<<", Could not find "<<lowerState->iState()->name()<<" ground state transition\n"; exit(-1); }

	// Set non-thermal flag for the ground-state transition - Changed! - Removed!
	//upTransBF->nonThermal(true);

	// Activate non-thermal processes for ground-state transition - New!
	upTransBF->activate(pNonThermal);

	cout<<"SFMapping::init: cell: "<<cell->index()<<", Added "<<lowerState->iState()->name()<<":"<<lowerState->getIndex()<<"->"<<upTransBF->upperState->iState()->name()<<":"<<upTransBF->upperState->getIndex()<<"\n";
      }

      // For each (SF) bound state
      for (k=0;k<iState->bState.size();k++) {
	
	// Set (SF) bound state pointer
	bState=&(iState->bState[k]);

	// Sum statistical weight - New!
	g=0; for (l=0;l<bState->index.size();l++) g+=cell->atom[atom[i].index()].iState[iState->index()].bState[bState->index[l]].g();

	// Resize distribution fraction vector - New!
	bState->distFrac.resize(bState->index.size());
	
	// For each index
	for (l=0;l<bState->index.size();l++)  {

	  // Set upper state pointer
	  upperState=&(cell->atom[atom[i].index()].iState[iState->index()].bState[bState->index[l]]);

	  // Set distribution fraction according to statistical weight - New!
	  bState->distFrac[l]=upperState->g()/g;

	  // Find ground state transition - New!
	  // Has to be improved to include ground-state multiplets!
	  downTransBB=findGroundStateBB(upperState);
	  // If not found - Log error and exit
	  if (downTransBB==NULL) { cout<<"SFMapping::init: cell: "<<cell->index()<<", Could not find "<<upperState->iState()->name()<<":"<<upperState->getIndex()<<" ground state transition\n"; exit(-1); }

	  // Set non-thermal flag for ground state transition - Changed! - Removed!
	  //downTransBB->nonThermal(true);

	  // Activate non-thermal processes for ground state transition - New!
	  downTransBB->activate(pNonThermal);

	  cout<<"SFMapping::init: cell: "<<cell->index()<<", Added "<<downTransBB->lowerState->iState()->name()<<":"<<downTransBB->lowerState->getIndex()<<"->"<<upperState->iState()->name()<<":"<<upperState->getIndex()<<" with distribution fraction "<<bState->distFrac[l]<<"\n";
	}  
      }

      // For each (SF) bound-bound transition
      for (k=0;k<iState->transBB.size();k++) {
	
	// Set (SF) bound-bound transition pointer
	sfTransBB=&(iState->transBB[k]);
	
	// Set lower and upper state pointer
	lowerState=&(cell->atom[atom[i].index()].iState[iState->index()].bState[sfTransBB->lowerIndex()]);
	upperState=&(cell->atom[atom[i].index()].iState[iState->index()].bState[sfTransBB->upperIndex()]);
	
	// Find transition 
	transBB=findTransBB(lowerState,upperState);
	// If not found - Log error and exit
	if (transBB==NULL) { cout<<"SFMapping::init: cell: "<<cell->index()<<", Could not find "<<lowerState->iState()->name()<<":"<<lowerState->getIndex()<<"->"<<upperState->iState()->name()<<":"<<upperState->getIndex()<<" transition\n"; exit(-1); }

	// Activate non-thermal processes for transition - New!
	transBB->activate(pNonThermal);

	cout<<"SFMapping::init: cell: "<<cell->index()<<", Added "<<lowerState->iState()->name()<<":"<<lowerState->getIndex()<<"->"<<upperState->iState()->name()<<":"<<upperState->getIndex()<<"\n";
      }
    }
  }
}

// Constructors
Koz92InterfaceConfig::Koz92InterfaceConfig() {};
Koz92InterfaceConfig::Koz92InterfaceConfig(const string& dataFile) { load(dataFile); };

// Load data
void Koz92InterfaceConfig::load(const YAML::Node& data) {

  // Load SF mapping
  if (data["sfMapping"]) sfMapping.load(data["sfMapping"]);
}

void Koz92InterfaceConfig::load(const string& dataFile) {
  YAML::Node data;

  // Load data from file
  data=YAML::LoadFile(dataFile);

  // Load data
  if (data["koz92Interface"]) load(data["koz92Interface"]);
}

// Log data
void Koz92InterfaceConfig::log() {};

// Load
void Koz92Interface::load(const string& dataFile) {
  config.load(dataFile);
}

// Update
void Koz92Interface::update(Cell* cell, bool initiate) {
  double xelin,ph,heatFrac,excFrac=0,ionFrac=0,excFracTrans;
  vector<double> ab(mIe,0);
  vector<vector<double> > xa(mIe),peie(mIe),pim(mIe);
  vector<vector<vector<double> > > pem(mIe);
  vector<vector<int> > eflag(mIe);
  TransBB* transBB;
  TransBF* transBF;
  SFAtom* atom;
  SFIState* iState;
  SFBState* bState;
  SFTransBB* sfTransBB;
  BState* lowerState;
  BState* upperState;
  int i,j,k,l,m;

  // If requested - Initiate
  if (initiate) init(cell);

  // Initialize SF vectors
  for (i=0;i<mIe;i++) {
    xa[i].resize(mIon,0);
    eflag[i].resize(mIon,0);
    peie[i].resize(mIon,0);
    pim[i].resize(mIon,0);
    pem[i].resize(mIon);
    for (j=0;j<mIon;j++) pem[i][j].resize(mIl,0);
  }

  // Initialize all non-thermal rates
  for (i=0;i<cell->atom.size();i++)
    for (j=0;j<cell->atom[i].iState.size();j++)
      for (k=0;k<cell->atom[i].iState[j].bState.size();k++) {
	for (l=0;l<cell->atom[i].iState[j].bState[k].upTransBB.size();l++) cell->atom[i].iState[j].bState[k].upTransBB[l]->CNU(0);
	for (l=0;l<cell->atom[i].iState[j].bState[k].upTransBF.size();l++) cell->atom[i].iState[j].bState[k].upTransBF[l]->CNU(0);
      }

  // Set electron fraction
  xelin=cell->xe();

  // For each (SF) atom
  for (i=0;i<config.sfMapping.atom.size();i++) {

    // Set (SF) atom pointer
    atom=&(config.sfMapping.atom[i]);

    // Set number fraction
    ab[atom->sfIndex()]=cell->atom[atom->index()].n()/cell->n();

    // For each (SF) ionization state
    for (j=0;j<atom->iState.size();j++) {

      // Set (SF) ionization state pointer
      iState=&(atom->iState[j]);

      // Set number fraction
      xa[atom->sfIndex()][iState->sfIndex()]=cell->atom[atom->index()].iState[iState->index()].X();

      // If any bound states or bound-bound transitions specified - Set excitation flag
      if (iState->transBB.size()!=0 or iState->bState.size()!=0) eflag[atom->sfIndex()][iState->sfIndex()]=1;
    }
  }

  // Run SF
  sfWrapper(xelin,ab,xa,eflag,ph,peie,pem,pim);

  // For each (SF) atom
  for (i=0;i<config.sfMapping.atom.size();i++) {

    // Set (SF) atom pointer
    atom=&(config.sfMapping.atom[i]);

    // For each (SF) ionization state
    for (j=0;j<atom->iState.size();j++) {

      // Set (SF) ionization state pointer
      iState=&(atom->iState[j]);

      // If ionization flag set
      if (iState->ionization()) {

	// Set lower state pointer
	lowerState=&(cell->atom[atom->index()].iState[iState->index()].bState[0]);

	// Find ground state transition - New!
	// Has to be improved to take into account ground-state multiplets!
	transBF=findGroundStateBF(lowerState);
	if (transBF==NULL) { cout<<"Koz92Interface::update: cell: "<<cell->index()<<", Error: Could not find"<<lowerState->iState()->name()<<" ground state transition\n"; exit(-1); }
      
	// Update non-thermal ionization rate
	transBF->CNU(cell->eRadio()*cell->rho()*pim[atom->sfIndex()][iState->sfIndex()]/transBF->Xi());

	//cout<<"Koz92Interface::update: cell: "<<cell->index()<<", "<<transBF->lowerState->iState()->name()<<":"<<transBF->lowerState->getIndex()<<"->"<<transBF->upperState->iState()->name()<<":"<<transBF->upperState->getIndex()<<", CNU: "<<transBF->CNU()<<", CU: "<<transBF->CU()<<", RU: "<<transBF->RU()<<"\n";

	// Sum total ionization fraction
	ionFrac+=pim[atom->sfIndex()][iState->sfIndex()];
      }

      // For each (SF) bound state
      for (k=0;k<iState->bState.size();k++) {

	// Set (SF) bound state pointer
	bState=&(iState->bState[k]);

	// For each index - New!
	for (l=0;l<bState->index.size();l++) {

	  // Set upper state pointer
	  upperState=&(cell->atom[atom->index()].iState[iState->index()].bState[bState->index[l]]);

	  // Find ground state transition - New!
	  // Has to be improved to take into account ground-state multiplets!
	  transBB=findGroundStateBB(upperState);
	  if (transBB==NULL) { cout<<"Koz92Interface::update: cell: "<<cell->index()<<", Error: Could not find"<<upperState->iState()->name()<<":"<<upperState->getIndex()<<" ground state transition\n"; exit(-1); }
	  
	  // For each SF index
	  for (m=0;m<bState->sfIndex.size();m++) {

	    // Get excitation fraction - New!
	    excFracTrans=pem[atom->sfIndex()][iState->sfIndex()][bState->sfIndex[m]]*bState->distFrac[l];
    
	    // Sum non-thermal excitation rate - Changed!
	    transBB->CNU(transBB->CNU()+cell->eRadio()*cell->rho()*excFracTrans/transBB->Xi());
	    
	    // Sum total excitation fraction - Changed!
	    excFrac+=excFracTrans;
	  }
	    
	  //cout<<"Koz92Interface::update: cell: "<<cell->index()<<", "<<transBB->lowerState->iState()->name()<<":"<<transBB->lowerState->getIndex()<<"->"<<transBB->upperState->iState()->name()<<":"<<transBB->upperState->getIndex()<<", CNU: "<<transBB->CNU()<<", CU: "<<transBB->CU()<<", RU: "<<transBB->RU()<<"\n";
	}
      }

      // For each (SF) bound-bound transition - New!
      for (k=0;k<iState->transBB.size();k++) {

	// Set (SF) bound-bound transition pointer
	sfTransBB=&(iState->transBB[k]);

	// Set lower and upper state pointers
	lowerState=&(cell->atom[atom->index()].iState[iState->index()].bState[sfTransBB->lowerIndex()]);
	upperState=&(cell->atom[atom->index()].iState[iState->index()].bState[sfTransBB->upperIndex()]);

	// Find transition 
	transBB=findTransBB(lowerState,upperState);
	if (transBB==NULL) { cout<<"Koz92Interface::update: cell: "<<cell->index()<<", Error: Could not find"<<lowerState->iState()->name()<<":"<<lowerState->getIndex()<<"->"<<upperState->iState()->name()<<":"<<upperState->getIndex()<<" transition\n"; exit(-1); }
	  
	// Get excitation fraction
	excFracTrans=pem[atom->sfIndex()][iState->sfIndex()][sfTransBB->sfIndex()];
    
	// Sum non-thermal excitation rate
	transBB->CNU(transBB->CNU()+cell->eRadio()*cell->rho()*excFracTrans/transBB->Xi());
	    
	// Sum total excitation fraction
	excFrac+=excFracTrans;
	    
	//cout<<"Koz92Interface::update: cell: "<<cell->index()<<", "<<transBB->lowerState->iState()->name()<<":"<<transBB->lowerState->getIndex()<<"->"<<transBB->upperState->iState()->name()<<":"<<transBB->upperState->getIndex()<<", CNU: "<<transBB->CNU()<<", CU: "<<transBB->CU()<<", RU: "<<transBB->RU()<<"\n";
      }
    }
  }

  // Make sure the fractions adds up to one
  heatFrac=1.0-(ionFrac+excFrac);

  // Update non-thermal heating rate
  cell->nonThermalHeating(cell->eRadio()*cell->rho()*heatFrac);
  cout<<"Koz92Interface::update: cell: "<<cell->index()<<", heatFrac: "<<heatFrac<<", nonThermalHeating: "<<cell->nonThermalHeating()<<"\n";

  // Update non-thermal ionization rate
  cell->nonThermalIonization(cell->eRadio()*cell->rho()*ionFrac);
  cout<<"Koz92Interface::update: cell: "<<cell->index()<<", ionFrac: "<<ionFrac<<", nonThermalIonization: "<<cell->nonThermalIonization()<<"\n";

  // Update non-thermal excitation rate
  cell->nonThermalExcitation(cell->eRadio()*cell->rho()*excFrac);
  cout<<"Koz92Interface::update: cell: "<<cell->index()<<", excFrac: "<<excFrac<<". nonThermalExcitation: "<<cell->nonThermalExcitation()<<"\n";
}

// Initiate
void Koz92Interface::init(Cell* cell) {

  cout<<"Koz92Interface::init: cell: "<<cell->index()<<", Enter\n";
  
  // Set SF mapping (atom and ionization state) indices
  // Note : Only one global mapping (therefore critical). Maybe we should have one per cell instead?
  #pragma omp critical
  config.sfMapping.init(cell);
  
  cout<<"Koz92Interface::init: cell: "<<cell->index()<<", Return\n";
}

// Constructors
Koz92SolverConfig::Koz92SolverConfig() : method(k92None), interface(""), fitter("") {};
Koz92SolverConfig::Koz92SolverConfig(const Koz92Method& _method, const string& _interface, const string& _fitter) : method(_method), interface(_interface), fitter(_fitter) {};
Koz92SolverConfig::Koz92SolverConfig(const string& dataFile) { load(dataFile); };

// Load data
void Koz92SolverConfig::load(const YAML::Node& data) {

  // Load method
  if (data["method"]) {
    if (data["method"].as<string>().compare("fitter")==0) method=k92Fitter;
    else if (data["method"].as<string>().compare("interface")==0) method=k92Interface;
    else method=k92None;
  }

  // Load interface configuration file
  if (data["interface"]) interface=data["interface"].as<string>();

  // Load fitter configuration file
  //if (data["fitter"]) fitter=data["fitter"].as<string>();
  if (data["fitter"]) fitter=data["fitter"].as<string>(); // Changed!
}

void Koz92SolverConfig::load(const string& dataFile) {
  YAML::Node data;

  // Load data from file
  data=YAML::LoadFile(dataFile);

  // Load data
  if (data["koz92Solver"]) load(data["koz92Solver"]);
}

// Log data
void Koz92SolverConfig::log() {};

// Constructors
Koz92Solver::Koz92Solver(Koz92SolverConfig _config) : config(_config), interface(), fitter() {
  if (config.interface.compare("")!=0) interface.load(config.interface);
  if (config.fitter.compare("")!=0) fitter.load(config.fitter);  
}
Koz92Solver::Koz92Solver(const string& dataFile) : interface(), fitter() {
  config.load(dataFile);
  if (config.interface.compare("")!=0) interface.load(config.interface);
  if (config.fitter.compare("")!=0) fitter.load(config.fitter);  
}

// Update
void Koz92Solver::update(Cell* cell, bool initiate) {
  double Z,C,X,excFracTrans,excFracTot=0,ionFracTrans,ionFracTot=0;
  TransBB* transBB;
  TransBF* transBF;
  int i,j,k,l;

  cout<<"Koz92Solver::update: cell: "<<cell->index()<<", eRadio (per volume): "<<cell->eRadio()*cell->rho()<<", xe: "<<cell->xe()<<"\n";

  // If requested - Initiate
  if (initiate) init(cell);

  // Set all non-thermal rates to zero
  cell->nonThermalHeating(0);
  cell->nonThermalIonization(0);
  cell->nonThermalExcitation(0);
  for (i=0;i<cell->atom.size();i++)
    for (j=0;j<cell->atom[i].iState.size();j++)
      for (k=0;k<cell->atom[i].iState[j].bState.size();k++) {
	for (l=0;l<cell->atom[i].iState[j].bState[k].upTransBB.size();l++) cell->atom[i].iState[j].bState[k].upTransBB[l]->CNU(0);
	for (l=0;l<cell->atom[i].iState[j].bState[k].upTransBF.size();l++) cell->atom[i].iState[j].bState[k].upTransBF[l]->CNU(0);
      }

  // Hack to avoid zero fractions !!! Fix in a better way !!!
  cell->eRadio(MAX(cell->eRadio(),1e-99));

  // Update using configured method
  if (config.method==k92Fitter) fitter.update(cell,initiate);
  else if (config.method==k92Interface) interface.update(cell,initiate);
  else cell->nonThermalHeating(cell->eRadio()*cell->rho());
}

// Initiate
void Koz92Solver::init(Cell* cell) {

  // Hack - Set eRadio to zero
  //cell->eRadio(0); // Changed!

  // If interface method - Initiate interface
  if (config.method==k92Interface) interface.init(cell);
}

// Load
void Koz92Solver::load(const string& dataFile) {
  // Load configuration
  config.load(dataFile);
  // If configuration file provided - Load fitter
  if (config.interface.compare("")!=0) interface.load(config.interface);
  // If configuration file provided - Load interface
  if (config.fitter.compare("")!=0) fitter.load(config.fitter);
}
