#ifndef _NTSOLVER_
#define _NTSOLVER_

#include <vector>
#include "mc.h"

using namespace std;
using namespace mc;

namespace ntsolver {

  // Constants for SF C++ wrapper
  const int mIe=15;
  const int mIon=5;
  const int mIl=1310;

  // SF C++ wrapper
  void sfWrapper(const double& xelin, const vector<double>& ab, const vector<vector<double> >& xa, const vector<vector<int> >& eflag, double& ph, vector<vector<double> >& peie, vector<vector<vector<double> > >& pem, vector<vector<double> >& pim);

  // Get SF atom index
  double sfIndexAtom(double Z);
  
  // Get SF ionization state index  
  double sfIndexIState(double C);

  // Find bound-free ground-state transition
  TransBF* findGroundStateBF(BState* bState);

  // Find bound-bound ground-state transition
  TransBB* findGroundStateBB(BState* bState);

  // Find bound-bound transition
  TransBB* findTransBB(BState* lowerState, BState* upperState);

  // Parameters for fitting formula (see Kozma 1992)
  class FitParam {
  public :
    // Constructors
    FitParam() {};
    FitParam(const double& _C, const double& a1, const double& a2, const double& a3, const double& a4);
    // Load data
    void load(const YAML::Node& data);
    double C;
    double a[4];
  };

  // Ionization table entry
  class IonEntry {
  public :
    // Constructors
    IonEntry() {};
    IonEntry(const double& _Z, const double& _C, const FitParam& _param);
    // Load data
    void load(const YAML::Node& data);
    double Z;
    double C;
    FitParam param;
  };

  // Ionization table
  class IonTable : public vector<IonEntry> {
  public :
    // Load data
    void load(const YAML::Node& data);
    // Find entry in table
    FitParam* find(const double& Z, const double& C);
  };

  // Excitation table entry
  class ExcEntry {
  public :
    // Constructors
    ExcEntry() {};
    ExcEntry(const double& _Z, const double& _C, const int& _upperStateIndex, const double& _fraction);
    // Load data
    void load(const YAML::Node& data);
    double Z;
    double C;
    int upperStateIndex;
    double fraction;
  };

  // Excitation table
  class ExcTable : public vector<ExcEntry> {
  public :
    // Load data
    void load(const YAML::Node& data);
    // Find entry in table
    double* find(const double& Z, const double& C, const int& upperStateIndex);
  };

  // Non-thermal heating, ionization and excitation following Kozma 1992
  class Koz92Fitter {
  private :
    FitParam heatParam;
    FitParam excParam;
    FitParam ionParam;
    ExcTable excTable;
    IonTable ionTable;
    // Apply fitting formula (see Kozma 1992)
    double fit(const double& xe, const FitParam& param) const;
    // Get effective ionization potential (see Kozma 1992)
    double XiEff(const double& Z, const double& C, const double& xe);
  public :
    // Load
    void load(const string& dataFile);
    // Get fractional energy going into heating
    double heatFrac(const double& xe) const;
    // Get fractional energy going into ionization
    double ionFrac(const double& xe) const;
    // Get fractional energy going into ionization (for a specified ion)
    double ionFrac(const double& Z, const double& C, const double& Xi, const double& X, const double& xe);
    // Get fractional energy going into excitation
    double excFrac(const double& xe) const;
    // Get fractional energy going into excitation (to a specified state for a specified ion)
    double excFrac(const double& Z, const double& C, const double& upperStateIndex, const double& X, const double& xe);
    // Update
    void update(Cell* cell, bool initiate);
  };

  const int noIndex=-1;

  // SF bound-bound transition
  class SFTransBB {
    int pLowerIndex;
    int pUpperIndex;
    int pSFIndex;
  public :
    // Constructors
    SFTransBB();
    // Load data
    void load(const YAML::Node& data);
    // Log data
    void log();
    // Get lower index
    double lowerIndex() const;
    // Get upper index
    double upperIndex() const;
    // Get SF index
    double sfIndex() const;
  };

  // SF bound state
  class SFBState {
    //int pIndex;
  public :
    vector<int> index;
    vector<int> sfIndex;
    vector<double> distFrac;
    // Constructors
    SFBState();
    // Load data
    void load(const YAML::Node& data);
    // Log data
    void log();
    // Get index
    //double index() const;
  };

  // SF ionization state
  class SFIState {
    double pC;
    int pIndex;
    int pSFIndex;
    bool pIonization;
  public :
    // SF bound state vector
    vector<SFBState> bState;
    // SF bound-bound transition vector
    vector<SFTransBB> transBB;
    // Constructors
    SFIState();
    // Load data
    void load(const YAML::Node& data);
    // Log data
    void log();
    // Get C
    double C() const;
    // Get index
    int index() const;
    // Set index
    void index(const int& _index);
    // Get SF index
    int sfIndex() const;
    // Get ionization flag
    bool ionization() const;
  };

  // SF atom
  class SFAtom {
    double pZ;
    int pIndex;
    int pSFIndex;
  public :
    // SF ionization state vector
    vector<SFIState> iState;
    // Constructors
    SFAtom();
    // Load data
    void load(const YAML::Node& data);
    // Log data
    void log();
    // Get Z
    double Z() const;
    // Get index
    int index() const;    
    // Set index
    void index(const int& _index);
    // Get SF index
    int sfIndex() const;
  };

  // SF mapping
  class SFMapping {
  public :
    // FS atom vector
    vector<SFAtom> atom;
    // Constructors
    SFMapping();
    // Load data
    void load(const YAML::Node& data);
    // Log data
    void log();
    // Initiate
    void init(Cell* cell);
  };

  // Kozma 1992 method
  enum Koz92Method { k92None, k92Fitter, k92Interface };

  // Kozma 1992 solver configuration
  class Koz92InterfaceConfig {
  public :
    // FS mapping
    SFMapping sfMapping;
    // Constructors
    Koz92InterfaceConfig();
    Koz92InterfaceConfig(const string& dataFile);
    // Load data
    void load(const YAML::Node& data);
    void load(const string& dataFile);
    // Log data
    void log();
  };

  // Non-thermal heating, ionization and excitation following Kozma 1992
  class Koz92Interface {
  private :
  public :
    Koz92InterfaceConfig config;
    // Load
    void load(const string& dataFile);
    // Update
    void update(Cell* cell, bool initiate=false);
    // Initiate
    void init(Cell* cell);
  };

  class NonThermalSolver {
  public :
    // Update
    virtual void update(Cell* cell, bool initiate) {};
    // Initiate - New!
    virtual void init(Cell* cell) {};
    // Load - New!
    virtual void load(const string& dataFile) {};
  };

  // Kozma 1992 solver configuration
  class Koz92SolverConfig {
  public :
    // Method
    Koz92Method method;
    // Interface configuration file
    string interface;
    // Fitter configuration file
    string fitter;
    // Constructors
    Koz92SolverConfig();
    Koz92SolverConfig(const Koz92Method& _method, const string& _interface, const string& _fitter);
    Koz92SolverConfig(const string& dataFile);
    // Load data
    void load(const YAML::Node& data);
    void load(const string& dataFile);
    // Log data
    void log();
  };

  class Koz92Solver : public NonThermalSolver {
  private :
    Koz92Interface interface;
    Koz92Fitter fitter;
  public :
    Koz92SolverConfig config;
    // Constructors
    Koz92Solver(Koz92SolverConfig _config);
    Koz92Solver(const string& dataFile);
    // Update
    virtual void update(Cell* cell, bool initiate=false);
    // Initiate - New!
    virtual void init(Cell* cell);
    // Load
    virtual void load(const string& dataFile); // Changed!
  };
}

#endif
