#include "mc.h"
#include "ntsolver.h"
#include "msolver.h"
#include "eos.h"

using namespace mc;
using namespace ntsolver;
using namespace msolver;
using namespace eos;

// Constructor
Path::Path(int i) { push_back(i); }

// Concatenate operator
Path Path::operator << (int& i) const {
  Path _path=(*this);
  _path.push_back(i);
  return _path;
}

// Initialize N/dN for cell
void msolver::initN(Cell* cell, vector<vector<vector<double> > >& N) {
  int i,j;
  N.resize(cell->atom.size());
  for (i=0;i<cell->atom.size();i++) {
    N[i].resize(cell->atom[i].iState.size());
    for (j=0;j<cell->atom[i].iState.size();j++) N[i][j].resize(cell->atom[i].iState[j].bState.size(),0);
  }
}

// Initialize N/dN for atom
void msolver::initN(const Atom& atom, vector<vector<double> >& N) {
  int i;
  N.resize(atom.iState.size());
  for (i=0;i<atom.iState.size();i++) N[i].resize(atom.iState[i].bState.size(),0);
}

// Initialize XIState for cell
void msolver::initXIState(Cell* cell, vector<vector<double> >& XIState) {
  int i,j;
  XIState.resize(cell->atom.size());
  for (i=0;i<cell->atom.size();i++) XIState[i].resize(cell->atom[i].iState.size(),0);
}

// Initialize XBState for cell
void msolver::initXBState(Cell* cell, vector<vector<vector<double> > >& XBState) {
  int i,j;
  XBState.resize(cell->atom.size());
  for (i=0;i<cell->atom.size();i++) {
    XBState[i].resize(cell->atom[i].iState.size());
    for (j=0;j<cell->atom[i].iState.size();j++) XBState[i][j].resize(cell->atom[i].iState[j].bState.size(),0);
  }
}

// Load data
void AtomConfigEntry::load(const YAML::Node& data) {
  int i;

  // Resize atom name vector
  atomName.resize(0);

  // Load atom names
  if (data["atoms"]) {
    if (data["atoms"].Type()==YAML::NodeType::Scalar) {
      atomName.push_back(data["atoms"].as<string>());
    }
    else {
      for (i=0;i<data["atoms"].size();i++) {
	atomName.push_back(data["atoms"][i].as<string>());
      }
    }
  }
}

// Log data
void AtomConfigEntry::log() {
  int i;
  cout<<" atoms:";
  if (atomName.size()>1) cout<<" [";
  for (i=0;i<atomName.size();i++) { 
    cout<<" "<<atomName[i];
    if (i<atomName.size()-1) cout<<",";
  }
  if (atomName.size()>1) cout<<" ]";
}
  
// Get entry
template<class AtomConfigEntryLike> AtomConfigEntryLike AtomConfig<AtomConfigEntryLike>::getEntry(const string& _atomName) {
  AtomConfigEntryLike defaultEntry;
  int i,j;

  // For each entry
  for (i=0;i<entry.size();i++)
    // For each atom
    for (j=0;j<entry[i].atomName.size();j++) {
      // If match return entry
      if (entry[i].atomName[j].compare(_atomName)==0 or entry[i].atomName[j].compare("all")==0) return entry[i];
      // Get default entry (if provided)
      if (entry[i].atomName[j].compare("other")==0) defaultEntry=entry[i];
    }

  // Return default entry
  return defaultEntry;
}

// Get entry
template<class AtomConfigEntryLike> AtomConfigEntryLike AtomConfig<AtomConfigEntryLike>::getEntry(Atom& atom) {
  return getEntry(atom.name());
}

// Get entry
template<class AtomConfigEntryLike> AtomConfigEntryLike AtomConfig<AtomConfigEntryLike>::getEntry(IState& iState) {
  return getEntry(iState.atom()->name());
}

// Load data
template<class AtomConfigEntryLike> void AtomConfig<AtomConfigEntryLike>::load(const YAML::Node& data) {
  AtomConfigEntryLike _entry;
  int i;

  // For each entry
  for (i=0;i<data.size();i++) {
    // Load entry
    _entry.load(data[i]);
    // Add to list
    entry.push_back(_entry);
  }
}

// Log data
template<class AtomConfigEntryLike> void AtomConfig<AtomConfigEntryLike>::log() {
  int i;
  for (i=0;i<entry.size();i++) entry[i].log();
}

// Load solver method data
SolverMethod msolver::loadSolverMethod(const YAML::Node& data) {
  if (data.as<string>().compare("nlte")==0) return sNLTE;
  else if (data.as<string>().compare("nebular")==0) return sNebular;
  else if (data.as<string>().compare("lte")==0) return sLTE;
  else if (data.as<string>().compare("comb")==0) return sComb;
  else return sNone;
}

// Constructor
LTEConvConfig::LTEConvConfig() : neTol(1e-3), TMin(noT), TMax(noT) {};

// Load data
void LTEConvConfig::load(const YAML::Node& data) {
  
  // Load electron number density tolerence
  if (data["neTol"]) neTol=data["neTol"].as<double>();

  // Load minimum temperature
  if (data["TMin"]) TMin=data["TMin"].as<double>();

  // Load maximum temperature
  if (data["TMax"]) TMax=data["TMax"].as<double>();
}

// Log data
void LTEConvConfig::log() {
  cout<<"LTEConvConfig::log: neTol: "<<neTol<<"\n";
}

// Constructor
LTETempConfig::LTETempConfig() : type(lteBB), typeIon(lteGas), typeBound(lteGas), update(true) {};

// Load data
void LTETempConfig::load(const YAML::Node& data) {
  
  // Load temperature type (used in update)
  if (data["type"]) {
    if (data["type"].as<string>().compare("DBB")==0) type=lteDBB;
    else type=lteBB;
  }

  // Load temperature type (used in bound population calculation)
  if (data["bound"]) {
    if (data["bound"].as<string>().compare("DBB")==0) typeBound=lteDBB;
    else if (data["bound"].as<string>().compare("BB")==0) typeBound=lteBB;
    else typeBound=lteGas;
  }

  // Load temperature type  (used in ion population calculation)
  if (data["ion"]) {
    if (data["ion"].as<string>().compare("DBB")==0) typeIon=lteDBB;
    else if (data["ion"].as<string>().compare("BB")==0) typeIon=lteBB;
    else typeIon=lteBB;
  }

  // Load update flag - New! - Testing!
  if (data["update"]) update=(data["update"].as<string>().compare("yes")==0);
}

// Log data
void LTETempConfig::log() {
  cout<<"LTETempConfig::log: type: "<<type<<", typeBound: "<<typeBound<<", typeIon: "<<typeIon<<"\n";
}

// Constructors
LTESolverConfig::LTESolverConfig() : convergence(), temperature(), testSwitch1(false) {};
LTESolverConfig::LTESolverConfig(const LTEConvConfig& _convergence, const LTETempConfig& _temperature, const bool& _testSwitch1) : convergence(_convergence), temperature(_temperature), testSwitch1(_testSwitch1) {};
LTESolverConfig::LTESolverConfig(const string& dataFile) { load(dataFile); }

// Load data
void LTESolverConfig::load(const YAML::Node& data) {
										    
  // Load convergence config
  if (data["convergence"]) convergence.load(data["convergence"]);  

  // Load temperature config
  if (data["temperature"]) temperature.load(data["temperature"]);  

  // Load test swith 1
  if (data["testSwitch1"]) testSwitch1=(data["testSwitch1"].as<string>().compare("yes")==0);
}

// Log data
void LTESolverConfig::log() {
  convergence.log();
  temperature.log();
}

// Load data (from file)
void LTESolverConfig::load(const string& dataFile) {
  YAML::Node data;

  // Load data from file
  data=YAML::LoadFile(dataFile);

  // Load data
  if (data["lteSolver"]) load(data["lteSolver"]);

  log();
}

// Constructors
LTESolver::LTESolver(LTESolverConfig _config) : config(_config), pLog(true) { }
LTESolver::LTESolver(const string& dataFile) : pLog(true) { config.load(dataFile); }

// Create copy
MStateSolver* LTESolver::copy() {
  return new LTESolver(*this);
}

// Solve for ionization state of atom
void LTESolver::solveIonState(Atom& atom, const double& ne, const double& T, const double& TR, const double& W, vector<double>& X) {
  int i;
  double n2n1,xsum;

  // If non-zero mass-fraction
  if (atom.X()!=0) {
    // Initialize number-fraction
    X[0]=1;
    xsum=X[0];
    // For each ionization state (except last)
    for (i=0;i<atom.iState.size()-1;i++) {
      // Calculate number-fraction for next ionization state
      n2n1=xIon(atom.iState[i],atom.iState[i+1],T,TR,W)/ne;
      X[i+1]=X[i]*n2n1;
      // Sum total number-fraction
      xsum+=X[i+1];
    }
    // Normalize number-fractions for each ionization state
    for (i=0;i<atom.iState.size();i++) X[i]=X[i]/xsum;
  }

  return;
}

// Update ionization state for atom
void LTESolver::updateIonState(Atom& atom, const double& ne, const double& T, const double& TR, const double& W) {
  vector<double> X(atom.iState.size());
  int i;

  // Solve for ionization state
  solveIonState(atom,ne,T,TR,W,X);

  // Update ionization state
  for (i=0;i<atom.iState.size();i++) atom.iState[i].X(X[i]);

  return;
}

// Update ionization state for cell (decoupled version)
void LTESolver::updateIonStateDec(Cell* cell) {
  double xsum,csum,n2n1,n1Min,n1Max,n2Min,n2Max,C,small;
  int i,j;
  // Initialize charge sum
  csum=0;
  // For each atom
  for (i=0;i<cell->atom.size();i++) {
    // Exclude atoms with zero abundance
    if (cell->atom[i].X()!=0) {
      small=cell->atom[i].n()*1e-6;
      cell->atom[i].iState[0].X(1);
      xsum=cell->atom[i].iState[0].X();
      // For each ionization state
      for (j=0;j<cell->atom[i].iState.size()-1;j++) {
	// Calculate n1 assuming n2=ne
	C=xIon(cell->atom[i].iState[j],cell->atom[i].iState[j+1],cell->T(),cell->TR(),cell->W());
	root_2p(1,-(2*cell->atom[i].n()+C),pow(cell->atom[i].n(),2),n1Min,n1Max);
	// Calculate n2
	n2Min=cell->atom[i].n()-n1Min;
	n2Max=cell->atom[i].n()-n1Max;
	// Calculate n1/n2
	if (n1Min>=0 && n2Min>=0) n2n1=(n2Min+small)/(n1Min+small);
	else n2n1=(n2Max+small)/(n1Max+small);
	// Update number fraction
	cell->atom[i].iState[j+1].X(cell->atom[i].iState[j].X()*n2n1);
	// Sum total number fraction
	xsum+=cell->atom[i].iState[j+1].X();
      }
      // For each ionization state
      for (j=0;j<cell->atom[i].iState.size();j++) {
	// Normalize number fraction
	cell->atom[i].iState[j].X(cell->atom[i].iState[j].X()/xsum);
	// Sum total charge
	csum+=cell->atom[i].iState[j].n()*fabs(cell->atom[i].iState[j].C());
      }
    }
  }
  // Update electrin number density
  cell->ne(csum);

  return;
}

// Update ionization state for cell
void LTESolver::updateIonState(Cell* cell) {
  double csum,neOld,neTol=config.convergence.neTol,damp=0.7;
  int i,j,n,nIter=1000;

  // Testing
  if (config.testSwitch1) {
    //if (log()) cout<<"LTESolver::updateIonState: testSwitch1\n";
    csum=0;
    for (i=0;i<cell->atom.size();i++) {
      for (j=0;j<cell->atom[i].iState.size()-1;j++) cell->atom[i].iState[j].X(1e-15);
      cell->atom[i].iState[j].X(1.0);
      csum+=cell->atom[i].n()*fabs(cell->atom[i].iState[j].C());
    }
    cell->ne(csum);
    return;
  }

  // Make initial guess using decoupled version
  updateIonStateDec(cell);

  // Iteration loop
  for (n=0;n<nIter;n++) {
    // Initialize charge sum
    csum=0;
    // For each atom
    for (i=0;i<cell->atom.size();i++) {
      // Update ionization state
      updateIonState(cell->atom[i],cell->ne(),cell->T(),cell->TR(),cell->W());
      // Sum charge
      for (j=0;j<cell->atom[i].iState.size();j++) csum+=cell->atom[i].iState[j].n()*fabs(cell->atom[i].iState[j].C());
    }
    // Store electron number density
    neOld=cell->ne();
    // Update electron number density
    cell->ne(cell->ne()+damp*(csum-cell->ne()));
    // If maximum error less than tolerance break
    if (fabs((cell->ne()-neOld)/cell->ne())<=neTol) break;
  }
  if (n==nIter and log()) cout<<"LTESolver::updateIonState: cell: "<<cell->index()<<", Electron density did not converge, ne="<<cell->ne()<<", neOld="<<neOld<<", fabs((ne-neOld)/ne)="<<fabs((cell->ne()-neOld)/cell->ne())<<", neTol="<<neTol<<"\n";
  else if (log()) cout<<"LTESolver::updateIonState: cell: "<<cell->index()<<", ne: "<<cell->ne()<<", Electron density converged in "<<n+1<<" iterations\n";

  return;
}

// Solve for bound state of ion
void LTESolver::solveBoundState(IState& iState, const double& T, const double& TR, const double& W, vector<double>& X) {
  int i;
  double xSum;
 
  // If at least one bound state
  if (iState.bState.size()>=1) {
    // Initialize number fraction
    X[0]=1;
    xSum=X[0];
    // For each bound state
    for (i=1;i<iState.bState.size();i++) {
      // Calculate number-fraction
      X[i]=xBound(iState.bState[0],iState.bState[i],T,TR,W);
      // Sum number-fraction
      xSum=xSum+X[i];
    }
    // Normalize number-fraction for each bound state
    for (i=0;i<iState.bState.size();i++) X[i]=X[i]/xSum;
  }

  // Testing!
  //if (iState.atom()->name().compare("Ni")==0 and iState.atom()->cell()->index()==24) { xSum=0; for (i=0;i<iState.bState.size();i++); xSum+=X[i]; cout<<"LTESolver::solveBoundState: cell: "<<iState.atom()->cell()->index()<<", "<<iState.name()<<", xSum: "<<xSum<<"\n"; }

  return;
}

// Update bound state for ion
void LTESolver::updateBoundState(IState& iState, const double& T, const double& TR, const double& W) {
  vector<double> X(iState.bState.size());
  int i;

  // Solve for bound state
  solveBoundState(iState,T,TR,W,X);

  // Update bound state
  for (i=0;i<iState.bState.size();i++) iState.bState[i].X(X[i]);

  return;
}

// Update bound state for atom
void LTESolver::updateBoundState(Atom& atom, const double& T, const double& TR, const double& W) {
  int i;
 
  // Update bound state for each ionization state
  for (i=0;i<atom.iState.size();i++) updateBoundState(atom.iState[i],T,TR,W);

  return;
}

// Update bound state for cell
void LTESolver::updateBoundState(Cell* cell) {
  double small=1e-100;
  int i,j,k;
  
  // Update bound state for each atom
  for (i=0;i<cell->atom.size();i++) updateBoundState(cell->atom[i],cell->T(),cell->TR(),cell->W());

  // Testing! - Prevent zero populations
  for (i=0;i<cell->atom.size();i++) {
    for (j=0;j<cell->atom[i].iState.size();j++) {
      cell->atom[i].iState[j].X(MAX(cell->atom[i].iState[j].X(),small));
      for (k=0;k<cell->atom[i].iState[j].bState.size();k++) cell->atom[i].iState[j].bState[k].X(MAX(cell->atom[i].iState[j].bState[k].X(),small));
    }
  }

  return;
}

// Enforce minimum (bound-state) number density
/*void LTESolver::enforceNMin(Atom& atom) {
  double bStateNMin=config.convergence.nMin*numeric_limits<double>::epsilon()*atom.n();
  double iStateNMin;
  int i,j;

  for (i=0;i<atom.iState.size();i++) {
    iStateNMin=bStateNMin*atom.iState[i].bState.size();
    if (atom.iState[i].n()<iStateNMin) atom.iState[i]->X(iStateNMin/atom.n());
    for (j=0;j<atom.iState[i].bState.size();j++) {
      if (atom.iState[i].bState[j].n()<bStateNMin) atom.iState[i].bState[j].X(bStateNMin/iState->n());
    }
  }

  // Normalize!!!
  }*/

// Update temperature for cell
void LTESolver::updateTemperature(Cell* cell) {
  double T;

  // Update temperature accordning to configuration
  //if (config.temperature.type==lteDBB) cell->T(cell->TR());
  //else cell->T(cell->TR()*pow(cell->W(),1.0/4.0));
  // Changed! - Testing!
  if (config.temperature.type==lteDBB) T=cell->TR();
  else T=cell->TR()*pow(cell->W(),1.0/4.0);
  if (config.convergence.TMin!=noT) T=MAX(T,config.convergence.TMin);
  if (config.convergence.TMax!=noT) T=MIN(T,config.convergence.TMax);
  cell->T(T);
    
  if (log()) cout<<"LTESolver::updateTemperature: cell: "<<cell->index()<<", T="<<cell->T()<<"\n";
}

// Update partition function for ion
void LTESolver::updateU(IState& iState, const double& T) {
  int i;
  //if (log()) cout<<"LTESolver::updateU: "<<iState.name()<<"\n";
  vector<double> E(iState.bState.size(),0),g(iState.bState.size(),0);
  for (i=0;i<iState.bState.size();i++) {
    //if (log()) cout<<"LTESolver::updateU: "<<iState.name()<<":"<<iState.bState[i].getIndex()<<"\n";
    E[i]=iState.bState[i].E();
    g[i]=iState.bState[i].g();
  }
  if (i!=0) iState.U(ULTE(T,E,g));
  else iState.U(1);
}

// Update partition function for atom
void LTESolver::updateU(Atom& atom, const double& T) {
  int i;
  //if (log()) cout<<"LTESolver::updateU: "<<atom.name()<<"\n";
  // Update partition function for each ionization state
  for (i=0;i<atom.iState.size();i++) updateU(atom.iState[i],T);
}

// Update partition function for cell
void LTESolver::updateU(Cell* cell) {
  int i;
  //if (log()) cout<<"LTESolver::updateU: cell: "<<cell->index()<<", T: "<<cell->T()<<"\n";
  // Update partition function for each atom
  for (i=0;i<cell->atom.size();i++) updateU(cell->atom[i],cell->T());
}

// Update matter state
//void LTESolver::update(Cell* cell, bool initiate) {
void LTESolver::update(Cell* cell, bool initiate, bool initiateState) {

  // If configured - Update temperature - Changed! - Testing!
  if (config.temperature.update) updateTemperature(cell);
  // Update partition function
  updateU(cell);
  // Update ionization state
  updateIonState(cell);
  // Update bound state
  updateBoundState(cell);
}

// Return the bound state number fraction (relative the ground state)
double LTESolver::xBound(const BState& groundState, const BState& upperState, const double& T, const double& TR, const double& W) {
  double X,TChoice; // Changed!

  // Choose temperatue type - New!
  if (config.temperature.typeBound==lteDBB) TChoice=TR;
  else if (config.temperature.typeBound==lteBB) TChoice=TR*pow(W,1.0/4.0);
  else TChoice=T;

  // Return the value as determined by LTE
  X=nBoundLTE(1,TChoice,groundState.E(),upperState.E(),groundState.g(),upperState.g()); // Changed!
  return X;
}

// Return the product of the ionization state number fraction and the electron density 
double LTESolver::xIon(const IState& iState1, const IState& iState2, const double& T, const double& TR, const double& W) {
  double Xne,TChoice; // Changed!

  // Choose temperature type - New!
  // Only work for TChoice=T yet as U calculated using T
  if (config.temperature.typeIon==lteDBB) TChoice=TR;
  else if (config.temperature.typeIon==lteBB) TChoice=TR*pow(W,1.0/4.0);
  else TChoice=T;

  // Return the value as determined by LTE
  Xne=nIonLTE(1,1,TChoice,iState1.Xi0(),iState1.U(),iState2.U()); // Changed!
  return Xne;
}

// Update matter state assuming LTE
void LTESolver::updateLTE(Cell* cell) {

  // Update partition function
  updateU(cell);
  // Update ionization state
  updateIonState(cell);
  // Update bound state
  updateBoundState(cell);
}

// Get logging flag
bool LTESolver::log() const {

  // Return logging flag
  return pLog;
}

// Set logging flag
void LTESolver::log(const bool& _log) {

  // Set logging flag
  pLog=_log;
}

// Constructor
NebPopConfig::NebPopConfig() : ionization(ionNebular), excitation(excNebular), EIonEdge(0), ionExcited(false) {};

// Load data
void NebPopConfig::load(const YAML::Node& data) {

  // Load ionization
  if (data["ionization"]) if (data["ionization"].as<string>().compare("lte")==0) ionization=ionLTE;
  
  // Load excitation
  if (data["excitation"]) if (data["excitation"].as<string>().compare("lte")==0) excitation=excLTE;
  
  // Load ionization edge energy
  if (data["EIonEdge"]) EIonEdge=data["EIonEdge"].as<double>();
  
  // Load ion excitated
  if (data["ionExcited"]) ionExcited=(data["ionExcited"].as<string>().compare("no")!=0);
}

// Log data
void NebPopConfig::log() {
  cout<<"NebPopConfig::log: ionization: "<<ionization<<", excitation: "<<excitation<<", EIonEdge: "<<EIonEdge<<", ionExcited: "<<ionExcited<<"\n";
}

// Constructor
NebTempConfig::NebTempConfig() : method(sNebular) {};

// Load data
void NebTempConfig::load(const YAML::Node& data) {

  // Load solver method
  if (data["method"]) method=loadSolverMethod(data["method"]);
};

// Log data
void NebTempConfig::log() {
  cout<<"NebTempConfig: method: "<<method<<"\n";  
}

// Constructor
NebConvConfig::NebConvConfig() : neTol(1e-3), TMin(noT), TMax(noT) {};

// Load data
void NebConvConfig::load(const YAML::Node& data) {
  
  // Load electron number density tolerence
  if (data["neTol"]) neTol=data["neTol"].as<double>();

  // Load minimum temperature
  if (data["TMin"]) TMin=data["TMin"].as<double>();

  // Load maximum temperature
  if (data["TMax"]) TMax=data["TMax"].as<double>();
}

// Log data
void NebConvConfig::log() {
  cout<<"NebConvConfig::log: neTol: "<<neTol<<"\n";
}

// Constructors
NebSolverConfig::NebSolverConfig() : population(), temperature(), convergence() {};
NebSolverConfig::NebSolverConfig(const NebPopConfig& _population, const NebTempConfig& _temperature, const NebConvConfig& _convergence) : population(_population), temperature(_temperature), convergence(_convergence) {};
NebSolverConfig::NebSolverConfig(const string& dataFile) { load(dataFile); }

// Load data
void NebSolverConfig::load(const YAML::Node& data) {

  // Load population config
  if (data["population"]) population.load(data["population"]);

  // Load temperature config
  if (data["temperature"]) temperature.load(data["temperature"]);
										    
  // Load convergence config
  if (data["convergence"]) convergence.load(data["convergence"]);  
}

// Log data
void NebSolverConfig::log() {
  population.log();
  temperature.log();
  convergence.log();
}

// Load data (from file)
void NebSolverConfig::load(const string& dataFile) {
  YAML::Node data;

  // Load data from file
  data=YAML::LoadFile(dataFile);

  // Load data
  if (data["nebSolver"]) load(data["nebSolver"]);

  log();
}

// Constructors
NebSolver::NebSolver(LTESolverConfig _config, NebSolverConfig _nebConfig) : nebConfig(_nebConfig), LTESolver(_config) { } // Changed!
NebSolver::NebSolver(const string& dataFile) : LTESolver(dataFile) { nebConfig.load(dataFile); } // Changed!

// Create copy
MStateSolver* NebSolver::copy() {
  return new NebSolver(*this);
}

// Return the bound state number fraction (relative the ground sate)
double NebSolver::xBound(const BState& groundState, const BState& upperState, const double& T, const double& TR, const double& W) {
  double AMetastable=10.0;
  double X;

  if (nebConfig.population.excitation==excLTE) X=LTESolver::xBound(groundState,upperState,T,TR,W);
  else {
    // If the upper state is metastable
    if (upperState.A()<=AMetastable) X=nBoundLTE(1,TR,groundState.E(),upperState.E(),groundState.g(),upperState.g());
    // Else
    else X=W*nBoundLTE(1,TR,groundState.E(),upperState.E(),groundState.g(),upperState.g());
  }  
    
  return X;
}

// Return the product of the ionization state fraction and the electron density
double NebSolver::xIon(const IState& iState1, const IState& iState2, const double& T, const double& TR, const double& W) {
  double X,zeta;

  if (nebConfig.population.ionization==ionLTE)  return LTESolver::xIon(iState1,iState2,T,TR,W);
  else {
    // Make this configurable !!!
    if (iState1.atom()->Z()<20) zeta=0.4;
    else {
      if (iState1.atom()->Z()==20) {
	if (TR>=5000) zeta=0.01;
	else zeta=0.005; 
      }
      else zeta=0.03;
    }

    return nIonMaz93(1,1,T,TR,iState1.Xi0(),iState1.U(),iState2.U(),W,zeta,nebConfig.population.EIonEdge,W,nebConfig.population.ionExcited);
  }
}

// Equation (10) in Mazzali & Lucy 93.
double NebSolver::nIonGround (const double& n1, const double& ne, const double& Te, const double& TR, const double& EIon, const double& g1, const double& g2, const double& W, const double& zeta) {
  double n2=(n1/ne)*zeta*W*pow(Te/TR,0.5)*nIonLTE(1,1,TR,EIon,g1,g2);
  return n2;
}

// Second term in Equation (11) in Mazzali & Lucy 93.
double NebSolver::nIonExcited (const double& n1, const double& ne, const double& Te, const double& TR, const double& EIon, const double& g1, const double& g2, const double& W, const double& zeta) {
  double n2=(n1/ne)*(1-zeta)*pow(W,2)*pow(Te/TR,0.5)*nIonLTE(1,1,TR,EIon,g1,g2);
  return n2;
}

// Equation (14) in Mazzali & Lucy 93.
double NebSolver::nIonMaz93 (const double& n1, const double& ne, const double& Te, const double& TR, const double& EIon, const double& g1, const double& g2, const double& W, const double& zeta, const double& EIonEdge, const double& b1, const bool& ionExcited) {
  double _delta=1;
  double n2;
  if (EIonEdge>0) _delta=delta(Te,TR,EIon,W,EIonEdge,b1);
  n2=_delta*nIonGround(n1,ne,Te,TR,EIon,g1,g2,W,zeta);
  if (ionExcited) n2=n2+nIonExcited(n1,ne,Te,TR,EIon,g1,g2,W,zeta);
  return n2;
}

// Delta factor as given by equations (15) and (20) in Mazzali & Lucy 93.
double NebSolver::delta (const double& Te, const double& TR, const double& EIon, const double& W, const double& EIonEdge, const double& b1) {
  double _delta;
  if (EIon >= EIonEdge) _delta=(Te/(b1*W*TR))*exp(EIon/(k*TR)-EIon/(k*Te));
  else _delta=1-exp(EIon/(k*TR)-EIonEdge/(k*TR))+(Te/(b1*W*TR))*exp(EIon/(k*TR)-EIonEdge/(k*Te));

  return _delta;
}

// Update temperature for cell
void NebSolver::updateTemperature(Cell* cell) {
  double T;

  if (nebConfig.temperature.method==sLTE) return LTESolver::updateTemperature(cell);
  else {
    // Set temperature according to radiation temperature
    //cell->T(0.9*cell->TR());
    // Changed! Testing!
    T=0.9*cell->TR();
    if (config.convergence.TMin!=noT) T=MAX(T,config.convergence.TMin);
    if (config.convergence.TMax!=noT) T=MIN(T,config.convergence.TMax);
    cell->T(T);

    if (log()) cout<<"NebSolver::updateTemperature: T="<<cell->T()<<"\n";
  }
}

// Update matter state assuming LTE
void NebSolver::updateLTE(Cell* cell) {

  // Update partition function
  lteSolver.updateU(cell);
  // Update ionization state
  lteSolver.updateIonState(cell);
  // Update bound state
  lteSolver.updateBoundState(cell);
}

// Constructor
NLTEPopConfigEntry::NLTEPopConfigEntry() : processes(), coupled(true) {}

// Load data
void NLTEPopConfigEntry::load(const YAML::Node& data) {
  int i;

  // Load base class data
  AtomConfigEntry::load(data);

  // Load coupled flag
  if (data["coupled"]) coupled=(data["coupled"].as<string>().compare("yes")==0);

  // Load processes
  if (data["processes"]) processes.load(data["processes"]);
}

// Log data
void NLTEPopConfigEntry::log() {
  cout<<"NLTEPopConfigEntry::log:";
  AtomConfigEntry::log();
  cout<<", coupled: "<<coupled<<",";
  processes.log();
  cout<<"\n";
}

// Constructor
//NLTEPopConfig::NLTEPopConfig() : coupledCell(false) {}

// Coupled
bool NLTEPopConfig::coupled (Atom& atom) {
  NLTEPopConfigEntry entry=getEntry(atom);
      
  return entry.coupled;
}

// Processes
Processes NLTEPopConfig::processes (Atom& atom) {
  NLTEPopConfigEntry entry=getEntry(atom);
    
  return entry.processes;
}

// Constructor
NLTETempConfigEntry::NLTETempConfigEntry() : processes() {}

// Load data
void NLTETempConfigEntry::load(const YAML::Node& data) {
  int i;

  // Load base class data
  AtomConfigEntry::load(data);

  // Load processes
  if (data["processes"]) processes.load(data["processes"]);
}

// Log data
void NLTETempConfigEntry::log() {
  cout<<"NLTETempConfigEntry::load:";
  AtomConfigEntry::log();
  cout<<",";
  processes.log();
  cout<<"\n";
}

// Processes
Processes NLTETempConfig::processes (Atom& atom) {
  NLTETempConfigEntry entry=getEntry(atom);
 
  return entry.processes;
}

// Load number density error method
ErrNMethod msolver::loadErrNMethod(const YAML::Node& data) {
  if (data.as<string>().compare("mean")==0) return enMean;
  else return enMax;
}

// Constructors
NLTEConvConfig::NLTEConvConfig() : nIter(1000), nIterMin(1), nTol(1e-2), TTol(1e-2), neTol(1e-2), dampN(0.7), dampT(0.5), dampNMin(0.3), dampTMax(1.0), dampTMin(0.3), nMin(1.0), TMin(500), TMax(1e99), dTMax(500), tauLim(1e-3), errNMethod(enMax), derivative(dPartial), dTDerivative(1e-6), scanT(false), scanTFac(0), scanTTol(1e-3), scanTMethod(scanTBisection), scanTBAcc(1.0), scanTNIter(1), scanTDampN(1.0), scanTNeTol(1e-2), scanTNTol(1e-2), scanTNTolInit(1e-2), scanTNTNeTol(1e99), scanTNTNeTolInit(1e99), iterT(true), NTNeTol(0.1), dNedN(false) { scanTFac.push_back(1.1); scanTFac.push_back(1.5); scanTFac.push_back(2.0); } // Changed!
//NLTEConvConfig::NLTEConvConfig(const int& _nIter, const int& _nIterMin, const double& _nTol, const double& _TTol, const double& _neTol, const double& _dampN, const double& _dampT, const double& _dampNMin, const double& _dampTMin, const double& _dampTMax, const double& _nMin, const double& _TMin, const double& _dTMax, const double& _tauLim, const ErrNMethod& _errNMethod, const DerivativeType& _derivative, const double& _dTDerivative, const bool& _scanT, const double& _scanTFac, const double& _scanTDiv, const double& _scanTFacSmall, const double& _scanTDivSmall, const double& _scanTNIter, const double& _scanTDampN, const double& _scanTNeTol, const double& _scanTNTol, const bool& _iterT) : nIter(_nIter), nIterMin(_nIterMin), nTol(_nTol), TTol(_TTol), neTol(_neTol), dampN(_dampN), dampT(_dampT), dampNMin(_dampNMin), dampTMin(_dampTMin), dampTMax(_dampTMax), nMin(_nMin), TMin(_TMin), dTMax(_dTMax), tauLim(_tauLim), errNMethod(_errNMethod), derivative(_derivative), dTDerivative(_dTDerivative), scanT(_scanT), scanTFac(_scanTFac), scanTDiv(_scanTDiv), scanTFacSmall(_scanTFacSmall), scanTDivSmall(_scanTDivSmall), scanTNIter(_scanTNIter), scanTDampN(_scanTDampN), scanTNeTol(_scanTNeTol), scanTNTol(_scanTNTol), iterT(_iterT) {}
NLTEConvConfig::NLTEConvConfig(const int& _nIter, const int& _nIterMin, const double& _nTol, const double& _TTol, const double& _neTol, const double& _dampN, const double& _dampT, const double& _dampNMin, const double& _dampTMin, const double& _dampTMax, const double& _nMin, const double& _TMin, const double& _TMax, const double& _dTMax, const double& _tauLim, const ErrNMethod& _errNMethod, const DerivativeType& _derivative, const double& _dTDerivative, const bool& _scanT, const vector<double>& _scanTFac, const double& _scanTTol, const ScanTMethod& _scanTMethod, const double& _scanTBAcc, const double& _scanTNIter, const double& _scanTDampN, const double& _scanTNeTol, const double& _scanTNTol, const double& _scanTNTolInit, const double& _scanTNTNeTol, const double& _scanTNTNeTolInit, const bool& _iterT, const double& _NTNeTol, const bool& _dNedN) : nIter(_nIter), nIterMin(_nIterMin), nTol(_nTol), TTol(_TTol), neTol(_neTol), dampN(_dampN), dampT(_dampT), dampNMin(_dampNMin), dampTMin(_dampTMin), dampTMax(_dampTMax), nMin(_nMin), TMin(_TMin), TMax(_TMax), dTMax(_dTMax), tauLim(_tauLim), errNMethod(_errNMethod), derivative(_derivative), dTDerivative(_dTDerivative), scanT(_scanT), scanTFac(_scanTFac), scanTTol(_scanTTol), scanTMethod(_scanTMethod), scanTBAcc(_scanTBAcc), scanTNIter(_scanTNIter), scanTDampN(_scanTDampN), scanTNeTol(_scanTNeTol), scanTNTol(_scanTNTol), scanTNTolInit(_scanTNTolInit), scanTNTNeTol(_scanTNTNeTol), scanTNTNeTolInit(_scanTNTNeTolInit), iterT(_iterT), NTNeTol(_NTNeTol), dNedN(_dNedN) {} // Changed!

// Load data
void NLTEConvConfig::load(const YAML::Node& data) {
  int i;

  // Load number of iterations
  if (data["nIter"]) nIter=data["nIter"].as<int>();

  // Load minimum number of iterations
  if (data["nIterMin"]) nIterMin=data["nIterMin"].as<int>();

  // Load number density tolerence
  if (data["nTol"]) nTol=data["nTol"].as<double>();

  // Load temperature tolerance
  if (data["TTol"]) TTol=data["TTol"].as<double>();

  // Load electron number density tolerence
  if (data["neTol"]) neTol=data["neTol"].as<double>();

  // Load number density damping factor
  if (data["dampN"]) dampN=data["dampN"].as<double>();

  // Load temperature damping factor
  if (data["dampT"]) dampT=data["dampT"].as<double>();

  // Load number density minimum damping factor
  if (data["dampNMin"]) dampNMin=data["dampNMin"].as<double>();

  // Load temperature minimum damping factor
  if (data["dampTMin"]) dampTMin=data["dampTMin"].as<double>();

  // Load temperature maximum damping factor
  if (data["dampTMax"]) dampTMax=data["dampTMax"].as<double>();

  // Load minimum number density (in units of atom.n*mprec)
  if (data["nMin"]) nMin=data["nMin"].as<double>();

  // Load minimum temperature
  if (data["TMin"]) TMin=data["TMin"].as<double>();

  // Load maximum temperature - New!
  if (data["TMax"]) TMax=data["TMax"].as<double>();

  // Load maximum temperature change
  if (data["dTMax"]) dTMax=data["dTMax"].as<double>();

  // Load optical depth limit
  if (data["tauLim"]) tauLim=data["tauLim"].as<double>();

  // Load number density error method
  if (data["errNMethod"]) errNMethod=loadErrNMethod(data["errNMethod"]);

  // Load derivative type
  if (data["derivative"]) {
    if (data["derivative"].as<string>().compare("full")==0) derivative=dFull;
    else if (data["derivative"].as<string>().compare("fullNe")==0) derivative=dFullNe;
    else derivative=dPartial;
  }

  // Load derivative step
  if (data["dTDerivative"]) dTDerivative=data["dTDerivative"].as<double>();

  // Load temperature scan flag
  if (data["scanT"]) scanT=(data["scanT"].as<string>().compare("yes")==0);

  // Load temperature scan fractions
  if (data["scanTFac"]) {
    if (data["scanTFac"].size()>0) {
      scanTFac.resize(data["scanTFac"].size());
      for (i=0;i<data["scanTFac"].size();i++) scanTFac[i]=data["scanTFac"][i].as<double>();
    }
  }

  // Load temperature scan tolerance
  if (data["scanTTol"]) scanTTol=data["scanTTol"].as<double>();

  // Load temperature scan method
  if (data["scanTMethod"]) {
    if (data["scanTMethod"].as<string>().compare("mixed")==0) scanTMethod=scanTMixed;
    else scanTMethod=scanTBisection;
  }

  // Load temperature scan bisection acceleration factor
  if (data["scanTBAcc"]) scanTBAcc=data["scanTBAcc"].as<double>();

  // Load temperature scan (number density) iterations
  if (data["scanTNIter"]) scanTNIter=data["scanTNIter"].as<int>();

  // Load temperature scan (number density) damping
  if (data["scanTDampN"]) scanTDampN=data["scanTDampN"].as<double>();

  // Load temperature scan (electron density) tolerance
  if (data["scanTNeTol"]) scanTNeTol=data["scanTNeTol"].as<double>();

  // Load temperature scan (number density) tolerance
  if (data["scanTNTol"]) scanTNTol=data["scanTNTol"].as<double>();

  // Load initial temperature scan (number density) tolerance
  if (data["scanTNTolInit"]) scanTNTolInit=data["scanTNTolInit"].as<double>();

  // Load temperature scan non-thermal solver (electron density) tolerance
  if (data["scanTNTNeTol"]) scanTNTNeTol=data["scanTNTNeTol"].as<double>();

  // Load initial temperature scan non-thermal solver (electron density) tolerance
  if (data["scanTNTNeTolInit"]) scanTNTNeTolInit=data["scanTNTNeTolInit"].as<double>();

  // Load temperature iteration flag
  if (data["iterT"]) iterT=(data["iterT"].as<string>().compare("yes")==0);

  // Load non-thermal solver (electron density) tolerance
  if (data["NTNeTol"]) NTNeTol=data["NTNeTol"].as<double>();

  // Load implicit electron number density derivative flag
  if (data["dNedN"]) dNedN=(data["dNedN"].as<string>().compare("yes")==0);
  else dNedN=false;
}

// Log data
void NLTEConvConfig::log() {
  //cout<<"NLTEConvConfig::log: nIter: "<<nIter<<", nIterMin: "<<nIterMin<<", nTol: "<<nTol<<", TTol: "<<TTol<<", neTol: "<<neTol<<", dampN: "<<dampN<<", dampT: "<<dampT<<", dampNMin: "<<dampNMin<<", dampTMin: "<<dampTMin<<", dampTMax: "<<dampTMax<<", nMin: "<<nMin<<", TMin: "<<TMin<<", dTMax: "<<dTMax<<", tauLim: "<<tauLim<<", errNMethod: "<<errNMethod<<", derivative: "<<derivative<<", dTDerivative: "<<dTDerivative<<", scanT: "<<scanT<<", scanTFac: "<<scanTFac<<", scanTDiv: "<<scanTDiv<<", scanTFacSmall: "<<scanTFacSmall<<", scanTDivSmall: "<<scanTDivSmall<<", scanTNIter: "<<scanTNIter<<", scanTDampN: "<<scanTDampN<<", scanTNeTol: "<<scanTNeTol<<", scanTNTol: "<<scanTNTol<<", iterT: "<<iterT<<"\n";
  cout<<"NLTEConvConfig::log: nIter: "<<nIter<<", nIterMin: "<<nIterMin<<", nTol: "<<nTol<<", TTol: "<<TTol<<", neTol: "<<neTol<<", dampN: "<<dampN<<", dampT: "<<dampT<<", dampNMin: "<<dampNMin<<", dampTMin: "<<dampTMin<<", dampTMax: "<<dampTMax<<", nMin: "<<nMin<<", TMin: "<<TMin<<", TMax: "<<TMax<<", dTMax: "<<dTMax<<", tauLim: "<<tauLim<<", errNMethod: "<<errNMethod<<", derivative: "<<derivative<<", dTDerivative: "<<dTDerivative<<", scanT: "<<scanT<<", scanTTol: "<<scanTTol<<", scanTNIter: "<<scanTNIter<<", scanTDampN: "<<scanTDampN<<", scanTNeTol: "<<scanTNeTol<<", scanTNTol: "<<scanTNTol<<", scanTNTolInit: "<<scanTNTolInit<<", scanTNTNeTol: "<<scanTNTNeTol<<", scanTNTNeTolInit: "<<scanTNTNeTolInit<<", iterT: "<<iterT<<", NTNeTol: "<<NTNeTol<<", dNedN: "<<dNedN<<"\n"; // Changed!
}

// Constructors
NLTESolverConfig::NLTESolverConfig() : population(), temperature(), convergence(), steadyState(true), coupled(false), ionCoupled(false) {}; // Changed!
NLTESolverConfig::NLTESolverConfig(const NLTEPopConfig& _population, const NLTETempConfig& _temperature, const NLTEConvConfig& _convergence, const bool& _steadyState, const bool& _coupled, const bool& _ionCoupled) : population(_population), temperature(_temperature), convergence(_convergence), steadyState(_steadyState), coupled(_coupled), ionCoupled(_ionCoupled) {}; // Changed!
NLTESolverConfig::NLTESolverConfig(const string& dataFile) { load(dataFile); }

// Load data
void NLTESolverConfig::load(const YAML::Node& data) {

  // Load population config
  if (data["population"]) population.load(data["population"]);

  // Load temperature config
  if (data["temperature"]) temperature.load(data["temperature"]);

  // Load convergence config
  if (data["convergence"]) convergence.load(data["convergence"]);

  // Load steady-state flag
  if (data["steadyState"]) steadyState=(data["steadyState"].as<string>().compare("yes")==0);

  // Load coupled flag
  if (data["coupled"]) coupled=(data["coupled"].as<string>().compare("yes")==0);

  // Load ion coupled flag - New!
  if (data["ionCoupled"]) ionCoupled=(data["ionCoupled"].as<string>().compare("yes")==0);

  log();
}

// Log data
void NLTESolverConfig::log() {
  population.log();
  temperature.log();
  convergence.log();
  cout<<"NLTESolverConfig::log: steadyState: "<<steadyState<<", coupled: "<<coupled<<", ionCoupled: "<<ionCoupled<<"\n"; // Changed!
}

// Load data (from file)
void NLTESolverConfig::load(const string& dataFile) {
  YAML::Node data;

  // Load data from file
  data=YAML::LoadFile(dataFile);

  // Load data
  if (data["nlteSolver"]) load(data["nlteSolver"]);
}

// Constructors
NLTESolver::NLTESolver(NLTESolverConfig _config, LTESolverConfig _lteConfig, Koz92SolverConfig _ntConfig ) : config(_config), lteSolver(_lteConfig), ntSolver(_ntConfig), pLog(true), tOld(0) { lteSolver.config.temperature=LTETempConfig(); } // Changed! - Force LTE solver default temperature config - Hack!
NLTESolver::NLTESolver(const string& dataFile) : lteSolver(dataFile), ntSolver(dataFile), pLog(true), tOld(0) { config.load(dataFile); lteSolver.config.temperature=LTETempConfig(); } // Changed! - Force LTE solver default temperature config - Hack!

// Copy constructor - New!
NLTESolver::NLTESolver(const NLTESolver& nlteSolver) : lteSolver(nlteSolver.lteSolver), ntSolver(nlteSolver.ntSolver), n(nlteSolver.n), rErrN(nlteSolver.rErrN), rErrT(nlteSolver.rErrT) ,rErrNe(nlteSolver.rErrNe), rErrNPrev(nlteSolver.rErrNPrev), rErrTPrev(nlteSolver.rErrTPrev), rErrNePrev(nlteSolver.rErrNePrev), dampN(nlteSolver.dampN), dampT(nlteSolver.dampT), excludeAtom(nlteSolver.excludeAtom), excludeIState(nlteSolver.excludeIState), excludeBState(nlteSolver.excludeBState), pLog(nlteSolver.pLog), config(nlteSolver.config), nOld(nlteSolver.nOld), TOld(nlteSolver.TOld), tOld(nlteSolver.tOld) { // Changed!
  //cout<<"NLTESolver::NLTESolver: this: "<<this<<"\n";
}

// Create copy
MStateSolver* NLTESolver::copy() {
  //cout<<"NLTESolver::copy: this: "<<this<<"\n";
  return new NLTESolver(*this);
}

// Count number of transitions for bound state
int NLTESolver::nTrans(const BState& bState, const Processes& processes) {
  double _nTrans=0;
  if (processes(pBoundBound)) _nTrans+=bState.upTransBB.size()+bState.downTransBB.size();
  if (processes(pBoundFree)) _nTrans+=bState.upTransBF.size()+bState.downTransBF.size();
  return _nTrans;
}

// Count number of transitions for ion
int NLTESolver::nTrans(const IState& iState, const Processes& processes) {
  double _nTrans=0;
  int i;
  for (i=0;i<iState.bState.size();i++) _nTrans+=nTrans(iState.bState[i],processes);
  return _nTrans;
}

// Count number of transitions for atom
int NLTESolver::nTrans(const Atom& atom, const Processes& processes) {
  double _nTrans=0;
  int i;
  for (i=0;i<atom.iState.size();i++) _nTrans+=nTrans(atom.iState[i],processes);
  return _nTrans;
}

// Calculate maximum optical depth for bound-bound transitions
double NLTESolver::bbTauMax(BState& bState) {
  double tauMax=0;
  int i;

  // Calculate maximum optical depth
  for (i=0;i<bState.upTransBB.size();i++) 
    if (bState.upTransBB[i]->tau()>tauMax) tauMax=bState.upTransBB[i]->tau();

  return tauMax;
}

// Calculate maximum optical depth for bound-bound transitions (ionization state version) - New!
double NLTESolver::bbTauMax(IState& iState) {
  double tauMax=0,bbTauMaxBState;
  int i;

  // Calculate maximum optical depth
  for (i=0;i<iState.bState.size();i++) {
    bbTauMaxBState=bbTauMax(iState.bState[i]);
    if (bbTauMaxBState>tauMax) tauMax=bbTauMaxBState;
  }

  return tauMax;
}

// Force neutral state of atom
double NLTESolver::forceNeutralState(Atom& atom) {
  int i,j;
  for (i=0;i<atom.iState.size();i++) { atom.iState[0].X(1); for (j=1;j<atom.iState.size();j++) atom.iState[i].X(0); }
}

// Force ground state of ion
double NLTESolver::forceGroundState(IState& iState) {
  int i,j;
  for (i=0;i<iState.bState.size();i++) { iState.bState[0].X(1); for (j=1;j<iState.bState.size();j++) iState.bState[i].X(0); }
}

// Exclude atom, ionization state or bound state
bool NLTESolver::exclude(const Path& path) {

  if (path.size()==1) {
    //cout<<"NLTESolver::exclude:: this: "<<this<<", path[0] "<<path[0]<<", excludeAtom.size: "<<excludeAtom.size()<<"\n";
    return excludeAtom[path[0]];
  }
  else if (path.size()==2) return excludeIState[path[0]][path[1]];
  else if (path.size()==3) return excludeBState[path[0]][path[1]][path[2]];

  return false;
}

// Make bound state NLTE matrix index map for ion
vector<int> NLTESolver::bMatrixIndex(IState& iState, int& nStates, const Path& path) {
  vector<int> index(iState.bState.size(),noIndex);
  int i;

  // For each bound state
  for (i=0;i<iState.bState.size();i++) {
    // If not excluded and no LTE reference state
    if (not exclude(path<<i) and iState.bState[i].LTERefState==NULL) {
      // Set matrix index
      index[i]=nStates;
      // Increase state counter
      nStates++;
    }
  }
  
  // For each bound state
  for (i=0;i<iState.bState.size();i++) {
    // If not excluded
    if (not exclude(path<<i)) {
      // If LTE reference state
      if (iState.bState[i].LTERefState!=NULL)
	// Set index to same as for LTE reference state
	index[i]=index[iState.bState[i].LTERefState->getIndex()];
    }
  }
  
  return index;
}

// Make bound state NLTE matrix index map for atom
vector<vector<int> > NLTESolver::bMatrixIndex(Atom& atom, int& nStates, const Path& path) {
  vector<vector<int> > index(atom.iState.size());
  int i,j; // Changed!
  BState* refState;

  // Testing!!!
  //cout<<"NLTESolver::bMatrixIndex: cell: "<<atom.cell()->index()<<", atom: "<<atom.name()<<", Enter\n";

  // Make index map for each ion
  for (i=0;i<atom.iState.size();i++) index[i]=bMatrixIndex(atom.iState[i],nStates,path<<i);

  // For each ion - New block!
  for (i=0;i<atom.iState.size();i++) 
    // For each bound state
    for (j=0;j<atom.iState[i].bState.size();j++) {
      // If not excluded
      if (not exclude(path<<i<<j)) {
	// If LTE reference state
	if (atom.iState[i].bState[j].LTERefState!=NULL) {
	  // Set reference state pointer
	  refState=atom.iState[i].bState[j].LTERefState;
	  // Set index to same as for LTE reference state
	  index[i][j]=index[refState->iState()->index()][refState->getIndex()];
	}
      }
    }

  return index;

  // Testing!!!
  //cout<<"NLTESolver::bMatrixIndex: cell: "<<atom.cell()->index()<<", atom: "<<atom.name()<<", Return\n";
}

// Make bound state NLTE matrix index map for cell
vector<vector<vector<int> > > NLTESolver::bMatrixIndex(Cell* cell, int& nStates) {
  vector<vector<vector<int> > > index(cell->atom.size());
  int i;

  // Make index map for each atom
  for (i=0;i<cell->atom.size();i++) index[i]=bMatrixIndex(cell->atom[i],nStates,Path(i));

  return index;
}

// Make ionization state NLTE matrix index map for atom
vector<int> NLTESolver::iMatrixIndex(Atom& atom, int& nStates, const Path& path) {
  vector<int> index(atom.iState.size(),noIndex);
  int i;

  // For each ionization state
  for (i=0;i<atom.iState.size();i++) {
    // If not excluded
    if (not exclude(path<<i)) {
      // Set matrix index
      index[i]=nStates;
      // Increase state counter
      nStates++;
    }
  }
  
  return index;
}

// Make ionization state NLTE matrix index map for cell - New!
vector<vector<int> > NLTESolver::iMatrixIndex(Cell* cell, int& nStates) {
  vector<vector<int> > index(cell->atom.size());
  int i;

  // Make index map for each atom
  for (i=0;i<cell->atom.size();i++) index[i]=iMatrixIndex(cell->atom[i],nStates,Path(i));

  return index;
}

// Populate bound state NLTE matrix for ion
void NLTESolver::populateBMatrix(const IState& iState, Mat_DP& matrix, Vec_DP& rhs, const vector<int>& index, const Processes& processes, const Path& path, bool nConserve, bool boundFree) {
  Cell* cell=iState.atom()->cell();
  TransBB* transBB;
  TransBF* transBF;
  BState *upperBState, *lowerBState, *refBState;
  IState *upperIState, *lowerIState;
  double x,xLower,xUpper;
  int i,j;
  
  // Get (bound-bound) collisional and non-thermal processes flags
  bool collisional=processes(pBoundBound,pCollisional);
  bool nonThermal=processes(pBoundBound,pNonThermal);
  bool twoPhoton=processes(pBoundBound,pTwoPhoton); // New!

  // For each bound state
  for (i=0;i<iState.bState.size();i++) {
    // If not excluded
    if (not exclude(path<<i)) {

      // Get LTE reference factor
      x=xLTERef(iState.bState[i],cell->T());

      // For each upward bound-bound transition
      for (j=0;j<iState.bState[i].upTransBB.size();j++) { 
	// Set transition pointer
	transBB=iState.bState[i].upTransBB[j];
	// If not LTE group transition
	//if (not LTEGroupTrans(transBB)) {
	// Get upper state LTE reference factor
	xUpper=xLTERef(*(transBB->upperState),cell->T());
	// Calculate equation coefficients
	matrix[index[i]][index[transBB->upperState->getIndex()]]+=transBB->gammaD(collisional,twoPhoton)*xUpper; // Changed!
	matrix[index[i]][index[i]]-=transBB->gammaU(collisional,nonThermal)*x;
	//}
      }
      
      // For each downward bound-bound transition
      for (j=0;j<iState.bState[i].downTransBB.size();j++) { 
	// Set transition pointer
	transBB=iState.bState[i].downTransBB[j];
	// If not LTE group transition
	//if (not LTEGroupTrans(transBB)) {
	// Get lower state LTE reference factor
	xLower=xLTERef(*(transBB->lowerState),cell->T());
	// Calculate equation coefficients
	matrix[index[i]][index[transBB->lowerState->getIndex()]]+=transBB->gammaU(collisional,nonThermal)*xLower;
	matrix[index[i]][index[i]]-=transBB->gammaD(collisional,twoPhoton)*x; // Changed!
	//}
      }

      // If include bound-free rates on the righ-hand-side
      if (boundFree) {
	// For each upward bound-free transition
	for (j=0;j<iState.bState[i].upTransBF.size();j++) {
	  // Set transition and state pointers
	  transBF=iState.bState[i].upTransBF[j]; 
	  upperBState=transBF->upperState;
	  upperIState=transBF->upperState->iState();
	  // Calculate right-hand-side
	  matrix[index[i]][index[i]]-=transBF->gammaU(processes(pBoundFree,pCollisional),processes(pBoundFree,pNonThermal),processes(pBoundFree,pChargeTransfer))*x; // Changed!
	  rhs[index[i]]-=transBF->gammaD(processes(pBoundFree,pCollisional),processes(pBoundFree,pChargeTransfer))*upperBState->X()*(upperIState->X()/iState.X()); // Changed!
	  // Need to include net rate in on-the-spot range on the right-hand side !!!
	}
	
	// For each downward bound-free transition
	for (j=0;j<iState.bState[i].downTransBF.size();j++) {
	  // Set transition and state pointers
	  transBF=iState.bState[i].downTransBF[j];
	  lowerBState=transBF->lowerState;
	  lowerIState=transBF->lowerState->iState();
	  // Calculate right-hand-side
	  matrix[index[i]][index[i]]-=transBF->gammaD(processes(pBoundFree,pCollisional),processes(pBoundFree,pChargeTransfer))*x; // Changed!
	  rhs[index[i]]-=transBF->gammaU(processes(pBoundFree,pCollisional),processes(pBoundFree,pNonThermal),processes(pBoundFree,pChargeTransfer))*lowerBState->X()*(lowerIState->X()/iState.X()); // Changed!
	  // Need to include net rate in on-the-spot range on the right-hand side !!!
	}
      }
    }
  }
  
  // If requested 
  if (nConserve) {
    // Add the number (fraction) conservation equation
    // Add reset the of the corresponding row in the matrix!!!
    //for (i=0;i<matrix.nrows();i++) matrix[matrix.nrows()-1][i]=1.0;
    for (i=0;i<matrix.nrows();i++) matrix[matrix.nrows()-1][i]+=xLTERef(iState.bState[i],cell->T());

    rhs[rhs.size()-1]=1.0;
  }
}

// Populate bound state NLTE matrix for ion
//void NLTESolver::populateBMatrixDN(const IState& iState, Mat_DP& matrix, Vec_DP& rhs, const vector<int>& index, const Processes& processes, const Path& path, bool nConserve, bool boundFree) {
void NLTESolver::populateBMatrixDN(IState& iState, Mat_DP& matrix, Vec_DP& rhs, const vector<int>& index, const Processes& processes, const Path& path, bool nConserve, bool boundFree) {
  Cell* cell=iState.atom()->cell();
  TransBB* transBB;
  TransBF* transBF;
  BState *upperBState, *lowerBState, *refBState;
  IState *upperIState, *lowerIState;
  double x,xLower,xUpper,n,nLower,nUpper;
  int i,j;
  
  // Get (bound-bound) collisional and non-thermal processes flags
  bool collisional=processes(pBoundBound,pCollisional);
  bool nonThermal=processes(pBoundBound,pNonThermal);
  bool twoPhoton=processes(pBoundBound,pTwoPhoton);

  // For each bound state
  for (i=0;i<iState.bState.size();i++) {
    // If not excluded
    if (not exclude(path<<i)) {

       // Get LTE reference factor
      x=xLTERef(iState.bState[i],cell->T());
      n=nLTERef(iState.bState[i],cell->T());

      // For each upward bound-bound transition
      for (j=0;j<iState.bState[i].upTransBB.size();j++) { 
	// Set transition pointer
	transBB=iState.bState[i].upTransBB[j];
	// If not LTE group transition
	//if (not LTEGroupTrans(transBB)) {
	// Get upper state LTE reference factor
	xUpper=xLTERef(*(transBB->upperState),cell->T());
	nUpper=nLTERef(*(transBB->upperState),cell->T());
	// Calculate equation coefficients
	matrix[index[i]][index[transBB->upperState->getIndex()]]+=transBB->gammaD(collisional,twoPhoton)*xUpper; // Changed!
	//if (log() and iState.atom()->cell()->index()==36 and !isnormal(	matrix[index[i]][index[transBB->upperState->getIndex()]])) cout<<"NLTESolver::populateBMatrixDN(1): cell: "<<iState.atom()->cell()->index()<<", trans: "<<transBB->upperState->iState()->name()<<":"<<transBB->upperState->getIndex()<<":"<<transBB->lowerState->iState()->name()<<":"<<transBB->lowerState->getIndex()<<", matrix["<<index[i]<<"]["<<index[transBB->upperState->getIndex()]<<"]: "<<matrix[index[i]][index[transBB->upperState->getIndex()]]<<"\n";
	matrix[index[i]][index[i]]-=transBB->gammaU(collisional,nonThermal)*x;
	//if (log() and iState.atom()->cell()->index()==36 and !isnormal(matrix[index[i]][index[i]])) cout<<"NLTESolver::populateBMatrixDN(2): cell: "<<iState.atom()->cell()->index()<<", trans: "<<transBB->upperState->iState()->name()<<":"<<transBB->upperState->getIndex()<<":"<<transBB->lowerState->iState()->name()<<":"<<transBB->lowerState->getIndex()<<", matrix["<<index[i]<<"]["<<index[i]<<"]: "<<matrix[index[i]][index[i]]<<"\n";
	// Add net rate based on previous populations to the right-hand side
	if (transBB->lowerState->LTERefState==NULL and transBB->upperState->LTERefState==NULL) {
	  rhs[index[i]]+=transBB->RNet(twoPhoton); // Changed!
	  if (collisional) rhs[index[i]]+=transBB->CNet();
	  if (nonThermal) rhs[index[i]]+=transBB->CNU();
	}
	else rhs[index[i]]-=(transBB->gammaD(collisional,twoPhoton)*nUpper-transBB->gammaU(collisional,nonThermal)*n); // Changed!
	//if (log() and iState.atom()->cell()->index()==36 and !isnormal(rhs[index[i]])) cout<<"NLTESolver::populateBMatrixDN(3): cell: "<<iState.atom()->cell()->index()<<", trans: "<<transBB->upperState->iState()->name()<<":"<<transBB->upperState->getIndex()<<":"<<transBB->lowerState->iState()->name()<<":"<<transBB->lowerState->getIndex()<<", rhs["<<index[i]<<"]: "<<rhs[index[i]]<<", RNet: "<<transBB->RNet()<<", CNet: "<<transBB->CNet()<<"\n";
	//if (log() and iState.atom()->name().compare("He")==0 and iState.atom()->cell()->index()==8) cout<<"NLTESolver::populateBMatrixDN: cell: "<<iState.atom()->cell()->index()<<", trans: "<<transBB->lowerState->iState()->name()<<":"<<transBB->lowerState->getIndex()<<"->"<<transBB->upperState->iState()->name()<<":"<<transBB->upperState->getIndex()<<", RNet: "<<transBB->RNet()<<", RU: "<<transBB->RU()<<", RD: "<<transBB->RD()<<", CNet: "<<transBB->CNet()<<", CU: "<<transBB->CU()<<", CD: "<<transBB->CD()<<", CNU: "<<transBB->CNU()<<"\n";
	//}
      }
      
      // For each downward bound-bound transition
      for (j=0;j<iState.bState[i].downTransBB.size();j++) { 
	// Set transition pointer
	transBB=iState.bState[i].downTransBB[j];
	// If not LTE group transition
	//if (not LTEGroupTrans(transBB)) {
	// Get lower state LTE reference factor
	xLower=xLTERef(*(transBB->lowerState),cell->T());
	nLower=nLTERef(*(transBB->lowerState),cell->T());
	// Calculate equation coefficients
	matrix[index[i]][index[transBB->lowerState->getIndex()]]+=transBB->gammaU(collisional,nonThermal)*xLower;
	//if (log() and iState.atom()->cell()->index()==36 and !isnormal(matrix[index[i]][index[transBB->lowerState->getIndex()]])) cout<<"NLTESolver::populateBMatrixDN(4): cell: "<<iState.atom()->cell()->index()<<", trans: "<<transBB->upperState->iState()->name()<<":"<<transBB->upperState->getIndex()<<":"<<transBB->lowerState->iState()->name()<<":"<<transBB->lowerState->getIndex()<<", matrix["<<index[i]<<"]["<<index[transBB->lowerState->getIndex()]<<"]: "<<matrix[index[i]][index[transBB->lowerState->getIndex()]]<<"\n";
	matrix[index[i]][index[i]]-=transBB->gammaD(collisional,twoPhoton)*x; // Changed!
	//if (log() and iState.atom()->cell()->index()==36 and !isnormal(matrix[index[i]][index[i]])) cout<<"NLTESolver::populateBMatrixDN(5): cell: "<<iState.atom()->cell()->index()<<", trans: "<<transBB->upperState->iState()->name()<<":"<<transBB->upperState->getIndex()<<":"<<transBB->lowerState->iState()->name()<<":"<<transBB->lowerState->getIndex()<<", matrix["<<index[i]<<"]["<<index[i]<<"]: "<<matrix[index[i]][index[i]]<<"\n";
	// Add net rate based on previous populations to the right-hand side
	if (transBB->lowerState->LTERefState==NULL and transBB->upperState->LTERefState==NULL) {
	  rhs[index[i]]-=transBB->RNet(twoPhoton); // Changed!
	  if (collisional) rhs[index[i]]-=transBB->CNet();
	  if (nonThermal) rhs[index[i]]-=transBB->CNU();
	}
	else rhs[index[i]]-=(transBB->gammaU(collisional,nonThermal)*nLower-transBB->gammaD(collisional,twoPhoton)*n); // Changed!
	//if (log() and iState.atom()->cell()->index()==36 and !isnormal(rhs[index[i]])) cout<<"NLTESolver::populateBMatrixDN(6): cell: "<<iState.atom()->cell()->index()<<", trans: "<<transBB->upperState->iState()->name()<<":"<<transBB->upperState->getIndex()<<":"<<transBB->lowerState->iState()->name()<<":"<<transBB->lowerState->getIndex()<<", rhs["<<index[i]<<"]: "<<rhs[index[i]]<<", RNet: "<<transBB->RNet()<<", CNet: "<<transBB->CNet()<<"\n";
	//if (log() and iState.atom()->name().compare("He")==0 and iState.atom()->cell()->index()==8) cout<<"NLTESolver::populateBMatrixDN: cell: "<<iState.atom()->cell()->index()<<", trans: "<<transBB->upperState->iState()->name()<<":"<<transBB->upperState->getIndex()<<"->"<<transBB->lowerState->iState()->name()<<":"<<transBB->lowerState->getIndex()<<", RNet: "<<transBB->RNet()<<", RU: "<<transBB->RU()<<", RD: "<<transBB->RD()<<", CNet: "<<transBB->CNet()<<", CU: "<<transBB->CU()<<", CD: "<<transBB->CD()<<", CNU: "<<transBB->CNU()<<"\n";
	//}
      }
    
      // If include bound-free rates at this stage
      if (boundFree) {
	// For each upward bound-free transition
	for (j=0;j<iState.bState[i].upTransBF.size();j++) {
	  // Set transition and state pointers
	  transBF=iState.bState[i].upTransBF[j]; 
	  upperBState=transBF->upperState;
	  upperIState=transBF->upperState->iState();
	  // Calculate equation coefficients
	  //matrix[index[i]][index[i]]-=transBF->gammaU(processes(pBoundFree,pCollisional),processes(pBoundFree,pNonThermal))*x;
	  matrix[index[i]][index[i]]-=transBF->gammaU(processes(pBoundFree,pCollisional),processes(pBoundFree,pNonThermal),processes(pBoundFree,pChargeTransfer),true)*x; // Changed!
	  // Add net rate based on previous populations to the right-hand side
	  // Need to be corrected for LTE reference states !!!
	  rhs[index[i]]+=(transBF->RNet()-transBF->RDAdd()); // Changed!
	  if (processes(pBoundFree,pCollisional)) rhs[index[i]]+=transBF->CNet();
	  if (processes(pBoundFree,pNonThermal)) rhs[index[i]]+=transBF->CNU();
	  if (processes(pBoundFree,pChargeTransfer)) rhs[index[i]]+=transBF->CTNet(); // New!
	}
	
	// For each downward bound-free transition
	for (j=0;j<iState.bState[i].downTransBF.size();j++) {
	  // Set transition and state pointers
	  transBF=iState.bState[i].downTransBF[j];
	  lowerBState=transBF->lowerState;
	  lowerIState=transBF->lowerState->iState();
	  // Calculate equation coefficients
	  //matrix[index[i]][index[i]]-=transBF->gammaD(processes(pBoundFree,pCollisional))*x;
	  matrix[index[i]][index[i]]-=transBF->gammaD(processes(pBoundFree,pCollisional),processes(pBoundFree,pChargeTransfer),true)*x; // Changed!
	  // Add net rate based on previous populations to the right-hand side
	  // Need to be corrected for LTE reference states !!!
	  rhs[index[i]]-=(transBF->RNet()-transBF->RDAdd()); // Changed!
	  if (processes(pBoundFree,pCollisional)) rhs[index[i]]-=transBF->CNet();
	  if (processes(pBoundFree,pNonThermal)) rhs[index[i]]-=transBF->CNU();
	  if (processes(pBoundFree,pChargeTransfer)) rhs[index[i]]-=transBF->CTNet(); // New!
	}
      }
    }
  }
  
  // If requested 
  if (nConserve) {
    // Add the number (fraction) conservation equation
    // Add reset the of the corresponding row in the matrix!!!
    for (i=0;i<matrix.nrows();i++) matrix[matrix.nrows()-1][i]+=xLTERef(iState.bState[i],cell->T());

    // Need to be corrected for LTE reference states !!!
    rhs[rhs.size()-1]=0.0;
  }
}

// Populate bound state NLTE matrix for atom
void NLTESolver::populateBMatrix(const Atom& atom, Mat_DP& matrix, Vec_DP& rhs, const vector<vector<int> >& index, const Processes& processes, const Path& path) {
  Cell* cell=atom.cell();
  TransBF* trans;
  double x,xLower,xUpper,dt; // Changed!
  //int i,j,k,indexNeutralLast=0;
  int i,j,k,indexFirstLast=0; // Changed!
  double nIstate,nIStateOld; // Testing!

  // Get (bound-free) collisional, non-thermal and charge-transfer processes flags
  bool collisional=processes(pBoundFree,pCollisional);
  bool nonThermal=processes(pBoundFree,pNonThermal);
  bool chargeTransfer=processes(pBoundFree,pChargeTransfer);

  // If not steady state - Get timestep - New!
  if (not config.steadyState) dt=atom.cell()->t()-tOld;

  // For each ionization state
  for (i=0;i<atom.iState.size();i++) {
    // If not excluded - Testing
    //if (not exclude(path<<i)) {    

    // Populate matrix with bound-bound rate coefficients
    populateBMatrix(atom.iState[i],matrix,rhs,index[i],processes,path<<i,false);

    // If not steady state - Testing!
    if (not config.steadyState) {
      nIstate=atom.iState[i].n(); nIStateOld=0;
      for (j=0;j<atom.iState[i].bState.size();j++) { nIStateOld+=nOld[atom.index()][i][j]; }
    }

    // For each bound state
    for (j=0;j<atom.iState[i].bState.size();j++) {
      // If not excluded
      if (not exclude(path<<i<<j)) {

	// If LTE reference state set fraction
	//if (atom.iState[i].bState[j].LTERefState!=NULL) x=nBoundLTE(1,cell->T(),atom.iState[i].bState[j].LTERefState->E(),atom.iState[i].bState[j].E(),atom.iState[i].bState[j].LTERefState->g(),atom.iState[i].bState[j].g());
	//else x=1;
	// Get LTE reference factor
	x=xLTERef(atom.iState[i].bState[j],cell->T());

	// If not steady state - Get old number density - New!
	//if (not config.steadyState) nOld=cellOld.atom[atom.index()].iState[i].bState[j].n();

	// If next state not excluded - Testing
	//if (not exclude(path<<(i+1))) {	
	// For each upward bound-free transition
	for (k=0;k<atom.iState[i].bState[j].upTransBF.size();k++) {
	  // Set transition pointer
	  trans=atom.iState[i].bState[j].upTransBF[k];
	  // Get upper state LTE reference factor
	  xUpper=xLTERef(*(trans->upperState),cell->T());
	  // Calculate equation coefficients
	  matrix[index[i][j]][index[i+1][trans->upperState->getIndex()]]+=trans->gammaD(collisional,chargeTransfer)*xUpper; // Changed!
	  matrix[index[i][j]][index[i][j]]-=trans->gammaU(collisional,nonThermal,chargeTransfer)*x; // Changed!
	  // Add net rate in the on-the-spot range to the right hand side
	  rhs[index[i][j]]+=trans->ROSP();
	}
	//}

	// If previous ionization state not excluded - Testing
	//if (not exclude(path<<(i-1))) {	
	// For each downward bound-free transition
	for (k=0;k<atom.iState[i].bState[j].downTransBF.size();k++) {
	  // Set transition pointer
	  trans=atom.iState[i].bState[j].downTransBF[k];
	  // Get lower state LTE reference factor
	  xLower=xLTERef(*(trans->lowerState),cell->T());
	  // Calculate equation coefficients
	  matrix[index[i][j]][index[i-1][trans->lowerState->getIndex()]]+=trans->gammaU(collisional,nonThermal,chargeTransfer)*xLower; // Changed!
	  matrix[index[i][j]][index[i][j]]-=trans->gammaD(collisional,chargeTransfer)*x; // Changed!
	  // Add net rate in the on-the-spot range to the right hand side
	  rhs[index[i][j]]-=trans->ROSP();
	}
	//}

	// If not steady state and non-zero timestep - New!
	//if (not steadyState and dt!=0) {
	if (not config.steadyState and dt!=0 and atom.name().compare("H")==0) {
	  // Add dn/dt term to matrix - (ground state equation) - Testing!
	  matrix[index[i][j]][index[i][j]]-=x/dt;
	  //matrix[index[i][0]][index[i][j]]-=x/dt;
	  // Add dn/dt term to right-hand side - (ground state equation) - Testing!
	  rhs[index[i][j]]-=nOld[atom.index()][i][j]/dt;
	  //rhs[index[i][0]]-=nOld[atom.index()][i][j]/dt;
	  cout<<"NLTESolver::poulateBMatrix: cell: "<<atom.cell()->index()<<", "<<atom.iState[i].name()<<":"<<atom.iState[i].bState[j].getIndex()<<", t: "<<atom.cell()->t()/day<<", tOld: "<<tOld/day<<", dt: "<<dt/day<<", n: "<<n<<", nOld: "<<nOld[atom.index()][i][j]<<", dn: "<<(nOld[atom.index()][i][j]-n)<<", dn/dt: "<<(n-nOld[atom.index()][i][j])/dt<<"\n";
	    if (atom.iState[i].bState[j].upTransBF.size()>0) cout<<"NLTESolver::poulateBMatrix: cell: "<<atom.cell()->index()<<", "<<atom.iState[i].name()<<":"<<atom.iState[i].bState[j].getIndex()<<", RD: "<<atom.iState[i].bState[j].upTransBF[0]->RD()<<", CD: "<<atom.iState[i].bState[j].upTransBF[0]->CD()<<"\n";
	}

	// Find matrix index for last non-excluded bound state of the neutral atom
	//if (atom.iState[i].C()==0) indexNeutralLast=index[i][j];

	// Find matrix index for last non-excluded bound state of the first ionization state - Changed!
	if (i==0) indexFirstLast=index[i][j];
      }
    }
    //}
  }

  // For each ionization state - Testing!
  /*for (i=0;i<atom.iState.size();i++) {
    // For each bound state
    for (j=0;j<atom.iState[i].bState.size();j++) {
      // If not excluded
      if (not exclude(path<<i<<j)) { 
	// If not ground state
	if (j!=0) {
	  // Add equation to ground state equation
	  for (k=0;k<atom.iState[i].bState.size();k++) matrix[index[i][0]][index[i][k]]+=matrix[index[i][j]][index[i][k]];
	}
      }
    }
    }*/

  // Add the number conservation equation
  // Add reset the of the corresponding row in the matrix!!!
  for (i=0;i<atom.iState.size();i++) {
    for (j=0;j<atom.iState[i].bState.size();j++)  {
      //if (not exclude(path<<i<<j)) matrix[indexNeutralLast][index[i][j]]+=xLTERef(atom.iState[i].bState[j],cell->T());
      if (not exclude(path<<i<<j)) matrix[indexFirstLast][index[i][j]]+=xLTERef(atom.iState[i].bState[j],cell->T()); // Changed!
    }
    //rhs[indexNeutralLast]=atom.n();
    rhs[indexFirstLast]=atom.n(); // Changed!
  }
}

// Populate bound state NLTE matrix for atom
void NLTESolver::populateBMatrixDN(Atom& atom, Mat_DP& matrix, Vec_DP& rhs, const vector<vector<int> >& index, const Processes& processes, const Path& path) { // Changed! Removed const before Atom.
  Cell* cell=atom.cell();
  TransBF* trans;
  double x,xLower,xUpper,xx,n,nLower,nUpper,dt; // Changed!
  bool includeOSP;
  //int i,j,k,ii,jj,indexNeutralLast=0; // Changed!
  int i,j,k,ii,jj,indexFirstLast=0; // Changed!
  double nIstate,nIStateOld; // Testing!

  // Testing!!!
  //cout<<"NLTESolver::populateBMatrixDN: cell: "<<atom.cell()->index()<<", atom: "<<atom.name()<<", Enter\n";

  // Get (bound-free) collisional, non-thermal and charge-transfer processes flags
  bool collisional=processes(pBoundFree,pCollisional);
  bool nonThermal=processes(pBoundFree,pNonThermal);
  bool chargeTransfer=processes(pBoundFree,pChargeTransfer);

  // If not steady state - Get timestep - New!
  if (not config.steadyState) dt=atom.cell()->t()-tOld;

  // For each ionization state
  for (i=0;i<atom.iState.size();i++) {

    // Testing!!!
    //cout<<"NLTESolver::populateBMatrixDN: cell: "<<atom.cell()->index()<<", ion: "<<atom.iState[i].name()<<"\n";
    
    // If not excluded - Testing
    //if (not exclude(path<<i)) {

    // Populate matrix with bound-bound rate coefficients
    populateBMatrixDN(atom.iState[i],matrix,rhs,index[i],processes,path<<i,false);

    // If not steady state - Testing!
    if (not config.steadyState) {
      nIstate=atom.iState[i].n(); nIStateOld=0;
      for (j=0;j<atom.iState[i].bState.size();j++) { nIStateOld+=nOld[atom.index()][i][j]; }
    }

    // For each bound state
    for (j=0;j<atom.iState[i].bState.size();j++) {
      // If not excluded
      if (not exclude(path<<i<<j)) {

	// Testing!!!
	//cout<<"NLTESolver::populateBMatrixDN: cell: "<<atom.cell()->index()<<", state: "<<atom.iState[i].name()<<":"<<atom.iState[i].bState[j].getIndex()<<"\n";
	
	// Get LTE reference factor
	x=xLTERef(atom.iState[i].bState[j],cell->T());
	n=nLTERef(atom.iState[i].bState[j],cell->T());

	// If not steady state - Get old number density - New!
	//if (not config.steadyState) nOld=cellOld.atom[atom.index()].iState[i].bState[j].n();

	// If next ionization state not excluded - Testing
	//if (not exclude(path<<(i+1))) {	
	// For each upward bound-free transition
	for (k=0;k<atom.iState[i].bState[j].upTransBF.size();k++) {
	  // Set transition pointer
	  trans=atom.iState[i].bState[j].upTransBF[k];
	  // Get upper state LTE reference factor
	  xUpper=xLTERef(*(trans->upperState),cell->T());
	  nUpper=nLTERef(*(trans->upperState),cell->T());
	  // Treat OSP region separately for OSP transitions
	  includeOSP=trans->onTheSpotNew()?false:true;
	  // Calculate equation coeffiecients for downward rates - Changed!
	  matrix[index[i][j]][index[i+1][trans->upperState->getIndex()]]+=trans->gammaD(collisional,chargeTransfer,includeOSP)*xUpper;
	  //if (log() and atom.cell()->index()==36 and !isnormal(matrix[index[i][j]][index[i+1][trans->upperState->getIndex()]])) cout<<"NLTESolver::populateBMatrixDN(1): cell: "<<atom.cell()->index()<<", trans: "<<trans->upperState->iState()->name()<<":"<<trans->upperState->getIndex()<<":"<<trans->lowerState->iState()->name()<<":"<<trans->lowerState->getIndex()<<", matrix["<<index[i][j]<<"]["<<index[i+1][trans->upperState->getIndex()]<<"]: "<<matrix[index[i][j]][index[i+1][trans->upperState->getIndex()]]<<"\n";
	  // Calculate equation coefficients for upward rates - Changed!
	  // Maybe we should not include nonThermal here as this rate is not updated in each iteration ...
	  matrix[index[i][j]][index[i][j]]-=trans->gammaU(collisional,nonThermal,chargeTransfer,includeOSP)*x;
	  //if (log() and atom.cell()->index()==36 and !isnormal(matrix[index[i][j]][index[i][j]])) cout<<"NLTESolver::populateBMatrixDN(2): cell: "<<atom.cell()->index()<<", trans: "<<trans->upperState->iState()->name()<<":"<<trans->upperState->getIndex()<<":"<<trans->lowerState->iState()->name()<<":"<<trans->lowerState->getIndex()<<", matrix["<<index[i][j]<<"]["<<index[i][j]<<"]: "<<matrix[index[i][j]][index[i][j]]<<"\n";
	  // If on-the-spot transition
	  if (trans->onTheSpotNew()) {
	    // Take the implicit dependence on the level populations in the upper and lower state into account
	    matrix[index[i][j]][index[i+1][trans->upperState->getIndex()]]-=trans->dROSPdN(trans->upperState)*xUpper;
	    matrix[index[i][j]][index[i][j]]-=trans->dROSPdN(trans->lowerState)*x;
	  }
	  // Add net rates based on the previous populations to the right hand side
	  if (trans->lowerState->LTERefState==NULL and trans->upperState->LTERefState==NULL) {
	    rhs[index[i][j]]+=(trans->RNet()-trans->RDAdd()); // Changed!
	    if (collisional) rhs[index[i][j]]+=trans->CNet();
	    if (nonThermal) rhs[index[i][j]]+=trans->CNU();
	    if (chargeTransfer) rhs[index[i][j]]+=trans->CTNet(); // New!
	  }
	  else rhs[index[i][j]]-=(trans->gammaD(collisional,chargeTransfer,true)*nUpper-trans->gammaU(collisional,nonThermal,chargeTransfer,true)*n); // Changed!
	  //if (log() and atom.cell()->index()==36 and !isnormal(rhs[index[i][j]])) cout<<"NLTESolver::populateBMatrixDN(3): cell: "<<atom.cell()->index()<<", trans: "<<trans->upperState->iState()->name()<<":"<<trans->upperState->getIndex()<<":"<<trans->lowerState->iState()->name()<<":"<<trans->lowerState->getIndex()<<", rhs["<<index[i][j]<<"]: "<<rhs[index[i][j]]<<", RNet: "<<trans->RNet()<<", RDAdd: "<<trans->RNet()<<", CNet: "<<trans->CNet()<<"\n";
	  //if (log() and atom.name().compare("He")==0 and atom.cell()->index()==8) cout<<"NLTESolver::populateBMatrixDN: cell: "<<atom.cell()->index()<<", trans: "<<trans->lowerState->iState()->name()<<":"<<trans->lowerState->getIndex()<<"->"<<trans->upperState->iState()->name()<<":"<<trans->upperState->getIndex()<<", RNet: "<<trans->RNet()<<", RU: "<<trans->RU()<<", RD: "<<trans->RD()<<", RDAdd: "<<trans->RDAdd()<<", CNet: "<<trans->CNet()<<", CU: "<<trans->CU()<<", CD: "<<trans->CD()<<", CNU: "<<trans->CNU()<<"\n";
	  // If requested - Add the implicit electron number density terms - New!
	  if (config.convergence.dNedN)
	    for (ii=0;ii<atom.iState.size();ii++)
	      for (jj=0;jj<atom.iState[ii].bState.size();jj++) {
		xx=xLTERef(atom.iState[ii].bState[jj],cell->T());
		matrix[index[i][j]][index[ii][jj]]-=trans->dRNetdN(&(atom.iState[ii].bState[jj]),true,true,false)*xx;
		matrix[index[i][j]][index[ii][jj]]-=trans->dCNetdN(&(atom.iState[ii].bState[jj]),noT,true,false)*xx;
	      }
	}
	//}

	// If previous ionization state not excluded - Testing
	//if (not exclude(path<<(i-1))) {	
	// For each downward bound-free transition
	for (k=0;k<atom.iState[i].bState[j].downTransBF.size();k++) {
	  // Set transition pointer
	  trans=atom.iState[i].bState[j].downTransBF[k];
	  // Get lower state LTE reference factor
	  xLower=xLTERef(*(trans->lowerState),cell->T());
	  nLower=nLTERef(*(trans->lowerState),cell->T());
	  // Treat OSP region separately for OSP transitions
	  includeOSP=trans->onTheSpotNew()?false:true;
	  // Calculate equation coefficients for upward rates - Changed! But should be like below, right?
	  // Maybe we should not include nonThermal here as this rate is not updated in each iteration ...
	  matrix[index[i][j]][index[i-1][trans->lowerState->getIndex()]]+=trans->gammaU(collisional,nonThermal,chargeTransfer,true)*xLower;
	  //if (log() and atom.cell()->index()==36 and !isnormal(matrix[index[i][j]][index[i-1][trans->lowerState->getIndex()]])) cout<<"NLTESolver::populateBMatrixDN(4): cell: "<<atom.cell()->index()<<", trans: "<<trans->upperState->iState()->name()<<":"<<trans->upperState->getIndex()<<":"<<trans->lowerState->iState()->name()<<":"<<trans->lowerState->getIndex()<<", matrix["<<index[i][j]<<"]["<<index[i-1][trans->lowerState->getIndex()]<<"]: "<<matrix[index[i][j]][index[i-1][trans->lowerState->getIndex()]]<<"\n";
	  //matrix[index[i][j]][index[i-1][trans->lowerState->getIndex()]]+=trans->gammaU(collisional,nonThermal,chargeTransfer,includeOSP)*xLower;
	  // Calculate equation coefficients for downward rates - Changed!
	  matrix[index[i][j]][index[i][j]]-=trans->gammaD(collisional,chargeTransfer,includeOSP)*x;
	  //if (log() and atom.cell()->index()==36 and !isnormal(matrix[index[i][j]][index[i][j]])) cout<<"NLTESolver::populateBMatrixDN(5): cell: "<<atom.cell()->index()<<", trans: "<<trans->upperState->iState()->name()<<":"<<trans->upperState->getIndex()<<":"<<trans->lowerState->iState()->name()<<":"<<trans->lowerState->getIndex()<<", matrix["<<index[i][j]<<"]["<<index[i][j]<<"]: "<<matrix[index[i][j]][index[i][j]]<<"\n";
	  // If on-the-spot transition
	  if (trans->onTheSpotNew()) {
	    // Take the implicit dependence on the level populations in the upper and lower state into account
	    matrix[index[i][j]][index[i-1][trans->lowerState->getIndex()]]+=trans->dROSPdN(trans->lowerState)*xLower;
	    matrix[index[i][j]][index[i][j]]+=trans->dROSPdN(trans->upperState)*x;
	  }
	  // Add net rates based on the previous populations to the right hand side
	  if (trans->lowerState->LTERefState==NULL and trans->upperState->LTERefState==NULL) {
	    rhs[index[i][j]]-=(trans->RNet()-trans->RDAdd()); // Changed!
	    if (collisional) rhs[index[i][j]]-=trans->CNet();
	    if (nonThermal) rhs[index[i][j]]-=trans->CNU();
	    if (chargeTransfer) rhs[index[i][j]]-=trans->CTNet(); // New!
	  }
	  else rhs[index[i][j]]-=(trans->gammaU(collisional,nonThermal,chargeTransfer,true)*nLower-trans->gammaD(collisional,chargeTransfer,true)*n); // Changed!
	  //if (log() and atom.cell()->index()==36 and !isnormal(rhs[index[i][j]])) cout<<"NLTESolver::populateBMatrixDN(6): cell: "<<atom.cell()->index()<<", trans: "<<", trans: "<<trans->upperState->iState()->name()<<":"<<trans->upperState->getIndex()<<":"<<trans->lowerState->iState()->name()<<":"<<trans->lowerState->getIndex()<<", rhs["<<index[i][j]<<"]: "<<rhs[index[i][j]]<<", RNet: "<<trans->RNet()<<", RDAdd: "<<trans->RNet()<<", CNet: "<<trans->CNet()<<"\n";
	  //if (log() and atom.name().compare("He")==0 and atom.cell()->index()==8) cout<<"NLTESolver::populateBMatrixDN: cell: "<<atom.cell()->index()<<", trans: "<<", trans: "<<trans->upperState->iState()->name()<<":"<<trans->upperState->getIndex()<<"->"<<trans->lowerState->iState()->name()<<":"<<trans->lowerState->getIndex()<<", RNet: "<<trans->RNet()<<", RU: "<<trans->RU()<<", RD: "<<trans->RD()<<", RDAdd: "<<trans->RDAdd()<<", CNet: "<<trans->CNet()<<", CU: "<<trans->CU()<<", CD: "<<trans->CD()<<", CNU: "<<trans->CNU()<<"\n";
	  // If requested - Add the implicit electron number density terms - New!
	  if (config.convergence.dNedN)
	    for (ii=0;ii<atom.iState.size();ii++)
	      for (jj=0;jj<atom.iState[ii].bState.size();jj++) {
		xx=xLTERef(atom.iState[ii].bState[jj],cell->T());
		matrix[index[i][j]][index[ii][jj]]+=trans->dRNetdN(&(atom.iState[ii].bState[jj]),true,true,false)*xx;
		matrix[index[i][j]][index[ii][jj]]+=trans->dCNetdN(&(atom.iState[ii].bState[jj]),noT,true,false)*xx;
	      }
	}
	//}

	// If not steady state and non-zero timestep - New!
	if (not config.steadyState and dt!=0 and atom.name().compare("H")==0) {
	  // Add dn/dt term to matrix - (ground state equation) - Testing!
	  matrix[index[i][j]][index[i][j]]-=x/dt;
	  //matrix[index[i][0]][index[i][j]]-=x/dt;
	  // Add dn/dt term to right-hand side - (ground state equation) - Testing!
	  rhs[index[i][j]]+=(n-nOld[atom.index()][i][j])/dt;
	  //rhs[index[i][0]]+=(n-nOld[atom.index()][i][j])/dt;
	  cout<<"NLTESolver::poulateBMatrix: cell: "<<atom.cell()->index()<<", "<<atom.iState[i].name()<<":"<<atom.iState[i].bState[j].getIndex()<<", t: "<<atom.cell()->t()/day<<", tOld: "<<tOld/day<<", dt: "<<dt/day<<", n: "<<n<<", nOld: "<<nOld[atom.index()][i][j]<<", dn: "<<(nOld[atom.index()][i][j]-n)<<", dn/dt: "<<(n-nOld[atom.index()][i][j])/dt<<"\n";
	  if (atom.iState[i].bState[j].upTransBF.size()>0) cout<<"NLTESolver::poulateBMatrix: cell: "<<atom.cell()->index()<<", "<<atom.iState[i].name()<<":"<<atom.iState[i].bState[j].getIndex()<<", RD: "<<atom.iState[i].bState[j].upTransBF[0]->RD()<<", CD: "<<atom.iState[i].bState[j].upTransBF[0]->CD()<<"\n";
	}

	// Find matrix index for last non-excluded bound state of the neutral atom
	// But what if the neutral atom is not included? Fix!
	//if (atom.iState[i].C()==0) indexNeutralLast=index[i][j];

	// Find matrix index for last non-excluded bound state of the first ionization state - Changed!
	if (i==0) indexFirstLast=index[i][j];
      }
    }
    //}
  }

  // For each ionization state - Testing!
  /*for (i=0;i<atom.iState.size();i++) {
    // For each bound state
    for (j=0;j<atom.iState[i].bState.size();j++) {
      // If not excluded
      if (not exclude(path<<i<<j)) { 
	// If not ground state
	if (j!=0) {
	  // Add equation to ground state equation
	  for (k=0;k<atom.iState[i].bState.size();k++) matrix[index[i][0]][index[i][k]]+=matrix[index[i][j]][index[i][k]];
	  rhs[index[i][0]]+=rhs[index[i][j]];
	}
      }
    }
    }*/

  // Add the number conservation equation to the right-hand side
  //rhs[indexNeutralLast]=atom.n();
  rhs[indexFirstLast]=atom.n(); // Changed!
  // Reset corresponding row in the matrix - New!
  //for (i=0;i<matrix.ncols();i++) matrix[indexNeutralLast][i]=0;
  for (i=0;i<matrix.ncols();i++) matrix[indexFirstLast][i]=0; // Changed!
  // Add the number conservation equation to the matrix
  for (i=0;i<atom.iState.size();i++) {
    for (j=0;j<atom.iState[i].bState.size();j++)  {
      if (not exclude(path<<i<<j)) {
	//matrix[indexNeutralLast][index[i][j]]+=xLTERef(atom.iState[i].bState[j],cell->T());
	matrix[indexFirstLast][index[i][j]]+=xLTERef(atom.iState[i].bState[j],cell->T()); // Changed!
	//rhs[indexNeutralLast]-=nLTERef(atom.iState[i].bState[j],cell->T());
	rhs[indexFirstLast]-=nLTERef(atom.iState[i].bState[j],cell->T()); // Changed!
      }
    }
  }

  // Testing!!!
  //cout<<"NLTESolver::populateBMatrixDN: cell: "<<atom.cell()->index()<<", atom: "<<atom.name()<<", Return\n";

  // Testing!!!
  /*if (atom.name().compare("Ti")==0 and atom.cell()->index()==54)
    for (i=0;i<atom.iState.size();i++)
      for (j=0;j<atom.iState[i].bState.size();j++) {
	if (atom.iState[i].name().compare("TiII")==0 and (j==39 or j==124)) {
	  for (ii=0;ii<atom.iState.size();ii++)
	    for (jj=0;jj<atom.iState[ii].bState.size();jj++)
	      if (matrix[index[i][j]][index[ii][jj]]!=0) cout<<"NLTESolver::populateBMatrixDN: cell: "<<atom.cell()->index()<<", "<<atom.iState[i].name()<<":"<<atom.iState[i].bState[j].getIndex()<<" - "<<atom.iState[ii].name()<<":"<<atom.iState[ii].bState[jj].getIndex()<<", matrix: "<<matrix[index[i][j]][index[ii][jj]]<<"\n";
	  cout<<"NLTESolver::populateBMatrixDN: cell: "<<atom.cell()->index()<<", "<<atom.iState[i].name()<<":"<<atom.iState[i].bState[j].getIndex()<<", rhs: "<<rhs[index[i][j]]<<"\n";
	}
	}*/
}

// Populate bound state NLTE matrix for cell
void NLTESolver::populateBMatrixDN(Cell* cell, Mat_DP& matrix, Vec_DP& rhs, const vector<vector<vector<int> > >& index) {
  Processes processes;
  TransBF* trans1;
  TransBF* trans2;
  double xLower,xUpper;
  int index1,i,j,k,l,ii,jj,kk,ll;

  //if (log()) cout<<"NLTESolver::populateBMatrixDN: cell: "<<cell->index()<<", thread_num: "<<omp_get_thread_num()<<", num_threads: "<<omp_get_num_threads()<<", Enter\n";

  // Testing - Update all required tables in advance
  if (cell->bBFactorUpdate) { cell->bBFactorTable(); cell->bBFactorUpdate=false; }
  if (cell->bMFactorUpdate) { cell->bMFactorTable(); cell->bMFactorUpdate=false; }
  if (cell->boundFreeCutOffUpdate) { cell->updateBoundFreeCutOffOMP(); cell->boundFreeCutOffUpdate=false; }
  if (cell->boundFreeTableUpdate) { cell->boundFreeAbsKappaTable(true); cell->boundFreeTableUpdate=false; }
  if (cell->boundFreeSTableUpdate) { cell->boundFreeSTable(true); cell->boundFreeSTableUpdate=false; }
  //if (cell->boundFreeCutOffUpdate) { cell->updateBoundFreeCutOff(); cell->boundFreeCutOffUpdate=false; }
  //if (cell->boundFreeTableUpdate) { cell->boundFreeAbsKappaTable(); cell->boundFreeTableUpdate=false; }
  //if (cell->boundFreeSTableUpdate) { cell->boundFreeSTable(); cell->boundFreeSTableUpdate=false; }
  //if (cell->bBFactorUpdate) cell->bBFactorTable();
  //if (cell->bMFactorUpdate) cell->bMFactorTable();
  //if (cell->boundFreeCutOffUpdate) cell->updateBoundFreeCutOffOMP();
  //if (cell->boundFreeTableUpdate) cell->boundFreeAbsKappaTable(true);
  //if (cell->boundFreeSTableUpdate) cell->boundFreeSTable(true);

  // For each atom
  //#pragma omp parallel for private(processes,trans1,trans2,xLower,xUpper,j,k,l,ii,jj,kk,ll) schedule(static,1) num_threads(2)
  //#pragma omp parallel for private(processes,trans1,trans2,xLower,xUpper,j,k,l,ii,jj,kk,ll) schedule(static,1) // Changed!
  #pragma omp parallel for private(processes,trans1,trans2,xLower,xUpper,j,k,l,ii,jj,kk,ll) schedule(dynamic,1) // Changed!
  for (i=0;i<cell->atom.size();i++) {

    //if (log()) cout<<"NLTESolver::populateBMatrixDN: cell: "<<cell->index()<<", "<<cell->atom[i].name()<<", thread_num: "<<omp_get_thread_num()<<", num_threads: "<<omp_get_num_threads()<<"\n";
    
    // If not exclude
    if (not exclude(Path(i))) {

      // Get processes from configuration
      processes=config.population.processes(cell->atom[i]);

      // Populate matrix
      populateBMatrixDN(cell->atom[i],matrix,rhs,index[i],processes,Path(i));

      //if (log()) cout<<"NLTESolver::populateBMatrixDN: cell: "<<cell->index()<<", Populate matrix with "<<cell->atom[i].name()<<" coupled rates\n";

      // For each ionization state
      for (j=0;j<cell->atom[i].iState.size();j++) {

	// For each bound state
	for (k=0;k<cell->atom[i].iState[j].bState.size();k++) {
	  // If not exclude
	  if (not exclude(Path(i)<<j<<k)) {
	    	    
	    // For each upward bound-free transition
	    for (l=0;l<cell->atom[i].iState[j].bState[k].upTransBF.size();l++) {

	      // Set pointer to first transition
	      trans1=cell->atom[i].iState[j].bState[k].upTransBF[l];

	      // If on-the-spot transition
	      if (trans1->onTheSpotNew()) {
	      // If on-the-spot transition and lower state not excluded - Testing
	      //if (trans1->onTheSpotNew() and (not exclude(Path(i)<<(j+1)))) {

		// For each bound state
		for (ii=0;ii<cell->atom.size();ii++) {
		  for (jj=0;jj<cell->atom[ii].iState.size();jj++) {
		    for (kk=0;kk<cell->atom[ii].iState[jj].bState.size();kk++) {
		      // If not exclude
		      if (not (exclude(Path(ii)) or exclude(Path(ii)<<jj) or exclude(Path(ii)<<jj<<kk))) {
			
			// For each upward bound-free transition
			for (ll=0;ll<cell->atom[ii].iState[jj].bState[kk].upTransBF.size();ll++) {

			  // Set pointer to second transition
			  trans2=cell->atom[ii].iState[jj].bState[kk].upTransBF[ll];

			  // If different atoms and on-the-spot transition
			  if (i!=ii and trans2->onTheSpotNew()) {
			  // If different atoms, on-the-spot transition and lower state not excluded
			  //if (i!=ii and trans2->onTheSpotNew() and (not exclude(Path(ii)<<(jj+1)))) {

			    // Get LTE reference factors
			    xUpper=xLTERef(*(cell->atom[ii].iState[jj].bState[kk].upTransBF[ll]->upperState),cell->T());
			    xLower=xLTERef(*(cell->atom[ii].iState[jj].bState[kk].upTransBF[ll]->lowerState),cell->T());

			    // Take the implicit dependence of the lower state of the first transition on the states of the second into account
			    index1=index[i][j][k];
			    #pragma omp atomic
			    matrix[index1][index[ii][jj+1][trans2->upperState->getIndex()]]-=trans1->dROSPdN(trans2->upperState)*xUpper;
			    #pragma omp atomic
			    matrix[index1][index[ii][jj][kk]]-=trans1->dROSPdN(trans2->lowerState)*xLower;			    

			    // Take the implicit dependence of the upper state of the first transition on the states of the second into account
			    index1=index[i][j+1][trans1->upperState->getIndex()];
			    #pragma omp atomic
			    matrix[index1][index[ii][jj+1][trans2->upperState->getIndex()]]+=trans1->dROSPdN(trans2->upperState)*xUpper;
			    #pragma omp atomic
			    matrix[index1][index[ii][jj][kk]]+=trans1->dROSPdN(trans2->lowerState)*xLower;
			  }
			}
		      }
		    }
		  }
		}
	      }
	    }
	  }
	}
	//}
      }
    }
  }
}

// Populate ionization state NLTE matrix for atom
void NLTESolver::populateIMatrix(Atom& atom, Mat_DP& matrix, Vec_DP& rhs, const vector<int>& index, const Processes& processes, const Path& path) {
  TransBF* trans;
  int i,j,k;

  // Get (bound-free) collisional and non-thermal processes flags
  bool collisional=processes(pBoundFree,pCollisional);
  bool nonThermal=processes(pBoundFree,pNonThermal);
  bool chargeTransfer=processes(pBoundFree,pChargeTransfer); // New!

  // For each ionization state
  for (i=0;i<atom.iState.size();i++) {
    // If not excluded
    if (not exclude(path<<i)) {
      // For each bound state
      for (j=0;j<atom.iState[i].bState.size();j++) {

	// If next ionization state not excluded - Testing
	//if (not exclude(path<<(i+1))) {     
	// For each upward bound-free transition
	for (k=0;k<atom.iState[i].bState[j].upTransBF.size();k++) { 
	  // Set transition pointer
	  trans=atom.iState[i].bState[j].upTransBF[k];
	  // Calculate equation coefficients
	  matrix[index[i]][index[i]]-=trans->gammaU(collisional,nonThermal,chargeTransfer)*atom.iState[i].bState[j].X(); // Changed!
	  matrix[index[i]][index[i+1]]+=trans->gammaD(collisional,chargeTransfer)*trans->upperState->X(); // Changed!
	  // Add net rate in the on-the-spot range to the right hand side - New!
	  rhs[index[i]]+=trans->ROSP();
	}
	//}
	// If lower state not excluded - Testing
	//if (not exclude(path<<(i-1))) {
	// For each downward bound-free transition
	for (k=0;k<atom.iState[i].bState[j].downTransBF.size();k++) {
	  // Set transition pointer
	  trans=atom.iState[i].bState[j].downTransBF[k];
	  // Calculate equation coefficients
	  matrix[index[i]][index[i]]-=trans->gammaD(collisional,chargeTransfer)*atom.iState[i].bState[j].X(); // Changed!
	  matrix[index[i]][index[i-1]]+=trans->gammaU(collisional,nonThermal,chargeTransfer)*trans->lowerState->X(); // Changed!
	  // Add net rate in the on-the-spot range to the right hand side - New!
	  rhs[index[i]]-=trans->ROSP();
	}
	//}
      }
    }
  }

  // Add the number (fraction) conservation equation
  for (i=0;i<matrix.nrows();i++) matrix[matrix.nrows()-1][i]=1.0;
  rhs[rhs.size()-1]=1.0;
}

// Populate ionization state NLTE matrix for atom - New!
// Note: This is just a rewrite of the BF part of populateBMatrixDN
void NLTESolver::populateIMatrixDN(Atom& atom, Mat_DP& matrix, Vec_DP& rhs, const vector<int>& index, const Processes& processes, const Path& path) {
  Cell* cell=atom.cell();
  TransBF* trans;
  double x,xLower,xUpper,xx,n,nLower,nUpper;
  bool includeOSP,hasOSPTrans; // Changed!
  int i,j,k,ii,jj;
  
  //if (cell->index()==36) cout<<"NLTESolver::populateIMatrixDN: cell: "<<atom.cell()->index()<<", "<<atom.name()<<", Enter\n";

  // Get (bound-free) collisional, non-thermal and charge-transfer processes flags
  bool collisional=processes(pBoundFree,pCollisional);
  bool nonThermal=processes(pBoundFree,pNonThermal);
  bool chargeTransfer=processes(pBoundFree,pChargeTransfer);

  // For each ionization state
  for (i=0;i<atom.iState.size();i++) {
    
    /*hasOSPTrans=false;
    for (j=0;j<atom.iState[i].bState.size();j++) {
      for (k=0;k<atom.iState[i].bState[j].upTransBF.size();k++) if (atom.iState[i].bState[j].upTransBF[k]->onTheSpotNew()) hasOSPTrans=true;
      for (k=0;k<atom.iState[i].bState[j].downTransBF.size();k++) if (atom.iState[i].bState[j].downTransBF[k]->onTheSpotNew()) hasOSPTrans=true;
      }*/
    hasOSPTrans=true;
    
    if (not hasOSPTrans) {

      for (j=MAX(i-1,0);j<MIN(i+2,int(atom.iState.size()));j++) {
	// Calculate matrix coefficients
	matrix[index[i]][index[j]]-=atom.iState[i].dRdN(&(atom.iState[j]),true,false);
	if (collisional) matrix[index[i]][index[j]]-=atom.iState[i].dCCdN(&(atom.iState[j]),noT,false);
	if (nonThermal) matrix[index[i]][index[j]]-=atom.iState[i].dCNdN(&(atom.iState[j]));
	if (chargeTransfer) matrix[index[i]][index[j]]-=atom.iState[i].dCTdN(&(atom.iState[j]));
      }
      
      // Add net rates based on the previous populations to the right hand side
      rhs[index[i]]+=atom.iState[i].R();
      if (collisional) rhs[index[i]]+=atom.iState[i].CC();
      if (nonThermal) rhs[index[i]]+=atom.iState[i].CN();
      if (chargeTransfer) rhs[index[i]]+=atom.iState[i].CT();
    }
    else {
      
      // For each bound state
      for (j=0;j<atom.iState[i].bState.size();j++) {
	// If not excluded
	if (not exclude(path<<i<<j)) {
	  
	  // Get LTE reference factor
	  x=atom.iState[i].bState[j].X();
	  n=atom.iState[i].bState[j].n();
	  
	  // For each upward bound-free transition
	  for (k=0;k<atom.iState[i].bState[j].upTransBF.size();k++) {
	    // Set transition pointer
	    trans=atom.iState[i].bState[j].upTransBF[k];
	    // Get upper state LTE reference factor
	    xUpper=trans->upperState->X();
	    nUpper=trans->upperState->n();
	    // Treat OSP region separately for OSP transitions
	    includeOSP=trans->onTheSpotNew()?false:true;
	    // Calculate equation coeffiecients for downward rates - Changed!
	    matrix[index[i]][index[i+1]]+=trans->gammaD(collisional,chargeTransfer,includeOSP)*xUpper;
	    // Calculate equation coefficients for upward rates - Changed!
	    // Maybe we should not include nonThermal here as this rate is not updated in each iteration ...
	    matrix[index[i]][index[i]]-=trans->gammaU(collisional,nonThermal,chargeTransfer,includeOSP)*x;
	    // If on-the-spot transition
	    if (trans->onTheSpotNew()) {
	      // Take the implicit dependence on the level populations in the upper and lower state into account
	      matrix[index[i]][index[i+1]]-=trans->dROSPdN(trans->upperState)*xUpper;
	      matrix[index[i]][index[i]]-=trans->dROSPdN(trans->lowerState)*x;
	    }
	    // Add net rates based on the previous populations to the right hand side
	    rhs[index[i]]+=(trans->RNet()-trans->RDAdd());
	    if (collisional) rhs[index[i]]+=trans->CNet();
	    if (nonThermal) rhs[index[i]]+=trans->CNU();
	    if (chargeTransfer) rhs[index[i]]+=trans->CTNet();
	    // If requested - Add the implicit electron number density terms 
	    if (config.convergence.dNedN)
	      for (ii=0;ii<atom.iState.size();ii++)
		for (jj=0;jj<atom.iState[ii].bState.size();jj++) {
		  xx=atom.iState[ii].bState[jj].X();
		  matrix[index[i]][index[ii]]-=trans->dRNetdN(&(atom.iState[ii].bState[jj]),true,true,false)*xx;
		  matrix[index[i]][index[ii]]-=trans->dCNetdN(&(atom.iState[ii].bState[jj]),noT,true,false)*xx;
		}
	  }
	  
	  // For each downward bound-free transition
	  for (k=0;k<atom.iState[i].bState[j].downTransBF.size();k++) {
	    // Set transition pointer
	    trans=atom.iState[i].bState[j].downTransBF[k];
	    // Get lower state LTE reference factor
	    xLower=trans->lowerState->X();
	    nLower=trans->lowerState->n();
	    // Treat OSP region separately for OSP transitions
	    includeOSP=trans->onTheSpotNew()?false:true;
	    // Calculate equation coefficients for upward rates - Changed! But should be like below, right?
	    // Maybe we should not include nonThermal here as this rate is not updated in each iteration ...
	    matrix[index[i]][index[i-1]]+=trans->gammaU(collisional,nonThermal,chargeTransfer,true)*xLower;
	    //matrix[index[i]][index[i-1]]+=trans->gammaU(collisional,nonThermal,chargeTransfer,includeOSP)*xLower;
	    // Calculate equation coefficients for downward rates - Changed!
	    matrix[index[i]][index[i]]-=trans->gammaD(collisional,chargeTransfer,includeOSP)*x;
	    // If on-the-spot transition
	    if (trans->onTheSpotNew()) {
	      // Take the implicit dependence on the level populations in the upper and lower state into account
	      matrix[index[i]][index[i-1]]+=trans->dROSPdN(trans->lowerState)*xLower;
	      matrix[index[i]][index[i]]+=trans->dROSPdN(trans->upperState)*x;
	    }
	    // Add net rates based on the previous populations to the right hand side
	    rhs[index[i]]-=(trans->RNet()-trans->RDAdd()); // Changed!
	    if (collisional) rhs[index[i]]-=trans->CNet();
	    if (nonThermal) rhs[index[i]]-=trans->CNU();
	    if (chargeTransfer) rhs[index[i]]-=trans->CTNet(); // New!
	    // If requested - Add the implicit electron number density terms - New!
	    if (config.convergence.dNedN)
	      for (ii=0;ii<atom.iState.size();ii++)
		for (jj=0;jj<atom.iState[ii].bState.size();jj++) {
		  xx=atom.iState[ii].bState[jj].X();
		  matrix[index[i]][index[ii]]+=trans->dRNetdN(&(atom.iState[ii].bState[jj]),true,true,false)*xx;
		  matrix[index[i]][index[ii]]+=trans->dCNetdN(&(atom.iState[ii].bState[jj]),noT,true,false)*xx;
		}
	  }
	}
      }
    }
  }

  /*if (cell->index()==36) 
    for (i=0;i<atom.iState.size();i++)
    for (j=MAX(i-1,0);j<MIN(i+2,int(atom.iState.size()));j++) cout<<"NLTESolver::populateMatrix: cell: "<<cell->index()<<", Matrix("<<atom.iState[i].name()<<","<<atom.iState[j].name()<<"): "<<matrix[index[i]][index[j]]<<", RHS("<<atom.iState[i].name()<<"): "<<rhs[index[i]]<<"\n";*/
    
  // Add the number conservation equation
  for (i=0;i<matrix.nrows();i++) matrix[matrix.nrows()-1][i]=1.0;
  rhs[rhs.size()-1]=0;

  //if (cell->index()==36) cout<<"NLTESolver::populateIMatrixDN: cell: "<<atom.cell()->index()<<", "<<atom.name()<<", Return\n";
}

// Populate ionization state NLTE matrix for cell - New!
void NLTESolver::populateIMatrix(Cell* cell, Mat_DP& matrix, Vec_DP& rhs, const vector<vector<int> >& index) {
  Processes processes;
  int i;

  // For each atom
  for (i=0;i<cell->atom.size();i++) {
    
    // If not exclude
    if (not exclude(Path(i))) {

      // Get processes from configuration
      processes=config.population.processes(cell->atom[i]);
      
      // Populate matrix
      populateIMatrix(cell->atom[i],matrix,rhs,index[i],processes,Path(i));
    }
  }
}

// Populate ionization state NLTE matrix for cell - New!
void NLTESolver::populateIMatrixDN(Cell* cell, Mat_DP& matrix, Vec_DP& rhs, const vector<vector<int> >& index) {
  Processes processes;
  bool collisional,nonThermal,chargeTransfer;
  int i,j,ii,jj;
  bool dNedN=config.convergence.dNedN;

  // For each atom
  for (i=0;i<cell->atom.size();i++) {
    // If not exclude
    if (not exclude(Path(i))) {

      // Get processes from configuration
      processes=config.population.processes(cell->atom[i]);

      // Get (bound-free) collisional, non-thermal and charge-transfer processes flags
      collisional=processes(pBoundFree,pCollisional);
      nonThermal=processes(pBoundFree,pNonThermal);
      chargeTransfer=processes(pBoundFree,pChargeTransfer);
      
      // For each ionization state
      for (j=0;j<cell->atom[i].iState.size();j++) {
        // If not excluded
	if (not exclude(Path(i)<<j)) {

	  // For each ionization state
	  for (ii=0;ii<cell->atom.size();ii++) {
	    for (jj=0;jj<cell->atom[ii].iState.size();jj++) {
	      // If not exclude
	      if (not (exclude(Path(ii)) or exclude(Path(ii)<<jj))) {

		// Calculate matrix coefficients
		matrix[index[i][j]][index[ii][jj]]-=cell->atom[i].iState[j].dRdN(&(cell->atom[ii].iState[jj]),true,dNedN);
		if (collisional) matrix[index[i][j]][index[ii][jj]]-=cell->atom[i].iState[j].dCCdN(&(cell->atom[ii].iState[jj]),noT,dNedN);
		// Maybe we should not include the nonThermal term here as this rate is not updated in each iteration ...
		if (nonThermal) matrix[index[i][j]][index[ii][jj]]-=cell->atom[i].iState[j].dCNdN(&(cell->atom[ii].iState[jj]));
		if (chargeTransfer) matrix[index[i][j]][index[ii][jj]]-=cell->atom[i].iState[j].dCTdN(&(cell->atom[ii].iState[jj]));
	      }
	    }
	  }

	  // Add net rates based on the previous populations to the right-hand side
	  rhs[index[i][j]]+=cell->atom[i].iState[j].R();
	  if (collisional) rhs[index[i][j]]+=cell->atom[i].iState[j].CC();
	  if (nonThermal) rhs[index[i][j]]+=cell->atom[i].iState[j].CN();
	  if (chargeTransfer) rhs[index[i][j]]+=cell->atom[i].iState[j].CT();
	}
      }
    }
  }
  
  // For each atom
  for (i=0;i<cell->atom.size();i++) {
    // If not excluded
    if (not exclude(Path(i))) {
      // Add the number conservation equation to the right-hand side
      rhs[index[i][index[i].size()-1]]=0;
      // Reset corresponding row in the matrix
      for (j=0;j<matrix.ncols();j++) matrix[index[i][index[i].size()-1]][j]=0;
      // Add the number conservation equation to the matrix
      for (j=0;j<cell->atom[i].iState.size();j++) {
	if (not exclude(Path(i)<<j)) {
	  matrix[index[i][index[i].size()-1]][index[i][j]]=1;
	}
      }
    }
  }
}

// Populate ionization state NLTE matrix for atom (test version 1)
void NLTESolver::populateIMatrixDN1(Atom& atom, Mat_DP& matrix, Vec_DP& rhs, const vector<int>& index, const Processes& processes, const Path& path) {
  Cell* cell=atom.cell();
  TransBF* trans;
  bool includeOSP;
  int i,j,k,ii,jj;

  // Get (bound-free) collisional, non-thermal and charge-transfer processes flags
  bool collisional=processes(pBoundFree,pCollisional);
  bool nonThermal=processes(pBoundFree,pNonThermal);
  bool chargeTransfer=processes(pBoundFree,pChargeTransfer);

  // For each ionization state
  for (i=0;i<atom.iState.size();i++) {

    // For each bound state
    for (j=0;j<atom.iState[i].bState.size();j++) {
      // If not excluded
      if (not exclude(path<<i<<j)) {

	// For each upward bound-free transition
	for (k=0;k<atom.iState[i].bState[j].upTransBF.size();k++) {
	  // Set transition pointer
	  trans=atom.iState[i].bState[j].upTransBF[k];
	  // Treat OSP region separately for OSP transitions
	  includeOSP=trans->onTheSpotNew()?false:true;
	  // Calculate equation coeffiecients for downward rates
	  matrix[index[i]][index[i+1]]+=trans->gammaD(collisional,chargeTransfer,includeOSP)*trans->upperState->X();
	  // Calculate equation coefficients for upward rates
	  // Maybe we should not include nonThermal here as this rate is not updated in each iteration ...
	  matrix[index[i]][index[i]]-=trans->gammaU(collisional,nonThermal,chargeTransfer,includeOSP)*trans->lowerState->X();
	  // Testing!
	  //if (log() and atom.cell()->index()==2 and atom.iState[i].name().compare("HeI")==0) cout<<"populateIMatrixDN1: cell: "<<atom.cell()->index()<<", "<<atom.iState[i].name()<<":"<<atom.iState[i].name()<<"."<<atom.iState[i].bState[j].getIndex()<<", gammaU: "<<(-trans->gammaU(collisional,nonThermal,chargeTransfer,includeOSP)*trans->lowerState->X())<<", gammaU(CT): "<<(-(trans->gammaU(collisional,nonThermal,chargeTransfer,includeOSP)-trans->gammaU(collisional,nonThermal,false,includeOSP))*trans->lowerState->X())<<", gammaU(NT): "<<(-(trans->gammaU(collisional,nonThermal,chargeTransfer,includeOSP)-trans->gammaU(collisional,false,chargeTransfer,includeOSP))*trans->lowerState->X())<<"\n";
	  // If on-the-spot transition
	  if (trans->onTheSpotNew()) {
	    // Take the implicit dependence on the level populations in the upper and lower state into account
	    matrix[index[i]][index[i+1]]-=trans->dROSPdN(trans->upperState)*trans->upperState->X();
	    matrix[index[i]][index[i]]-=trans->dROSPdN(trans->lowerState)*trans->lowerState->X();
	  }
	  // Add net rates based on the previous populations to the right hand side
	  rhs[index[i]]+=(trans->RNet()-trans->RDAdd());
	  if (collisional) rhs[index[i]]+=trans->CNet();
	  if (nonThermal) rhs[index[i]]+=trans->CNU();
	  if (chargeTransfer) rhs[index[i]]+=trans->CTNet();
	  // If requested - Add the implicit electron number density terms - New!
	  if (config.convergence.dNedN)
	    for (ii=0;ii<atom.iState.size();ii++)
	      for (jj=0;jj<atom.iState[ii].bState.size();jj++) {
		matrix[index[i]][index[ii]]-=trans->dRNetdN(&(atom.iState[ii].bState[jj]),true,true,false)*atom.iState[ii].bState[jj].X();
		matrix[index[i]][index[ii]]-=trans->dCNetdN(&(atom.iState[ii].bState[jj]),noT,true,false)*atom.iState[ii].bState[jj].X();
	      }
	}

	// For each downward bound-free transition
	for (k=0;k<atom.iState[i].bState[j].downTransBF.size();k++) {
	  // Set transition pointer
	  trans=atom.iState[i].bState[j].downTransBF[k];
	  // Treat OSP region separately for OSP transitions
	  includeOSP=trans->onTheSpotNew()?false:true;
	  // Calculate equation coefficients for upward rates - Should be like below, right?
	  // Maybe we should not include nonThermal here as this rate is not updated in each iteration ...
	  matrix[index[i]][index[i-1]]+=trans->gammaU(collisional,nonThermal,chargeTransfer,true)*trans->lowerState->X();
	  //matrix[index[i][j]][index[i-1][trans->lowerState->getIndex()]]+=trans->gammaU(collisional,nonThermal,chargeTransfer,includeOSP)*xLower;
	  // Calculate equation coefficients for downward rates
	  matrix[index[i]][index[i]]-=trans->gammaD(collisional,chargeTransfer,includeOSP)*trans->upperState->X();
	  // Testing!
	  //if (log() and atom.cell()->index()==2 and atom.iState[i].name().compare("HeI")==0) cout<<"populateIMatrixDN1: cell: "<<atom.cell()->index()<<", "<<atom.iState[i].name()<<":"<<atom.iState[i].name()<<"."<<atom.iState[i].bState[j].getIndex()<<", gammaD: "<<(-trans->gammaD(collisional,chargeTransfer,includeOSP)*trans->upperState->X())<<", gammaD(CT): "<<(-(trans->gammaD(collisional,chargeTransfer,includeOSP)-trans->gammaD(collisional,false,includeOSP))*trans->upperState->X())<<"\n";
	  // If on-the-spot transition
	  if (trans->onTheSpotNew()) {
	    // Take the implicit dependence on the level populations in the upper and lower state into account
	    matrix[index[i]][index[i-1]]+=trans->dROSPdN(trans->lowerState)*trans->lowerState->X();
	    matrix[index[i]][index[i]]+=trans->dROSPdN(trans->upperState)*trans->upperState->X();
	  }
	  // Add net rates based on the previous populations to the right hand side
	  rhs[index[i]]-=(trans->RNet()-trans->RDAdd());
	  if (collisional) rhs[index[i]]-=trans->CNet();
	  if (nonThermal) rhs[index[i]]-=trans->CNU();
	  if (chargeTransfer) rhs[index[i]]-=trans->CTNet();
	  // If requested - Add the implicit electron number density terms - New!
	  if (config.convergence.dNedN)
	    for (ii=0;ii<atom.iState.size();ii++)
	      for (jj=0;jj<atom.iState[ii].bState.size();jj++) {
		matrix[index[i]][index[ii]]+=trans->dRNetdN(&(atom.iState[ii].bState[jj]),true,true,false)*atom.iState[ii].bState[jj].X();
		matrix[index[i]][index[ii]]+=trans->dCNetdN(&(atom.iState[ii].bState[jj]),noT,true,false)*atom.iState[ii].bState[jj].X();
	      }
	}
      }
    }
  }

  // Add number conservation equation
  rhs[index[index.size()-1]]=0;
  for (i=0;i<atom.iState.size();i++) {
    if (not exclude(Path(i))) {
      matrix[index[index.size()-1]][index[i]]=1;
    }
  }
}

// Populate ionization state NLTE matrix for cell (test version 1)
void NLTESolver::populateIMatrixDN1(Cell* cell, Mat_DP& matrix, Vec_DP& rhs, const vector<vector<int> >& index) {
  Processes processes;
  int i;

  // For each atom
  for (i=0;i<cell->atom.size();i++) {
    
    // If not exclude
    if (not exclude(Path(i))) {

      // Get processes from configuration
      processes=config.population.processes(cell->atom[i]);

      // Populate matrix
      populateIMatrixDN1(cell->atom[i],matrix,rhs,index[i],processes,Path(i));
    }
  }
}

// Populate ionization state NLTE matrix for cell (test version 2)
void NLTESolver::populateIMatrixDN2(Cell* cell, Mat_DP& matrix, Vec_DP& rhs, const vector<vector<int> >& index) {
  Processes processes;
  bool collisional,nonThermal,chargeTransfer;
  bool dNedN=config.convergence.dNedN;
  TransBF* trans;
  int i,j,k,l,ii,jj,kk;

  // For each atom
  for (i=0;i<cell->atom.size();i++) {
    // If not exclude
    if (not exclude(Path(i))) {

      // Get processes from configuration
      processes=config.population.processes(cell->atom[i]);

      // Get (bound-free) collisional, non-thermal and charge-transfer processes flags
      collisional=processes(pBoundFree,pCollisional);
      nonThermal=processes(pBoundFree,pNonThermal);
      chargeTransfer=processes(pBoundFree,pChargeTransfer);

      // For each bound state
      for (j=0;j<cell->atom[i].iState.size();j++) {
	for (k=0;k<cell->atom[i].iState[j].bState.size();k++) {
	  // If not excluded
	  if (not exclude(Path(i)<<j<<k)) {

	    // For each upward bound-free transition
	    for (l=0;l<cell->atom[i].iState[j].bState[k].upTransBF.size();l++) {
	      // Set transition pointer
	      trans=cell->atom[i].iState[j].bState[k].upTransBF[l];
	    
	      // For each bound state
	      for (ii=0;ii<cell->atom.size();ii++) {
		for (jj=0;jj<cell->atom[ii].iState.size();jj++) {
		  for (kk=0;kk<cell->atom[kk].iState.size();kk++) {
		    // If not exclude
		    if (not (exclude(Path(ii)) or exclude(Path(ii)<<jj) or exclude(Path(ii)<<jj<<kk))) {

		      // Add net rate number derivatives to the coefficient matrix
		      matrix[index[i][j]][index[ii][jj]]-=trans->dRNetdN(&(cell->atom[ii].iState[jj].bState[kk]),true,dNedN);
		      if (collisional) matrix[index[i][j]][index[ii][jj]]-=trans->dCNetdN(&(cell->atom[ii].iState[jj].bState[kk]),noT,dNedN);
		      // Maybe we should not include the nonThermal term here as this rate is not updated in each iteration ...		      
		      //if (nonThermal) matrix[index[i][j]][index[ii][jj]]-=trans->dCNUdN(&(cell->atom[ii].iState[jj].bState[kk]));
		      if (chargeTransfer) matrix[index[i][j]][index[ii][jj]]-=trans->dCTNetdN(&(cell->atom[ii].iState[jj].bState[kk]));
		    }
		  }
		}
	      }
	      
	      // Add net rates based on the previous populations to the right-hand side
	      rhs[index[i][j]]+=trans->RNet();
	      if (collisional) rhs[index[i][j]]+=trans->CNet();
	      if (nonThermal) rhs[index[i][j]]+=trans->CNU();
	      if (chargeTransfer) rhs[index[i][j]]+=trans->CTNet();
	    }
	    
	    // For each downward bound-free transition
	    for (l=0;l<cell->atom[i].iState[j].bState[k].downTransBF.size();l++) {
	      // Set transition pointer
	      trans=cell->atom[i].iState[j].bState[k].downTransBF[l];
	      
	      // For each bound state
	      for (ii=0;ii<cell->atom.size();ii++) {
		for (jj=0;jj<cell->atom[ii].iState.size();jj++) {
		  for (kk=0;kk<cell->atom[kk].iState.size();kk++) {
		    // If not exclude
		    if (not (exclude(Path(ii)) or exclude(Path(ii)<<jj) or exclude(Path(ii)<<jj<<kk))) {
		      
		      // Add net rate derivatives to the coefficient matrix
		      matrix[index[i][j]][index[ii][jj]]+=trans->dRNetdN(&(cell->atom[ii].iState[jj].bState[kk]),true,dNedN);
		      if (collisional) matrix[index[i][j]][index[ii][jj]]+=trans->dCNetdN(&(cell->atom[ii].iState[jj].bState[kk]),noT,dNedN);
		      // Maybe we should not include the nonThermal term here as this rate is not updated in each iteration ...		      
		      //if (nonThermal) matrix[index[i][j]][index[ii][jj]]+=trans->dCNUdN(&(cell->atom[ii].iState[jj].bState[kk]));
		      if (chargeTransfer) matrix[index[i][j]][index[ii][jj]]+=trans->dCTNetdN(&(cell->atom[ii].iState[jj].bState[kk]));
		    }
		  }
		}
	      }
	      
	      // Add net rates based on the previous populations to the right-hand side
	      rhs[index[i][j]]-=trans->RNet();
	      if (collisional) rhs[index[i][j]]-=trans->CNet();
	      if (nonThermal) rhs[index[i][j]]-=trans->CNU();
	      if (chargeTransfer) rhs[index[i][j]]-=trans->CTNet();
	    }
	  }
	}
      }
    }
  }
      
  // For each atom
  for (i=0;i<cell->atom.size();i++) {
    // If not excluded
    if (not exclude(Path(i))) {
      // Add the number conservation equation to the right-hand side
      rhs[index[i][index[i].size()-1]]=0;
      // Reset corresponding row in the matrix
      for (j=0;j<matrix.ncols();j++) matrix[index[i][index[i].size()-1]][j]=0;
      // Add the number conservation equation to the matrix
      for (j=0;j<cell->atom[i].iState.size();j++) {
	if (not exclude(Path(i)<<j)) {
	  matrix[index[i][index[i].size()-1]][index[i][j]]=1;
	}
      }
    }
  }
}

// Log transition rates for cell
void NLTESolver::logGamma(Cell* cell) {
  TransBF* transBF;
  TransBB* transBB;
  int i,j,k,l;

  // For each atom
  for (i=0;i<cell->atom.size();i++) {
    // Testing!
    if (cell->atom[i].name().compare("He")==0) {
      // If not excluded
      if (not exclude(Path(i))) {
	
	// For each ionization state
	for (j=0;j<cell->atom[i].iState.size();j++) {
	  // If not excluded
	  if (not exclude(Path(i)<<j)) {
	    
	    // For each bound state
	    for (k=0;k<cell->atom[i].iState[j].bState.size();k++) {
	      // If not excluded
	      if (not exclude(Path(i)<<j<<k)) {
		
		// For each upward bound-free transition
		for (l=0;l<cell->atom[i].iState[j].bState[k].upTransBF.size();l++) {
		  // Set transition pointer
		  transBF=cell->atom[i].iState[j].bState[k].upTransBF[l]; 
		  cout<<"NLTESolver::logGamma: cell: "<<cell->index()<<", "<<transBF->lowerState->iState()->name()<<":"<<transBF->lowerState->getIndex()<<"->"<<transBF->upperState->iState()->name()<<":"<<transBF->upperState->getIndex()<<", lambda: "<<(c/transBF->nu())/AA<<", bound-free (radiative): "<<transBF->gammaU(false,false,false)<<", bound-free (collisional): "<<transBF->gammaU(true,false,false)-transBF->gammaU(false,false,false)<<", bound-free (non-thermal): "<<transBF->gammaU(false,true,false)-transBF->gammaU(false,false,false)<<"\n"; // Changed!
		  cout<<"NLTESolver::logGamma: cell: "<<cell->index()<<", "<<transBF->upperState->iState()->name()<<":"<<transBF->upperState->getIndex()<<"->"<<transBF->lowerState->iState()->name()<<":"<<transBF->lowerState->getIndex()<<", lambda: "<<(c/transBF->nu())/AA<<", bound-free (radiative): "<<transBF->gammaD(false,false)<<", bound-free (collisional): "<<transBF->gammaD(true,false)-transBF->gammaD(false,false)<<"\n"; // Changed!
		}
		
		// For each upward bound-bound transition
		for (l=0;l<cell->atom[i].iState[j].bState[k].upTransBB.size();l++) {
		  // Set transition pointer
		  transBB=cell->atom[i].iState[j].bState[k].upTransBB[l]; 
		  cout<<"NLTESolver::logGamma: cell: "<<cell->index()<<", "<<transBB->lowerState->iState()->name()<<":"<<transBB->lowerState->getIndex()<<"->"<<transBB->upperState->getIndex()<<", lambda: "<<(c/transBB->nu())/AA<<", bound-bound (radiative): "<<transBB->gammaU(false,false)<<", bound-bound (collisional): "<<transBB->gammaU(true,false)-transBB->gammaU(false,false)<<", bound-bound (non-thermal): "<<transBB->gammaU(false,true)-transBB->gammaU(false,false)<<"\n"; // Changed!
		  cout<<"NLTESolver::logGamma: cell: "<<cell->index()<<", "<<transBB->upperState->iState()->name()<<":"<<transBB->upperState->getIndex()<<"->"<<transBB->lowerState->getIndex()<<", lambda: "<<(c/transBB->nu())/AA<<", bound-bound (radiative): "<<transBB->gammaD(false,true)<<", bound-bound (collisional): "<<transBB->gammaD(true,true)-transBB->gammaD(false,true)<<"\n"; // Changed!
		}
	      }
	    }
	  }
	}
      }
    }
  }
}

// Matrix singular
bool matrixSingular(Mat_DP& matrix) {
  bool singular=false;
  double max;
  int i,j;
  
  for (i=0;i<matrix.nrows();i++) {
    max=0;
    for (j=0;j<matrix.ncols();j++) {
      if (abs(matrix[i][j])>max) max=abs(matrix[i][j]); // Changed!
    }
    if (max==0) { singular=true; cout<<"NLTESolver::matrixSingular: Only zeros in row "<<i<<"\n"; break; }
  }

  if (singular) 
    for (i=0;i<matrix.nrows();i++)
      for (j=0;j<matrix.ncols();j++)
	cout<<"NLTESolver::matrixSingular: matrix["<<i<<","<<j<<"]: "<<matrix[i][j]<<"\n";

  return singular;
}

// Get LTE reference factor
double NLTESolver::xLTERef(const BState& bState, const double& T) {
  BState* refState=bState.LTERefState;

  if (refState!=NULL) {
    //return nBoundLTE(1,T,refState->E(),bState.E(),refState->g(),bState.g());
    // Changed!
    if (refState->iState()==bState.iState()) return nBoundLTE(1,T,refState->E(),bState.E(),refState->g(),bState.g());
    else { // New block!
      Atom* atom=bState.iState()->atom();
      Cell* cell=atom->cell();
      int i,iStateIndex=bState.iState()->index(),LTERefIStateIndex=refState->iState()->index();
      double Xi0,g1,g2,x=1.0;
      // Note: Assumes LTERefStateIndex<iStateIndex!
      for (i=LTERefIStateIndex;i<iStateIndex;i++) {
	if (i==LTERefIStateIndex) { Xi0=atom->iState[i].Xi0()-refState->E(); g1=refState->g(); g2=1.0; }
	else if (i==(iStateIndex-1)) { Xi0=atom->iState[i].Xi0()+bState.E(); g1=1.0; g2=bState.g(); }
	else { Xi0=atom->iState[i].Xi0(); g1=1.0; g2=1.0; }
	// Why is T a parameter and ne not? Fix!
	x=x*nIonLTE(1,cell->ne(),T,Xi0,g1,g2);
      }
      return x;
    }
  }
  else return 1.0;
}

// Get LTE reference number density
//double NLTESolver::nLTERef(const BState& bState, const double& T) {
double NLTESolver::nLTERef(BState& bState, const double& T) { // Changed! - Testing!
  BState* refState=bState.LTERefState;

  if (refState!=NULL) return xLTERef(bState,T)*refState->n();
  else return bState.n();
}

// Get old LTE reference factor
/*double NLTESolver::xLTERefOld(const BState& bState, const double& T) {
  BState* refState=bState.LTERefState;

  if (refState!=NULL) return bState.n()/refState->n();
  else return 1.0;
  }*/

// (Internal) LTE group transition
bool NLTESolver::LTEGroupTrans(TransBB* transBB) const {
  BState* lowerState=transBB->lowerState;
  BState* upperState=transBB->upperState;

  // If lower state LTE reference
  if (lowerState->LTERefState!=NULL)
    // If lower state LTE reference equals upper state or upper state LTE reference - Return true
    if (lowerState->LTERefState==upperState or lowerState->LTERefState==upperState->LTERefState) return true;

  // If upper state LTE reference
  if (upperState->LTERefState!=NULL)
    // If upper state LTE reference equals lower state or lower state LTE reference - Return true    
    if (upperState->LTERefState==lowerState or upperState->LTERefState==lowerState->LTERefState) return true;

  // Return false
  return false;
}

// Expand LTE groups (in terms of number fraction/density)
void NLTESolver::expandLTEGroups(IState& iState, vector<double>& X, const Path& path) {
  Cell* cell=iState.atom()->cell();
  double x;
  //vector<double> xSum(iState.bState.size(),0);
  BState* refState;
  int i;

  // For each bound state
  for (i=0;i<iState.bState.size();i++) {
    // If not excluded and LTE reference state
    if ((not exclude(path<<i)) and iState.bState[i].LTERefState!=NULL) {
      refState=iState.bState[i].LTERefState;
      // Get relative number fraction
      //x=nBoundLTE(1,cell->T(),refState->E(),iState.bState[i].E(),refState->g(),iState.bState[i].g());
      // Get LTE reference factor
      x=xLTERef(iState.bState[i],cell->T());
      // Set number fraction/density
      X[i]=X[refState->getIndex()]*x;
      // Sum relative number fraction
      //xSum[refState->getIndex()]+=x;
    }
  }
}

// Expand LTE groups (in terms of number density) - New!
void NLTESolver::expandLTEGroupsN(Atom& atom, vector<vector<double> >& N, const Path& path) {
  Cell* cell=atom.cell();
  double x;
  //vector<double> xSum(iState.bState.size(),0);
  BState* refState;
  int i,j;

  // For each ion
  for (i=0;i<atom.iState.size();i++) {
    // For each bound state
    for (j=0;j<atom.iState[i].bState.size();j++) {
      // If not excluded and LTE reference state
      if ((not exclude(path<<i<<j)) and atom.iState[i].bState[j].LTERefState!=NULL) {
	refState=atom.iState[i].bState[j].LTERefState;
	// Get relative number fraction
	//x=nBoundLTE(1,cell->T(),refState->E(),iState.bState[i].E(),refState->g(),iState.bState[i].g());
	// Get LTE reference factor
	x=xLTERef(atom.iState[i].bState[j],cell->T());
	// Set number fraction/density
	N[i][j]=N[refState->iState()->index()][refState->getIndex()]*x;
	// Sum relative number fraction
	//xSum[refState->getIndex()]+=x;
      }
    }
  }
}

// Expand LTE groups (in terms of number density corrections)
void NLTESolver::expandLTEGroupsDN(IState& iState, vector<double>& dN, const Path& path) {
  Cell* cell=iState.atom()->cell();
  double x;
  BState* refState;
  int i;

  // For each bound state
  for (i=0;i<iState.bState.size();i++) {
    // If not excluded and LTE reference state
    if ((not exclude(path<<i)) and iState.bState[i].LTERefState!=NULL) {
      refState=iState.bState[i].LTERefState;
      // Get LTE reference factor
      x=xLTERef(iState.bState[i],cell->T());
      // Calculate number fraction/density correction
      dN[i]=(refState->n()+dN[refState->getIndex()])*x-iState.bState[i].n();
    }
  }
}

// Expand LTE groups (in terms of number density corrections) - New!
void NLTESolver::expandLTEGroupsDN(Atom& atom, vector<vector<double> >& dN, const Path& path) {
  Cell* cell=atom.cell();
  double x;
  BState* refState;
  int i,j;

  //cout<<"NLTESolver::expandLTEGroupsDN: cell: "<<atom.cell()->index()<<", atom: "<<atom.name()<<", Enter\n";

  // For each ion
  for (i=0;i<atom.iState.size();i++) {
    // For each bound state
    for (j=0;j<atom.iState[i].bState.size();j++) {
      // If not excluded and LTE reference state
      if ((not exclude(path<<i<<j)) and atom.iState[i].bState[j].LTERefState!=NULL) {
	refState=atom.iState[i].bState[j].LTERefState;
	// Get LTE reference factor
	x=xLTERef(atom.iState[i].bState[j],cell->T());
	// Calculate number fraction/density correction
	dN[i][j]=(refState->n()+dN[refState->iState()->index()][refState->getIndex()])*x-atom.iState[i].bState[j].n();
      }
    }
  }

  //cout<<"NLTESolver::expandLTEGroupSDN: cell: "<<atom.cell()->index()<<", atom: "<<atom.name()<<", Return\n";
}

// Solve for bound state of ion
void NLTESolver::solveBoundState(IState& iState, vector<double>& X, const Processes& processes, const Path& path) {

  vector<int> index;
  int i,nStates=0;
  Vec_DP rhs;
  Mat_DP matrix;
  Vec_INT ii;
  DP dd;

  // Make NLTE matrix index map
  index=bMatrixIndex(iState,nStates,path);
      
  // Initialize
  matrix=Mat_DP(0.0,nStates,nStates);
  rhs=Vec_DP(0.0,nStates);
  ii=Vec_INT(0.0,nStates);
    
  // Populate NLTE matrix
  populateBMatrix(iState,matrix,rhs,index,processes,path);
  
  // Return if singular NLTE matrix
  if (matrixSingular(matrix)) { if (log()) cout<<"NLTESolver::solveBoundState: Warning: Singular matrix\n"; return; }
    
  // Invert equation system using LU decomposition
  ludcmp(matrix,ii,dd);
  lubksb(matrix,ii,rhs);

  // For each bound state
  for (i=0;i<iState.bState.size();i++) {
    // If not excluded and no LTE reference state
    if ((not exclude(path<<i)) and iState.bState[i].LTERefState==NULL) 
      // Set number fraction
      X[i]=rhs[index[i]];
  }
  // Expand LTE groups
  expandLTEGroups(iState,X,path);

  return;
}

// Solve for ionization state of atom - New!
void NLTESolver::solveIonState(Atom& atom, vector<double>& dN, const Processes& processes, const Path& path) {

  vector<int> index;
  int i,nStates=0;
  Vec_DP rhs;
  Mat_DP matrix;
  Vec_INT ii;
  DP dd;

  // Make NLTE matrix index map
  index=iMatrixIndex(atom,nStates,path);
      
  // Initialize
  matrix=Mat_DP(0.0,nStates,nStates);
  rhs=Vec_DP(0.0,nStates);
  ii=Vec_INT(0.0,nStates);

  //if (atom.cell()->index()==36) cout<<"NLTESolver::solveIonState: cell: "<<atom.cell()->index()<<", "<<atom.name()<<", populate matrix\n";
    
  // Populate NLTE matrix
  populateIMatrixDN(atom,matrix,rhs,index,processes,path);

  // Return if singular NLTE matrix
  if (matrixSingular(matrix)) { if (log()) cout<<"NLTESolver::solveIonState: Warning: Singular matrix\n"; return; }
    
  //if (atom.cell()->index()==36) cout<<"NLTESolver::solveIonState: cell: "<<atom.cell()->index()<<", "<<atom.name()<<", invert matrix\n";

  // Invert equation system using LU decomposition
  ludcmp(matrix,ii,dd);
  lubksb(matrix,ii,rhs);

  //if (atom.cell()->index()==36) cout<<"NLTESolver::solveIonState: cell: "<<atom.cell()->index()<<", "<<atom.name()<<", matrix inverted\n";

  // For each ionization state
  for (i=0;i<atom.iState.size();i++ ) {
    // If not excluded - Set number density
    if (not exclude(path<<i)) dN[i]=rhs[index[i]];
  }
  
  return;
}

// Solve for ionization state of atom (in terms of number fractions) - Renamed (from solveIonState)!
void NLTESolver::solveIonStateX(Atom& atom, vector<double>& X, const Processes& processes, const Path& path) {

  vector<int> index;
  int i,nStates=0;
  Vec_DP rhs;
  Mat_DP matrix;
  Vec_INT ii;
  DP dd;

  // Make NLTE matrix index map
  index=iMatrixIndex(atom,nStates,path);
      
  // Initialize
  matrix=Mat_DP(0.0,nStates,nStates);
  rhs=Vec_DP(0.0,nStates);
  ii=Vec_INT(0.0,nStates);
    
  // Populate NLTE matrix
  populateIMatrix(atom,matrix,rhs,index,processes,path);
  
  // Return if singular NLTE matrix
  if (matrixSingular(matrix)) { if (log()) cout<<"NLTESolver::solveIonState: Warning: Singular matrix\n"; return; }
    
  // Invert equation system using LU decomposition
  ludcmp(matrix,ii,dd);
  lubksb(matrix,ii,rhs);

  // For each ionization state
  for (i=0;i<atom.iState.size();i++ ) {
    // If not excluded - Set number fraction
    if (not exclude(path<<i)) X[i]=rhs[index[i]];
  }
  
  return;
}

// Solve for bound state of atom
void NLTESolver::solveBoundState(Atom& atom, vector<vector<double> >& dN, const Processes& processes, const Path& path) {

  vector<vector<int> > index;
  int i,j,nStates=0;
  Vec_DP rhs;
  Mat_DP matrix;
  Vec_INT ii;
  DP dd;

  // Testing!!!
  //cout<<"NLTESolver::solveBoundState: cell: "<<atom.cell()->index()<<", atom: "<<atom.name()<<", Enter\n";

  // Make NLTE matrix index map
  index=bMatrixIndex(atom,nStates,path);
      
  // Initialize
  matrix=Mat_DP(0.0,nStates,nStates);
  rhs=Vec_DP(0.0,nStates);
  ii=Vec_INT(0.0,nStates);
    
  //if (log() and atom.name().compare("Fe")==0) cout<<"NLTESolver::solveBoundState: cell: "<<atom.cell()->index()<<", "<<atom.name()<<", Populate Matrix\n";

  // Populate NLTE matrix
  populateBMatrixDN(atom,matrix,rhs,index,processes,path);

  // Return if singular NLTE matrix
  if (matrixSingular(matrix)) { if (log()) cout<<"NLTESolver::solveBoundState: Warning: Singular matrix, cell: "<<atom.cell()->index()<<", atom: "<<atom.name()<<"\n"; return; }

  //if (log() and atom.name().compare("Fe")==0) cout<<"NLTESolver::solveBoundState: cell: "<<atom.cell()->index()<<", "<<atom.name()<<", Invert Matrix\n";
   
  // Invert equation system using LU decomposition
  ludcmp(matrix,ii,dd);
  lubksb(matrix,ii,rhs);

  //if (log() and atom.name().compare("Fe")==0) cout<<"NLTESolver::solveBoundState: cell: "<<atom.cell()->index()<<", "<<atom.name()<<", Matrix inverted\n";

  // For each ionization state
  for (i=0;i<atom.iState.size();i++) {
    // Resize number density vector
    // Note: This is not a proper reset!
    dN[i].resize(atom.iState[i].bState.size(),0);
    // For each bound state
    for (j=0;j<atom.iState[i].bState.size();j++) {
      // If not excluded and no LTE reference state
      if ((not exclude(path<<i<<j)) and atom.iState[i].bState[j].LTERefState==NULL) 
	// Set number density
	dN[i][j]=rhs[index[i][j]];
    }
    // Expand LTE groups
    //expandLTEGroupsDN(atom.iState[i],dN[i],path<<i);
  }

  // Expand LTE groups - Changed and moved!
  expandLTEGroupsDN(atom,dN,path);

  // Testing!
  /*for (i=0;i<atom.iState.size();i++) {
    for (j=0;j<atom.iState[i].bState.size();j++) {
      if (!isnumber(dN[i][j])) cout<<"NLTESolver::solveBoundState(3): : cell: "<<atom.cell()->index()<<", "<<atom.name()<<", dN["<<i<<"]["<<j<<"]: "<<dN[i][j]<<"\n";	 
    }
    }*/

  // Testing!!!
  //cout<<"NLTESolver::solveBoundState: cell: "<<atom.cell()->index()<<", atom: "<<atom.name()<<", Return\n";

  return;
}

// Solve for bound state of atom (in terms of number densities)
void NLTESolver::solveBoundStateN(Atom& atom, vector<vector<double> >& N, const Processes& processes, const Path& path) {

  vector<vector<int> > index;
  int i,j,nStates=0;
  Vec_DP rhs;
  Mat_DP matrix;
  Vec_INT ii;
  DP dd;

  // Make NLTE matrix index map
  index=bMatrixIndex(atom,nStates,path);
      
  // Initialize
  matrix=Mat_DP(0.0,nStates,nStates);
  rhs=Vec_DP(0.0,nStates);
  ii=Vec_INT(0.0,nStates);
    
  // Populate NLTE matrix
  populateBMatrix(atom,matrix,rhs,index,processes,path);
  
  // Return if singular NLTE matrix
  if (matrixSingular(matrix)) { if (log()) cout<<"NLTESolver::solveBoundState: Warning: Singular matrix, atom: "<<atom.name()<<"\n"; return; }
    
  // Invert equation system using LU decomposition
  ludcmp(matrix,ii,dd);
  lubksb(matrix,ii,rhs);

  // For each ionization state
  for (i=0;i<atom.iState.size();i++) {
    // Resize number density vector
    // Note: This is not a proper reset!
    N[i].resize(atom.iState[i].bState.size(),0);
    // For each bound state
    for (j=0;j<atom.iState[i].bState.size();j++) {
      // If not excluded and no LTE reference state
      if ((not exclude(path<<i<<j)) and atom.iState[i].bState[j].LTERefState==NULL) 
	// Set number density
	N[i][j]=rhs[index[i][j]];
    }
    // Expand LTE groups
    //expandLTEGroups(atom.iState[i],N[i],path<<i);
  }
  
  // Expand LTE groups - Changed and moved!
  expandLTEGroupsN(atom,N,path);

  return;
}

// Solve for bound state of atom (in terms of number fractions)
void NLTESolver::solveBoundStateX(Atom& atom, vector<double>& XIState, vector<vector<double> >& XBState, const Processes& processes, const Path& path) {
  vector<vector<double> > N;
  double nMin=config.convergence.nMin*numeric_limits<double>::epsilon()*atom.n();
  double nAtom,nIState;
  int i,j;  

  // Initialize N
  initN(atom,N);

  // Solve for bound state in terms of number densities
  solveBoundStateN(atom,N,processes,path);

  // Expand LTE groups missing!

  // Limit number density
  for (i=0;i<atom.iState.size();i++)     
    for (j=0;j<atom.iState[i].bState.size();j++)
      if (not exclude(path<<i<<j)) N[i][j]=MAX(N[i][j],nMin);

  // Sum number density
  // Unclear if muliple initializers work - Fix!
  nAtom=0; // New!
  for (i=0;i<atom.iState.size();i++) // Changed!
    for (j=0;j<atom.iState[i].bState.size();j++) nAtom+=N[i][j];

  // For each ionization state
  for (i=0;i<atom.iState.size();i++) {
     
    // Sum number density
    // Unclear if muliple initializers work - Fix!
    nIState=0; // New!
    for (j=0;j<atom.iState[i].bState.size();j++) nIState+=N[i][j]; // Changed!
      
    // If not excluded - Update number fraction
    if (not exclude(path<<i)) {
      XIState[i]=nIState/nAtom;
    }
    
    // For each bound state
    for (j=0;j<atom.iState[i].bState.size();j++) {
      // If not excluded - Update number fraction
      if (not exclude(path<<i<<j)) {
	// If non-zero ionization state number density - Update
	if (nIState!=0) XBState[i][j]=N[i][j]/nIState;
	// Else - Set dummy (but consistent) value
	else XBState[i][j]=(j==0?1:0);
      }
    }
  }

  return;
}

// Solve for bound state of atom (Decoupled version)
void NLTESolver::solveBoundStateDecoupled(Atom& atom, vector<vector<double> >& dN, const Processes& processes, const Path& path) {
  vector<vector<double> > XBState(atom.iState.size());
  //vector<double> XIState(atom.iState.size(),0);
  vector<double> dNI(atom.iState.size(),0); // Changed!
  int i,j;

  // For each ion
  for (i=0;i<atom.iState.size();i++) {
    // Resize solution vector
    XBState[i].resize(atom.iState[i].bState.size(),0);
    // If not excluded - Solve for the bound states
    if (not exclude(path<<i)) solveBoundState(atom.iState[i],XBState[i],processes,path<<i);
  }
  
  // Solve for the ionization states
  //solveIonState(atom,XIState,processes,path);
  solveIonState(atom,dNI,processes,path); // Changed!

  // Testing!
  //if (log() and atom.cell()->index()==24 and atom.name().compare("Ni")==0) for (i=0;i<atom.iState.size();i++) cout<<"NLTESolver::solveBoundStateDecoupled: cell: "<<atom.cell()->index()<<", "<<atom.iState[i].name()<<", dNI: "<<dNI[i]<<"\n";

  // Testing!
  //vector<double> XBStateTot(atom.iState.size(),0);
  //vector<double> nIState(atom.iState.size(),0);  

  // For each ionization state
  for (i=0;i<atom.iState.size();i++) {
    // If not excluded
    if (not exclude(path<<i)) {
      // For each bound state
      for (j=0;j<atom.iState[i].bState.size();j++) {
	// If not excluded - Calculate change in number density
	if (not exclude(path<<i<<j)) {
	  //dN[i][j]=XIState[i]*XBState[i][j]*atom.n()-atom.iState[i].bState[j].n();
	  dN[i][j]=(atom.iState[i].n()+dNI[i])*XBState[i][j]-atom.iState[i].bState[j].n(); // Changed!
	  // Testing!
	  //XBStateTot[i]+=XBState[i][j];
	  //nIState[i]+=atom.iState[i].bState[j].n();
	}
      }
    }
  }

  // Testing!
  //if (log() and atom.cell()->index()==24 and atom.name().compare("Ni")==0) for (i=0;i<atom.iState.size();i++) { dNI[i]=0; for (j=0;j<atom.iState[i].bState.size();j++) dNI[i]+=dN[i][j]; cout<<"NLTESolver::solveBoundStateDecoupled: cell: "<<atom.cell()->index()<<", "<<atom.iState[i].name()<<", dNI: "<<dNI[i]<<", n: "<<atom.iState[i].n()<<", nIState: "<<nIState[i]<<"<<, XBStateTot: "<<XBStateTot[i]<<"\n"; }

  return;
}

// Solve for bound state of atom (in terms of number fractions, Decoupled version)
void NLTESolver::solveBoundStateDecoupledX(Atom& atom, vector<double>& XIState, vector<vector<double> >& XBState, const Processes& processes, const Path& path) {
  int i,j;

  // For each ion
  for (i=0;i<atom.iState.size();i++) {
    // Resize solution vector
    XBState[i].resize(atom.iState[i].bState.size(),0);
    // If not excluded - Solve for the bound states
    if (not exclude(path<<i)) solveBoundState(atom.iState[i],XBState[i],processes,path<<i);
  }
  
  // Solve for the ionization states
  //solveIonState(atom,XIState,processes,path);
  solveIonStateX(atom,XIState,processes,path); //Changed!

  return;
}

// Solve for bound state of cell
void NLTESolver::solveBoundState(Cell* cell, vector<vector<vector<double> > >& dN) {
  bool coupled;
  Processes processes;
  int i;

  //if (log()) cout<<"NLTESolver::solveBoundState: cell: "<<cell->index()<<", thread_num: "<<omp_get_thread_num()<<", num_threads: "<<omp_get_num_threads()<<", Enter\n";
  if (cell->index()==36) cout<<"NLTESolver::solveBoundState: cell: "<<cell->index()<<", Enter\n";

  // If coupled cell flag set - Solve for bound state of cell
  if (config.coupled) solveBoundStateCoupled(cell,dN);
  // Else
  else {

    //if (log()) cout<<"NLTESolver::solveBoundState: cell: "<<cell->index()<<", Update tables\n";
    //if (cell->index()==36) cout<<"NLTESolver::solveBoundState: cell: "<<cell->index()<<", Update tables\n";

    // Testing - Update all required tables in advance
    if (cell->bBFactorUpdate) { cell->bBFactorTable(); cell->bBFactorUpdate=false; }
    if (cell->bMFactorUpdate) { cell->bMFactorTable(); cell->bMFactorUpdate=false; }
    // Testing!!!
    //if (cell->boundFreeCutOffUpdate) { cell->updateBoundFreeCutOffOMP(); cell->boundFreeCutOffUpdate=false; }
    // Testing!!!
    //if (cell->boundFreeCutOffUpdate) { cell->updateBoundFreeCutOff(); cell->boundFreeCutOffUpdate=false; }
    if (cell->index()==36) cout<<"Cell::solveBoundState: cell: "<<cell->index()<<", Update bound-free opacity table\n";
    if (cell->boundFreeTableUpdate) { cell->boundFreeAbsKappaTable(true); cell->boundFreeTableUpdate=false; }
    if (cell->index()==36) cout<<"Cell::solveBoundState: cell: "<<cell->index()<<", Update bound-free source-function table\n";
    if (cell->boundFreeSTableUpdate) { cell->boundFreeSTable(true); cell->boundFreeSTableUpdate=false; }
    if (cell->index()==36) cout<<"Cell::solveBoundState: cell: "<<cell->index()<<", Tables updated\n";
    //if (cell->bBFactorUpdate) cell->bBFactorTable();
    //if (cell->bMFactorUpdate) cell->bMFactorTable();
    //if (cell->boundFreeCutOffUpdate) cell->updateBoundFreeCutOffOMP();
    //if (cell->boundFreeTableUpdate) cell->boundFreeAbsKappaTable(true);
    //if (cell->boundFreeSTableUpdate) cell->boundFreeSTable(true);

    //if (log()) cout<<"NLTESolver::solveBoundState: cell: "<<cell->index()<<", Tables updated\n";
    //if (cell->index()==36) cout<<"NLTESolver::solveBoundState: cell: "<<cell->index()<<", Tables updated\n";

    // For each atom
    //#pragma omp parallel for private(coupled,processes) schedule(static,1) num_threads(2)
    //#pragma omp parallel for private(coupled,processes) schedule(static,1) // Changed!
    #pragma omp parallel for private(coupled,processes) schedule(dynamic,1) // Changed!
    //schedule(static,1) num_threads(2)
    for (i=0;i<cell->atom.size();i++) {
      // If not exclude
      if (not exclude(Path(i))) {

	//if (log()) cout<<"NLTESolver::solveBoundState: cell: "<<cell->index()<<", "<<cell->atom[i].name()<<", thread_num: "<<omp_get_thread_num()<<", num_threads: "<<omp_get_num_threads()<<"\n";
	
	// Get coupled solution flag and processes from configuration
	coupled=config.population.coupled(cell->atom[i]);
	processes=config.population.processes(cell->atom[i]);
	
	// Solve for bound state according to configuration
	if (not coupled) solveBoundStateDecoupled(cell->atom[i],dN[i],processes,Path(i));
	else solveBoundState(cell->atom[i],dN[i],processes,Path(i));
      }
    }
 }

  //if (log()) cout<<"NLTESolver::solveBoundState: cell: "<<cell->index()<<", Return\n";
  if (cell->index()==36) cout<<"NLTESolver::solveBoundState: cell: "<<cell->index()<<", Return\n";
 
  return;
}

// Solve for bound state of cell (in terms of number fractions)
void NLTESolver::solveBoundStateX(Cell* cell, vector<vector<double> >& XIState, vector<vector<vector<double> > >& XBState) {
  bool coupled;
  Processes processes;
  int i;

  // For each atom
  for (i=0;i<cell->atom.size();i++) {
    // If not exclude
    if (not exclude(Path(i))) {

      // Get coupled solution flag and processes from configuration
      coupled=config.population.coupled(cell->atom[i]);
      processes=config.population.processes(cell->atom[i]);

      // Solve for bound state according to configuration
      if (not coupled) solveBoundStateDecoupledX(cell->atom[i],XIState[i],XBState[i],processes,Path(i));
      else solveBoundStateX(cell->atom[i],XIState[i],XBState[i],processes,Path(i));
    }
  }
  
  return;
}

// Solve for bound state of cell (Coupled version)
void NLTESolver::solveBoundStateCoupled(Cell* cell, vector<vector<vector<double> > >& dN) {

  vector<vector<vector<int> > > index;
  int i,j,k,nStates=0;
  Vec_DP rhs;
  Mat_DP matrix;
  Vec_INT ii;
  DP dd;

  // Make NLTE matrix index map
  index=bMatrixIndex(cell,nStates);
      
  // Initialize
  matrix=Mat_DP(0.0,nStates,nStates);
  rhs=Vec_DP(0.0,nStates);
  ii=Vec_INT(0.0,nStates);

  //cout<<"NLTESolver::solveBoundStateCoupled: cell: "<<cell->index()<<", nStates: "<<nStates<<", Populate Matrix\n";
    
  // Populate NLTE matrix
  populateBMatrixDN(cell,matrix,rhs,index);

  //cout<<"NLTESolver::solveBoundStateCoupled: cell: "<<cell->index()<<", nStates: "<<nStates<<", Invert Matrix\n";

  // Testing
  BState* bState;
  IState* iState;
  Mat_DP matrixOrig=matrix;
  Vec_DP rhsOrig=rhs;
  
  // Return if singular NLTE matrix
  if (matrixSingular(matrix)) { if (log()) cout<<"NLTESolver::solveBoundState: Warning: Singular matrix, cell: "<<cell->index()<<"\n"; return; }
   
  // Invert equation system using LU decomposition
  ludcmp(matrix,ii,dd);
  lubksb(matrix,ii,rhs);

  //cout<<"NLTESolver::solveBoundStateCoupled: cell: "<<cell->index()<<", Matrix inverted\n";

  // For each atom
  for (i=0;i<cell->atom.size();i++) {
    // If not exclude
    if (not exclude(Path(i))) {
      // Resize number density vector
      dN[i].resize(cell->atom[i].iState.size());
      // For each ionization state
      for (j=0;j<cell->atom[i].iState.size();j++) {
	// Resize number density vector
	// Note: This is not a proper reset!
	dN[i][j].resize(cell->atom[i].iState[j].bState.size(),0);
	// For each bound state
	for (k=0;k<cell->atom[i].iState[j].bState.size();k++) {
	  // If not excluded and no LTE reference state
	  if ((not exclude(Path(i)<<j<<k)) and cell->atom[i].iState[j].bState[k].LTERefState==NULL) 
	    // Set number density
	    dN[i][j][k]=rhs[index[i][j][k]];
	}
	// Expand LTE groups
	//expandLTEGroupsDN(cell->atom[i].iState[j],dN[i][j],Path(i)<<j);
      }
    }
    // Expand LTE groups - Changed and moved!
    expandLTEGroupsDN(cell->atom[i],dN[i],Path(i));
  }
  
  //cout<<"NLTESolver::solveBoundStateCoupled: cell: "<<cell->index()<<", 4\n";

  return;
}

// Solve for ionization state of cell (coupled version) - New!
// Note: dN returned in terms of bound state populations
void NLTESolver::solveIonStateCoupled(Cell* cell, vector<vector<vector<double> > >& dN) {

  vector<vector<int> > index;
  int i,j,k,nStates=0;
  Vec_DP rhs;
  Mat_DP matrix;
  Vec_INT ii;
  DP dd;

  // Make NLTE matrix index map
  index=iMatrixIndex(cell,nStates);
      
  // Initialize
  matrix=Mat_DP(0.0,nStates,nStates);
  rhs=Vec_DP(0.0,nStates);
  ii=Vec_INT(0.0,nStates);

  // Populate NLTE matrix
  populateIMatrixDN(cell,matrix,rhs,index);
  
  // Return if singular NLTE matrix
  if (matrixSingular(matrix)) { if (log()) cout<<"NLTESolver::solveIonStateCoupled: Warning: Singular matrix\n"; return; }
    
  // Invert equation system using LU decomposition
  ludcmp(matrix,ii,dd);
  lubksb(matrix,ii,rhs);

  //cout<<"NLTESolver::solveIonStateCoupled: cell: "<<cell->index()<<", Matrix inverted\n";

  // For each atom
  for (i=0;i<cell->atom.size();i++) {
    // If not exclude
    if (not exclude(Path(i))) {
      // Resize number density vector
      dN[i].resize(cell->atom[i].iState.size());
      // For each ionization state
      for (j=0;j<cell->atom[i].iState.size();j++ ) {
	// Resize number density vector
	// Note: This is not a proper reset!
	dN[i][j].resize(cell->atom[i].iState[j].bState.size(),0);
	// For each bound state
	for (k=0;k<cell->atom[i].iState[j].bState.size();k++)
	  // If not excluded - Set number density
	  if (not exclude(Path(i)<<j<<k)) dN[i][j][k]=rhs[index[i][j]]*cell->atom[i].iState[j].bState[k].X();
      }
    }
  }
  
  return;
}

// Calculate heating-cooling balance
//double NLTESolver::G(const IState& iState, const vector<double>& dN, const Processes& processes, const Path& path, const double& T) {
double NLTESolver::G(IState& iState, const vector<double>& dN, const Processes& processes, const Path& path, const double& T) {
  Cell* cell=iState.atom()->cell();
  double h,c,g,_G=0;
  int i;

  // For each bound state
  for (i=0;i<iState.bState.size();i++) {
    // if not excluded
    if (not exclude(path<<i)) {
      
      // Calculate heating coefficient
      h=0;
      if (processes(pBoundFree,pRadiative)) h+=iState.bState[i].boundFreeHeating(T)/iState.bState[i].n();
      if (processes(pBoundBound|pBoundFree,pCollisional)) h+=iState.bState[i].collHeating(T)/iState.bState[i].n();
      if (processes(pFreeFree)) h+=iState.freeFreeHeating(T)/iState.n();
      if (processes(pNonThermal)) h+=cell->nonThermalHeating()/cell->n();
      // Note: This does not take into account the dependence on the ion number densities! - Fix!
      if (processes(pBoundFree,pChargeTransfer)) h+=cell->CTHeating(T)/cell->n();
      
      // Calculate cooling coefficient
      c=0;
      if (processes(pBoundFree,pRadiative)) c+=iState.bState[i].boundFreeCooling(T)/iState.bState[i].n();
      if (processes(pBoundBound|pBoundFree,pCollisional)) c+=iState.bState[i].collCooling(T)/iState.bState[i].n();
      if (processes(pFreeFree)) c+=iState.freeFreeCooling(T)/iState.n();
      // Note: This does not take into account the dependence on the ion number densities! - Fix!
      if (processes(pBoundFree,pChargeTransfer)) c+=cell->CTCooling(T)/cell->n();
      
      // Calculate heating-cooling balance (coefficient)
      g=h-c;
      
      // Add upp total heating-cooling balance
      _G+=g*(iState.bState[i].n()+dN[i]);
    }
  }

  return _G;
}

// Calculate heating-cooling balance
double NLTESolver::G(Cell* cell, const vector<vector<vector<double> > >& dN, Processes processes, const double& T) {
  double h,c,g,_G=0;
  Processes processesAtom;
  int i,j,k,l;

  //if (log()) cout<<"NLTESolver::G: cell: "<<cell->index()<<", thread_num: "<<omp_get_thread_num()<<", num_threads: "<<omp_get_num_threads()<<", Enter\n";
  if (cell->index()==36) cout<<"NLTESolver::G: cell: "<<cell->index()<<", Enter\n";

  // Testing - Update all required tables in advance
  if (cell->bBFactorUpdate) { cell->bBFactorTable(); cell->bBFactorUpdate=false; }
  if (cell->bMFactorUpdate) { cell->bMFactorTable(); cell->bMFactorUpdate=false; }
  //if (cell->boundFreeCutOffUpdate) { cell->updateBoundFreeCutOffOMP(); cell->boundFreeCutOffUpdate=false; }
  //if (cell->boundFreeSTableUpdate) { cell->boundFreeSTable(true); cell->boundFreeSTableUpdate=false; }
  //if (cell->boundFreeCutOffUpdate) { cell->updateBoundFreeCutOff(); cell->boundFreeCutOffUpdate=false; }
  //if (cell->boundFreeSTableUpdate) { cell->boundFreeSTable(); cell->boundFreeSTableUpdate=false; }
  //if (cell->bBFactorUpdate) cell->bBFactorTable();
  //if (cell->bMFactorUpdate) cell->bMFactorTable();
  //if (cell->boundFreeCutOffUpdate) cell->updateBoundFreeCutOffOMP();
  //if (cell->boundFreeSTableUpdate) cell->boundFreeSTable(true);
    
  // For each atom
  //#pragma omp parallel for private(j,k,g,h,c,processesAtom) schedule(static,1) num_threads(2)
  //#pragma omp parallel for private(j,k,g,h,c,processesAtom) schedule(static,1) // Changed!
  #pragma omp parallel for private(j,k,g,h,c,processesAtom) schedule(dynamic,1) // Changed!
  for (i=0;i<cell->atom.size();i++) {
    //if (log()) cout<<"NLTESolver::G: cell: "<<cell->index()<<", "<<cell->atom[i].name()<<", thread_num: "<<omp_get_thread_num()<<", num_threads: "<<omp_get_num_threads()<<"\n";
    // Get processes from configuration
    processesAtom=config.temperature.processes(cell->atom[i]);
    // If not excluded
    if (not exclude(Path(i))) {
      // For each ionization state
      for (j=0;j<cell->atom[i].iState.size();j++) {
	// If not excluded
	if (not exclude(Path(i)<<j)) {
	  // For each bound state
	  for (k=0;k<cell->atom[i].iState[j].bState.size();k++) {
	    // if not excluded and non-zero number density
	    // The latter condition prevents division by zero - Improve?
	    if (not exclude(Path(i)<<j<<k) and cell->atom[i].iState[j].bState[k].n()!=0) {

	      // Calculate heating coefficient
	      h=0;
	      if (processes(pBoundFree,pRadiative) and processesAtom(pBoundFree,pRadiative))
		h+=cell->atom[i].iState[j].bState[k].boundFreeHeating(T)/cell->atom[i].iState[j].bState[k].n();
	      if (processes(pBoundBound|pBoundFree,pCollisional) and processesAtom(pBoundBound|pBoundFree,pCollisional))
		h+=cell->atom[i].iState[j].bState[k].collHeating(T)/cell->atom[i].iState[j].bState[k].n();
	      if (processes(pFreeFree) and processesAtom(pFreeFree))
		h+=cell->atom[i].iState[j].freeFreeHeating(T)/cell->atom[i].iState[j].n();
	      if (processes(pNonThermal) and processesAtom(pNonThermal)) 
		h+=cell->nonThermalHeating()/cell->n();
	      // Note: This does not take into account the dependence on the ion number densities! - Fix!
	      if (processes(pBoundFree,pChargeTransfer) and processesAtom(pBoundFree,pChargeTransfer)) 
		h+=cell->CTHeating(T)/cell->n();
	      
	      // Calculate cooling coefficient
	      c=0;
	      if (processes(pBoundFree,pRadiative) and processesAtom(pBoundFree,pRadiative))
		c+=cell->atom[i].iState[j].bState[k].boundFreeCooling(T)/cell->atom[i].iState[j].bState[k].n();
	      if (processes(pBoundBound|pBoundFree,pCollisional) and processesAtom(pBoundBound|pBoundFree,pCollisional))
		c+=cell->atom[i].iState[j].bState[k].collCooling(T)/cell->atom[i].iState[j].bState[k].n();
	      if (processes(pFreeFree) and processesAtom(pFreeFree))
		c+=cell->atom[i].iState[j].freeFreeCooling(T)/cell->atom[i].iState[j].n();
	      // Note: This does not take into account the dependence on the ion number densities! - Fix!
	      if (processes(pBoundFree,pChargeTransfer) and processesAtom(pBoundFree,pChargeTransfer)) 
		c+=cell->CTCooling(T)/cell->n();
	      
	      // Calculate heating-cooling balance (coefficient)
	      g=h-c;
	      
	      // Add upp total heating-cooling balance
              #pragma omp atomic
	      _G+=g*(cell->atom[i].iState[j].bState[k].n()+dN[i][j][k]);
	    }
	  }
	}
      }
    }
  }

  // If not steady-state - Add expansion cooling - Remove - Testing!
  //if (not config.steadyState) _G-=((EOSBasic*) cell->eos)->PGas(cell)*3/cell->t();

  //if (log()) cout<<"NLTESolver::G: cell: "<<cell->index()<<", Return\n";
  if (cell->index()==36) cout<<"NLTESolver::G: cell: "<<cell->index()<<", Return\n";
  
  return _G;
}

// Log heating-cooling
// Not up-to-date!
void NLTESolver::logHC(Cell* cell, const vector<vector<vector<double> > >& dN, Processes processes, const double& T) {
  double h_bf,h_c,h_ff,c_bf,c_c,c_ff,H_BF,H_C,H_FF,C_BF,C_C,C_FF,H=0,C=0,C_Exp=0;
  Processes processesAtom;
  int i,j,k,l;

  if (log()) cout<<"NLTESolver::logHC: cell: "<<cell->index()<<", T: "<<cell->T()<<", ne: "<<cell->ne()<<"\n";

  if (log()) {
    for (j=0;j<cell->atom[0].iState.size();j++) {
      cout<<"NLTESolver::logHC: cell: "<<cell->index()<<", "<<cell->atom[0].iState[j].name()<<", X: "<<cell->atom[0].iState[j].X()<<", n: "<<cell->atom[0].iState[j].n()<<"\n";
      for (k=0;k<cell->atom[0].iState[j].bState.size();k++) {
	cout<<"NLTESolver::logHC: cell: "<<cell->index()<<", "<<cell->atom[0].iState[j].name()<<":"<<k<<", X: "<<cell->atom[0].iState[j].bState[k].X()<<", n: "<<cell->atom[0].iState[j].bState[k].n()<<"\n";
      }
    }
  }

  for (i=0;i<cell->atom.size();i++) {
    processesAtom=config.temperature.processes(cell->atom[i]);
    if (not exclude(Path(i))) {
      for (j=0;j<cell->atom[i].iState.size();j++) {
	if (not exclude(Path(i)<<j)) {

	  H_BF=0; C_BF=0;
	  H_C=0; C_C=0;
	  H_FF=0; C_FF=0;
	  for (k=0;k<cell->atom[i].iState[j].bState.size();k++) {
	    if (not exclude(Path(i)<<j<<k) and cell->atom[i].iState[j].bState[k].n()!=0) {

	      if (processes(pBoundFree,pRadiative) and processesAtom(pBoundFree,pRadiative)) {
		h_bf=cell->atom[i].iState[j].bState[k].boundFreeHeating(T)/cell->atom[i].iState[j].bState[k].n();
		H_BF+=h_bf*(cell->atom[i].iState[j].bState[k].n()+dN[i][j][k]);
	      } 
	      if (processes(pBoundBound|pBoundFree,pCollisional) and processesAtom(pBoundBound|pBoundFree,pCollisional)) {
		h_c=cell->atom[i].iState[j].bState[k].collHeating(T)/cell->atom[i].iState[j].bState[k].n();
		H_C+=h_c*(cell->atom[i].iState[j].bState[k].n()+dN[i][j][k]);
	      }
	      if (processes(pFreeFree) and processesAtom(pFreeFree)) {
		h_ff=cell->atom[i].iState[j].freeFreeHeating(T)/cell->atom[i].iState[j].n();
		H_FF+=h_ff*(cell->atom[i].iState[j].bState[k].n()+dN[i][j][k]);
	      }
	      if (processes(pBoundFree,pRadiative) and processesAtom(pBoundFree,pRadiative)) {
		c_bf=cell->atom[i].iState[j].bState[k].boundFreeCooling(T)/cell->atom[i].iState[j].bState[k].n();
		C_BF+=c_bf*(cell->atom[i].iState[j].bState[k].n()+dN[i][j][k]);
	      }
	      if (processes(pBoundBound|pBoundFree,pCollisional) and processesAtom(pBoundBound|pBoundFree,pCollisional)) {
		c_c=cell->atom[i].iState[j].bState[k].collCooling(T)/cell->atom[i].iState[j].bState[k].n();
		C_C+=c_c*(cell->atom[i].iState[j].bState[k].n()+dN[i][j][k]);
	      }
	      if (processes(pFreeFree) and processesAtom(pFreeFree)) {
		c_ff=cell->atom[i].iState[j].freeFreeCooling(T)/cell->atom[i].iState[j].n();		
		C_FF+=c_ff*(cell->atom[i].iState[j].bState[k].n()+dN[i][j][k]);
	      }
	    }
	  }
	  if (processes(pBoundFree,pRadiative) and processesAtom(pBoundFree,pRadiative)) {
	    if (log()) cout<<"NLTESolver::logHC: cell: "<<cell->index()<<", "<<cell->atom[i].iState[j].name()<<", bound-free: H: "<<H_BF<<", C: "<<C_BF<<"\n";
	    H+=H_BF; C+=C_BF;
	  }
	  if (processes(pBoundBound|pBoundFree,pCollisional) and processesAtom(pBoundBound|pBoundFree,pCollisional)) {
	    if (log()) cout<<"NLTESolver::logHC: cell: "<<cell->index()<<", "<<cell->atom[i].iState[j].name()<<", collisional: H: "<<H_C<<", C: "<<C_C<<"\n";
	    H+=H_C; C+=C_C;
	  }
	  if (processes(pFreeFree) and processesAtom(pFreeFree)) {
	    if (log()) cout<<"NLTESolver::logHC: cell: "<<cell->index()<<", "<<cell->atom[i].iState[j].name()<<", free-free: H: "<<H_FF<<", C: "<<C_FF<<"\n";
	    H+=H_FF; C+=C_FF;
	  }
	}
      }
    }
  }

  // Remove - Testing!
  /*if (not config.steadyState) {
    C_Exp=((EOSBasic*) cell->eos)->PGas(cell)*3/cell->t();
    C+=C_Exp;
    if (log()) cout<<"NLTESolver::logG: cell: "<<cell->index()<<", C_Exp: "<<C_Exp<<"\n";
    }*/

  if (log()) cout<<"NLTESolver::logG: cell: "<<cell->index()<<", H: "<<H<<", C: "<<C<<"\n";
}

// Calculate derivative of heating-cooling balance
//double NLTESolver::dGdT(const IState& iState, const Processes& processes, const Path& path) {
double NLTESolver::dGdT(IState& iState, const Processes& processes, const Path& path) {
  Cell* cell=iState.atom()->cell();
  double _dGdT,dTDerivative=config.convergence.dTDerivative;
  vector<double> dN(iState.bState.size(),0);
  int i;

  // Calculate derivative of the heating-cooling difference
  _dGdT=(G(iState,dN,processes,path,cell->T()*(1+0.5*dTDerivative))-G(iState,dN,processes,path,cell->T()*(1-0.5*dTDerivative)))/(cell->T()*dTDerivative);

  return _dGdT;
}

// Calculate derivative of heating-cooling balance
double NLTESolver::dGdT(Cell* cell, Processes processes) {
  double _dGdT,dTDerivative=config.convergence.dTDerivative;
  double T,ne,dNe,G1,G2,xLTE;
  vector<vector<vector<double> > > dN(cell->atom.size());
  int i,j,k,l;
  TransBB* transBB;
  TransBF* transBF;
  
  // Initiatlize dN to zero
  for (i=0;i<cell->atom.size();i++) {
    dN[i].resize(cell->atom[i].iState.size());
    for (j=0;j<cell->atom[i].iState.size();j++) dN[i][j].resize(cell->atom[i].iState[j].bState.size(),0);
  }

  // If partial derivative
  if (config.convergence.derivative==dPartial) {

    // Calculate derivative of the heating-cooling difference
    _dGdT=(G(cell,dN,processes,cell->T()*(1+0.5*dTDerivative))-G(cell,dN,processes,cell->T()*(1-0.5*dTDerivative)))/(cell->T()*dTDerivative);

  }
  // Else (i.e. full derivative)
  else {

    vector<vector<vector<double> > > dNZero=dN;

    // Get current temperature
    T=cell->T();
    ne=cell->ne();
    
    // Evaluate G att higher temperature
    cell->T(T*(1+0.5*dTDerivative)); cell->reset(dpT);
    solveBoundState(cell,dN);
    updateBoundState(cell,dN,false); cell->reset(dpN);
    if (config.convergence.derivative==dFullNe) { cell->ne(cell->ne()+getdNe(cell,dN)); cell->reset(dpNe); }
    G1=G(cell,dNZero,processes,T*(1+0.5*dTDerivative));

    // Reverse dN
    for (i=0;i<cell->atom.size();i++)
      for (j=0;j<cell->atom[i].iState.size();j++)
	for (k=0;k<cell->atom[i].iState[j].bState.size();k++) dN[i][j][k]=-dN[i][j][k];

    // Rewind bound state
    updateBoundState(cell,dN,false); cell->reset(dpN);
    if (config.convergence.derivative==dFullNe) { cell->ne(cell->ne()+getdNe(cell,dN)); cell->reset(dpNe); }

    // Evaluate G at lower temperature
    cell->T(T*(1-0.5*dTDerivative)); cell->reset(dpT);
    solveBoundState(cell,dN);
    updateBoundState(cell,dN,false); cell->reset(dpN);
    if (config.convergence.derivative==dFullNe) { cell->ne(cell->ne()+getdNe(cell,dN)); cell->reset(dpNe); }
    G2=G(cell,dNZero,processes,T*(1-0.5*dTDerivative));

    // Reverse dN
    for (i=0;i<cell->atom.size();i++)
      for (j=0;j<cell->atom[i].iState.size();j++)
	for (k=0;k<cell->atom[i].iState[j].bState.size();k++) dN[i][j][k]=-dN[i][j][k];

    // Rewind bound state
    updateBoundState(cell,dN,false); cell->reset(dpN);
    if (config.convergence.derivative==dFullNe) { cell->ne(cell->ne()+getdNe(cell,dN)); cell->reset(dpNe); }

    // Reset temperature and electron density
    cell->T(T);
    cell->ne(ne);
    cell->reset(dpT|dpNe);

    // Calculate derivative of G
    _dGdT=(G1-G2)/(T*dTDerivative);
  }

  return _dGdT;
}

// Solve for temperature
void NLTESolver::solveTemperature(Cell* cell, const vector<vector<vector<double> > >& dN, double& dT) {
  int i,j;
  Processes processes;
  double H=0,C=0;
  
  // Calculate dT
  dT=-G(cell,dN)/dGdT(cell);

  //if (log()) cout<<"NLTESolver::solveTemperature: cell: "<<cell->index()<<", G: "<<G(cell,dN)<<", G(bound-free): "<<G(cell,dN,Processes(pBoundFree,pRadiative))<<", G(collisional): "<<G(cell,dN,Processes(pBoundFree|pBoundBound,pCollisional))<<", G(free-free): "<<G(cell,dN,Processes(pFreeFree))<<"\n";
  //if (log()) cout<<"NLTESolver::solveTemperature: cell: "<<cell->index()<<", dGdT: "<<dGdT(cell)<<", dGdT(bound-free): "<<dGdT(cell,Processes(pBoundFree,pRadiative))<<", dGdT(collisional): "<<dGdT(cell,Processes(pBoundFree|pBoundBound,pCollisional))<<", dGdT(free-free): "<<dGdT(cell,Processes(pFreeFree))<<"\n";
  //if (log()) logHC(cell,dN);
}

// Scan for temperature
/*void NLTESolver::scanTemperature(Cell* cell, double& dT, bool update) {
  vector<vector<vector<double> > > dNZero;
  MStateMPI mState;
  double TCur,TMin,TMax,TNew,GCur,GMin,GMax,GNew;
  int n;
  int scanTDiv;

  // Initialize
  initN(cell,dNZero);

  // Get temperature
  TCur=cell->T();

  // Calculate scan temperatures (small)
  TMax=TCur*config.convergence.scanTFacSmall;
  TMin=TCur/config.convergence.scanTFacSmall;
  scanTDiv=config.convergence.scanTDivSmall;

  // Store matter state
  mState.pack(cell);  
  // Reset (with respect to all quantities)
  cell->reset(dpAll);

  // Update matter state at current temperature
  updateMatterStateT(cell,TCur,config.convergence.scanTNIter,config.convergence.scanTDampN,config.convergence.scanTNeTol,config.convergence.scanTNTol);
  // Evaluate G at the current temperature
  GCur=G(cell,dNZero);

  // Restore matter state
  mState.unpack(cell);
  // Reset (with respect to temperature, level populations and electron density)
  cell->reset(dpT|dpN|dpNe);

  // Update matter state at minimum temperature
  updateMatterStateT(cell,TMin,config.convergence.scanTNIter,config.convergence.scanTDampN,config.convergence.scanTNeTol,config.convergence.scanTNTol);
  // Evaluate G at minimum temperature
  GMin=G(cell,dNZero);

  // Restore matter state
  mState.unpack(cell);
  // Reset (with respect to temperature, level populations and electron density)
  cell->reset(dpT|dpN|dpNe);

  // Update matter state at maximum temperature
  updateMatterStateT(cell,TMax,config.convergence.scanTNIter,config.convergence.scanTDampN,config.convergence.scanTNeTol,config.convergence.scanTNTol);
  // Evaluate G at maximum temperature
  GMax=G(cell,dNZero);

  // Restore matter state
  mState.unpack(cell);
  // Reset (with respect to temperature, level populations and electron density)
  cell->reset(dpT|dpN|dpNe);

  // If no zero found
  if ((GMin>0 and GCur>0 and GMax>0) or (GMin<0 and GCur<0 and GMax<0))  {

    // Calculate scan temperatures
    TMax=TCur*config.convergence.scanTFac;
    TMin=TCur/config.convergence.scanTFac;
    scanTDiv=config.convergence.scanTDiv;
    
    // Store matter state
    mState.pack(cell);  
    // Reset (with respect to all quantities)
    cell->reset(dpAll);
    
    // Update matter state at current temperature
    updateMatterStateT(cell,TCur,config.convergence.scanTNIter,config.convergence.scanTDampN,config.convergence.scanTNeTol,config.convergence.scanTNTol);
    // Evaluate G at the current temperature
    GCur=G(cell,dNZero);

    // Restore matter state
    mState.unpack(cell);
    // Reset (with respect to temperature, level populations and electron density)
    cell->reset(dpT|dpN|dpNe);
    
    // Update matter state at minimum temperature
    updateMatterStateT(cell,TMin,config.convergence.scanTNIter,config.convergence.scanTDampN,config.convergence.scanTNeTol,config.convergence.scanTNTol);
    // Evaluate G at minimum temperature
    GMin=G(cell,dNZero);
    
    // Restore matter state
    mState.unpack(cell);
    // Reset (with respect to temperature, level populations and electron density)
    cell->reset(dpT|dpN|dpNe);
    
    // Update matter state at maximum temperature
    updateMatterStateT(cell,TMax,config.convergence.scanTNIter,config.convergence.scanTDampN,config.convergence.scanTNeTol,config.convergence.scanTNTol);
    // Evaluate G at maximum temperature
    GMax=G(cell,dNZero);

    // Restore matter state
    mState.unpack(cell);
    // Reset (with respect to temperature, level populations and electron density)
    cell->reset(dpT|dpN|dpNe);
    
    // If no zero found
    if ((GMin>0 and GCur>0 and GMax>0) or (GMin<0 and GCur<0 and GMax<0))  {
      dT=0; mState.unpack(cell); cell->reset(dpT|dpN|dpNe);
      if (log()) cout<<"NLTESolver::scanTemperature: cell: "<<cell->index()<<", Warning: No zero found, T: ["<<TMin<<","<<TMax<<"], G: ["<<GMin<<","<<GMax<<"]\n";
      return;
    }
  }

  if ((GCur>=0 and GMax<=0) or (GCur<=0 and GMax>=0)) { TMin=TCur; GMin=GCur; }
  else { TMax=TCur; GMax=GCur; }  

  if (log()) cout<<"NLTESolver::scanTemperature: cell: "<<cell->index()<<", T: ["<<TMin<<","<<TMax<<"], G: ["<<GMin<<","<<GMax<<"]\n";

  // Iteration loop
  //for (n=0;n<config.convergence.scanTDiv;n++) {
  for (n=0;n<scanTDiv;n++) {

    TNew=(TMin+TMax)/2;

    // Restore matter state
    // Probably better not to restore the state here ...
    //mState.unpack(cell);
    // Reset (with respect to temperature, level populations and electron density)
    cell->reset(dpT|dpN|dpNe);

    // Update matter state at new temperature
    updateMatterStateT(cell,TNew,config.convergence.scanTNIter,config.convergence.scanTDampN,config.convergence.scanTNeTol,config.convergence.scanTNTol);

    // Evaluate G at new temperature
    GNew=G(cell,dNZero);

    if ((GNew>=0 and GMax<=0) or (GNew<=0 and GMax>=0)) { TMin=TNew; GMin=GNew; }
    else { TMax=TNew; GMax=GNew; }

    if (log()) cout<<"NLTESolver::scanTemperature: cell: "<<cell->index()<<", T: ["<<TMin<<","<<TMax<<"], G: ["<<GMin<<","<<GMax<<"]\n";
  }

  // Interpolate zero
  if (GMin!=GMax) dT=(TMin-GMin/((GMax-GMin)/(TMax-TMin)))-TCur;
  else dT=(TMin+TMax)/2-TCur;
  
  // If not update
  if (not update) {
    
    // Restore matter state
    mState.unpack(cell);
    // Reset (with respect to temperature, level populations and electron density)
    cell->reset(dpT|dpN|dpNe);
  }
  // Else
  else {
    
    // Update matter state
    updateMatterStateT(cell,TCur+dT,config.convergence.scanTNIter,config.convergence.scanTDampN,config.convergence.scanTNeTol,config.convergence.scanTNTol);
  }
  }*/

// Scan for temperature
void NLTESolver::scanTemperature(Cell* cell, double& dT, bool update) {
  vector<vector<vector<double> > > dNZero;
  MStateMPI mState;
  double TCur,TMin,TMax,TNew,TMinOld=-1e99,TMaxOld=1e99,GCur,GMin,GMax,GNew,GMinOld,GMaxOld,dGdT;
  int n,i;
  int scanTDiv;

  // Initialize
  initN(cell,dNZero);

  // Get temperature
  TCur=cell->T();

  // For each scan factor
  for (i=0;i<config.convergence.scanTFac.size();i++) {

    // Calculate scan temperatures
    //TMax=TCur*config.convergence.scanTFac[i];
    TMax=MIN(TCur*config.convergence.scanTFac[i],config.convergence.TMax); // Changed!
    //TMin=TCur/config.convergence.scanTFac[i];
    TMin=MAX(TCur/config.convergence.scanTFac[i],config.convergence.TMin); // Changed!
    
    // If first scan factor
    if (i==0) {

      // Store matter state
      mState.pack(cell);  
      // Reset (with respect to all quantities)
      cell->reset(dpAll);
      // Testing
      //cell->reset(dpAll,false);

      // Update matter state at current temperature
      updateMatterStateT(cell,TCur,config.convergence.scanTNIter,config.convergence.scanTDampN,config.convergence.scanTNeTol,config.convergence.scanTNTolInit,config.convergence.scanTNTNeTolInit);
      // Evaluate G at the current temperature
      GCur=G(cell,dNZero);
    }

    // Testing !!!
    //exit(-1);
    
    // Restore matter state
    mState.unpack(cell);
    // Reset (with respect to temperature, level populations and electron density)
    cell->reset(dpT|dpN|dpNe);
    // Testing
    //cell->reset(dpT|dpN|dpNe,false);
    
    // Update matter state at minimum temperature
    updateMatterStateT(cell,TMin,config.convergence.scanTNIter,config.convergence.scanTDampN,config.convergence.scanTNeTol,config.convergence.scanTNTolInit,config.convergence.scanTNTNeTolInit);
    // Evaluate G at minimum temperature
    GMin=G(cell,dNZero);
    
    // Restore matter state
    mState.unpack(cell);
    // Reset (with respect to temperature, level populations and electron density)
    cell->reset(dpT|dpN|dpNe);
    // Testing
    //cell->reset(dpT|dpN|dpNe,false);
    
    // Update matter state at maximum temperature
    updateMatterStateT(cell,TMax,config.convergence.scanTNIter,config.convergence.scanTDampN,config.convergence.scanTNeTol,config.convergence.scanTNTolInit,config.convergence.scanTNTNeTolInit);
    // Evaluate G at maximum temperature
    GMax=G(cell,dNZero);
    
    // Restore matter state
    mState.unpack(cell);
    // Reset (with respect to temperature, level populations and electron density)
    cell->reset(dpT|dpN|dpNe);
    // Testing
    //cell->reset(dpT|dpN|dpNe,false);

    // If zero bounded by intervall - break loop
    if (not ((GMin>0 and GCur>0 and GMax>0) or (GMin<0 and GCur<0 and GMax<0))) break;
  }


  // If no zero found
  if ((GMin>0 and GCur>0 and GMax>0) or (GMin<0 and GCur<0 and GMax<0))  {
    dT=0; mState.unpack(cell); cell->reset(dpT|dpN|dpNe);
    if (log()) {
      cout<<"NLTESolver::scanTemperature: cell: "<<cell->index()<<", Warning: No zero found, T: ["<<TMin<<","<<TMax<<"], G: ["<<GMin<<","<<GMax<<"]\n";
      /*double T1=TMin;
      while (T1<TMax) {
	cell->T(T1); 
	cell->reset(dpT);
	cout<<"NLTESolver::scanTemperature: cell: "<<cell->index()<<", T: "<<T1<<", G: "<<G(cell,dNZero)<<"\n";
	T1=T1*1.1;
	}
      mState.unpack(cell);
      cell->reset(dpT|dpN|dpNe);*/
    }
    return;
  }

  // Update temperature intervall
  if ((GCur>=0 and GMax<=0) or (GCur<=0 and GMax>=0)) { TMinOld=TMin; GMinOld=GMin; TMin=TCur; GMin=GCur; }
  else { TMaxOld=TMin; GMaxOld=GMax; TMax=TCur; GMax=GCur; }  

  if (log()) cout<<"NLTESolver::scanTemperature: cell: "<<cell->index()<<", T: ["<<TMin<<","<<TMax<<"], G: ["<<GMin<<","<<GMax<<"]\n";

  // Calculate number of divisions needed to achieve the requested tolerence
  //scanTDiv=int(ceil(log10((config.convergence.scanTFac[i]-1)/config.convergence.scanTTol)/log10(2)));
  // Testing
  scanTDiv=100;

  // Iteration loop
  for (n=0;n<scanTDiv;n++) {

    // If mixed method and even iterations
    if (n%2==0 and config.convergence.scanTMethod==scanTMixed) {
      // Update evaluation temperature (using secant rule)
      dGdT=(GMax-GMin)/(TMax-TMin);
      TNew=TMin-GMin/dGdT;
    }
    // Else
    else {
      // Update evaluation temperature (using bisection rule)
      if (fabs(GMin)<fabs(GMax)) TNew=((2*config.convergence.scanTBAcc-1)*TMin+TMax)/(2*config.convergence.scanTBAcc);
      else TNew=((2*config.convergence.scanTBAcc-1)*TMax+TMin)/(2*config.convergence.scanTBAcc);
    }
 
    // Reset (with respect to temperature, level populations and electron density)
    cell->reset(dpT|dpN|dpNe);
    // Testing
    //cell->reset(dpT|dpN|dpNe,false);

    //cout<<"NLTESolver::scanTemperature: cell: "<<cell->index()<<", 1\n";
    
    // Update matter state at new temperature
    updateMatterStateT(cell,TNew,config.convergence.scanTNIter,config.convergence.scanTDampN,config.convergence.scanTNeTol,config.convergence.scanTNTol,config.convergence.scanTNTNeTol);

    //cout<<"NLTESolver::scanTemperature: cell: "<<cell->index()<<", 2\n";

    // Evaluate G at new temperature
    GNew=G(cell,dNZero);

    // Update temperature intervall
    if ((GNew>=0 and GMax<=0) or (GNew<=0 and GMax>=0)) { TMinOld=TMin; GMinOld=GMin; TMin=TNew; GMin=GNew; }
    else { TMaxOld=TMax; GMaxOld=GMax; TMax=TNew; GMax=GNew; }

    if (log()) cout<<"NLTESolver::scanTemperature: cell: "<<cell->index()<<", T: ["<<TMin<<","<<TMax<<"], G: ["<<GMin<<","<<GMax<<"]\n";

    // If requested tolerance achieved - Break loop
    if (fabs(2*(TMax-TMin)/(TMax+TMin))<config.convergence.scanTTol) break;
  }

  // Interpolate zero
  if (GMin!=GMax) dT=(TMin-GMin/((GMax-GMin)/(TMax-TMin)))-TCur;
  else dT=(TMin+TMax)/2-TCur;
  
  // If not update
  if (not update) {
    
    // Restore matter state
    mState.unpack(cell);
    // Reset (with respect to temperature, level populations and electron density)
    cell->reset(dpT|dpN|dpNe);
    // Testing
    //cell->reset(dpT|dpN|dpNe,false);
  }
  // Else
  else {
    
    // Update matter state
    updateMatterStateT(cell,TCur+dT,config.convergence.scanTNIter,config.convergence.scanTDampN,config.convergence.scanTNeTol,config.convergence.scanTNTol,config.convergence.scanTNTNeTol);
  }
}

// Update matter state at given temperature
void NLTESolver::updateMatterStateT(Cell* cell, const double& T, const int& nIter, double _dampN, double neTol, double NTol, double NTNeTol) { // Changed
  vector<vector<vector<double> > > dN;
  vector<vector<double> > XIState;
  vector<vector<vector<double> > > XBState;
  double ne,dampNOld;
  int i;
  // Testing
  int j,k,l,m;
  double ROSP,ROSPAtom,RD,RDAtom,RU,RUAtom;
  TransBF* trans1;
  TransBF* trans2;
  TransBF* transBF;
  TransBB* transBB;
  double rErrNI;

  dampNOld=dampN;
  dampN=_dampN;

  // Set temperature
  cell->T(T);
  // Reset (with respect to temperature)
  cell->reset(dpT);
  // Testing
  //cell->reset(dpT,false);

  // Testing
  // Update bound-free cut-off
  if (cell->index()==36) cout<<"Cell::updateMatterStateT: cell: "<<cell->index()<<", Update bound-free cut-off\n";
  cell->updateBoundFreeCutOffOMP(); cell->boundFreeCutOffUpdate=false;

  // For each iteration
  for (i=0;i<MAX(nIter,1);i++) {

    // Testing - If ion coupled flag (and coupled flag not set)
    if (config.ionCoupled and (not config.coupled)) {
      // Initialize
      initN(cell,dN);
      // Solve for coupled ion state
      solveIonStateCoupled(cell,dN);
      // Update ion state
      updateMatterState(cell,dN,0.0);
      // Calculate relative error in electron number density
      rErrNe=fabs(getdNe(cell,dN)/cell->ne());
      // Calculate relative error in bound state number density
      rErrN=rErrNMean(cell,dN);
      // Calculate relative error in ionization state number density
      rErrNI=rErrNIMean(cell,dN);
      //if (log()) cout<<"NLTESolver::updateMatterStateT: 1, cell: "<<cell->index()<<", iteration "<<i<<", ne: "<<cell->ne()<<", rErrNe: "<<rErrNe<<", rErrN: "<<rErrN<<", rErrNI: "<<rErrNI<<"\n";
    }

    // Initialize
    initN(cell,dN);
    initXIState(cell,XIState);
    initXBState(cell,XBState);

    // If no damping - Unclear if this still works - Removed!
    /*if (dampN==1.0) {
      // Get electron density
      ne=cell->ne();
      // Solve for bound state (using X)
      solveBoundStateX(cell,XIState,XBState);
      // Update bound state state (using X)
      updateMatterStateX(cell,XIState,XBState,0.0);
      // Calculate relative error in electron number density
      rErrNe=fabs((ne-cell->ne())/ne);
      // Ignore relative error in bound state number density
      rErrN=0;
      }*/
    // Else
    //else {
    // Solve for bound state (using dN)
    solveBoundState(cell,dN);
    // Update bound state state (using dN)
    updateMatterState(cell,dN,0.0);
    // Calculate relative error in electron number density
    rErrNe=fabs(getdNe(cell,dN)/cell->ne());
    // Calculate relative error in bound state number density
    rErrN=rErrNMean(cell,dN);
    // Calculate relative error in ionization state number density - New!
    rErrNI=rErrNIMean(cell,dN);
    //}
    //if (log()) cout<<"NLTESolver::updateMatterStateT: 2, cell: "<<cell->index()<<", iteration "<<i<<", ne: "<<cell->ne()<<", rErrNe: "<<rErrNe<<", rErrN: "<<rErrN<<", rErrNI: "<<rErrNI<<"\n";
    //if (log()) cout<<"NLTESolver::updateMatterStateT: cell: "<<cell->index()<<", iteration "<<i<<", ne: "<<cell->ne()<<", rErrNe: "<<rErrNe<<", rErrN: "<<rErrN<<", rErrNMean: "<<rErrNMean(cell,dN)<<", rErrNMax: "<<rErrNMax(cell,dN)<<"\n";
    if (cell->index()==36) cout<<"NLTESolver::updateMatterStateT: cell: "<<cell->index()<<", iteration "<<i<<", T: "<<T<<", ne: "<<cell->ne()<<", rErrNe: "<<rErrNe<<", rErrN: "<<rErrN<<", rErrNMean: "<<rErrNMean(cell,dN)<<", rErrNMax: "<<rErrNMax(cell,dN)<<"\n";
    if (fabs(rErrNe)<neTol and fabs(rErrN)<NTol) break;
    // Testing!
    /*if (cell->index()==36) {
      int j,k,l;
      for (j=0;j<cell->atom.size();j++) 
	for (k=0;k<cell->atom[j].iState.size();k++)
	  for (l=0;l<cell->atom[j].iState[k].bState.size();l++) {
	    if (!isnormal(dN[j][k][l]) or !isnormal(cell->atom[j].iState[k].bState[l].n()) or fabs(dN[j][k][l])>cell->atom[j].n())
	      cout<<"NLTESolver::updateMatterStateT: cell: "<<cell->index()<<", dN["<<cell->atom[j].iState[k].name()<<":"<<cell->atom[j].iState[k].bState[l].getIndex()<<"]: "<<dN[j][k][l]<<", N["<<cell->atom[j].iState[k].name()<<":"<<cell->atom[j].iState[k].bState[l].getIndex()<<"]: "<<cell->atom[j].iState[k].bState[l].n()<<"\n";
	    //if (cell->atom[j].iState[k].bState[l].LTERefState!=NULL) cout<<"NLTESolver::updateMatterStateT: cell: "<<cell->index()<<", bState: "<<cell->atom[j].iState[k].name()<<":"<<cell->atom[j].iState[k].bState[l].getIndex()<<", refState: "<<cell->atom[j].iState[k].bState[l].LTERefState->iState()->name()<<":"<<cell->atom[j].iState[k].bState[l].LTERefState->getIndex()<<", xLTE: "<<xLTERef(cell->atom[j].iState[k].bState[l],cell->T())<<"\n";
	  }
	  }*/
    // Update non-thermal rates - New!
    if (fabs(rErrNe)>NTNeTol) ntSolver.update(cell);
  }

  if (log()) cout<<"NLTESolver::updateMatterStateT: cell: "<<cell->index()<<", Matter state converged in "<<i+1<<" iterations, T: "<<cell->T()<<", ne: "<<cell->ne()<<", rErrNe: "<<rErrNe<<", rErrN: "<<rErrN<<"\n";	  

  dampN=dampNOld;
}

// Get change in electron number density
double NLTESolver::getdNe(Cell* cell, const vector<vector<vector<double> > >& dN) {
  double dNe=0;
  int i,j,k;

  // Calculate change in electron number density
  for (i=0;i<cell->atom.size();i++) 
    for (j=0;j<cell->atom[i].iState.size();j++)
      for (k=0;k<cell->atom[i].iState[j].bState.size();k++)
	dNe+=dN[i][j][k]*fabs(cell->atom[i].iState[j].C());

  return dNe;
}

// Update bound state for atom
void NLTESolver::updateBoundState(Atom& atom, const vector<vector<double> >& dN, const Path& path, bool damp) {
//void NLTESolver::updateBoundState(Atom& atom, const vector<vector<double> >& dN, const Path& path) {
  vector<vector<double> > n(atom.iState.size());
  double nAtom,nIState;
  double nMin=config.convergence.nMin*numeric_limits<double>::epsilon()*atom.n();
  int i,j;
  
  // For each ionization state
  for (i=0;i<atom.iState.size();i++) {
    // Resize number fraction vector
    n[i].resize(atom.iState[i].bState.size(),0);
    // For each bound state
    for (j=0;j<atom.iState[i].bState.size();j++) {
      // If not excluded
      if (not exclude(path<<i<<j)) {
	// Calculate new guess for number fraction
	if (damp) {
	  n[i][j]=MAX(atom.iState[i].bState[j].n()+dN[i][j]*dampN,nMin);  
	  //if (atom.name().compare("Ti")==0 and n[i][j]!=(atom.iState[i].bState[j].n()+dN[i][j]*dampN)) cout<<"NLTESolver::updateBoundState: cell: "<<atom.cell()->index()<<", 1, "<<atom.iState[i].name()<<":"<<j<<", bState.n: "<<atom.iState[i].bState[j].n()<<", n: "<<n[i][j]<<", dN: "<<dN[i][j]<<", dampN: "<<dampN<<", nMin: "<<nMin<<"\n";
	}
	else {
	  n[i][j]=MAX(atom.iState[i].bState[j].n()+dN[i][j],nMin);
	  //if (atom.name().compare("Ti")==0 and n[i][j]!=(atom.iState[i].bState[j].n()+dN[i][j])) cout<<"NLTESolver::updateBoundState: cell: "<<atom.cell()->index()<<", 1, "<<atom.iState[i].name()<<":"<<j<<", bState.n: "<<atom.iState[i].bState[j].n()<<", n: "<<n[i][j]<<", dN[i][j]: "<<dN[i][j]<<", nMin: "<<nMin<<"\n";
	}
      }
    }
  }

  // Sum number density
  // Unclear if muliple initializers work - Fix!
  nAtom=0; // New!
  for (i=0;i<atom.iState.size();i++) // Changed!
    for (j=0;j<atom.iState[i].bState.size();j++) nAtom+=n[i][j];

  // Testing!!!
  double dNAtom=0;
  for (i=0;i<atom.iState.size();i++)
    for (j=0;j<atom.iState[i].bState.size();j++) dNAtom+=dN[i][j];

  // For each ionization state
  for (i=0;i<atom.iState.size();i++) {
     
    // Sum number density
    // Unclear if muliple initializers work - Fix!
    nIState=0; // New!
    for (j=0;j<atom.iState[i].bState.size();j++) nIState+=n[i][j]; // Changed!
      
    //if (atom.iState[i].name().compare("TiII")==0) cout<<"NLTESolver::updateBoundState: cell: "<<atom.cell()->index()<<", TiII:0, bState.n: "<<atom.iState[i].bState[0].n()<<", iState.n: "<<atom.iState[i].n()<<", atom.n: "<<atom.n()<<", n: "<<n[i][0]<<", dN: "<<dN[i][0]<<", nIState: "<<nIState<<", nAtom: "<<nAtom<<", dNAtom: "<<dNAtom<<", excludeIState: "<<exclude(path<<i)<<", excludeBState: "<<exclude(path<<i<<j)<<"\n";

    // If not excluded - Update number fraction
    if (not exclude(path<<i)) atom.iState[i].X(nIState/nAtom);
    
    // For each bound state
    for (j=0;j<atom.iState[i].bState.size();j++) {
      // If not excluded - Update number fraction
      if (not exclude(path<<i<<j)) {

	// If non-zero ionization state number density - Update
	if (nIState!=0) atom.iState[i].bState[j].X(n[i][j]/nIState);
	// Else - Set dummy (but consistent) value
	else atom.iState[i].bState[j].X(j==0?1:0);
      }
    }

    //if (atom.iState[i].name().compare("TiII")==0) cout<<"NLTESolver::updateBoundState: cell: "<<atom.cell()->index()<<", TiII:0, bState.n: "<<atom.iState[i].bState[0].n()<<", iState.n: "<<atom.iState[i].n()<<"\n";
  }
}

// Update bound state for atom (in terms of number fractions)
void NLTESolver::updateBoundStateX(Atom& atom, const vector<double>& XIState, const vector<vector<double> >& XBState, const Path& path) {
  int i,j;

  // For each ionization state
  for (i=0;i<atom.iState.size();i++) {
    
    // If not excluded - Update number fraction
    if (not exclude(path<<i)) atom.iState[i].X(XIState[i]);
    
    // For each bound state
    for (j=0;j<atom.iState[i].bState.size();j++) {
      
      // If not excluded - Update number fraction
      if (not exclude(path<<i<<j)) atom.iState[i].bState[j].X(XBState[i][j]);
    }
  }
}

// Update bound state for cell
void NLTESolver::updateBoundState(Cell* cell, const vector<vector<vector<double> > >& dN, bool damp) {
  int i;

  // For each atom
  for (i=0;i<cell->atom.size();i++)
    // If not excluded - Update bound state
    if (not exclude(Path(i))) updateBoundState(cell->atom[i],dN[i],Path(i),damp);
}

// Update bound state for cell (in terms of number fractions)
void NLTESolver::updateBoundStateX(Cell* cell, const vector<vector<double> >& XIState, const vector<vector<vector<double> > >& XBState) {
  int i,j,k;
  vector<vector<double> > dN;

  // For each atom
  for (i=0;i<cell->atom.size();i++)
    // If not excluded
    if (not exclude(Path(i))) updateBoundStateX(cell->atom[i],XIState[i],XBState[i],Path(i));
}

// Update bound state for cell (takes care of excluded atoms and states)
void NLTESolver::updateBoundStateExcluded(Cell* cell) {
  int i,j,k;

  // For each atom
  for (i=0;i<cell->atom.size();i++) {
    // If excluded
    if (exclude(Path(i))) {
      // Force neutral state
      forceNeutralState(cell->atom[i]);
      // Force ground state for each ion
      for (j=0;j<cell->atom[i].iState.size();j++) forceGroundState(cell->atom[i].iState[j]);
    }
    // Else
    else {
      // For each ionization state
      for (j=0;j<cell->atom[i].iState.size();j++) {
	// If excluded
	if (exclude(Path(i)<<j)) { 
	  // Set number fraction to zero
	  cell->atom[i].iState[j].X(0);
	  // Force ground state
	  forceGroundState(cell->atom[i].iState[j]); 
	}
	// Else
	else {
	  // For each bound state
	  for (k=0;k<cell->atom[i].iState[j].bState.size();k++) {
	    // If excluded - Set number fraction to zero
	    if (exclude(Path(i)<<j<<k)) cell->atom[i].iState[j].bState[k].X(0);
	  }
	}
      }
    }
  }
}
  
// Update matter state for cell
void NLTESolver::updateMatterState(Cell* cell, const vector<vector<vector<double> > >& dN, const double& dT, bool damp) {
  double dTMax=config.convergence.dTMax;
  double TMin=config.convergence.TMin;
  double ne=0;
  int i,j,k;
  
  // Update bound state
  updateBoundState(cell,dN,damp);
    
  // Take care of excluded atoms and states
  updateBoundStateExcluded(cell);
	
  // Update temperature
  if (damp) cell->T(MAX(cell->T()+MAX(MIN(dT*dampT,dTMax),-dTMax),TMin));
  else cell->T(cell->T()+dT);

  // Calculate electron number density
  for (i=0;i<cell->atom.size();i++) 
    for (j=0;j<cell->atom[i].iState.size();j++) {
      ne+=cell->atom[i].iState[j].n()*fabs(cell->atom[i].iState[j].C());
    }

  // Testing!!!
  /*if (cell->index()==36) {
    int j,k,l;
    for (j=0;j<cell->atom.size();j++) 
      for (k=0;k<cell->atom[j].iState.size();k++)
	for (l=0;l<cell->atom[j].iState[k].bState.size();l++)
	  if (!isnormal(dN[j][k][l]) or !isnormal(cell->atom[j].iState[k].bState[l].n()) or fabs(dN[j][k][l])>cell->atom[j].n())
	    cout<<"NLTESolver::updateMatterState: cell: "<<cell->index()<<", dN["<<cell->atom[j].iState[k].name()<<":"<<cell->atom[j].iState[k].bState[l].getIndex()<<"]: "<<dN[j][k][l]<<", N["<<cell->atom[j].iState[k].name()<<":"<<cell->atom[j].iState[k].bState[l].getIndex()<<"]: "<<cell->atom[j].iState[k].bState[l].n()<<"\n";
	    }*/
  
  // Update electron number density
  cell->ne(ne);

  // Reset (with respect to temperature, level populations and electron density)
  cell->reset(((dT!=0)?dpT:dpNone)|dpN|dpNe);
  // Testing
  //cell->reset(((dT!=0)?dpT:dpNone)|dpN|dpNe,false);
}

// Update matter state for cell (in terms of number fractions)
void NLTESolver::updateMatterStateX(Cell* cell, const vector<vector<double> >& XIState, const vector<vector<vector<double> > >& XBState, const double& dT) {
  double ne=0;
  int i,j;
  
  // Update bound state
  updateBoundStateX(cell,XIState,XBState);
    
  // Take care of excluded atoms and states
  updateBoundStateExcluded(cell);
	
  // Update temperature
  cell->T(cell->T()+dT);

  // Calculate electron number density
  for (i=0;i<cell->atom.size();i++) 
    for (j=0;j<cell->atom[i].iState.size();j++) {
      ne+=cell->atom[i].iState[j].n()*fabs(cell->atom[i].iState[j].C());
    }

  // Update electron number density
  cell->ne(ne);

  // Reset (with respect to temperature, level populations and electron density)
  cell->reset(((dT!=0)?dpT:dpNone)|dpN|dpNe);
  // Testing
  //cell->reset(((dT!=0)?dpT:dpNone)|dpN|dpNe,false);
}

// Get maximum relative error in number density
double NLTESolver::rErrNMax(Cell* cell, const vector<vector<vector<double> > >& dN) {
  double tauLim=config.convergence.tauLim;
  double rErrN,_rErrNMax;
  double tauMax;
  int i,j,k,l;
  
  // Reset number fraction maximum relative error
  _rErrNMax=0;
  // For each bound state
  for (i=0;i<cell->atom.size();i++) {
    for (j=0;j<cell->atom[i].iState.size();j++) {
      for (k=0;k<cell->atom[i].iState[j].bState.size();k++) {
	// Get maximum optical depth for bound-bound transitions
	tauMax=bbTauMax(cell->atom[i].iState[j].bState[k]);
	// Calculate relative error in number density
	rErrN=dN[i][j][k]/cell->atom[i].iState[j].bState[k].n();
	// Calculate maximum relative error in number density
	if (fabs(rErrN)>_rErrNMax and tauMax>tauLim) _rErrNMax=fabs(rErrN);
	//if (log() and tauMax>tauLim and rErrN>=config.convergence.nTol) cout<<"NLTESolver::rErrMax: cell: "<<cell->index()<<", "<<cell->atom[i].iState[j].name()<<":"<<cell->atom[i].iState[j].bState[k].getIndex()<<", Above limit, rErrN : "<<rErrN<<"="<<dN[i][j][k]<<"/"<<cell->atom[i].iState[j].bState[k].n()<<"\n";
	// Testing!
	//if (log() and tauMax>tauLim and rErrN>=config.convergence.nTol and cell->index()==24) cout<<"NLTESolver::rErrNMax: cell: "<<cell->index()<<", "<<cell->atom[i].iState[j].name()<<":"<<cell->atom[i].iState[j].bState[k].getIndex()<<", Above limit, rErrN : "<<rErrN<<"="<<dN[i][j][k]<<"/"<<cell->atom[i].iState[j].bState[k].n()<<"\n";
      }
    }
  }

  return _rErrNMax;
}

// Get average relative error in number density
double NLTESolver::rErrNMean(Cell* cell, const vector<vector<vector<double> > >& dN) {
  double tauLim=config.convergence.tauLim;
  double rErrN,_rErrNMean;
  double tauMax;
  int i,j,k,nStates=0;
  //double dNI; // Testing!
  
  // Reset number fraction maximum relative error
  _rErrNMean=0;
  // For each bound state
  for (i=0;i<cell->atom.size();i++) {
    for (j=0;j<cell->atom[i].iState.size();j++) {
      // Testing!
      //dNI=0;
      for (k=0;k<cell->atom[i].iState[j].bState.size();k++) {
	// Testing!
	//dNI+=dN[i][j][k];
	// Get maximum optical depth for bound-bound transitions
	tauMax=bbTauMax(cell->atom[i].iState[j].bState[k]);
	// Calculate relative error in number density
	rErrN=dN[i][j][k]/cell->atom[i].iState[j].bState[k].n();
	// Calculate sum of relative error in number density
	if (tauMax>tauLim) {
	  _rErrNMean=_rErrNMean+fabs(rErrN);
	  nStates++;
	  // Testing!
	  //if (log() and cell->index()==2) cout<<"NLTESolver::rErrNMean: cell: "<<cell->index()<<", "<<cell->atom[i].iState[j].name()<<":"<<cell->atom[i].iState[j].bState[k].getIndex()<<", rErrN: "<<rErrN<<"="<<dN[i][j][k]<<"/"<<cell->atom[i].iState[j].bState[k].n()<<"\n";
	}
	// Testing!
	//if (log() and tauMax>tauLim and rErrN>=config.convergence.nTol and cell->index()==24) cout<<"NLTESolver::rErrNMean: cell: "<<cell->index()<<", "<<cell->atom[i].iState[j].name()<<":"<<cell->atom[i].iState[j].bState[k].getIndex()<<", Above limit, rErrN : "<<rErrN<<"="<<dN[i][j][k]<<"/"<<cell->atom[i].iState[j].bState[k].n()<<", n: "<<cell->atom[i].iState[j].bState[k].n()<<"\n";
      }
      // Testing!
      //if (log() and cell->index()==24 and (cell->atom[i].name().compare("Ni")==0 or cell->atom[i].name().compare("Co")==0 or cell->atom[i].name().compare("Fe")==0)) cout<<"NLTESolver::rErrNMean: cell: "<<cell->index()<<", "<<cell->atom[i].iState[j].name()<<", rErrNI : "<<dNI/cell->atom[i].iState[j].n()<<"="<<dNI<<"/"<<cell->atom[i].iState[j].n()<<", n: "<<cell->atom[i].iState[j].n()<<"\n";
    }
  }

  // Calculate mean of relative error in number density
  _rErrNMean=_rErrNMean/nStates;

  return _rErrNMean;
}

// Get average relative error in ion number density - New!
double NLTESolver::rErrNIMean(Cell* cell, const vector<vector<vector<double> > >& dN) {
  double tauLim=config.convergence.tauLim;
  double rErrNI,_rErrNIMean,dNI;
  double tauMax;
  int i,j,k,nStates=0;
 
  // Reset number fraction maximum relative error
  _rErrNIMean=0;
  // For each ionization state
  for (i=0;i<cell->atom.size();i++) {
    for (j=0;j<cell->atom[i].iState.size();j++) {
      // Reset change in ion number density
      dNI=0;
      // Sum change in ion number density
      for (k=0;k<cell->atom[i].iState[j].bState.size();k++) dNI+=dN[i][j][k];
      // Get maximum optical depth for bound-bound transitions
      tauMax=bbTauMax(cell->atom[i].iState[j]);
      // Calculate relative error in ion number density
      rErrNI=dNI/cell->atom[i].iState[j].n();
      // Calculate sum of relative error in ion number density
      if (tauMax>tauLim) {
	_rErrNIMean=_rErrNIMean+fabs(rErrNI);
	nStates++;
	// Testing!
	//if (log() and cell->index()==2) cout<<"NLTESolver::rErrNIMean: cell: "<<cell->index()<<", "<<cell->atom[i].iState[j].name()<<", rErrNI: "<<rErrNI<<"="<<dNI<<"/"<<cell->atom[i].iState[j].n()<<"\n";
      }
    }
  }
  
  // Calculate mean of relative error in number density
  _rErrNIMean=_rErrNIMean/nStates;
  
  return _rErrNIMean;
}

// Update errors
void NLTESolver::updateErrors(Cell* cell, const vector<vector<vector<double> > >& dN, const double& dT) {
  double tauLim=config.convergence.tauLim;
  ErrNMethod errNMethod=config.convergence.errNMethod;
  double dNe;
  double tauMax;
  int i,j,k,l;
  
  // Get change in electron number density
  dNe=getdNe(cell,dN);

  // Calculate relative error in temperature
  rErrT=dT/cell->T();

  // Calculate relative error in electron number density
  rErrNe=dNe/cell->ne();

  // Calculate relative error in bound state number density according to configuration
  switch(errNMethod) {
  case enMean : rErrN=rErrNMean(cell,dN); break;
  default: rErrN=rErrNMax(cell,dN);
  }

  //cout<<"NLTESolver::updateErrors: rErrMMean: "<<rErrNMean(cell,dN)<<", rErrNMax: "<<rErrNMax(cell,dN)<<"\n"; 
}

// Update damping factors
void NLTESolver::updateDamping() {
  double dampNMin=config.convergence.dampNMin;
  double dampTMin=config.convergence.dampTMin;
  double dampTMax=config.convergence.dampTMax;

  // If error in electron number density decrease slowly (and change sign)
  if (n>0 and fabs(rErrNe/rErrNePrev)>0.9 and (rErrNe/rErrNePrev)<0 and dampN>dampNMin) {
    // Decrease level population damping factor
    dampN=MAX(dampN*0.9,dampNMin);
    if (log()) cout<<"NLTESolver::updateDamping: Error in level populations decrease slowly - Reduce damping factor: "<<dampN/0.9<<" -> "<<dampN<<"\n";
  }

  // If error in temperature decrease slowly
  if (n>0 and fabs(rErrT/rErrTPrev)>0.9) {
    // If sign-change
    if ((rErrT/rErrTPrev)<0 and dampT>dampTMin) {
      // Decrease temperature damping factor
      dampT=MIN(MAX(dampT*0.9,dampTMin),1.0);
      if (log()) cout<<"NLTESolver::updateDamping: Error in temperature decrease slowly - Reduce damping factor: "<<dampT/0.9<<" -> "<<dampT<<"\n";
    }
    // If no sign-change
    if ((rErrT/rErrTPrev)>0 and dampT<dampTMax) {
      // Increase temperature damping factor
      dampT=MIN(dampT*1.1,dampTMax);
      if (log()) cout<<"NLTESolver::updateDamping: Error in temperature decrease slowly - Increase damping factor: "<<dampT/1.1<<" -> "<<dampT<<"\n";
    }
  }
}

// Update convergence criteria
void NLTESolver::updateConvergence(Cell* cell, const vector<vector<vector<double> > >& dN, const double& dT) {
  int i,j;

  // Update errors
  updateErrors(cell,dN,dT);

  // Update damping
  updateDamping();

  // Store old errors
  rErrNPrev=rErrN;
  rErrNePrev=rErrNe;
  rErrTPrev=rErrT;
}

// Solution converged
bool NLTESolver::converged() {

  // If error in level populations, temperature and electron number density less than tolerance return true
  if (fabs(rErrN)<config.convergence.nTol and fabs(rErrT)<config.convergence.TTol and fabs(rErrNe)<config.convergence.neTol and n>=config.convergence.nIterMin) return true;
    
  return false;
}

// Log level inversions
void NLTESolver::logInversion(Cell* cell) {
  BState* bState1;
  BState* bState2;
  int i,j,k;

  // For each bound state
  for (i=0;i<cell->atom.size();i++) {
    for (j=0;j<cell->atom[i].iState.size();j++) {
      for (k=1;k<cell->atom[i].iState[j].bState.size();k++) {

	// Set bound state pointers
	bState1=&(cell->atom[i].iState[j].bState[k-1]);
	bState2=&(cell->atom[i].iState[j].bState[k]);

	// If population inverted - Log
	if ((bState2->n()/bState2->g())>(bState1->n()/bState1->g()) and log()) cout<<"NLTESolver::logLevelInversion: cell: "<<cell->index()<<", Warning: Level "<<bState2->iState()->name()<<":"<<bState2->getIndex()<<" inverted\n";
      }
    }
  }
}

// Update matter state
void NLTESolver::update(Cell* cell, bool initiate, bool initiateState) {
  int nIter=config.convergence.nIter;
  int i,j,k;
  double dT;
  double rErrNeIon; // Testing!
  vector<vector<vector<double> > > dN;
  
  // If requested initiate
  //if (initiate) init(cell);
  if (initiate) init(cell,initiateState);

  // Reset
  reset(cell,dN);

  // If not steady state and new timestep - Set old state to last state
  if ((not config.steadyState) and cell->t()!=tLast) { tOld=tLast; TOld=TLast; nOld=nLast; }

  // Set bound-free OSP table only - Testing!
  cell->config->boundFreeTableOSPOnly=true;

  // If initial scan requested
  if (config.convergence.scanT) {
    // Update non-thermal rates - New!
    ntSolver.update(cell);
    // Scan for temparture and update matter state
    scanTemperature(cell,dT,true);
    if (log()) cout<<"CombSolver::update: cell: "<<cell->index()<<", Matter state updated - Initial scan, ne: "<<cell->ne()<<", T: "<<cell->T()<<"\n";
  }
  
  // Update non-thermal rates - New!
  ntSolver.update(cell);

  // Iteration loop
  for (n=0;n<nIter;n++) {

    // Testing - If ion coupled flag (and coupled flag not set)
    if (config.ionCoupled and (not config.coupled)) {
      // Solve for coupled ion state
      solveIonStateCoupled(cell,dN);
      // Update ion state
      updateMatterState(cell,dN,0.0);
      // Calculate relative error in electron number density
      rErrNeIon=fabs(getdNe(cell,dN)/cell->ne());
      if (log()) cout<<"NLTESolver::updateMatterStateT: cell: "<<cell->index()<<", iteration "<<n<<", ne: "<<cell->ne()<<", rErrNe: "<<rErrNeIon<<"\n";
    }

    // Solve for bound state
    solveBoundState(cell,dN);

    // If temperature iteration requested - Solve for temperature
    if (config.convergence.iterT) solveTemperature(cell,dN,dT);
    // Else - No change in temperature
    else dT=0;
    
    // Update convergence criteria
    updateConvergence(cell,dN,dT);

    // Update matter state
    updateMatterState(cell,dN,dT);

    // Reset (with respect to temperature, level populations and electron density)
    cell->reset(((dT!=0)?dpT:dpNone)|dpN|dpNe);
    // Testing
    //cell->reset(((dT!=0)?dpT:dpNone)|dpN|dpNe,false);

    // If converged break
    if (converged()) break;

    if (log()) cout<<"NLTESolver::update: cell: "<<cell->index()<<", Matter state updated - Iteration: "<<n<<", ne: "<<cell->ne()<<", T: "<<cell->T()<<", rErrNe: "<<rErrNe<<", rErrT: "<<rErrT<<", rErrN: "<<rErrN<<"\n";

    // Update non-thermal rates
    if (fabs(rErrNe)>config.convergence.NTNeTol) ntSolver.update(cell);
  }
  if (n==nIter and log()) cout<<"NLTESolver::update: cell: "<<cell->index()<<", Matter state did not converge\n";
  else if (log()) cout<<"NLTESolver::update: cell: "<<cell->index()<<", Matter state converged in "<<n+1<<" iterations\n";

  // Unset bound-free OSP table only - Testing!
  cell->config->boundFreeTableOSPOnly=false;

  // If not steady state - Backup state
  if (not config.steadyState) backupState(cell);

  // Log level inversions
  //logInversion(cell);
}

// Initiate
void NLTESolver::init(Cell* cell, bool initiateState) {
  Processes processes;
  int i,j,k;

  //if (log()) cout<<"NLTESolver::init: cell: "<<cell->index()<<"\n";

  // If requested - Initiate matter state using LTE solver
  if (initiateState) lteSolver.update(cell);

  // Initiate non-thermal state
  ntSolver.update(cell,true);

  // Resize exclude vectors
  excludeAtom.resize(cell->atom.size(),false); excludeIState.resize(cell->atom.size()); excludeBState.resize(cell->atom.size());
  // For each atom
  for (i=0;i<cell->atom.size();i++) {

    // Get processes from configuration
    processes=config.population.processes(cell->atom[i]);

    // Exclude if no transitions or zero mass fraction
    if (nTrans(cell->atom[i],processes)==0 or cell->atom[i].X()==0) excludeAtom[i]=true;

    // Resize exclude vectors
    excludeIState[i].resize(cell->atom[i].iState.size(),false); excludeBState[i].resize(cell->atom[i].iState.size());
    // For each ionization state
    for (j=0;j<cell->atom[i].iState.size();j++) {

      // Exclude if no transitions or atom excluded
      if (nTrans(cell->atom[i].iState[j],processes)==0 or excludeAtom[i]) excludeIState[i][j]=true;

      // Resize exclude vectors
      excludeBState[i][j].resize(cell->atom[i].iState[j].bState.size(),false);
      // For each bound state
      for (k=0;k<cell->atom[i].iState[j].bState.size();k++) {

	// Exclude if no transitions or atom or ionization state excluded
	if (nTrans(cell->atom[i].iState[j].bState[k],processes)==0 or excludeAtom[i] or excludeIState[i][j]) excludeBState[i][j][k]=true;
      }
    }
  }

  // If not steady state 
  if (not config.steadyState) { 
    
    // Backup state
    backupState(cell);
    
    // Set old state to last state to old cell
    tOld=tLast; TOld=tLast; nOld=nLast;
  }
}

// Reset
void NLTESolver::reset(Cell* cell, vector<vector<vector<double> > >& dN) {
  int i,j,k;

  // Reset damping factors
  dampT=config.convergence.dampT;
  dampN=config.convergence.dampN;

  // Resize solution vector
  dN.resize(cell->atom.size());
  // For each atom
  for (i=0;i<cell->atom.size();i++) {

    // Resize solution vector
    dN[i].resize(cell->atom[i].iState.size());
    // For each ionization state
    for (j=0;j<cell->atom[i].iState.size();j++) {

      // Resize and reset solution vector
      dN[i][j].resize(cell->atom[i].iState[j].bState.size(),0);
    }
  }
}

// Update matter state assuming LTE
// Note: We need to force lteSolver.config.temperature.typeBound/Ion=lteGas, right?
void NLTESolver::updateLTE(Cell* cell) {

  // Update partition function
  lteSolver.updateU(cell);
  // Update ionization state
  lteSolver.updateIonState(cell);
  // Update bound state
  lteSolver.updateBoundState(cell);
}

// Load data
void NLTESolver::load(const string& dataFile) {
  // Load non-thermal solver data
  ntSolver.load(dataFile);
}

// Get logging flag
bool NLTESolver::log() const {

  // Return logging flag
  return pLog;
}

// Set logging flag
void NLTESolver::log(const bool& _log) {

  // Set logging flag
  pLog=_log;
}

// Backup state - New!
void NLTESolver::backupState(Cell* cell) {
  int i,j,k;

  // Backup time
  tLast=cell->t();

  // Backup temperature
  TLast=cell->T();

  // Resize number density vector
  nLast.resize(cell->atom.size());

  // For each atom
  for (i=0;i<cell->atom.size();i++) {

    // Resize number density vector
    nLast[i].resize(cell->atom[i].iState.size());
  
    // For each ionization state
    for (j=0;j<cell->atom[i].iState.size();j++) {

      // Resize number density vector
      nLast[i][j].resize(cell->atom[i].iState[j].bState.size());

      // For each bound state
      for (k=0;k<cell->atom[i].iState[j].bState.size();k++) {

	// Backup number denisty
	nLast[i][j][k]=cell->atom[i].iState[j].bState[k].n();
      }
    }
  }
}

// Copy cell - New!
// Workaround for memory leak in grid copy
/*void NLTESolver::copyCell(Cell* cell1, Cell* cell2) {
  MStateMPI mState;
  
  // Copy matter state
  mState.pack(cell1);
  mState.unpack(cell2);
  // Reset
  cell2->reset();
  // Copy border fluxes
  ((CellSS*) cell2)->FBack(((CellSS*) cell1)->FBack());
  ((CellSS*) cell2)->FFront(((CellSS*) cell1)->FFront());
  // Copy spatial borders
  ((CellSS*) cell2)->rMin(((CellSS*) cell1)->rMin());
  ((CellSS*) cell2)->rMax(((CellSS*) cell1)->rMax());
  // Copy time borders
  cell2->tMin(cell1->tMin());
  cell2->tMax(cell1->tMax());
  // Copy DBB field
  cell2->TR(cell1->TR());
  cell2->W(cell1->W());
  // Copy TJ
  cell2->TJ(cell1->TJ());
  // Copy radioactive energy deposition - New!
  cell2->eRadio(cell1->eRadio());
  }*/

// Constructor
CombPopConfigEntry::CombPopConfigEntry() : method(sNLTE), excitationMethod(sNLTE), ionizationMethod(sNLTE) {}

// Load data
void CombPopConfigEntry::load(const YAML::Node& data) {

  // Load base class data
  AtomConfigEntry::load(data);

  // If provided
  if (data["method"]) {

    // Load solver method
    if (data["method"].Type()==YAML::NodeType::Scalar) {
      method=loadSolverMethod(data["method"]);
      excitationMethod=method;
      ionizationMethod=method;
    }
    else {
      // Use combined solver method
      method=sComb;

      // Load excitation solver method
      if (data["method"]["excitation"]) excitationMethod=loadSolverMethod(data["method"]["excitation"]);

      // Load ionization solver method
      if (data["method"]["ionization"]) ionizationMethod=loadSolverMethod(data["method"]["ionization"]);
    }
  }
}

// Log data
void CombPopConfigEntry::log() {
  cout<<"PopConfigEntry::log:";
  AtomConfigEntry::log();
  cout<<", method: "<<method<<", excitationMethod: "<<excitationMethod<<", ionizationMethod: "<<ionizationMethod<<"\n";
}

// Get method
SolverMethod CombPopConfig::method (Atom& atom) {
  CombPopConfigEntry entry=getEntry(atom);
      
  return entry.method;
}

// Get excitation method
SolverMethod CombPopConfig::excitationMethod (Atom& atom) {
  CombPopConfigEntry entry=getEntry(atom);
      
  return entry.excitationMethod;
}

// Get excitation method
SolverMethod CombPopConfig::excitationMethod (IState& iState) {
  CombPopConfigEntry entry=getEntry(iState);
      
  return entry.excitationMethod;
}

// Get method
SolverMethod CombPopConfig::ionizationMethod (Atom& atom) {
  CombPopConfigEntry entry=getEntry(atom);
      
  return entry.ionizationMethod;
}

// Constructor
CombTempConfig::CombTempConfig() : method(sNLTE) {}

// Load data
void CombTempConfig::load(const YAML::Node& data) {

  // Load solver method
  if (data["method"]) method=loadSolverMethod(data["method"]);
}

// Log data
void CombTempConfig::log() {
  cout<<"TempConfig: method: "<<method<<"\n";
}

// Constructors
CombSolverConfig::CombSolverConfig() : population(), temperature() {};
CombSolverConfig::CombSolverConfig(const CombPopConfig& _population, const CombTempConfig& _temperature) : population(_population), temperature(_temperature) {};
CombSolverConfig::CombSolverConfig(const string& dataFile) { load(dataFile); }

// Load data
void CombSolverConfig::load(const YAML::Node& data) {

  // Load population config
  if (data["population"]) population.load(data["population"]);

  // Load temperature config
  if (data["temperature"]) temperature.load(data["temperature"]);

  log();
}

// Log data
void CombSolverConfig::log() {
  population.log();
  temperature.log();
}

// Load data (from file)
void CombSolverConfig::load(const string& dataFile) {
  YAML::Node data;

  // Load data from file
  data=YAML::LoadFile(dataFile);

  // Load data
  if (data["combSolver"]) load(data["combSolver"]);
}

// Constructors
CombSolver::CombSolver(NLTESolverConfig _config, CombSolverConfig _combConfig, NebSolverConfig _nebConfig, LTESolverConfig _lteSolverConfig) : combConfig(_combConfig), nebSolver(_lteSolverConfig,_nebConfig), NLTESolver(_config) { }
CombSolver::CombSolver(const string& dataFile) : nebSolver(dataFile), NLTESolver(dataFile) { combConfig.load(dataFile); }

// Copy constructor - New!
CombSolver::CombSolver(const CombSolver& combSolver) : nebSolver(combSolver.nebSolver), combConfig(combSolver.combConfig), NLTESolver(combSolver) { 
  //cout<<"CombSolver::CombSolver: this: "<<this<<"\n";
}

// Create copy
MStateSolver* CombSolver::copy() {
  //cout<<"CombSolver::copy, this: "<<this<<"\n";
  return new CombSolver(*this);
}

// Solve for bound state of ion (overloaded)
void CombSolver::solveBoundState(IState& iState, vector<double>& X, const Processes& processes, const Path& path) {
  Cell* cell=iState.atom()->cell();

  // Use NLTE solver depending on configuration
  if (combConfig.population.excitationMethod(iState)==sNLTE) NLTESolver::solveBoundState(iState,X,processes,path);
  // Otherwise use nebular solver
  else nebSolver.solveBoundState(iState,cell->T(),cell->TR(),cell->W(),X);

  return;
}

// Solve for ionization state of atom (overloaded) - New!
void CombSolver::solveIonState(Atom& atom, vector<double>& dN, const Processes& processes, const Path& path) {
  Cell* cell=atom.cell();
  vector<double> X(dN.size());
  int i;

  // Use NLTE solver depending on configuration
  if (combConfig.population.ionizationMethod(atom)==sNLTE) NLTESolver::solveIonState(atom,dN,processes,path);
  // Otherwise use nebular solver
  else {
    // Update partition function - Hack !!!
    nebSolver.updateU(atom,cell->T());
    // Update ionization state
    nebSolver.solveIonState(atom,cell->ne(),cell->T(),cell->TR(),cell->W(),X);
    // Convert to number density
    for (i=0;i<atom.iState.size();i++) dN[i]=X[i]*atom.n()-atom.iState[i].n();
  }

  return;
}

// Solve for ionization state of atom (in terms of number fractions, overloaded) - Renamed (from solveIonState)!
void CombSolver::solveIonStateX(Atom& atom, vector<double>& X, const Processes& processes, const Path& path) {
  Cell* cell=atom.cell();

  // Use NLTE solver depending on configuration
  if (combConfig.population.ionizationMethod(atom)==sNLTE) NLTESolver::solveIonStateX(atom,X,processes,path);
  // Otherwise use nebular solver
  else {
    // Update partition function - Hack !!!
    nebSolver.updateU(atom,cell->T());
    // Update ionization state
    nebSolver.solveIonState(atom,cell->ne(),cell->T(),cell->TR(),cell->W(),X);
  }

  return;
}

// Solve for bound state of atom (overloaded)
void CombSolver::solveBoundState(Atom& atom, vector<vector<double> >& dN, const Processes& processes, const Path& path) {

  // Use NLTE solver depending on configuration
  if (combConfig.population.method(atom)==sNLTE) NLTESolver::solveBoundState(atom,dN,processes,path);
  // Otherwise use the decoupled solver (which will call the nebular solver depending on configuration)
  else solveBoundStateDecoupled(atom,dN,processes,path);

  return;
}

// Solve for bound state of atom (in terms of number fractions, overloaded)
void CombSolver::solveBoundStateX(Atom& atom, vector<double>& XIState, vector<vector<double> >& XBState, const Processes& processes, const Path& path) {

  // Use NLTE solver depending on configuration
  if (combConfig.population.method(atom)==sNLTE) NLTESolver::solveBoundStateX(atom,XIState,XBState,processes,path);
  // Otherwise use the decoupled solver (which will call the nebular solver depending on configuration)
  else solveBoundStateDecoupledX(atom,XIState,XBState,processes,path);

  return;
}

// Solve for bound state of cell (overloaded) 
void CombSolver::solveBoundState(Cell* cell, vector<vector<vector<double> > >& dN) {
  NLTESolver::solveBoundState(cell,dN);
}

// Solve for bound state of cell (in terms of number fractions, overloaded) 
void CombSolver::solveBoundStateX(Cell* cell, vector<vector<double> >& XIState, vector<vector<vector<double> > >& XBState) {
  NLTESolver::solveBoundStateX(cell,XIState,XBState);
}

// Update matter state (overloaded)
void CombSolver::update(Cell* cell, bool initiate, bool initiateState) {
  int nIter=config.convergence.nIter;
  double dT;
  double rErrNeIon; // Testing!
  vector<vector<vector<double> > > dN;
  int i;
  
  // If requested initiate
  if (initiate) init(cell,initiateState);

  // Reset
  reset(cell,dN);

  // If not steady state and new timestep - Set old state to last state
  if ((not config.steadyState) and cell->t()!=tLast) { tOld=tLast; TOld=TLast; nOld=nLast; }

  // Set bound-free OSP table only - Testing!
  cell->config->boundFreeTableOSPOnly=true;

  // If temperature method is NLTE and initial scan requested
  if (combConfig.temperature.method==sNLTE and config.convergence.scanT) {
    // Testing!!!
    cout<<"CombSolver::update: cell: "<<cell->index()<<", 1\n";
    // Update non-thermal rates - New!
    ntSolver.update(cell);
    // Testing!!!
    cout<<"CombSolver::update: cell: "<<cell->index()<<", 2\n";
    // Scan for temperature and update
    scanTemperature(cell,dT,true);
    // Testing!!!
    cout<<"CombSolver::update: cell: "<<cell->index()<<", 3\n";
    if (log()) cout<<"CombSolver::update: cell: "<<cell->index()<<", Matter state updated - Initial scan, ne: "<<cell->ne()<<", T: "<<cell->T()<<"\n";
  }

  // Update non-thermal rates - New!
  ntSolver.update(cell);
  
  // Iteration loop
  for (n=0;n<nIter;n++) {

    // Testing - If ion coupled flag (and coupled flag not set)
    if (config.ionCoupled and (not config.coupled)) {
      // Solve for coupled ion state
      solveIonStateCoupled(cell,dN);
      // Update ion state
      updateMatterState(cell,dN,0.0);
      // Calculate relative error in electron number density
      rErrNeIon=fabs(getdNe(cell,dN)/cell->ne());
      if (log()) cout<<"CombSolver::update: cell: "<<cell->index()<<", iteration "<<n<<", ne: "<<cell->ne()<<", rErrNe: "<<rErrNeIon<<"\n";
    }

    // Solve for bound state
    solveBoundState(cell,dN);
    
    // If temperature method is NLTE and iteration requested - Solve for temperature
    if (combConfig.temperature.method==sNLTE and config.convergence.iterT) solveTemperature(cell,dN,dT);
    // Else - No change in temperature
    else dT=0;

    // Update convergence criteria
    updateConvergence(cell,dN,dT);

    // If converged break
    if (converged()) break;

    if (log()) cout<<"CombSolver::update: cell: "<<cell->index()<<", Iteration: "<<n<<", rErrNe: "<<rErrNe<<", rErrT: "<<rErrT<<", rErrN: "<<rErrN<<", rErrNMean: "<<rErrNMean(cell,dN)<<", rErrNMax: "<<rErrNMax(cell,dN)<<"\n";

    // Update matter state
    updateMatterState(cell,dN,dT);

    if (log()) cout<<"CombSolver::update: cell: "<<cell->index()<<", Matter state updated - Iteration: "<<n<<", ne: "<<cell->ne()<<", T: "<<cell->T()<<"\n";
    
    // Update non-thermal rates - New!
    if (fabs(rErrNe)>config.convergence.NTNeTol) ntSolver.update(cell);

    // If converged break
    //if (converged()) break;
  }
  if (n==nIter and log()) cout<<"CombSolver::update: cell: "<<cell->index()<<", Matter state did not converge\n";
  else if (log()) cout<<"CombSolver::update: cell: "<<cell->index()<<", Matter state converged in "<<n+1<<" iterations\n";

  // Unset bound-free OSP table only - Testing!
  cell->config->boundFreeTableOSPOnly=false;

  // If not steady state - Backup state
  if (not config.steadyState) backupState(cell);

  // Update non-thermal rates - New!
  //ntSolver.update(cell);

  // Log level inversions
  //logInversion(cell);

  //if (log()) cout<<"CombSolver::update: cell: "<<cell->index()<<", Return\n";

  // Testing!
  //logGamma(cell);
}

// Initiate (overloaded)
void CombSolver::init(Cell* cell, bool initiateState) {
  Processes processes;
  SolverMethod method,ionizationMethod,excitationMethod;
  int i,j,k;

  // Initiate NLTE solver
  NLTESolver::init(cell,initiateState);

  // Re-initiate with nebular solver ??
  
  // For each atom
  for (i=0;i<cell->atom.size();i++) {

    // Get solver methods and feedback flag from configuration
    method=combConfig.population.method(cell->atom[i]);
    ionizationMethod=combConfig.population.ionizationMethod(cell->atom[i]);
    excitationMethod=combConfig.population.excitationMethod(cell->atom[i]);

    // Do not exclude atom if solver method differs from NLTE and mass fraction is non-zero
    if (method!=sNLTE and cell->atom[i].X()!=0) excludeAtom[i]=false;

    // For each ionization state
    for (j=0;j<cell->atom[i].iState.size();j++) {

      // Do not exclude ionization state if solver method differes from NLTE
      if (ionizationMethod!=sNLTE) excludeIState[i][j]=false;

      // For each bound state
      for (k=0;k<cell->atom[i].iState[j].bState.size();k++) {
	
	// Do not exclude bound state if solver method differs from NLTE
	if (excitationMethod!=sNLTE) excludeBState[i][j][k]=false;
      }
    }
  }
}
